#ifndef CXXOMFORT_ALGORITHM_HPP
#define CXXOMFORT_ALGORITHM_HPP
/**
 * @file
 */

#include "base.hpp"
#include <algorithm>
#include <functional>

// 
// c++03 
// 

/* 
 * Algorithms for compilers missing C++98 support 
 * */

#if (0 && CXXOMFORT_CXX_STD < 1997)
namespace cxxomfort { namespace cxxostd {
	template <typename T>
	T const& min (T const& a, T const& b) CXXO_NOEXCEPTNOTHROW {
		return a < b ? a : b;
	}

	template <typename T>
	T const& max (T const& a, T const& b) CXXO_NOEXCEPTNOTHROW {
		return a > b ? a : b;
	}

} }

	#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC)
namespace std {
	using cxxomfort::cxxostd::min;
	using cxxomfort::cxxostd::max;
}
	#endif

#endif

namespace cxxomfort { namespace cxxostd {

/**
 * @ingroup std0algorithm
 * @param tv The value to use to fill.
 * @cxx03b Fills a sequence with a given value
 * @xrefitem std0algorithm "" "" Fills a sequence with a given value
 * @see @cppref{algorithm/fill}
 * 
 * @note This algorithm is provided here only for enhancements in later Standards or for compilers that lack it, and is otherwise not directly used.
 * 
 **/
template <typename It, typename T>
CXXO_CONSTEXPR14 
void fill (It ini, It fin, T const& tv) {
	for (; ini != fin; ++ini) { *ini= tv; }
}

/**
 * @param n How many elements to fill.
 * @param tv The value to use to fill.
 * @ingroup std0algorithm
 * @cxx03b Fills @a n elements of a sequence with a given value
 * @xrefitem std0algorithm "" "" Fills @a n elements of a sequence with a given value
 * @see @cppref{algorithm/fill}
 * */
template <typename It, typename Numeric, typename T>
CXXO_CONSTEXPR14 
void fill_n (It ini, Numeric n, T const& tv) {
	for (Numeric i= 0; i<n; ++i, ++ini) { *ini= tv; }
}

} } 

// 
// c++11 algorithms
// 

#if defined(DOXYGEN_DOC)
namespace std {
/**
 * @brief Copy elements conditionally to @a dest if they fulfill @a f .
 * @ingroup std0algorithm
 * @param f A predicate that will evaluate elements in the sequence.
 * @return the advanced @a dest iterator.
 * @cxx1b Copy elements from one sequence to another
 * @see @cppref{algorithm/copy_if}
 * @see @cppref{algorithm/copy}
 * 
 * From the sequence <tt>[ini, fin)</tt>, copies elements @e e to @a dest for which <tt>f(e)</tt> is @c true .
 * 
 */
template <typename InpI, typename OutI, typename Predicate>
OutI copy_if (InpI ini, InpI fin, OutI dest, Predicate f)  {}

/**
 * @brief Copy a given amount of elements to @a dest .
 * @ingroup std0algorithm
 * @return the iterated @a dest iterator.
 * @cxx1b Copy "n" elements from one sequence to another
 * @see @cppref{algorithm/copy_n}
 * @see @cppref{algorithm/copy}
 * 
 * Copies @a n elements from @a ini to @dest , unconditionally.
 * 
 **/
template <typename InpI, typename Integral, typename OutI>
inline OutI copy_n (InpI ini, Integral n, OutI dest) {}

/**
 * @brief Copies elements from a sequence to @a Dtrue or @a Dfalse conditionally.
 * @param f A predicate that maps <tt>InpI::value_type -> bool</tt>.
 * @return the advanced <code>std::pair</code> (@a Dtrue , @a Dfalse ).
 * @ingroup std0algorithm
 * 
 * @xrefitem std0algorithm "Data Transfer" "" Conditionally copy data to either of two destinations.
 * @cxx1b Conditionally copy data to either of two destinations.
 * @see @cppref{algorithm/partition_copy}
 * @see @cppref{algorithm/copy_if}
 * 
 * This algorithm is a variation of @c copy_if() . Where copy_if copies to one destination @e dest if the condition is @c true , @c partition_copy() copies elements @a e in <tt>[ini,fin)</tt> to either @a Dtrue or @a Dfalse depending on the return value of <tt>f(e)</tt>.
 * 
 **/
template <typename InpI, typename OutI1, typename OutI2, typename Predicate>
std::pair<OutI1,OutI2> 
partition_copy (InpI ini, InpI const fin, OutI1 Dtrue, OutI2 Dfalse, Predicate f) { }

} // std

#endif

namespace cxxomfort { namespace cxxostd {

template <typename InpI, typename OutI, typename Predicate>
OutI copy_if (InpI ini, InpI fin, OutI dest, Predicate f) {
    if (ini != fin) {
        for (; ini != fin; ++ini) {
            if (f(*ini)) { *dest++= *ini; }
        } // for
        
    }
    return dest;
}

template <typename InpI, typename Integral, typename OutI>
inline OutI copy_n (InpI ini, Integral n, OutI dest) {
    for (; n --> 0; ++ini, ++dest) {
        *dest= *ini;
    }
    return dest;
}

template <typename InpI, typename OutI1, typename OutI2, typename Predicate>
std::pair<OutI1,OutI2> 
partition_copy (InpI ini, InpI const fin, OutI1 Dtrue, OutI2 Dfalse, Predicate f) {
    for (; ini != fin; ++ini) {
        if (f(*ini)) *Dtrue++= *ini; 
        else *Dfalse++= *ini;
    } // end for
    return std::pair<OutI1,OutI2>(Dtrue,Dfalse);
}


/**
 * @fn std::find_if
 * @brief Finds an element in a sequence that matches a predicate.
 * @param f A predicate that maps <tt>It::value_type -> bool</tt>.
 * @ingroup std0algorithm
 * @cxx11b Find an element matching a predicate
 * @sa @cppref{algorithm/find_if}
 * 
 * @note This algorithm is already provided in C++03. It is provided here only
 * for enhancements in later Standards that might be desirable to invoke 
 * backports for explicitly.
 * 
 * @note This algorithm is <tt>constexpr</tt> starting C++20; here it is 
 * tentatively marked <tt>constexpr</tt> for C++14 as well. We can't modify 
 * the C++17 version.
 * 
 **/
template<class It, class Pred>
CXXO_CONSTEXPR14
It find_if (It ini, It fin, Pred f) {
    for (; ini != fin; ++ini) {
        if (f(*ini)) {
            return ini;
        }
    }
    return fin; // < none of the elements fulfill f
}

/**
 * @fn std::find_if_not 
 * @brief Finds like @c find_if but the predicate is reversed.
 * @param f A predicate that maps <tt>It::value_type -> bool</tt>.
 * @ingroup std0algorithm
 * @cxx11b Find an element @e not matching a predicate
 * @sa @cppref{algorithm/find_if}
 * 
 * @note This algorithm is <tt>constexpr</tt> starting C++20; here it is 
 * tentatively marked <tt>constexpr</tt> for C++14 as well. We can't modify 
 * the C++17 version.
 * 
 **/
template<class It, class Pred>
CXXO_CONSTEXPR14
It find_if_not(It ini, It fin, Pred f) {
    for (; ini != fin; ++ini) {
        if (!f(*ini)) {
            return ini;
        }
    }
    return fin; // < all the elements fulfill f
}

/**
 * @brief Counts the number of times in a sequence the criteria @a p is matched.
 * @param f A predicate that maps <tt>It::value_type -> bool</tt>.
 * @cxx17b Count the number of elements matching a predicate in a sequence
 * @xrefitem std0algorithm "" "Search" Count the number of elements matching a predicate in a sequence
 * @sa @cppref{algorithm/count}
 * @sa @ref count
 * 
 * @note This algorithm is <tt>constexpr</tt> starting C++20; here it is 
 * tentatively marked <tt>constexpr</tt> for C++14 as well. We can't modify 
 * the C++17 version.
 * 
 **/
template<class It, class Pred>
CXXO_CONSTEXPR14 typename std::iterator_traits<It>::difference_type 
count_if (It ini, It fin, Pred p) {
    typename std::iterator_traits<It>::difference_type ret= 0;
    for (; ini != fin; ++ini) {
        if (p(*ini)) {
            ++ret;
        }
    }
    return ret; 
}

/**
 * @brief Counts the number of times in a sequence a given value is found.
 * @cxx17b Count the number of elements matching a given value in a sequence
 * @xrefitem std0algorithm "" "Search" Count the number of elements matching a given value in a sequence
 * @see @cppref{algorithm/count}
 * @see @ref count_if
 * 
 * @note This algorithm is <tt>constexpr</tt> starting C++20; here it is 
 * tentatively markes <tt>constexpr</tt> for C++14 as well. We can't modify 
 * the C++17 version.
 * 
 **/
template<class It, typename T>
CXXO_CONSTEXPR14 typename std::iterator_traits<It>::difference_type 
count (It ini, It fin, T const& tv) {
    typename std::iterator_traits<It>::difference_type ret= 0;
    for (; ini != fin; ++ini) {
        if (*ini == tv) {
            ++ret;
        }
    }
    return ret; 
}


// detail helpers for shuffle
//! @internal
namespace detail_algorithm {

    template <typename URNG, typename E>
    E uniform_in (URNG& u, E m1, E m2) {
        //typedef typename std::result_of<URNG>::type T;
        E output = m1 + (u() % (E)(m2 - m1 + 1));
        return output;
    }

} // detail_algorithm


/**
 * @brief Shuffles a sequence according to a generator call g().
 * @param g A @e writable object that acts like a (P)RNG and returns a value via <code>operator()</code>.
 * @xrefitem std0algorithm "" "" Randomly shuffles elements in a sequence
 * @cxx11b Randomly shuffle elements in a sequence
 * @ingroup std0algorithm
 * 
 * @note Requires support from <random>.
 * 
 * */
template< class RAIterator, class URNG >
void shuffle( RAIterator ini, RAIterator fin, URNG& g ) {
    typedef typename std::iterator_traits<RAIterator>::difference_type diff_t;
    diff_t n = fin - ini; // guaranteed for RAIterator
    for (diff_t i = n-1; i > 0; --i) {
        using std::swap;
        //swap(ini[i], ini[D(g, param_t(0, i))]);
        swap (ini[i], ini[ detail_algorithm::uniform_in(g, 0, i) ]);
    }

}

} } //~namespace cxxomfort::cxxostd


// 
// sorting algorithms (C++11)
// 

#if defined(DOXYGEN_DOC)
namespace std {
/**
 * @param less Criteria for sorting, a binary function f(x,y) that returns @c true if <code>x<y</code>.
 * @return A bool response about if the sequence is sorted.
 * @xrefitem std0algorithm "" "" Determine if a sequence is sorted.
 * @cxx11b Determine if a sequence is sorted.
 * @sa @cppref{algorithm/is_sorted}
 */
template <typename Iterator, typename Compare>
inline bool is_sorted (Iterator ini, Iterator fin, Compare less) {}
//! @overload is_sorted
template <typename Iterator, typename Compare>
inline bool is_sorted (Iterator ini, Iterator fin) {}


/**
 * @param less Criteria for sorting, a binary function f(x,y) that returns @c true if <code>x<y</code>.
 * @return The position of the last element within the sequence that is in sorted order.
 * @xrefitem std0algorithm "" "" Determine up to which element a sequence is sorted.
 * @cxx11b Determine up to which element a sequence is sorted.
 * @sa @cppref{algorithm/is_sorted}
 */
template <class FIt, class Compare>
FIt is_sorted_until (FIt first, FIt last, Compare less)  { }

//! @overload is_sorted_until
template <typename Iterator>
inline bool is_sorted_until (Iterator ini, Iterator fin) { }

} // std

#endif

namespace cxxomfort { namespace cxxostd {

template <typename Iterator, typename Compare>
inline bool is_sorted (Iterator ini, Iterator fin, Compare less) {
    if (ini == fin) return true;
    Iterator prev= ini++;
    for (; ini != fin; ++ini, ++prev) {
        if (less(*ini,*prev)) return false;
    }
    return true;
}

template <typename Iterator>
inline bool is_sorted (Iterator ini, Iterator fin) {
    std::less<typename std::iterator_traits<Iterator>::value_type> less;
    return is_sorted(ini,fin,less);
}

template <class FIt, class Compare>
FIt is_sorted_until (FIt first, FIt last, Compare comp)  {
    if (first != last) {
        FIt next_ = first;
        while (++next_ != last) {
            if (comp(*next_, *first))
                return next_;
            first = next_;
        }
    }
    return last;
}

template <typename Iterator>
inline bool is_sorted_until (Iterator ini, Iterator fin) {
    std::less<typename std::iterator_traits<Iterator>::value_type> less;
    return is_sorted(ini,fin,less);
}


} }

#include "impl/11-is_xxx_of.hpp"
//! @include{doc} impl/11-minmax.hpp
#include "impl/11-minmax.hpp"
//! @include{doc} impl/11-permutations.hpp
#include "impl/11-permutations.hpp"

// 
// namespace std 
// 
#if (defined( CXXOFLAG_NO_STD_USING ) && CXXOFLAG_NO_STD_USING>0)
#else

	#if (CXXOMFORT_CXX_STD >= 2011)
	#elif (CXXO_COMPILER_SUPPORT_std_cxx11_algorithms>0)
	#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_DMC,>=850))
namespace std {
	using ::cxxomfort::cxxostd::copy_if;
	using ::cxxomfort::cxxostd::partition_copy;
	using ::cxxomfort::cxxostd::find_if_not;
	using ::cxxomfort::cxxostd::minmax;
	using ::cxxomfort::cxxostd::minmax_element;
}
	#else
namespace std {
	using ::cxxomfort::cxxostd::copy_if;
	using ::cxxomfort::cxxostd::copy_n;
	using ::cxxomfort::cxxostd::partition_copy;
	using ::cxxomfort::cxxostd::find_if_not;
	using ::cxxomfort::cxxostd::shuffle;
	//using ::cxxomfort::cxxostd::all_of;
	//using ::cxxomfort::cxxostd::any_of;
	//using ::cxxomfort::cxxostd::none_of;
	//using ::cxxomfort::cxxostd::minmax;
	//using ::cxxomfort::cxxostd::minmax_element;
}
	#endif

	#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_std_cxx11_algorithms>0)
		#undef  CXXOMFORT_IMPLEMENTS_algorithm_sorted
		#define CXXOMFORT_IMPLEMENTS_algorithm_sorted CXXO_IMPLSTATUS_NATIVE()
	#else
		#undef  CXXOMFORT_IMPLEMENTS_algorithm_sorted
		#define CXXOMFORT_IMPLEMENTS_algorithm_sorted CXXO_IMPLSTATUS_BACKPORT()
namespace std {
	using ::cxxomfort::cxxostd::is_sorted;
	using ::cxxomfort::cxxostd::is_sorted_until;
}
	#endif

#endif

// 
// c++14 algorithms
// 

#include "impl/14-algorithm-equal.hpp"

#if (defined( CXXOFLAG_NO_STD_USING ) && CXXOFLAG_NO_STD_USING>0)
#else

#endif

// 
// c++17 algorithms
// 

#if (defined(DOXYGEN_DOC))
namespace std {

/**
 * @param f A function or function-like.
 * @cxx17b Invoke a function for each element in a sequence.
 * @xrefitem std0algorithm "" "" Invoke a function for each element in a sequence.
 * @see @cppref{algorithm/for_each_n}
 * 
 * @note The return value of @a f is ignored. The "non-ignoring" equivalent is @c accumulate() .
 * 
 * */
template<class Iterator, class Size, class UnaryFunction>
Iterator for_each_n (Iterator first, Size n, UnaryFunction f)  {}

}

#endif

namespace cxxomfort { namespace cxxostd {

template<class II, class Size, class UnaryFunction>
II for_each_n (II first, Size n, UnaryFunction f) {
    for (Size i = 0; i < n; ++first, (void) ++i) {
        f(*first);
    }
    return first;
}

} }

#if (CXXOMFORT_CXX_STD >= 2017)
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1800) && CXXOMFORT_CXX_EMULATION>=2017)
#else
namespace std {
	using ::cxxomfort::cxxostd::for_each_n;
}
#endif

#include "impl/17-clamp.hpp"

// 
// c++20 algorithms
// 

// ! @include{doc} impl/20-shift.hpp
#include "impl/20-shift.hpp"


// 
// compilers needing ^ 
// 

// 
// static_clamp (from cxxo-utils)
// 

#include <cxxomfort/impl/static_math.hpp>

namespace cxxomfort { namespace fix {

	using ::cxxomfort::st_math::static_clamp;

} }

// 
// constexpr invocation of clamp
// 

namespace cxxomfort { namespace fix {

#if defined(DOXYGEN_DOC)
//! Implementation of @c std::clamp with @c constexpr
//! @ingroup cxxomfort-fix
template <typename T>
constexpr inline 
T clamp (cxxomfort::constexpr_t, T x, T Min, T Max) noexcept {
}

#elif (CXXO_COMPILER_SUPPORT_constexpr_relaxed)
template <typename T>
constexpr inline 
T clamp (cxxomfort::constexpr_t const, T x, T Min, T Max) noexcept {
	static_assert (Min<Max , "const_clamp:(min<max)");
	return ( x < Min ? Min : x <= Max ? x : Max );
}

#elif (CXXO_COMPILER_SUPPORT_constexpr)
template <typename T>
CXXO_CONSTEXPR  inline 
T clamp (cxxomfort::constexpr_t const, T x, T Min, T Max) CXXO_NOEXCEPTNOTHROW {
	return (!(Min<Max) ? throw std::logic_error("const_clamp:(min,max)") : 
		x < Min ? Min : x <= Max ? x : Max );
}

#else
#endif

}} // cxxomfort::fix

// 
// constexpr_find
// 

#if (defined(__cpp_lib_constexpr_algorithms) || defined(DOXYGEN_DOC))

// std::find() is constexpr since C++20
namespace cxxomfort { namespace fix {

// std::find() is constexpr since C++20

/**
 * @ingroup cxxomfort-fix
 * @cxx20b Adding @c constexpr capability to <tt>std::find()</tt>
 * 
 * Adds an "overload" to <tt>std::find()</tt> so that pre-C++20 can invoke it in constexpr in contexts where the compiler makes it possible.
 * 
 * @code
 * using cxxomfort::fix::find;
 * constexpr int Data[N] = { ...};
 * constexpr int const* p = find( cxxomfort::constexpr_tag, Data, Data+N, 9);
 * @endcode
 * 
 * @note This is defined only starting with C++11.
 * 
 * */
template <typename It, typename T>
constexpr It
find (cxxomfort::constexpr_t const, It ini, It fin, T const& value) {
	return std::find(ini,fin,value);
}

/**
 * @ingroup cxxomfort-fix
 * @cxx20b Adding @c constexpr capability to <tt>std::find_if()</tt>
 * 
 * Adds an "overload" to <tt>std::find_if()</tt> so that pre-C++20 can invoke it in constexpr in contexts where the compiler makes it possible.
 * 
 * @note This is defined only starting with C++11.
 * 
 * */
template <typename It, typename Fn>
constexpr It
find_if (cxxomfort::constexpr_t const, It ini, It fin, Fn&& fn) {
	return std::find_if(ini,fin,fn);
}

} }

#elif (CXXO_COMPILER_SUPPORT_constexpr )
	#include <cxxomfort/iterator.hpp>

// std::find() is constexpr since C++20
namespace cxxomfort { namespace fix {

// std::find() is constexpr since C++20

template <typename It, typename T>
constexpr It
find (cxxomfort::constexpr_t const, It ini, It fin, T const& value) {
	using namespace std;
	return ini==fin ? fin : 
		((*ini==value) ? ini : find(cxxomfort::constexpr_tag,static_next(ini),fin,value));
}

template <typename It, typename Fn>
constexpr It
find_if (cxxomfort::constexpr_t const, It ini, It fin, Fn&& fn) {
	using namespace std;
	return ini==fin ? fin : 
		(fn(*ini) ? ini : find_if(cxxomfort::constexpr_tag,static_next(ini),fin,fn));
}

} }

#endif


// 
// constexpr_count 
// 

#if (defined(__cpp_lib_constexpr_algorithms) || defined(DOXYGEN_DOC))

namespace cxxomfort { namespace fix {


/**
 * @ingroup cxxomfort-fix
 * @cxx20b Adding @c constexpr capability to <tt>std::count()</tt>
 * 
 * Adds an "overload" to <tt>std::count()</tt> so that pre-C++20 can invoke it in constexpr in contexts where the compiler makes it possible.
 * 
 * @code
 * using cxxomfort::fix::count;
 * constexpr int Data[N] = { ...};
 * constexpr int const* p = count( cxxomfort::constexpr_tag, Data, Data+N, 9);
 * @endcode
 * 
 * @note This is defined only starting with C++11.
 * 
 * */
template <typename It, typename T>
constexpr size_t 
count (cxxomfort::constexpr_t const, It ini, It fin, T const& value) noexcept {
	return std::count(ini,fin,value);
}

/**
 * @ingroup cxxomfort-fix
 * @cxx20b Adding @c constexpr capability to <tt>std::count_if()</tt>
 * 
 * Adds an "overload" to <tt>std::count_if()</tt> so that pre-C++20 can invoke it in constexpr in contexts where the compiler makes it possible.
 * 
 * @note This is defined only starting with C++11.
 * 
 * */
template <typename It, typename Fn>
constexpr size_t 
count_if (cxxomfort::constexpr_t const, It ini, It fin, Fn&& fn) noexcept {
	return std::count_if(ini,fin,fn);
}

} }

#elif (0 && CXXO_COMPILER_SUPPORT_constexpr_relaxed)

#elif (CXXO_COMPILER_SUPPORT_constexpr)
	#include <cxxomfort/iterator.hpp>

namespace cxxomfort { namespace fix {

template <typename It, typename T>
constexpr size_t 
count (cxxomfort::constexpr_t const, It ini, It fin, T const& value) noexcept {
	using namespace std;
	return ini==fin ? 0 : 
		((*ini==value) + count(cxxomfort::constexpr_tag,static_next(ini),fin,value));
}

template <typename It, typename Fn>
constexpr size_t 
count_if (cxxomfort::constexpr_t const, It ini, It fin, Fn&& fn) noexcept {
	using namespace std;
	return ini==fin ? 0 : 
		((fn(*ini)) + count_if(cxxomfort::constexpr_tag,static_next(ini),fin,fn));
}

} }
#endif


// 
// constexpr_equal
// 

#if (CXXO_COMPILER_SUPPORT_constexpr>0)
	#include <cxxomfort/iterator.hpp>

namespace cxxomfort { namespace fix {


template <typename It1, typename It2, typename Eq>
constexpr inline bool
equal_impl (cxxomfort::constexpr_t const, It1 ini1, It1 fin1, It2 ini2, It2 fin2, Eq&& eq, bool pre_) noexcept {
	using namespace std;
	return (ini1==fin1 || ini2==fin2) ? pre_ :
		(eq(*ini1,*ini2)) ? equal_impl(cxxomfort::constexpr_tag, static_next(ini1), fin1, static_next(ini2), fin2, eq, true) : false;
}

template <typename It1, typename It2>
constexpr inline bool
equal_impl (cxxomfort::constexpr_t const, It1 ini1, It1 fin1, It2 ini2, It2 fin2, bool pre_) 
noexcept {
	using namespace std;
	return (ini1==fin1 || ini2==fin2) ? pre_ :
		(*ini1==*ini2) ? equal_impl(cxxomfort::constexpr_tag, static_next(ini1), fin1, static_next(ini2), fin2, true) : false;
}

template <typename It1, typename It2, typename Eq>
constexpr bool
equal (cxxomfort::constexpr_t const, It1 ini1, It1 fin1, It2 ini2, It2 fin2, Eq&& eq) 
noexcept {
	return equal_impl( cxxomfort::constexpr_tag, ini1, fin1, ini2, fin2, eq, false);
}

template <typename It1, typename It2>
constexpr bool
equal (cxxomfort::constexpr_t const, It1 ini1, It1 fin1, It2 ini2, It2 fin2) 
noexcept {
	return equal_impl( cxxomfort::constexpr_tag, ini1, fin1, ini2, fin2, false);
}


} }
#endif


#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
	#pragma message CXXO_NOTICE("enabled <algorithm> support.")
#endif

#endif

/**
 * @page std0algorithm <algorithm>
 * @brief Backports related to Standard header <tt><algorithm></tt>
 * 
 * @code
 * #include <cxxomfort/algorithm.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces 
 * 
 * Backports defined in this section:
 * 
 * * @c Some C++98 algorithms that might be missing in older compilers, such as @c fill() .
 * * @c all_of() , @c any_of() , @c none_of() (C++11) - check if elements of a sequence fit a predicate.
 * * @c copy_if() , @c copy_n() , @c partition_copy() (C++11) - copy elements from a sequence.
 * * @c find_if_not() (C++11) - check if elements @e don't fit a predicate.
 * * @c is_permutation() , @c next_permutation(), @c prev_permutation() (C++11) - check the sorting status of a sequence.
 * * @c is_sorted() , @c is_sorted_until() (C++11) - check the sorting status of a sequence.
 * * @c minmax() , @c minmax_element() (C++11) - minimum and maximum elements in a sequence.
 * * @c equal() , @c mismatch() overloads (C++14) - compare sequences.
 * * @c clamp() (C++17) - constrain values.
 * * @c count() , @c count_if() (C++17) - count elements in a sequence.
 * * @c for_each_n() (C++17) - execute functors across a sequence.
 * * @c shift_left(), @c shift_right() (C++20) - slide a sequence.
 * 
 * Ommitted:
 * 
 * * @c shuffle
 * 
 * Non-backport features ( @ref cxxomfort-fixes and others ):
 * 
 * * Static / Compile-time Algorithms - see static_algorithms .
 * + @c constexpr algorithms - see @ref algorithm_constexpr 
 * 
 * @section algorithm_constexpr Constexpr-capable algorithms
 * 
 * Starting with C++20, a high number of features in <tt><algorithm></tt> are made @c constexpr capable.
 * 
 * This header offers a number of @c constexpr -capable implementations for some algorithms that can be made @c constexpr with only C++11 or C++14 requirement but weren't in their corresponding Standard. Because the names can not be reused by cxxomfort without UB as they might end up sharing the same number of arguments (see eg.: @c equal() / @c mismatch() in C++14) , they are instead provided via a new mechanism:
 * 
 * The functions are provided in namespace @c cxxomfort::fix and lead with a @ref @c constexpr_tag argument. For example, the following invocations will call the constexpr-capable functions in cxxomfort:
 * 
 * @code
 * // may or may not work in constexpr, 
 * // depending on *stdlib* (not on compiler)
 * constexpr int* px = std::find( arr, arr+N, 8 );
 * // should work always in constexpr 
 * // if the *compiler* is capable
 * using cxxomfort::constexpr_tag
 * constexpr int* px = cxxomfort::fix::find( constexpr_tag, arr, arr+N, 8 );
 * @endcode
 * 
 * These algorithms are:
 * 
 * * @c cxxomfort::fix::clamp()
 * * @c cxxomfort::fix::count()
 * * @c cxxomfort::fix::count_if()
 * * @c cxxomfort::fix::find()
 * * @c cxxomfort::fix::find_if()
 * * @c cxxomfort::fix::equal()
 *
 * @note Without relaxed constexpr, how much work can these algorithms do depends on the compiler's recursion level for @c constexpr . Consult your compiler manual for more information.
 * 
 * @note These algorithms can only work constexpr with native arrays, and with @c constexpr types whose iterators are pointer types (eg.: <tt>basic_string</tt> with default allocator).
 * 
 * @section demo Demonstration:
 * 
 * Copies and prints elements from a list via 
 * @c copy_if() and @c for_each_n() .
 * 
 * @code

struct  {
    template <typename Numeric> 
    bool operator() (Numeric n) const { 
        return n % 2 == 0; 
    }
}   even= {};

struct {
    template <typename T>
    SS& operator() (T const& t) const {
        return os<< t;
    }
  
    std::ostream& os;
}   print (std::cout);

list<int> s1;
generate( begin(s1), end(s1), rand);
vector<int> s2;
copy_if( begin(s1), end(s1), back_inserter(s2), even);
for_each_n( begin(s1), end(s1), print);

 * @endcode
 * 
 * @section algorithm_examples Examples 
 * 
 * For other examples see the following entries in @c cxxomfort-examples : 
 * 
 * * @c <a href="https://notabug.org/lmachucabezzaza/cxxomfort-examples/src/master/examples/randomly.cpp">randomly</a> - usage of various algorithms.
 * 
 * * <em>More pending...</em>
 *
 * @section also See Also
 * 
 * @see @ref cxxomfort/algorithm.hpp (header reference)
 * @see @cppref{header/algorithm} (cppreference) 
 * @see @ref std0numeric (std numeric algorithms) , @ref cxxo-sup-algorithm (supplemental)
 * 
 * */
