#ifndef CXXOMFORT_BASE_HPP
#define CXXOMFORT_BASE_HPP
/**
 * @file
 * @brief Minimal cxxomfort setup.
 * @ingroup cxxo-base
 * 
 * For more complete cxxomfort setup (includes more headers, adds new names 
 * and functions), use one of these:
 * 
 * @code
 * #include <cxxomfort/backports.hpp> // base + backports
 * #include <cxxomfort/cxxomfort.hpp> // base + backports + library supplements
 * @endcode
 * 
 */

//#include <cstddef>
//#include <ciso646>
#include "config.hpp"
#include <cstddef>

// -- Cxxomfort basic features, always enabled by default --
#include "./base/alignof.hpp" // alignof impl
#include "./base/nullptr.hpp" // nullptr
#include "./base/static_assert.hpp" // static_assert
#include "./base/cstdint.hpp" // c++11 integer types ([u]intNN_t)
#include "./base/iterator.hpp" // iter. accessors (begin/end, next/prev)
#include "./base/move.hpp" // move emulation
#include "./base/explicit_cast.hpp" // explicit cast operator
#endif

/**
 * @defgroup cxxo-base Base Features of cxxomfort
 * 
 * The set of basic features of cxxomfort. 
 * These features are included automatically when 
 * <code>#include <cxxomfort/base.hpp></code> 
 * or any of the cxxomfort headers are @c #include -d-.
 * 
 * Most of the features here involve setting up a common ground 
 * between C++03 and C++11; of particular interest is the backporting 
 * of the two keywords @c nullptr and @c static_assert .
 * 
 * * @c alignof()
 * * @c #nullptr - see @ref cxxo_nullptr .
 * * @ref cxxo_static_assert "compile time assertions (static_assert)".
 * * @link cxxo:explicit_cast @c explicit_cast<> @endlink .
 * * @link backported_move_operations Move operations @endlink .
 * * rvalue-reference emulation.
 * * @ref cxxo_global_beginend "global iterator accessors begin/end"
 * 
 * 
 * @section cxxo_nullptr nullptr - null pointer literal
 * 
 * cxxomfort includes among its facilities the "official" backport implementation for the special literal @c nullptr which represents a "null pointer constant". 
 * 
 * @c nullptr can be used anywhere where a pointer could be initialized to a "null" value, replacing the usage of eg.: <tt>int * p = NULL</tt> or <tt>0</tt> in older code.
 * 
 * The Standard also provides the name @c std::nullptr_t for the type of the of the null pointer literal and can be used for overloading. Cxxomfort also provides this name dirctly.
 * 
 * The macro <tt>CXXOMFORT_IMPLEMENTS_nullptr</tt> informs of the status of implementation:
 * 
 * * <tt> >= ...NATIVE() </tt> - found as a reserved word (C++11 onwards)
 * * <tt> >= ...BACKPORT() </tt> - not found, implemented by cxxomfort as a reserved object.
 * * <tt> >= ...EMULATION() </tt> - not found, implemented by cxxomfort as a @b macro with the value <tt>0</tt>.
 * * <tt>else </tt> - not implemented by cxxomfort.
 * 
 * @subsection also1 See Also
 * 
 * * @wg21{n1601,"null pointer literal"}
 * * @cppref{keyword/nullptr}
 * * the "official" backport at http://en.wikibooks.org/wiki/More_C++_Idioms/nullptr#Solution_and_Sample_Code and the <a href="https://stackoverflow.com/questions/8747005/backporting-nullptr-to-c-pre-c0x-programs">StackOverflow question on the subject of the backport</a>.
 * 
 * 
 * 
 * @section cxxo_static_assert static_assert - Compile time assertions 
 * 
 * cxxomfort includes among its facilities a backport implementation 
 * for the "compile time assertions" feature: the C++11 keyword 
 * @c static_assert can be used to issue a check that is performed a 
 * compile time, resulting in an error message if the condition 
 * is not met.
 * 
 * @note Because the backport is implemented as a macro, it is sensitive to commas and parentheses; see usage examples below.
 * 
 * Example usage:
 * 
 * @code

// example static_assert
static_assert( is_same<typename std::string::value_type, char>::value, "string?");
// since it is a macro, tests that contain commas must be reformulated
static_assert((sizeof(pair<int,int>) > sizeof(int)), "pair?");
enum foo { ... };
enum { complicated_test = 
	is_same< typename underlying_type<foo>::type, 
		typename decay< map< int, vector<foo> >::value_type >::type
	>::value
};
static_assert (complicated_test, "complicated?");
 * @endcode
 * 
 * As a cxxomfort-specific extension, the macro @c static_assert0 is also provided that emulates C*+17's "short static_assert" feature:
 * 
 * @code
 * static_assert ( test, "message" );
 * static_assert0 ( test ); // roughly equals static_assert ( test, "test" ); 
 * @endcode
 * 
 * The macro @c CXXOMFORT_IMPLEMENTS_static_assert informs of the status of @c static_assert as found by the library:
 * 
 * * <tt> >= CXXO_IMPLSTATUS_NATIVE() </tt> - found as a keyword.
 * * <tt> >= CXXO_IMPLSTATUS_EMULATION() </tt> - implemented by cxxomfort as a macro.
 * * <tt> >0 </tt> - found as a macro not defined by cxxomfort (possibly vendored)
 * * <tt> == 0 </tt> - not found, and not implemented.
 * 
 * See @c #static_assert for more details.
 * 
 * @section cxxo_explicit_cast explicit_cast<To>
 * 
 * cxxomfort library automatically includes an implementation 
 * of <em>Imperfect C++</em>'s @c explicit_cast<>  
 * for use in types that need to enable explicit conversion 
 * in both C++11 (where <code>explicit operator...</code> is available) 
 * and C++03. 
 * 
 * @code
 * struct A {
 * 	...
 * };
 * 
 * struct B {
 * 	...
 *  CXXO_EXPLICIT_OPERATOR(A) () const {
 * 		return A();
 *  }
 * 
 * B b;
 * A a = explicit_cast<A>(b);
 * // equivalent to static_cast in C++>=11.
 * 
 * @endcode
 * 
 * See @ref explicit_cast for more information and 
 * <a href="https://stackoverflow.com/questions/13193766/isset-or-operator-void-or-explicit-opertor-bool-or-something-else">
 * this StackOverflow question</a> for more rationale on implementation and mapping to C++11.
 * 
 * @section cxxo_movable "Move" emulation
 * 
 * cxxomfort implements a simplified variant of Boost:Move aka 
 * "move semantics" in C++03.
 * 
 * @section cxxo_global_beginend Global Iterator Accessors
 * <h3>Global iterator accessors</h3>
 * 
 * cxxomfort provides backports for the global iterator accessors 
 * <code>std::begin</code> and <code>std::end</code> in C++11 which 
 * allow for uniform access of sequences and container objects.
 * 
 * @code

array<int, 4> S1;
char S2[27];
vector<float> S3;
map<int,string> S4;

// for each of these, the following expression works replacing the object:

transform ( begin(S1), end(S1), begin(S1), somefunction );

 * @endcode
 * 
 * For related features see: @c #size() , @c #move() , @c fixed_vector . 
 * 
 * The global begin, end are documented at 
 * <a href="https://en.cppreference.com/w/cpp/iterator/begin">begin</a> and 
 * <a href="https://en.cppreference.com/w/cpp/iterator/end">end</a> 
 * \@ @e cppreference .
 * 
 * @section inclusion1 Inclusion Details
 * 
 * Except for this page all features should be explicitly included to the 
 * project by @c #include -ing the necessary header.
 * 
 * * @ref cxxomfort
 * * @ref cxx11-backports 
 * * @ref cxx14-backports
 * * @ref cxx17-backports
 * * @ref cxx20-backports 
 * * @ref cxx23-backports 
 * 
 * 
 */
