#ifndef CXXOMFORT_BASE_EXPLICIT_CAST_HPP
#define CXXOMFORT_BASE_EXPLICIT_CAST_HPP
/**
 * @file
 * @brief Explicit cast, and @c explicit_cast conversion operator for C++03
 * @ingroup cxxo-base
 * 
 * Interfaces defined in this header:
 * 
 * * @c @ref explicit_cast (as a synonym in C++11)
 * * @c CXXO_EXPLICIT_OPERATOR
 * 
 */

#include "../config.hpp"
#define CXXOMFORT_IMPLEMENTS_n2333 CXXO_IMPLSTATUS_EMULATION()
#define CXXOMFORT_IMPLEMENTS_n2437 CXXO_IMPLSTATUS_EMULATION()

#undef  CXXOMFORT_IMPLEMENTS_explicit_operator
#define CXXOMFORT_IMPLEMENTS_explicit_operator CXXOMFORT_IMPLEMENTS_n2333


namespace cxxomfort {
namespace detail {

	template <typename T> 
	struct addref_t {
		typedef T& type;
	};

	template <typename T> 
	struct addref_t<T&> {
		typedef T& type;
	};

	template <typename A, typename B> struct detail_is_same { enum { value = false }; };
	template <typename A> struct detail_is_same<A,A> { enum { value = true }; };

} } 

#if (defined(DOXYGEN_DOC))

namespace cxxomfort { namespace detail {

 /** 
 * 
 * @def CXXO_EXPLICIT_OPERATOR(T)
 * @brief Implements a member conversion operator that provides explicit conversion
 * @ingroup base-features
 * 
 * This macro creates a conversion operator that provides an explicit conversion 
 * to a type of argument type @p T . It is a cross-Standards alternative to <tt>explicit operator</tt> of C++11.
 * 
 * This is defined to be an accessory conversion that the caller 
 * invokes via <tt>explicit_cast<T>(object)</tt>.
 * 
 * @see <tt>explicit operator</tt> introduced in C++11 via @wg21{n2333} .
 * 
 * 
 */
#define CXXO_EXPLICIT_OPERATOR(T)  implementation_defined
/**
 * @brief Explicitly cast an object of type F to a type T
 * @ingroup base-features
 * 
 * This pseudo-keyword performs an @c explicit -using cast on an object @em f 
 * to a target type @em T . This is implemented via a passthrough-only 
 * holder type, that in C++11 uses move operation to construct when available.
 * 
 */
template <typename T, typename F>
T explicit_cast (F f) { return implementation_defined_cast<T>(f); }

} }

#elif (CXXO_COMPILER_SUPPORT_explicit_operator)
namespace cxxomfort { namespace detail {

template <typename T, typename F> CXXO_CONSTEXPR inline 
T explicit_cast (F const& f) {
	return static_cast<T>(f);
}

} }

using ::cxxomfort::detail::explicit_cast;

	#define CXXO_EXPLICIT_OPERATOR(T) explicit operator T


#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,<1800))
namespace cxxomfort { namespace detail {

template <typename T>
struct explicit_cast_t {
    T value;
	explicit_cast_t (T const& t) CXXO_NOEXCEPT : value(t) { }
	explicit_cast_t & operator= (T t) { value= t; return *this; }
    CXXO_CONSTEXPR 
    operator T const& () const CXXO_NOEXCEPTNOTHROW  { return value; }
    //operator T () const CXXO_NOEXCEPTNOTHROW { return value; }
    
    private:
    //explicit_cast_t& operator= (explicit_cast_t const&) ;
};

} }
	#define explicit_cast ::cxxomfort::detail::explicit_cast_t
	#define CXXO_EXPLICIT_OPERATOR(T) operator ::cxxomfort::detail::explicit_cast_t<T> 


#else
	#if (CXXOMFORT_CXX_STD >= 2011 || \
	(CXXO_COMPILER_SUPPORT_explicit_operator && CXXO_COMPILER_SUPPORT_rvref)\
	)
		#include <utility> // std::move
	#endif
namespace cxxomfort { namespace detail {

template <typename T>
struct explicit_cast_t {
    T value;

	explicit_cast_t (T const& t) CXXO_NOEXCEPT : value(t) { }
	explicit_cast_t & operator= (T t) { value= t; return *this; }
    CXXO_CONSTEXPR 
    operator T const& () const CXXO_NOEXCEPTNOTHROW  { return value; }
    //operator T () const CXXO_NOEXCEPTNOTHROW { return value; }
    
    private:
    //explicit_cast_t& operator= (explicit_cast_t const&) ;
};

template <typename T, typename F>
T explicit_cast (F const& f) CXXO_NOEXCEPT { 
	//explicit_cast_t<T> x= f;
	//return x;
	return explicit_cast_t<T>(f); ;
	// return (f.operator explicit_cast_t<T>());  
}

} }

	#define CXXO_EXPLICIT_OPERATOR(T) operator ::cxxomfort::detail::explicit_cast_t<T> 

//! @c explicit_cast as a "keyword"
using ::cxxomfort::detail::explicit_cast;


#endif

#if defined(DOXYGEN_DOC)
/**
 * 
 * @page explicit_cast explicit_cast
 * 
 * The cxxomfort library implements an emulation of the "explicit operator" feature of C++11 that allows classes to provide explicit conversion to a defined set of types. 
 * A customized conversion "cast" that works like C++>=11's <tt>explicit operator</tt> and is invoked explicitly via a pseudokeyword "<tt>explicit_cast</tt>". The implementation is built upon based on the explicit_cast implementation from <em>Imperfect C++: Solutions for the Real World</em>, and "adapted out" in C++11.
 * 
 * An example usage:
 * 
 * @code

// lib code

class MyClass {
  ....
  CXXO_EXPLICIT_OPERATOR(bool) () const {
      ...return a bool somehow
  }
};

// client code

MyClass myobject;
...
if (explicit_cast<bool>(myobject)) {
    ...do something
} else {
    ...do something else
}
 * 
 * @endcode
 * 
 * 
 * @section Code Generation
 *
 * In order to invoke this feature, two special keywords are used:
 * 
 * @e Library code adds a special member function to the class using the macro 
 * <tt>CXXO_EXPLICIT_OPERATOR(target_type)</tt> to declare its signature (potentially adding cv-qualification), then continues with the body of the conversion operator.
 * 
 * @e Client code evaluates the conversion via the pseudokeyword <tt>explicit_cast<target_type>(variable)</tt>.
 * 
 * 
 * When in a C++<11 context, using this feature resolves to a special function call that provides a @c bool ; when in C++>=11 context, it resolves instead to the native <tt>explicit operator bool()</tt>.
 * 
 * Example result code for the above code in C++03:
 * 
 * @code

class MyClass {
    operator explicit_cast<bool> () const {
        ...return a bool somehow
    }
};

if (explicit_cast<bool>(myobject)) {
    ....
} else {
	....
}

 * @endcode
 * 
 * Whereas resulting code in a C++>=11 context looks like this:
 * 
 * @code

class MyClass {
    explicit operator bool () const {
        ...return a bool somehow
    }
}

if (static_cast<bool>(myobject)) {
    ....
} else { 
	....
}

 * @endcode
 * 
 * @section Known Limitations
 * 
 * Because the feature is implemented as a pseudo-keyword, it can not be used to create variables at <tt>if(...)</tt> scope, as C++<03 does not have the ability to autoinvoke bool conversions in such contexts.
 * 
 * The solution is simple: implement the guard variable in an outer scope:
 * 
 * @code

// This won't work in C++03 unless MyClass *implicitly* converts to bool
if (MyClass myobject = ....) { ...}
// This always work, in C++98, C++03, C++11, ...
{ MyClass myobject = ....;
if (explicit_cast<bool>(myobject)) { .... }
} // "myobject" dies at this point

 * @endcode
 * 
 * @see http://ryan.gulix.cl/fossil.cgi/cxxomfort/wiki?name=Features/explicit_cast 
 * 
 * */
 
#endif

#if defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 2)
    #pragma message CXXO_NOTICE("enabled base 'explicit operator' emulation")
#endif


#endif
