#ifndef CXXOMFORT_BASE_ITERATOR_HPP
#define CXXOMFORT_BASE_ITERATOR_HPP
/**
 * @file
 * @brief Implements some C++11 functions in C++03 from the "iterator" header.
 * @author Luis Machuca Bezzaza <luis [dot] machuca [at] gulix [dot] cl>
 *
 * Interfaces defined in this header:
 *
 * * @c begin() , @c end() (also for namespace std)
 * * @c next() , @c prev() (also for namespace std)
 * * @c cbegin() (also for namespace std)
 * * @c cend() (also for namespace std)
 *
 */

#include "../config.hpp"
#include "../base.hpp"

#undef CXXOMFORT_USING_ITERATOR
#if (CXXOMFORT_CXX_STD >= 2011)
	#define CXXOMFORT_USING_ITERATOR 0 // none to do
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_11 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILER_SUPPORT_iterator_accessors_11> 0)
	#define CXXOMFORT_USING_ITERATOR 0 // none to do
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_11 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC)
	#define CXXOMFORT_USING_ITERATOR 1 // simple

#elif (CXXOMFORT_CXX_STD >= 1997 && CXXOMFORT_CXX_EMULATION<= 2011)
    #define CXXOMFORT_USING_ITERATOR 2 // complex
	#include <iterator>

#else
	#define CXXOMFORT_USING_ITERATOR 1 // simple

#endif

// 
// implementation
// 

#if (CXXOMFORT_USING_ITERATOR >= 1)

#endif
#if (CXXOMFORT_USING_ITERATOR==2)
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_11 CXXO_IMPLSTATUS_BACKPORT()

namespace cxxomfort { namespace iterator {

template <typename T>
struct check_type {
    enum { s = sizeof(T) };
};

//
// Detect the presence of begin, end, size, front and back member functions
//
template <typename T>
struct iterator_accessors_helper {

    typedef struct { int a; } yes;
    typedef char no;

    private:
    template <typename U, U> struct really_has;

    template <typename C> static yes& Test_begin(really_has <typename C::const_iterator (C::*)() const,
                            &C::begin>*);

    template <typename C> static yes& Test_begin(really_has <typename C::iterator (C::*)() const,
                            &C::begin>*);

    template <typename C> static yes& Test_end(really_has <typename C::const_iterator (C::*)() const,
                            &C::end>*);

    template <typename C> static yes& Test_end(really_has <typename C::iterator (C::*)() const,
                            &C::end>*);

    template <typename C> static yes& Test_size(really_has <typename C::difference_type (C::*)() const,
                            &C::size>*);

    template <typename C> static yes& Test_size(really_has <size_t (C::*)() const,
                            &C::size>*);

    template <typename> static no& Test_begin(...);
    template <typename> static no& Test_end(...);
    template <typename> static no& Test_size(...);
    
    public:
    static const bool has_begin    = (sizeof(Test_begin<T>(0)) == sizeof(yes));
    static const bool has_end      = (sizeof(Test_end<T>(0)) == sizeof(yes));
    static const bool has_size     = (sizeof(Test_size<T>(0)) == sizeof(yes));


};

/*
 * This is marked as a system header mostly because of 
 * the generation of intended but annoying 
 * "typedef defined but not used" warnings that can turn 
 * into errors in strict builds.
 */
	#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC \
		&& CXXOMFORT_COMPILER_VERSION >= 406)
	#pragma GCC system_header
	#endif

/**
 * @ingroup iterator
 * @{
 */

//! Returns the "begin" iterator for a container-like object.
//! @ref cxx11-backports
template <typename C>
typename C::iterator inline
begin (C & c) {
    static_assert(sizeof(typename C::iterator), "");
    static_assert(sizeof(typename C::value_type), "");
    return c.begin();
}

//! @overload cxxomfort::begin
template <typename C> inline
typename C::const_iterator
begin (C const& c) {
    static_assert(sizeof(typename C::const_iterator), "");
    static_assert(sizeof(typename C::value_type), "");
    return c.begin();
}

//! @overload cxxomfort::begin 
template <typename T, ::std::size_t N> inline
T* begin (T(&arr)[N]) { return arr; }

//! Returns the "end" iterator for a container-like object.
//! @ref cxx11-backports
template <typename C> inline
typename C::iterator
end (C & c) {
    static_assert(sizeof(typename C::iterator), "");
    static_assert(sizeof(typename C::value_type), "");
    return c.end();
}

//! @overload cxxomfort::end 
template <typename C> inline
typename C::const_iterator
end (C const& c) {
    static_assert(sizeof(typename C::const_iterator), "");
    static_assert(sizeof(typename C::value_type), "");
    using namespace std;
    return c.end();
}

//! @overload end
template <typename T, size_t N> inline
T* end (T(&arr)[N]) { return arr+N; }

// cbegin, cend moved to impl/14-iterator-access.hpp

// rbegin, rend moved to impl/14-iterator-access.hpp


/**
 * @brief Moves an iterator forwards an amount of steps, returning the result.
 * @ingroup iterator
 * @ref cxx03-backports
 *
 * This is reflected as @c std::next .
 */
template <typename FIterator> inline
FIterator next
(FIterator i, typename std::iterator_traits<FIterator>::difference_type n = 1
) {
    std::advance(i, n);
    return i;
}

/**
 * @brief Moves an iterator backwards an amount of steps, returning the result.
 * @ingroup cxx11-backports
 * @ref iterator
 *
 * This is reflected as @c std::prev .
 */
template <typename FIterator> inline
FIterator prev (FIterator i
, typename std::iterator_traits<FIterator>::difference_type n = 1
) {
    std::advance(i, -n);
    return i;
}



	#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC \
		&& CXXOMFORT_COMPILER_VERSION >= 408)
	#pragma GCC diagnostic pop
	#endif
} } // cxxomfort::iterator


/*
 * Provide begin, end support for std::valarray as well.
 */
#include <iterator>

namespace std {
template <typename T> class valarray;
}

namespace cxxomfort { namespace iterator {

template <typename T> inline
T* begin (std::valarray<T> & c) { return addressof(c[0]); }

template <typename T> inline
T const* begin (std::valarray<T> const& c) { return addressof(c[0]); }

template <typename T> inline
T* end (std::valarray<T> & c) { return addressof(c[0])+c.size(); }

template <typename T> inline
T const* end (std::valarray<T> const& c) { return addressof(c[0])+c.size(); }


}
} //~::cxxomfort


#elif (CXXOMFORT_USING_ITERATOR == 1) // simple
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_11 CXXO_IMPLSTATUS_EMULATION()
    // compilers like OW, that have very little template features

namespace cxxomfort { namespace iterator {

template <typename T>
typename T::iterator begin (T& t) {
    return (t.begin());
}

template <typename T>
typename T::iterator end (T& t) {
    return (t.end());
}

template <typename T>
typename T::const_iterator begin (T const& t) {
    return (t.begin());
}

template <typename T>
typename T::const_iterator end (T const& t) {
    return (t.end());
}

template <typename U, size_t N>
U* begin (U(&arr)[N]) { return arr; }
template <typename U, size_t N>
U* end (U(&arr)[N]) { return arr+N; }

template <typename FIterator> 
FIterator next (FIterator i, int n = 1
) {
    std::advance(i, n);
    return i;
}

template <typename FIterator> 
FIterator prev (FIterator i, int n = 1
) {
    std::advance(i, -n);
    return i;
}


} }  // cxxomfort::iterator

#else

#endif

//#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC)
//#pragma GCC diagnostic warning "-Wunused-local-typedefs
//#endif



namespace std {

#if (CXXOMFORT_CXX_STD < 2011 && CXXO_COMPILER_SUPPORT_iterator_accessors_11 == 0)
    using ::cxxomfort::iterator::begin;
    using ::cxxomfort::iterator::end;
// GCC 4.4 already has next, prev in c++0x mode
    #if (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_GCC && CXXOMFORT_CXX_EMULATION == 2011)
    #else
    using ::cxxomfort::iterator::next;
    using ::cxxomfort::iterator::prev;
    #endif

#else

#endif
}

#endif
