#ifndef CXXOMFORT_NULLPTR_HPP
#define CXXOMFORT_NULLPTR_HPP
/**
 * @file
 * @brief Implementation detaisl for "nullptr" backport.
 * @author Luis Machuca Bezzaza <luis [dot] machuca [at] gulix [dot] cl>
 * @ingroup cxxo-base
 *
 * Interfaces defined in this file:
 * 
 * * nullptr_t (also in namespace std)
 * * nullptr (also in namespace std)
 */
#include "../config.hpp"


#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_nullptr>0)
    #define CXXOMFORT_IMPLEMENTS_n1488 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD >= 1997 && CXXO_COMPILER_SUPPORT_nullptr==0)
    #define CXXOMFORT_USING_nullptr 2
    #define CXXOMFORT_IMPLEMENTS_n1488 CXXO_IMPLSTATUS_BACKPORT()

#else
	#define CXXOMFORT_USING_nullptr 1 // simple
    #define CXXOMFORT_IMPLEMENTS_n1488 CXXO_IMPLSTATUS_EMULATION()
#endif


/*
 * This is marked as a system header mostly because of 
 * bug https://gcc.gnu.org/bugzilla/show_bug.cgi?id=48914
 * where it is not possible to omit "nullptr is reserved" warnings.
 */
#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC \
	&& CXXOMFORT_COMPILER_VERSION >= 406)
	#pragma GCC system_header
#endif


#if (CXXO_COMPILER_SUPPORT_nullptr == 0 && defined(CXXOMFORT_USING_nullptr) && CXXOMFORT_USING_nullptr == 2)
    #if defined (CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 2)
		#pragma message CXXO_NOTICE("enabled 'nullptr' recommendation backport")
    #endif
namespace std {

// This is pretty much borrowed from the Standards recommendation verbatim.

class nullptr_t {
public:
    template<typename T>           // convertible to any type
    operator T*() const            // of null non-member
    { return 0; }                  // pointer...
    template<typename C, typename T> // or any type of null
    operator T C::*() const          // member pointer...
    { return 0; }
private:
    void operator&() const;          // whose address can't be taken
};

} //~namespace std

const std::nullptr_t nullptr = {};

#elif (defined(CXXOMFORT_USING_nullptr) && CXXOMFORT_USING_nullptr==1) // pre-C++98
    #if defined (CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 2)
		#pragma message CXXO_NOTICE("enabled 'nullptr' emulation")
    #endif

	// some compilers don't have support for templates that would allow the 
	// templated conversion operator.
	// here, we use a simpler nullptr that converts at least 
	// to (void*)0-

namespace std {

	struct nullptr_t {
		operator int() const { return NULL; }
	};

} // std

	#define nullptr NULL

#endif

#if defined (CXXOMFORT_USING_nullptr)



/**
 * @brief Null pointer literal (backport)
 * @ingroup cxx11-backports
 * @see http://stackoverflow.com/questions/8747005/backporting-nullptr-to-c-pre-c0x-programs
 * 
 * The null pointer literal @c nullptr allows for explicit comparison or 
 * assignment with pointer types, to disambiguate from contexts where eg.: 
 * @c 0 (zero) or a macro like @c NULL are used.
 * 
 * @code
int n = 0;    // valid
int q = nullptr; // invalid
int* ptrn = 0; // valid, but not recommended
int* ptrq = nullptr; // valid
if (ptrq != nullptr) { ... }
 * @endcode
 */


#else

#endif // using emulation

#define CXXOMFORT_IMPLEMENTS_n1601 CXXOMFORT_IMPLEMENTS_n1488
#define CXXOMFORT_IMPLEMENTS_n2214 CXXOMFORT_IMPLEMENTS_n1601
#undef  CXXOMFORT_IMPLEMENTS_nullptr
#define CXXOMFORT_IMPLEMENTS_nullptr CXXOMFORT_IMPLEMENTS_n2214

#endif // file
