#ifndef CXXOMFORT_BIT_HPP
#define CXXOMFORT_BIT_HPP
/**
 * @file
 * @brief Implementations and completion of names tied to <tt>bit</tt>.
 * @xrefitem std0bit "<bit>" ""
 *
 */

#include "config.hpp"
#include "base.hpp"

// 
// std::endian 
// 

#if (CXXOMFORT_CXX_STD>=2020 || CXXO_CXXSTDHEADER_bit)
	#include <bit>
#else
	#include "impl/20-endian.hpp"
#endif // c++

// 
// fix:: bit_mask versions for compile-time
// 

#include <limits>
#if (CXXOMFORT_CXX_STD >= 2011)
	#include <type_traits>
#else
	#include <cxxomfort/type_traits.hpp>
#endif
#include <cxxomfort/impl/math.hpp>

namespace cxxomfort { namespace fix {

/**
 * @brief Defines commonly-used bitmasks.
 * @cxxomfort-fixes Defines commonly-used bitmasks.
 * @sa @ref std0bit
 * */
template <typename U>
struct bit_masks {
	#if (CXXOMFORT_CXX_STD >= 2011)
	static_assert ( std::is_unsigned<U>::value , "bit_masks needs an unsigned type!" );
	#endif

	enum e { bits = std::numeric_limits<U>::digits };
	//! Value with only the left-most bit set.
	static const U lmb1 = U(1u)<< (bits-1); // leftmostbit1
	//! Value with only the "half-way" bit set.
	static const U hmb1 = U(1u)<< (bits/2); // half-most bit 1
	//! Value with only the lowest half of bits set.
	static const U lo_half = hmb1-1; // lowest bits set
	//! Value with only the highest half of bits set.
	static const U hi_half = ~lo_half; // highest bits set
};


#if defined(DOXYGEN_DOC)
namespace cxxomfort { namespace fix {

//! Defines member @c value with the number of lefmost @c 0 s in @a u .
//! @xrefitem Exposes in <tt>::value</tt> the number of leftmost @c 0 in @a u .
template <typename U, U u>
struct count_l_0_t {};

//! Defines member @c value with the number of lefmost @c 1 s in @a u .
//! @xrefitem Exposes in <tt>::value</tt> the number of leftmost @c 1 in @a u .
template <typename U, U u>
struct count_l_1_t
: public count_l_0_t<U,static_cast<U>(~u)> {};

//! Defines member @c value with the position of the leftmost @c 1 bit in @a u .
//! @xrefitem Integral trait that exposes the position of the leftmost @c 1 in @a u .
template <typename U, U u>
struct pos_l_1_t {};

} } // cxxomfort::fix

#else

namespace detail {

template <typename U, U u>
struct complement_value {
	static U const value = ~u;
};

template <typename U, size_t N, U u>
struct count_l_0_iterate {
	static const U value = N==0 ? 0 : (bit_masks<U>::lmb1 & u ? 0 : U(1 + count_l_0_iterate<U,N-1,U(u<<1)>::value) );
};

template <typename U, U u>
struct count_l_0_iterate<U,0,u> {
	static const U value = 0; 
};

} // detail 

template <typename U, U u>
struct count_l_0_t
: std::integral_constant<uint32_t, (u == 0 ? U(bit_masks<U>::bits) 
	: U(detail::count_l_0_iterate<U,bit_masks<U>::bits,u>::value) )>
{ };

template <typename U, U u>
struct count_l_1_t
: public std::integral_constant<uint32_t, count_l_0_t<U, detail::complement_value<U,u>::value >::value >
{ };

// find the position of leftmost bit 1

template <typename U, U u>
struct pos_l_1_t {
	static const U value = u == 0 ? U(bit_masks<U>::bits+1) 
	: bit_masks<U>::bits - count_l_0_t<U,u>::value - 1;
};

#endif


} } // cxxomfort::fix

#include <cxxomfort/impl/mathbit.hpp>

// 
// bit_width, bit_floor, bit_ceil 
// 

#if defined(DOXYGEN_DOC)
namespace std {

/**
 * @ingroup std0bit
 * @{
 * */

/** 
 * @cxx20b Bit operations: number of bits required to represent the value of @a u .
 * @xrefitem std0bit "" "" Bit operations: number of bits required to represent the value of @a u .
 * @pre <tt>std::is_unsigned<U>::value</tt> 
 * */
template <typename U>
constexpr U bit_width (U u) noexcept {};

/**
 * @cxx20b Bit operations: round @a u down to a power of 2.
 * @xrefitem std0bit "" "" Bit operations: round @a u down to a power of 2.
 * @pre <tt>std::is_unsigned<U>::value</tt> 
 * */
template <typename U>
constexpr U bit_floor (U u) noexcept {};

/**
 * @cxx20b Bit operations: round @a u up to a power of 2.
 * @xrefitem std0bit "" "" Bit operations: round @a u up to a power of 2.
 * @pre <tt>std::is_unsigned<U>::value</tt> 
 * */
template <typename U>
constexpr U bit_ceil (U u) noexcept {};

/**
 * @}
 * */
} } 

#elif (defined(__cpp_lib_bitops)) 
namespace cxxomfort { namespace cxxostd {
	using ::std::bit_width;
	using ::std::bit_floor;
	using ::std::bit_ceil;

} } 

#elif (CXXOMFORT_CXX_STD>=2011)
	#include <limits>
	#include <type_traits>

namespace cxxomfort { namespace cxxostd {

// 1 + floor(log2(x))
template <typename U, typename = typename std::enable_if< std::is_unsigned<U>::value > >
CXXO_CONSTEXPR 
U 	bit_width (U x) CXXO_NOEXCEPT {
	return (x==U(0U)) ? 0 : 1 + ::cxxomfort::impl::intlog(x);
}

template <typename U, typename = typename std::enable_if< std::is_unsigned<U>::value > >
CXXO_CONSTEXPR 
U 	bit_floor (U const u) CXXO_NOEXCEPT {
	return (u == 0) ? 0 : (1U<<(std::numeric_limits<U>::digits - 1 - ::cxxomfort::bits::cntl0u(u)) );
}

// from Hacker's Delight Ch.3
template <typename U, typename = typename std::enable_if< std::is_unsigned<U>::value > >
CXXO_CONSTEXPR 
U 	bit_ceil (U const u) CXXO_NOEXCEPT {
	return (u < 2U) ? 1 : (1U<<(std::numeric_limits<U>::digits - ::cxxomfort::bits::cntl0u(u-1)) );
}

} }

namespace std {
	using ::cxxomfort::cxxostd::bit_floor;
	using ::cxxomfort::cxxostd::bit_ceil;
	using ::cxxomfort::cxxostd::bit_width;
}

#else


namespace std {

template <typename U>
U 	bit_floor (U const u) {
	return (u == 0) ? 0 : (1U<<(std::numeric_limits<U>::digits - 1 - ::cxxomfort::bits::cntl0u(u)) );
}

template <typename U>
U 	bit_ceil (U const u) CXXO_NOEXCEPT {
	return (u < 2U) ? 1 : (1U<<(std::numeric_limits<U>::digits - ::cxxomfort::bits::cntl0u(u-1)) );
}


}


#endif // c++

// 
// popcount, has_single_bit (C++20)
// 
// here we can leverage __builtin_popcount intrinsic


#if defined(DOXYGEN_DOC)
#elif defined(__cpp_lib_bitops)
	#define CXXOMFORT_IMPLEMENTS_bit_popcount CXXO_IMPLSTATUS_NATIVE()
namespace cxxomfort { namespace cxxostd {
	using ::std::popcount;
} }
#elif (defined(__builtin_popcount))
	#define CXXOMFORT_IMPLEMENTS_bit_popcount CXXO_IMPLSTATUS_BACKPORT()

namespace cxxomfort { namespace cxxostd {

/**
 * @cxx20b Bit operations: return the number of 1-bit set in @a u .
 * @xrefitem std0bit "" "" Bit operations: return the number of 1-bit set in @a u .
 * @pre <tt>std::is_unsigned<U>::value</tt> 
 * */
template <typename U>
constexpr int popcount (U u) noexcept {};

template <typename U>
CXXO_CONSTEXPR int popcount (U u
, typename std::enable_if< std::is_unsigned<U>::value, bool* >::type = 0
) CXXO_NOEXCEPT ;

template<>
CXXO_CONSTEXPR int 
popcount<unsigned> (unsigned u) CXXO_NOEXCEPT { return __builtin_popcount(u); }

template<>
CXXO_CONSTEXPR int 
popcount<unsigned long> (unsigned long u) CXXO_NOEXCEPT { return __builtin_popcountl(u); }

template<>
CXXO_CONSTEXPR int 
popcount<unsigned long long> (unsigned long long u) CXXO_NOEXCEPT { return __builtin_popcountll(u); }

template <typename U>
CXXO_CONSTEXPR bool has_single_bit (U u) CXXO_NOEXCEPT {
	return popcount(u)==1;
}

} }

namespace std {
	using ::cxxomfort::cxxostd::popcount;
	using ::cxxomfort::cxxostd::has_single_bit;
}

#else
	#include <cxxomfort/type_traits.hpp>
	#define CXXOMFORT_IMPLEMENTS_bit_popcount CXXO_IMPLSTATUS_EMULATION()

namespace cxxomfort { namespace cxxostd {

template <typename U>
CXXO_CONSTEXPR14 int popcount (U u
, typename std::enable_if< std::is_unsigned<U>::value, bool* >::type = 0
) CXXO_NOEXCEPT {
	uint16_t const qu_bits[16] 
	= {0, 1, 1, 2, 1, 2, 2, 3, 
		1, 2, 2, 3, 2, 3, 3, 4
	};
	return (u == U(0) ) ? 0 : (static_cast<int>(qu_bits[u % 16]) + popcount(u>>4));
}

template <typename U>
CXXO_CONSTEXPR bool 
has_single_bit (U u) CXXO_NOEXCEPT {
	return u != 0 && (u & (u - 1)) == 0;
}

} }

namespace std {
	using ::cxxomfort::cxxostd::popcount;
	using ::cxxomfort::cxxostd::has_single_bit;
}

#endif

// 
// countl_zero, countr_zero
// 

#if (defined(DOXYGEN_DOC) )

#elif defined(__cpp_lib_bitops)
namespace cxxomfort { namespace cxxostd {
	using ::std::countl_zero;
	using ::std::countr_zero;
}

#elif (CXXOMFORT_CXX_STD >= 1997)

namespace cxxomfort { namespace cxxostd {

template <typename U> CXXO_CONSTEXPR inline 
int countl_zero (U u) CXXO_NOEXCEPTNOTHROW {
	return ::cxxomfort::bits::cntl0u(u);
}

// we use "mask for trailing zeros", from Hacker's Delight
template <typename U> CXXO_CONSTEXPR inline 
int countr_zero (U u) CXXO_NOEXCEPTNOTHROW {
	return ::cxxomfort::bits::cntt0u(u);
}

}}

namespace std {
	using ::cxxomfort::cxxostd::countl_zero;
	using ::cxxomfort::cxxostd::countr_zero;
}

#endif

// 
// bit_cast (C++20)
// 

#if defined(DOXYGEN_DOC)
namespace std {

/**
 * @brief Reinterpret the bit pattern content of an object (C++20) - 
 * @ingroup std0bit
 * @cxx20b Reinterpret the bit pattern of an object as an object of a different type.
 * @xrefitem std0bit "" "" Reinterpret the bit pattern of a type.
 * @see @cppref{numeric/bit_cast}
 * 
 * @pre @a From , @a To need both to be trivially copyable.
 * 
 * @pre In C++<14 in cxxomfort's implementation, @a To needs to also have a trivial default constructor.
 * 
 * */
template <typename To, typename From>
constexpr To bit_cast (From const& from);

}
#elif CXXO_COMPILER_SUPPORT_bit_cast
	#define CXXOMFORT_IMPLEMENTS_p0476 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILER_SUPPORT_constexpr_relaxed)
	#define CXXOMFORT_IMPLEMENTS_p0476 CXXO_IMPLSTATUS_BACKPORT()
	#define __cpp_lib_bit_cast 201401
	// compilers with constexpr relaxed would handle memcpy

	#include <type_traits>

namespace std {

template <class To, class From>
CXXO_CONSTEXPR14 
typename std::enable_if<
    sizeof(To) == sizeof(From) &&
    std::is_trivially_copy_constructible<From>::value &&
    std::is_trivially_copy_constructible<To>::value,
    To>::type
bit_cast(const From& src) noexcept {
    static_assert(std::is_trivially_constructible<To>::value,
        "This implementation additionally requires destination type to be trivially constructible");
    To dst;
    std::memcpy(&dst, &src, sizeof(To));
    return dst;
}

} // std

#elif (CXXOMFORT_CXX_STD >= 2011)
	#define CXXOMFORT_IMPLEMENTS_p0476 CXXO_IMPLSTATUS_EMULATION()

	#include <cxxomfort/type_traits.hpp>

namespace std {

template <class To, class From>
typename std::enable_if<
    sizeof(To) == sizeof(From) &&
    std::is_trivially_copy_constructible<From>::value &&
    std::is_trivially_copy_constructible<To>::value && 
    std::is_trivially_default_constructible<To>::value 
    , To>::type
bit_cast(const From& src) CXXO_NOEXCEPTNOTHROW {
    static_assert(std::is_trivially_default_constructible<To>::value,
        "This implementation additionally requires destination type to be trivially constructible");
    To dst;
    std::memcpy(&dst, &src, sizeof(To));
    return dst;
}

} // std

#else
	#define CXXOMFORT_IMPLEMENTS_p0476 0

#endif

#define CXXOMFORT_IMPLEMENTS_bit_cast CXXOMFORT_IMPLEMENTS_p0476


// 


#endif


/**
 * @page std0bit <bit>
 * @brief Backports related to Standard header <tt><bit></tt>
 * 
 * @code
 * #include <cxxomfort/bit.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces 
 * 
 * The following names here are also available in namespace @c std:
 * 
 * * <tt>std::endian</tt>
 * * @c bit_ceil() , @c bit_floor() - round to powers of two.
 * * @c popcount() - find the number of @c 1 bits in the given value.
 * * @c has_single_bit() - checks if number is a power of two.
 * * @c bit_width() — number of bits needed to fully represent an integer.
 * * @c bit_cast() (C++20) (down to C++11) - reinterpret bit pattern of an object.
 * 
 * The following interfaces used internally are also made available in namespace @c cxxomfort::fix :
 * 
 * * 'bit_masks' — common compile-time bitmasks for unsigned operations.
 * 
 * The following interfaces are made available in namespace @c cxxomfort:: .
 * 
 * * @c pos_l_1() , @c pos_r_1() — position of the (left-most, right-most) 1-bit in a number.
 * * @c mask_r_0() — bitmask for the rightmost sequential 0-bits in a number.
 * * 'count_l_0_t' — count of leading 0-bits in a compile-time integer.
 * * 'count_r_0_t' — count of trailing 0-bits in a compile-time integer.
 * 
 * @section also See Also
 * 
 * @see @cppref{header/bit} (cppreference) 
 * @see @ref cxxo-sup-bit 
 * 
 * */
