#ifndef CXXOMFORT_COMPLEX_HPP
#define CXXOMFORT_COMPLEX_HPP
/**
 * @file
 * @brief Backports for features in <tt>complex</tt>.
 * @xrefitem std0complex "<complex>" ""
 * @ingroup std0complex
 *
 */

#include <cxxomfort/base.hpp>
#include <complex>
#include <cxxomfort/utils/traits_basic.hpp>

// 
// per-compiler config updates 
// 

#if 0
#elif (!defined(__cpp_lib_complex_udls) && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1900) )
	#define __cpp_lib_complex_udls 1
#endif

// 
// i_c helper
// 

namespace cxxomfort { namespace fix {
CXXOMFORT_CXX11_CODE(inline, /* */) namespace complex_helpers {

#if (defined(DOXYGEN_DOC))
#elif (CXXOMFORT_CXX_STD >= 2014 || defined(__cpp_lib_complex_udls))
template <typename Floating=double
, typename = typename ::cxxomfort::traits::enable_if<::cxxomfort::traits::is_floating_point<Floating>::value>
>
struct icomplex_t {
	constexpr std::complex<Floating> operator() (Floating f) const noexcept {
		return std::complex<Floating>(0.0, f);
	}
};
constexpr const icomplex_t<> i_c {};
constexpr const icomplex_t<float> if_c {};
constexpr const icomplex_t<long double> il_c {};


#elif (CXXOMFORT_CXX_STD >= 2011)

template <typename Floating=double>
struct icomplex_t {
	constexpr std::complex<Floating> operator() (Floating f) const noexcept {
		return std::complex<Floating>(0.0, f);
	}
};
constexpr const icomplex_t<> i_c {};
constexpr const icomplex_t<float> if_c {};
constexpr const icomplex_t<long double> il_c {};

#else

template <typename Floating=double>
struct icomplex_t {
	std::complex<Floating> operator() (Floating f) const CXXO_NOEXCEPTNOTHROW {
		return std::complex<Floating>(0.0, f);
	}
};
const icomplex_t<> i_c = {};
const icomplex_t<float> if_c = {};
const icomplex_t<long double> il_c = {};

#endif

} // complex_helpers 
} } // fix

// 
// i UDL, if UDL, il UDL literals
// 

#if defined(DOXYGEN_DOC)
	/**
	 * @ingroup std00featuremacros
	 * @cppfeature{<tt>""i</tt> user-defined literal and similar for <tt>std::complex</tt>}
	 * @xrefitem std0complex "" "" Feature Testing macros.
	 * */
	#define __cpp_lib_complex_udls YYYY00L
#elif defined(__cpp_lib_complex_udls)
	// nothing to do
#elif defined(CXXOFLAG_NO_STD_USING)
	// nothing to do
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG,<800) && CXXOMFORT_CXX_STD>=2011)
	// Clang++ chokes on operator ""if due to lexer, presumably
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_GCC,<800) && CXXOMFORT_CXX_STD==2011)
	// GCC chokes on operator ""if due to lexer, presumably
#elif (defined(__cpp_user_defined_literals) && CXXO_COMPILER_SUPPORT_udl_lexer_nospace>0)
	#define __cpp_lib_complex_udls (CXXOMFORT_CXX_STD*100)

namespace std { CXXOMFORT_CXX11_CODE(inline, /* */) namespace  literals { namespace complex_literals {

	constexpr std::complex<double> operator ""i (long double ix) {
		return std::complex<double>(0.0, ix);
	}
	constexpr std::complex<double> operator ""i (unsigned long long ix) {
		return std::complex<double>(0.0, ix);
	}

	constexpr std::complex<long double> operator ""il (long double ix) {
		return std::complex<long double>(0.0, ix);
	}
	constexpr std::complex<long double> operator ""il (unsigned long long ix) {
		return std::complex<long double>(0.0, ix);
	}

	#if (CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG,<600) && CXXOMFORT_CXX_STD>=2011)
	// Clang chokes on operator ""if due to lexer using a reserved word, presumably
	#else
	constexpr std::complex<float> operator ""if (long double ix) {
		return std::complex<float>(0.0, ix);
	}
	constexpr std::complex<float> operator ""if (unsigned long long ix) {
		return std::complex<float>(0.0, ix);
	}
	#endif
	
} // complex_literals

	using complex_literals::operator ""i;
	using complex_literals::operator ""il;

}} // std::literals
#endif

// 
// ""_i helpers, always unambiguous
// 

#if (defined(__cpp_user_defined_literals) && CXXO_COMPILER_SUPPORT_udl_lexer_nospace>0)

namespace cxxomfort { namespace fix {
CXXOMFORT_CXX11_CODE(inline, ) namespace complex_literals {

	CXXO_CONSTEXPR std::complex<double> operator ""_i (long double ix) {
		return std::complex<double>(0.0, ix);
	}
	CXXO_CONSTEXPR std::complex<double> operator ""_i (unsigned long long ix) {
		return std::complex<double>(0.0, ix);
	}

	CXXO_CONSTEXPR std::complex<long double> operator ""_il (long double ix) {
		return std::complex<long double>(0.0, ix);
	}
	CXXO_CONSTEXPR std::complex<long double> operator ""_il (unsigned long long ix) {
		return std::complex<long double>(0.0, ix);
	}

	CXXO_CONSTEXPR std::complex<float> operator ""_if (long double ix) {
		return std::complex<float>(0.0, ix);
	}
	CXXO_CONSTEXPR std::complex<float> operator ""_if (unsigned long long ix) {
		return std::complex<float>(0.0, ix);
	}

} // complex_literals

} }

#endif

#endif


/**
 * @page std0complex <complex>
 * @brief Backports related to Standard header <tt><complex></tt>
 * 
 * @code
 * #include <cxxomfort/complex.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Interfaces defined in this section:
 * 
 * * @ref icomplex_literal - backports <tt>""i</tt> family of user-defined literals to C++11.
 * * The @ref __cpp_lib_complex_udls macro if the above UDL is implemented.
 * 
 * Non-backport interfaces:
 * 
 * * @ref icomplex_helper "Complex helper" , for systems where UDL is not available.
 * 
 * @section icomplex_literal  ""i family of UDL for C++14
 * 
 * C++14 adds the family of UDL-literals <tt>""i, ""if, ""il</tt> , which allows literal construction of <tt>std::complex<></tt> values. However, in some compilers in modes previous to C++14, in particular in Clang, this is a breaking change since those compilers <a href="https://gcc.gnu.org/onlinedocs/gcc/Complex.html">implement the literal extension</a> <tt>i</tt> (as well as others) to construct built-in complex numbers such as Clang's <tt>_Complex double</tt>, etc.
 * 
 * Furthermore, <tt>""if</tt> causes issues in some compilers without full C++14 support because <tt>if</tt> is a reserved word.
 * 
 * In order to clean up this behaviour as well as provide a C++11-compatible UDL, cxxomfort adds the UDL family <tt>""_i</tt> (note the underscore), which is allowed at user-defined level and is unambiguous.
 * 
 * @code 
 * #include <cxxomfort/complex.hpp>
 * 
 * using namespace std::literals;
 * auto cc1 = 2.3-3.2i; 
 * // ^ ambiguous pre-C++14 in most systems; 
 * // cc1 can be std::complex<...> or _Complex..., etc.
 * using namespace cxxomfort::literals; // or cxxomfort::fix::complex_literals 
 * auto cc2 = 2.3-3.2_i; 
 * // unambiguous: cc2 is std::complex<double> everywhere it is supported
 * @endcode
 * 
 * @section std0complex_also See Also
 * 
 * @see @cppref{header/complex} (cppreference) 
 * 
 * */

