#if (defined(CXXOMFORT_BASE_HPP) || defined(CXXOMFORT_MINIMAL_CONFIG))
#else
    #error "Include cxxomfort's headers to include this header."
#endif
#ifndef CXXOMFORT_CONFIG_HPP
#define CXXOMFORT_CONFIG_HPP
//! @file 
//! @brief Starting configuration file.

//! @ingroup cxxomfort-macros
//! @xrefitem cxxomfort-macros "Informative Macros" "" Library version and release date as @c YYYYMMDD .
#define CXXOMFORT_DATE 20221208UL

// License Info
// SPDX-License-Identifier: LGPL-3.0-or-later

#include "config/_pre.hpp"
#include "config/_support.hpp"


#if defined(DOXYGEN_DOC)
/**
 * @def CXXOFLAG_NO_STD_USING
 * @brief Disables adding names to <code>namespace std</code>
 * @ingroup cxxomfort-macros-behaviour
 * @xrefitem cxxomfort-macros-behaviour "" "" Prevents adding some names to namespace std.
 * 
 * If this macro is defined and has a value other than @c 0 ,
 * cxxomfort headers will where possible avoid adding the backported features to <tt>namespace std</tt>.
 * 
 * @warning Using this macro is unsupported and provided only to fix potential compatibility issues. The @c entire point of the cxxomfort library is that it adds stuff to namespace @c std since there is stuff that depends on it (eg.: anything that used <tt>std::enable_if</tt>).
 * 
 */
	#define CXXOFLAG_NO_STD_USING 0

/**
 * @def CXXOFLAG_USE_INTERNAL_TR1
 * @brief Declare this macro to tell the library to use its own implementation of TR1 assets.
 * @ingroup cxxomfort-macros-behaviour
 * @xrefitem cxxomfort-macros-behaviour "" "" Make the library provide its own TR1 on older compilers.
 * 
 * This macro is intended to be used with pre-C++03 compilers that don't implement C++03 TR1, and is ignored when the compiler supports C++11 or above.
 * 
 * See also: @ref std0array , @ref std0functional , @ref std0random , @ref std0tuple , @ref std0utility .
 * 
 */
	#define CXXOFLAG_USE_INTERNAL_TR1 0

/**
 * @def CXXOFLAG_WITHOUT_LONGLONG
 * @brief Declare this macro to tell the library that the compiler does not support <tt>long long</tt>.
 * @ingroup cxxomfort-macros-behaviour
 * @xrefitem cxxomfort-macros-behaviour "" "" Disable use of overloads and definitions that require <tt>long long</tt> where possible.
 * 
 * @warning Compiler configurations without <tt>long long</tt> are @e not supported.
 * 
 */
	#define CXXOFLAG_WITHOUT_LONGLONG 0

/**
 * @brief Makes the reserved name @c byte a typedef.
 * @ingroup cxxomfort-macros-behaviour
 * @xrefitem cxxomfort-macros-behaviour "" "" Make @c std::byte be an alias for <tt>unsigned char</tt>.
 * 
 * When pre- C++17 mode, if this macro is enabled, <tt><cxxomfort/cstddef.hpp></tt> will set up the name @c byte (<tt>std::byte</tt>) as a typedef to the type <tt>unsigned char</tt>, instead of using one of the backport / emulation mechanisms made available to cxxomfort.<br/>The resulting code is not necessarily supported (eg.: ambiguous overloads).
 * 
 * */
	#define CXXOFLAG_MAKE_TYPEDEF_byte 0

#endif // doxygen

//! @xrefitem cxxomfort-codegen "" "" Enables a code block if the compiler matches @c CXXO_COMPILER_ID @a com_ and version compares with @a ifver_ .
#define CXXO_COMPILERREQ(com_,ifver_) ((CXXOMFORT_COMPILER_ID == com_ && CXXOMFORT_COMPILER_VERSION ifver_))

#define CXXO_NOTICE_IMPL_(msg) (msg)

//! Emits a notice at compilation time.
//! @ingroup cxxomfort-behaviour
#define CXXO_NOTICE(msg) CXXO_NOTICE_IMPL_(  __FILE__ "(" CXXO_STRINGIZE1(__LINE__) ") : cxxomfort notice: " msg)
//! Emits a warning at compilation time.
//! @ingroup cxxomfort-behaviour
#define CXXO_WARNING(msg) CXXO_NOTICE_IMPL_( __FILE__ "(" CXXO_STRINGIZE1(__LINE__) ") : cxxomfort warning: " msg)


#include <cstddef>

/*
 * Detect compilers and apply specific configurations.
 */
#if 0
#elif (defined(__clang_major__) && defined(__clang_minor__))
    #include "./config/clang.hpp"
#elif defined(__DJGPP__)
	#include "./config/djgpp.hpp"
#elif (defined(__GNUC__) && defined(__GNUC_MINOR__))
    #include "./config/gcc.hpp"
#elif defined(_MSC_VER)
    #include "./config/msc.hpp"
#elif defined(__DMC__)
    #include "./config/dmc.hpp"
#elif defined(__WATCOMC__)
    #include "./config/ow.hpp"
#else

// this is an unknown / unsupported compiler, assume things just work™.
    #define CXXOMFORT_COMPILER_ID 0
    #define CXXOMFORT_COMPILER_VERSION 0
    #define CXXOMFORT_COMPILER_VERSION_FULL 0

    #define CXXO_NOTICE(msg) _Pragma(message ("WARNING - " msg))
    #if defined(CXXOMFORT_NOTICES)
        #pragma message CXXO_WARNING("Found an undetermined compiler.")
    #endif
#endif

#if (defined(CXXOMFORT_NOTICES))
    #pragma message ("compiler : " \
    CXXO_STRINGIZE1(CXXOMFORT_COMPILER_ID) ":" \
    CXXO_STRINGIZE1(CXXOMFORT_COMPILER_VERSION)  \
    " cxxo ver:" CXXO_STRINGIZE1(CXXOMFORT_DATE) \
    " cxx:" CXXO_STRINGIZE1(CXXOMFORT_CXX_STD) ":" CXXO_STRINGIZE1(CXXOMFORT_CXX_EMULATION) \
    )

/*
 * Set up more friendly messages
 */
    #if (CXXOMFORT_NOTICES > 1)

        #if 0
        #elif (CXXOMFORT_CXX_STD >= 2026)
			#pragma message CXXO_NOTICE("C++26 or newer")
        #elif (CXXOMFORT_CXX_STD >= 2023)
            #pragma message CXXO_NOTICE("C++23 mode");
        #elif (CXXOMFORT_CXX_STD >= 2020)
            #pragma message CXXO_NOTICE("C++20 mode");
        #elif (CXXOMFORT_CXX_STD >= 2017)
            #pragma message CXXO_NOTICE("C++17 mode");
        #elif (CXXOMFORT_CXX_STD >= 2014)
            #pragma message CXXO_NOTICE("C++14 mode");
        #elif (CXXOMFORT_CXX_STD >= 2011)
            #pragma message CXXO_NOTICE("C++11 mode");
        #elif (CXXOMFORT_CXX_STD >=1997 && CXXOMFORT_CXX_STD < 2011)
            #pragma message CXXO_NOTICE("C++03 or basic mode");
        #else
            #pragma message CXXO_NOTICE("detected C++ not at least C++03; expect hard compilation fail!")
        #endif

        #if 0
        #elif (CXXOMFORT_CXX_EMULATION == 2007)
            #pragma message CXXO_NOTICE("emulation / c++98+tr1 mode (2007)")
        #elif (CXXOMFORT_CXX_EMULATION == 2011) 
            #pragma message CXXO_NOTICE("emulation / c++0x mode (c++11)")
        #elif (CXXOMFORT_CXX_EMULATION == 2014)
            #pragma message CXXO_NOTICE("emulation / c++1y mode (c++14)")
        #elif (CXXOMFORT_CXX_EMULATION >= 2017)
            #pragma message CXXO_NOTICE("emulation / c++1z mode (c++17)")
        #elif (CXXOMFORT_CXX_EMULATION >= 2020)
            #pragma message CXXO_NOTICE("emulation / c++2x mode (c++20)")
        #elif (CXXOMFORT_CXX_EMULATION >= 2023)
            #pragma message CXXO_NOTICE("emulation / c++2y mode (c++23)")
        #elif (CXXOMFORT_CXX_EMULATION >= 2026)
			#pragma message CXXO_NOTICE("emulation / c++2z mode (c++26)")
        #endif
        
    #endif // friendly notices

#endif

// 
// cxxomfort independent implementation bases 
// 

#if (CXXO_COMPILER_SUPPORT_decltype || defined(__cpp_decltype))
	#undef  CXXOMFORT_IMPLEMENTS_n1978	
	#define CXXOMFORT_IMPLEMENTS_n1978 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXO_COMPILER_SUPPORT_typeof)
	#undef  CXXOMFORT_IMPLEMENTS_n1978	
	#define CXXOMFORT_IMPLEMENTS_n1978 CXXO_IMPLSTATUS_EMULATION()
#else
	#undef  CXXOMFORT_IMPLEMENTS_n1978	
	#define CXXOMFORT_IMPLEMENTS_n1978 0
#endif

#if (CXXOMFORT_CXX_STD>=2011 || defined(__cpp_attributes))
	#undef  CXXOMFORT_IMPLEMENTS_attribute
	#define CXXOMFORT_IMPLEMENTS_attribute CXXO_IMPLSTATUS_NATIVE()
#endif


//
// Macros that conditionally generate code for C++11, C++14, C++17
//

#if (defined(DOXYGEN_DOC))
	//! Evaluates to @a code1 if at least in C++11 mode, @a code2 otherwise
	//! @ingroup cxxomfort-codegen
	#define CXXOMFORT_CXX11_CODE( code1 , code2 ) 
	//! Evaluates to @a code1 if at least in C++14 mode, @a code2 otherwise
	//! @ingroup cxxomfort-codegen
	#define CXXOMFORT_CXX14_CODE( code1 , code2 ) 
	//! Evaluates to @a code1 if at least in C++17 mode, @a code2 otherwise
	//! @ingroup cxxomfort-codegen
	#define CXXOMFORT_CXX17_CODE( code1 , code2 ) 
#elif (CXXOMFORT_CXX_STD < 2011)
    #define CXXOMFORT_CXX11_CODE(cxx11,none) none
    #define CXXOMFORT_CXX14_CODE(cxx14,none) none
    #define CXXOMFORT_CXX17_CODE(cxx17,none) none

#elif (CXXOMFORT_CXX_STD < 2014)
    #define CXXOMFORT_CXX11_CODE(cxx11,none) cxx11
    #define CXXOMFORT_CXX14_CODE(cxx14,none) none
    #define CXXOMFORT_CXX17_CODE(cxx17,none) none

#elif (CXXOMFORT_CXX_STD < 2017)
    #define CXXOMFORT_CXX11_CODE(cxx11,none) cxx11
    #define CXXOMFORT_CXX14_CODE(cxx14,none) cxx14
    #define CXXOMFORT_CXX17_CODE(cxx17,none) none

#elif (CXXOMFORT_CXX_STD < 2020)
    #define CXXOMFORT_CXX11_CODE(cxx11,none) cxx11
    #define CXXOMFORT_CXX14_CODE(cxx14,none) cxx14
    #define CXXOMFORT_CXX17_CODE(cxx17,none) cxx17

#endif

//
// ---------------------------------------------------------
//

// check for exceptions 
#if (defined(__cpp_exceptions) || defined(_CPPUNWIND) || defined(__EXCEPTIONS) )
	#undef  CXXOMFORT_EXCEPTIONS_DISABLED
#else
	#define CXXOMFORT_EXCEPTIONS_DISABLED
#endif

// Macros to conditionally generate keywords

#if (defined(__cpp_constexpr))
    #define CXXO_CONSTEXPR constexpr
#else
    #define CXXO_CONSTEXPR 
#endif

#if (CXXOMFORT_CXX_STD >= 2014 || (defined(__cpp_constexpr) && __cpp_constexpr >= 201304L))
    #define CXXO_CONSTEXPR14 constexpr
#else
    #define CXXO_CONSTEXPR14 
#endif

#if (CXXOMFORT_CXX_STD >= 2017 && CXXO_COMPILER_SUPPORT_constexpr_relaxed > 0)
    #define CXXO_CONSTEXPR17 constexpr 
#else
    #define CXXO_CONSTEXPR17 
#endif

#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_noexcept>0)
    #define CXXO_NOEXCEPT noexcept
#else
    #define CXXO_NOEXCEPT
#endif

/**
 * @def CXXO_NOEXCEPTNOTHROW
 * @brief Evaluates to either @c noexcept  or @c throw()  when available.
 * */
#if (defined(CXXOMFORT_EXCEPTIONS_DISABLED))
    #define CXXO_NOEXCEPTNOTHROW
#elif (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_noexcept > 0)
    #define CXXO_NOEXCEPTNOTHROW noexcept
#else
    #define CXXO_NOEXCEPTNOTHROW throw() 
#endif

//
// CXXO_NOEXCEPT_COND(...)
// Evaluates to the noexcept of the condition in C++11 onwards, and to nothing in C++03
#if (defined(CXXOMFORT_EXCEPTIONS_DISABLED))
    #define CXXO_NOEXCEPT_COND(...)
#elif (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_noexcept > 0)
    #define CXXO_NOEXCEPT_COND(...) noexcept(__VA_ARGS__)
#else
    #define CXXO_NOEXCEPT_COND(...)
#endif

#if (!defined(CXXO_NOTHROW))
	#if (CXXOMFORT_CXX_STD >= 2020)
		#define CXXO_NOTHROW() [[nothrow]] 
	#else
		#define CXXO_NOTHROW() __attribute__((nothrow))
	#endif
#endif

/*
 * Check flags
 * */

#ifdef CXXOFLAG_NO_INITIALIZERS
	#if (CXXOMFORT_NOTICES > 1)
		#pragma message CXXO_NOTICE("compiler missing support for some forms of initializers. Compilation might fail.")
	#endif
#endif


// 
// Code Generation
// 


//
// ---------------------------------------------------------
//

#include "config/_codegen.hpp"
#include "config/_has.hpp"
#define _CXXOMFORT_IMPLEMENTS(nn) (( CXXO_JOIN(CXXOMFORT_IMPLEMENTS_n,nn) > 0))
#define CXXOMFORT_IMPLEMENTS(nn) _CXXOMFORT_IMPLEMENTS(nn)

// this command in a shell reveals all nxxxx proposals implemented by this library:
// find -type f -iname '*.hpp' | xargs egrep -o "CXXOMFORT_IMPLEMENTS_n([0-9]+)\b"

/*
 * Utility macro to declare variadic (pseudovariadic in C++03) template members
 */
//#include "impl/pseudovariadic.hpp"


//
// ---------------------------------------------------------
//

#include <cstdio>

//! The default namespace of C++ Standard stuff.
namespace std {}
//! Namespace of the @c cxxomfort library.
namespace cxxomfort {
/**
 * @brief Backports made available by cxxomfort.
 * 
 * Namespace that holds the library's backports of named C++ features.
 * 
 * */
namespace cxxostd {}
/** 
 * @brief Fixes for implementation issues in std
 * 
 * Namespace that holds fixes and alternative implementations or names 
 * for stuff in C++ std that can not be modified or reimplemented.
 * 
 * See @ref cxxomfort-fixes for more details.
 * 
 * @ingroup cxxomfort-fixes
 * 
 * */
namespace fix {}

/**
 * @brief Supplements to backports and other utilities.
 * 
 * Namespace that holds the library's own implementations, emulations and supplements.
 * 
 * @ingroup cxxo-sup
 * */
namespace library {}

//! Information about the library support
//! @ingroup cxxomfort-config
struct info { 
    enum {
    //! library version
    version = CXXOMFORT_DATE , 
    date = CXXOMFORT_DATE, 
    //! compiler identification code
    compilerid = CXXOMFORT_COMPILER_ID ,
    //! compiler version code (usually <code>100*major+minor</code>)
    compilerversion = static_cast<unsigned>(CXXOMFORT_COMPILER_VERSION) ,
    compilerversionfull = static_cast<unsigned long>(CXXOMFORT_COMPILER_VERSION_FULL) ,
    //! value of the standard
    cxx_std = static_cast<unsigned>(CXXOMFORT_CXX_STD) ,
    //! level of support of C++11 features in C++11 emulation / c++0x mode
    cxx_emulation = static_cast<unsigned>(CXXOMFORT_CXX_EMULATION) , 
    cplusplus = __cplusplus
    };
    
    /* struct */ 
}  ;

#if (CXXOMFORT_CXX_STD < 1997)
const struct info _;
#else
const struct info _ = {};
#endif

// output_info moved to cstdio.hpp


// ~::cxxomfort
}


// 
// tag_constexpr 
// tags fix functions that can run @c constexpr in C++11 .
// 

namespace cxxomfort { namespace fix {

struct tag_constexpr_t {} ;
#if (CXXOMFORT_CXX_STD >= 2011)
static constexpr tag_constexpr_t tag_constexpr = {};
#endif

} }

#endif
