#ifndef CXXOMFORT_CONFIG_HPP
    #error "This file should only be #included via <cxxomfort/config.hpp>"
#endif
#ifndef CXXOMFORT_CONFIG_CODEGEN_HPP
#define CXXOMFORT_CONFIG_CODEGEN_HPP
/**
 * @file
 */

/**
 * @page cxxomfort-codegen Code Generation Utilities
 * @brief Describes cxxomfort macros for code generation.
 * 
 * Cxxomfort includes a number of macros that help generate code. This includes in particular code constructs that can change across C++ Standard versions, such as @c noexcept conditionals, constructors definition, and sequence construction.
 * 
 * @section codegen-Parsing Parsing
 * 
 * A few macros for working with the preprocessor are available: 
 * 
 * * @ref @c CXXO_JOIN()  
 * * @ref @c CXXO_STRINGIZE()
 * * @ref @c CXXO_EXPAND()
 * * among others.
 * 
 * @section codegen-Types Types 
 * 
 * These macros are made available to generate and name types:
 * 
 * * @ref @c CXXO_TYPEOF() , @ref @c CXXO_DECLTYPE()
 * * @ref @c CXXO_AUTO()
 * * @ref @c CXXO_CONSTEXPR and related
 * * @ref @c CXXO_NOEXCEPT and related
 * 
 * * among others.
 * 
 * <tt>CXXO_TYPEOF(expr)</tt> evaluates to the type of the given @a expr expression if it can be found via a @c __typeof__ -like feature (C++<11) or @c decltype (C++>=11). <tt>CXXO_DECLTYPE(expr)</tt> also evaluates to the type, but @b only makes use of @c decltype , and fails if @c decltype is not available.
 * 
 * 
 * @section codegen-Functions Other Code Generation 
 * 
 * <b>Relational Operators</b>:
 * 
 * * <tt>CXXO_GENERATE_RELATIONALS()</tt>: automatically generate non-member 
 * relational operators <tt>> , >=, <=, !=</tt> for a class that already implements 
 * operators <tt> == , < </tt>. 
 * 
 * @code
 * class Type {
 *   bool operator== (Type const&) const;
 *   bool operator< (Type const&) const;
 * };
 * CXXO_GENERATE_RELATIONALS(Type);
 * // automatically generates the following:
 * // bool operator!= (Type const&, Type const&);
 * // bool operator<= (Type const&, Type const&);
 * // bool operator>= (Type const&, Type const&);
 * // bool operator> (Type const&, Type const&);
 * 
 * @endcode
 * 
 * <b>Typed enumerators</b>:
 * 
 * The dual macro <tt>CXXO_OPEN_ENUM_OR_ENUMCLASS(E,utype)</tt> and its corresponding @c _CLOSE form, create a type @c E that functions as a typed / scoped enumerator with underlying type @a utype . This resolves to a <tt>enum class</tt> if the compiler supports it, or a normal @c struct with type name @a E otherwise.
 * 
 * The dual macro <tt>CXXO_OPEN_TYPESAFE_ENUM(E,utype)</tt> and its corresponding @c _CLOSE form are similar to the above, except that if <tt>enum class</tt> is not available, they create @a E as a @c cxxomfort::typesafe_enum with underlying type @a utype instead. @c typesafe_enum is slightly better integrated with cxxomfort's libraries, as it also provides the accessory functions @c to_value() and @c to_enum() as well as relational operators for completeness.
 * 
 * */

#if (CXXOMFORT_CXX_STD >= 2011 || \
(CXXO_COMPILER_SUPPORT_default_delete && CXXO_COMPILER_SUPPORT_noexcept))
    #define CXXO_DEFAULT_DEFAULT_CONSTRUCTOR(Type) Type () = default
#else
    #define CXXO_DEFAULT_DEFAULT_CONSTRUCTOR(Type) Type () throw() {} 
#endif

#if (CXXOMFORT_CXX_STD >= 2011 || \
(CXXO_COMPILER_SUPPORT_default_delete && CXXO_COMPILER_SUPPORT_noexcept))
	#define CXXO_DELETED_DEFAUT_CONSTRUCTOR(Type) Type () noexcept = delete
#else
	#define CXXO_DELETED_DEFAUT_CONSTRUCTOR(Type) private: Type() 
#endif

#if (CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_default_delete)
    #define CXXO_DEFAULT_COPY_CONSTRUCTOR(Type, Su, ...) \
    Type (Type const& ) Su = default 
#else 
	//! @internal
    #define CXXO_DEFAULT_COPY_CONSTRUCTOR(Type, Su, ...) \
    Type (Type const& From) Su \
    : __VA_ARGS__ {} 
#endif


/**
 * @def CXXO_DELETED_COPY_CONSTRUCTOR
 * @brief Declares a "<tt>=delete</tt>" copy constructor.
 * @ingroup cxxomfort-codegen
 * @brief Announces a deleted (unavailable) copy constructor.
 * */
#if (CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_default_delete)
    #define CXXO_DELETED_COPY_CONSTRUCTOR(Type) Type (Type const&) = delete
    #define CXXO_DELETED_COPY_ASSIGN(Type) Type& operator= (Type const&) = delete
#else
    #define CXXO_DELETED_COPY_CONSTRUCTOR(Type) private: Type (Type const&);public: 
    #define CXXO_DELETED_COPY_ASSIGN(Type) private: Type& operator= (Type const&); public: 
#endif

/**
 * */
// uses macro CXXO_SIG_CONDITION, to be set at call site, 
// to create conditional template/function arguments for SFINAE
#if (CXXOMFORT_CXX_STD >= 2011)
    #define CXXO_ENABLEIF_TEMPLATE() \
    , typename = typename std::enable_if< CXXO_SIGCONDITION >::type 
    #define CXXO_ENABLEIF_FUNCTION() 
#else
    #define CXXO_ENABLEIF_TEMPLATE() 
    #define CXXO_ENABLEIF_FUNCTION() \
    , typename std::enable_if< CXXO_SIGCONDITION >::type * = nullptr 
    
#endif

#if defined(DOXYGEN_DOC)
/**
 * @brief Returns the type of an expression if a @c typeof -like feature is available.
 * @pre <tt>__typeof__</tt> or a similar intrinsic is available and has been detected (ie.: <tt>CXXO_COMPILER_SUPPORT_typeof > 0</tt>).
 * @xrefitem cxxomfort-codegen "" "Types" Obtains the type of an expression @a expr using a typeof-like feature.
 * @sa @ref CXXO_DECLTYPE
 * @sa @ref CXXO_TYPEOF_OR 
 * 
 * In C++>=11, this will expand to <tt>decltype(expr)</tt> and will be exactly equivalent to <tt>CXXO_DECLTYPE(expr)</tt>.
 * 
 * */
	#define CXXO_TYPEOF(expr) __typeof__(expr) [or equivalent]
#elif (CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_decltype)
    #undef  CXXO_TYPEOF
    #define CXXO_TYPEOF(...) decltype(__VA_ARGS__)
    #undef  CXXOMFORT_IMPLEMENTS_typeof
    #define CXXOMFORT_IMPLEMENTS_typeof CXXO_IMPLSTATUS_NATIVE()
#elif (CXXO_COMPILER_SUPPORT_typeof && !defined(CXXO_TYPEOF))
    #undef  CXXO_TYPEOF
    #define CXXO_TYPEOF(...) __typeof__(__VA_ARGS__)
    #undef  CXXOMFORT_IMPLEMENTS_typeof
    #define CXXOMFORT_IMPLEMENTS_typeof CXXO_IMPLSTATUS_NATIVE()
#elif (!defined(CXXO_TYPEOF))
    // a shot in the dark?
    #define CXXO_TYPEOF(expr) cxxo_typeof_not_available_t 
    #undef  CXXOMFORT_IMPLEMENTS_typeof
    #define CXXOMFORT_IMPLEMENTS_typeof 0
#endif

/**
 * @def CXXO_TYPEOF_OR
 * @brief Declared type of an expression, or a hint.
 * */
#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof)
	#define CXXO_TYPEOF_OR(T,E) CXXO_TYPEOF(E)
#else
	#define CXXO_TYPEOF_OR(T,E) T
#endif

#if (CXXO_COMPILER_SUPPORT_enum_class>0)
	#define CXXO_OPEN_ENUM_OR_ENUMCLASS(Name,U) enum class Name : U
	#define CXXO_CLOSE_ENUM_OR_ENUMCLASS(Name,U) 
#else
	#define CXXO_OPEN_ENUM_OR_ENUMCLASS(Name,U) enum Name
	#define CXXO_CLOSE_ENUM_OR_ENUMCLASS(Name,U) 
#endif

#if (CXXO_COMPILER_SUPPORT_enum_class>0)
	#define CXXO_OPEN_TYPESAFE_ENUM(Name, Underlying) enum class Name : Underlying 
	#define CXXO_CLOSE_TYPESAFE_ENUM(Name, Underlying) 
#else 
	#define CXXO_OPEN_TYPESAFE_ENUM(Name, Underlying) \
	namespace impl { struct CXXO_JOIN(enum_,Name) { enum type 
	#define CXXO_CLOSE_TYPESAFE_ENUM(Name, Underlying) \
	/*}*/; }; } /*impl*/ struct Name : public ::cxxomfort::typesafe_enum< CXXO_JOIN(impl::enum_,Name) , Underlying > { \
		typedef ::cxxomfort::typesafe_enum< CXXO_JOIN(impl::enum_,Name), Underlying > _Base; \
		Name () CXXO_NOEXCEPT : _Base() {} \
		Name (_Base::enum_type e) CXXO_NOEXCEPT : _Base(e) {} \
	}  
#endif


#if defined(DOXYGEN_DOC)
/**
 * @brief Returns the type of an expression, obtained using @c decltype .
 * @pre <tt>decltype</tt> is available (ie.: <tt>CXXO_COMPILER_SUPPORT_decltype > 0</tt>).
 * @xrefitem cxxomfort-codegen "" "Types" Obtains the type of an expression @a expr using C++>=11 @c decltype .
 * @sa @ref CXXO_TYPEOF
 * @sa @ref CXXO_TYPEOF_OR 
 * 
 * */
	#define CXXO_DECLTYPE(expr) decltype(expr)
#elif (CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_decltype)
    #undef  CXXO_DECLTYPE
    #define CXXO_DECLTYPE(...) decltype(__VA_ARGS__)
    #undef  CXXOMFORT_IMPLEMENTS_decltype
    #define CXXOMFORT_IMPLEMENTS_decltype CXXO_IMPLSTATUS_NATIVE()
#else
    // a shot in the dark?
    #define CXXO_DECLTYPE(expr) cxxo_decltype_not_available_t 
    #undef  CXXOMFORT_IMPLEMENTS_decltype
    #define CXXOMFORT_IMPLEMENTS_decltype 0
#endif

#if defined(DOXYGEN_DOC)
/**
 * @brief Defines a variable with an automatically computed type.
 * @pre C++11 @c auto is available, or at least @c CXXO_TYPEOF() is available.
 * @xrefitem cxxomfort-codegen "Types" "Codegen" Create a vaiable with an automatically computed type.
 * @sa @ref CXXO_TYPEOF() ,  @ref CXXO_DECLTYPE()
 * 
 * @warning In some cases, the precomputed type may differ beween @c auto (C++11) and typeof-like utilities.
 * 
 * */
	#define CXXO_AUTO(name, expr...) auto name = expr; [or equivalent]
#elif (CXXOMFORT_CXX_STD >= 2011)
	#define CXXO_AUTO(name, ...) auto name = __VA_ARGS__
#else
	#define CXXO_AUTO(name, ...) CXXO_TYPEOF(__VA_ARGS__) name = CXXO_EXPAND(__VA_ARGS__ )
#endif

/**
 * @def CXXO_AUTO_OR
 * @brief Declares a variable with the type of an expression, or a hinted type.
 * */
#if (CXXOMFORT_CXX_STD >= 2011)
	#define CXXO_AUTO_OR(Name, Expr, T) auto Name = Expr
#elif (CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof)
	#define CXXO_AUTO_OR(Name, Expr, T) CXXO_TYPEOF(Expr) Name = Expr 
#else
	#define CXXO_AUTO_OR(Name, Expr, T) T Name = Expr 
#endif

#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_override_final>0)
/**
 * @def CXXO_OVERRIDE override
 * @brief Expands to @c override in C++>=11.
 * @xrefitem cxxomfort-codegen Expands to @c override in C++>=11.
 * */
	#define CXXO_OVERRIDE override
#else
	#define CXXO_OVERRIDE 
#endif


#if defined(__cpp_exceptions)
	#define CXXO_THROW_STD(...) throw __VA_ARGS__
	#define CXXO_RETHROW_STD() throw
	#define CXXO_try() try 
	#define CXXO_catch(exc) catch (exc)
#else
	#define CXXO_THROW_STD(...) assert(false && CXXO_STRINGIZE(__VA_ARGS__))
	#define CXXO_RETHROW_STD() assert(false)
	#define CXXO_try() if (true)
	#define CXXO_catch(exc) if (false)
#endif

/*
#define CXXO_DEFAULT_DEFAULT_CONSTRUCTOR(Type) Type () = default [or equivalent]

#define CXXO_DEFAULT_COPY_CONSTRUCTOR(Type, Su, ...) \
    [if >=C++11] Type (Type const& ) Su = default  \
    [else] Type (Type const& From) Su  : __VA_ARGS__ {} 

#define CXXO_DELETED_DEFAUT_CONSTRUCTOR(Type) Type () noexcept = delete [or equivalent]

*/


/**
 * @def CXXO_DEFAULT_DEFAULT_CONSTRUCTOR
 * @brief Declares a “ <tt>=default</tt> ” default constructor.
 * @ingroup cxxomfort-codegen
 * @xrefitem cxxomfort-codegen "Codegen" "default-ctor"
 * 
 * Using this macro in the body of a class will invoke, where available, 
 * the <tt>=default</tt> default constructor. In pre-C++11, it will invoke 
 * a <tt>throw()</tt> empty constructor.
 * 
 * @code
 * class Foo {
 * 		CXXO_DEFAULT_DEFAULT_CONSTRUCTOR(Foo);
 * 		//In >= C++11 equivalent to:
 * 		Foo () noexcept = default;
 * 
 * };
 * @endcode
 * 
 * */

/**
 * @def CXXO_DEFAULT_COPY_CONSTRUCTOR
 * @brief Declares a <tt>=default</tt> copy constructor.
 * @ingroup cxxomfort-codegen
 * @param Su constructor suffix (eg.: noexcept)
 * @param ... Explicit constructor body (for C++03)
 * @xrefitem cxxomfort-codegen "Codegen" "copy-ctor"
 * 
 * Using this macro in the body of a class invokes, starting with C++11, 
 * a <tt>=default</tt> copy constructor. In C++03, it generates the code 
 * for a copy constructor using the subsequent arguments to the macro 
 * as code for the member initializer list, using the variable name 
 * <tt>From</tt> for the <b>argument</b> of the copy constructor.
 * 
 * @code
 * class Foo {
 * 		CXXO_DEFAULT_COPY_CONSTRUCTOR(Foo,  , member(From.member) );
 * 		//In >= C++11 equivalent to:
 * 		Foo (Foo const&)  = default;
 * 		//In < C++11 equivalent to:
 * 		Foo (Foo const& From) : member(From.member) {}
 * 
 * 		private:
 * 		float member;
 * 
 * };
 * @endcode
 * 
 * */


/**
 * @def CXXO_DELETED_DEFAULT_CONSTRUCTOR
 * @brief Declares a <tt>=delete</tt> default constructor.
 * @ingroup cxxomfort-codegen
 * @xrefitem cxxomfort-codegen "Codegen" "default-ctor"
 * 
 * Using this macro in the body of a class <tt>Foo</tt> 
 * declares the default constructor (<tt>Foo ();</tt>) 
 * as <tt>=delete</tt>-d or unavailable.
 * 
 * */


#if defined(DOXYGEN_DOC)
/**
 * @def CXXO_CONSTEXPR
 * @brief Evaluates to @c constexpr when it's available.
 * @ingroup cxxomfort-codegen
 * */
    #define CXXO_CONSTEXPR constexpr
/**
 * @def CXXO_CONSTEXPR14
 * @brief Evaluates to @c constexpr when it's available in its C++14 extended / "relaxed" form.
 * @ingroup cxxomfort-codegen
 * */
    #define CXXO_CONSTEXPR14 constexpr
/**
 * @def CXXO_NOEXCEPT
 * @brief Evaluates to @c noexcept when available.
 * @ingroup cxxomfort-codegen
 * 
 * */
    #define CXXO_NOEXCEPT noexcept
/**
 * @def CXXO_NOEXCEPT_COND
 * @brief Evaluates to @c noexcept(condition...) when available.
 * @ingroup cxxomfort-codegen
 * 
 * */
    #define CXXO_NOEXCEPT_COND(condition) noexcept(condition)
/**
 * @def CXXO_NOEXCEPTNOTHROW
 * @brief Evaluates to @c noexcept when available, and to an empty throw declaration (<tt>throw()</tt>) otherwise.
 * @ingroup cxxomfort-codegen
 * 
 * */
    #define CXXO_NOEXCEPTNOTHROW noexcept
/**
 * @def CXXO_EXPLICIT_COND
 * @brief Evaluates to <tt>explicit(condition...)</tt> when available.
 * @ingroup cxxomfort-codegen
 * 
 * */
    #define CXXO_EXPLICIT_COND(...) explicit(__VA_ARGS__)

#endif // doc

#endif



