#ifndef CXXOMFORT_CONFIG_HPP
    #error "This file must not be included directly (see cxxomfort/config.hpp)"
#endif
#ifndef CXXOMFORT_CONFIG_GCC_HPP
#define CXXOMFORT_CONFIG_GCC_HPP

/*
 * GCC version history:
 * https://www.gnu.org/software/gcc/releases.html
 * 
 * 5.5 - 2017-10
 * 7.2 - 2017-08
 * 6.4 - 2017-07
 * 5.4 - 2016-06
 * 6.0 - 2016-04
 * 5.0 - 2015-04
 * 4.8 - 2013-03 **preferred, in Stretch / Trusty**
 * 4.6 - 2011-06 **preferred, in Jessie / Trusty**
 * 4.4 - 2009-04
 * 4.2 - 2007-07
 * 
 * 
Here follows a list of C++11 features for GCC:

 * static_assert - 403
 * decltype - 403
 * variadic templates - 404
 * auto with new semantics - 404
 * char16_t, char32_t - 404
 * rvalue references - 404
 * std::initializer_list - 404
 * defaulted and deleted members - 404
 * scoped enums - 404
 * explicit operator T - 405
 * local types as template arguments - 405
 * for (elem : range) - 406
 * nullptr, std::nullptr_t - 406
 * user-defined literals - 407
 * alignment - 408
 * rvalue references for *this - 408
 * decltype for function calls - 408
 * Attributes - 408
 * 
 * See: 
 * http://gcc.gnu.org/projects/cxx0x.html
 * 
 * For correct usage of this library emulation mode
 * ("C++0x") should not be used below GCC 4.6; 
 * normal C++03 plus a TR1 implementation should be used instead.
 */

//
// setup identification information and pragmas
//

#define CXXOMFORT_COMPILER_ID CXXO_COMPILERID_GCC
#define CXXOMFORT_COMPILER_VERSION (__GNUC__*100UL+__GNUC_MINOR__)
#define CXXOMFORT_COMPILER_VERSION_FULL (__GNUC__*100000UL+__GNUC_MINOR__*1000UL+__GNUC_PATCHLEVEL__)

#undef  CXXO_COMPILER_SUPPORT_type_intrinsics
#define CXXO_COMPILER_SUPPORT_type_intrinsics 1

#if (CXXOMFORT_COMPILER_VERSION >= 404)
	#undef  CXXO_COMPILER_SUPPORT_expression_sfinae
	#define CXXO_COMPILER_SUPPORT_expression_sfinae 1
#endif

#if (CXXOMFORT_COMPILER_VERSION >= 408)
    #undef  CXXO_NOTICE_IMPL_ 
    #define CXXO_NOTICE_IMPL_(msg) msg
#else
    #undef  CXXO_NOTICE_IMPL_ 
    #define CXXO_NOTICE_IMPL_(msg) msg
    #undef  CXXO_NOTICE
    #define CXXO_NOTICE(msg) CXXO_NOTICE_IMPL_( "cxxomfort notice: " msg)
    #undef  CXXO_WARNING
    #define CXXO_WARNING(msg) CXXO_NOTICE_IMPL_( "cxxomfort warning: " msg)
#endif

#if defined(CXXOMFORT_NOTICES)
    #pragma message CXXO_NOTICE("Found the GNU C++ Compiler.")
#endif

#if (CXXOMFORT_NOTICES > 1 && defined(_LIBCPP))
	#pragma message CXXO_NOTICE("with libcxx")
#endif

/* 
 * GCC old versions define __cplusplus to be 1 but still have 
 * some good C++1997 and C++2003 support.
 * For those compilers, we set std and emulation to 1997.
 * 
 * */
#if (CXXOMFORT_COMPILER_VERSION >= 400 && CXXOMFORT_COMPILER_VERSION <= 406 \
&& CXXOMFORT_CXX_STD<10)
	#undef  CXXOMFORT_CXX_STD
	#define CXXOMFORT_CXX_STD 1997
	#undef  CXXOMFORT_CXX_EMULATION
	#define CXXOMFORT_CXX_EMULATION 1997
#endif

// GCC advertises C++11 mode with "-std=c++0x" and a macro
#if (CXXOMFORT_CXX_STD < 2011 && defined(__GXX_EXPERIMENTAL_CXX0X__))
    #undef  CXXOMFORT_CXX_EMULATION
    #define CXXOMFORT_CXX_EMULATION 2011
#endif

/**
 * GCC headers tend to be organized via a 
 * path with subfolders {bits,tr1}
 */
#if defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1)
    #pragma message( CXXO_NOTICE("Using bits/../tr1/../<version> include paths"))
#endif

// 
// general support found in GCC
// 

#undef  CXXO_COMPILER_SUPPORT___alignof
#define CXXO_COMPILER_SUPPORT___alignof 1
#undef  CXXO_COMPILER_SUPPORT_typeof
#define CXXO_COMPILER_SUPPORT_typeof 1
#undef  CXXO_COMPILER_SUPPORT_va_args
#define CXXO_COMPILER_SUPPORT_va_args 1

// 
// GCC in C++03 mode includes TR1 support 
// and headers in all the versions we support, so far.
// 

#if (CXXOMFORT_CXX_STD >= 1997 && CXXOMFORT_CXX_STD < 2011 && CXXOMFORT_CXX_EMULATION < 2011)
	#undef  CXXOMFORT_CXX_EMULATION
	#define CXXOMFORT_CXX_EMULATION 2007
	#undef  CXXOFLAG_tr1
	#define CXXOFLAG_tr1 1
#endif

// set up TR1 headers
#if (CXXOFLAG_tr1 > 0)
	// TR1 stuff is under directory tr1/
	#undef  CXXOFLAG_directory_tr1
	#define CXXOFLAG_directory_tr1 1
	// TR1 stuff is defined under namespace std::tr1
	#undef  CXXOFLAG_namespace_tr1 
	#define CXXOFLAG_namespace_tr1 1
	#undef  CXXO_COMPILER_SUPPORT_tr1_array
	#define CXXO_COMPILER_SUPPORT_tr1_array 1
    #undef  CXXO_COMPILER_SUPPORT_tr1_cstdint
    #define CXXO_COMPILER_SUPPORT_tr1_cstdint 1
    #undef  CXXO_COMPILER_SUPPORT_tr1_random
    #define CXXO_COMPILER_SUPPORT_tr1_random 1
    #undef  CXXO_COMPILER_SUPPORT_tr1_type_traits
    #define CXXO_COMPILER_SUPPORT_tr1_type_traits 1
    #undef  CXXO_COMPILER_SUPPORT_tr1_tuple
    #define CXXO_COMPILER_SUPPORT_tr1_tuple 1
	#define CXXO_CXXSTDHEADER_tr1_array 1
	#define CXXO_CXXSTDHEADER_tr1_cstdint 1
	#define CXXO_CXXSTDHEADER_tr1_cinttypes 1
	#define CXXO_CXXSTDHEADER_tr1_random 1
	#define CXXO_CXXSTDHEADER_tr1_tuple 1
	#define CXXO_CXXSTDHEADER_tr1_type_traits 1
	
	#undef  CXXO_COMPILER_SUPPORT_shared_ptr
	#define CXXO_COMPILER_SUPPORT_shared_ptr 1

#endif

#if (CXXOFLAG_tr1  && CXXOMFORT_CXX_EMULATION == 2007)
	#define CXXOFLAG_namespace_tr1 1
	#undef  CXXO_STDTR1_NAME
	#define CXXO_STDTR1_NAME() std::tr1
	#undef  CXXO_STDTR1_OPEN
	#define CXXO_STDTR1_OPEN() namespace std { namespace tr1 {
	#undef  CXXO_STDTR1_CLOSE
	#define CXXO_STDTR1_CLOSE() } }
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES > 1)
		#pragma message CXXO_NOTICE ("namespace tr1")
	#endif

#endif

// 
// features per version
// 

#if (CXXOMFORT_COMPILER_VERSION >= 403)
    #if (defined(__GXX_EXPERIMENTAL_CXX0X__))
        #if (CXXOMFORT_COMPILER_VERSION==403)
            #pragma message( CXXO_WARNING( "Configuration not supported (nullptr and other basic features might be missing)" ))
        #endif
        #undef  CXXO_COMPILER_SUPPORT_static_assert
        #define CXXO_COMPILER_SUPPORT_static_assert 1
        #undef  CXXO_COMPILER_SUPPORT_decltype
        #define CXXO_COMPILER_SUPPORT_decltype 2343
    #endif
    #undef  CXXO_COMPILER_SUPPORT_type_intrinsics
    #define CXXO_COMPILER_SUPPORT_type_intrinsics 1

	#undef  CXXO_COMPILER_SUPPORT_std_hash_basicstring
	#define CXXO_COMPILER_SUPPORT_std_hash_basicstring 1

#endif

// starting with GCC 4.x, checking TR1 headers

#if (CXXOMFORT_COMPILER_VERSION >= 404)
	#undef  CXXO_COMPILER_SUPPORT_std_hash
	#define CXXO_COMPILER_SUPPORT_std_hash 1
    #if (CXXOMFORT_CXX_STD< 2011) && (CiXXOMFORT_CXX_EMULATION==2011)
        // Note: *Unsupported* c++0x emulation mode
        #undef  CXXO_COMPILER_SUPPORT_auto
        #define CXXO_COMPILER_SUPPORT_auto 1984
        #undef  CXXO_COMPILER_SUPPORT_conditional_enableif 
        #define CXXO_COMPILER_SUPPORT_conditional_enableif 1
        #undef  CXXO_COMPILER_SUPPORT_default_delete
        #define CXXO_COMPILER_SUPPORT_default_delete 1
        #undef  CXXO_COMPILER_SUPPORT_enum_class
        #define CXXO_COMPILER_SUPPORT_enum_class 1
        #undef  CXXO_COMPILER_SUPPORT_initializer_list
        #define CXXO_COMPILER_SUPPORT_initializer_list 1
        #undef  CXXO_COMPILER_SUPPORT_rvref
        #define CXXO_COMPILER_SUPPORT_rvref 2218
        #undef  CXXO_COMPILER_SUPPORT_std_cxx11_algorithms
        #define CXXO_COMPILER_SUPPORT_std_cxx11_algorithms 1
        #undef  CXXO_COMPILER_SUPPORT_std_metaprogramming_helpers
        #define CXXO_COMPILER_SUPPORT_std_metaprogramming_helpers 1
        #undef  CXXO_COMPILER_SUPPORT_unique_ptr
        #define CXXO_COMPILER_SUPPORT_unique_ptr 1
        #undef  CXXO_COMPILER_SUPPORT_variadic
        #define CXXO_COMPILER_SUPPORT_variadic 1
    #else
//        #undef  CXXO_COMPILER_SUPPORT_tr1_random
//        #undef  CXXO_COMPILER_SUPPORT_tr1_tuple
//        #define CXXO_COMPILER_SUPPORT_tr1_random 1
//        #define CXXO_COMPILER_SUPPORT_tr1_tuple 1
        
        #if !defined(CXXOFLAG_MAKE_TYPEDEF_charn_t)
            #define CXXOFLAG_MAKE_TYPEDEF_charn_t // necessary because of SSO in <string>
        #endif

    #endif
#endif
#if (CXXOMFORT_COMPILER_VERSION >= 405)
    #if (CXXOMFORT_CXX_STD< 2011) && (CXXOMFORT_CXX_EMULATION==2011)
        #undef  CXXO_COMPILER_SUPPORT_alignof
        #define CXXO_COMPILER_SUPPORT_alignof 1
        #undef  CXXO_COMPILER_SUPPORT_explicit_operator
        #define CXXO_COMPILER_SUPPORT_explicit_operator 2437
        #undef  CXXO_COMPILER_SUPPORT_rvref
        #define CXXO_COMPILER_SUPPORT_rvref 2844
    #endif
#endif

#undef  CXXO_COMPILER_SUPPORT_alignof
#define CXXO_COMPILER_SUPPORT_alignof (CXXOMFORT_CXX_STD>=2011)

#if (CXXOMFORT_COMPILER_VERSION >= 406)
    #if (CXXOMFORT_CXX_STD< 2011) && (CXXOMFORT_CXX_EMULATION>=2011)
        #undef  CXXO_CXXSTDHEADER_cstdint
        #define CXXO_CXXSTDHEADER_cstdint 1
        #undef  CXXO_CXXSTDHEADER_system_error
        #define CXXO_CXXSTDHEADER_system_error 1
        #undef  CXXO_CXXSTDHEADER_tuple
        #define CXXO_CXXSTDHEADER_tuple 1
        #undef  CXXO_CXXSTDHEADER_type_traits
        #define CXXO_CXXSTDHEADER_type_traits 1

        #undef  CXXO_COMPILER_SUPPORT_alignof
        #define CXXO_COMPILER_SUPPORT_alignof 1
        #undef  CXXO_COMPILER_SUPPORT_auto
        #define CXXO_COMPILER_SUPPORT_auto 1
        #undef  CXXO_COMPILER_SUPPORT_trailing_returns
        #define CXXO_COMPILER_SUPPORT_trailing_returns 1
        #undef  CXXO_COMPILER_SUPPORT_charn_t
        #define CXXO_COMPILER_SUPPORT_charn_t 2 // defined as their own types
        #undef  CXXO_COMPILER_SUPPORT_default_delete
        #define CXXO_COMPILER_SUPPORT_default_delete 1
        #undef  CXXO_COMPILER_SUPPORT_enum_class
        #define CXXO_COMPILER_SUPPORT_enum_class 1
        #undef  CXXO_COMPILER_SUPPORT_foreach
        #define CXXO_COMPILER_SUPPORT_foreach 1
        #undef  CXXO_COMPILER_SUPPORT_local_types
        #define CXXO_COMPILER_SUPPORT_local_types 1
        #undef  CXXO_COMPILER_SUPPORT_noexcept
        #define CXXO_COMPILER_SUPPORT_noexcept 1
        #undef  CXXO_COMPILER_SUPPORT_nullptr
        #define CXXO_COMPILER_SUPPORT_nullptr 1
        #undef  CXXO_COMPILER_SUPPORT_raw_strings
        #define CXXO_COMPILER_SUPPORT_raw_strings 2442
        #undef  CXXO_COMPILER_SUPPORT_rvref
        #define CXXO_COMPILER_SUPPORT_rvref 3053
        #undef  CXXO_COMPILER_SUPPORT_static_assert
        #define CXXO_COMPILER_SUPPORT_static_assert 1
        #undef  CXXO_COMPILER_SUPPORT_std_cxx11_constructible_traits
        #define CXXO_COMPILER_SUPPORT_std_cxx11_constructible_traits 1
        #undef  CXXO_COMPILER_SUPPORT_variadic
        #define CXXO_COMPILER_SUPPORT_variadic 1

		#undef  CXXO_COMPILER_SUPPORT_alignment_tools
        #define CXXO_COMPILER_SUPPORT_alignment_tools 1
        #undef  CXXO_COMPILER_SUPPORT_common_type
        #define CXXO_COMPILER_SUPPORT_common_type 1
        #undef  CXXO_COMPILER_SUPPORT_conditional_enableif 
        #define CXXO_COMPILER_SUPPORT_conditional_enableif 1
        #undef  CXXO_COMPILER_SUPPORT_std_cxx11_algorithms
        #define CXXO_COMPILER_SUPPORT_std_cxx11_algorithms 1
        #undef  CXXO_COMPILER_SUPPORT_std_iterator_helpers
        #define CXXO_COMPILER_SUPPORT_std_iterator_helpers 1
        #undef  CXXO_COMPILER_SUPPORT_unique_ptr
        #define CXXO_COMPILER_SUPPORT_unique_ptr 1
        
    #else
    #endif
#endif

#if (CXXOMFORT_COMPILER_VERSION >= 407)
    #if (CXXOMFORT_CXX_STD< 2011) && (CXXOMFORT_CXX_EMULATION==2011)
        #undef  CXXO_COMPILER_SUPPORT_constexpr
        #define CXXO_COMPILER_SUPPORT_constexpr 2235
        #undef  CXXO_COMPILER_SUPPORT_template_aliases
        #define CXXO_COMPILER_SUPPORT_template_aliases 1
        #undef  CXXO_COMPILER_SUPPORT_user_defined_literals
        #define CXXO_COMPILER_SUPPORT_user_defined_literals 1
    #endif
#endif

// GCC 4.8 onwards set up support for C++14 (--std=c++1y or --std=c++14)
// via usage of the __cplusplus value
#if (CXXOMFORT_CXX_STD == 2013)
    #define CXXOMFORT_CXX_EMULATION 2014
#endif


#if (CXXOMFORT_COMPILER_VERSION >= 408)
    #undef  CXXO_COMPILER_SUPPORT__Pragma
    #define CXXO_COMPILER_SUPPORT__Pragma 1
    
    #undef  CXXO_COMPILER_SUPPORT_alignment_tools
    #define CXXO_COMPILER_SUPPORT_alignment_tools (CXXOMFORT_CXX_STD >= 2011 || CXXOMFORT_CXX_EMULATION>=2011)
    // 4.8 announces c++0x support as c++11, but still misses some things
    #if (defined(__GXX_EXPERIMENTAL_CXX0X__) && CXXOMFORT_CXX_EMULATION < 2011)
        #undef  CXXOMFORT_CXX_EMULATION
        #define CXXOMFORT_CXX_EMULATION 2011
    #endif
    #if (CXXOMFORT_CXX_STD< 2011) && (CXXOMFORT_CXX_EMULATION>=2011)
        #undef  CXXO_COMPILER_SUPPORT_attribute
        #define CXXO_COMPILER_SUPPORT_attribute 1
        #undef  CXXO_COMPILER_SUPPORT_decltype
        #define CXXO_COMPILER_SUPPORT_decltype 3276
        // 4.8 is considered to have full C++11 support for practical purposes
        // but it's untested so far
    #endif

#endif

#if (CXXOMFORT_COMPILER_VERSION >= 409)
    #if (CXXOMFORT_CXX_STD< 2011) && (CXXOMFORT_CXX_EMULATION==2011)
        // 4.8 is considered to have full C++11 support for practical purposes
        // but it's untested so far
    #else
    //    #undef CXXOMFORT_CXX_STD
    //    #define CXXOMFORT_CXX_STD 2003
    #endif
#endif

#if (CXXOMFORT_COMPILER_VERSION >= 500)
    #if (CXXOMFORT_CXX_STD>=2011)
        #undef  CXXO_COMPILER_SUPPORT_attribute_deprecated
        #define CXXO_COMPILER_SUPPORT_attribute_deprecated 1
		#undef  CXXO_COMPILER_SUPPORT_macro_has_include
		#define CXXO_COMPILER_SUPPORT_macro_has_include 1
    #endif
#endif

#if (CXXOMFORT_COMPILER_VERSION >= 600)
	// GCC >= 6.1 reports C++17 emulation mode as __cplusplus==201500
	#if (CXXOMFORT_CXX_STD >= 2011)
	#endif
	#if (__cplusplus >= 201402 )
		#undef  CXXOMFORT_CXX_EMULATION
		#define CXXOMFORT_CXX_EMULATION 2014
		#undef  CXXO_COMPILER_SUPPORT_udl_lexer_nospace
		#define CXXO_COMPILER_SUPPORT_udl_lexer_nospace 1
	#endif
	#if (__cplusplus == 201500 )
		#undef CXXOMFORT_CXX_EMULATION
		#define CXXOMFORT_CXX_EMULATION 2017
		#undef  CXXO_COMPILER_SUPPORT_std_voidt
		#define CXXO_COMPILER_SUPPORT_std_voidt 1
	#endif
#endif

/*
#if (CXXOMFORT_COMPILER_VERSION >= 700)
    #if (CXXOMFORT_CXX_STD>=2015)
        // these are present in the GCC7 libstdc++
        #undef  CXXO_COMPILER_SUPPORT_std_byte
        #define CXXO_COMPILER_SUPPORT_std_byte 1
        #undef  CXXO_COMPILER_SUPPORT_std_voidt
        #define CXXO_COMPILER_SUPPORT_std_voidt 1
    #endif
#endif
*/

#if (CXXOMFORT_COMPILER_VERSION>=800)
#endif

#if (CXXOMFORT_COMPILER_VERSION>=800)
#endif

#if (CXXOMFORT_COMPILER_VERSION>=1000)
	#undef  CXXO_COMPILER_SUPPORT_destroying_delete
	#define CXXO_COMPILER_SUPPORT_destroying_delete (CXXOMFORT_CXX_STD>=2020 || CXXOMFORT_CXX_EMULATION>=2020)
	#undef  CXXO_COMPILER_SUPPORT_explicit_condition
	#define CXXO_COMPILER_SUPPORT_explicit_condition (CXXOMFORT_CXX_STD>=2020 || CXXOMFORT_CXX_EMULATION>=2020)
#endif

// detect lack of a <cuchar> header
#if (CXXOMFORT_COMPILER_VERSION < 600)
    #undef  CXXO_CXXSTDHEADER_cuchar
    #define CXXO_CXXSTDHEADER_cuchar 0
#endif

// enable the exclusive __cpp macros
#if (!defined(__cpp_decltype) && CXXO_COMPILER_SUPPORT_decltype>0)
	#define __cpp_decltype (CXXOMFORT_CXX_STD*100)
#endif
#if (!defined(__cpp_rvalue_references) && CXXO_COMPILER_SUPPORT_rvref>0)
	#define __cpp_rvalue_references (CXXOMFORT_CXX_STD*100)
#endif
#if (!defined(__cpp_user_defined_literals) && CXXO_COMPILER_SUPPORT_user_defined_literals>0)
	#define __cpp_user_defined_literals (CXXOMFORT_CXX_STD*100)
#endif

#if (CXXOMFORT_CXX_STD >= 2011)
	#if !defined(__cpp_static_assert)
		#define __cpp_static_assert 200410L
	#endif
	#if !defined(__cpp_inheriting_constructors)
		#define __cpp_inheriting_constructors 200802L
	#endif
#endif


#endif
