#ifndef CXXOMFORT_CONFIG_HPP
    #error "This file must not be included directly (see cxxomfort/config.hpp)"
#endif
#ifndef CXXOMFORT_CONFIG_MSC_HPP
#define CXXOMFORT_CONFIG_MSC_HPP

/*
 * 
 * MSVC version history
 * 
 * 14.0 / 2015 (2015-07). Generates telemetry!
 * 12.0 / 2013 (2013-10) - _MSC_VER==1800; **to be tested**
 * 11.0 / 2012 (2012-08) - _MSC_VER==1700; **preferred, Windows 7**
 * 10.0 / 2010 (2010-04) - _MSC_VER==1600;
 * 9.0 / 2008 (2007-11) - _MSC_VER==1500; **preferred**
 * 8.0 (2005-11)
 * 
 * 
 * In VC < 2010, there is no such thing as C++11 emulation mode. What's more
 * important, in some versions the implementation of TR1 is broken or, in the
 * case of VC9 Express (2008) it won't install at all. 
 * Meaning that to be usable, this library needs to implement many
 * type_traits on its own.
 * 
 * In VC >= 2010, the compiler works in what Microsoft calls C++11 mode, 
 * but it is incomplete or broken in various ways; 
 * for example std::errc is not defined as a type, one version resolves 
 * rvalue references incorrectly, and many of them up to 2013 have no 
 * initializer_list support. As such this library does not recorgnize 
 * VC >= 10 as C++11 mode, but as C++0x emulation.
 * (with CXXOMFORT_CXX_EMULATION=2007)
 * 
 * Starting with VC 2017, the entirely of C++11 is implemented 
 * as far as cxxomfort is concerned.
 * 
 * See also:
 * http://msdn.microsoft.com/en-us/library/ms177194%28v=VS.100%29.aspx
 * https://dev.to/yumetodo/list-of-mscver-and-mscfullver-8nd
*/

//
// setup identification information and pragmas
//


#define CXXOMFORT_COMPILER_ID CXXO_COMPILERID_MSC
#define CXXOMFORT_COMPILER_VERSION (_MSC_VER)
#define CXXOMFORT_COMPILER_VERSION_FULL (_MSC_FULL_VER)
// eg.: 120 for VC6 (ugh!), 150 for VC9 (Studio / Express 2008)

#define CXXO_PRAGMA(x) _Pragma (#x)

#undef CXXO_NOTICE
#define CXXO_NOTICE(msg) (msg)
#if defined(CXXOMFORT_NOTICES)
    #pragma message CXXO_NOTICE("Found the Microsoft C++ Compiler [" CXXO_STRINGIZE(_MSC_VER) "]" )
#endif

#if (CXXOMFORT_CXX_STD < 100)
	#undef CXXOMFORT_CXX_STD
	#define CXXOMFORT_CXX_STD (__cplusplus/100)
#endif

// We won't even make an attempt to support VC6
#if (CXXOMFORT_COMPILER_VERSION < 1300)
    #error "cxxomfort -- Detected an horribly broken compiler! (MSVC6 or earlier)"
#endif

/**
 * MSVC headers tend to be organized via a 
 * path like ${VISUAL_STUDIO}/VC/include
 * 
 * TR1 headers live in the same directory as the normal ones, 
 * but features are in namespace std::tr1
 */

// At least VC2005
#if (CXXOMFORT_COMPILER_VERSION >= 1400)
	#undef  CXXOFLAG_tr1
	#define CXXOFLAG_tr1 0
    #undef  CXXO_COMPILER_SUPPORT_type_intrinsics
    #define CXXO_COMPILER_SUPPORT_type_intrinsics 1
    #if (_MSC_FULL_VER == 140050727) // SP1
		
    #endif
#endif

// VC 2008 or higher
#if (CXXOMFORT_COMPILER_VERSION >= 1500) 
    #undef  CXXO_COMPILER_SUPPORT_va_args
    #define CXXO_COMPILER_SUPPORT_va_args 1

    #undef  CXXO_COMPILER_SUPPORT___alignof
    #define CXXO_COMPILER_SUPPORT___alignof 1

    #if (_MSC_FULL_VER == 150030729) // SP1 has TR1
        #undef  CXXOMFORT_CXX_EMULATION
        #define CXXOMFORT_CXX_EMULATION 2007

        #undef  CXXOFLAG_tr1
        #define CXXOFLAG_tr1 1
        #undef  CXXOFLAG_namespace_tr1
        #define CXXOFLAG_namespace_tr1 1
        #undef  CXXO_STDTR1_NAME
        #define CXXO_STDTR1_NAME() std::tr1
        #undef  CXXO_STDTR1_OPEN
        #define CXXO_STDTR1_OPEN() namespace std { namespace tr1 {
        #undef  CXXO_STDTR1_CLOSE
        #define CXXO_STDTR1_CLOSE() } }
        #if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES > 1)
            #pragma message CXXO_NOTICE ("namespace tr1")
        #endif

        #undef  CXXO_CXXSTDHEADER_array
        #define CXXO_CXXSTDHEADER_array 1
        #undef  CXXO_CXXSTDHEADER_random
        #define CXXO_CXXSTDHEADER_random 1
        #undef  CXXO_CXXSTDHEADER_tuple
        #define CXXO_CXXSTDHEADER_tuple 1
        #undef  CXXO_CXXSTDHEADER_type_traits
        #define CXXO_CXXSTDHEADER_type_traits 1
        
        #undef  CXXO_COMPILER_SUPPORT_array
        #define CXXO_COMPILER_SUPPORT_array 0
        #undef  CXXO_COMPILER_SUPPORT_tr1_array 
        #define CXXO_COMPILER_SUPPORT_tr1_array 1
        #undef  CXXO_COMPILER_SUPPORT_tr1_tuple 
        #define CXXO_COMPILER_SUPPORT_tr1_tuple 1
        #undef  CXXO_COMPILER_SUPPORT_tr1_type_traits
        #define CXXO_COMPILER_SUPPORT_tr1_type_traits 1
        #undef  CXXO_COMPILER_SUPPORT_type_traits_basic
        #define CXXO_COMPILER_SUPPORT_type_traits_basic 1

		#undef  CXXO_COMPILER_SUPPORT_inheritance_override
		#define CXXO_COMPILER_SUPPORT_inheritance_override 2928

        #undef  CXXO_COMPILER_SUPPORT_std_hash
        #define CXXO_COMPILER_SUPPORT_std_hash 1
        #undef  CXXO_COMPILER_SUPPORT_std_hash_basicstring
        #define CXXO_COMPILER_SUPPORT_std_hash_basicstring 1

        #undef  CXXO_COMPILER_SUPPORT_shared_ptr
        #define CXXO_COMPILER_SUPPORT_shared_ptr 1
    #endif
    
#endif

// At least VC2010
#if (CXXOMFORT_COMPILER_VERSION >= 1600) // VC10 or higher
    #undef  CXXOMFORT_CXX_EMULATION
    #define CXXOMFORT_CXX_EMULATION 2011
    #undef  CXXO_COMPILER_SUPPORT__Pragma
    #define CXXO_COMPILER_SUPPORT__Pragma 1

    #undef  CXXO_CXXSTDHEADER_array
    #define CXXO_CXXSTDHEADER_array 1
    #undef  CXXO_CXXSTDHEADER_random
    #define CXXO_CXXSTDHEADER_random 1
    #undef  CXXO_CXXSTDHEADER_system_error
    #define CXXO_CXXSTDHEADER_system_error 1
    #undef  CXXO_CXXSTDHEADER_typeindex
    #define CXXO_CXXSTDHEADER_typeindex 1
    #undef  CXXO_CXXSTDHEADER_tuple
    #define CXXO_CXXSTDHEADER_tuple 1

    #undef  CXXO_COMPILER_SUPPORT_auto
    #define CXXO_COMPILER_SUPPORT_auto 1984
    #undef  CXXO_COMPILER_SUPPORT_charn_t
    #define CXXO_COMPILER_SUPPORT_charn_t 2 // char16_t as typedef in yvals.h
    #undef  CXXO_COMPILER_SUPPORT_charn_t_IS_TYPEDEF
    #define CXXO_COMPILER_SUPPORT_charn_t_IS_TYPEDEF 1
    #undef  CXXO_COMPILER_SUPPORT_conditional_enableif
    #define CXXO_COMPILER_SUPPORT_conditional_enableif 1
    #undef  CXXO_COMPILER_SUPPORT_decltype
    #define CXXO_COMPILER_SUPPORT_decltype 2343
    #undef  CXXO_DECLTYPE
    #define CXXO_DECLTYPE(...) decltype(__VA_ARGS__)
    #undef  CXXO_COMPILER_SUPPORT_typeof
    #define CXXO_COMPILER_SUPPORT_typeof 1
    #undef  CXXO_TYPEOF
    #define CXXO_TYPEOF(...) decltype(__VA_ARGS__)
    #undef  CXXO_COMPILER_SUPPORT_iterator_accessors_11
    #define CXXO_COMPILER_SUPPORT_iterator_accessors_11 1
    #undef  CXXO_COMPILER_SUPPORT_local_types
    #define CXXO_COMPILER_SUPPORT_local_types 2657
    #undef  CXXO_COMPILER_SUPPORT_nullptr
    #define CXXO_COMPILER_SUPPORT_nullptr 2431
    #undef  CXXO_COMPILER_SUPPORT_rvref
    #define CXXO_COMPILER_SUPPORT_rvref 2844
    #undef  CXXO_COMPILER_SUPPORT_shared_ptr
    #define CXXO_COMPILER_SUPPORT_shared_ptr 1
    #undef  CXXO_COMPILER_SUPPORT_static_assert
    #define CXXO_COMPILER_SUPPORT_static_assert 1720
    #undef  CXXO_COMPILER_SUPPORT_std_cxx11_algorithms
    #define CXXO_COMPILER_SUPPORT_std_cxx11_algorithms 1 
    #undef  CXXO_COMPILER_SUPPORT_std_cxx11_constructible_traits
    #define CXXO_COMPILER_SUPPORT_std_cxx11_constructible_traits 1
    #undef  CXXO_COMPILER_SUPPORT_std_hash
    #define CXXO_COMPILER_SUPPORT_std_hash 1
    #undef  CXXO_COMPILER_SUPPORT_iterator_accessors_11
    #define CXXO_COMPILER_SUPPORT_iterator_accessors_11 1
    #undef  CXXO_COMPILER_SUPPORT_std_metaprogramming_helpers
    #define CXXO_COMPILER_SUPPORT_std_metaprogramming_helpers 1
    #undef  CXXO_COMPILER_SUPPORT_trailing_returns
    #define CXXO_COMPILER_SUPPORT_trailing_returns 1
    #undef  CXXO_COMPILER_SUPPORT_type_traits_basic
    #define CXXO_COMPILER_SUPPORT_type_traits_basic 1
    #undef  CXXO_COMPILER_SUPPORT_unique_ptr
    #define CXXO_COMPILER_SUPPORT_unique_ptr 1

    #undef  CXXOFLAG_tr1
    #define CXXOFLAG_tr1 1
    #undef  CXXOFLAG_namespace_tr1
    #define CXXOFLAG_namespace_tr1 1
    #undef  CXXO_STDTR1_NAME
    #define CXXO_STDTR1_NAME() std::tr1
    #undef  CXXO_STDTR1_OPEN
    #define CXXO_STDTR1_OPEN() namespace std { namespace tr1 {
    #undef  CXXO_STDTR1_CLOSE
    #define CXXO_STDTR1_CLOSE() } }

    #if (_MSC_FULL_VER == 160040219) // SP1
        
    #endif
    
    #if (!defined(__cpp_static_caaser))
		#define __cpp_static_assert 199711L
	#endif
#endif

// At least VC 2012
#if (CXXOMFORT_COMPILER_VERSION >= 1700)
	#undef  CXXOFLAG_tr1
	#define CXXOFLAG_tr1 1

    #undef  CXXO_CXXSTDHEADER_cstdint 
    #define CXXO_CXXSTDHEADER_cstdint 1
    #undef  CXXO_CXXSTDHEADER_typeindex
    #define CXXO_CXXSTDHEADER_typeindex 1

    #undef  CXXO_COMPILER_SUPPORT_common_type
    #define CXXO_COMPILER_SUPPORT_common_type 1
    #undef  CXXO_COMPILER_SUPPORT_enum_class
    #define CXXO_COMPILER_SUPPORT_enum_class 1
    #undef  CXXO_COMPILER_SUPPORT_override_final
    #define CXXO_COMPILER_SUPPORT_override_final 1
    #undef  CXXO_COMPILER_SUPPORT_foreach
    #define CXXO_COMPILER_SUPPORT_foreach 1
    #undef  CXXO_COMPILER_SUPPORT_inheritance_final
    #define CXXO_COMPILER_SUPPORT_inheritance_final 3272
	#undef  CXXO_COMPILER_SUPPORT_inheritance_override
	#define CXXO_COMPILER_SUPPORT_inheritance_override 3272
    #undef  CXXO_COMPILER_SUPPORT_std_common_type
    #define CXXO_COMPILER_SUPPORT_std_common_type 1
    #undef  CXXO_COMPILER_SUPPORT_std_is_trivially
    #define CXXO_COMPILER_SUPPORT_std_is_trivially 1
    #undef  CXXO_COMPILER_SUPPORT_std_is_constructible_traits
    #define CXXO_COMPILER_SUPPORT_std_is_constructible_traits 1
    
    #if (!defined(__cpp_rvalue_references))
		#define __cpp_rvalue_references (CXXOMFORT_CXX_STD*100)
	#endif

    #if (_MSC_FULL_VER >= 170060610) // update 3
		#undef  CXXO_COMPILER_SUPPORT_lambda_expressions
		#define CXXO_COMPILER_SUPPORT_lambda_expressions 1        
    #endif
    
    #if (_MSC_FULL_VER >= 170051025) // November CTP
    #endif

	#if !defined(__cpp_decltype)
		#define __cpp_decltype __cplusplus 200707L
	#endif
	#if !defined(__cpp_range_based_for)
		#define __cpp_range_based_for 200907L
	#endif

#endif

// At least VC2013
#if (CXXOMFORT_COMPILER_VERSION >= 1800) 
    #undef  CXXO_COMPILER_SUPPORT_alignof
    #define CXXO_COMPILER_SUPPORT_alignof 1
    #undef  CXXO_COMPILER_SUPPORT_decltype
    #define CXXO_COMPILER_SUPPORT_decltype 3276
    #undef  CXXO_COMPILER_SUPPORT_raw_strings
    #define CXXO_COMPILER_SUPPORT_raw_strings 1
    
	#undef  CXXO_COMPILER_SUPPORT_transparent_operators_functional
	#define CXXO_COMPILER_SUPPORT_transparent_operators_functional 1
	#undef  CXXO_COMPILER_SUPPORT_iterator_accessors_14
	#define CXXO_COMPILER_SUPPORT_iterator_accessors_14 1

    #if (_MSC_FULL_VER >= 180030720) // update 2
    #endif
    #if (_MSC_FULL_VER >= 180031101) // update 4
    #endif
    #if (_MSC_FULL_VER >= 180021114) // November CTP
    #endif
    
#endif

// At least VC2015
#if (CXXOMFORT_COMPILER_VERSION >= 1900)

	// MSVC 2015 U3 adds the /std:c++xx Switch
	#if (_MSC_FULL_VER >= 190024210) // update 3, support for std switch
		#undef  CXXOMFORT_CXX_EMULATION
		#define CXXOMFORT_CXX_EMULATION (_MSVC_LANG / 100)
	#endif

    #undef  CXXOFLAG_tr1
    #define CXXOFLAG_tr1 0
    #undef  CXXOFLAG_namespace_tr1
    #define CXXOFLAG_namespace_tr1 0
    #undef  CXXO_STDTR1_NAME
    #define CXXO_STDTR1_NAME() std
    #undef  CXXO_STDTR1_OPEN
    #define CXXO_STDTR1_OPEN() namespace std {
    #undef  CXXO_STDTR1_CLOSE
    #define CXXO_STDTR1_CLOSE() }

    #undef  CXXO_COMPILER_SUPPORT_alignment_tools
    #define CXXO_COMPILER_SUPPORT_alignment_tools 1
    #undef  CXXO_COMPILER_SUPPORT_alignof 
    #define CXXO_COMPILER_SUPPORT_alignof 1
    #undef  CXXO_COMPILER_SUPPORT_attribute
    #define CXXO_COMPILER_SUPPORT_attribute 1
    #define __cpp_attribute 200809L
    #undef  CXXO_COMPILER_SUPPORT_charn_t
    #define CXXO_COMPILER_SUPPORT_charn_t 1
    #undef  CXXO_COMPILER_SUPPORT_constexpr
    #define CXXO_COMPILER_SUPPORT_constexpr 1
    #undef  CXXO_COMPILER_SUPPORT_explicit_operator
    #define CXXO_COMPILER_SUPPORT_explicit_operator 1
    #undef  CXXO_COMPILER_SUPPORT_inheriting_constructors
    #define CXXO_COMPILER_SUPPORT_inheriting_constructors 1
    #undef  CXXO_COMPILER_SUPPORT_integer_sequence
    #define CXXO_COMPILER_SUPPORT_integer_sequence 1
    #undef  CXXO_COMPILER_SUPPORT_noexcept
    #define CXXO_COMPILER_SUPPORT_noexcept 1
    #undef  CXXO_COMPILER_SUPPORT_rvref
    #define CXXO_COMPILER_SUPPORT_rvref 3053
    #undef  CXXO_COMPILER_SUPPORT_template_aliases
    #define CXXO_COMPILER_SUPPORT_template_aliases 1
    #undef  CXXO_COMPILER_SUPPORT_unicode_string_literals 
    #define CXXO_COMPILER_SUPPORT_unicode_string_literals 1
    #undef  CXXO_COMPILER_SUPPORT_user_defined_literals
    #define CXXO_COMPILER_SUPPORT_user_defined_literals 1
    #undef  CXXO_COMPILER_SUPPORT_variadic
    #define CXXO_COMPILER_SUPPORT_variadic 1

    #undef  CXXO_COMPILER_SUPPORT_iterator_accessors_17
    #define CXXO_COMPILER_SUPPORT_iterator_accessors_17 1

	#if !defined(__cpp_decltype_auto)
		#define __cpp_decltype_auto 201103L
	#endif
	#if !defined(__cpp_inheriting_constructors)
		#define __cpp_inheriting_constructors 199711L
	#endif
	#if !defined(__cpp_rvalue_references)
		#define __cpp_rvalue_references __cplusplus 200610L
	#endif
	#if !defined(__cpp_user_defined_literals)
		#define __cpp_user_defined_literals __cplusplus 200809L
	#endif
	#if !defined(__cpp_variadic_templates)
		#define __cpp_variadic_templates __cplusplus 200704L
	#endif


#endif


// At least VC2017
#if (CXXOMFORT_COMPILER_VERSION >= 1910)

	#if defined(_MSVC_TRADITIONAL) && (_MSVC_TRADITIONAL)
		#pragma message CXXO_WARNING("MSVC configured to use the VS-specific, non-compatible preprocessor")
	#endif
	
	#undef  CXXO_COMPILER_SUPPORT_std_clamp
	#define CXXO_COMPILER_SUPPORT_std_clamp 1
    #undef  CXXO_COMPILER_SUPPORT_constexpr_relaxed 
    #define CXXO_COMPILER_SUPPORT_constexpr_relaxed 1
    
    #if (_MSC_FULL_VER >= 191125547) // U4
		#undef  CXXO_COMPILER_SUPPORT_static_assert_17
		#define CXXO_COMPILER_SUPPORT_static_assert_17 (CXXOMFORT_CXX_EMULATION>=2017)
        
    #endif
    
    #if (CXXOMFORT_COMPILER_VERSION >= 1913 && CXXOMFORT_CXX_EMULATION >= 2017) // U6
		#undef  CXXO_COMPILER_SUPPORT_macro_has_include
		#define CXXO_COMPILER_SUPPORT_macro_has_include 1
		#undef  CXXO_CXXSTDHEADER_string_view
		#define CXXO_CXXSTDHEADER_string_view 1
		#undef  CXXO_COMPILER_SUPPORT_std_byte
		#define CXXO_COMPILER_SUPPORT_std_byte (CXXOMFORT_CXX_EMULATION >= 2017)
        #undef  CXXO_COMPILER_SUPPORT_invoke
        #define CXXO_COMPILER_SUPPORT_invoke (CXXOMFORT_CXX_EMULATION >= 2017)
        #undef  CXXO_COMPILER_SUPPORT_std_hash
        #define CXXO_COMPILER_SUPPORT_std_hash 1

    #endif

    #if (_MSC_FULL_VER >= 191326132 && CXXOMFORT_CXX_EMULATION >= 2017) // update 6 final
		#undef  CXXO_CXXSTDHEADER_any
		#define CXXO_CXXSTDHEADER_any 1
		#undef  CXXO_CXXSTDHEADER_optional
		#define CXXO_CXXSTDHEADER_optional 1
		#undef  CXXO_CXXSTDHEADER_variant
		#define CXXO_CXXSTDHEADER_variant 1
        
    #endif

    #if (CXXOMFORT_COMPILER_VERSION >= 1916) // U9

    #endif
    
    #undef  CXXO_COMPILER_SUPPORT_std_voidt
    #define CXXO_COMPILER_SUPPORT_std_voidt (CXXOMFORT_CXX_EMULATION >= 2014)
    #undef  CXXO_COMPILER_SUPPORT_iterator_accessors_17
    #define CXXO_COMPILER_SUPPORT_iterator_accessors_17 (CXXOMFORT_CXX_EMULATION >= 2014)

#endif

// MSVC 2017 15.7 *finally* brings in Expr-SFINAE
#if (CXXOMFORT_COMPILER_VERSION>=1914)
	#undef  CXXO_COMPILER_SUPPORT_expression_sfinae
	#define CXXO_COMPILER_SUPPORT_expression_sfinae 1
	#undef  CXXO_COMPILER_SUPPORT_bool_constant
	#define CXXO_COMPILER_SUPPORT_bool_constant 1 // (CXXOMFORT_CXX_EMULATION >= 2017)
#endif

// At least VC2019
#if (CXXOMFORT_COMPILER_VERSION>=1920)
	#undef  CXXO_COMPILER_SUPPORT_destroying_delete
	#define CXXO_COMPILER_SUPPORT_destroying_delete (CXXOMFORT_CXX_EMULATION>=2020 && CXXOMFORT_COMPILER_VERSION>=1927)

#endif

// -- END Microsoft C++ Compiler --
#endif
