#ifndef CXXOMFORT_CSTDDEF_HPP
#define CXXOMFORT_CSTDDEF_HPP
/**
 * @file
 * @brief Implementations and additions tied to Standard header <tt>cstddef</tt>.
 * @ingroup std0cstddef 
 * 
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/config.hpp>
#include <cstddef>

// 
// ssize_t
// 
#undef  CXXOMFORT_IMPLEMENTS_lwg2251
#if (CXXOMFORT_CXX_STD >= 9999) // sometime in the future
	#define CXXOMFORT_IMPLEMENTS_lwg2251 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_OW)
	#define CXXOMFORT_IMPLEMENTS_lwg2251 CXXO_IMPLSTATUS_NATIVE()
	// already defined ssize_t
#elif (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_DMC)
	#define CXXOMFORT_IMPLEMENTS_lwg2251 CXXO_IMPLSTATUS_NATIVE()
    // already defined
#else
	#include <sys/types.h>
	#if (defined(__unix__))
		#include <unistd.h>
		#define CXXOMFORT_IMPLEMENTS_lwg2251 CXXO_IMPLSTATUS_BACKPORT()
	#else
	#endif
	#if (!defined(_POSIX_VERSION))
		typedef intmax_t ssize_t;
		#define CXXOMFORT_IMPLEMENTS_lwg2251 CXXO_IMPLSTATUS_EMULATION()
	#endif
#endif
#undef  CXXOMFORT_IMPLEMENTS_ssize_t
//! Informs whether cxxomfort has found, or has implemented, LWG2251's @c ssize_t .
#define CXXOMFORT_IMPLEMENTS_ssize_t CXXOMFORT_IMPLEMENTS_lwg2251

// 
// max_align_t 
// 

#if defined(DOXYGEN_DOC)
namespace std {
/**
 * @typedef max_align_t 
 * @brief Alias for the native type with the largest alignment available.
 * @ingroup std0cstddef
 * @sa https://en.cppreference.com/w/cpp/types/max_align_t
 * @cxx1b Alias for the native type with the largest alignment available.
 * @xrefitem std0cstddef max_align_t definition
 */
	struct max_align_t;
}
#elif (CXXOMFORT_CXX_STD<2011 && CXXOMFORT_CXX_EMULATION<2011)
namespace std {
    typedef long double max_align_t;
}
#endif

namespace cxxomfort {
#if (CXXOMFORT_CXX_STD>=2011)
	typedef decltype(NULL) NULL_t;
#else
	typedef int NULL_t;
#endif

/**
 * @brief Sample enum that can be used to figure out size and signedness of enums at compiletime.
 * */
struct enum_sample {
	enum type { value0 = 0, value1 = 1};
	enum { value = sizeof(type) };
};



}

#include "base/nullptr.hpp"  // nullptr_t
#include "base/static_assert.hpp"

// byte and helpers

#include "impl/17-byte.hpp" // c++17's "byte"

namespace cxxomfort { 
	using fix::to_byte;
	#if (CXXOMFORT_IMPLEMENTS_byte >= CXXO_IMPLSTATUS_NATIVE())
	static inline CXXO_CONSTEXPR 
	unsigned char 	to_value (std::byte const b) {
		return ::cxxomfort::fix::to_value(b);
	}
	#endif
}
#endif

/**
 * @page std0cstddef <cstddef>
 * @brief Features related to Standard header <tt><cstddef></tt>.
 * 
 * @section Interfaces 
 * 
 * Features included in this section and mapped to global nsmespace or <tt>std::</tt>:
 * 
 * * @c max_align_t (C++11)
 * * @link cxxomfort::cxxostd::byte @c std::byte @endlink , @c to_integral() (C++17)
 * 
 * Non-backport features (<tt>cxxomfort::fix</tt>):
 * 
 * * @c to_byte() - obtain a <tt>std::byte</tt> value for a given integral.
 * * @c to_integer() - obtain the value of a <tt>std::byte</tt>.
 * * @c to_value() - overload synonym for <tt>to_integral(std::byte)</tt>.
 * 
 * Other features:
 * 
 * * @c ssize_t - This signed type is defined for non-POSIX systems, see @ref cxxomfort_lwg2251 .
 * 
 * **Implementation Macros**
 * 
 * * <tt>std::byte</tt> implementation level: @c CXXOMFORT_IMPLEMENTS_std_byte 
 * * <tt>ssize_t</tt> implementation level: @c CXXOMFORT_IMPLEMENTS_ssize_t 
 * 
 * @section byte_conversion std::byte conversions
 * 
 * C++17's @c byte can not be directly initialized with integral values in C++14 or previous due to the requirements on conversions to <tt>enum class</tt> types. The cxxomfort library exposes the following names under <tt>cxxomfort::fix</tt> in order to work with variables of type @ref @c std::byte :
 * 
 * @code
 * byte     b = cxxomfort::fix::to_byte('c'); // creates byte from uchar
 * unsigned d = cxxomfort::fix::to_value(b); // converts to uchar
 * @endcode
 * 
 * Both functions are also available under @ref @c cxxomfort:: .
 * 
 * @see @ref cxxomfort-fixes
 * 
 * @section cxxomfort_lwg2251 LWG2251 "ssize_t" 
 * 
 * POSIX systems expose the @c ssize_t type which is essentially a @c signed version of @c size_t . This type is exposed by cxxomfort when available, otherwise it's calculated from already existing types as basically <tt>make_signed<size_t>::type</tt>.
 * 
 * @see @wg21{lwg2251,} for the formal definition of @c ssize_t .
 * 
 * The macro @c CXXOMFORT_IMPLEMENTS_ssize_t informs how this fetaure is provided:
 * 
 * * <tt> >= CXXO_IMPLSTATUS_NATIVE()</tt> - if @c ssize_t already exists in the system headers (POSIX or otherwise).
 * * <tt> >= CXXO_IMPLSTATUS_BACKPORT()</tt> - if cxxomfort provides it.
 * * <tt> == 0 </tt> - if it can not be added.
 * 
 * @section std0cstddef_also See Also
 * 
 * * @see @cppref{header/cstddef}
 * * @see @cppref{types/byte} (C++17)
 * 
 * See also: @ref std0cstdio <cstdio> , @ref std0cstdint <cstdint> 
 * 
 * 
 * **/
