#ifndef CXXOMFORT_CSTDINT_HPP
#define CXXOMFORT_CSTDINT_HPP
/**
 * @file
 * @brief Implementations and additions tied to Standard header <tt>\<cstdint\></tt>.
 * 
 * This file will, when included, attempt to find and @c \#include  whichever 
 * headers provide the integer named types such as @c int16_t .
 * 
 */


#include <cxxomfort/base.hpp>
#include <cxxomfort/cxxo-utils.hpp> // tinyint_
#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
	#pragma message CXXO_NOTICE("enabled <cstdint> support.")
#endif
#if (CXXOMFORT_CXX_STD>=2011)
	#include <cstdint>
	#undef  CXXOMFORT_IMPLEMENTS_cstdint_integrals
	#define CXXOMFORT_IMPLEMENTS_cstdint_integrals CXXO_IMPLSTATUS_NATIVE()
#else
	//#include <inttypes.h>
	#include <cxxomfort/base/cstdint.hpp>
#endif

// 
// Numeric type macros (C++03 onward)
// These are supplements to cstdint's U/INTn_C() macros, but instead of merely affixing compiler-specified literal suffixes, they cast to the defined integer type. The macro names are chosen to reflect printf-like formatting flags for ints
// These features are macros and not functions because they also have to work in compile-time contexts in C++03-
// 

namespace cxxomfort {
namespace fix {
	#define CXXOI_hi(x) (static_cast<signed short>(x))
	#define CXXOI_hu(x) (static_cast<unsigned short>(x))

	#define CXXOI_li(x) (static_cast<signed long>(x))
	#define CXXOI_lu(x) (static_cast<unsigned long>(x))

	#define CXXOI_lli(x) (static_cast<signed long long>(x))
	#define CXXOI_llu(x) (static_cast<unsigned long long>(x))

	#define CXXOI_ji(x) (static_cast<intmax_t>(x))
	#define CXXOI_ju(x) (static_cast<uintmax_t>(x))

	#define CXXOI_zi(x) (static_cast<ssize_t>(x))
	#define CXXOI_zu(x) (static_cast<size_t>(x))

namespace integer_literals {
#if (CXXO_COMPILER_SUPPORT_user_defined_literals && CXXO_COMPILER_SUPPORT_udl_lexer_nospace)

signed short operator ""_hi (unsigned long long int x) noexcept {
	return static_cast<signed short>(x);
}
unsigned short operator ""_hu (unsigned long long int x) noexcept {
	return static_cast<unsigned short>(x);
}
signed long operator ""_li (unsigned long long int x) noexcept {
	return static_cast<signed long>(x);
}
unsigned long operator ""_lu (unsigned long long int x) noexcept {
	return static_cast<unsigned long>(x);
}
intmax_t operator ""_ji (unsigned long long int x) noexcept {
	return static_cast<intmax_t>(x);
}
uintmax_t operator ""_ju (unsigned long long int x) noexcept {
	return static_cast<uintmax_t>(x);
}
ssize_t operator ""_zi (unsigned long long int x) noexcept {
	return static_cast<ssize_t>(x);
}
size_t operator ""_zu (unsigned long long int x) noexcept {
	return static_cast<size_t>(x);
}

#endif
}

}
} // cxxomfort

/**
 * @page std0cstdint <cstdint>
 * @brief Features related to Standard header <code>\<cstdint\></code>.
 *
 * @code
 * #include <cxxomfort/cstdint.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Features declared here include mostly the special type names for the supported integer ("integral") types that are defined by the compiler:
 * 
 * * Candidate size integer typedefs <code>[u]int_least<em>n</em>_t</code> such as @c int_least32_t .
 * * Exact size integer typedefs <code>[u]int<em>n</em>_t</code> such as @c int32_t .
 * * Fast arithmetic integer typedefs <code>uint_fast<em>n</em>_t</code> such as @c int_fast32_t .
 * * Maximum-size integer typedef @c intmax_t , @c uintmax_t .
 * * Pointer-sized integer typedefs @c intptr_t , @c uintptr_t .
 * * <tt>ssize_t</tt> (via @ref std0cstddef ).
 * 
 * The following non-backport interfaces are defined:
 * 
 * * @ref udl_integer_literals - defined in namespace <tt>cxxomfort::fix::integer_literals</tt>, these allow for creating integer values of exact types following <tt>INTx_C()</tt> and <tt>printf("%...")</tt> style definitions for types.
 * 
 * @section udl_integer_literals 
 * 
 * In addition to cstdint's <tt>INTn_C()</tt> family macros, cxxomfort also adds the following eight macros for creating integer constants of specified types.
 * 
 * * @c CXXOI_hi() and @c CXXOI_hu() - create @c signed or @c unsigned @b @c short s.
 * * @c CXXOI_li() and @c CXXOI_lu() - create @c signed or @c unsigned @b @c long s.
 * * @c CXXOI_lli() and @c CXXOI_llu() - create @c signed or @c unsigned <b><tt>long long</tt></b>s.
 * * @c CXXOI_ji() and @c CXXOI_ju() - create @c signed or @c unsigned @b @c (u)intmax_t s.
 * * @c CXXOI_zi() and @c CXXOI_zu() - create @c signed or @c unsigned @b @c (s)size_t s.
 * 
 * <tt>cxxomfort::fix::integer_literals</tt>
 * 
 * @section Notes Notes
 * 
 * For platforms without access to a Standard <code><stdint.h></code>, cxxomfort makes this possible by providing a copy of <em>pstdint.h the Portable stdint</em>.
 * 
 * @warning For the definition of these types, cxxomfort assumes that the native type <code>long long</code> is available, even in C++03 mode. Using a platform without such type is <b>unsupported</b> but can be tried by defining the @link cxxomfort-behaviour "Configuration macro" @endlink <tt>CXXOFLAG_WITHOUT_LONGLONG</tt>.
 * 
 * @section std0cstdint_also See Also
 * 
 * @see @cppref{header/cstdint}
 * @see @cppref{types/integer} - description of the various integer types
 * 
 * 
 * **/

#endif
