#ifndef CXXOMFORT_CSTRING_HPP
#define CXXOMFORT_CSTRING_HPP
/**
 * @file 
 * @brief Implementations and additions tied to <tt>\<cstring\></tt>.
 * @xrefitem std0cstring "" ""
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/algorithm.hpp>
#include <cstring>
#include <algorithm>
#include <cstdlib>
#include <cerrno>

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
    #pragma message CXXO_NOTICE("enabled <cstring> support.")
#endif

namespace cxxomfort { namespace fix {

	#if defined(DOXYGEN_DOC)
	/**
	 * @brief Wraps calls to strerror_r to fix the order and number of arguments across implementations.
	 * @ingroup cxxomfort-fix 
	 * 
	 * Some systems, in particular GNU, expose @c strerror_r() with a different number or order of arguments. Cxxomfort fixes this by implementing a wrapper that also covers some systems that do not implement the function, such as pre-UCRT Win32.
	 * 
	 * */
static inline 
int strerror_r (int const errnum, char* dest, size_t destsz);

	#elif defined(_GNU_SOURCE)
		//#pragma message ("_GNU_SOURCE")

static inline
int 
strerror_r (int const errnum, char *destbuf, size_t const destsz) {
	errno= 0;
	// here we MUST use the return value or strerror_r instead of the passed buffer 
	// see https://en.cppreference.com/w/cpp/string/byte/strerror for rationale
	char *gbuffer= ::strerror_r(errnum, destbuf, destsz);
	int const errv= errno;
	if (errv) return errv;
	
	size_t gsz= strlen(gbuffer)+1;
	if (destbuf != nullptr && destsz > 0) {
		size_t slen=  gsz<destsz ? gsz : destsz;
		std::copy_n( gbuffer, slen, destbuf );
	}

	return errv;
}

	#elif defined(_POSIX_VERSION)
		#pragma message ("_POSIX")

using ::strerror_r;
	
	#else
	// eg.: Windows, where POSIX is not used

static inline
int 
strerror_r (int const errnum, char *destbuf, size_t const destsz) {
	errno= 0;
	char * ev = strerror(errnum);
	int const errv= errno;
	if (errv) return errv;
	if (destbuf != nullptr && destsz > 0) {
		using std::copy_n;
		size_t slen= strlen(ev) + 1< destsz ? strlen(ev) + 1: destsz;
		copy_n( ev, slen, destbuf );
	}
	return errv;
}

	#endif
	
} }

#endif // file

/**
 * @page std0cstring <cstring>
 * @brief Backports related to Standard header <tt><cstring></tt>
 * 
 * @code
 * #include <cxxomfort/cstring.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Interfaces defined in this section:
 * 
 * @e None
 * 
 * Interfaces not in std:
 * 
 * * @c cxxomfort::fix::strerror_r() — Fixes wrong GNU form of @c strerror_r() .
 * 
 * @section std0cstring_also See Also
 * 
 * * @see @cppref{header/cstring} (cppreference)
 *
 * */
