#ifndef CXXOMFORT_CXXO_UTILS_HPP
#define CXXOMFORT_CXXO_UTILS_HPP

/**
 * @file
 * @brief Small utilities used by various components of cxxomfort, but low-level and generic enough that they don't get categorized into the backport headers.
 *
 * @ingroup cxxomfort-utils
 * @xrefitem cxxomfort-utils "Utilities" "" Header File
 *
 * @code
#include <cxxomfort/cxxo-utils.hpp>
 * @endcode
 *
 * Including this header brings up the base cxxomfort
 * utilities as well.
 *
 * */
#include "base.hpp"
#include "config.hpp"

#include <cxxomfort/cstdio.hpp>
//#include <cxxomfort/impl/p0146-regular_void.hpp>
#include <cxxomfort/impl/relationals.hpp>
#include <cxxomfort/utils/traits_basic.hpp> // conditional, is_void
#include <cstdlib> // std::abort()
#include <cstdio>
#include <utility>

/**
 * @{
 * @ingroup cxxomfort-utils
 *
 * */

namespace cxxomfort {

//
// Static_Const
// https://ericniebler.com/2014/10/21/customization-point-design-in-c11-and-beyond/
//

#if (defined(DOXYGEN_DOC) || CXXO_COMPILER_SUPPORT_constexpr)
	//#pragma message "Static_Const 2"

/**
 * @cxxoutils Declares static objects to avoid ODR violation.
 * */
template <typename T>
struct Static_Const {
	constexpr const static T value {};
};
template <typename T>
constexpr T Static_Const<T>::value;
#elif (defined(__cpp_constexpr) && defined(__cpp_nsdmi))
	//#pragma message "Static_Const 1"
template <typename T>
struct Static_Const {
	constexpr const static T value {};
};
template <typename T>
constexpr T Static_Const<T>::value;
#elif (defined(__cpp_constexpr))
	#pragma message "Static_Const 0"
template <typename T>
struct Static_Const {
	constexpr const static T value {};
};
template <typename T>
T Static_Const<T>::value ;
#else
	//#pragma message "Static_Const -1"
template <typename T>
struct Static_Const {
	static T const value;
};
template <typename T>
T const Static_Const<T>::value ;
#endif

/*
 * used as follows:
 * namespace detail { // important!
 *   struct T { };
 * }
 * namespace {
 *   static detail::T const& myconst = Static_Const<detail::T>::value; // C++03
 *   constexpr detail::T&& myconst = Static_Const<detail::T>::value; // C++>=11
 * }
 *
 * */


//
// priority tag
// https://quuxplusone.github.io/blog/2021/07/09/priority-tag
//

/**
 * @cxxoutils Priority tag for ordering specializations and overloads
 *
 * @see https://quuxplusone.github.io/blog/2021/07/09/priority-tag
 * */
template <size_t I>
struct prio
: prio <I-1> { enum { value = I}; };

template <>
struct prio<0> { enum { value = 0 }; };

//
// regular void
//

/**
 * @brief Type that models a semi-regular @c void
 *
 * It is intended to be used in contexts where you need to do nothing but
 * can't use @c void for that (for example, for returning a value from a function).
 *
 * Regular void is based on proposal @e p0146r1 which defines an operable void type with the following properties:
 *
 * * @c rvoid can be default-constructed
 * * @c rvoid can be copy-constructed and copy-assigned
 * * All instances of @c rvoid compare equal
 * * Arrays of @c rvoid can be made
 * * Functions can take @c rvoid as arguments, and return @c rvoid
 *
 * @showrefs
 *
 * @section rvoid_also See Also
 *
 * @see @wg21{p0146r1,regular void} .
 *
 * @see https://stackoverflow.com/questions/53197340/how-much-existing-c-code-would-break-if-void-was-actually-defined-as-struct-v
 *
 * */
struct rvoid {
#if (CXXO_COMPILER_SUPPORT_constexpr || defined(DOXYGEN_DOC))
	constexpr rvoid () noexcept = default;
	constexpr rvoid (rvoid const&) = default;
	CXXO_CONSTEXPR14 rvoid& operator= (rvoid const&) = default;
#else
	rvoid () {};
#endif
#if (CXXOMFORT_CXX_STD >= 2011)
	template <typename T>
	explicit CXXO_CONSTEXPR rvoid (T&&) noexcept {}
#else
	template <typename T>
	explicit rvoid (T const&) {}
#endif
};

static_assert(sizeof(rvoid)==1, "rvoid check 1");

//! All instances of @c rvoid compare equal:
CXXO_CONSTEXPR static inline
bool operator== (rvoid const, rvoid const) CXXO_NOEXCEPT { return true; }

CXXO_CONSTEXPR static inline
bool operator!= (rvoid const, rvoid const) CXXO_NOEXCEPT { return false; }

// A metafunction that is an identity metafunction for all types except cv-void, in which case it converts to cv-Void

template <typename T> struct filter_void_identity {
	typedef T type;
};

template <> struct filter_void_identity<void>  {
	typedef rvoid type;
};

template <> struct filter_void_identity<void const>  {
	typedef rvoid const type;
};

// The inverse of the above metafunction

template <typename T> struct filter_rvoid_identity {
	typedef T type;
};

template <> struct filter_rvoid_identity<rvoid> {
	typedef void type;
};

template <> struct filter_rvoid_identity<rvoid const> {
	typedef void const type;
};

//
// None
//

typedef rvoid None_t;

#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD >= 2014)
//! @cxxoutils "None" keyword used by the cxxomfort library.
static constexpr None_t None = {};
#else
namespace {
	CXXO_CONSTEXPR None_t const& None = Static_Const<None_t>::value;
}
#endif


/**
 * @brief noop function
 * */
struct noop_t {
    void operator() (...) const CXXO_NOEXCEPTNOTHROW {  }
};

namespace {
	CXXO_CONSTEXPR noop_t const& noop = Static_Const<noop_t>::value;
}

} // cxxomfort

// 
// Empty types
// 

namespace cxxomfort {

//! @cxxoutils A generic empty type.
struct Empty { explicit Empty() CXXOMFORT_CXX11_CODE(=default, {}); };

CXXO_CONSTEXPR static inline bool operator== (Empty const, Empty const) CXXO_NOEXCEPTNOTHROW { return true; }
CXXO_CONSTEXPR static inline bool operator!= (Empty const, Empty const) CXXO_NOEXCEPTNOTHROW { return false; }

template <typename T>
struct EmptyT { };

}

// 
// Status codes
// 

namespace cxxomfort {

//! Holds message codes for cxxomfort's implementation of status codes, system_error2, etc.
struct message_codes {
	enum type {
		success= 0, //!< Operation successful
		unknown= -1,
		forbidden= 1,
		default_constructor_error, //!< Invoking the default constructor for an object resulted in unrecoverable error / would have thrown
		copy_constructor_error, //!< Invoking the copy constructor for an object resulted in unrecoverable error / would have thrown
		use_before_init, //!< Operation on an object before it has been declared prepared / initialized
		use_after_finish, //!< Operation on an object after it has been declared released / decommissioned
		invalid_argument, //!< Argument not permitted, see eg.: @c EINVAL
		invalid_encoding, //!< Argument does not have the required text encoding
		div_by_zero, //!< A division by zero would have been invoked
		domain_error, //!< Argument out of domain, see eg.: @c EDOM
		out_of_range, //!< Result outside of type range, see eg.: @c ERANGE 
		out_of_bounds, //!< Attempt to access element at unavailable index, see eg.: @c EINVAL 
		dereference_error, //!< An object is not dereferenceable
		invalid_logic, //!< Attempt to use a binary logic operation on an object
		out_of_memory, //! No more memory can be allocated from a global arena
		out_of_local_memory, //! No more memory can be allocated from a local arena
		
	};
};

}

// 
// Parameter types
// 

namespace cxxomfort { 

/**
 * @cxxoutils  Parameter type to prevent unwanted conversions and promotions of function arguments.
 * 
 * @code
 * int foo (uint32_t x);
 * int bar (cxxomfort::Exactly<uint32_t> x);
 * 
 * foo(15); // wrong; compiles, but 15 is of type 'int'
 * foo(15,9); // wrong; compiles
 * 
 * bar(15); // ok; fails to compile
 * bar(15.9); // ok; fails to compile
 * bar(uint32_t(15)); // ok; compiles
 * @endcode
 * 
 * */
template <typename T>
class Exactly {
	public:
	#if (CXXOMFORT_CXX_STD >= 2011)
	Exactly (T x) noexcept : value(std::move(x)) {}
	#else
	Exactly (T x) : value(x) {}
	#endif
	
	operator T const& () const { return value; }

	private:
	template <typename U>
	Exactly(U);
	
	public:
	T value;
};


// use as f (T, Deleted = 0);

/** 
 * @cxxoutils Parameter type to pass to a function so that its signature is deleted.
 * 
 * @code
 * int bar (int a, int b);
 * int bar (int a, double b, cxxomfort::Deleted = 0);
 * 
 * bar(8, 15.0); // ok; fails to compile
 * bar(8, 15); // ok; compiles
 * @endcode
 * 
 * @see https://stackoverflow.com/questions/12877546/how-do-i-avoid-implicit-conversions-on-non-constructing-functions
 * */
class Deleted {
	private:
	Deleted (void*);
};


/**
 * @brief denotes a reference used as an output argument to a function.
 * @see https://foonathan.net/2016/10/output-parameter/
 * */
template <typename T>
class Outp {
	public:
	typedef T value_type;
#if (CXXOMFORT_CXX_STD >= 2011)
	Outp () = delete;
	~Outp () = default;
#endif
	explicit Outp (T& ref) CXXO_NOEXCEPTNOTHROW : t(ref) {}
	T& operator= (T const& v) { return t=v; }
	operator T& () CXXO_NOEXCEPT { return t; }
	operator T const& () const CXXO_NOEXCEPTNOTHROW { return t; }

	private:
	T& t;
};

}

// 
// composition types and tags 
// 

namespace cxxomfort { 

// a converse of nothrow_t, to mark functions that *can* throw when the default is not to

struct dothrow_t { };
#if defined(DOXYGEN_DOC)
//! @cxxoutils Tag dispatch identifier for @c dothrow marked functions.
static constexpr dothrow_t const dothrow_tag = {};
#else
namespace {
	CXXO_CONSTEXPR dothrow_t const& dothrow_tag = Static_Const<dothrow_t const>::value;
}
#endif

// a tag type to select constexpr overloads of some functions

struct constexpr_t { };
#if defined(DOXYGEN_DOC)
//! @cxxoutils Tag dispatch identifier for @c constexpr .
static constexpr constexpr_t const constexpr_tag = {};
#else
namespace {
	CXXO_CONSTEXPR constexpr_t const& constexpr_tag = Static_Const<constexpr_t const>::value;
}
#endif

/*
 * tag types for insert, remove operations
 * 
 * those are for being able to overload expressions such as
 * container[at_back] = 4; // PHP-like, equivalent to container.push_back(4);
 * 
 * */

struct at_front_t { };
struct at_back_t { };
//! @cxxoutils Tag dispatch identifier for operations at the front of a sequence.
static CXXO_CONSTEXPR at_front_t const at_front = {};
//! @cxxoutils Tag dispatch identifier for operations at the back of a sequence.
static CXXO_CONSTEXPR at_back_t const at_back = {};

// with_list_t
// a tag type to select constructors with initializer_list unambiguously

struct with_list_t { };
#if defined(DOXYGEN_DOC)
/** 
 * @cxxoutils Tag dispatch identifier for constructors using @c initializer_list .
 * 
 * @c initializer_list constructors are problematic due to ambiguity that is forced with C++11's not-so-uniform "\"uniform\"" initialization. To avoid ambiguities, every constructor that cxxomfort implements that uses @c initializer_list tags them with this extra parameter.
 * 
 * @code
 * @endcode 
 * 
 * */
static constexpr constexpr_t with_list_tag = {};
#else
namespace {
	CXXO_CONSTEXPR with_list_t const& with_list_tag = Static_Const<with_list_t const>::value;
}
#endif


//
// noncopyable
//

struct not_copy_constructible {
	#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_default_delete)
	not_copy_constructible (not_copy_constructible const&) = delete;
	#else
	private:
	not_copy_constructible (not_copy_constructible const&);
	#endif
};

//
// opaque_native
//

/**
 * @brief Wraps and tags a native C type as an opaque type.
 * @cxxoutils Wraps a native type to give it @c class - like properties.
 * @see https://nullptr.nl/2018/02/phantom-types/
 *
 * @c OpaqueNative takes a native C type, like @c int or @c float and wraps it with a @a Tag type as an @e opaque (named) type. The result is a type that can be used by explicitly accessing the underlying type where desired.
 *
 * The wrapper provides access to the underlying object via the @c value member, but the opaque interface is used via the @c value() and @c opaque() functions.
 *
 * Because the resulting type is a class type, it can participate in eg.: inheritance.
 *
 * @code
 * typedef OpaqueNative<double, height_tag> Height;
 * typedef OpaqueNativ<unsigned, age_tag> Age;
 *
 * Handle_type match_person (string name, Height h, Age a) { ... }
 *
 * match_person("John Doe", 32, 33); // fails
 * match_person("John Doe", opaque<Height>(32), opaque<Age>(33) ); // works
 * match_person("John Doe", opaque<Age>(33), opaque<Age>(33) ); // fails
 * @endcode
 *
 * */
template <typename Ty, typename Tag>
struct OpaqueNative {
    typedef Ty value_type;
    typedef Tag tag_type;
    Ty value;
/*
    constexpr
    explicit TaggedNative (Ty y) noexcept : value(y)
    { }

    constexpr
    explicit operator Ty () const noexcept { return value; }
*/
};

template <typename Ty, typename Tag>
CXXO_CONSTEXPR
Ty value (OpaqueNative<Ty,Tag> const t) CXXO_NOEXCEPTNOTHROW {
    return t.value;
}

template <typename Ty, typename Tag>
Ty& get(OpaqueNative<Ty,Tag>& t) CXXO_NOEXCEPTNOTHROW {
    return t.value;
}

template <typename Opaque, typename Ty>
CXXO_CONSTEXPR14
Opaque opaque (Ty y) CXXO_NOEXCEPTNOTHROW {
    OpaqueNative< typename Opaque::value_type, typename Opaque::tag_type >
    r = { y };
    return r;
}

}

// 
// extensible underlying type for enumeration types 
// 

namespace cxxomfort {

template <typename, typename=void> struct enumeration_underlying_type { typedef void type; };

}

// 
// absolute difference 
// 

namespace cxxomfort {

template <typename T>
T absdiff (T const& t1, T const& t2) {
	return t1 < t2 ? t2 - t1 : t1 - t2; 
}

template <typename T, typename Compare, typename Diff>
T absdiff (T const& t1, T const& t2, Compare less, Diff diff) {
	return less(t1,t2) ? diff(t2,t1) : diff(t1,t2);
}

}

//
// constexpr min, max
//

namespace cxxomfort {

//! constexpr variant of @c min()
//! @cxxoutils @c constexpr version of @c std::min()
template <typename T>
CXXO_CONSTEXPR
T min (constexpr_t, T a, T b) CXXO_NOEXCEPTNOTHROW {
	return b < a ? b : a;
}
//! constexpr variant of @c max()
//! @cxxoutils @c constexpr version of @c std::max()
template <typename T>
CXXO_CONSTEXPR
T max (constexpr_t, T a, T b) CXXO_NOEXCEPTNOTHROW {
	return a < b ? b : a;
}

} // cxxomfort

//
// tinyint_ impl
//

namespace cxxomfort {

template <typename Base>
struct tinyint_ {
	static_assert( sizeof(Base)==1, "only basic types signed char, unsigned char allowed");
	public:
#if (CXXOMFORT_CXX_STD>=2014)
	constexpr tinyint_ () noexcept = default;
	constexpr tinyint_ (tinyint_ const&) noexcept = default;
	constexpr tinyint_& operator= (tinyint_ const&) noexcept = default;
#elif (CXXOMFORT_CXX_STD>=2011)
	constexpr tinyint_ () noexcept = default;
	constexpr tinyint_ (tinyint_ const&) noexcept = default;
	tinyint_& operator= (tinyint_ const&) noexcept = default;
#else
	// tinyint_ (tinyint_ const& t) ; // autogenerated
	//tinyint_& operator= (tinyint_ const&) throw() ; // autogenerated
#endif
	CXXO_CONSTEXPR
	tinyint_ (Base const b= 0) CXXO_NOEXCEPTNOTHROW
	: v(b) {}
	template <typename Others> CXXO_CONSTEXPR
	explicit tinyint_ (Others const o) CXXO_NOEXCEPTNOTHROW
	: v(static_cast<Base>(o)) {}

	tinyint_& operator= (Base const b) CXXO_NOEXCEPTNOTHROW {
		v= b;
		return *this;
	}

	// tinyints are explicitly convertible to any type to which their base form casts
	template <typename T>
	CXXO_EXPLICIT_OPERATOR(T) () const CXXO_NOEXCEPTNOTHROW {
		return v;
	}
	CXXO_CONSTEXPR
	Base       value () const CXXO_NOEXCEPTNOTHROW { return v; }

	tinyint_ const& operator+ () const CXXO_NOEXCEPTNOTHROW { return *this; }
	CXXO_CONSTEXPR
	tinyint_   operator- () const CXXO_NOEXCEPTNOTHROW { return -v; }
	CXXO_CONSTEXPR
	tinyint_   operator~ () const CXXO_NOEXCEPTNOTHROW { return ~v; }
	CXXO_CONSTEXPR
	bool       operator! () const CXXO_NOEXCEPTNOTHROW { return !v; }

	tinyint_&  operator++ () CXXO_NOEXCEPTNOTHROW { ++v; return *this; }
	tinyint_   operator++ (int) CXXO_NOEXCEPTNOTHROW { tinyint_ ret(*this); ++*this; return ret; }
	tinyint_&  operator-- () CXXO_NOEXCEPTNOTHROW { --v; return *this; }
	tinyint_   operator-- (int) CXXO_NOEXCEPTNOTHROW { tinyint_ ret(*this); --*this; return ret; }

	friend  CXXO_CONSTEXPR
	bool operator== (tinyint_ const n1, tinyint_ const n2) CXXO_NOEXCEPTNOTHROW {
		return n1.v == n2.v;
	}
	friend CXXO_CONSTEXPR
	bool operator< (tinyint_ const n1, tinyint_ const n2) CXXO_NOEXCEPTNOTHROW {
		return n1.v < n2.v;
	}

	tinyint_& operator+= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v+= x2.v; return *this; }
	tinyint_& operator-= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v-= x2.v; return *this; }
	tinyint_& operator*= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v*= x2.v; return *this; }
	tinyint_& operator/= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v/= x2.v; return *this; }
	tinyint_& operator%= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v%= x2.v; return *this; }
	tinyint_& operator<<= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v<<= x2.v; return *this; }
	tinyint_& operator>>= (tinyint_ const x2) CXXO_NOEXCEPTNOTHROW { v>>= x2.v; return *this; }


	Base v;
};

typedef tinyint_<signed char> tinyint;
typedef tinyint_<unsigned char> tinyuint;

template <typename T> CXXO_CONSTEXPR
T to_integral (tinyint const x) CXXO_NOEXCEPTNOTHROW {
	return x.value();
}

template <typename T> CXXO_CONSTEXPR
T to_integral (tinyuint const x) CXXO_NOEXCEPTNOTHROW {
	return x.value();
}

CXXO_CONSTEXPR static inline
signed int to_value (tinyint const x) CXXO_NOEXCEPTNOTHROW {
	return x.value();
}

CXXO_CONSTEXPR static inline
unsigned int to_value (tinyuint const x) CXXO_NOEXCEPTNOTHROW {
	return x.value();
}

// relationals

CXXO_GENERATE_RELATIONALS(tinyint);
CXXO_GENERATE_RELATIONALS(tinyuint);


//
// bool_once
//

/**
 * @brief Manages a bool flag so that it evaluates to @c true once.
 * @param flag A @e writable @c bool type value.
 * @returns The previously held value of @a flag .
 * @throws Never throws.
 * @cxxoutils Manages a bool flag so that it evaluates to @c true once.
 *
 * The use case is such as performing repetitive tasks eg.: inside a loop, but where the very first iteration of the loop has a different, added code path:
 *
 * @code
 *
 * // before
 *
 * // this falls outside of the loop and can cause code desync due to repetition
 * repetitive_task(x, y, 0);
 * for (size_t idx= 1; idx < last; ++i) {
 * 		separation_task(w);
 * 		repetitive_task(x, y, idx);
 * }
 *
 * // after
 *
 * bool first= true;
 * for (size_t idx= 0; idx < last; ++idx) {
 * 		if (!bool_once(first)) {
 * 			separation_task(w);
 * 		}
 * 		// now the task is strictly part of the loop
 * 		repetitive_task(x, y, idx);
 * }
 * @endcode
 *
 * See also <tt>std::exchange()</tt>.
 *
 * */
bool bool_once (bool& flag) CXXO_NOEXCEPTNOTHROW {
	return flag ? (flag=false), true : false;
}

//
// assert_failure
//

struct assert_failure {
    template <typename F>
    explicit assert_failure (std::nullptr_t, F f) {
        f();
#if (CXXOMFORT_CXX_STD >= 2014)
        std::quick_exit(EXIT_FAILURE);
#else
        std::abort();
#endif
    }

    explicit assert_failure (const char *smsg) {
        using namespace std;
        fprintf(stderr, "Assertion failure: %s\n", smsg);
        abort();
    }
};

} // cxxomfort

//
// exchange, take (n3668, p2226)
//

#if (defined(DOXYGEN_DOC))
#elif (CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_template_aliases && CXXO_COMPILER_SUPPORT_constexpr)

	#include <utility>
	#include <cxxomfort/utils/traits_basic.hpp>
	#include <cxxomfort/impl/11-type_traits.hpp>

namespace cxxomfort {

template <typename T, typename U=T> inline
T exchange (T& obj, U&& new_value) noexcept(std::is_scalar<T>::value && std::is_scalar<U>::value) {
    T old_value = std::move(obj);
    obj = std::forward<U>(new_value);
    return old_value;
}

template <typename T> inline
T take (T& obj)
noexcept(std::is_nothrow_copy_assignable<T>::value) {
    T old_val = std::move(obj);
    obj = T();
    return old_val;
}

template <typename T> inline
void take_assign (T& target, T& source)
noexcept(std::is_nothrow_copy_assignable<T>::value)
{
    target = std::move(source);
    source = T();
}

}
#else

namespace cxxomfort {

template <typename T, typename U> inline
T exchange (T& obj, U& new_value) {
	T old_value= obj;
	obj= T(new_value);
	return old_value;
}

template <typename T> inline
T take (T& obj) {
	T old_val(obj);
	obj= T();
	return old_val;
}

template <typename T> inline
void take_assign (T& target, T& source) {
	target=source;
	source= T();
}

}

#endif

//
// coalesce , coalesce_ptrs
//

namespace cxxomfort {
#if defined(DOXYGEN_DOC)
/**
 * @brief Examines a series of pointers and returns the value in the first non-NULL element among them.
 * @param p1,p2,... A list of pointers to @a T .
 * @return The first p that is not NULL, or else the last p.
 * @throws Never throws.
 * @ingroup cxxomfort-utils
 *
 * The function is variadic in C++≥11; in versions before it can take up to 5 arguments.
 *
 * The funcion is @c noexcept in C++≥11, and marked @c throw() in versions before if the compiler supports it.
 *
 * */
template <typename T>
T const* coalesce_ptrs (T*... ptrs) noexcept {}
#elif (CXXOMFORT_CXX_STD >= 2011)
template <typename T>
T const* coalesce_ptrs (T const* p1, T const* p2, T const* args...) noexcept {
	return p1 ? p1 : p2 ? p2 : coalesce_ptrs(args);
}
template <typename T>
T const* coalesce_ptrs (T const* p1, T const* p2) noexcept {
	return p1 ? p1 : p2;
}
#else
template <typename T>
T const* coalesce_ptrs (T const* p1, T const* p2, T const* p3=nullptr, T const* p4=nullptr, T const* p5 = nullptr) CXXO_NOEXCEPTNOTHROW {
	return p1 ? p1 : p2 ? p2 : p3 ? p3 : p4 ? p4 : p5;
}

#endif

/**
 * @brief Coalesces a sequence, returning the first non-NULL element, or the last element.
 * @param ini,fin Iterators working as pointer-like elements.
 * @throws Never throws.
 * @ingroup cxxomfort-utils
 *
 * The function assumes pointer-like interface for the iterator type and thus never throws (is marked @c noexcept for C++≥11). Attempting to use it with an iterator type that can throw on dereference is unsupported.
 * */
template <typename Iter>
Iter coalesce (Iter ini, Iter fin) CXXO_NOEXCEPTNOTHROW {
	Iter preini= ini;
	for (; ini != fin; (preini= ini), ++ini) {
		if (ini) return ini;
	}
	return preini;
}

}

// 
// noexcept_call_cast - casts a function call to its C-compatible, noexcept nothrow version
// from https://thephd.dev/freestanding-noexcept-allocators-vector-memory-hole
// 

namespace cxxomfort {

#if (CXXOMFORT_CXX_STD >= 2014)
template <typename Fnc> struct noexcept_cast_helper;

template <typename Ret, typename... Args> struct noexcept_cast_helper<Ret(*)(Args...)> {
    using type = Ret(*)(Args...) CXXOMFORT_CXX17_CODE(noexcept,);
};

template <typename T> auto 
noexcept_call_cast (T const obj) noexcept {
    return reinterpret_cast<typename noexcept_cast_helper<T>::type>(obj);
};

#elif (CXXOMFORT_CXX_STD >= 2011)

template <typename Fnc> struct noexcept_cast_helper;

template <typename Ret, typename... Args> struct noexcept_cast_helper<Ret(*)(Args...)> {
    using type = Ret(*)(Args...) CXXOMFORT_CXX17_CODE(noexcept,);
};

template <typename T> 
typename noexcept_cast_helper<T>::type 
noexcept_call_cast (T const obj) noexcept {
    return reinterpret_cast<typename noexcept_cast_helper<T>::type>(obj);
};

#else

#endif

}


#if defined(DOXYGEN_DOC)
	#include <cxxomfort/impl/functionmeta.hpp>
	#include <cxxomfort/impl/static_math.hpp>
#elif (0 && CXXOMFORT_CXX_STD >= 1997)
	#include <cxxomfort/impl/functionmeta.hpp>
	namespace cxxomfort {
		using impl::functiontype;
		using impl::function_caller;
	}
	#include <cxxomfort/impl/static_math.hpp>
#endif

#include <cxxomfort/utils/pf-literals.hpp>
#include <cxxomfort/utils/array_ref.hpp>
#include <cxxomfort/impl/i12n.hpp>
#include <cxxomfort/utils/typelist.hpp>
#include <cxxomfort/impl/foreach.hpp>
#include <cxxomfort/impl/p0792-function_ref.hpp>

/**
 * @}
 *
 * */

//
// macros for using std::... annoyance
//

#define CXXO_USING_ALGORITHM() using std::find; using std::find_if; using std::find_if_not; using std::find_end; using std::count; using std::count_if; using std::transform; using std::copy; using std::copy_n; using std::copy_if; using std::copy_backward; using std::partition_copy; using std::lower_bound; using std::upper_bound; using std::minmax; using std::min_element; using std::max_element; using std::sort; using std::partial_sort; using std::for_each;
#define CXXO_USING_IOSTREAMS() using std::istream; using std::ostream; using std::cin; using std::cout; using std::endl;
#define CXXO_USING_IOMANIP() using std::hex; using std::dec; using std::setw; using std::boolalpha; using std::lowercase; using std::uppercase; using std::fixed; using std::scientific;
#define CXXO_USING_MATH() using std::exp; using std::log; using std::log2; using std::pow; using std::sin; using std::cos; using std::tan;
#define CXXO_USING_TUPLES() using std::pair; using std::tuple; using std::make_pair; using std::make_tuple; using std::get; using std::ignore; using std::tie; using std::make_from_tuple;


#endif

/**
 * @page cxxomfort-utils
 * @brief Cxxomfort independent utilities
 *
 * @c cxxomfort ships a few independent utilities that are intended to help write other parts of the code and assist with backports, or supplement some parts of the basic featureset when it can be done withotu depending on anything too advanced in the standard headers. These features are exposed by the header @c cxxo-utils.hpp for completeness.
 *
 * @code
 * #include <cxxomfort/cxxo-utils.hpp>
 * // alternatively included from general cxxomfort tools
 * #include <cxxomfort/backports.hpp>
 * @endcode
 *
 * The header is not intended to be directly included, but can be done so nonetheless. Instead, including the adequate header for each feature from <tt><cxxomfort/library/...></tt> guarantees their inclusion.
 *
 * All class/function features here are exposed in namespace @c cxxomfort .
 *
 * @section Interfaces
 *
 * The following is an incomplete listing of features available under this namespace:
 *
 * <b>Generalistic Tools</b>
 *
 * * @ref cxxomfort::rvoid "@c rvoid " - regular @c void type based on early @c p0146
 * * @ref cxxomfort::None "@c None " - generalised "None" Python-like keyword using @c rvoid .
 * * @ref @c cxxomfort::Static_Const<T> — wrapper around headerlib static objects to avoid ODR violation.
 * * @ref @c cxxomfort::prio - priority tag used for specialization and overload.
 * * @ref @c cxxomfort::dothrow — tag used to dispatch on cxxomfort's added throwable overloads.
 * * @ref @c cxxomfort::constexpr_tag — tag used to dispatch on cxxomfort's added @c constexpr overloads.
 * * @c noop_fn , @c noop() noop generalised function.
 * * @ref @c cxxomfort::bool_once() - checks and unsets a @c bool variable in one step
 * * @ref @c cxxomfort::Exactly - input parameter wrapper that prevents implicit conversions.
 * * @ref @c cxxomfort::Outpar - output parameter wrapper ordering.
 * * @ref @c cxxomfort::OpaqueNative — wraps a native type to give it class-like properties (so it can be inherited from etc).
 * * @ref cxxomfort::array_ref , @ref cxxomfort::array_ref_n — view over a contiguous data buffer.
 *
 * <b>Code Generation</b>
 *
 * * @c CXXO_GENERATE_RELATIONALS , @c CXXO_GENERATE_RELATIONALS2 - generate non.member relational operators given existing @c operator== and/or @c operator< .
 * * @c CXXO_OPEN_TYPESAFE_ENUM , @c CXXO_CLOSE_TYPESAFE_ENUM — macros to create typesafe enums implemented as native <tt>enum class</tt> if possible.
 * * @ref cxxo-foreach — Iteration over Sequences.
 * * @ref cxxo-i12n — Initialization of Sequences.
 *
 * <b>Utilities</b>
 *
 * * @ref @c constexpr_tag - a tag type similar to @c std::nothrow , to select constexpr function overloads
 * * @c coalesce() - coalesces a sequence of pointer-like objects.
 * * @c coalesce_ptrs() - return the first element in a collection of possibly-NULL pointers.
 * * @ref cxxomfort::siesta — Simple, throwable status codes, and @ref siesta.hpp "related utilities".
 * * @ref @c cxxomfort::array - wrapper over native arrays.
 * * @ref @c cxxomfort::array_ref - views over contiguous data.
 * * @ref @c cxxomfort::functiontype - template with information about function types.
 * * @ref @c cxxomfort::function_caller - wrap a C function as their own type.
 * * @ref @c cxxomfort::scoped_ptr - basic managed pointer
 * * @ref @c cxxomfort::basic_cstring_view and derived like @c cstring_view - views over native C strings.
 * * @ref @c cxxomfort::st_math - compile-time operations on integers, pre-C++11.
 * * @ref @c cxxomfort::typelist - implementation of typelists, and @ref typelist.hpp "related utilities". 
 * * @ref @c cxxomfort::typesafe_enum — wrap C native enums in a typesafe construct.
 *
 * */

