/**
 * @defgroup cxx03-backports C++03 Backports
 * @brief Backports from C++03 and C++ TR1 (aka "2007").
 * 
 * A listing of some of the features backported:
 * 
 * * From <tt><array></tt>: @ref cxxomfort::cxxostd::array , etc.
 * * From <tt><functional></tt>: @ref bind1st , @ref reference_wrapper , etc.
 * * A basic implementation of <tt>type_traits</tt> (not directly used).
 * 
 * These features are not used directly by cxxomfort, as it assumes that the underlying compiler is at least C+03/TR1 compatible. However, these facilities are offered for internal use by the library to reduce dependencies where there is no certainty that the compiler provides the adequate tooling in older versions (eg.: @c type_traits ), and to build up basic cxxomfort support in cases where TR1 is not available (see @c CXXOMFORT_USING_INTERNAL_TR1 ).
 *
 * See also: @ref cxx11-backports  .
 * 
 */

/**
 * @defgroup cxx11-backports C++11 Backports
 * @brief Backports from C++11.
 * 
 * C++11 adds a large number of features such as variadic templates, @c constexpr , static asserts and much improved template support and metaprogramming. While those changes can not directly be backported as they are implemented at the language and lexer level, the @e library level features that use them can be backported, to a point.
 * 
 * The following names and keywords are backported:
 * 
 * * <tt>#static_assert</tt> — compile-time assertion checks.
 * * @c std::nullptr_t , @c nullptr — typed null pointer literal.
 * 
 * The features backported from C++11 and added to 'namespace std' include but are not limited to the following names:
 * 
 * * @ref std0cstdint — The header proper and the names for the specific integer types.
 * * @ref std0cuchar — @c char16_t , @c char32_t , if they would be usable <b>as typedefs</b> in pre-C++11.
 * * @ref std0algorithm — A number of generic functions such as @c copy_if(), @c partition_copy(), @c minmax() , @c is_sorted() , etc. 
 * * @ref std0forward_list — The header proper and a basic implementation of singly-linked list.
 * * @ref std0functional — <tt>#reference_wrapper</tt>, @c ref() / @c cref() .
 * * @ref std0iterator — Global iterator accessors @c begin() , @c end() , rbegin() , @c rend() .
 * * @ref std0memory — @c addressof() , @c pointer_traits , @c allocator_traits , and @c unique_ptr .
 * * @ref std0numeric — iota() .
 * * @ref std0ratio — The header proper and the @c ratio class.
 * * @ref std0random — <tt>#shuffle_order_engine</tt> and @c knuth_b , <tt>#uniform_int_distribution</tt>, etc.
 * * @ref std0string — @c to_string() , @c stoul() and siblings, and where possible <tt>std::hash<></tt> for all available string types.
 * * @ref std0system_error — The header proper, the @c errc enumerator, and the basic utilities @c generic_error and @c system_error .
 * * @ref std0type_traits — @c enable_if , @c conditional ; the basic type_traits (already as part of the base Cxxomfort package); and some nothrow and triviality related traits.
 * * @ref std0utility — @c declval() , @c move() (for move emulation where possible).
 * 
 * @tableofcontents 
 * 
 * See also: @ref cxx03-backports @ref cxx14-backports @ref cxx17-backports .
 * 
 */

/**
 * @defpage cxx14-backports C++14 Backports
 * @brief Backports from C++14.
 * 
 * C++14 adds a number of features regarding algorithms and function object support.
 * In very general sense it is a "bugfix" upgrade to C++11, with the major game changer being the increased capability of @c constexpr support.
 * 
 * The features backported from this version and added to 'namespace std' include but are not limited to the following names:
 * 
 * * @ref std0algorithm — the full (heterogeneous) range versions of @c equal() and @c mismatch() .
 * * @ref std0functional — @ref cpp_functional_transparent_functors .
 * * @ref std0memory — @c make_unique() .
 * * @ref std0tuple — @ref cpp_tuple_get_type .
 * * @ref std0utility — @c integer_sequence and related types and helpers.
 * 
 * See also: @ref cxx11-backports @ref cxx17-backports .
 * 
 */

/**
 * @defgroup cxx17-backports C++17 Backports
 * @brief Backports from C++17.
 * 
 * C++17 adds a number of changes to the core language and the lexer than can not be backported, most importantly structured bindings and fold-expressions. Other than that, it adds a number of library-level changes that can be backported at about full usability level down to C++11, most notably @c std::invoke() .
 * 
 * The features backported from this version and added to 'namespace std' include but are not limited to the following names:
 * 
 * * @ref std0algorithm — @c clamp() function (including a helper for C++11 constexpr).
 * * @ref std0cstddef — @ref std::byte .
 * * @ref std0functional — @c not_fn() , @c invoke() (partial before C++11).
 * * @ref std0iterator — non-member container accessors @c data() , @c size() , @c empty() . 
 * * @ref std0numeric — @c gcd() / @c lcm() algorithms.
 * * @ref std0string_view — The header proper and its basic utilities.
 * * @ref std0tuple — @c make_from_tuple() , @c apply() .
 * * @ref std0type_traits — @c make_void , 'void_t' (down to C++11 only) , @c bool_constant , etc.
 * * @ref std0utility — @c as_const(), @c in_place_type , @c in_place_index etc.
 * 
 * C++17 also adds three new headers with "vocabulary types" - <tt><any></tt>, <tt><optional></tt> and <tt><variant></tt>. Backports for those are not part of cxxomfort's normal distribution, instead they are offered in their own package @c cxxo-vocabulary17 .
 * 
 * See also: @ref cxx14-backports @ref cxx20-backports .
 * 
 */

/**
 * @defgroup cxx20-backports C++20 Backports
 * @brief Backports from C++20.
 * 
 * C++20 includes some major changes to the language, such as modules, generic non-type template parameters, and a slightly different memory mode. There are also some library-level additions that can be backported fully down to C++17, or with various levels of conformance down to C++03.
 * 
 * The features backported from this version and added to 'namespace std' include but are not limited to the following names:
 * 
 * * @ref std0algorithm — Shift algorithms @c shift_left() , @c shift_right() ; overloads for consistent container erase-remove idiom.
 * * @ref std0bit — The header proper, and some basic bit operation functions; @c std::endian .
 * * @ref std0functional — The @c bind_front() helper (down to C++11 mostly).
 * * @ref std0iterator — The non-member helper @c ssize() .
 * * @ref std0type_traits — bound/unbound-array traits.
 * * @ref std0utility — safe integral comparisons.
 *
 * See also: @ref cxx17-backports @ref cxx23-backports .
 * 
 */

/**
 * @defgroup cxx23-backports C++23 Backports
 * @brief Backports from C++23.
 * 
 * The features backported from C++11 and added to 'namespace std' include but are not limited to the following names:
 * 
 * * From <tt><functional></tt>: <tt>invoke_r()</tt>.
 * * From <tt><type_traits></tt>: <tt>is_scoped_enum</tt>.
 * 
 * See also: @ref cxx20-backports .
 * 
 */

/**
 * 
 * @page cxxo-macros
 * @brief Listing of support macros
 * 
 * The following macros determine what features does cxxomfort implement and whether it does so by forwarding to native facilities or by providing its own forms of emulation.
 * 
 * The possible values for these macros are:
 * 
 * * @c CXXO_IMPLSTATUS_NATIVE() -- cxxomfort doesn't do anything; instead, the compiler's native facility is in use.
 * * @c CXXO_IMPLSTATUS_BACKPORT() -- cxxomfort provides a @em backport of the given feature, using other facilities either already native or existing in cxxomfort.
 * * @c CXXO_IMPLSTATUS_EMULATION() -- cxxomfort provides an alternative or @em emulation of the given feature that differs in syntax, semantics and capabilities; most likely using a different name.
 * * @c 0 -- the feature is not supported, or the required header has not been included.
 * 
 * 
 * See also:
 * 
 * * @sa std00attribute
 * * @sa std00featuremacros
 * * @sa cxxomfort-behaviour 
 * * @sa cxxomfort-codegen
 * 
 * */


/**
 * @page cxxomfort-behaviour Behaviour Configuration Macros
 * @brief Macros used to control cxxomfort's behaviour
 * 
 * @section featurelevel Feature Level Macros
 * 
 * * @ref @c CXXOFLAG_NO_STD_USING  - tells the compiler to avoid adding names to <tt>namespace std</tt> where possible.
 * * @ref @c CXXOFLAG_USE_INTERNAL_TR1 - tells the compiler to use its own TR1 implementation.
 * * @ref @c CXXOFLAG_NONOMINMAX - tells the compiler to not define @c NOMINMAX in Visual C++ where possible.
 * * @ref @c CXXOFLAG_WITHOUT_LONGLONG - tells the library that <tt>long long</tt> is not available (works only in pre-C++11 mode).
 * 
 * @section loglevel Logging Macros
 * 
 * * @c @ref CXXOMFORT_NOTICES  - controls log level. Defined as a numeric value, larger values increase the level of verbosity.
 * 
 * When @c CXXOMFORT_NOTICES is defined:
 * 
 * * A value of @c 1 will output basic notifications about library status and detected compiler, as well as detected C++ version.
 * * A value of @c 2 will @e also output notifications about what headers are being backported, and what fixes is the library adding for known issues (see @ref @c cxxomfort-fix ).
 * * A value of @c 3 or above will @e also output notifications about specific features being added or adjusted depending on each header.
 * 
 * * @ref @c CXXOFLAG_USE_INTERNAL_TR1 - enables (only in pre-C++03 mode) support for cxxomfort's backports to TR1 components.
 * 
 * In this release, cxxomfort's internal TR1 covers the following six components: 
 * 
 * * <tt>array<T,N></tt>
 * * <tt>reference_wrapper</tt>, @c ref() , @c cref() 
 * * <tt>mem_fn()</tt> (partial)
 * * <tt>bind()</tt> (partial)
 * * <tt>tuple<T1,...></tt> (partial) (up to 5 elements), @c make_tuple() , @c tie() , @c get() .
 * * Basic Category <tt>type_traits</tt>
 * 
 * For macros that pertain to code generation see @ref cxxomfort-codegen .
 * 
 * */



/**
 * @page cxxomfort-fixes "Implementation Fixes"
 * @brief Fixes to features in @c std 
 * 
 * Some features in <tt>namespace std</tt> are implemented incomplete, with issues or in unusable fashion in some compilers, and as such cxxomfort can not reuse the names or extend them (with member functions, for example) to provide corrected or updated functionality.
 * 
 * To deal with this, namespace <t>cxxomfort::fix</t>  offers a number of  alternatives or "forwarders" for the features in namespace std. These names either redirect to the @c std one if it's working, or to an implementation in cxxomfort when they are not.
 * 
 * Examples of features here:
 * 
 * * @link s_helper <tt>s("text")</tt> @endlink to construct @c basic_string s.
 * * @link sv_helper <tt>sv("text")</tt> @endlink to construct @c basic_string_view s.
 * * @link byte_conversion <tt>to_byte()</tt> @endlink - Allows @c std::byte construction pre-C++14 where uniform initialization for enum classes without constructors is not available.
 * * @link smartptr_accessors <tt>to_bool()</tt> for smart pointers @endlink to check their managed status.
 * * @link typetrait_accessors @c <tt>to_value()</tt> etc for @c integral_constant @endlink for versions of C++ where the integral conversion operator is not available.
 * * @link cxxomfort::fix::to_errc <tt>to_errc()</tt> @endlink - Allows @c std::errc code construction from integers pre-C++14.
 * * @ref cxxo_nonmember_string_accessors .
 * * @c common_type - Bypasses incorrect implementation in MSVC 2010 where it misses out variadic support and some specific decay case conventions.
 * * @c errc - Bypasses incorrect @c std::errc definition in MSVC versions due to it being declared there as a namespace.
 * * @c hash hash<> - An extended, extensible version of @c std::hash from <tt><functional></tt> that forwards to the std one but also can beb specialized for families ot types.
 * * @c randmo() - Obtains a "random" number the same way as @c rand does, except it discards some bias.
 * 
 * For a more comprehensive listing of fixes, check the various headers' notes on "<b>Interfaces not in std</b>".
 * 
 * Some of the fixes are detailed below:
 * 
 * <hr></hr>
 * 
 * @section cxxo_nonmember_string_accessors Nonmember helpers for string types
 * 
 * The <tt>basic_string</tt> interface is a mess (over 100 @e member functions). Also, the fact that the operations are defined as members means cxxomfort (nor any other library) can't extend or backport members added a posteriori, such as what's done in C++20.
 * 
 * cxxomfort gets around this by defining a number of <b>nonmember accessors</b> for the members in <tt>basic_string</tt> and <tt>basic_string_view</tt>, analogue to what <tt><iterator></tt> does in C++17 for containers general. This has the advantage that such interface is extensible.
 * 
 * The accessors are provided in namespace <tt>cxxomfort::string</tt> (maps to <tt>cxxomfort::fix::string</tt>) and include but are not limited to the following:
 * 
 * Assume that @c str1 is a @c basic_string , @c pos1 is an iterator of type of @a str1 's @c iterator or @c const_iterator , @c idx1 is an integral value of type of @a str1 's @c size_type , and @c char1 is a character value of the type of @a str1 's character value, or convertible. Similarly, assume that @c strlike1 is a custom string-like type implicitly convertible to <tt>std::basic_string_view</tt>.
 * 
 * 
 * * <tt>c_str (basic_string const&)</tt> 
 * * <tt>compare (basic_string const&, basic_string const&)</tt>
 * * <tt>compare (basic_string_view, basic_string_view)</tt>
 * * <tt>insert (str1&, pos1, strlike2)</tt>
 * * <tt>insert (str1&, pos1, char2, count2)</tt>
 * * <tt>erase (str1&, idx1, count1)</tt>
 * * <tt>erase (str1&, pos1, pos2)</tt>
 * * <tt>starts_with (basic_string_view str, basic_string_view prefix)</tt>
 * * <tt>ends_with (basic_string_view str, basic_string_view suffix)</tt>
 * * <tt>contains (basic_string_view str, basic_string_view needle)</tt>
 * 
 * @section fix_errc Fixes to <tt>std::errc</tt>
 * 
 * Some compilers, in particular MSVC 2010, 2012, have an erroneous definition of <tt>std::errc</tt> that makes it impossible to use errc as a type in function signatures or declare variables of this type. For example in MSVC 2010, it is defined as:
 * 
 * @code
 * namespace std {
 * namespace errc { ... }
 * }
 * @endcode
 * 
 * 
 * <hr></hr>
 * 
 * 
 * */



/**
 * @page std00attribute [[attribute]] support
 * @brief Features related to C++ Standard Attributes support
 * 
 * Cxxomfort adds basic support for tagging functions and entities with attributes.
 * 
 * The macros @c @ref CXXO_ATTRIBUTE and @c @ref CXXO_ATTR take arguments @a arg and expand in the following way:
 * 
 * * C++11 onwards: to an attribute in C++11 notation <tt>[[arg]]</tt>;
 * * pre-C++11, if the desired attribute is supported, to a compiler-specific expression that brands with the attribute, such as <tt>__attribute__((arg))</tt>.
 * * otherwise they expand to nothing.
 * 
 * @section std00attribute_interfaces Interfaces
 * 
 * The library provides the following interfaces:
 * 
 * @b @c CXXO_ATTRIBUTE 
 * 
 * This macro takes as arguments an attribute name and a series of arguments, and expands to an attribute expression with those arguments.
 * 
 * @code
 * CXXO_ATTRIBUTE(nodiscard,"because of side effects") int decl();
 * // with C++20 message support, expands to:
 * [[nodiscard("because of side effects")]] int decl();
 * // with only C++17 definition of the attribute, expands to:
 * [[nodiscard]] int decl();
 * // otherwise, it might expand to something like these:
 * __attribute__((nodiscard)) int decl();
 * int decl();
 * @endcode
 * 
 * As a general rule, <tt>CXXO_ATTRIBUTE(attr,a1,a2,...)</tt> expands to the chosen declaration <tt>CXXO_ATTRIBUTE_attr(a1,a2,...)</tt>. Invoking the macro for an attribute @a attr for which support is not provided is ill-defined.
 * 
 * @b @c CXXO_ATTR
 * 
 * This macro takes a single argument that is an attribute name, and expands to an attribute expression for that attribute.
 * 
 * @code
 * CXXO_ATTR(pure) int calculation (int a, int b);
 * // in a futre version of C++, this would expand to:
 * [[pure]] int calculation (int a, int b);
 * // otherwise, it might expand to something like these:
 * __attribute__((pure)) int calculation (int a, int b);
 * int calculation (int a, int b);
 * @endcode
 * 
 * Invoking the macro for an attribute @a attr for which support is not provided is ill-defined.
 * 
 * */


/**
 * @page std00featuremacros C++ Feature Macros
 * @brief Features related to SD6-style feature macros.
 * 
 * Starting with commit 2022-02-20, cxxomfort sets @link https://en.cppreference.com/w/cpp/feature_test feature testing macros @endlink for the features that the library implements (basically, a subset of those macros that begin with <tt>__cpp_lib</tt>). This allows third-party code and user code that uses the feature macros to also be able to detect that cxxomfort has added these capabilities.
 * 
 * Unless stated otherwise, cxxomfort sets the macro to the year of the value of the C++ standard detected (ie.: to <tt>(CXXOMFORT_CXX_STD*100)</tt>. This allows using relational comparisons at preprocessor level to determine if the feature comes native, or cxxomfort is implementing it.
 * 
 * @section langfeatures Language Features
 * 
 * In general terms, cxxomfort does not set language feature macros (ie.: those that start with <tt>__cpp_</tt> instead of <tt>__cpp_lib</tt>). An exception is made for the following <b>three</b> macros related to language features that cxxomfort more heavily uses, and that are set <b>only</b> when missing in compilers that otherwise advertise C++11 compatibility:
 * 
 * * <tt>__cpp_decltype</tt>, used for @ref CXXO_DECLTYPE() and @ref CXXO_AUTO() .
 * * <tt>__cpp_rvalue_references</tt>.
 * * <tt>__cpp_unicode_characters</tt>, used for @ref std0cuchar features.
 * 
 * In all these cases, when the feature is detected the macro is defined to the value <tt>1</tt>, so as to avoid issues with any other relational compare at the preprocessor level.
 * 
 * */

/**
 * @page cxxomfort-libraryfeaturemacros Cxxomfort Feature Macros
 * @brief Feature macros for the cxxomfort lirary
 * 
 * Starting with commit 2022-02-20, cxxomfort sets SD6-like feature macros for some extensions and features that are exposed over the @ref cxxomfort-library "Supplementals Section" . See also @ref std00featuremacros for a comparison.
 * 
 * Unless stated otherwise, cxxomfort sets the macro to the year of the value of the C++ standard detected (ie.: to <tt>(CXXOMFORT_CXX_STD*100)</tt>. This allows using relational comparisons at preprocessor level to determine if the feature comes native, or cxxomfort is implementing it.
 * 
 * 
 * */
