#ifndef CXXOMFORT_EXPECTED_HPP
#define CXXOMFORT_EXPECTED_HPP
/**
 * @file 
 * @brief Implementations and Backports for Standard header <tt><expected></tt>.
 * @ingroup std0expected
 * 
 */

#include <cxxomfort/utility.hpp>
#include <cxxomfort/cxxo-utils.hpp> // type traits
#include <limits> // numeric_limits
#if (CXXOMFORT_CXX_STD>=2011)
	#include <type_traits>
#endif

namespace cxxomfort { namespace cxxostd {

namespace {

template <typename T, bool=false> 
struct is_errnolike_impl {
	enum { value = false };
};

#if (CXXOMFORT_CXX_STD >= 2011)
template <typename T> 
struct is_errnolike_impl<T,true> {
	enum { value = (std::is_scalar<T>::value && (T()==T(0)) ) };
};

template <typename T> 
struct is_errno_like 
: std::integral_constant<bool, is_errnolike_impl<T, std::is_scalar<T>::value>::value > { };

#else
template <typename T> 
struct is_errnolike_impl<T,true> {
	enum { value = (::cxxomfort::traits::is_arithmetic<T>::value && (T()==T(0)) ) };
};

template <typename T> 
struct is_errno_like 
: std::integral_constant<bool, is_errnolike_impl<T, ::cxxomfort::traits::is_arithmetic<T>::value>::value > { };

#endif

} // detail



} } // cxxomfort::cxxostd

// 
// include header
// 

#if (defined(DOXYGEN_DOC))
namespace std {
	template <typename T, typename E> class expected {};
}

#elif (defined(__has_include))
	#if (__has_include(<expected>))
		#define CXXO_has_expected
		#include <expected>
	#endif
#endif

// 
// base impl
// 

#if (!defined(CXXO_has_expected))
	//#include <cxxomfort/memory.hpp>
	#include <cxxomfort/cstdio.hpp>
	#include <exception>
	#if (CXXOMFORT_CXX_STD>=2020)
		#include <variant>
	#endif

namespace cxxomfort { namespace cxxostd {

namespace {
	template <typename T, typename E>
	struct are_simple_storage 
	: std::integral_constant<bool, CXXOMFORT_CXX11_CODE(std::is_trivial<T>::value && std::is_trivial<E>::value , std::is_pod<T>::value && std::is_scalar<E>::value) > {};
}

struct unexpect_t {};
unexpect_t const unexpect = {};

template <typename E>
struct unexpected {
    E value;
    CXXO_CONSTEXPR 
    explicit unexpected (E const& e)
    : value(e) {}
};

template <typename E> CXXO_CONSTEXPR 
unexpected<E> make_unexpected (E const& e) { return unexpected<E>(e); }

template <typename>
struct bad_expected_access;

template <> struct bad_expected_access<void> 
: public std::exception {
	protected:
	bad_expected_access () {}
};

template <typename T, typename E, bool If=false>
struct expected_storage {
	std::pair<T,E> store; bool const B;
	expected_storage () : store() , B(true) {}
	explicit expected_storage(T const& tx) : store(tx,E()), B(true) {}
	explicit expected_storage(unexpected<E> const& ex) : store(T(), ex.value), B(false) {}
	
	CXXO_CONSTEXPR 
	T const& 	value () const CXXO_NOEXCEPT { return store.first; }
	CXXO_CONSTEXPR 
	E const&	error () const CXXO_NOEXCEPT { return store.second; }
};

#pragma pack(push)
#pragma pack(2)
template <typename T, typename E>
struct expected_storage<T,E,true> {
	union { T t; E e; } /*store*/; bool const B;
	expected_storage () : t(), B(true) {}
	explicit expected_storage(T const& tx) : t(tx) , B(true) {}
	explicit expected_storage(unexpected<E> const& ex) : e(ex.value), B(false) {}
	CXXO_CONSTEXPR 
	T const& 	value () const CXXO_NOEXCEPT { return t; }
	CXXO_CONSTEXPR 
	E const& 	error () const CXXO_NOEXCEPT { return e; }
};
#pragma pack(pop)


template <typename E> struct bad_expected_access 
: bad_expected_access<void> {
	bad_expected_access (E ev)
	: e(std::move(ev )) {}
	
	E const& error () const CXXO_NOEXCEPT { return e; }
	
	virtual char const* what () const CXXO_NOEXCEPTNOTHROW {
		return "bad_expected_access for an object that contains an error";
	}
	
	private:
	E e;
};

template <class T, class E, typename=void>
class expected {
	private:
	typedef expected_storage<T,E, are_simple_storage<T,E>::value> 	_store_t;
    public:
    static_assert(!std::is_array<E>::value, "E must not be an array type");
    typedef T   value_type;
    typedef E   error_type;
    typedef unexpected<E> unexpected_type;
    
    CXXO_CONSTEXPR expected (std::in_place_t = std::in_place) CXXO_NOEXCEPT
    : X() {}
    
    CXXO_CONSTEXPR 
    explicit expected (T const& t)
    : X(t) {}
    
    #if (CXXOMFORT_CXX_STD >= 2011)
    template <typename U, typename std::enable_if < 
		std::is_constructible<T,U>::value == true
     >::type >
    CXXO_CONSTEXPR 
    explicit expected (U&& v) 
    : X(std::forward<U>(v)) { }
    #endif
    
    #if (0 && CXXOMFORT_CXX_STD>=2011)
    template <typename... Args>
    CXXO_CONSTEXPR 
    explicit expected (std::in_place_t, Args&&... args)
    : X((std::forward<Args>(args)...)) { }
    #endif
    
    CXXO_CONSTEXPR 
    expected (unexpect_t const, E e)
    : X(unexpected<E>(e)) { }
    
    CXXO_CONSTEXPR 
    expected (unexpected<E> const& e)
    : X(e) { }
    
    CXXO_CONSTEXPR 
    CXXO_EXPLICIT_OPERATOR(bool) () const CXXO_NOEXCEPT { return has_value(); }
    
    CXXO_CONSTEXPR14 
    value_type const*   operator-> () const CXXO_NOEXCEPT {
        return has_value() ? &(X.value()) : nullptr; 
    }
    
    CXXO_CONSTEXPR14 
    value_type const&   operator* () const CXXO_NOEXCEPT {
        return X.value();
    }
    
    CXXO_CONSTEXPR 
    bool    has_value () const CXXO_NOEXCEPT { return X.B; }
    
    CXXO_CONSTEXPR14 
    value_type const&   value () const { 
		if (has_value()) return X.value(); else CXXO_THROW_STD( bad_expected_access<E>(error()) ); 
	}
    
    CXXO_CONSTEXPR 
    error_type const&   error () const { return X.error(); }
    
    friend bool 	operator== (expected const a, expected const b) CXXO_NOEXCEPT {
		return a.X==b.X;
	}

    friend bool 	operator!= (expected const a, expected const b) CXXO_NOEXCEPT {
		return !(a==b);
	}
    
    private:
    _store_t X;
};

} 
using ::cxxomfort::cxxostd::unexpected;

} // cxxomfort::cxxostd 

	#if (!defined(CXXOFLAG_NO_STD_USING))
namespace std {
	using ::cxxomfort::cxxostd::unexpect_t;
	using ::cxxomfort::cxxostd::unexpect;
	// using ::cxxomfort::cxxostd::unexpected; // taken pre-C++17
	using ::cxxomfort::cxxostd::make_unexpected;
	
	using ::cxxomfort::cxxostd::bad_expected_access;
	using ::cxxomfort::cxxostd::expected;
	
}
		#define __cpp_lib_expected (CXXOMFORT_CXX_STD*100)
	#endif


#endif


#if (defined(__cpp_lib_expected))
namespace cxxomfort { namespace fix {
	#if (CXXOMFORT_CXX_STD >= 2023)
	using ::std::unexpected;
	#else
	using ::cxxomfort::cxxostd::unexpected;
	#endif
} }
#endif


/**
 * @page std0expected <expected>
 * @brief Backports related to Standard header <tt><expected></tt> (C++23)
 * 
 * @code
 * #include <cxxomfort/expected.hpp>
 * // or as part of:
 * #include <cxxomfort/expected.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Interfaces defined in this section:
 * 
 * * @ref @c std::expected 
 * * @c std::make_unexpected() 
 * * the @c __cpp_lib_expected macro if @c expected is backported
 * 
 * Non-backport interfaces:
 * 
 * <tt>cxxomfort::fix::unexpected</tt> — access the <tt>std::unexpected</tt> type in C++23, or the equivalent type pre-C++23 where the name "unexpected" is taken by the termination function in <tt><exception></tt>.
 * 
 * @section std0expected_also See Also
 * 
 * @see @cppref{header/expected} (cppreference) 
 * 
 * */
#endif
