#ifndef CXXOMFORT_FORWARD_LIST_HPP
#define CXXOMFORT_FORWARD_LIST_HPP
/**
 * @file
 *
 * Interfaces defined in this header:
 *
 * * forward_list (from C++11)
 *
**/
// An implementation of forward_list<> based off
// my former ListaEnlazada class.
//
// Dependencies:
//  * cxxomfort
//  * allocator_traits en TR1


#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD >= 1997)

#include <cxxomfort/base.hpp>
#include <cxxomfort/config.hpp>
#include <iterator>
#include <memory>
#include <stdexcept>

#define CXXOMFORT_IMPLEMENTS_n2543 CXXO_IMPLSTATUS_BACKPORT()
#define CXXOMFORT_IMPLEMENTS_singly_linked_list CXXOMFORT_IMPLEMENTS_n2543

namespace cxxomfort { namespace cxxostd {

namespace detail_fwlist {
    struct sl_node_base {
        sl_node_base *next;
        CXXO_CONSTEXPR sl_node_base () CXXO_NOEXCEPTNOTHROW : next(nullptr) {}
        explicit sl_node_base (sl_node_base * p) CXXO_NOEXCEPTNOTHROW : next(p) {}
    };
    template <typename T> struct sl_node;
    template <typename T, bool Constness> class iterator_base;
}

/**
 * @class forward_list
 * @brief Implementation of C++11's forward_list , made compatible with C++03.
 * @cxx11b Singly-linked list implementation, see @cppref{container/forward_list} .
 * @xrefitem std0forward_list "" "" Singly-linked list implementation.
 *
 * Differences from the standard's provided:
 *
 * * Operator move-assignment is not available.
 * * Members merge, sort, unique and resize are not yet available.
 * * A member push_back is provided, which allows for tail insertion in O(1) time.
 * * Relational operators other than (in)equality are not provided.
 *
 */
template <typename T, typename A = std::allocator<T> >
class forward_list {
    private:
    typedef detail_fwlist::sl_node_base   node_t;
    typedef detail_fwlist::sl_node<T>     nodeT_t;
    typedef typename A::template rebind<nodeT_t>::other node_alloc_t;
    typedef std::allocator_traits<node_alloc_t> alloc_traits;


    public:
    typedef typename A::size_type size_type;
    typedef ptrdiff_t difference_type;
    typedef T                 value_type;
    typedef A                 allocator_type;
    typedef value_type&       reference;
    typedef const value_type& const_reference;
    typedef typename std::allocator_traits<A>::pointer pointer;
    typedef typename std::allocator_traits<A>::const_pointer const_pointer;
    typedef detail_fwlist::iterator_base<T,false>    iterator;
    typedef detail_fwlist::iterator_base<T,true>     const_iterator;

    private:
    CXXO_COPYABLE_MOVABLE(forward_list);

    public:
    
    //! Default constructor.
    forward_list () CXXO_NOEXCEPTNOTHROW 
    : _p_bb(nullptr), _p_last(nullptr),  alloc(A()) 
    { } ;
    //! Copy constructor.
    forward_list (forward_list const&);
    //! Allocator constructor.
    forward_list (A const&); 
    //! Allocator constructor.
    forward_list (forward_list const&, A const&);


    /**
     * @brief Sequence constructor. Create a list containing @a c copies of @a t .
     * @see @ref forward_list::assign()
     * */
    explicit forward_list
    (size_t c, T const& t= T(), A const& a = A());

    /**
     * @brief Sequence constructor. Create a list containing a copy of the sequence <tt>[ini,fin)</tt>.
     * */
    template <typename I> forward_list
    (I ini, I fin, A const& a = A());                     //< sequence-ctor
    
    forward_list (CXXO_RV_REF(forward_list));     //< move-ctor

#if (CXXOMFORT_CXX_STD >= 2011)
    explicit forward_list (std::initializer_list<T> const&);          //< init-list-ctor
#endif

    //! Destructor.
    ~forward_list ();                             //< destructor

    // Assignment
    ////////
    
    //! Assignment operator.
    forward_list& operator=
    (CXXO_COPY_ASSIGN_REF(forward_list));
    //! Move-Assign operator.
    forward_list& operator= (CXXO_RV_REF(forward_list));
    
    //! Assign a new list, containing @a c copies of @a t .
    void  assign (size_type c, T const& t); 
    //! Assign a new list, containing the elements in <tt>[ini,fin)</tt>.
    template <typename Iterator>
    void  assign (Iterator ini, Iterator fin);
    
#if (CXXOMFORT_CXX_STD >= 2011)
    forward_list& operator= (std::initializer_list<T> const&);        //< init-list assignment
    void  assign (std::initializer_list<T> const&);    //< assign a init-list
#endif

    allocator_type  get_allocator () const;  //< direct access to allocator
    bool       empty () const;          //< check if empty
    size_type  max_size () const;       //< max size


    const_reference front () const;     //< direct access to front
    reference       front ();           //< direct access to front
    
    const_iterator  cbegin () const CXXO_NOEXCEPT;     //< read-only iteration from begin
    const_iterator  begin ()  const CXXO_NOEXCEPT { return cbegin(); }
    const_iterator  cend ()   const CXXO_NOEXCEPT;     //< read-only iteration mark end
    const_iterator  end ()    const CXXO_NOEXCEPT { return cend(); }
    iterator        begin () CXXO_NOEXCEPT;            //< mutating iteration from begin
    iterator        end ()   CXXO_NOEXCEPT;            //< mutating iteration mark end
    
    //const_iterator  cbefore_begin () const;  //< read-only iteration mark before begin
    const_iterator  before_begin () const CXXO_NOEXCEPT;    //< read-only iteration mark before begin
    iterator        before_begin () CXXO_NOEXCEPT;          //< mutating iteration mark before begin

    // Insertions
    ////////
    
    //! Insert an element at the beginning.
    void  push_front (T const&);
    //! Insert an element at the beginning, via move / move emulation.
    void  push_front (CXXO_RV_REF(T));
    //! Insert an element at the end.
    void  push_back (T const&);
//    void  push_back (CXXO_RV_REF(T));
    
    //! Insert an element after the position of @a p .
    iterator   insert_after (const_iterator p, T const&);
    //! Insert an element after the position of @a p , via move / move emulation.
    iterator   insert_after (const_iterator p, CXXO_RV_REF(T));
    //! Insert @a c copies of @a t , after the position of @a p .
    iterator   insert_after (const_iterator p, size_type c, T const& t);
    //! insert copies of the sequence <tt>[ini,fin)</tt>, after the position of @a p .
    template< class I >
    iterator   insert_after (const_iterator pos, I ini, I fin);  //< insert a sequence

#if (CXXOMFORT_CXX_STD >= 2011)
    template <typename... Args>
    iterator emplace_after (iterator, Args&&...);
#endif

    // Deletions
    ////////
    
    //! Erase the element at the front.
    void  pop_front ();
    //! Erase the element after the position of @a p .
    iterator   erase_after (const_iterator);
    //! Erase the elements after the position of @a p1 , up to @a p2 .
    iterator   erase_after (const_iterator p1, const_iterator p2);
    //! Erase everything, resetting the list to its default state.
    void       clear ();


    // emplace_after( p, ...)

    // Sort
    ////////
    
    //! Sort by <tt>operator<()</tt>.
    void  sort ();
    //! Sort by the comparator @a lt .
    template <typename Compare>
    void  sort (Compare lt);

    // General Modifications
    ////////
    
    void       swap (forward_list&);    //< swap contents with another list
    void       resize (size_type);
    void       resize (size_type, T const&);

#if 0 // pending

    void  merge (forward_list&);
    void  merge (CXXO_RV_REF(forward_list));
    template <typename Compare>
    void  merge (forward_list&, Compare);
    template <typename Compare>
    void  merge (CXXO_RV_REF(forward_list), Compare);

    void  splice (const_iterator, forward_list&);
    void  splice (const_iterator, forward_list&, const_iterator src);
    void  splice (const_iterator, forward_list&, const_iterator ini, const_iterator fin);
#endif

    void       splice_after (const_iterator, forward_list &); // 1
    void       splice_after (const_iterator, forward_list &, const_iterator); // 2

    void       remove (T const&);
    template <typename Cond>
    void       remove_if (Cond c);

#if 0 // pending

    void  reverse ();

    void  unique ();
    template <typename Compare>
    void  unique (Compare);
#endif

    private:
    node_t          _p_bb;
    node_t *_p_last;
    node_alloc_t    alloc;

    static nodeT_t *N (node_t* p) { return static_cast<nodeT_t*>(p); }
    static node_t  *U (nodeT_t *p) { return static_cast<node_t*>(p); }

    node_t*    insert_after (node_t *p, T const&);
    node_t*    in_unlink_after (node_t *p);

    template <typename I> void in_uninitialized_copy(I,I);

    void            in_uninitialized_fill_n (size_t, T const&);
    void            in_link_after (node_t*, node_t*);
    void            in_readjust_last ();
    node_t*    in_destroy (nodeT_t*);
    void       in_report_list ();

} /* forward_list */ ;

template <typename T1, typename A1, typename T2, typename A2>
bool operator== (forward_list<T1,A1> const& a, forward_list<T2,A2> const& b);

template <typename T1, typename A1, typename T2, typename A2>
bool operator!= (forward_list<T1,A1> const& a, forward_list<T2,A2> const& b);


}} // cxxomfort::cxxostd

#include "impl/forward_list-impl.hpp"


#if (CXXOMFORT_CXX_STD>=2011)
    #include <forward_list>
#else

#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#else
    namespace std {
        using ::cxxomfort::cxxostd::forward_list;
    }
#endif // no std using

#endif // c++11 already has <forward_list>


#endif

#endif // c++1997

/**
 * @page std0forward_list <forward_list>
 * @brief Backports related to Standard Header <tt><forward_list></tt>
 * 
 * Interfaces defined in this section:
 * 
 * * @ref cxxomfort::cxxostd::forward_list  - Singly Linked List (C++11).
 * 
 * Note the following differences between the C++11 <tt>forward_list<></tt> and cxxomfort <tt>forward_list</tt>:
 * 
 * * cxxomfort uses move emulation for move constructor and move assignment.

 * @section also See Also
 * 
 * * @see @cppref{header/forward_list} 
 * 
 */
