#ifndef CXXOMFORT_IMPL_07_STD_ARRAY_HPP
#define CXXOMFORT_IMPL_07_STD_ARRAY_HPP
/*
 * @file
 */
#include <cxxomfort/base.hpp>

#define CXXOMFORT_IMPLEMENTS_n1479 CXXO_IMPLSTATUS_BACKPORT()
#undef  CXXOMFORT_IMPLEMENTS_array
#define CXXOMFORT_IMPLEMENTS_array CXXOMFORT_IMPLEMENTS_n1479

namespace cxxomfort {
namespace cxxostd {

namespace detail {
    template <typename T>
    struct add_pointer {
        typedef T* type;
    };
    
} // detail

/**
 * @ingroup std0array
 * @{
 */

/**
 * @brief std::array<T,N>  implementation
 * @tparam N the size of the array at compile time
 * @cxx03b Wraps a native array of type <tt>T[N]</tt>.
 * @xrefitem std0array "" "" Wraps a native array of type <tt>T[N]</tt>.
 * 
 * An implementation of C++03-TR1/C++11's <code>std::array</code> with some minor changes.
 * 
 * While this implementation is mostly intended for compilers that don't ship their own @c std::array , cxxomfort can also use it to supplement some features missing from @c std::array in given C++ versions, such as @c constexpr operator access. In this case, it's exposed also as @ref @c cxxomfort::array .
 * 
 * Main differences from TR1's array:
 * 
 * * <tt>operator[] (non-const)</tt> is @c constexpr when relaxed-constexpr support is detected.
 * * <tt>.front() const</tt> and <tt>.back() const</tt>  are @c constexpr when relaxed-constexpr
 * * <tt>.fill()</tt> is @c constexpr when relaxed-constexpr support is detected.
 * 
 * @sa https://wg21.link/n1479
 * 
 */
template <typename T, size_t N>
struct array {
    static_assert (N , "array<T,N> with N>0");
    
    typedef T value_type;
    typedef T const           const_value_type;
    enum{ const_size = N};

    typedef size_t            size_type;
    typedef ptrdiff_t         difference_type;

    typedef typename detail::add_pointer<T>::type    pointer;
    typedef typename detail::add_pointer<T const>::type const_pointer;
    
    typedef T&           reference;
    typedef T const&     const_reference;
    
    typedef pointer           iterator;
    typedef const_pointer     const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    
    
    //! Access to the underlying array
    CXXO_CONSTEXPR14
    pointer         data () CXXO_NOEXCEPTNOTHROW { return elems; }
    //! Access to the underlying array
    CXXO_CONSTEXPR 
    const_pointer   data () const CXXO_NOEXCEPTNOTHROW { return elems; }
    //! Size of the array (the same as the constant @a N )
    CXXO_CONSTEXPR 
    size_t     size () const CXXO_NOEXCEPTNOTHROW { return N; }
    //! Maximum size of the array (the same as @e size )
    CXXO_CONSTEXPR 
    size_t     max_size () const CXXO_NOEXCEPTNOTHROW { return N; }
    //! Returns @c true if the array is empty 
    CXXO_CONSTEXPR 
    bool       empty () const CXXO_NOEXCEPTNOTHROW { return N==0; }

	//! Direct access to the first element
    reference       front () CXXO_NOEXCEPTNOTHROW { return elems[0]; }
    CXXO_CONSTEXPR14 
    const_reference front () const CXXO_NOEXCEPTNOTHROW { return elems[0]; }
    //! Direct access to the last element
    reference       back () CXXO_NOEXCEPTNOTHROW { return elems[N-1]; }
    CXXO_CONSTEXPR14 
    const_reference back () const CXXO_NOEXCEPTNOTHROW { return elems[N-1]; }
    
	//! Indexed access (non-throwing)
    CXXO_CONSTEXPR14 
    reference       operator[] (size_t i) CXXO_NOEXCEPTNOTHROW { return elems[i]; }
    CXXO_CONSTEXPR 
    const_reference operator[] (size_t i) const CXXO_NOEXCEPTNOTHROW { return elems[i]; }
    
    //! Indexed access (throwing)
    reference       at (size_t i) { 
		return (i < size()) ? elems[i]  : throw std::out_of_range("array");
	}

    CXXO_CONSTEXPR14 
    const_reference at (size_t i) const { 
		if (i < size()) return elems[i];
		else throw std::out_of_range("array const");
	}
    
    iterator        begin () CXXO_NOEXCEPT { return elems; }
    CXXO_CONSTEXPR 
    const_iterator  begin () const CXXO_NOEXCEPT { return elems; }
    CXXO_CONSTEXPR 
    const_iterator  cbegin () const CXXO_NOEXCEPT { return elems; }
    iterator        end () CXXO_NOEXCEPT { return elems+N; }
    CXXO_CONSTEXPR 
    const_iterator  end () const CXXO_NOEXCEPT { return elems+N; }
    CXXO_CONSTEXPR 
    const_iterator  cend () const CXXO_NOEXCEPT { return elems+N; }
    
    reverse_iterator     rbegin() CXXO_NOEXCEPT { return reverse_iterator(elems+N); }
    CXXO_CONSTEXPR14 
	const_reverse_iterator     rbegin() const CXXO_NOEXCEPT { 
        return const_reverse_iterator(elems+N); 
    }
    reverse_iterator     rend() CXXO_NOEXCEPT { return reverse_iterator(elems); }
    CXXO_CONSTEXPR14 
    const_reverse_iterator     rend() const CXXO_NOEXCEPT { 
        return const_reverse_iterator(elems); 
    }
    
    CXXO_CONSTEXPR14 
    void fill (T const& t) {
        for (size_t i=0; i < N; ++i, elems[i]=t) {}
    }
    
    CXXO_CONSTEXPR14 
    friend bool operator== (array const& a1, array const& a2) {
        //return std::equal(a1.begin(), a1.end(), a2.begin());
        for (size_t i= 0; i < N; ++i) { if (a1[i]!=a2[i]) return false; }
        return true;
    }

    friend bool operator< (array const& a1, array const& a2) {
        return std::lexicographical_compare(a1.begin(), a1.end(), a2.begin(), a2.end());
    }

    public:
    T elems[N];
};

template <typename T, size_t N>
void swap (array<T,N>& li, array<T,N>& lr) {
    for (size_t i=0; i < N; ++i) { using std::swap; swap(li.elems[i], lr.elems[i]); }
}

/**
 * @brief Tuple-like interface to access an element in an @c array .
 * @tparam I Index to be accessed.
 * @returns Direct access to the element in the array.
 * */
template <size_t I, typename T, size_t N>
CXXO_CONSTEXPR14 
T& get (array<T,N>& a) CXXO_NOEXCEPTNOTHROW {
    static_assert(I<N, "invalid get<> in array");
    return a[I];
}

//! @overload get
template <size_t I, typename T, size_t N>
CXXO_CONSTEXPR14 
T const& get (array<T,N> const& a) CXXO_NOEXCEPTNOTHROW {
    static_assert(I<N, "invalid get<> in array const");
    return a[I];
}


/**
 * @}
 * */
} }

#endif
