#ifndef CXXOMFORT_IMPL_07_LCG_HPP
#define CXXOMFORT_IMPL_07_LCG_HPP
/**
 * @file
 * 
 * An implementation of LCG for std::linear_congruential_engine, drawing from stackoverflow and libcpp.
 * 
 */


#include <cxxomfort/config.hpp>
#include <cstddef>
#include <cassert>
#include <cxxomfort/impl/to_signed.hpp>
#include <limits>
#include <cxxomfort/limits.hpp>

namespace cxxomfort { namespace cxxostd { namespace random {

// 
// const_mod
// 

namespace detail {

//
// integer_log2
//

template <int p, int n> struct max_pow2_less {
    enum { c = 2*n < p };
    static const int value = c ? (max_pow2_less< c*p, 2*c*n>::value) : n;
};

template <> struct max_pow2_less<0, 0> {
    static const int value = 0;
};

template<int Shift> struct integer_log2_impl {
    template<class T>
    CXXOMFORT_CXX11_CODE(constexpr,) 
    static int apply(T t, int accum, int update = 0) {
        return update = ((t >> Shift) != 0) * Shift, integer_log2_impl<Shift / 2>::apply(t >> update, accum + update);
    }
};

template<> struct integer_log2_impl<1> {
    template<class T>
    CXXOMFORT_CXX11_CODE(constexpr,) static int apply(T t, int accum) {
        return int(t >> 1) + accum;
    }
};

template<class T>
CXXOMFORT_CXX11_CODE(constexpr,) int integer_log2(T t) {
    return integer_log2_impl<
        max_pow2_less<
            ::std::numeric_limits<T>::digits, 4
        >::value
    >::apply(t, 0);
}



struct div_t {
    uintmax_t quotient;
    uintmax_t remainder;
};

inline div_t muldivmod(uintmax_t a, uintmax_t b, uintmax_t m) {
    typedef std::numeric_limits<uintmax_t> NL;
    static const int bits = NL::digits / 2;
    static const uintmax_t mask = (uintmax_t(1) << bits) - 1;
    //typedef ::boost::uint_t<bits>::fast digit_t;
    typedef uintmax_t digit_t;

    int shift = NL::digits - 1 - integer_log2(m);

    a <<= shift;
    m <<= shift;

    digit_t product[4] = { 0, 0, 0, 0 };
    digit_t a_[2] = { digit_t(a & mask), digit_t((a >> bits) & mask) };
    digit_t b_[2] = { digit_t(b & mask), digit_t((b >> bits) & mask) };
    digit_t m_[2] = { digit_t(m & mask), digit_t((m >> bits) & mask) };

    // multiply a * b
    for(int i = 0; i < 2; ++i) {
        digit_t carry = 0;
        for(int j = 0; j < 2; ++j) {
            uint64_t temp = uintmax_t(a_[i]) * b_[j] + carry + product[i + j];
            product[i + j] = digit_t(temp & mask);
            carry = digit_t(temp >> bits);
        }
        if(carry != 0) {
            product[i + 2] += carry;
        }
    }

    digit_t quotient[2];

    if(m == 0) {
        div_t result = {
            ((uintmax_t(product[3]) << bits) | product[2]),
            ((uintmax_t(product[1]) << bits) | product[0]) >> shift,
        };
        return result;
    }

    // divide product / m
    for(int i = 3; i >= 2; --i) {
        uintmax_t temp = uintmax_t(product[i]) << bits | product[i - 1];

        digit_t q = digit_t((product[i] == m_[1]) ? mask : temp / m_[1]);

        uintmax_t rem = ((temp - uintmax_t(q) * m_[1]) << bits) + product[i - 2];

        uintmax_t diff = m_[0] * uintmax_t(q);

        int error = 0;
        if(diff > rem) {
            if(diff - rem > m) {
                error = 2;
            } else {
                error = 1;
            }
        }
        q -= error;
        rem = rem + error * m - diff;

        quotient[i - 2] = q;
        product[i] = 0;
        product[i-1] = (rem >> bits) & mask;
        product[i-2] = rem & mask;
    }

    div_t result = {
        ((uintmax_t(quotient[1]) << bits) | quotient[0]),
        ((uintmax_t(product[1]) << bits) | product[0]) >> shift,
    };
    return result;
}

inline uintmax_t muldiv(uintmax_t a, uintmax_t b, uintmax_t m)
{ return muldivmod(a, b, m).quotient; }

inline uintmax_t mulmod(uintmax_t a, uintmax_t b, uintmax_t m)
{ return muldivmod(a, b, m).remainder; }


template<class IntType, IntType m>
class const_mod {
public:
  static IntType apply(IntType x)
  {
    if(((unsigned_m() - 1) & unsigned_m()) == 0)
      return (unsigned_type(x)) & (unsigned_m() - 1);
    else {
      IntType supress_warnings = (m == 0);
      assert(supress_warnings == 0);
      return x % (m + supress_warnings);
    }
  }

  static IntType add(IntType x, IntType c)   {
    if(((unsigned_m() - 1) & unsigned_m()) == 0)
      return (unsigned_type(x) + unsigned_type(c)) & (unsigned_m() - 1);
    else if(c == 0)
      return x;
    else if(x < m - c)
      return x + c;
    else
      return x - (m - c);
  }

  static IntType mult(IntType a, IntType x)   {
    if(((unsigned_m() - 1) & unsigned_m()) == 0)
      return unsigned_type(a) * unsigned_type(x) & (unsigned_m() - 1);
    else if(a == 0)
      return 0;
    else if(a == 1)
      return x;
    else if(m <= traits::const_max/a)      // i.e. a*m <= max
      return mult_small(a, x);
    else if(traits::is_signed && (m%a < m/a))
      return mult_schrage(a, x);
    else
      return mult_general(a, x);
  }

  static IntType mult_add(IntType a, IntType x, IntType c) {
    if(((unsigned_m() - 1) & unsigned_m()) == 0)
      return (unsigned_type(a) * unsigned_type(x) + unsigned_type(c)) & (unsigned_m() - 1);
    else if(a == 0)
      return c;
    else if(m <= (traits::max()-c)/a) {  // i.e. a*m+c <= max
      IntType supress_warnings = (m == 0);
      assert(supress_warnings == 0);
      return (a*x+c) % (m + supress_warnings);
    } else
      return add(mult(a, x), c);
  }

  static IntType pow(IntType a, uintmax_t exponent)   {
      IntType result = 1;
      while(exponent != 0) {
          if(exponent % 2 == 1) {
              result = mult(result, a);
          }
          a = mult(a, a);
          exponent /= 2;
      }
      return result;
  }

  static IntType invert(IntType x)
  { return x == 0 ? 0 : (m == 0? invert_euclidian0(x) : invert_euclidian(x)); }

private:
  typedef ::cxxomfort::fix::integral_limits<IntType> traits;
  typedef typename ::cxxomfort::traits::make_unsigned<IntType>::type unsigned_type;

  const_mod();      // don't instantiate

  static IntType mult_small(IntType a, IntType x)   {
    IntType supress_warnings = (m == 0);
    assert(supress_warnings == 0 && "suppress_warnings in mult_small");
    return a*x % (m + supress_warnings);
  }

  static IntType mult_schrage(IntType a, IntType value)   {
    const IntType q = m / a;
    const IntType r = m % a;

    assert(r < q && "overflow check in mult_schrage");        // check that overflow cannot happen

    return sub(a*(value%q), r*(value/q));
  }

  static IntType mult_general(IntType a, IntType b)
  {
    IntType suppress_warnings = (m == 0);
    assert(suppress_warnings == 0 && "suppress_warnings in mult_general");
    IntType modulus = m + suppress_warnings;
    assert(modulus == m);
    if(uintmax_t(modulus) <=
        (::std::numeric_limits< uintmax_t>::max)() / modulus) {
      return static_cast<IntType>(uintmax_t(a) * b % modulus);
    } else {
      return static_cast<IntType>(mulmod(a, b, modulus));
    }
  }

  static IntType sub(IntType a, IntType b) {
    if(a < b)
      return m - (b - a);
    else
      return a - b;
  }

  static unsigned_type unsigned_m()   {
      if(m == 0) {
          typedef std::numeric_limits<IntType> L;
          return unsigned_type(L::max()) + 1;
      } else {
          return unsigned_type(m);
      }
  }

  // invert c in the finite field (mod m) (m must be prime)
  static IntType invert_euclidian(IntType c)
  {
    // we are interested in the gcd factor for c, because this is our inverse
    assert(c > 0 && "c > 0 in invert_euclidean");
    IntType l1 = 0;
    IntType l2 = 1;
    IntType n = c;
    IntType p = m;
    for(;;) {
      IntType q = p / n;
      l1 += q * l2;
      p -= q * n;
      if(p == 0)
        return l2;
      IntType q2 = n / p;
      l2 += q2 * l1;
      n -= q2 * p;
      if(n == 0)
        return m - l1;
    }
  }

  // invert c in the finite field (mod m) (c must be relatively prime to m)
  static IntType invert_euclidian0(IntType c)
  {
    // we are interested in the gcd factor for c, because this is our inverse
    assert(c > 0 && "c > 0 in invert_euclidean0");
    if(c == 1) return 1;
    IntType l1 = 0;
    IntType l2 = 1;
    IntType n = c;
    IntType p = m;
    IntType max = (std::numeric_limits<IntType>::max)();
    IntType q = max / n;
    assert((max % n != n - 1) && "c must be relatively prime to m.");
    l1 += q * l2;
    p = max - q * n + 1;
    for(;;) {
      if(p == 0)
        return l2;
      IntType q2 = n / p;
      l2 += q2 * l1;
      n -= q2 * p;
      if(n == 0)
        return m - l1;
      q = p / n;
      l1 += q * l2;
      p -= q * n;
    }
  }
};


// lcg_ta from libcpp 

template <unsigned long long __a, unsigned long long __c,
          unsigned long long __m, unsigned long long _Mp,
          bool _MightOverflow = (__a != 0 && __m != 0 && __m-1 > (_Mp-__c)/__a)>
struct lce_ta;

template <unsigned long long __a, unsigned long long __c, unsigned long long __m>
struct lce_ta<__a, __c, __m, (unsigned long long)(~0), true>
{
    typedef unsigned long long result_type;
    static inline result_type next(result_type _x) {
        // Schrage's algorithm
        const result_type __q = __m / __a;
        const result_type __r = __m % __a;
        const result_type __t0 = __a * (_x % __q);
        const result_type __t1 = __r * (_x / __q);
        _x = __t0 + (__t0 < __t1) * __m - __t1;
        _x += __c - (_x >= __m - __c) * __m;
        return _x;
    }
};

template <unsigned long long __a, unsigned long long __m>
struct lce_ta<__a, 0, __m, (unsigned long long)(~0), true>
{
    typedef unsigned long long result_type;
    //_LIBCPP_INLINE_VISIBILITY
    static result_type next(result_type _x)
    {
        // Schrage's algorithm
        const result_type _q = __m / __a;
        const result_type _r = __m % __a;
        const result_type _t0 = __a * (_x % _q);
        const result_type _t1 = __r * (_x / _q);
        _x = _t0 + (_t0 < _t1) * __m - _t1;
        return _x;
    }
};

template <unsigned long long __a, unsigned long long __c, unsigned long long __m>
struct lce_ta<__a, __c, __m, (unsigned long long)(~0), false>
{
    typedef unsigned long long result_type;
    //_LIBCPP_INLINE_VISIBILITY
    static result_type next(result_type _x)
    {
        return (__a * _x + __c) % __m;
    }
};

template <unsigned long long __a, unsigned long long __c>
struct lce_ta<__a, __c, 0, (unsigned long long)(~0), false>
{
    typedef unsigned long long result_type;
    //_LIBCPP_INLINE_VISIBILITY
    static result_type next(result_type _x)
    {
        return __a * _x + __c;
    }
};


} // detail

template<class UInteger, UInteger a, UInteger c, UInteger m_> 
class linear_congruential_engine {
    static_assert (::cxxomfort::traits::is_unsigned<UInteger>::value
        , "LCE requisite: 'UInteger' must be an unsigned integral type");
    static_assert ((m_ == 0) || (a < m_), "LCE requisite: a < m");
    static_assert ((m_ == 0) || (c < m_), "LCE requisite: c < m");

    public:
    typedef UInteger result_type;
    static const UInteger multiplier = a;
    static const UInteger increment = c;
    static const UInteger modulus = m_;
    static const UInteger default_seed= 1;

    static const result_type const_min = increment==0 ? 1 : 0;
    static const result_type const_max = modulus - 1;

    result_type     min() const CXXO_NOEXCEPTNOTHROW { return increment==0 ? 1 : 0; }
    result_type     max() const CXXO_NOEXCEPTNOTHROW  { return modulus - result_type(1); }

    //! Crea el generador con la semila @p x0 .
    explicit linear_congruential_engine () {
        this->seed(default_seed);
    }

    explicit linear_congruential_engine (result_type x0) { 
        this->seed(x0); 
    }

    linear_congruential_engine (linear_congruential_engine const& lce)
    : in_x(lce.in_x) {
        //std::cerr<< "[ lce::copy_constructor ]"<< std::endl;
    }

/*
    //! Crea el generador usando el generador @p g como semilla.
    template <typename Gen>
    linear_congruential_engine (Gen& g) { 
        //std::cerr<< "[ lce::constructor (Generator&) ]"<< std::endl;
        this->seed(g); 
    }
*/

    void seed (result_type s) CXXO_NOEXCEPT {
		in_x= s;
	}

    template <typename GEN>
    void seed_gen (GEN& g
/*
    , typename cxxomfort::traits::enable_if< (false == cxxomfort::traits::is_fundamental<GEN>::value), bool >::type* = 0
*/
    ) {
		//static_assert (cxxomfort::traits::is_fundamental<GEN>::value == false, "class");
		std::cout<< "GEN: "<< typeid(GEN).name()<< std::endl;
		result_type in_s= (::cxxomfort::traits::is_fundamental<GEN>::value) ? g % modulus : g() % modulus;
		//result_type	 in_s= g() % modulus;
		if (in_s == 0) in_s = default_seed;
        this->seed(in_s);
    }

    //! Obtiene el siguiente valor en la secuencia.
    result_type     operator() () {
		//in_x= (a * in_x) % m_;
		//in_x= (in_x + c) % m_;
        in_x = detail::const_mod<UInteger, m_>::mult_add(multiplier, in_x, increment);
        return in_x;
    }

    void discard (uintmax_t z) {
		while (z-->0) { this->operator()(); }
	}

    bool operator== (linear_congruential_engine const& argder) const {
        return in_x == argder.in_x;
    }

    bool operator!= (linear_congruential_engine const &argder) const {
        return !(*this == argder);
    }

/*
    template<typename C, typename CT>
    friend std::basic_ostream<C,CT>&  
    operator<< (std::basic_ostream<C,CT>& os, LC_TEMPLATE const& G) {
        G.stream_out(os);
        return os;
    }

    template <typename C, typename CT>
    friend std::basic_istream<C,CT>& 
    operator>> (std::basic_istream<C,CT>& is, LC_TEMPLATE & G) {
        G.stream_in(is);
        return is;
    }
*/

    //
    //
    private:

    UInteger in_x;

/*
    template<typename GEN>
    void seed (GEN& g, cxxomfort::traits::true_type);
    template<typename GEN>
    void seed (GEN& g, cxxomfort::traits::false_type);
*/

/*    
    template <typename C, typename CT>
    void stream_out (std::basic_ostream<C,CT> &os) const {
        os<< in_x;
    }
    template <typename C, typename CT>
    void stream_in (std::basic_istream<C,CT> &is) {
        UInteger x;
        if (is>> x) {
            if (this->min() <= x && x <= this->max()) {
                in_x= x;
            } else {
                is.setstate(std::ios_base::failbit);
            }
        }
    }
*/

};

/*
// Seed from a generator
template<class UInteger, UInteger a, UInteger c, UInteger m> 
template <typename Gen>
void linear_congruential_engine<UInteger, a, c, m>::seed (GEN& _seq, cxxomfort::traits::false_type) {
    std::cout<< "[ lce::seed(gen&,false_type) ]"<< std::endl;
    UInteger x0 = _seq();
    if ((detail::mod<UInteger, 1, 0, m>(c) == 0)
        && (detail::mod<UInteger, 1, 0, m>(x0) == 0))
      in_x = detail::mod<UInteger, 1, 0, m>(1);
    else
      in_x = detail::mod<UInteger, 1, 0, m>(x0);
}

template<class UInteger, UInteger a, UInteger c, UInteger m> 
template <typename GEN>
void linear_congruential_engine<UInteger, a, c, m>::seed (GEN& _g, cxxomfort::traits::true_type) {
    std::cout<< "[ lce::seed(gen&,true_type) ]"<< std::endl;
    return seed(static_cast<unsigned long>(_g));
}
*/


//! @c minstd_rand0
typedef linear_congruential_engine<uint32_t, 16807, 0, 2147483647> 
    minstd_rand0;
//! @c minstd_rand
typedef linear_congruential_engine<uint32_t, 48271, 0, 2147483647> 
    minstd_rand;

} // random

} } // cxxomfort::cxxostd

#endif
