#ifndef CXXOMFORT_IMPL_RANLUX_HPP
#define CXXOMFORT_IMPL_RANLUX_HPP
/**
 * @file
 */

#include "../config.hpp"
#include "../base.hpp"
#include <cxxomfort/impl/07-type_traits.hpp>
#include <cxxomfort/impl/11-conditional.hpp>

/* 
 * for C++14 onwards we can simply use the compiler-provided version 
 * but sometimes for compilers advertising C++11 implementation may be broken 
 * and even on TR1 some implementations like GCC 4.4-4.6 are broken.
 * */
#if (CXXOMFORT_CXX_STD >= 2014)
	// nothing to do
#elif (CXXOMFORT_CXX_STD >= 0)
	#include <cstddef>
	#include <limits>
namespace cxxomfort { namespace cxxostd { namespace random {

namespace detail {

template <class UInteger, int _w>
struct Shift { 
	enum { s = ((_w) < static_cast<UInteger>(std::numeric_limits<UInteger>::digits ) ) };
    static const UInteger value = s ? UInteger(1) << _w : UInteger(0); 
};

template <typename T, T a, T c, T m> 
struct ModF {
    static T calc(T t) {
        if (a == 1) t %= m;
        else {
            static const T hq = m / a;
            static const T hr = m % a;
            T t1 = a * (t % hq); 
            T t2 = hr * (t / hq);
            if (t2 <= t1) t = t1 - t2;
            else t = m - t2 + t1;
        }
        if (c != 0) {
            const T hd = m - t;
            if (c < hd) t += c;
            else t = c - hd;
        }
        return t;
    }
};

// Special case for m == 0 -- use unsigned integer overflow as modulo operator.
template<typename T, T a, T c, T m> 
struct ModT {
    static T calc (T t) { return (a * t) + c;   }
};


// Dispatch based on modulus value to prevent divide-by-zero compile-time
// errors when m == 0.
template<typename T, T a, T c, T m> inline 
T mod(T t) {
	return (m == 0) ? ModT<T,a,c,m>::calc(t) : ModF<T,a,c,m>::calc(t);
}


} // detail

template <typename II, size_t _w, size_t _s, size_t _r>
class subtract_with_carry_engine {
    static_assert (::cxxomfort::traits::is_unsigned<II>::value, "II must be unsigned type");
    public:
    typedef II  result_type;
    static const size_t word_size = _w;
    static const size_t short_lag = _s;
    static const size_t long_lag = _r;
    static const II default_seed = II(19780503u);
    static const II modulus = II(1) << _w;

    private:
#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG)
    typedef ::cxxomfort::cxxostd::random_renames::linear_congruential_engine<result_type, 40014u, 0u, 2147483563u> lcg_t;
#else
    typedef linear_congruential_engine<result_type, 40014u, 0u, 2147483563u> lcg_t;
#endif

    public:
	/**
	* @brief Constructs an explicitly seeded % subtract_with_carry_engine
	*        random number generator.
	*/
	explicit subtract_with_carry_engine (result_type sd = default_seed) { 
		seed(sd); 
	}

      /**
       * @brief Constructs a %subtract_with_carry_engine random number engine
       *        seeded from the seed sequence @p __q.
       *
       * @param __q the seed sequence.
       */
/*
    template<typename Sseq , typename E = typename
        detail::enable_if<!detail::is_same<_Sseq, subtract_with_carry_engine>::value, void>
        ::type >
    explicit subtract_with_carry_engine (Sseq& __q) { 
        static const bool valid= !std::is_same<Sseq, subtract_with_carry_engine>::value;
        static_assert (valid, "Not Sseq");
        seed(__q); 
    }
*/

      /**
       * @brief Seeds the initial state @f$x_0@f$ of the random number
       *        generator.
       *
       * N1688[4.19] modifies this as follows.  If @p val == 0,
       * sets value to 19780503.  In any case, with a linear
       * congruential generator lcg(i) having parameters @f$ m_{lcg} =
       * 2147483563, a_{lcg} = 40014, c_{lcg} = 0, and lcg(0) = value
       * @f$, sets @f$ x_{-r} \dots x_{-1} @f$ to @f$ lcg(1) \bmod m
       * \dots lcg(r) \bmod m @f$ respectively.  If @f$ x_{-1} = 0 @f$
       * set carry to 1, otherwise sets carry to 0.
       */
    void seed (result_type s0 = default_seed) {
		//std::cout<< "[seeding with "<< s0<< "]"<< std::endl;
		lcg_t lcg(s0 == 0u ? default_seed : s0);
		//std::cout<< "[lcg pass " << lcg<< "]"<< std::flush;
		 this->seed(lcg);
	}

      /**
       * @brief Seeds the initial state @f$x_0@f$ of the
       * % subtract_with_carry_engine random number generator.
       */


	template < typename GEN >
	void seed (GEN& gen) {
		for (size_t j = 0; j < long_lag; ++j) {
			//std::cout<< "["<< j<< "]"<< std::flush;
			M_x[j] = gen() % modulus;
		}
		M_carry = (M_x[long_lag-1] == 0);
		M_p = 0;
	}
  
    static result_type min()  { return result_type(0); }
    static result_type max()  { return detail::Shift<II, _w>::value - 1; /*(uintmax_t(1)<< _w) - 1; */  }
    static const result_type  static_min = 0;
    static const result_type  static_max= detail::Shift<II, _w>::value - 1; 

    void discard(unsigned long long z) {
        while (z-->0) { (*this)(); }
    }

    result_type operator() () {
		// Derive short lag index from current index.
		intmax_t psv = M_p - (intmax_t)short_lag;
		if (psv < 0)
			psv += long_lag;

	  // Calculate new x(i) without overflow or division.
	  // NB: Thanks to the requirements for UInt, M_x[_M_p] + _M_carry
	  // cannot overflow.
		II __xi;
		if (M_x[psv] >= M_x[M_p] + M_carry) {
			__xi = M_x[psv] - M_x[M_p] - M_carry;
			M_carry = 0;
		} else {
			__xi = (modulus /*detail::Shift<II, _w>::value*/ - M_x[M_p] - M_carry + M_x[psv]);
			M_carry = 1;
		}
		M_x[M_p] = __xi;

	  // Adjust current index to loop around in ring buffer.
		if (++M_p >= (intmax_t)long_lag) 
			M_p = 0;

		return __xi;
	}

    bool operator== (subtract_with_carry_engine const& rhs) const {
        return std::equal(M_x, M_x + long_lag, rhs.M_x);
    }

    private:
    II       M_x [long_lag];
    II       M_carry; // carry
    size_t     M_p; // k

};

typedef subtract_with_carry_engine<uint_fast32_t, 24, 10, 24> ranlux24_base;
typedef subtract_with_carry_engine<uint_fast64_t, 48,  5, 12> ranlux48_base;

} // random
} }

#endif

#endif // file
