#ifndef CXXOMFORT_IMPL_SHUFFLE_ORDER_HPP
#define CXXOMFORT_IMPL_SHUFFLE_ORDER_HPP

#include "../config.hpp"
#include "../base.hpp"
#include <cxxomfort/array.hpp>
#include <cxxomfort/iterator.hpp>
#include <cxxomfort/impl/11-conditional.hpp>
#include <algorithm>

/* 
 * for C++14 onwards we can simply use the compiler-provided version 
 * but sometimes for compilers advertising C++11 implementation may be broken 
 * and even on TR1 some implementations like GCC 4.4-4.6 are broken.
 * */
#if (CXXOMFORT_CXX_STD >= 2014)
	// nothing to do
#elif (CXXOMFORT_CXX_STD >= 0)
	#include <cstddef>
	#include <limits>
	#include <ctime>
// 
// shuffle_order_engine 
// 

namespace cxxomfort { namespace cxxostd { 

/**
 * @brief An adapter that caches and returns results from a generator in a random-ish order.
 * @cxx1b 
 * @xrefitem std0random "<random>" "Engnies"
 * @anchor shuffle_order_engine
 * 
 * Given an @a Engine such as @c ranlux24_base , the instantiation 
 * <code>Foo : shuffle_order_engine<ranlux24_base,44></code> creates an 
 * adaptor that acts as an Engine, except that it caches @a 44 values 
 * that are then returned randomly instead of in the order given by the 
 * generator. Each new invocation of <code>::operator()</code> replaces 
 * the value returned from the cache with a new value from the @a Engine .
 * 
 * C++11 defines a specific instantiation with the name @c knuth_b .
 * 
 */
template <typename Engine, size_t K>
class shuffle_order_engine {
    public:
    typedef typename Engine::result_type result_type;

    shuffle_order_engine ()
    : cache_(), engine_() {}

    shuffle_order_engine (result_type s)
    : cache_(), engine_(s) {
        using namespace std;
        generate_n (begin(cache_), K-1, engine_);
    }
    shuffle_order_engine (Engine const& e)
    : cache_(), engine_(e) {}


    void seed () { engine_.seed(); }
    void seed (result_type s) { engine_.seed(s); }

    Engine const& base () const CXXO_NOEXCEPT {
        return engine_;
    }

    result_type operator() () {
        result_type ri = engine_();
        size_t idx= (ri ^ std::time(0)) % K;
        std::swap(ri, cache_[idx]);
        return ri;
    }

#if (CXXOMFORT_CXX_STD >= 2011)
    CXXO_CONSTEXPR 
    result_type min () const CXXO_NOEXCEPTNOTHROW { return engine_.min(); }
    CXXO_CONSTEXPR 
    result_type max () const CXXO_NOEXCEPTNOTHROW { return engine_.max(); }
#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG)
    static result_type min () CXXO_NOEXCEPTNOTHROW { return Engine::min(); }
    static result_type max () CXXO_NOEXCEPTNOTHROW { return Engine::max(); }
#else
    result_type min () const CXXO_NOEXCEPTNOTHROW { return engine_.min(); }
    result_type max () const CXXO_NOEXCEPTNOTHROW { return engine_.max(); }
#endif

    private:
    std::array<typename Engine::result_type, K> cache_;
    Engine engine_;

};

typedef shuffle_order_engine<std::minstd_rand0, 256> knuth_b;

} }

#endif

#endif // file
