#ifndef CXXOMFORT_IMPL_07_TYPE_TRAITS_HPP
#define CXXOMFORT_IMPL_07_TYPE_TRAITS_HPP
/**
 * @file
 * 
 * This is an implementation of the basic type traits, starting with the featureset in TR1 (:2007) and some cxxomfort additions. It is made available so that some simple trait metaprogramming can be used without requiring the whole of <type_traits>.
 * 
 * 
 */

#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <cxxomfort/base/nullptr.hpp>
#include <cxxomfort/base/cstdint.hpp>
#include <cxxomfort/utils/traits_basic.hpp>

#if (CXXOMFORT_CXX_STD >= 2011)
	#include <type_traits>
#else

namespace cxxomfort { namespace traits {

// 
// rank = number of dimensions
// 
template <typename T> struct rank : integral_constant< size_t, 0> {};
template <typename T, size_t N> struct rank<T[N]> : integral_constant< size_t, rank<T>::value + 1> {};
#if (CXXOMFORT_CXX_STD<1997)
#else
template <typename T> struct rank<T[]> : integral_constant< size_t, rank<T>::value + 1> {};
#endif

//
// extent = size of topmost dimension
//
template<typename T, size_t N=0> struct extent 
: integral_constant<size_t, 0> {};
#if (CXXOMFORT_CXX_STD<1997)
#else
template<typename T>                    struct extent<T[],0> 
: integral_constant<size_t, 0> {};
template<typename T, unsigned N>        struct extent<T[],N> 
: integral_constant<size_t, extent<T,N-1>::value> {};
#endif
template<typename T, size_t N>          struct extent<T[N],0> 
: integral_constant<size_t, N> {};
template<typename T, size_t I, unsigned N> struct extent<T[I],N> 
: integral_constant<size_t, extent<T,N-1>::value> {};

//
// remove extent
//
template <typename T>
struct remove_extent { typedef T type; };
template <typename T, std::size_t N>
struct remove_extent<T[N]> { typedef T type;};
#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC)
#else
template <typename T>
struct remove_extent<T[]> { typedef T type; };
#endif

//
// remove all extents
//
template <typename T>
struct remove_all_extents { typedef T type;};
template <typename T, std::size_t N>
struct remove_all_extents<T[N]> {
    typedef typename remove_all_extents<T>::type type;
};
#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC)
#else
template <typename T>
struct remove_all_extents<T[]> {
    typedef typename remove_all_extents<T>::type type;
};
#endif



// 
// is_member_pointer and associates
// 

namespace detail {
// this gets a little wieldly, we handle up to 5 arguments (A0 to A4)

    template <typename T> struct f_helper : false_type {};
    template <typename R, typename Class>
    struct f_helper< R(Class::*)()> : true_type {};
    template <typename R, typename Class, typename A1>
    struct f_helper< R(Class::*)(A1)> : true_type {};
    template <typename R, typename Class, typename A1>
    struct f_helper< R(Class::*)(A1,...)> : true_type {};

	template <typename T> struct m_helper : false_type {};
	template <typename T, typename Class>
	struct m_helper< T Class::* > : true_type {};

} // member function pointer helper


template <typename T>
struct is_member_function_pointer 
: detail::f_helper< typename drop_cv<T>::type > {};

template <typename T>
struct is_member_pointer 
: detail::m_helper< typename drop_cv<T>::type > {};

template <typename T>
struct is_member_object_pointer
: integral_constant< bool, is_member_pointer<T>::value == true && is_member_function_pointer<T>::value == false> {};


// 
// traits that require compiler support
// 

#if (CXXOMFORT_CXX_STD >= 2011)
	#define Trait2(name,intr) 
	// nothing to do
	using std::is_enum;
	using std::is_class;
	using std::is_union;
	using std::is_pod;
	
#elif (CXXOMFORT_CXX_STD>=1997 && CXXOMFORT_CXX_EMULATION>=2007)

	#define Trait2(name,intr) template <typename T> struct name : integral_constant<bool, intr(T) ? true : false> {}

Trait2(is_enum, __is_enum);
Trait2(is_empty, __is_empty);
Trait2(is_union, __is_union);
Trait2(is_class, __is_class);
Trait2(is_pod, __is_pod);

	#undef Trait2

#elif (CXXO_COMPILER_SUPPORT_type_intrinsics)

	#define Trait2(name,intr) template <typename T> struct name : integral_constant<bool, intr(T) ? true : false> {}

Trait2(is_enum, __is_enum);
Trait2(is_empty, __is_empty);
Trait2(is_union, __is_union);
Trait2(is_class, __is_class);
Trait2(is_pod, __is_pod);

#else

	#define Trait2(name) template <typename T> struct name : false_type {}

Trait2(is_enum);
Trait2(is_empty);
Trait2(is_union);
Trait2(is_class);

// quick & dirty
template <typename T> struct is_pod : is_arithmetic<T> {};

#endif
#undef Trait2


/*
 * Advanced type properties
 */

// given the above, it should be possible to 
// implement is_scalar.
template<class T>
struct is_scalar 
: integral_constant < bool,
    is_arithmetic<T>::value || is_pointer<T>::value ||
    is_member_pointer<T>::value || is_enum<T>::value ||
    is_null_pointer<T>::value
> {};

// as well as is_object
template <class T> 
struct is_object 
: conditional< 
    !(is_function<T>::value 
     || is_reference<T>::value 
     || is_void<T>::value)
, true_type, false_type>::type     {};


} } // cxxomfort::traits

#endif // c++11
#endif
