#ifndef CXXOMFORT_IMPL_07_UNIFORM_ARITHMETIC_HPP
#define CXXOMFORT_IMPL_07_UNIFORM_ARITHMETIC_HPP
/**
 * @file
 */


#include <cxxomfort/config.hpp>
#include <cstddef>
#include <cassert>
#include <cxxomfort/impl/07-type_traits.hpp>

namespace cxxomfort { namespace cxxostd { namespace random {

template <typename Integer = int>
class uniform_int_distribution{
    static_assert (::cxxomfort::traits::is_integral<Integer>::value, "Integer must be an integer type");

    public:
    /** The type of the parameters of the distribution. */
    typedef Integer input_type;
    /** The type of the range of the distribution. */
    typedef Integer result_type;

    public:
    struct param_type {
        public:
        typedef uniform_int_distribution distribution_type;
        typedef Integer element_type;
        param_type (element_type a, element_type b) : x0_(a), x1_(b) {}

        private:
        element_type x0_, x1_;
    };

    //! Constructs an object
    explicit uniform_int_distribution (Integer m= 0, Integer M= 9)
    : M_min(m), M_max(M) {
        assert (m <= M && "'uniform_int_distribution' requires min <= max");
    }

    result_type min () const CXXO_NOEXCEPTNOTHROW { return M_min; }
    result_type max () const CXXO_NOEXCEPTNOTHROW { return M_max; }

    void reset () {
	}
    
    //! Obtiene un aleatorio distribuido en (min,max)
    template <typename URNG> 
    result_type operator () (URNG& u) {
        typedef typename URNG::result_type u_result_type;
        typedef typename ::cxxomfort::traits::is_integral<u_result_type> is_intg_t;
        static const bool is_intg= is_intg_t::value;
        static_assert (is_intg, "yay");
        return (is_intg 
        ? this->M_call (u, M_min, M_max, ::cxxomfort::traits::true_type() )
        : this->M_call (u, M_min, M_max, ::cxxomfort::traits::false_type() )
        );
    }

    //! Obtiene un aleatorio distribuido en (0, n) (para random_shuffle)
    template <typename URNG>
    result_type operator() (URNG& u, result_type n) {
        typedef typename URNG::result_type u_result_type;
        typedef typename ::cxxomfort::traits::is_integral<u_result_type> is_int_t;
        return this->M_call(u, 0, n-1, is_int_t());
    }

    param_type      param () const { return param_type(M_min, M_max); }
    

    // --- IMPL ---
    private:
    template <typename URNG> 
    static result_type M_call (URNG& u, result_type m, result_type M
    , ::cxxomfort::traits::false_type) {
        return result_type((u() - u.min())
            / (u.max() - u.min())
            * (M - m + 1)) + m
        ;
    }

    template <typename URNG> 
    static result_type M_call (URNG& u, result_type m, result_type M, 
    ::cxxomfort::traits::true_type) {
        typedef typename ::cxxomfort::traits::make_unsigned<typename URNG::result_type>::type urng_uresult_type;
        typedef typename ::cxxomfort::traits::make_unsigned<result_type>::type uresult_type;
        typedef typename ::cxxomfort::traits::conditional<(sizeof(urng_uresult_type) > sizeof(uresult_type))
            , urng_uresult_type, uresult_type
            >::type uctype;

        result_type ret;

        const urng_uresult_type u_min = u.min();
        const urng_uresult_type __urnmax = u.max();
        const urng_uresult_type __urnrange = __urnmax - u_min;
        const uctype prange = M - m;
        const uctype __udenom = (__urnrange <= prange
                       ? 1 : __urnrange / (prange + 1));
        do {
            ret = (urng_uresult_type(u()) -  u_min) / __udenom;
        } while (ret > M - m);

        return ret + m;
    }
    
    Integer M_min;
    Integer M_max;

}; //class uniform_int_distribution


template <typename Real = double>
class uniform_real_distribution {
    public:
    // types
    typedef Real input_type;
    typedef Real result_type;
    public:
    struct param_type {
        public:
        typedef uniform_real_distribution distribution_type;
        typedef Real element_type;
        param_type (element_type a, element_type b) : x0_(a), x1_(b) {}
        private:
        element_type x0_, x1_;
    };

    public:
    explicit uniform_real_distribution (Real m = Real(0), Real M = Real(1))
    : in_min(m), in_max(M) {
        assert (in_min <= in_max);
    }

    result_type     min() const CXXO_NOEXCEPTNOTHROW { return in_min; }
    result_type     max() const CXXO_NOEXCEPTNOTHROW { return in_max; }
    void            reset() { }
    
    param_type      param () const CXXO_NOEXCEPTNOTHROW { return param_type(in_min,in_max); }
    
    template <class URNG>
    result_type operator() (URNG& urng) {
        result_type numerator = 
        static_cast<result_type>(urng() - urng.min());
        result_type divisor = 
        static_cast<result_type>(urng.max() - urng.min());
        //std::cout<< "[divisor "<< divisor<< " ]"<< std::flush;
        return numerator / divisor * (in_max - in_min) + in_min;
//        return (urng() * (in_max - in_min)) + in_min;
    }

    private:
    Real in_min;
    Real in_max;
};



} // random

} } // cxxomfort::cxxostd


#endif
