#ifndef CXXOMFORT_IMPL_ALIGNED_STORAGE_HPP
#define CXXOMFORT_IMPL_ALIGNED_STORAGE_HPP
/**
 * @file
 * @brief Implementation of std::aligned_storage for C++03.
 * @author Luis Machuca Bezzaza <luis [dot] machuca [at] gulix [dot] cl>
 *
 */

#include "../config.hpp"

/* For aligned_storage, we can usually do well if we have __align(N) available. 
 * That however is an intrinsic, so it has to be detected on a per-compiler basis.
 * For compilers without align(N) or __alignof(), we can usually do "well enough" by reserving 
 * [2*sizeof(T)+1] for compound types, and [sizeof(T)] for fundamental types, but we do give away the requirement that the resulting storage is of size N.
 */

#undef  CXXOMFORT_IMPLEMENTS_aligned_storage
#if (CXXOMFORT_CXX_STD>=2011)
	// nothing to do
	#define CXXOMFORT_IMPLEMENTS_aligned_storage CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILER_SUPPORT_alignN)
	// compiler supports __align(N)
	#define CXXOMFORT_IMPLEMENTS_aligned_storage (CXXO_IMPLSTATUS_EMULATION())

namespace cxxomfort { namespace cxxostd {

template <size_t S, size_t A=sizeof(int)>
struct aligned_storage {
	typedef struct { unsigned char align(A) data[S]; } 
	type;
};


} }

// 
// GCC, Clang
//
#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC) \
 || (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG)

	#define CXXOMFORT_IMPLEMENTS_aligned_storage (CXXO_IMPLSTATUS_EMULATION()+1)

namespace cxxomfort { namespace cxxostd {

    #if (CXXOMFORT_COMPILER_VERSION<=406) // align(N) can only take literals


        #if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES > 1)
            #pragma message CXXO_NOTICE("aligned_storage: A")
        #endif

template <size_t S, size_t A = alignof(long double)> struct aligned_storage;

#define ALIGNED_STORAGE_ALIGN(N) \
template <size_t S> struct aligned_storage<S, N > { \
    typedef struct { unsigned char __attribute__((aligned( N ))) data[S]; } \
    type; \
}
ALIGNED_STORAGE_ALIGN(1);
ALIGNED_STORAGE_ALIGN(2);
ALIGNED_STORAGE_ALIGN(4);
ALIGNED_STORAGE_ALIGN(8);
ALIGNED_STORAGE_ALIGN(16);
ALIGNED_STORAGE_ALIGN(32);
ALIGNED_STORAGE_ALIGN(64);
ALIGNED_STORAGE_ALIGN(128);
ALIGNED_STORAGE_ALIGN(256);
	#undef ALIGNED_STORAGE_ALIGN
    #define CXXO_DEFINED_ALIGNED_STORAGE

    #else // gcc > 4.6

        #if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES > 1)
            #pragma message CXXO_NOTICE("aligned_storage: B")
        #endif

template <size_t S, size_t A = alignof(char[S])> struct aligned_storage {
    typedef struct { unsigned char __attribute__((aligned(A))) data[S]; }
    type;
};

        #if (CXXOMFORT_CXX_EMULATION<2011)
            #define CXXO_DEFINED_ALIGNED_STORAGE
        #endif

    #endif // gcc version

} }

#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC) 
// 
// MSC, declspec() can only take literals
// 

	#define CXXOMFORT_IMPLEMENTS_aligned_storage (CXXO_IMPLSTATUS_EMULATION()+2)

namespace cxxomfort { namespace cxxostd {

template <size_t I> struct np2 {
    static size_t const n0 = I-1;
    static size_t const value = (n0 | (n0>>1UL) | (n0>>2UL) | (n0>>4UL) | (n0>>8UL) | (n0>>16UL) | (n0>>32UL) ) +1;
};

template <size_t S, size_t A =( S > sizeof(long double) ? alignof(long double) : alignof(cxxomfort::impl::pack_t<sizeof(int)>)) > struct aligned_storage;
	#define ALIGNED_STORAGE_ALIGN(N) \
template <size_t S> struct aligned_storage<S, N > { \
    typedef struct __declspec(align( N )) { unsigned char elem[S]; } \
    type;  \
}

ALIGNED_STORAGE_ALIGN(1);
ALIGNED_STORAGE_ALIGN(2);
ALIGNED_STORAGE_ALIGN(4);
ALIGNED_STORAGE_ALIGN(8);
ALIGNED_STORAGE_ALIGN(16);
ALIGNED_STORAGE_ALIGN(32);
ALIGNED_STORAGE_ALIGN(64);
	#undef ALIGNED_STORAGE_ALIGN
	#define CXXO_DEFINED_ALIGNED_STORAGE

} }

#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC)
	#define CXXOMFORT_IMPLEMENTS_aligned_storage (CXXO_IMPLSTATUS_EMULATION()+3)
    #define CXXO_DEFINED_ALIGNED_STORAGE
	// only supports alignment via pragma

namespace cxxomfort { namespace cxxostd {

template <size_t S, size_t A=sizeof(int) > struct aligned_storage;

#pragma pack(push)
#pragma pack(1)
template <size_t S> struct aligned_storage<S,1> {
	typedef struct { unsigned char data[S]; } type;
};
#pragma pack(2)
template <size_t S> struct aligned_storage<S,2> {
	typedef struct { union { unsigned char data[S]; uint16_t dummy; } u; } type;
};
#pragma pack(4)
template <size_t S> struct aligned_storage<S,4> {
	typedef struct { union { unsigned char data[S]; uint32_t dummy; } u; } type;
};
#pragma pack(8)
template <size_t S> struct aligned_storage<S,8> {
	typedef struct { unsigned char data[S]; } type;
};
#pragma pack(pop)

} } 

////
#else
	#define CXXOMFORT_IMPLEMENTS_aligned_storage (0)
namespace cxxomfort { namespace cxxostd {
	template <size_t S, size_t A> struct aligned_storage; // not implemented
} } 
#endif // std / compiler version


namespace cxxomfort {
namespace cxxostd {


#if (0 && !defined(CXXOMFORT_ALIGNMENT_TOOLS_SUPPORTED))
template <size_t S, size_t A = -1> struct aligned_storage {
    typedef union { 
        unsigned char storage[S]; 
        struct struct_;
        double dd[2]; 
        long double ld;
        double d;
        long l;
        int i;
        char c;
        void (*fn_a)();
        int struct_::memobj;
        double (struct_::*memfn)(int);
    } type;
};
#endif


} // memory
} // cxxomfort

#if (CXXOMFORT_CXX_STD < 2011 && CXXOMFORT_CXX_EMULATION<2011 && CXXOMFORT_IMPLEMENTS_aligned_storage>0 && CXXOMFORT_IMPLEMENTS_aligned_storage<CXXO_IMPLSTATUS_NATIVE())
namespace std {
    using cxxomfort::cxxostd::aligned_storage;
}
#endif


#endif // file
