#ifndef CXXOMFORT_IMPL_STD_COMMON_TYPE_HPP
#define CXXOMFORT_IMPL_STD_COMMON_TYPE_HPP
/**
 * @file
 * */

#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
    #pragma message CXXO_NOTICE("common_type implementation")
#endif

#include <cxxomfort/config.hpp>

#if defined(DOXYGEN_DOC)

namespace std {

/**
 * @brief Determines the common type between two types, if any.
 * @ingroup std0typetraits
 * @cxx11b Determines the common convertible type between two types, if any.
 * @xrefitem std0typetraits "" "" Determines the common convertible type between two types, if any.
 * 
 * For implementation of @c common_type the cxxomfort library switches 
 * between a number of approaches depending on the level of support 
 * that is found.
 * 
 * * For C++>=11 cxxomfort just forwards to @c std::common_type , guaranteed to work.
 * * For compilers that support at least support variadic templates and @c decltype , a straight backport is provided.
 * * For compilers that support @c decltype or @c typeof , without variadics, we provide a limited version that takes up to 5 types.
 * * For any other compiler, we offer a much more limited version that works only with fundamental types and takes up to 5 types.
 * 
 * Without at least @c typeof support, the most that can be done is using eg.: @c std::is_convertible to check possible type conversions, although this can give false results since in those earlier systems @c is_convertible is also incomplete or is missing. 
 * 
 * @subsection common_type_stdname Usage of name in namespace std 
 * 
 * All implementations still retain the possibility to be extended by user, by specializing the main template. For this reason, and unlike some other features in cxxomfort, cxxomfort's @c commong_type is <b>defined in namespace std</b> regardless of wether @c CXXOFLAG_NO_STD_USING is defined.
 * 
 * @see @c cxxomfort::fix::common_type .
 * 
 * 
**/
template <typename... Types>
struct common_type {
    typedef _implementation_defined_ type;
};

}

#elif (CXXOMFORT_CXX_STD >= 2011)

	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type CXXO_IMPLSTATUS_NATIVE()
	#define CXXOMFORT_FIX_common_type 0

	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("std::common_type")
	#endif

namespace cxxomfort { namespace fix {
	using ::std::common_type;
}}

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_GCC,>=405) && CXXO_COMPILERREQ(CXXO_COMPILERID_GCC,<407) && CXXOMFORT_CXX_EMULATION==2011)
	// GCC works but has a non-implemented issue in some C++0x versions (4.5 and 4.6)
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type CXXO_IMPLSTATUS_NATIVE()
	#define CXXOMFORT_FIX_common_type 0
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("common_type fix for older GCC")
	#endif

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1800))
	// MSVC variadic common_type starting with 2015
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type CXXO_IMPLSTATUS_NATIVE()
	#define CXXOMFORT_FIX_common_type 0
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("MSVC variadic common_type")
	#endif

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1700))
	// MSVC fixes common_type starting MSVC 2013
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type CXXO_IMPLSTATUS_NATIVE()
	#define CXXOMFORT_FIX_common_type 0
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("MSVC non-variadic common_type")
	#endif


#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600))
	// MSVC broken std::common_type in MSVC 2010
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type (CXXO_IMPLSTATUS_EMULATION()+2)
	#define CXXOMFORT_FIX_common_type 1
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("MSVC broken common_type 2010")
	#endif

#elif ((CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof) && CXXO_COMPILER_SUPPORT_variadic)
	// c++03 with at least __typeof__ and variadic

	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("common_type variadic + typeof/decltype")
	#endif
	#include "./11-decay.hpp" // decay
	#include <cxxomfort/utility.hpp> // declval
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type (CXXO_IMPLSTATUS_BACKPORT())
	#define CXXOMFORT_FIX_common_type 0

namespace std {

template <typename A1, typename A2, typename... Args>
struct common_type 
: public common_type < typename common_type<A1,A2>::type, Args...> {}

template <typename A1, typename A2>
struct common_type <A1,A2> {
	typedef typename std::decay<A1>::type D1;
	typedef typename std::decay<A2>::type D2;
	typedef CXXO_TYPEOF(true ? std::declval<D1>() : std::declval<D2>()) type;
};

// as per cppreference
template <typename A1>
struct common_type <A1> : common_type<A1,A1> {};

} // std

#elif ((CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof))
	// decltype or typeof, but no variadics (eg.: MSVC 2012 to 2015)
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("common_type typeof/decltype")
	#endif
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type (CXXO_IMPLSTATUS_EMULATION()+2)
	#define CXXOMFORT_FIX_common_type 0

	#include "./11-decay.hpp"
	#include <cxxomfort/utility.hpp> // std::declval
	#include <ciso646>

namespace std {

template <typename A1, typename A2, typename A3=A2, typename A4=A2, typename A5=A2>
struct common_type 
: public common_type < typename common_type<A1,A2>::type, A3, A4, A5, A5> 
{ };

template <typename A1, typename A2>
struct common_type <A1,A2 /*defaults*/ > {
	typedef typename std::decay<A1>::type D1;
	typedef typename std::decay<A2>::type D2;
	typedef CXXO_TYPEOF(true ? std::declval<D1>() : std::declval<D2>()) type;
};

// as per cppreference
//template <typename A1>
//struct common_type <A1> : public common_type<A1,A1> 
//{ };

} 

#else // no support
	#undef  CXXOMFORT_IMPLEMENTS_common_type 
	#define CXXOMFORT_IMPLEMENTS_common_type (CXXO_IMPLSTATUS_EMULATION())
	#define CXXOMFORT_FIX_common_type 0
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("common_type limited")
	#endif

	#include "../impl/11-decay.hpp"
	#include "../impl/11-conditional.hpp"

namespace std {

template <typename T1, typename T2, typename T3=T2, typename T4=T2, typename T5=T2>
struct common_type 
: public common_type < typename common_type<T1,T2>::type, T3, T4, T5, T5> 
{ };

/* using is_convertible to convert */

template <typename T1, typename T2> 
struct common_type <T1,T2> {
	private:
	struct common_type_cant_resolve; 
	// template <typename Q> static typename std::add_reference<Q>::type dv ();
	enum convvalues_ { 
		samey = cxxomfort::traits::is_same<T1,T2>::value, 
		convertible_12 = std::tr1::is_convertible<T1,T2>::value ,
		convertible_21 = std::tr1::is_convertible<T2,T1>::value
	};
	static_assert (convertible_12 || convertible_21, "cxxomfort: common_type<...> instantiation can not resolve");
	typedef typename std::conditional< 
		convertible_12, 
		T2, 
		typename std::conditional<
			convertible_21, T1, common_type_cant_resolve
			>::type
		>::type convto_t;
	
	public:
	typedef typename std::decay<convto_t>::type type;
	
};

/* enumerating known common_type cases */

// we specialize promotion pairs 
// that define common_type<T1,T2> -> type = T2
#define DEF_COMFORT_COMMONTYPE_PROMOTE(T1,T2) \
template <> struct common_type<T1,T2> { typedef T2 type; }; \
template <> struct common_type<T2,T1> { typedef T2 type; }  \


    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,short);
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,int) ;
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,long) ;
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,long long) ;
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,unsigned short);
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,unsigned int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,unsigned long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(bool,unsigned long long);

    DEF_COMFORT_COMMONTYPE_PROMOTE(signed char,short);
    DEF_COMFORT_COMMONTYPE_PROMOTE(signed char,int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(signed char,long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned char,unsigned short);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned char,unsigned int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned char,unsigned long);

    DEF_COMFORT_COMMONTYPE_PROMOTE(short,int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(short,long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(short,long long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned short,unsigned int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned short,unsigned long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned short,unsigned long long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned short,int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned short,long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned short,long long);

    DEF_COMFORT_COMMONTYPE_PROMOTE(int,long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(int,long long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned int,unsigned long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned int,unsigned long long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned int,int);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned int,long);
    DEF_COMFORT_COMMONTYPE_PROMOTE(unsigned int,long long);

    DEF_COMFORT_COMMONTYPE_PROMOTE(float,double);
    DEF_COMFORT_COMMONTYPE_PROMOTE(float,long double);
    DEF_COMFORT_COMMONTYPE_PROMOTE(double,long double);

    DEF_COMFORT_COMMONTYPE_PROMOTE(int,float);

    //DEF_COMFORT_COMMONTYPE_PROMOTE(char const*, std::string);
    //DEF_COMFORT_COMMONTYPE_PROMOTE(wchar const*, std::wstring);

#undef DEF_COMFORT_COMMONTYPE_PROMOTE


} 

#endif // cxx version


// 
// cxxomfort fix for common_type (MSVC etc)
// 

namespace cxxomfort { namespace fix {
/*
 * Unlike other utilities, common_type has to be implemented in namespace std 
 * because clients have to be able to specialize. This means that in systems where it's broken (such as MSVC 2010) it's not possible to reuse or extend the name; the only ways around it are to either hack the header file, implementing a whole alternative <type_traits> header file via transparent headers, or provide an alternative name.
 * 
 * cxxomfort::fix::common_type is offered as an alternative name.
 * 
 */

#if (defined(CXXOMFORT_FIX_common_type))
	#if (CXXOMFORT_FIX_common_type > 0)

template <typename A1, typename A2, typename A3=A2, typename A4=A2, typename A5=A2>
struct common_type 
: public common_type < typename common_type<A1,A2>::type, A3, A4, A5, A5> 
{ };

template <typename A1, typename A2>
struct common_type <A1,A2 /*defaults*/ > {
	typedef typename std::decay<A1>::type D1;
	typedef typename std::decay<A2>::type D2;
	typedef CXXO_TYPEOF(true ? std::declval<D1>() : std::declval<D2>()) type;
};

	
	#else
	using ::std::common_type;
	#endif
#else
	using ::std::common_type;
#endif

} } // cxxomfort::fix


#endif
