#ifndef CXXOMFORT_IMPL_MINMAX_HPP
#define CXXOMFORT_IMPL_MINMAX_HPP
/**
 * @file 
 */

#define CXXOMFORT_IMPLEMENTS_n1990 0

#include <cxxomfort/config.hpp>

#if defined(DOXYGEN_DOC)
namespace std {

/**
 * @brief Return a <code>pair(min,max)</code> from two arguments and a comparator @a Less .
 * @cxx11b Find the minimum and maximum elements in a sequence.
 * @xrefitem std0algorithm "" "" Find the minimum and maximum elements in a sequence.
 * @return A @c pair (min, max)
 * @see @cppref{algorithm/minmax}
 * */
template <typename T, typename Comparator> std::pair<T const&, T const&> 
minmax (T const& a, T const& b, Comparator Less) {}

//! @overload minmax
template <typename T>
std::pair<T const&, T const&> minmax (T const& a, T const& b)  {}

/**
 * @brief Returns the pair (minimum,maximum) for the given sequence and comparator @p less .
 * @param less A comparator object with the same semantics as <code>std::less</code> .
 * @xrefitem std0algorithm "<algorithm> backports" "Search"
 * @xrefitem cxx11-backports "" "Algorithms" minmax_elements
 * @return A <code>std::pair</code> containing the minimum and maximum.
 */
template <typename FIterator, typename Comparator>
std::pair<FIterator,FIterator> 
minmax_element (FIterator ini, FIterator fin, Comparator less) {}

//! @overload minmax_element
template <typename FIterator>
std::pair<FIterator,FIterator> 
minmax_element (FIterator ini, FIterator fin) {}

} 

#elif (CXXOMFORT_CXX_STD >= 2011)
    #undef  CXXOMFORT_IMPLEMENTS_n1990
    #define CXXOMFORT_IMPLEMENTS_n1990 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC && CXXOMFORT_COMPILER_VERSION >=404 && CXXOMFORT_COMPILER_VERSION < 600 && CXXOMFORT_CXX_EMULATION==2011 )
    // GCC already has defined these in 4.4-4.6 c++0x mode
    #undef  CXXOMFORT_IMPLEMENTS_n1990
    #define CXXOMFORT_IMPLEMENTS_n1990 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600) )
    // MSVC 2010 onwards
    #undef  CXXOMFORT_IMPLEMENTS_n1990
    #define CXXOMFORT_IMPLEMENTS_n1990 CXXO_IMPLSTATUS_NATIVE()
#else // pre-c++11
    #undef  CXXOMFORT_IMPLEMENTS_n1990
    #define CXXOMFORT_IMPLEMENTS_n1990 CXXO_IMPLSTATUS_BACKPORT()

namespace cxxomfort { namespace cxxostd {

/**
 * @ingroup cxx11-backports
 * @{
 */

template <typename T, typename Comparator>
std::pair<T const&, T const&> minmax (T const& a, T const& b, Comparator Less) {
    return Less(b,a) ? 
    std::pair<T const&, T const&>(b, a) : std::pair<T const&, T const&>(a, b)
    ;
}

//! @overload minmax
template <typename T>
std::pair<T const&, T const&> minmax (T const& a, T const& b)  {
    return (b < a) ? 
    std::pair<T const&,T const&>(b, a) : std::pair<T const&,T const&>(a, b)
    ;
}

template <typename FIterator, typename Comparator>
std::pair<FIterator,FIterator> 
minmax_element (FIterator ini, FIterator fin, Comparator less) {
    std::pair<FIterator, FIterator> result(ini, fin);
 
    if (ini == fin) return result;
    if (++ini == fin) return result;

    if (less(*ini, *result.first)) {
        result.second = result.first;
        result.first = ini;
    } else {
        result.second = ini;
    }
    while (++ini != fin) {
        FIterator i = ini;
        if (++ini == fin) {
            if (less(*i, *result.first)) result.first = i;
            else if (!(less(*i, *result.second))) result.second = i;
            break;
        } else {
            if (less(*ini, *i)) {
                if (less(*ini, *result.first)) result.first = ini;
                if (!(less(*i, *result.second))) result.second = i;
            } else {
                if (less(*i, *result.first)) result.first = i;
                if (!(less(*ini, *result.second))) result.second = ini;
            }
        }
    }
    return result;

}

template <typename FIterator>
std::pair<FIterator,FIterator> 
minmax_element (FIterator ini, FIterator fin) {
    std::less <typename std::iterator_traits<FIterator>::value_type> less;
    return minmax_element(ini, fin, less);
}

/**
 * @}
 */

} } // cxxomfort


#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_std_cxx11_algorithms>0)
#else
    #if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC && CXXOMFORT_COMPILER_VERSION >=404 && CXXOMFORT_COMPILER_VERSION < 600 && CXXOMFORT_CXX_EMULATION==2011 )
    // GCC already has defined these in 4.4-4.6 c++0x mode
    #else
namespace std {
    using ::cxxomfort::cxxostd::minmax;
    using ::cxxomfort::cxxostd::minmax_element;
}
    #endif
#endif // no-using

#endif

#undef  CXXOMFORT_IMPLEMENTS_algorithm_minmax
#define CXXOMFORT_IMPLEMENTS_algorithm_minmax CXXOMFORT_IMPLEMENTS_n1990


#endif

