#ifndef CXXOMFORT_IMPL_ALGORITHM_14EQUAL_HPP
#define CXXOMFORT_IMPL_ALGORITHM_14EQUAL_HPP
/**
 * @file
 * @brief Implementation of C++14's overloads for @c equal(), @c mismatch() .
 * */
#include <cxxomfort/base.hpp>
#include <cxxomfort/algorithm.hpp>
#include <cxxomfort/utils/traits_basic.hpp>
#include "./11-conditional.hpp"

#undef  CXXOMFORT_IMPLEMENTS_n3671
//#define CXXOMFORT_IMPLEMENTS_n3671 CXXO_IMPLSTATUS_BACKPORT()

#if defined(__cpp_lib_robust_nonmodifying_seq_ops)
	#define CXXOMFORT_IMPLEMENTS_n3671 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD>=2014)
	#define CXXOMFORT_IMPLEMENTS_n3671 CXXO_IMPLSTATUS_NATIVE()
#else
	#define CXXOMFORT_IMPLEMENTS_n3671 CXXO_IMPLSTATUS_BACKPORT()
	#if (CXXOMFORT_CXX_STD >= 1997)
		#define __cpp_lib_robust_nonmodifying_seq_ops 200701
	#endif
	#include <cxxomfort/type_traits.hpp>

namespace cxxomfort { namespace cxxostd {

//
// explicit sized ranges for equal, from C++14
//
namespace detail_algorithm {
template <typename T1, typename T2> struct heterogenous_equal_to {
    bool operator() (T1 const& t1, T2 const& t2) const { return t1 == t2; }
};

template <typename RangeIt1, typename RangeIt2, typename Compare> inline 
bool equal_impl (RangeIt1 ini1, RangeIt1 fin1, RangeIt2 ini2, RangeIt2 fin2, Compare c, std::false_type /*iterator_tags are non-random-access*/) {
    for (; ini1 != fin1 && ini2 != fin2 ; ++ini1, ++ini2) {
        if (!c(*ini1, *ini2)) return false;
    }
    if ( (ini1 != fin1)|| (ini2 != fin2) ) return false;
    return true;
}

template <typename RangeIt1, typename RangeIt2, typename Compare> inline 
bool equal_impl (RangeIt1 ini1, RangeIt1 fin1, RangeIt2 ini2, RangeIt2 fin2, Compare c, std::true_type) {
    if ((fin1-ini1)!=(fin2-ini2)) return false; // ranges are of different size
    return std::equal( ini1, fin1, ini2, c);
}


} //detail_algorithm

/**
 * @brief Compare two sequences [ini1,fin1) and [ini2,fin2) for equality given comparator @p eq . Four-iterators overload.
 * @param[in] ini1,fin1 Delimiters for the first sequence argument to compare.
 * @param[in] ini2,fin2 Delimiters for the second sequence argument to compare.
 * @param eq An @e equality compare object.
 * @ingroup std0algorithm
 * @cxx14b Full-ranges implementation of @c std::equal .
 * @return @c true if both ranges have equal elements.
 */
template <typename RangeIt1, typename RangeIt2, typename Compare> inline 
bool 
equal (RangeIt1 ini1, RangeIt1 fin1, RangeIt2 ini2, RangeIt2 fin2, Compare eq) {
    typedef typename std::iterator_traits<RangeIt1>::iterator_category cat_1;
    typedef typename std::iterator_traits<RangeIt2>::iterator_category cat_2;
    typedef std::random_access_iterator_tag rtag;
    using traits::is_same;
    typedef typename std::conditional< 
        (is_same<cat_1,rtag>::value && is_same<cat_2,rtag>::value)
        , ::std::true_type, ::std::false_type
    >::type dispatcher_t;
    return detail_algorithm::equal_impl(ini1,fin1,ini2,fin2,eq,dispatcher_t());
}

//! @overload equal(ini1,fin1,ini2,fin2,eq)
template <typename RangeIt1, typename RangeIt2> inline 
bool 
equal (RangeIt1 ini1, RangeIt1 fin1, RangeIt2 ini2, RangeIt2 fin2) {
    typedef typename std::iterator_traits<RangeIt1>::iterator_category cat_1;
    typedef typename std::iterator_traits<RangeIt2>::iterator_category cat_2;
    typedef typename std::iterator_traits<RangeIt1>::value_type T1;
    typedef typename std::iterator_traits<RangeIt2>::value_type T2;
    typedef std::random_access_iterator_tag rtag;
    using traits::is_same;
    typedef typename std::conditional< 
        (is_same<cat_1,rtag>::value && is_same<cat_2,rtag>::value)
        , ::std::true_type, ::std::false_type
    >::type dispatcher_t;
    detail_algorithm::heterogenous_equal_to< T1, T2 > this_eq;
    return detail_algorithm::equal_impl(ini1, fin1, ini2, fin2, this_eq, dispatcher_t());
}

//
// explicit sized ranges for mismatch, from C++14
//

/**
 * @brief Finds the first difference (dif1,dif2) between two sequences [ini1,fin1) and [ini2,fin2), given comparator @p eq . Four-iterator overload.
 * @param p An @e equality compare object.
 * @ingroup std0algorithm
 * @cxx14b Full-ranges implementation of @c std::mismatch .
 */
template<class InputIt1, class InputIt2, class BinaryPredicate>
std::pair<InputIt1, InputIt2>
mismatch (InputIt1 ini1, InputIt1 fin1, InputIt2 ini2, InputIt2 fin2, BinaryPredicate p) {
    while (ini1 != fin1 && ini2 != fin2 && p(*ini1, *ini2)) {
        ++ini1, ++ini2;
    }
    return std::make_pair(ini1, ini2);
}

//! @overload mismatch
template<class InputIt1, class InputIt2>
std::pair<InputIt1, InputIt2>
mismatch (InputIt1 ini1, InputIt1 fin1, InputIt2 ini2, InputIt2 fin2) {
    typedef typename std::iterator_traits<InputIt1>::value_type T1;
    typedef typename std::iterator_traits<InputIt2>::value_type T2;
    detail_algorithm::heterogenous_equal_to< T1, T2 > this_eq;
    return mismatch(ini1, fin1, ini2, fin2, this_eq);
}

} } // cxxomfort

namespace std {
	//! Uses @c equal supplement from cxxomfort
	using cxxomfort::cxxostd::equal;
	//! Uses @c mismatch supplement from cxxomfort
	using cxxomfort::cxxostd::mismatch;
}

#endif
#undef  CXXOMFORT_IMPLEMENTS_equal_mismatch_full_ranges
#define CXXOMFORT_IMPLEMENTS_equal_mismatch_full_ranges CXXOMFORT_IMPLEMENTS_n3671

#endif
