#ifndef CXXOMFORT_IMPL_14_INTEGER_SEQUENCE_HPP
#define CXXOMFORT_IMPL_14_INTEGER_SEQUENCE_HPP
#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <utility>

/**
 * @file
 * @brief Backports for C++14's "integer_sequence" facilities.
 * @ingroup std0utility
 */

#ifdef DOXYGEN_DOC
namespace std {

/**
 * @brief C++14 integer_sequence utility
 * @xrefitem std0utility "" "" Processes parameter packs using integer values
 * @cxx14b Process parameter packs using integer values
 * 
 * Given <tt>integer_sequence<type, N></tt>, it generates the 
 * object with parameter pack <tt><type, 0, 1, ..., N-1></tt>.
 * 
 * For example, <tt>integer_sequence<int, 5></tt> "contains" the parameter pack <tt><0, 1, 2, 3, 4></tt>.
 * 
 * @see @cppref{utility/integer_sequence}
 * @see @ref __cpp_lib_integer_sequence is defined when this feature is added.
 * 
 * @note This is backported down to C++11 only.
 * 
 * */
template <typename T, T... I> 
struct integer_sequence {
    template <T N> struct append : integer_sequence<T, I..., N> {};
    static CXXO_CONSTEXPR std::size_t size () { return sizeof...(I); }
    static const size_t _size; //< The size (length) of the sequence.
    typedef append<_size> next; //< The next subyacent sequence.
    typedef T type;
}; 

//! Helper for @c integer_sequence
//! @cxx14b Helper to create @c integer_sequence .
template <typename T, size_t N> struct make_integer_sequence {};

/**
 * @brief Integer sequence of <tt>size_t</tt>.
 * @cxx14b Process parameter packs using integer values
 */
template <size_t N> struct index_sequence : integer_sequence<size_t, N> {};

//! Helper for @c index_sequence
//! @cxx14b Helper to create @c index_sequence .
template <typename T, size_t N> struct make_index_sequence {};

} 

	/**
	 * @ingroup std00featuremacros
	 * @cppfeature{@ref integer_sequence std::integer_sequence}
	 * *//
	#define __cpp_lib_integer_sequence YYYY00L

#elif (defined(__cpp_lib_integer_sequence))
    #define CXXOMFORT_IMPLEMENTS_n3658 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD>= 2014 || CXXO_COMPILER_SUPPORT_integer_sequence>0)
    #define CXXOMFORT_IMPLEMENTS_n3658 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD >= 2011 || \
	(CXXOMFORT_CXX_EMULATION >= 2011 && CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_template_aliases)\
	)
	//#pragma message ("integer_sequence c++11esque")
    #define CXXOMFORT_IMPLEMENTS_n3658 CXXO_IMPLSTATUS_BACKPORT()
    
    #undef  CXXOMFORT_IMPLEMENTS_integer_sequence
    #define CXXOMFORT_IMPLEMENTS_integer_sequence CXXOMFORT_IMPLEMENTS_n3658

namespace cxxomfort { namespace cxxostd {

/**
 * @ingroup cxx14-backports
 * @{
 */
template <typename T, T... Elems>
struct integer_sequence {
    typedef T type;
    static constexpr size_t size () { return sizeof...(Elems); }
}; 

template <std::size_t... N>
using index_sequence = integer_sequence<std::size_t, N...>;

template <typename I, size_t N, size_t... Next>
struct ISHelper : public ISHelper<I, N-1U, N-1U, Next...>
{ 
};

template <typename I, size_t... Next>
struct ISHelper<I, 0U, Next...> { 
	using type= integer_sequence<I, Next...>;
};


template <typename I, std::size_t N>
using make_integer_sequence = typename ISHelper<I, N>::type;

template <std::size_t N>
using make_index_sequence = typename ISHelper<std::size_t, N>::type;

template <typename... T>
using index_sequence_for = make_index_sequence<sizeof...(T)>;


/**
 * @}
 */

} }

#else
	// nothing to do
    #define CXXOMFORT_IMPLEMENTS_n3658 0

#endif



#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (CXXOMFORT_IMPLEMENTS_n3658 >= CXXO_IMPLSTATUS_NATIVE())
#elif (CXXOMFORT_IMPLEMENTS_n3658 >= CXXO_IMPLSTATUS_BACKPORT())
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("integer_sequence and allies")
	#endif
	#define __cpp_lib_integer_sequence (CXXOMFORT_CXX_STD*100)
	
namespace std {
    //! Brings @c integer_sequence to std.
    using cxxomfort::cxxostd::integer_sequence;
    using cxxomfort::cxxostd::index_sequence;
    using cxxomfort::cxxostd::make_integer_sequence;
    using cxxomfort::cxxostd::make_index_sequence;
    using cxxomfort::cxxostd::index_sequence_for;
} // namespace std
#endif // std using

#undef  CXXOMFORT_IMPLEMENTS_integer_sequence
#define CXXOMFORT_IMPLEMENTS_integer_sequence CXXOMFORT_IMPLEMENTS_n3658 

#endif
