#ifndef CXXOMFORT_IMPL_14_ITERATOR_ACCESSORS_HPP
#define CXXOMFORT_IMPL_14_ITERATOR_ACCESSORS_HPP
/**
 * @file
 * @brief Implementation of C++14's iterator accesors std::cbegin , std::cend , etc.
 */
#include <cxxomfort/config.hpp>
#include <cxxomfort/base/iterator.hpp>
#include <cxxomfort/utils/traits_basic.hpp> // as_const
#include <cstddef>

namespace cxxomfort { namespace cxxostd { 
namespace iterator {

// as_const, from utility

namespace {
	using ::cxxomfort::traits::add_const;
	template <typename T>
	typename add_const<T>::type& as_const (T& r) CXXO_NOEXCEPTNOTHROW { return r; }
}

// 
// cbegin, cend 
//

#if defined(DOXYGEN_DOC)
/**
 * @brief Returns an iterator to the beginning of a given container or sequence.
 * @ingroup std0iterator
 * @cxx14b Iterator accessor
 * @xrefitem std0iterator "" "" Iterator accessor
 * 
 * @note Starting with year 2022 releases, the behaviour of @c cbegin() , @c cend() is slightly modified to match the original proposal's intended behaviour. In particular, in pre-C++14 mode, invoking @c cbegin(c) for a given container now more closely matches <tt>as_const(c).begin()</tt>.
 * 
 * 
 * */
template <typename Seq>
typename Seq::const_iterator cbegin (Seq&&);

/**
 * @ingroup std0iterator
 * @brief Returns an iterator to the end of a given container or sequence.
 * @cxx14b Iterator accessor
 * @xrefitem std0iterator "" "" Iterator accessor
 * */
template <typename Seq>
typename Seq::const_iterator cend (Seq&&);

#elif (CXXO_COMPILER_SUPPORT_iterator_accessors_14>0)
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_14 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILER_SUPPORT_constexpr && (CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof ))
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_14 CXXO_IMPLSTATUS_BACKPORT()

template <typename C>
constexpr auto inline
cbegin (C const& c)  -> decltype(std::begin(c)) {
    return std::begin(c);
} ;

template <typename C>
CXXO_CONSTEXPR  auto inline
cend (C const& c) -> decltype(std::end(c)) {
    return std::end(c);
} ;

#else

template <typename C>
inline typename C::const_iterator
cbegin (C const& c) {
    return std::begin(c);
} 

template <typename C>
inline typename C::const_iterator
cend (C const& c) {
    return std::end(c);
} 

#endif

// 
// reverse accessors 
// 

#if (CXXOMFORT_CXX_STD >= 2011)
template <typename C>
constexpr auto inline
rbegin (C& c)  -> decltype(c.rbegin()) {
    return c.rbegin();
} ;

template <typename C>
constexpr auto inline
rend (C& c)  -> decltype(c.rend()) {
    return c.rend();
} ;


template <typename C>
constexpr auto inline
crbegin (C const& c)  -> decltype(c.crbegin()) {
    return c.crbegin();
} ;

template <typename C>
constexpr auto inline
crend (C const& c)  -> decltype(c.crend()) {
    return c.crend();
} ;

#elif (CXXOMFORT_CXX_STD>=1997)

template <typename C>
inline typename C::reverse_iterator 
rbegin (C& c) {
	return c.rbegin();
}

template <typename C>
inline typename C::reverse_iterator 
rend (C& c) {
	return c.rend();
}

template <typename C>
inline typename C::const_reverse_iterator 
crbegin (C const& c) {
	return c.crbegin();
}

template <typename C>
inline typename C::const_reverse_iterator 
crend (C const& c) {
	return c.crend();
}

#else
	// very old containers likely don't have cr accessors

template <typename C>
inline typename C::reverse_iterator 
rbegin (C& c) {
	return c.rbegin();
}

template <typename C>
inline typename C::reverse_iterator 
rend (C& c) {
	return c.rend();
}

template <typename C>
inline typename C::const_reverse_iterator 
crbegin (C const& c) {
	return c.rbegin();
}

template <typename C>
inline typename C::const_reverse_iterator 
crend (C const& c) {
	return c.rend();
}

#endif

//! @overload cxxomfort::rbegin 
template <typename T, ::std::size_t N> inline
::std::reverse_iterator<T*> rbegin (T(&arr)[N]) { return ::std::reverse_iterator<T*>(arr+N); }

//! @overload cxxomfort::rend 
template <typename T, ::std::size_t N> inline
::std::reverse_iterator<T*> rend (T(&arr)[N]) { return ::std::reverse_iterator<T*>(arr); }


}
}} // cxxomfort::cxxostd

#undef CXXOMFORT_IMPLEMENTS_iterator_accessors_14
#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_14 0
#elif (CXXO_COMPILER_SUPPORT_iterator_accessors_14>0)
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_14 CXXO_IMPLSTATUS_NATIVE()
#else
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++14 iterator accessors")
	#endif

    namespace std {
        using ::cxxomfort::cxxostd::iterator::cbegin;
        using ::cxxomfort::cxxostd::iterator::cend;
        using ::cxxomfort::cxxostd::iterator::rbegin;
        using ::cxxomfort::cxxostd::iterator::rend;
        using ::cxxomfort::cxxostd::iterator::crbegin;
        using ::cxxomfort::cxxostd::iterator::crend;
    }
	#define CXXOMFORT_IMPLEMENTS_iterator_accessors_14 CXXO_IMPLSTATUS_BACKPORT()

#endif // no std using

#endif
