#ifndef CXXOMFORT_IMPL_TUPLEGETTYPE_HPP
#define CXXOMFORT_IMPL_TUPLEGETTYPE_HPP
/**
 * @file
 */

#include <cxxomfort/base.hpp>

#if (defined(CXXOMFORT_NOTICES))
    #if (CXXOMFORT_NOTICES > 1)
		#pragma message CXXO_NOTICE("enabled tuple get<type> implementation")
    #endif
#endif

#if (CXXOMFORT_CXX_STD >= 2011)
	#include <tuple>
#endif


/*
 * Support for getting elements of a tuple by type
 * (C++14 proposal n3584 "Wording for Addressing Tuples by Type")
 * based off of n3404 and n3584
 * http://www.open-std.org/JTC1/SC22/WG21/docs/papers/2013/n3584.html
 * 
 * Interfaces defined in this header:
 * 
 * * std::get <T> (std::tuple<...>)
 * 
 */
#if defined(DOXYGEN_DOC)
namespace std {

/**
 * @brief Return a member of a tuple given its type as a template argument.
 * @ingroup std0tuple
 * @cxx14b Get a tuple element by type, as per <tt>get< type >(std::tuple&& sometuple)</tt>
 * @tparam T a type, such that it exists in the type pack @c TupleArgs... .
 * @anchor std_get_type_tuple
 * 
 * @pre The type @a T exists in @a TupleArgs
 * 
 * Given a <tt>std::tuple</tt> type with component types @a TupleArgs... , the new C++14 overload of <tt>std::get()</tt> is provided that gives access to the element of type @a T in the tuple, assuming there is one element in @a TupleArgs with type @a T .
 * 
 * @code
tuple<int, string, float> tu (...);
get<string>(tu); // gets get<1>(tu), which is the std::string member
get<bool>(tu); // does not compile, no bool type in tu
 * @endcode
 * 
 * @note This functionality is backported mostly to C++03 compatible compilers with high template support. The macro @c @ref CXXOMFORT_IMPLEMENTS_tuple_get_type notifies if this function can be enabled.
 * 
 */
template <typename T, typename... TupleArgs>
T& get (std::tuple<TupleArgs...>&) {}

//! @overload get
template <typename T, typename... TupleArgs>
T const& get (std::tuple<TupleArgs...> const&) {}

} // std

	/**
	 * @ingroup std00featuremacros
	 * @cppfeature{type <tt>get<T></tt> for tuples}
	 * */
	#define __cpp_lib_tuples_by_type YYYY00L

#elif (defined(__cpp_lib_tuples_by_type))
    // already available, do nothing
	#define CXXOMFORT_IMPLEMENTS_n3584 CXXO_IMPLSTATUS_NATIVE()
	#define CXXOMFORT_IMPLEMENTS_n3670 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD >= 2011 || (CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_rvref) )

	#define CXXOMFORT_IMPLEMENTS_n3584 CXXO_IMPLSTATUS_BACKPORT()
	#define CXXOMFORT_IMPLEMENTS_n3670 CXXO_IMPLSTATUS_BACKPORT()
	#define __cpp_lib_tuples_by_type (CXXOMFORT_CXX_STD*100)

	#include <utility> // integer_sequence
	#include <cxxomfort/impl/tuplemeta.hpp> // is_tuple, tuple_index

namespace {
	using cxxomfort::impl::is_tuple;
	using cxxomfort::impl::tuple_index;
}

namespace std {

// three overloads as per cppreference

/*
 * @brief Return a member of a tuple given its type as a template argument.
 * @ingroup cxx14-backports
 */
template <typename C, typename... Targs
, typename = std::enable_if< cxxomfort::impl::tuple_index<C,std::tuple<Targs...> >::value != -1 
> 
>
C&& get (std::tuple<Targs...>&& t) CXXO_NOEXCEPT {
    enum { N = tuple_index<C, std::tuple<Targs...> >::value };
    return std::forward<C>(get<N>(t));
}

template <typename C, typename... Targs
, typename = std::enable_if< cxxomfort::impl::tuple_index<C,std::tuple<Targs...> >::value != -1 
>
>
C const& get (std::tuple<Targs...> const& t) CXXO_NOEXCEPT {
    enum { N = cxxomfort::impl::tuple_index<C, std::tuple<Targs...> >::value };
    return (get<N>(t));
}

template <typename C, typename... Targs
, typename = std::enable_if< cxxomfort::impl::tuple_index<C,std::tuple<Targs...> >::value != -1 
>
>
C& get (std::tuple<Targs...> & t) CXXO_NOEXCEPT {
    enum { N = cxxomfort::impl::tuple_index<C, std::tuple<Targs...> >::value };
    return (get<N>(t));
}


} // std

#elif (CXXOMFORT_CXX_STD>=1997)

	#define CXXOMFORT_IMPLEMENTS_n3584 CXXO_IMPLSTATUS_EMULATION()
	#define CXXOMFORT_IMPLEMENTS_n3670 CXXO_IMPLSTATUS_EMULATION()
	#define __cpp_lib_tuples_by_type (CXXOMFORT_CXX_STD*100)

	#include <cxxomfort/impl/tuplemeta.hpp> // is_tuple, tuple_index
	#include <cxxomfort/impl/11-conditional.hpp> // enable_if

namespace std {

namespace {
	using cxxomfort::impl::is_tuple;
	using cxxomfort::impl::tuple_index;
}

// at this point, c++03
template <typename C, typename Tuple>
C const&
get (Tuple const& t
, typename std::enable_if <(is_tuple<Tuple>::value==true), bool* >::type = 0
) {
    enum { N = tuple_index<C,Tuple>::value }; 
    static_assert( N < tuple_size<Tuple>::value , "get<Type> to type not in Tuple");
    return get<N>(t);
}

template <typename C, typename Tuple>
C &
get (Tuple & t
, typename std::enable_if <(is_tuple<Tuple>::value==true), bool* >::type = 0
) {
    enum { N = tuple_index<C,Tuple>::value }; 
    static_assert( N < tuple_size<Tuple>::value , "get<Type> to type not in Tuple");
    return get<N>(t);
}

/*
template <typename C, typename T0>
C& get (std::tuple<T0>& t
, typename std::enable_if < std::is_same<C,T0>::type , bool*>::value 
) {
	using namespace std;
	return get<0>(t);
}

template <typename C, typename T0, typename T1>
C& get (std::tuple<T0,T1>& t
, typename std::enable_if < 
	(cxxomfort::impl::tuple_index<C,std::tuple<T0,T1> >::value != -1)
	, bool*>::value 
) {
	typedef std::tuple<T0,T1> Tuple;
    enum { N = cxxomfort::impl::tuple_index<C,Tuple>::value }; 
	return get<N>(t);
}

template <typename C, typename T0, typename T1, typename T2>
C& get (std::tuple<T0,T1,T2>& t
, typename std::enable_if < 
	(cxxomfort::impl::tuple_index<C,std::tuple<T0,T1,T2> >::value != -1)
	, bool*>::value 
) {
	typedef std::tuple<T0,T1,T2> Tuple;
    enum { N = cxxomfort::impl::tuple_index<C,Tuple>::value }; 
	return get<N>(t);
}
*/

} // std


#elif (CXXOMFORT_IMPLEMENTS_tuple == CXXO_IMPLSTATUS_EMULATION())

	#include <cxxomfort/impl/tuplemeta.hpp> // is_tuple, tuple_index
	#include <cxxomfort/impl/11-conditional.hpp> // enable_if

namespace std {

template <typename C, typename Tuple>
typename C const& get (Tuple const& t) {
    enum { S = std::tuple_size<Tuple>::value , N = cxxomfort::impl::tuple_index<C,Tuple>::value }; 
    static_assert ( N >= 0 && N < S , "Type not present or unique in tuple" );
    return (C const&)(get<N>(t));
}

}

	#define CXXOMFORT_IMPLEMENTS_n3584 CXXO_IMPLSTATUS_EMULATION()
	#define CXXOMFORT_IMPLEMENTS_n3670 CXXO_IMPLSTATUS_EMULATION()
	#define __cpp_lib_tuples_by_type 1

#endif

#undef  CXXOMFORT_IMPLEMENTS_tuple_get_type
#define CXXOMFORT_IMPLEMENTS_tuple_get_type CXXOMFORT_IMPLEMENTS_n3584


// Implement interface for std::pair

/*
namespace std {

template <typename C, typename T2>
C const& get(std::pair<C,T2> const& p) { 
    enum { value= detail::tuple_element2index_helper_pair<C,T2>::value };
    return p.first; 
}
template <typename C, typename T1>
C const& get(std::pair<T1,C> const& p) { 
    enum { value= detail::tuple_element2index_helper_pair<T1,C>::value };
    return p.second;
}

template <typename C, typename T2>
C & get(std::pair<C,T2> & p) { 
    enum { value= detail::tuple_element2index_helper_pair<C,T2>::value };
    return p.first; 
}
template <typename C, typename T1>
C & get(std::pair<T1,C> & p) { 
    enum { value= detail::tuple_element2index_helper_pair<T1,C>::value };
    return p.second;
}

}
*/

#endif
