#ifndef CXXOMFORT_IMPL_CSTDDEF_BYTE_HPP
#define CXXOMFORT_IMPL_CSTDDEF_BYTE_HPP

#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include "../impl/07-type_traits.hpp"
#include <cstddef>
#include <cstring>

#if (defined(__cpp_lib_byte) && __cpp_lib_byte>=201603)
	#undef  CXXO_COMPILER_SUPPORT_std_byte
	#define CXXO_COMPILER_SUPPORT_std_byte 1
#endif

#if (defined(DOXYGEN_DOC))
namespace std {

/** 
 * @enum byte
 * @brief Implementation of a raw byte data type.
 * @ingroup std0cstddef
 * @cxx17b Standard declaration of a raw data, non-integral @e byte type.
 * @see  @wg21{p0298r3,byte type}
 * @see  @cppref{types/byte}
 * @see  @ref __cpp_lib_byte is defined when this feature is added.
 * 
 * This implementation-defined type implements the functionality 
 * for <tt>std::byte</tt> where required (eg.: in C++03 mode).
 * 
 * In general, a <tt>byte</tt> is implemented as follows:
 * 
 * @code
 * enum class byte : unsigned char {};
 * unsigned to_integral (byte); // implementation_defined
 * 
 * @endcode
 * 
 * Only comparison operator, bit operations (via bitmask operators) and explicit 
 * conversion to <tt>unsigned char</tt> are defined for it. Just 
 * like <tt>std::byte</tt>, it has no value-taking constructors.
 * 
 * The implementation changes according to the language features 
 * that cxxomfort can find. In C++11 onwards mode, an actual 
 * <tt>enum class</tt> is used.
 * 
 * In general, the following operations are allowed on a @c byte :
 * 
 * @code
 * byte b = to_byte(0xnn);
 * b | c;
 * b & c;
 * b ^ c;
 * b >>= 2;
 * c <<= 3;
 * c= ~b;
 * int cint = to_integer<int>(c);
 * byte d { 0x23 }; // c++17 only
 * @endcode
 * 
 * Except where explicitly indicated, all operations on <tt>byte</tt> are nothrow and (C++14) constexpr, including conversions to integrals.
 * 
 * The helper function @c to_byte() is used to construct a @c byte pre-C++17.
 * 
 * The helper functions @c to_integer() and @c to_value() can recover the value stored by a @c byte into an integral or <tt>unsigned char</tt> variable respectively.
 * 
 * See @ref @c std::byte for more information.
 * */
enum class byte : unsigned char {
};

//! Compares the numerical expression of two bytes
inline bool operator== (byte, byte);

//! Compares the numerical expression of two bytes
inline bool operator!= (byte, byte);

/** 
 * @brief Converts the byte value into an @a Integral .
 * @ingroup std0cstddef
 * @cxx17b Converts a @c byte to an integral value.
 * 
 * @xrefitem byte_conversion "" "" Converts a @c byte value to an Integral type.
 * 
 * */
template <typename Integral>
Integral to_integer (byte b);

//! Converts the byte value into an <tt>unsigned char</tt>.
inline unsigned char to_value (byte b);


} // std

namespace cxxomfort { namespace fix {

/**
 * @brief Constructs a @c std::byte using an integral value.
 * @ingroup cxxomfort-fixes
 * 
 * @pre @e Integral is a C++ @e integral type.
 * 
 * @xrefitem byte_conversion "" "" Converts an integral value to @c byte .
 * 
 * @xrefitem cxxomfort-fixes "" "" Create  @c byte values.
 * 
 * */
template <typename Integral> constexpr 
std::byte 	to_byte (Integral i) noexcept {
	return static_cast<std::byte>(i);
}

} }


#elif (CXXO_COMPILER_SUPPORT_std_byte>0)
	#define CXXOMFORT_IMPLEMENTS_p0298 CXXO_IMPLSTATUS_NATIVE()

namespace cxxomfort { namespace fix {

using std::byte;

template <typename I, typename std::enable_if<std::is_integral<I>::value, bool>::type = true >
constexpr std::byte 
to_byte (I i) noexcept {
	return static_cast<std::byte>(i);
}

} }  // cxxomfort::fix

#elif (CXXO_COMPILER_SUPPORT_enum_class && CXXO_COMPILER_SUPPORT_template_aliases)
	#define CXXOMFORT_IMPLEMENTS_p0298 CXXO_IMPLSTATUS_BACKPORT()
	#define __cpp_lib_byte __cplusplus

// check if may_alias is available
	#if ( CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG,>=300) || CXXO_COMPILERREQ(CXXO_COMPILERID_GCC, >= 500) )
		#define BYTE_ATTR __attribute__((__may_alias__))
	#else
		#define BYTE_ATTR 
	#endif

namespace cxxomfort { namespace cxxostd {

enum class BYTE_ATTR byte : unsigned char { };

template <typename I, typename = typename std::enable_if<std::is_integral<I>::value>::type >
CXXO_CONSTEXPR  inline 
I to_integer (byte b) noexcept {
	return I(b);
}

} }

namespace cxxomfort { namespace fix {

using ::cxxomfort::cxxostd::byte;

template <typename I, typename std::enable_if<std::is_integral<I>::value, bool>::type = true >
CXXO_CONSTEXPR  inline 
::cxxomfort::cxxostd::byte to_byte (I i) noexcept {
	return static_cast<::cxxomfort::cxxostd::byte>(i);
}

} } // cxxomfort::fix
	//#pragma message ("backport byte as enum class + template aliases")

#elif ( \
	(CXXO_COMPILER_SUPPORT_enum_class))
	// pre-C++11 compilers that at least pick up constexpr and enum class
	#define CXXOMFORT_IMPLEMENTS_p0298 CXXO_IMPLSTATUS_BACKPORT()
	#define __cpp_lib_byte __cplusplus

namespace cxxomfort { namespace cxxostd {

// check if may_alias is available
#if ( CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG,>=300) || CXXO_COMPILERREQ(CXXO_COMPILERID_GCC, >= 500) )
    #define BYTE_ATTR __attribute__((__may_alias__))
#else
    #define BYTE_ATTR 
#endif

enum class BYTE_ATTR byte 
: unsigned char {
};

template <typename I>
CXXO_CONSTEXPR  inline 
I to_integer (byte b) CXXO_NOEXCEPT {
	static_assert0(std::is_integral<I>::value);
	return I(b);
}

} } // cxxomfort::cxxostd

namespace cxxomfort { namespace fix {

using ::cxxomfort::cxxostd::byte;

template <typename I >
CXXO_CONSTEXPR  inline 
::cxxomfort::cxxostd::byte to_byte (I i) CXXO_NOEXCEPT {
	static_assert0(std::is_integral<I>::value);
	return static_cast<::cxxomfort::cxxostd::byte>(i);
}

} } // cxxomfort::fix

	#pragma message ("backport byte as enum class + template aliases")

#elif (CXXOMFORT_CXX_STD>=1997) 

	#define CXXOMFORT_IMPLEMENTS_p0298 (CXXO_IMPLSTATUS_EMULATION()+1)
	#define __cpp_lib_byte __cplusplus

	#include <cxxomfort/utils/typesafe_enum.hpp>

namespace cxxomfort { namespace cxxostd {

namespace detail {
	struct mybyte_def { enum type {};  };
}

struct byte : public ::cxxomfort::typesafe_enum<detail::mybyte_def,unsigned char> {
	typedef ::cxxomfort::typesafe_enum<detail::mybyte_def,unsigned char> _Base;
	byte () {}
	template <typename A>
	explicit byte (A a) : _Base(a) {}
	//using _Base::value;
	template <typename Int> CXXO_EXPLICIT_OPERATOR(Int) () const { return value(); }
	CXXO_EXPLICIT_OPERATOR(unsigned char) () const { return value(); }
};

template <typename Integral>
CXXO_CONSTEXPR 
Integral to_integer (byte b) CXXO_NOEXCEPTNOTHROW {
	using namespace std;
	static_assert (::cxxomfort::traits::is_integral<Integral>::value, "to_integer");
	return explicit_cast<unsigned char>(b);
}

} }  // cxxomfort::cxxostd

namespace cxxomfort { namespace fix {

using ::cxxomfort::cxxostd::byte;

template <typename Integral>
::cxxomfort::cxxostd::byte to_byte (Integral i) CXXO_NOEXCEPTNOTHROW {
	using namespace std;
	::cxxomfort::cxxostd::byte b;
	(unsigned char&)(b) = static_cast<unsigned char>(i);
	return b;
}

} } // cxxomfort::fix
	#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES>2))
		#pragma message ("emulate byte as typesafe_enum")
	#endif
#else 

#endif
#undef  CXXOMFORT_IMPLEMENTS_byte 
#define CXXOMFORT_IMPLEMENTS_byte CXXOMFORT_IMPLEMENTS_p0298

// 
// end of byte impls
// 


#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (defined(CXXOFLAG_NO_STD_USING_STD_BYTE))
#elif (CXXOMFORT_IMPLEMENTS_p0298>=CXXO_IMPLSTATUS_EMULATION() && CXXOMFORT_IMPLEMENTS_p0298<CXXO_IMPLSTATUS_NATIVE())
namespace std {
	using ::cxxomfort::cxxostd::byte;
}
#endif

#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (CXXOMFORT_IMPLEMENTS_p0298>=CXXO_IMPLSTATUS_EMULATION() && CXXOMFORT_IMPLEMENTS_p0298<CXXO_IMPLSTATUS_NATIVE())
namespace std {
	using ::cxxomfort::cxxostd::to_integer;
}
#endif

// 
// internal checks 
// 

static_assert(sizeof(cxxomfort::fix::byte) == sizeof(char), "byte: sizeof check");
static_assert (::cxxomfort::traits::is_integral<cxxomfort::fix::byte>::value==false, "byte: separate-type check");

// 
// operators
// 

namespace cxxomfort { namespace cxxostd {

/*
 * MSVC has a compiler bug for enum and enum classes for the comparison operators 
 * See https://stackoverflow.com/questions/44515148/why-is-operator-overload-of-enum-ambiguous-in-msvc
 * */
#if (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC, >= 1600) && CXXOMFORT_IMPLEMENTS_p0298 == CXXO_IMPLSTATUS_BACKPORT() )

#else

CXXO_CONSTEXPR static inline 
bool operator!= (byte b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	return !(b1==b2);
}

#endif

#if (CXXOMFORT_IMPLEMENTS_byte >= CXXO_IMPLSTATUS_NATIVE())
#elif (CXXOMFORT_IMPLEMENTS_byte > 0)
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>1)
		#pragma message ("binary operators for byte")
	#endif

CXXO_CONSTEXPR static inline 
byte operator& (byte b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return to_byte( to_integer<unsigned>(b1) & to_integer<unsigned>(b2) );
}

CXXO_CONSTEXPR static inline 
byte operator| (byte b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return to_byte( to_integer<unsigned>(b1) | to_integer<unsigned>(b2) );
}

CXXO_CONSTEXPR static inline 
byte operator^ (byte b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return to_byte( to_integer<unsigned>(b1) ^ to_integer<unsigned>(b2) );
}

CXXO_CONSTEXPR static inline 
byte operator~ (byte b1) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return to_byte( ~to_integer<unsigned>(b1) );
}

static inline 
byte& operator&= (byte& b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return b1= b1 & b2;
}

static inline 
byte& operator|= (byte& b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	b1= b1|b2;
	return b1;
}

static inline 
byte& operator^= (byte& b1, byte b2) CXXO_NOEXCEPTNOTHROW {
	return b1= b1 ^ b2;
}

CXXO_CONSTEXPR static inline 
byte operator>> (byte b1, unsigned s) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return to_byte(to_integer<unsigned>(b1) >> s);
}

CXXO_CONSTEXPR static inline 
byte operator<< (byte b1, unsigned s) CXXO_NOEXCEPTNOTHROW {
	using ::cxxomfort::fix::to_byte;
	return to_byte(to_integer<unsigned>(b1) << s);
}

static inline 
byte& operator>>= (byte& b1, unsigned s) CXXO_NOEXCEPTNOTHROW {
	return b1= b1 >> s;
}

static inline 
byte& operator<<= (byte& b1, unsigned s) CXXO_NOEXCEPTNOTHROW {
	return b1= b1 << s;
}

#endif // native

} } 


namespace cxxomfort { namespace cxxostd {

/**
 * @brief Construct a 'byte' value from an integer.
 * @ingroup cxxomfort-fix
 * 
 */
	//using cxxomfort::fix::to_byte;

} } // cxxomfort::fix


#endif
