#ifndef CXXOMFORT_IMPL_CLAMP_HPP
#define CXXOMFORT_IMPL_CLAMP_HPP
/**
 * @file
 * @brief Implementation of std::clamp 
 * @sa http://en.cppreference.com/w/cpp/algorithm/clamp
 * 
 */

#include <cxxomfort/config.hpp>
#include <cxxomfort/cxxo-utils.hpp>
#include <cassert>
#include <algorithm>

#if (defined(__cpp_lib_clamp))
	#undef  CXXO_COMPILER_SUPPORT_std_clamp
	#define CXXO_COMPILER_SUPPORT_std_clamp 1
#endif

#define CXXOMFORT_IMPLEMENTS_n4536 0

namespace cxxomfort { namespace cxxostd {

#if defined(DOXYGEN_DOC)
/**
 * @brief Clamps a value.
 * @return An altered value from @a val to fit into the open interval (@a lo , @a hi ).
 * @ingroup std0algorithm
 * @xrefitem std0algorithm "" "" Clamps a value @a a within a range
 * @cxx17b Clamps a value @a val within a range
 * @sa http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2015/n4536.html
 * 
 * @note Behaviour is undefined if <tt>hi < lo</tt>.
 * 
 * @note Pre-C++14, @c constexpr functions can not perform assert or throw checks in their 
 * code. As such in C++11, while @c clamp() is @c constexpr it has no precondition checks.
 * 
 */
template<class T, class Compare>
constexpr 
T const& clamp ( T const& val, T const& lo, T const& hi, Compare comp) {
    return val;
}

//! @overload clamp 
template<class T>
CXXO_CONSTEXPR
T const& clamp ( T const& val, T const& lo, T const& hi) {
    return clamp(val,lo,hi, std::less<T>() );
}

// implementations

#elif (CXXO_COMPILER_SUPPORT_std_clamp>0)
    // nothing to do
    #undef  CXXOMFORT_IMPLEMENTS_n4536
    #define CXXOMFORT_IMPLEMENTS_n4536 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD>=2017 && (CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG,>=600)) )
    #undef  CXXOMFORT_IMPLEMENTS_n4536
    #define CXXOMFORT_IMPLEMENTS_n4536 CXXO_IMPLSTATUS_BACKPORT()
#elif (CXXOMFORT_CXX_STD>=2017)
    // nothing to do
    #undef  CXXOMFORT_IMPLEMENTS_n4536
    #define CXXOMFORT_IMPLEMENTS_n4536 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD>=2014)
    #undef  CXXOMFORT_IMPLEMENTS_n4536
    #define CXXOMFORT_IMPLEMENTS_n4536 CXXO_IMPLSTATUS_BACKPORT()

#elif (CXXOMFORT_CXX_STD>=2011)
    #undef  CXXOMFORT_IMPLEMENTS_n4536
    #define CXXOMFORT_IMPLEMENTS_n4536 CXXO_IMPLSTATUS_EMULATION()

/*
 * Pre-C++14 we can not make clamp constexpr without removing the assert check?
 * 
 * */

namespace {
    static inline void _clamp_exit () {
        assert(!"clamp() assert condition fail");
    }
}

template <typename T, typename Compare>
CXXO_CONSTEXPR
T const& clamp ( T const& val, T const& lo, T const& hi, Compare comp) {
    return ( 
	!comp(hi, lo) ?
	(comp(val, lo) ? lo : comp(hi, val) ? hi : val) :
	throw cxxomfort::assert_failure(nullptr, _clamp_exit)
    );
}

//! @overload clamp
//! @xrefitem std0algorithm "<algorithm> backports" ""
template <typename T>
CXXO_CONSTEXPR14
T const& clamp ( T const& val, T const& lo, T const& hi) {
    return (
	!(hi<lo) ?
	((val < lo) ? lo : hi < val ? hi : val) : lo
    );
}

#else // pre-c++11
    #undef  CXXOMFORT_IMPLEMENTS_n4536
    #define CXXOMFORT_IMPLEMENTS_n4536 CXXO_IMPLSTATUS_EMULATION()
template <typename T, typename Compare>
T const& clamp ( T const& val, T const& lo, T const& hi, Compare comp) {
    assert( !comp(hi,lo) );
    return ( 
        (comp(val, lo) ? lo : comp(hi, val) ? hi : val) 
    );
}

template <typename T>
T const& clamp ( T const& val, T const& lo, T const& hi) {
    assert( !(hi<lo) );
    return ( 
        ((val< lo) ? lo : (hi< val) ? hi : val) 
    );
}

#endif

} }

#if (CXXOMFORT_IMPLEMENTS_n4536 == CXXO_IMPLSTATUS_BACKPORT() )
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("clamp() backport")
	#endif

namespace cxxomfort { namespace cxxostd {

template <typename T, typename Compare>
constexpr 
T const& clamp ( T const& val, T const& lo, T const& hi, Compare comp) {
    assert( !comp(hi,lo) );
    return ( 
        (comp(val, lo) ? lo : comp(hi, val) ? hi : val) 
    );
}

template <typename T>
constexpr 
T const& clamp ( T const& val, T const& lo, T const& hi) {
    assert( !(hi<lo) );
    return ( 
        (val<lo ? lo : hi<val ? hi : val) 
    );
}

} }

#endif

#define CXXOMFORT_IMPLEMENTS_clamp CXXOMFORT_IMPLEMENTS_n4536

#if (defined(CXXOFLAG_NO_STD_USING))
#elif (defined(__cpp_lib_clamp))
#else
	#define __cpp_lib_clamp (CXXOMFORT_CXX_STD*100)
namespace std {
    using cxxomfort::cxxostd::clamp;
}
#endif


#endif
