#ifndef CXXOMFORT_IMPL_17_INVOKE_HPP
#define CXXOMFORT_IMPL_17_INVOKE_HPP
/**
 * @file
 * 
 * The implementation for C++11-esque invoke from 
 * https://en.cppreference.com/mwiki/index.php?title=cpp/utility/functional/invoke&oldid=92023
 * 
 * The implementation for C++11-esque invoke_result from 
 * https://en.cppreference.com/mwiki/index.php?title=cpp/types/result_of&oldid=102232
 */


#include "../config.hpp"
#include <cxxomfort/base.hpp>
#include <functional>

// 
// 
// 

#if (defined(__cpp_lib_invoke) && CXXO_COMPILER_SUPPORT_invoke==0)
	#undef  CXXO_COMPILER_SUPPORT_invoke
	#define CXXO_COMPILER_SUPPORT_invoke 1
#endif

#if defined(DOXYGEN_DOC)

namespace std {

//! @private 
struct dependent_type_t {};

/**
 * @brief Invokes a callable. See also @c invoke_r() .
 * @ingroup std0functional
 * @cxx17b Invokes callables via <tt>callable( args... )</tt>, accounting for eg.: member-pointer callables.
 * @xrefitem std0functional "" "" Invokes callables via <tt>callable( args... )</tt>, accounting for eg.: member-pointer callables.
 * @see @cppref{utility/functional/invoke} 
 * @see @c invoke_r()
 * @see @c is_invocable 
 *
 * @c invoke() calls a functionlike or callable @a f with the given arguments @a args... , taking care of forwarding, member function pointers, etc.
 * 
 * <tt>invoke(f, 1, 2, 3, ...)</tt> is in general equivalent to <tt>f(1, 2, 3, ...)</tt>. 
 * 
 * With member functions, <tt>invoke(C::*f, a, b, c, ...)</tt> is in general equivalent to <tt>a.*f(b, c, ...)</tt>.
 * 
 * The @c invoke() feature is backported down to C++11 with full support, and down to C++03 with @e partial support. The macro @c CXXOMFORT_IMPLEMENTS_invoke informs about implementation status.
 * 
 * The macro @c __cpp_lib_invoke is defined when the feature is backported.
 * 
 * @note In C++11 and C++14 mode, @c invoke() may not necessarily be @c constexpr .
 * 
 * @note Pre-C++11, <tt>invoke()</tt> is partially supported; in particular, @c invoke() is mostly / only supported for compilers that have <tt>CXXO_COMPILER_SUPPORT_typeof == true</tt> , and is limted to C functions and object-like callables used in expressions where a return is explicitly requested (eg.: at the right-hand side of an assignment expression). Older compilers without typeof or any other similar measure should be using @c invoke_r() instead. 
 * 
 * @note For pre-C++11 compilers without typeof, @em if @c invoke() is supported, <tt>invoke()</tt> only supports object-like callables. To use with eg.: member function pointers, wrap those in <tt>mem_fn()</tt>, <tt>std::function</tt>, etc.
 * 
 * */
template <typename F, typename... Args>
dependent_type_t invoke (F&& callable, Args&&... args) {
}

/**
 * @anchor cxxostd_invoke_r
 * @brief Invokes a callable and specifies a return type. See also @c invoke() .
 * @ingroup std0functional
 * @cxx23b Invoke callables via <tt>(R)callable( args... )</tt>, specifying the return type.
 * @xrefitem std0functional "" "" Invoke callables via <tt>(R)callable( args... )</tt>, specifying the return type.
 * @see @cppref{utility/functional/invoke} 
 * @see @c invoke()
 * @see @c is_invocable
 *
 * @c invoke_r() calls a functionlike or callable @a f with the given arguments @a args... , taking care of forwarding, member function pointers, and implicitly converts the resulting value to type @a R during return.
 * 
 * <tt>invoke_r<R>(f, 1, 2, 3, ...)</tt> is in general equivalent to <tt>static_cast<R>(f(1, 2, 3, ...))</tt>. 
 * 
 * The @c invoke_r() feature is backported down to C++11 with full support, and down to C++98 with @e partial support. Unlike @c invoke() which requires type deduction for return, @c invoke_r() can be used even in compilers that lack @c typeof . 
 * 
 * The macro @c __cpp_lib_invoke_r is defined when the feature is backported.
 * 
 * @note In C++11 and C++14 mode, @c invoke_r() may not necessarily be @c constexpr .
 * 
 * @note Pre-C++11, <tt>invoke_r()</tt> only supports object-like callables. To use with eg.: member function pointers, wrap those in <tt>mem_fn()</tt>, <tt>std::function</tt>, etc.
 * 
 * */
template <typename R, typename F, typename... Args>
R invoke_r (F&& callable, Args&&... args) {
}

/**
 * @brief Trait that describes if a functor of type @a F is invocable.
 * @ingroup std0type_traits
 * 
 * */
template <typename F, typename... Args>
struct is_invocable{};

template <typename R, typename F, typename... Args>
struct is_invocable_r{};

} }

#elif (defined(__cpp_lib_invoke))
	#undef  CXXOMFORT_IMPLEMENTS_invoke
	#define CXXOMFORT_IMPLEMENTS_invoke CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1900))
	#undef  CXXOMFORT_IMPLEMENTS_invoke
	#define CXXOMFORT_IMPLEMENTS_invoke CXXO_IMPLSTATUS_NATIVE()
	#define __cpp_lib_invoke (CXXOMFORT_CXX_STD*100)

#elif (CXXO_COMPILER_SUPPORT_auto && CXXO_COMPILER_SUPPORT_decltype && CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_trailing_returns && CXXO_COMPILER_SUPPORT_variadic)
	/* C++11 support required: auto, variadics, trailing returns, forwarding arguments.
	 * */
	#undef  CXXOMFORT_IMPLEMENTS_invoke
	#define CXXOMFORT_IMPLEMENTS_invoke (CXXO_IMPLSTATUS_BACKPORT()+10)
	#define __cpp_lib_invoke (CXXOMFORT_CXX_STD*100)

	#include <functional>
	#include <utility>

namespace cxxomfort {
namespace cxxostd {

template <typename F, typename... Args>
auto invoke(F&& f, Args&&... args) 
-> decltype(std::forward<F>(f)(std::forward<Args>(args)...)) {
    return std::forward<F>(f)(std::forward<Args>(args)...);
}

template <typename B, typename T, typename... Args>
auto invoke (T B::*pmf, Args&&... args) 
-> decltype( invoke (std::mem_fn(pmf), std::forward<Args>(args)...) ) {
//	return std::mem_fn(pmf)
//	(std::forward<P>(p), std::forward<Args>(args)... );
	return invoke (std::mem_fn(pmf), std::forward<Args>(args)...);
}

} } // cxxomfort::cxxostd

namespace std {
	using ::cxxomfort::cxxostd::invoke;
}

#elif ((CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof) && CXXO_COMPILER_SUPPORT_rvref)
	#undef  CXXOMFORT_IMPLEMENTS_invoke
	#define CXXOMFORT_IMPLEMENTS_invoke (CXXO_IMPLSTATUS_BACKPORT())
	#define __cpp_lib_invoke (CXXOMFORT_CXX_STD*100)
	
	#include <functional>
	#include <utility> // forward

namespace cxxomfort { namespace cxxostd {

namespace detail {
	template <typename T> T& declval ();

// base INVOKE

template <typename F>
inline decltype(declval<F>()()) 
INVOKE( F&& f) {
	return std::forward<F>(f)();
}

template <typename F, typename A1>
inline decltype(declval<F>()( declval<A1>() )) 
INVOKE( F&& f, A1&& a1 ) {
	return std::forward<F>(f)
	(std::forward<A1>(a1)/*,*/);
}

template <typename F, typename A1, typename A2>
inline decltype(declval<F>()( declval<A1>() , declval<A2>() )) 
INVOKE( F&& f, A1&& a1, A2&& a2 ) {
	return std::forward<F>(f)
	(std::forward<A1>(a1), std::forward<A2>(a2)/*,*/);
}

template <typename F, typename A1, typename A2, typename A3>
inline decltype(declval<F>()( declval<A1>() , declval<A2>() , declval<A3>() )) 
INVOKE( F&& f, A1&& a1, A2&& a2, A3&& a3 ) {
	return std::forward<F>(f)
	(std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3)/*,*/);
}

template <typename F, typename A1, typename A2, typename A3, typename A4>
inline decltype(declval<F>()( declval<A1>(), declval<A2>(), declval<A3>(), declval<A4>() )) 
INVOKE( F&& f, A1&& a1, A2&& a2, A3&& a3, A4&& a4 ) {
	return std::forward<F>(f)
	(std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3),std::forward<A4>(a4)/*,*/);
}

/* 
template <typename Base, typename T, typename Derived>
inline decltype(std::declval<Derived>().*) 
INVOKE(T Base::*pmd, Derived&& ref) {
	return std::forward<Derived>(ref).*pmd;
}
 
template <class PMD, class Pointer>
inline auto INVOKE(PMD pmd, Pointer&& ptr) ->
    decltype((*std::forward<Pointer>(ptr)).*pmd) {
      return (*std::forward<Pointer>(ptr)).*pmd;
}
 
*/

// pointer to member function INVOKE

/*
template <class Base, class T, class Derived, class... Args>
inline auto INVOKE(T Base::*pmf, Derived&& ref, Args&&... args) ->
    decltype((std::forward<Derived>(ref).*pmf)(std::forward<Args>(args)...)) {
      return (std::forward<Derived>(ref).*pmf)(std::forward<Args>(args)...);
}
*/

template <class Base, class T, class Derived>
inline decltype((std::forward<Derived>(ref).*pmf)() )
INVOKE(T Base::*pmf, Derived&& ref) {
	return (std::forward<Derived>(ref).*pmf)( );
}

template <class Base, class T, class Derived, typename A1>
inline decltype((std::forward<Derived>(ref).*pmf)( declval<A1>() ) )
INVOKE(T Base::*pmf, Derived&& ref, A1&& a1) {
	return (std::forward<Derived>(ref).*pmf)
	(std::forward<A1>(a1) );
}

template <class Base, class T, class Derived, typename A1, typename A2>
inline decltype((declval<Derived>().*)( declval<A1>(), declval<A2>() ) )
INVOKE(T Base::*pmf, Derived&& ref, A1&& a1, A2&& a2) {
	return (std::forward<Derived>(ref).*pmf)
	(std::forward<A1>(a1), std::forward<A2>(a2) );
}

template <class Base, class T, class Derived, typename A1, typename A2, typename A3>
inline decltype((std::forward<Derived>(ref).*pmf)( declval<A1>(), declval<A2>(), declval<A3>() ) )
INVOKE(T Base::*pmf, Derived&& ref, A1&& a1, A2&& a2, A3&& a3) {
	return (std::forward<Derived>(ref).*pmf)
	(std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3) );
}


 

/*template <class PMF, class Pointer, class... Args>
inline auto INVOKE(PMF pmf, Pointer&& ptr, Args&&... args) ->
    decltype(((*std::forward<Pointer>(ptr)).*pmf)(std::forward<Args>(args)...)) {
      return ((*std::forward<Pointer>(ptr)).*pmf)(std::forward<Args>(args)...);
}
*/

template <class PMF, class Pt>
inline decltype(std::mem_fn(declval<PMF>())( declval<Pt>() ) ) 
INVOKE(PMF pmf, Pt&& ptr) {
	return std::mem_fn(pmf)( std::forward<Pt>(ptr) );
}

// invoke(&type::memfun, obj, a1)
template <class PMF, class Pt, typename A1>
inline decltype(std::mem_fn(declval<PMF>())( declval<Pt>(), declval<A1>() ) ) 
INVOKE(PMF pmf, Pt&& ptr, A1&& a1) {
	return std::mem_fn(pmf)
	(std::forward<Pt>(ptr), std::forward<A1>(a1) );
}

template <class PMF, class Pt, typename A1, typename A2>
inline decltype(((*std::declval<Pt>()).*)( declval<A1>(), declval<A2>() ) ) 
INVOKE(PMF pmf, Pt&& ptr, A1&& a1, A2&& a2) {
	return std::mem_fn(pmf)
	(std::forward<A1>(a1), std::forward<A2>(a2) );
}

template <class PMF, class Pt, typename A1, typename A2, typename A3>
inline decltype(((*std::declval<Pt>()).*)( declval<A1>(), declval<A2>(), declval<A3>() ) ) 
INVOKE(PMF pmf, Pt&& ptr, A1&& a1, A2&& a2, A3&& a3) {
	return std::mem_fn(pmf)
	(std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3) );
}

} // detail


template <typename F>
CXXO_TYPEOF( detail::INVOKE(detail::declval<F>() ))
invoke (F&& f) {
	return detail::INVOKE(std::forward<F>(f));
}


template <typename F, typename A1>
CXXO_TYPEOF( detail::INVOKE(detail::declval<F>(), detail::declval<A1>() ))
invoke (F&& f, A1&& a1) {
	return detail::INVOKE(std::forward<F>(f),std::forward<A1>(a1));
}

template <typename F, typename A1, typename A2>
CXXO_TYPEOF( detail::INVOKE(detail::declval<F>(), detail::declval<A1>(), detail::declval<A2>() ))
invoke (F&& f, A1&& a1, A2&& a2) {
	return detail::INVOKE(std::forward<F>(f),std::forward<A1>(a1),std::forward<A2>(a2));
}

template <typename F, typename A1, typename A2, typename A3>
CXXO_TYPEOF( detail::INVOKE(detail::declval<F>(), detail::declval<A1>(), detail::declval<A2>(), detail::declval<A3>() ))
invoke (F&& f, A1&& a1, A2&& a2, A3&& a3) {
	return detail::INVOKE(std::forward<F>(f),std::forward<A1>(a1),std::forward<A2>(a2),std::forward<A3>(a3));
}


template <typename F, typename A1, typename A2, typename A3, typename A4>
CXXO_TYPEOF( detail::declval<F>() ( detail::declval<A1>() , detail::declval<A2>() , detail::declval<A3>() , detail::declval<A4>() ) )
invoke (F&& f, A1&& a1, A2&& a2, A3&& a3, A4&& a4) {
	return std::forward<F>(f)
	(std::forward<A1>(a1),std::forward<A2>(a2),std::forward<A3>(a3),std::forward<A4>(a4)) ;
}

} }

namespace std {
	using ::cxxomfort::cxxostd::invoke;
}

// 
// basic C++11 support (no variadic)
// 
#elif (CXXOMFORT_CXX_STD>=1997 && ( CXXO_COMPILER_SUPPORT_decltype || CXXO_COMPILER_SUPPORT_typeof) )
	#undef  CXXOMFORT_IMPLEMENTS_invoke
	#define CXXOMFORT_IMPLEMENTS_invoke (CXXO_IMPLSTATUS_EMULATION()+1)
	#define __cpp_lib_invoke (CXXOMFORT_CXX_STD*100)
	/* Compilers that support typeof. In this case, 
	 * invoke (f, t1, t2, ..., tN) is, more or less:
 * 
 * * if f is function member pointer:
 *   (t1.*f)(t2, ...)
 * 
 * * elif f is data member pointer and N==1
 *   (t1.*f)
 * 
 * * else
 *   f(t1,t2, ...)
 * 
 * 
 * */
	#include <cxxomfort/utility.hpp> // declval
	#include <cxxomfort/impl/11-conditional.hpp> // conditional
	#include <cxxomfort/type_traits.hpp> 

namespace cxxomfort {
namespace cxxostd {

// some older compilers need a definition
#if (CXXOMFORT_CXX_STD>=1997)
template <typename T> T& declval1 ();
#else
template <typename T> T& declval1 () {};
#endif

// invoke of something without arguments calls as a nullary

template <typename F>
CXXO_TYPEOF( declval1<F>() () ) 
invoke (F f) {
	return std::ref(f)();
}

/*
 * invoke f with n>1 args is:
 * * (a1.f)(a2, ...) if f is a mfp
 * * (a1.*f)(a2, ...) if f is a mop
 * * f(a1, a2, ...) otherwise
 * 
 * */

//! @internal
template <typename F, typename A1, bool M= std::is_member_pointer<F>::value >
struct result_of_invoke1 {
	enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef CXXO_TYPEOF( declval1<F>()( declval1<A1>() ) )
	result_type;
	
	static result_type	
	call (F f, A1 a1) {
		return f(a1);
	};
	
};

// dispatches depending on if F is a member function or member object pointer

template <typename F, typename A1, bool MF = std::is_member_function_pointer<F>::value >
struct mp_helper1 {
	typedef typename std::tr1::result_of<F(A1)>::type result_type;

	static result_type
	call (F f, A1 a1) {
		return (a1.*f)();
	}
};

template <typename F2, typename A1>
struct mp_helper1<F2,A1,false> /* object pointer*/ {
	typedef typename std::tr1::result_of<F2(A1)>::type result_type;

	static result_type
	call (F2 f, A1 a1) {
		return (a1.*f);
	}
};

//! @internal
template <typename F, typename A1>
struct result_of_invoke1<F,A1,true> /* member pointe */ {
	enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef typename std::tr1::result_of<F(A1)>::type result_type;
	
	static result_type 
	call (F f, A1 a1) {
		return mp_helper1<F,A1>::call(f,a1);
	};
	
};


template <typename F, typename A1>
typename result_of_invoke1<F,A1>::result_type
invoke (F f, A1 a1) {
	return result_of_invoke1<F,A1>::call(f,a1);
}

// invoke (2 args)

//! @internal
template <typename F, typename A1, typename A2, bool M= std::is_member_pointer<F>::value>
struct result_of_invoke2 {
	enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef CXXO_TYPEOF( declval1<F>() ( declval1<A1>() , declval1<A2>() ) ) 
	result_type;

	static result_type	
	call (F& f, A1& a1, A2& a2) {
		return f(a1,a2);
	};

};


//! @internal
template <typename F, typename A1, typename A2>
struct result_of_invoke2<F,A1,A2,true> /* member pointe */ {
	// enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef typename std::tr1::result_of<F(A1,A2)>::type result_type;
	
	static result_type 
	call (F& f, A1& a1, A2& a2) {
		return std::mem_fn(f)(a1,a2);
	};
	
};

template <typename F, typename A1, typename A2>
typename result_of_invoke2<F,A1,A2>::result_type
invoke (F f, A1 a1, A2 a2) {
	return result_of_invoke2<F,A1,A2>::call(f,a1,a2);
}

	#if (CXXOMFORT_CXX_STD < 2011 && CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC)
template <typename F, typename A1, typename A2>
typename result_of_invoke2<F*,A1,A2>::result_type
invoke (F* f, A1 a1, A2 a2) {
	return result_of_invoke2<F*,A1,A2>::call(f,a1,a2);
}

	#endif

// invoke (3 args)

//! @internal
template <typename F, typename A1, typename A2, typename A3, bool M= std::is_member_pointer<F>::value>
struct result_of_invoke3 {
	typedef CXXO_TYPEOF( declval1<F>() ( declval1<A1>() , declval1<A2>(), declval1<A3>() ) ) 
	result_type;

	static result_type	
	call (F f, A1 a1, A2 a2, A3 a3) {
		return f(a1,a2,a3);
	};

};

//! @internal
template <typename F, typename A1, typename A2, typename A3>
struct result_of_invoke3<F,A1,A2,A3,true> /* member pointe */ {
	enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef typename std::tr1::result_of<F(A1,A2,A3)>::type result_type;
	
	static result_type 
	call (F f, A1 a1, A2 a2, A3 a3) {
		return (a1.*f)(a2,a3);
	};
	
};


template <typename F, typename A1, typename A2, typename A3>
typename result_of_invoke3<F,A1,A2,A3>::result_type
invoke (F f, A1 a1, A2 a2, A3 a3) {
	return result_of_invoke3<F,A1,A2,A3>::call(f,a1,a2,a3);
}

// invoke (4 args)

//! @internal
template <typename F, typename A1, typename A2, typename A3, typename A4, bool M= std::is_member_pointer<F>::value>
struct result_of_invoke4 {
	typedef CXXO_TYPEOF( declval1<F>() ( declval1<A1>() , declval1<A2>(), declval1<A3>(), declval1<A4>() ) ) 
	result_type;

	static result_type	
	call (F f, A1 a1, A2 a2, A3 a3, A4 a4) {
		return f(a1,a2,a3,a4);
	};

};

//! @internal
template <typename F, typename A1, typename A2, typename A3, typename A4>
struct result_of_invoke4<F,A1,A2,A3,A4,true> /* member pointe */ {
	enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef typename std::tr1::result_of<F(A1,A2,A3,A4)>::type result_type;
	
	static result_type 
	call (F f, A1 a1, A2 a2, A3 a3,A4 a4) {
		return (a1.*f)(a2,a3,a4);
	};
	
};


template <typename F, typename A1, typename A2, typename A3, typename A4>
typename result_of_invoke4<F,A1,A2,A3,A4>::result_type
invoke (F f, A1 a1, A2 a2, A3 a3, A4 a4) {
	return result_of_invoke4<F,A1,A2,A3,A4>::call(f,a1,a2,a3,a4);
}


// invoke (5 args)

//! @internal
template <typename F, typename A1, typename A2, typename A3, typename A4, typename A5, bool M= std::is_member_pointer<F>::value>
struct result_of_invoke5 {
	typedef CXXO_TYPEOF( declval1<F>() ( declval1<A1>() , declval1<A2>(), declval1<A3>(), declval1<A4>(), declval1<A5>() ) ) 
	result_type;

	static result_type	
	call (F f, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5) {
		return f(a1,a2,a3,a4,a5);
	};

};

//! @internal
template <typename F, typename A1, typename A2, typename A3, typename A4, typename A5>
struct result_of_invoke5<F,A1,A2,A3,A4,A5,true> /* member pointe */ {
	enum { mfp = std::is_member_function_pointer<F>::value , dfp = std::is_member_object_pointer<F>::value };
	typedef typename std::tr1::result_of<F(A1,A2,A3,A4,A5)>::type result_type;
	
	static result_type 
	call (F f, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5) {
		return (a1.*f)(a2,a3,a4,a5);
	};
	
};


template <typename F, typename A1, typename A2, typename A3, typename A4, typename A5>
typename result_of_invoke5<F,A1,A2,A3,A4,A5>::result_type
invoke (F f, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5) {
	return result_of_invoke5<F,A1,A2,A3,A4,A5>::call(f,a1,a2,a3,a4,a5);
}


} } // cxxomfort::cxxostd

namespace std {
	using ::cxxomfort::cxxostd::invoke;
}

#else // not even typeof support - nothing to do
	#undef  CXXOMFORT_IMPLEMENTS_invoke
	#define CXXOMFORT_IMPLEMENTS_invoke (0)

namespace cxxomfort { namespace cxxostd {

/* without typeof or any means to determine the return type, all we can do 
 * is to invoke the required function and ignore the result type.
 * 
 * In theory we *could* return the type we're requested via a template conversion operator, but this would work in only but a few cases. As soon as the caller side requests via a template (eg.: "cout<< invoke(...)"), requests an overloaded call, or doesn't request a type at all (for eg.: void invocables), this would break with hard compile errors that would be impossible to pin down by compiler/version. 
 * 
 * Pre-typeof code wanting to invoke a callable and retrieve a result should be using invoke_r(), which is supported.
 * */

template <typename Fn>
struct invoke_wrapper_0 {
	Fn& f;
	explicit invoke_wrapper_0( Fn& fn_)
	: f(fn_) 
	{ }
	
};

template <typename Fn>
invoke_wrapper_0<Fn> 
invoke (Fn fn) {
	return invoke_wrapper_0<Fn>(fn);
}


template <typename Fn, typename A1>
struct invoke_wrapper_1 {
	Fn f;
	A1 const& a1;
	bool invoked;
	
	typedef typename std::tr1::result_of<Fn(A1)>::type Ret;
	
	invoke_wrapper_1 (Fn& fn_, A1 const& a1_)
	: f(fn_), a1(a1_), invoked(false)
	{ }
	
	operator Ret () {
		invoked= true;
		return f(a1);
	}
	
	~invoke_wrapper_1 () {
		if (!invoked) { f(a1); }
	}
};

template <typename Fn, typename A1>
invoke_wrapper_1<Fn,A1> 
invoke (Fn fn, A1 const& a1) {
	std::cout<< "[invoke fn(A1) : "<< cxxomfort::typeid_demangle(typeid(fn))<< " ]"<< std::flush;
	return invoke_wrapper_1<Fn,A1>(fn,a1);
}

template <typename Fn, typename A1, typename A2>
struct invoke_wrapper_2 {
	//typedef typename std::decay<Fn>::type DFn;
	Fn f;
	A1 const& a1;
	A2 const& a2;
	bool invoked;
	invoke_wrapper_2 (Fn& fn_, A1 const& a1_, A2 const& a2_)
	: f(fn_), a1(a1_), a2(a2_), invoked(false)
	{
		//std::cout<< "[A1: "<<a1<< "]"<< std::flush<< "[A2: "<< a2<< "]"<< std::flush; 
	}
	
	template <typename T>
	operator T () {
		//std::cout<< "[invoke operator()]"<< std::flush;
		invoked= true;
		return f(a1,a2);
	}
	
	~invoke_wrapper_2 () {
		if (!invoked) { 
			//std::cout<< "[invoke dtor]"<< std::flush;
			f(a1,a2); 
		}
	}
};

template <typename Fn, typename A1, typename A2>
invoke_wrapper_2<Fn,A1,A2> 
invoke (Fn fn, A1 const& a1, A2 const& a2) {
	//std::cout<< "[invoke fn(A1,A2) : "<< cxxomfort::typeid_demangle(typeid(fn))<< " ]"<< std::flush;
	return invoke_wrapper_2<Fn,A1,A2>(fn,a1,a2);
}

} } // cxxomfort::cxxostd

namespace std {
	using ::cxxomfort::cxxostd::invoke;
}


#endif

// 
// is_invocable (C++11) 
// 

#if (CXXO_COMPILER_SUPPORT_invoke)
#elif (CXXOMFORT_CXX_STD >= 2011)
	#include "./n4502-detection_idiom.hpp"

namespace cxxomfort { namespace cxxostd {

template <typename F, typename... Args>
using f_invoke_check = decltype(std::invoke(std::declval<F>(), std::declval<Args>()... ) );

template <typename F, typename... Args>
struct is_invocable 
: std::integral_constant< bool, cxxomfort::impl::is_detected<f_invoke_check, F, Args...>::value >
{ };


} }

namespace std {
	using cxxomfort::cxxostd::is_invocable;
}


#elif (CXXO_COMPILER_SUPPORT_typeof)

#else

#endif



// 
// invoke_result 
// 

#if (defined(__cpp_lib_is_invocable))

#elif (defined(__cpp_lib_invoke) && CXXO_COMPILER_SUPPORT_typeof > 0 && CXXO_COMPILER_SUPPORT_variadic)
	#define __cpp_lib_is_invocable 1

namespace std {
namespace detail {
template <typename AlwaysVoid, typename, typename...>
struct invoke_result { };

template <typename F, typename...Args>
struct invoke_result<decltype(void(std::invoke(std::declval<F>(), std::declval<Args>()...))),
                 F, Args...> {
    using type = decltype(std::invoke(std::declval<F>(), std::declval<Args>()...));
};
} // namespace detail
 
template <class F, class... ArgTypes>
struct invoke_result : detail::invoke_result<void, F, ArgTypes...> {};

}

#endif

// 
// 
// 

#endif
