#ifndef CXXOMFORT_IMPL_17_ITERATOR_ACCESS_HPP
#define CXXOMFORT_IMPL_17_ITERATOR_ACCESS_HPP
/**
 * @file
 * @brief Implementation of std::size , std::data , std::empty 
 */
#include <cxxomfort/config.hpp>
#include <cxxomfort/base/iterator.hpp> // also includes <iterator>
#include <cstddef>

#if (defined(__cpp_lib_nonmember_container_access))
	#undef  CXXO_COMPILER_SUPPORT_iterator_accessors_17
	#define CXXO_COMPILER_SUPPORT_iterator_accessors_17 1
#endif

#if (CXXO_COMPILER_SUPPORT_iterator_accessors_17)
	#define CXXOIMPL_size_ CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC, >=1900))
	#define CXXOIMPL_size_ CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD >= 2011 && CXXO_COMPILER_SUPPORT_trailing_returns>0)
	#define CXXOIMPL_size_ CXXO_IMPLSTATUS_BACKPORT()
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++17 size() backport")
	#endif

namespace cxxomfort { namespace cxxostd {

template <typename C> inline constexpr 
auto size (C const& container) -> decltype(container.size()) {
    return container.size();
}

// overload for native arrays
template <typename T, size_t N> inline constexpr 
size_t size (T(&)[N]) {
    return N;
}

} } 

namespace std {
	using ::cxxomfort::cxxostd::size;
}

#elif (CXXOMFORT_CXX_STD >= 1997)
	#define CXXOIMPL_size_ CXXO_IMPLSTATUS_EMULATION()
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++17 size() emulation")
	#endif

namespace cxxomfort { namespace cxxostd {

namespace detail_size {

template <typename T>
struct iterator_accessors_helper_size {

    typedef struct { int a; } yes;
    typedef char no;

    private:
    template <typename U, U> struct really_has;

    template <typename C> static yes& Test_size(really_has <typename C::difference_type (C::*)() const,
                            &C::size>*);

    template <typename C> static yes& Test_size(really_has <size_t (C::*)() const,
                            &C::size>*);

    template <typename> static no& Test_size(...);
    
    public:
    static const bool has_size     = (sizeof(Test_size<T>(0)) == sizeof(yes));


};


template <typename C, bool F=false> struct call_size_has {
    typename C::difference_type operator() (C const& c) const { 
        return std::distance(c.begin() , c.end());
    }
}; // call_size_has

template <typename C> struct call_size_has<C,true> {
    typename C::size_type operator() (C const& c) const { 
        return c.size();
    }
}; // call_size_has

} // detail_size

template <typename C> inline
typename C::size_type
size (C & c) {
    //typedef typename C::iterator iterator;
    const detail_size::call_size_has< 
        C, detail_size::iterator_accessors_helper_size<C>::has_size 
    > s = {};
    return s(c);
}
} }

namespace std {
	using ::cxxomfort::cxxostd::size;
}

#else 
namespace std {
template <typename C> inline 
typename C::size_type size (C const& c) { return c.size(); }

}
#endif


/*
 * data, from c++17
 */

#if (defined(DOXYGEN_DOC))
namespace std {

/**
 * @brief provides access to an internal "<code>.data()</code>" member.
 * @ingroup std0iterator
 * @cxx17b Invoked <tt>c.data()</tt> or a similar construct for contiguous-storage sequences.
 * 
 * <tt>data(c)</tt> invokes <tt>c.data()</tt> for container for which it exists. The result type is assumed to be a pointer type or equivalent that provides direct access to a contiguous storage.
 * 
 * @warning For C++ pre-C++11, this relies on the container type declaring this pointer type as the member types <tt>pointer</tt> and <tt>const_pointer</tt>.
 * 
 * */
template <typename C>
constexpr auto data (C& container) -> decltype(container.data()) {}

template <typename C>
constexpr auto data (C const& container) -> decltype(container.data()) {}

}
#elif (CXXOMFORT_CXX_STD >= 2017 || CXXO_COMPILER_SUPPORT_iterator_accessors_17)
    // nothing to do
	#define CXXOIMPL_std_data_ CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD >= 2011 && CXXO_COMPILER_SUPPORT_trailing_returns>0)
	#include <initializer_list>
	#define CXXOIMPL_std_data_ CXXO_IMPLSTATUS_BACKPORT()

namespace cxxomfort { namespace cxxostd {

template <typename C> inline constexpr 
auto data (C& container) -> decltype(container.data()) {
    return container.data();
}

template <typename C> inline constexpr 
auto data (C const& container) noexcept -> decltype(container.data()) {
    return container.data();
}

template <typename T>  inline constexpr 
const T* data (std::initializer_list<T> const& i) noexcept {
    return i.begin();
}

} } // namespace
#else
    // pre-C++11
	#define CXXOIMPL_std_data_ CXXO_IMPLSTATUS_EMULATION()

namespace cxxomfort { namespace cxxostd {

template <typename C>
typename C::pointer data (C& container) {
    return container.data();
}

template <typename C>
typename C::const_pointer data (C const& container) {
    return container.data();
}

} }
#endif

// overloads for native arrays
namespace cxxomfort { namespace cxxostd {

template <typename T, size_t N> inline CXXO_CONSTEXPR 
T* data (T (&a)[N]) CXXO_NOEXCEPT {
    return a;
}
} } // namespace

#if (!defined(CXXOFLAG_NO_STD_USING) && CXXOIMPL_std_data_ > 0 && CXXOIMPL_std_data_ < CXXO_IMPLSTATUS_NATIVE) 
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++17 data() backport/emulation")
	#endif
namespace std {
	using ::cxxomfort::cxxostd::data;
}

#endif


/*
 * empty, from c++17
 */

#if (defined(DOXYGEN_DOC))
namespace std {
/**
 * @brief provides access to an internal "<code>.empty()</code>" member.
 * @ingroup std0iterator
 * @cxx17b "Iterators" empty
 * 
 * <tt>empty(c)</tt> invokes <tt>c.empty()</tt> for container for which it exists. The result type is assumed to be a @c bool .
 * 
 * @warning For C++ pre-C++11, this relies on the container type declaring a <tt>value_type</tt> member type.
 * 
 * */
template <typename C>
bool empty (C const& container) {}

}
#elif (CXXOMFORT_CXX_STD >= 2017 || CXXO_COMPILER_SUPPORT_iterator_accessors_17>0)
    // nothing to do
	#define CXXOIMPL_empty_ CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD >= 2011)
	#include <initializer_list>
	#define CXXOIMPL_empty_ CXXO_IMPLSTATUS_BACKPORT()
	
namespace cxxomfort { namespace cxxostd {

template <typename C> inline 
bool empty (C const& container) {
    return container.empty();
}

template <typename T> inline 
constexpr bool empty (std::initializer_list<T> const& i) noexcept {
    return i.size()==0;
}

} }
#else
    // pre-C++11
	#define CXXOIMPL_empty_ CXXO_IMPLSTATUS_EMULATION()

namespace cxxomfort { namespace cxxostd {

template <typename C> inline 
bool empty (C const& container) {
    return container.empty();
}

} }
#endif

// overload for native arrays
namespace cxxomfort { namespace cxxostd {

template <typename T, size_t N> inline CXXO_CONSTEXPR 
bool empty (T const (&)[N]) CXXO_NOEXCEPT {
    return false;
}

} } // cxxomfort::cxxostd

#if (!defined(CXXOFLAG_NO_STD_USING) && CXXOIMPL_std_data_ > 0 && CXXOIMPL_std_data_ < CXXO_IMPLSTATUS_NATIVE) 
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++17 empty() backport/emulation")
	#endif
namespace std {
	using ::cxxomfort::cxxostd::empty;
}

#endif


#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (CXXOMFORT_CXX_STD>=2017 || CXXO_COMPILER_SUPPORT_iterator_accessors_17>0)
	#define CXXOMFORT_IMPLEMENTS_n4280 CXXO_IMPLSTATUS_NATIVE()
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1904) && CXXOMFORT_CXX_EMULATION >= 2014)
	#define CXXOMFORT_IMPLEMENTS_n4280 CXXO_IMPLSTATUS_BACKPORT()
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++17 iterator accessors backport")
	#endif
	namespace std {
		// size already is in xutility
        using ::cxxomfort::cxxostd::data;
        using ::cxxomfort::cxxostd::empty;
	}
#else
	#if (CXXOIMPL_size_ == CXXO_IMPLSTATUS_BACKPORT() && CXXOIMPL_data_ == CXXO_IMPLSTATUS_BACKPORT() || CXXOIMPL_empty_ == CXXO_IMPLSTATUS_BACKPORT())
		#define CXXOMFORT_IMPLEMENTS_n4280 CXXO_IMPLSTATUS_BACKPORT()
	#else
		#define CXXOMFORT_IMPLEMENTS_n4280 CXXO_IMPLSTATUS_EMULATION()
	#endif
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("c++17 iterator accessors backport/emulation")
	#endif

    namespace std {
        using ::cxxomfort::cxxostd::data;
        using ::cxxomfort::cxxostd::empty;
    }
#endif // no std using


#ifdef DOXYGEN_DOC
namespace cxxomfort { namespace cxxostd {

/**
 * @brief Accesses and returns the @c .empty() member of the container.
 * @return A bool, equal to calling @c .empty() on the container.
 * @ingroup cxx17-backports
 * 
 * */
template <typename C>
bool empty (C const& c) { return c.empty(); }

/**
 * @brief Accesses and returns the @c .size() member of the container.
 * @return A numeric type, equal to calling @c .size() on the container.
 * 
 * */
template <typename C>
size_t size (C const& c) { return c.size(); }

/**
 * @brief Accesses and returns the @c .data() member of the container.
 * @return A pointer, likely of type <tt>{containertype}::pointer</tt> .
 * @ingroup cxx17-backports
 * 
 * */
template <typename C>
typename C::pointer data (C const& c) { return c.data(); }


} }

#endif

#undef  CXXOMFORT_IMPLEMENTS_iterator_accessors_17
#define CXXOMFORT_IMPLEMENTS_iterator_accessors_17 CXXOMFORT_IMPLEMENTS_n4280

#endif
