#ifndef CXXOMFORT_IMPL_STRING_VIEW_HPP
#define CXXOMFORT_IMPL_STRING_VIEW_HPP
/**
 * @file
 * @brief Header implementing C++17's @c basic_string_view .
 * @ingroup cxx17-backports
 *
 */

#include <cstring>
#include <string> // TODO: try and avoid this
#include <cxxomfort/utils/char_traits.hpp>
#include <cxxomfort/utils/array_ref.hpp>
#include <cxxomfort/algorithm.hpp>
#include <cxxomfort/impl/07-type_traits.hpp>

#if (1 || CXXOMFORT_CXX_STD>=1997)

#define CXXOMFORT_IMPLEMENTS_n3605 CXXO_IMPLSTATUS_BACKPORT()
#define CXXOMFORT_IMPLEMENTS_n3685 CXXO_IMPLSTATUS_BACKPORT()
#define CXXOMFORT_IMPLEMENTS_n3762 CXXO_IMPLSTATUS_BACKPORT()
#undef  CXXOMFORT_IMPLEMENTS_string_view
#define CXXOMFORT_IMPLEMENTS_string_view CXXOMFORT_IMPLEMENTS_n3685
#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
    #pragma message CXXO_NOTICE("basic_string_view internal implementation")
#endif

#if defined(DOXYGEN_DOC)
namespace std {
#else
namespace cxxomfort { namespace cxxostd {
#endif

namespace {
	CXXO_CONSTEXPR14 inline
	size_t ctlength (char const* str) CXXO_NOEXCEPT {
		size_t s= 0;
		while (*str!=0) {++s; ++str;}
		return s;
	}
}

/**
 * @brief Read-only view of a @c string or string-like object, from C++17. See also @cppref{string/basic_string_view} .
 * @cxx17b A view over a string of character.
 * @xrefitem std0string_view "Classes" "" A view over a string of character.
 * 
 * A @c basic_string_view is a view (non-owning) to a buffer of characters representing a string object, and given a set of operations.
 * 
 * @section differences Differences with std::basic_string_view
 * 
 * Cxxomfort's @c basic_string_view has a number of differences from the std version in C++17 onwards, documented here for clarity. Largely the differences lie in the member functions, as cxxomfort intentionally chooses to not implement the hundreditude of member functions that @c std::basic_string_view inherits from the bad design of @c std::basic_string .
 * 
 * Compared to the std version, the cxxomfort implementation:
 * 
 * * incorporates an @c element_type member type equivalent to @c std::span 's  @c element_type
 * * incorporates a constructor from <tt>CharTy[N]</tt> for C++<11
 * * incorporates a constructor from @c std::basic_string for C++<17, to supplement the conversion operator in @c std::basic_string post-C++17 that can not be injected in previous versions
 * * incorporates an @c at_or() member that returns an alternative value instead of throwing an exception, when the index is out of bounds. (This is inherited from @c array_ref )
 * * does not implement @c remove_prefix() and @c remove_suffix() 
 * * does not implement @c starts_with() and @c end_with() from C++20, leaving them instead to the global accessors in @c cxxomfort/string.hpp .
 * * does not implement @c contains() from C++23, leaving them instead to the global accessors in @c cxxomfort/string.hpp .
 * 
 * Cxxomfort also offers the helper @ref @c sv_helper in order to construct @c basic_string_view s easily; and starting with C++11 it implements the <tt>""sv</tt> user-defined literal as well.
 * 
 * @see @cppref{string/basic_string_view} .
 * @see @ref sv_helper sv("text")
 * 
 * */

template <
class CharTy, class CTraits = std::char_traits<CharTy> 
> class basic_string_view 
: protected ::cxxomfort::array_ref<CharTy const> {
    private:
    typedef ::cxxomfort::array_ref<CharTy const> _Base;

    public:
    //! Character type employed by the string
    typedef CharTy      value_type;
    //! @c char_traits corresponding to the character type
	typedef CTraits     traits_type;
    //! Element type representing the underlying buffer view
    typedef typename _Base::element_type       element_type;
    //! Type representing the amount of characters the string can have
    typedef typename _Base::size_type          size_type;
    //! Integral type representing the difference between two iterators
    typedef typename _Base::difference_type    difference_type;
    //! Pointer type for the underlying character storage
    typedef CharTy *         pointer;
    //! Pointer type to a @c const view of the underlying character storage
    typedef CharTy const *   const_pointer;
    //! Reference type for the underlying characters
    typedef CharTy &         reference;
    //! Reference type to a @c const view of the underlying characters
    typedef CharTy const &   const_reference;
    //! Iterator type
    typedef const_pointer    const_iterator;
    typedef const_iterator                 iterator;
    //! Reverse iterator type
    typedef typename _Base::const_reverse_iterator  const_reverse_iterator;
    typedef const_reverse_iterator         reverse_iterator;

    //! Flag that indicates a position not found.
    static const size_type npos= size_type(-1);

    
    public:
    // def-ctor

#if (defined(DOXYGEN_DOC)) 
    //! default-ctor: empty string.
    constexpr basic_string_view () noexcept = default;
    //! copy-ctor
    constexpr basic_string_view (basic_string_view const&) noexcept = default;
#elif (CXXOMFORT_CXX_STD >= 2011)
    constexpr basic_string_view () noexcept = default;
    constexpr basic_string_view (basic_string_view const&) noexcept = default;
#else
    CXXO_CONSTEXPR 
    basic_string_view () CXXO_NOEXCEPTNOTHROW
    : _Base()
    { }

    CXXO_CONSTEXPR 
    basic_string_view (basic_string_view const& c) CXXO_NOEXCEPTNOTHROW
    : _Base(c) 
    { }

#endif
    
    //! @c string_view to characters at position @a px with length @a sz .
    CXXO_CONSTEXPR 
    basic_string_view (CharTy const* px, size_t sz) CXXO_NOEXCEPTNOTHROW
    : _Base(px,sz)
    { }

    //! String view to an array of characters (cxxomfort-specific constructor)
#if (CXXOMFORT_CXX_STD>=1997)
    template <size_t N>
    CXXO_CONSTEXPR 
    basic_string_view (std::nullptr_t, CharTy const (&arr)[N]) CXXO_NOEXCEPTNOTHROW 
    : _Base(arr,N)
    { }
#endif
    
    //! @c string_view to characters at position @a px and length reported by @c CTraits::length .
    // TODO: check if this should be explicit
    // NOTE: we cheat when Ch is 'char' because char_traits<>::length is not constexpr pre C++17
    CXXO_CONSTEXPR14
    basic_string_view (const_pointer const px) CXXO_NOEXCEPTNOTHROW 
    : _Base(px, (std::is_same<CTraits, std::char_traits<char> >::value)? ctlength(px) : CTraits::length(px) )
    { };
    

#if defined(DOXYGEN_DOC)
    //! Iterator constructor, added in C++20
    template <typename IBegin, typename IEnd
    , typename std::enable_if < !std::is_arithmetic<IBegin>::value && !std::is_arithmetic<IEnd>::value && std::is_same< CharTy , typename std::iterator_traits<IBegin>::value_type >::value, bool*  >::type= 0
    >
    basic_string_view (IBegin ini, IEnd fin)
    : _Base(ini, (fin-ini)) 
    { };
#elif (CXXOMFORT_CXX_STD >= 1997)
	#if (CXXOMFORT_NOTICES > 2)
		#pragma message CXXO_NOTICE("string_view impl with range constructor")
	#endif
    template <typename IBegin, typename IEnd>
    basic_string_view (
    IBegin ini, IEnd fin
, typename std::enable_if < !std::is_arithmetic<IBegin>::value && !std::is_arithmetic<IEnd>::value && std::is_same< CharTy , typename std::iterator_traits<IBegin>::value_type >::value, bool*  >::type= 0
    )
    : _Base(ini, std::distance(ini,fin)) 
    { };
#else
#endif


    /*
     * This constructor is needed pre-C++17 to match the implicit conversion operator in std::string to std::string_view
     * See https://stackoverflow.com/questions/63166194/implicitly-convert-string-to-string-view
     * */
#if (CXXOMFORT_CXX_STD >= 1997 && CXXOMFORT_CXX_STD < 2017)
	#if (CXXOMFORT_NOTICES > 2)
		#pragma message CXXO_NOTICE("string_view impl with basic_string() constructor to match c++17")
	#endif
    template <typename Allocator>
    basic_string_view (std::basic_string<CharTy,CTraits,Allocator> const& s) CXXO_NOEXCEPT 
    : _Base(s.data(),s.size())
    { }
#endif
	 
	 // C++23 deleted this constructor
#if (CXXOMFORT_CXX_STD >= 2011)
	constexpr basic_string_view (std::nullptr_t) = delete;
#endif

    using _Base::max_size;
    //! Returns string size.
    using _Base::size;
    //! Returns indicating if the string is empty.
    using _Base::empty;
    //! Returns pointer to the character sequence.
    using _Base::data;
    //! Returns the length of the viewed string.
    CXXO_CONSTEXPR 
    size_type  length () const CXXO_NOEXCEPTNOTHROW { return this->size(); }

    using _Base::front;
    using _Base::back;


    using _Base::cbegin; //< begin iterator
    using _Base::cend; //< end iterator
    using _Base::begin; //< begin iterator
    using _Base::end; //< end iterator
    
    using _Base::crbegin;
    using _Base::crend;
    using _Base::rbegin;
    using _Base::rend;

#if (CXXOMFORT_CXX_STD>=1997)
    //! Accesses characters in the string, unchecked.
    using _Base::operator[];
#else
    const_reference operator[] (size_t idx) const CXXO_NOTHROW() { 
		return static_cast<_Base const*>(this)->operator[](idx); 
	}
#endif
    //! Accesses characters in the string, unchecked.
    //using _Base::at_or;
#if defined(DOXYGEN_DOC)
    //! Accesses characters in the string, checked.
    const_reference at (size_type idx) const; 
#else
    const_reference at (size_type idx) const { 
		if (idx >= size()) CXXO_THROW_STD( std::out_of_range("string_view") );
		return this->operator[](idx); 
	}
#endif

    //! Copies contents from the viewed string to a @c CharTy array.
    size_type  copy (CharTy* d, size_type count, size_type pos=0) const;
    //! Returns a view of part of the string.
    CXXO_CONSTEXPR  
    basic_string_view    substr (size_type pos, size_type count = npos) const CXXO_NOEXCEPTNOTHROW {
		// rcount is the smaller of count and size() - pos. 
		// size_type const rcount = std::min(count, size()-pos);
		return (pos >= size() || count >= size()) ? basic_string_view() : basic_string_view(data()+pos, count < size()-pos ? count : size()-pos);
	}

    //! @c strcmp -like compare against another string view
    int 	compare (basic_string_view const r) const CXXO_NOEXCEPTNOTHROW {
		size_type rlen = size() < r.size() ? size() : r.size();
		int compareres= CTraits::compare(this->data(), r.data(), rlen);
		if (compareres==0) {
			return size() == r.size() ? 0 : (size() < r.size() ? -1 : +1);
		}
		else return compareres;
	}
    int 	compare (pointer r) const CXXO_NOEXCEPTNOTHROW {
        return compare ( basic_string_view(r) ); 
    }
    
    
    //! Find a substring.
    CXXO_CONSTEXPR14
    size_type 	find (basic_string_view const s, size_type pos=0) const CXXO_NOEXCEPTNOTHROW {
		if (pos >= this->length()) return npos;
		if (s.size() == 0 || s.size() > size()) return npos;
		using cxxomfort::fix::find;
		const_iterator st = this->begin()+pos;
		const_iterator p1 = find(cxxomfort::constexpr_tag, st, this->end(), s[0]);
		while (!( p1==end() )) {
			basic_string_view t1= substr(p1-this->begin(), s.size());
			if (t1==s) 
				return p1-this->begin();
			p1= find(cxxomfort::constexpr_tag, p1+1, this->end(), s[0]);
		}
		return npos;
		/*
		(void)p1;
		const_iterator loc = std::search(
			st, this->end(), s.begin(), s.end());
		if (loc==this->end()) return npos;
		else return static_cast<size_type>(loc - this->begin());
		* */
	}

    //! Find a character.
    CXXO_CONSTEXPR14 
    size_type 	find (CharTy const ch, size_type pos= 0) const CXXO_NOEXCEPTNOTHROW {
		if (pos >= length()) return npos;
		for (size_t i= pos; i < length(); ++i) { if ((*this)[i]==ch) return i; }
		return npos;
	}
    //! Find a substring.
    CXXO_CONSTEXPR14 
    size_type 	find (const_pointer s, size_type pos= 0) const CXXO_NOEXCEPTNOTHROW {
        return find(basic_string_view(s), pos);
    }
    //! Find a substring.
    size_type 	find (const_pointer s, size_type pos, size_type count) const CXXO_NOEXCEPTNOTHROW {
        return find(basic_string_view(s,count), pos);
    }
    
    //! Find last instance of a string.
    size_type 	rfind (basic_string_view, size_type pos=0) const CXXO_NOEXCEPTNOTHROW ;
    //! Find last instance of a character.
    size_type 	rfind (CharTy, size_type pos= 0) const CXXO_NOEXCEPTNOTHROW ;
    //! Find last instance of a string.
    size_type 	rfind (const_pointer s, size_type pos= 0) const CXXO_NOEXCEPTNOTHROW {
        return rfind(basic_string_view(s), pos);
    }
    //! Find last instance of a string.
    size_type 	rfind (const_pointer s, size_type pos, size_type count) const CXXO_NOEXCEPTNOTHROW {
        return rfind(basic_string_view(s,count), pos);
    }

    CXXO_CONSTEXPR14 friend 
    bool operator< (basic_string_view const c1, basic_string_view const c2) CXXO_NOEXCEPTNOTHROW {
        return c1.compare(c2)<0;
    }
    
    // allow conversion to a std::basic_string
    // to supplement the assignment operator for std::string in C++17
    
    CXXO_CONSTEXPR14 
    operator std::basic_string<CharTy, CTraits> () const {
		return std::basic_string<CharTy, CTraits>(this->begin(), this->end());
	}
    
};

// Copies the substring [pos, pos + rcount) to the character string pointed to by dest, 
// where rcount is the smaller of count and size() - pos
template <typename Ch, typename ChT>
typename basic_string_view<Ch,ChT>::size_type 
basic_string_view<Ch,ChT>::copy 
(Ch* dest, size_type count, size_type pos) const {
    size_t rcount = std::min(count, size()-pos);
#if (CXXOMFORT_CXX_STD>=1997)
    Ch* d= std::copy(data()+pos, data()+pos+rcount, dest);
#else
    Ch* d= dest+rcount;
    for (size_t i=0; i < rcount; ++i) { dest[i]= data()[pos+i]; }
#endif
    return d-dest;
}

// searches the last occurrence

template <typename Ch, typename ChT>
typename basic_string_view<Ch,ChT>::size_type 
basic_string_view<Ch,ChT>::rfind 
(basic_string_view<Ch,ChT> s, size_type pos) const CXXO_NOEXCEPTNOTHROW {
    if (pos >= this->length()) return npos;
    const_iterator st = this->begin()+pos;
    const_iterator loc = std::find_end(
        st, this->end(), s.begin(), s.end());
    if (loc==this->end()) return npos;
    else return static_cast<size_type>(loc - this->begin());
}

// 
// relational operators
// 

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator== (basic_string_view<Ch,ChT> const c1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
    //return c1.compare(c2)==0;
    return cxxomfort::fix::equal(cxxomfort::constexpr_tag, c1.begin(), c1.end(), c2.begin(), c2.end());
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator!= (basic_string_view<Ch,ChT> s1, basic_string_view<Ch,ChT> s2) {
    return !(s1==s2);
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator< (basic_string_view<Ch,ChT> const c1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
	return c1.compare(c2)<0;
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator> (basic_string_view<Ch,ChT> const c1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
    return c1.compare(c2)>0;
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator<= (basic_string_view<Ch,ChT> const c1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
    return c1.compare(c2)<=0;
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator>= (basic_string_view<Ch,ChT> const c1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
    return c1.compare(c2)>=0;
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator== (basic_string_view<Ch,ChT> const c1, Ch const* s2) CXXO_NOEXCEPTNOTHROW {
    return c1.compare( basic_string_view<Ch,ChT>(s2) )==0;
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator== (Ch const* s1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
    return basic_string_view<Ch,ChT>(s1).compare( c2 )==0;
}


template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator!= (basic_string_view<Ch,ChT> const c1, Ch const* s2) CXXO_NOEXCEPTNOTHROW {
    return c1.compare( basic_string_view<Ch,ChT>(s2) )!=0;
}

template <typename Ch, typename ChT>
CXXO_CONSTEXPR14 inline 
bool operator!= (Ch const* s1, basic_string_view<Ch,ChT> const c2) CXXO_NOEXCEPTNOTHROW {
    return basic_string_view<Ch,ChT>(s1).compare( c2 )!=0;
}


//! Name for a strng view of @c char .
//! @xrefitem std0string_view "" "" @c basic_string_view of @c char .
typedef basic_string_view<char, std::char_traits<char> > string_view;
//! Name for a string view of @c wchar_t .
//! @xrefitem std0string_view "" "" @c basic_string_view of @c wchar_t .
typedef basic_string_view<wchar_t, std::char_traits<wchar_t> > wstring_view;

#if (CXXOMFORT_CXX_STD >= 2011)
//! Name for a string view of @c char16_t .
typedef basic_string_view<char16_t> u16string_view;
//! Name for a string view of @c char32_t .
typedef basic_string_view<char32_t> u32string_view;
#endif

#if (CXXOMFORT_CXX_STD >= 2020)
//! Name for a string view of @c char8_t .
typedef basic_string_view<char8_t> u8string_view;
#endif

} } // cxxomfort::cxxostd


#elif 0 // older compilers like OpenWatcom

	#if (CXXOMFORT_NOTICES > 2)
		#pragma message CXXO_NOTICE("string_view zero impl")
	#endif
    #define CXXOMFORT_IMPLEMENTS_n3605 CXXO_IMPLSTATUS_EMULATION()
    #define CXXOMFORT_IMPLEMENTS_n3685 CXXO_IMPLSTATUS_EMULATION()
    #define CXXOMFORT_IMPLEMENTS_n3762 0
    #if !defined(CXXOMFORT_IMPLEMENTS_string_view)
        #define CXXOMFORT_IMPLEMENTS_string_view CXXOMFORT_IMPLEMENTS_n3685
    #endif

namespace cxxomfort { namespace cxxostd {

template <typename Char>
struct basic_string_view  {
	private:
	typedef cxxomfort::impl::array_ref::array_ref<Char const> impl_type;
	impl_type str;
	public:
	
	typedef size_t size_type;
	typedef Char value_type;
	typedef Char* pointer;
	typedef Char const* const_pointer;
	typedef Char const& reference;
	typedef Char const& const_reference;
	
	typedef const_pointer iterator;
	typedef const_pointer const_iterator;
	
	explicit basic_string_view (Char const* pt)
	: str(pt, strlen(pt)) 
	{ }
	
	basic_string_view (Char const* pt, size_t sz)
	: str(pt, sz)
	{ }

	public:
	size_type  size () const { return str.size(); }
	unsigned long  length () const { return size(); }
	
	const_iterator begin () { return str.begin(); }
	const_iterator end () { return str.end(); }

	
	const_reference at (size_type idx) const { return str.at(idx); }
	const_reference operator[] (size_type idx) const { return str[idx]; }
	
};

typedef basic_string_view<char> string_view;
typedef basic_string_view<wchar_t> wstring_view;


#if defined(DOXYGEN_DOC)
}
#else
} }
#endif

#endif

#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
    #pragma message CXXO_NOTICE("basic_string_view internal implementation successfully parsed")
#endif

#endif
