#ifndef CXXOMFORT_IMPL_TUPLEAPPLY_HPP
#define CXXOMFORT_IMPL_TUPLEAPPLY_HPP
/**
 * @file
 * @brief Backport implementation of @c std::apply 
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/utility.hpp> // forward, move

#if defined(DOXYGEN_DOC)
namespace cxxomfort { namespace cxxostd {

//! @internal 
struct implementation_defined_t {};

/**
 * @brief Applies the tuple @a t as the arguments to the functionlike / callable @a f .
 * @param f A callable / function-like argument.
 * @param t A @c std::tuple or similar construct.
 * @ingroup std0tuple
 * @cxx17b Invokes a callable @a f using the elements of the tuple @a t as arguments.
 * @returns Whatever <tt>f(t...)</tt> returns.
 * @see @cppref{utility/apply}
 * 
 * Example usage:
 * 
 * @code
 * auto logdata= make_tuple("Handle %x code %u with result %s", 0x10, 300, "Hello World");
 * invoke(printf, logdata);
 * // prints: 
 * // "Handle 0x10 code 300 with result Hello World"
 * @endcode
 * 
 * */
template <typename F, typename Tuple>
implementation_defined_t apply (F f, Tuple& t) { }

} }

#elif defined(__cpp_lib_apply)
	#undef  CXXOMFORT_IMPLEMENTS_n3829
	#define CXXOMFORT_IMPLEMENTS_n3829 CXXO_IMPLSTATUS_NATIVE()

#elif (0 && CXXOMFORT_CXX_STD == 2017 && CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG,>=600))

	// Clang is missing this when in C++17 mode
	// ref is https://en.cppreference.com/mwiki/index.php?title=cpp/utility/apply&oldid=95943 

    #undef  CXXOMFORT_IMPLEMENTS_n3829 
    #define CXXOMFORT_IMPLEMENTS_n3829 CXXO_IMPLSTATUS_BACKPORT()
	#include <cxxomfort/impl/17-invoke.hpp> // invoke(f, ...)

namespace cxxomfort { namespace cxxostd {

namespace impl {
template <class F, class Tuple, std::size_t... I>
constexpr decltype(auto) apply_impl(F&& f, Tuple&& t, std::index_sequence<I...>) {
    return std::invoke(std::forward<F>(f), std::get<I>(std::forward<Tuple>(t))...);
}

}  // impl

template <class F, class Tuple>
constexpr decltype(auto) apply(F&& f, Tuple&& t) {
    return impl::apply_impl(
        std::forward<F>(f), std::forward<Tuple>(t),
        std::make_index_sequence<std::tuple_size<std::decay_t<Tuple>>::value>{});
}
} }

// 
// full C++11 (variadic, trailing returns)
// 
#elif ((CXXOMFORT_CXX_STD >= 2011) || \
    (	CXXO_COMPILER_SUPPORT_decltype && CXXO_COMPILER_SUPPORT_rvref && \
		CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_trailing_returns )  \
		)

    #if (defined(CXXOMFORT_NOTICES))
        #if (CXXOMFORT_NOTICES > 1)
        #pragma message CXXO_NOTICE("enabled tuple apply(fn,tuple) variadic implementation")
        #endif
    #endif

    #include <cxxomfort/impl/14-integer_sequence.hpp>
    #include <cxxomfort/impl/17-invoke.hpp> // invoke(f, ...)

// http://stackoverflow.com/questions/687490/how-do-i-expand-a-tuple-into-variadic-template-functions-arguments/12650100#12650100

namespace cxxomfort { namespace cxxostd {

template<typename F, typename Tuple, size_t... I>
auto apply_(F&& f, Tuple&& args, std::index_sequence<I...>)
-> decltype(std::invoke( std::forward<F>(f) , std::get<I>(std::forward<Tuple>(args))...)){
    return std::invoke( std::forward<F>(f) , std::get<I>(std::forward<Tuple>(args))...);
}

/**
 * @brief Applies a tuple as an argument pack to a function.
 * @ingroup tuple
 * @return The return type from the function @a f .
 * 
 * Given the function @a f and the tuple @a args , 
 * @c apply  performs the function call 
 * <code>f (args[0], args[1], ...)</code>.
 * 
 */

template<typename F, typename Tuple,
typename Dec= typename std::decay<Tuple>::type, 
typename Indices = std::make_index_sequence<std::tuple_size<Dec>::value>>
auto apply(F&& f, Tuple&& args)
-> decltype(apply_(std::forward<F>(f), std::forward<Tuple>(args), Indices())) {
    return apply_(std::forward<F>(f), std::forward<Tuple>(args), Indices());
}

} }

    #undef  CXXOMFORT_IMPLEMENTS_n3829 
    #define CXXOMFORT_IMPLEMENTS_n3829 (CXXO_IMPLSTATUS_BACKPORT()+2)
    #define __cpp_lib_apply (CXXOMFORT_CXX_STD*100)

// 
// C++0x version with rvref and trailing returns, but no variadics
// 
#elif (CXXO_COMPILER_SUPPORT_decltype && CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_trailing_returns)
    #include <cxxomfort/impl/17-invoke.hpp> // invoke(f, ...)

namespace cxxomfort { namespace cxxostd {

template<typename F, typename Tuple>
auto apply_(F&& f, Tuple&& tup, std::integral_constant<size_t,1>)
-> decltype( std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) ) ) {
    return std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) );
}

template<typename F, typename Tuple>
auto apply_(F&& f, Tuple&& tup, std::integral_constant<size_t,2>)
-> decltype( std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) , std::get<1>(std::forward<Tuple>(tup)) )) {
    return std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) , std::get<1>(std::forward<Tuple>(tup)) );
}

template<typename F, typename Tuple>
auto apply_(F&& f, Tuple&& tup, std::integral_constant<size_t,3>)
-> decltype( std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) , std::get<1>(std::forward<Tuple>(tup)) , std::get<2>(std::forward<Tuple>(tup)) )) {
    return std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) , std::get<1>(std::forward<Tuple>(tup)) , std::get<2>(std::forward<Tuple>(tup)) );
}

template<typename F, typename Tuple>
auto apply_(F&& f, Tuple&& tup, std::integral_constant<size_t,4>)
-> decltype( std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) , std::get<1>(std::forward<Tuple>(tup)) , std::get<2>(std::forward<Tuple>(tup)) , std::get<3>(std::forward<Tuple>(tup)) )) {
    return std::invoke( std::forward<F>(f) , std::get<0>(std::forward<Tuple>(tup)) , std::get<1>(std::forward<Tuple>(tup)) , std::get<2>(std::forward<Tuple>(tup)) , std::get<3>(std::forward<Tuple>(tup)) );
}

template<typename F, typename Tuple>
auto apply(F&& f, Tuple&& args)
-> decltype( apply_(std::forward<F>(f), std::forward<Tuple>(args), std::integral_constant<size_t, std::tuple_size<typename std::decay<Tuple>::type>::value>() ) ) {
    return apply_(std::forward<F>(f), std::forward<Tuple>(args), std::integral_constant<size_t, std::tuple_size<typename std::decay<Tuple>::type>::value>() );
}

template<typename T, typename B, typename Tuple>
auto apply(T B::* mf, Tuple&& args)
-> decltype( apply_(std::mem_fn(mf), std::forward<Tuple>(args), std::integral_constant<size_t, std::tuple_size<typename std::decay<Tuple>::type>::value>() ) ) {
    return apply_(std::mem_fn(mf), std::forward<Tuple>(args), std::integral_constant<size_t, std::tuple_size<typename std::decay<Tuple>::type>::value>() );
}

} }

    #undef  CXXOMFORT_IMPLEMENTS_n3829 
    #define CXXOMFORT_IMPLEMENTS_n3829 (CXXO_IMPLSTATUS_BACKPORT()+1)
    #define __cpp_lib_apply (CXXOMFORT_CXX_STD*100)

#elif (CXXOMFORT_IMPLEMENTS_tuple > 0 && CXXOMFORT_CXX_STD >= 1997)
    #undef  CXXOMFORT_IMPLEMENTS_n3829 
    #define CXXOMFORT_IMPLEMENTS_n3829 CXXO_IMPLSTATUS_EMULATION()

//implementation for c++03 for up to 10 elements

namespace cxxomfort { namespace cxxostd {

template <typename Sig> struct FnReturnType { typedef typename Sig::result_type type; };
template <typename R, typename A0> struct FnReturnType<R(A0)> {
    typedef R type;
};
template <typename R, typename A0> struct FnReturnType<R(*)(A0)> {
    typedef R type;
};
template <typename R, typename A0, typename A1> struct FnReturnType<R(A0,A1)> 
: FnReturnType<R(A0)> {};
template <typename R, typename A0, typename A1> struct FnReturnType<R(*)(A0,A1)> 
: FnReturnType<R(A0)> {};

template <typename R, typename A0, typename A1, typename A2> struct FnReturnType<R(A0,A1,A2)> 
: FnReturnType<R(A0)> {};
template <typename R, typename A0, typename A1, typename A2> struct FnReturnType<R(*)(A0,A1,A2)> 
: FnReturnType<R(A0)> {};

template <typename R, typename A0, typename A1, typename A2
    , typename A3> struct FnReturnType<R(A0,A1,A2,A3)> 
: FnReturnType<R(A0)> {};
template <typename R, typename A0, typename A1, typename A2
    , typename A3> struct FnReturnType<R(*)(A0,A1,A2,A3)> 
: FnReturnType<R(A0)> {};

template <typename R, typename A0, typename A1, typename A2
    , typename A3, typename A4> 
struct FnReturnType<R(A0,A1,A2,A3,A4)> 
: FnReturnType<R(A0)> {};
template <typename R, typename A0, typename A1, typename A2
    , typename A3, typename A4> 
struct FnReturnType<R(*)(A0,A1,A2,A3,A4)> 
: FnReturnType<R(A0)> {};

template <typename R, typename A0, typename A1, typename A2
    , typename A3, typename A4, typename A5> 
struct FnReturnType<R(A0,A1,A2,A3,A4,A5)> 
: FnReturnType<R(A0)> {};
template <typename R, typename A0, typename A1, typename A2
    , typename A3, typename A4, typename A5> 
struct FnReturnType<R(*)(A0,A1,A2,A3,A4,A5)> 
: FnReturnType<R(A0)> {};


template <typename R, typename A0, typename A1, typename A2
    , typename A3, typename A4, typename A5> 
struct FnReturnType<R(A0,A1,A2,A3,A4,A5,...)> 
: FnReturnType<R(A0)> {};
template <typename R, typename A0, typename A1, typename A2
    , typename A3, typename A4, typename A5> 
struct FnReturnType<R(*)(A0,A1,A2,A3,A4,A5,...)> 
: FnReturnType<R(A0)> {};


template <typename F, typename Tuple, unsigned int S= std::tuple_size< typename std::remove_cv<Tuple>::type>::value> 
struct call_impl {
};

	#define GET_(x) get<x>(t)

template <typename F, typename Tuple> struct call_impl<F, Tuple,1> {
    static void call (F& f, Tuple& t) {
        using std::get;
        f(GET_(0));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,2> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,3> {
    template <typename R> static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,4> {
    template <typename R> static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,5> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3), GET_(4));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,6> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3), GET_(4), GET_(5));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,7> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3), GET_(4), GET_(5), GET_(6));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,8> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3), GET_(4), GET_(5), GET_(6), GET_(7));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,9> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3), GET_(4), GET_(5), GET_(6), GET_(7), GET_(8));
    }
};

template <typename F, typename Tuple> struct call_impl<F,Tuple,10> {
    template <typename R>
    static R call (F& f, Tuple& t) {
        using std::get;
        return f(GET_(0), GET_(1), GET_(2), GET_(3), GET_(4), GET_(5), GET_(6), GET_(7), GET_(8), GET_(9));
    }
};

	#undef GET_

template <typename F, typename Tuple>
typename FnReturnType<F>::type
apply (F f, Tuple& t) {
    //typedef typename FnReturnType< typename std::remove_pointer<F>::type >::type ret_type;
    typedef typename FnReturnType<F>::type R;
    return call_impl<F,Tuple>::template call<R>( f, t );
}

template <typename F, typename Tuple>
typename FnReturnType<F>::type
apply (F f, Tuple const& t) {
    typedef typename FnReturnType<F>::type R;
    return call_impl<F,Tuple const>::template call<R>( f, t );
}

} } // cxxomfort::cxxostd

#endif


#if (CXXOMFORT_IMPLEMENTS_n3829 > 0)
	#undef  CXXOMFORT_IMPLEMENTS_tuple_apply
	#define CXXOMFORT_IMPLEMENTS_tuple_apply  CXXOMFORT_IMPLEMENTS_n3829
#endif

#if (defined(CXXOMFORT_NO_STD_USING))
#elif ((CXXOMFORT_IMPLEMENTS_tuple_apply > 0) && CXXOMFORT_IMPLEMENTS_tuple_apply < CXXO_IMPLSTATUS_NATIVE())
	//#pragma message CXXO_NOTICE("1")
namespace std {
    using cxxomfort::cxxostd::apply;
}
#endif

#endif
