#ifndef CXXOMFORT_IMPL_BINDFRONT_HPP
#define CXXOMFORT_IMPL_BINDFRONT_HPP
/**
 * @file
 */
#include <cxxomfort/base.hpp>
#include <functional>

#define CXXOMFORT_IMPLEMENTS_p0356 0

#if (defined(__cpp_lib_bind_front))
	#undef  CXXO_COMPILER_SUPPORT_bind_front
	#define CXXO_COMPILER_SUPPORT_bind_front 1
#endif

#if defined(DOXYGEN_DOC)
namespace cxxomfort { namespace cxxostd {

/**
 * @brief Wraps a function and its first @a n arguments in a callable object.
 * @param args A sequence of arguments that are bound to the callable.
 * @ingroup std0functional
 * @cxx20b Wraps a function and its first @e n arguments as a callable.
 * 
 * @see @cppref{utility/functional/bind_front}
 * 
 * Given a callable object or function @a fn taking @e n arguments and a sequence @a args of at most @e n arguments,  @c bind_front(fn,args...) returns a callable object @a wf of undetermined type that wraps @a fn and @a args such as at a later point of use the invocation <tt>wf(call_args...)</tt>, @a fn is invoked via @c invoke() with the lists of arguments @a args and @a call_args in sequence.
 * 
 * In other words, @c bind_front() functions exactly like @c bind() except that it only binds the first few arguments.
 * 
 * @code
 * auto wf = bind_front(fn, a1, a2, a3);
 * wf(b1, b2); // calls fn(a1, a2, a3, b1, b2);
 * @endcode
 * 
 * @note This feature from C++20 is backported down to at most C++11.
 * 
 * 
 * */
template <typename Fn, typename... Binds>
implementation_defined_t bind_front (Fn&& callable, Binds&&... args) {
}

} }
#elif (CXXO_COMPILER_SUPPORT_bind_front>0)

    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 CXXO_IMPLSTATUS_NATIVE()

// 
// C++11 support +  trailing return or decltype(auto)
// this severely eases forwarding and return type
// 
#elif (  \
( (CXXO_COMPILER_SUPPORT_decltype>0) && CXXO_COMPILER_SUPPORT_decltypeauto && CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_trailing_returns) \
)
    /*
     * MSVC excluded due to a "operator decltype" bug: 
     * https://developercommunity.visualstudio.com/content/problem/295931/error-c2833-operator-decltype-is-not-a-recognized.html 
     * */
    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 (CXXO_IMPLSTATUS_BACKPORT()+4)
    #include <cxxomfort/impl/17-tuple_apply.hpp>
    #include <cxxomfort/utility.hpp> // forward

namespace cxxomfort { namespace cxxostd {
namespace impl {

template <typename F, typename... Binds>
struct bf {
    typedef typename std::decay<F>::type DF;
    F f;
    std::tuple< typename std::decay<Binds>::type...> tu;

    bf (F&& f, Binds&&... args)
    : f(std::forward<F>(f)), tu(std::forward<Binds>(args)...)
    {}

    template <typename... Args>  auto   
    operator() (Args&&... args) 
    -> decltype(std::apply(f, /* xtu */ std::tuple_cat(tu, std::forward_as_tuple(args...) ) ) )
    {
        using namespace std;
        auto xtu = std::tuple_cat(tu, std::forward_as_tuple(args...));
        return ::std::apply(f, xtu );
    } 
};

} // impl

template <typename F, typename... Args> inline 
impl::bf<F,Args...> bind_front (F&& f, Args&&... args) {
    return impl::bf<F,Args...>(std::forward<F>(f), std::forward<Args>(args)...);
}

} } // cxxomfort::cxxostd


// 
// MSVC 2010-2013 crash with a decltype() implementation
// 
#elif (0 && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600) && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,<1800))

    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 0

// 
// partial C++11, or C++0x with at least decltype and rvref
// 
#elif (CXXO_COMPILER_SUPPORT_decltype && CXXO_COMPILER_SUPPORT_rvref)

    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 (CXXO_IMPLSTATUS_BACKPORT()+2)
    #include <cxxomfort/utility.hpp> // forward
    #include <cxxomfort/tuple.hpp>
    #include <cxxomfort/impl/17-tuple_apply.hpp>

namespace cxxomfort { namespace cxxostd {
namespace impl {

template <typename A1>
std::tuple<A1&&> fwd_as_tuple_impl( A1&& a1 ) {
    return std::tuple<A1&&>( std::forward<A1>(a1));
}

template <typename A1, typename A2>
std::tuple<A1&&,A2&&> fwd_as_tuple_impl( A1&& a1, A2&& a2 ) {
    return std::tuple<A1&&>( std::forward<A1>(a1), std::forward<A2>(a2));
}


template <typename F, typename Tuple>
struct bf;

template <typename F, typename T1>
struct bf<F, std::tuple<T1> > {
    typedef typename std::decay<F>::type DF;
    typedef std::tuple<T1> Tuple;
    DF f;
    Tuple tu;

    bf (F&& f, Tuple&& tt)
    : f(std::forward<F>(f)), tu(std::move(tt)) /*tu(std::forward<Binds>(args)...)*/
    {}

    template <typename F2, typename B1> 
    struct Res2 {
        typedef std::tuple<T1,B1> P;
        typedef decltype(std::apply(std::declval<F2>(),std::declval<P>())) type;
    };

    template <typename B1> 
    typename Res2<F,B1>::type
    operator() (B1&& b1) {
        return std::apply(f, std::tuple_cat(tu, std::make_tuple(b1) ) );
    }

    template <typename B1, typename B2> 
    decltype(std::apply(f, std::tuple_cat(tu, std::forward_as_tuple( std::declval<B1>(), std::declval<B2>() )) ) )
    operator() (B1&& b1, B2&& b2) {
        auto r= std::tuple_cat(tu, std::tuple<B1&&,B2&&>( std::forward<B1>(b1), std::forward<B2>(b2) ) );
        return std::apply(f,r);
    }

};

template <typename F, typename T1, typename T2>
struct bf<F, std::tuple<T1,T2> > {
    typedef typename std::decay<F>::type DF;
    typedef std::tuple<T1,T2> Tuple;
    DF f;
    Tuple tu;

    bf (F&& f, Tuple&& tt)
    : f(std::forward<F>(f)), tu(std::move(tt)) /*tu(std::forward<Binds>(args)...)*/
    {}

    template <typename P1, typename P2, typename B1> 
    struct Res2 {
        typedef std::tuple<P1,P2,B1> P;
        typedef decltype(std::apply(f,std::declval<P>())) type;
    };

    template <typename B1> auto
    operator() (B1&& b1) 
    -> decltype(std::apply(f, std::tuple_cat(tu, fwd_as_tuple_impl(b1) ) ))
    {
        return std::apply(f, std::tuple_cat(tu, fwd_as_tuple_impl(b1) ) );
    }

    template <typename B1, typename B2> auto
    operator() (B1&& b1, B2&& b2) 
    -> decltype(std::apply(f, std::tuple_cat(tu, fwd_as_tuple_impl(b1,b2) ) ))
    {
        return std::apply(f, std::tuple_cat(tu, fwd_as_tuple_impl(b1,b2) ) );
    }

};


} // impl

template <typename F, typename A1>
inline impl::bf<F, std::tuple<A1> > 
bind_front (F&& f, A1&& arg1) {
    return impl::bf<F, std::tuple<A1> >(std::forward<F>(f), std::tuple<A1&&>(std::forward<A1>(arg1)) );
}

template <typename T, typename B, typename A1>
inline auto 
bind_front (T B::* mf, A1&& arg1) 
-> decltype( bind_front(std::mem_fn(mf), std::forward<A1>(arg1) ) ) {
    return bind_front(std::mem_fn(mf), std::forward<A1>(arg1) );
}

template <typename F, typename A1, typename A2>
inline impl::bf<F, std::tuple<A1,A2> > 
bind_front (F&& f, A1&& arg1, A2&& arg2) {
    return impl::bf<F, std::tuple<A1,A2> >(std::forward<F>(f), std::tuple<A1&&,A2&&>(std::forward<A1>(arg1),std::forward<A2>(arg2) ) );
}

template <typename T, typename B, typename A1, typename A2>
inline auto 
bind_front (T B::* mf, A1&& arg1, A2&& arg2) 
-> decltype( bind_front(std::mem_fn(mf), std::forward<A1>(arg1), std::forward<A2>(arg2)) ) {
    return bind_front(std::mem_fn(mf), std::forward<A1>(arg1), std::forward<A2>(arg2));
}

} } // cxxomfort::cxxostd



// 
// MSVC 2017-2019 with native invoke, but with the "operator decltype" bug
// 
#elif (0 && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1910) && CXXOMFORT_CXX_EMULATION >= 2017)

// 
// C++03 or later with typeof, plus cxxomfort support for invoke
// (this version takes the functor by value)
// 
#elif (CXXOMFORT_IMPLEMENTS_tuple >= CXXO_IMPLSTATUS_BACKPORT() && CXXOMFORT_IMPLEMENTS_invoke >= CXXO_IMPLSTATUS_EMULATION() && CXXOMFORT_IMPLEMENTS_tuple_apply > 0 && CXXO_COMPILER_SUPPORT_typeof > 0 ) // TR1 support for result_of
    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 (CXXO_IMPLSTATUS_EMULATION()+3)
    
    #include <cxxomfort/type_traits.hpp>
	#include "./17-invoke.hpp"
    //#pragma message ("cool2")

namespace cxxomfort { namespace cxxostd {

namespace impl {

template <typename F, typename Wrap>
struct pbf_proxy {
	int value;
};

template <typename F, typename T1>
struct pbf_proxy<F, std::tuple< T1> > {
	typedef std::tuple< typename std::decay<T1>::type > TWrap;
	typedef typename std::decay<F>::type DF;
	DF f;
	TWrap b;

	explicit pbf_proxy (F& fn, TWrap const& bs)
	: f(fn), b(bs) {}

	template <class TW, class Arg1> struct Par1 {
		typedef std::tuple<Arg1> TA;
		typedef CXXO_TYPEOF( std::tuple_cat( std::declval<TW>(), std::declval< TA >() ) )
		type;
		//typedef CXXO_TYPEOF( std::invoke(std::declval<F>(),t1,std::declval<A1>() ) ) type;
	};

	template <typename A1>
	CXXO_TYPEOF( std::apply(std::declval<F>(), std::declval<typename Par1<TWrap,A1>::type>()) )
	operator() (A1 a1) const {
		return std::apply( f, std::tuple_cat(b, std::make_tuple(a1) ) );
	}

	template <class TW, typename A1, typename A2> struct Par2 {
		typedef CXXO_TYPEOF( std::tuple_cat( std::declval<TW>(), std::declval< std::tuple<A1,A2> >() ) )
		type;
	};

	template <typename A1, typename A2>
	CXXO_TYPEOF( std::apply(std::declval<F>(), std::declval<typename Par2<TWrap,A1,A2>::type>()) )
	operator() (A1 a1, A2 a2) const {
		return std::apply( f, std::tuple_cat(b, std::make_tuple(a1,a2) ) );
	}

	template <class TW, typename A1, typename A2, typename A3> struct Par3 {
		typedef CXXO_TYPEOF( std::tuple_cat( std::declval<TW>(), std::declval< std::tuple<A1,A2,A3> >() ) )
		type;
	};

	template <typename A1, typename A2, typename A3>
	CXXO_TYPEOF( std::apply(std::declval<F>(), std::declval<typename Par3<TWrap,A1,A2,A3>::type>()) )
	operator() (A1 a1, A2 a2, A3 a3) const {
		return std::apply( f, std::tuple_cat(b, std::make_tuple(a1,a2,a3) ) );
	}

};


template <typename F, typename T1, typename T2>
struct pbf_proxy<F, void(T1,T2)> {
	typedef typename ::std::decay<F>::type DF;
	DF f;
	T1 const t1;
	T2 const t2;

	template <typename A1> struct Res3 {
		typedef CXXO_TYPEOF( std::invoke(std::declval<F>(),std::declval<T1>(),std::declval<T2>(),std::declval<A1>()) ) type;
	};

	template <typename A1, typename A2> struct Res4 {
		typedef CXXO_TYPEOF( std::invoke(std::declval<F>(),std::declval<T1>(),std::declval<T2>(),std::declval<A1>(),std::declval<A2>()) ) type;
	};

	template <typename A1, typename A2, typename A3> struct Res5 {
		typedef CXXO_TYPEOF( std::invoke(std::declval<F>(),std::declval<T1>(),std::declval<T2>(),std::declval<A1>(),std::declval<A2>(),std::declval<A3>() ) ) type;
	};

	explicit pbf_proxy (F& fn, T1 const& v1, T2 const& v2)
	: f(fn), t1(v1), t2(v2) {}


	template <typename A1>
	/// typename result_of_invoke3<F,T1,T2,A1>::result_type 
	typename Res3<A1>::type  // CXXO_TYPEOF( std::invoke(f,t1,t2,std::declval<A1>()) )
	operator() (A1 a1) const {
		return invoke(f,t1,t2,a1);
	}
	template <typename A1, typename A2>
	typename Res4<A1,A2>::type // CXXO_TYPEOF( std::invoke(f,t1,t2,std::declval<A1>(),std::declval<A2>()) )
	operator() (A1 a1, A2 a2) const {
		return invoke(f,t1,t2,a1,a2);
	}
	template <typename A1, typename A2, typename A3>
	typename Res5<A1,A2,A3>::type // CXXO_TYPEOF( std::invoke(f,t1,t2,std::declval<A1>(),std::declval<A2>(),std::declval<A3>()) )
	operator() (A1 a1, A2 a2, A3 a3) const {
		return invoke(f,t1,t2,a1,a2,a3);
	}
};

} // impl

template <typename F, typename T1>
inline 
impl::pbf_proxy<F, std::tuple<T1> >
bind_front (F f, T1 const& t1) {
	return impl::pbf_proxy<F, std::tuple<T1> >(f, std::make_tuple(t1) );
}

/*
template <typename F, typename T1, typename T2>
inline 
impl::pbf_proxy<F, std::tuple<T1,T2> >
bind_front (F f, T1 const& t1, T2 const& t2) {
	return impl::pbf_proxy<F>(f, std::make_tuple(t1,t2) );
}

template <typename F, typename T1, typename T2, typename T3>
inline 
impl::pbf_proxy<F, std::tuple<T1,T2,T3> >
bind_front (F f, T1 const& t1, T2 const& t2, T3 const& t3) {
	return impl::pbf_proxy<F>(f, std::make_tuple(t1,t2,t3) );
}
*/

} } 


// 
// C++03 with TR1 or compatible 
// (note this version takes the functor by value)
// 

#elif (CXXOMFORT_CXX_STD >= 1997 ) // TR1 support for result_of

    #include <cxxomfort/impl/11-decay.hpp>

namespace cxxomfort { namespace cxxostd {

namespace impl {

template <typename F, typename Sig> 
struct pbf_proxy {
	int value;
};

template <typename F, typename T1, typename T2, typename T3>
struct pbf_proxy<F, void(T1,T2,T3)> {
	typedef typename std::decay<F>::type DF;
	DF f;
	T1 t1;
	T2 t2;
	T3 t3;
	pbf_proxy (F& f_, T1 const& v1, T2 const& v2, T3 const& v3)
	: f(f_), t1(v1), t2(v2), t3(v3) {}

	template <typename A1>
	typename std::tr1::result_of<DF(T1,T2,T3,A1)>
	::type operator() (A1 a1) const {
		return f(t1,t2,t3,a1);
	}
	template <typename A1, typename A2>
	typename std::tr1::result_of<DF(T1,T2,T3,A1,A2)>
	::type operator() (A1 a1, A2 a2) const {
		return f(t1,t2,t3,a1,a2);
	}
	template <typename A1, typename A2, typename A3>
	typename std::tr1::result_of<DF(T1,T2,T3,A1,A2,A3)>
	::type operator() (A1 a1, A2 a2, A3 a3) const {
		return f(t1,t2,t3,a1,a2,a3);
	}

};


template <typename F, typename T1, typename T2>
struct pbf_proxy<F, void(T1,T2)> {
	typedef typename std::decay<F>::type DF;
	DF f;
	T1 const& t1;
	T2 const& t2;
	pbf_proxy (F& f_, T1 const& v1, T2 const& v2)
	: f(f_), t1(v1), t2(v2) {}

	template <typename A1>
	typename std::tr1::result_of<DF(T1,T2,A1)>
	::type operator() (A1 a1) const {
		return f(t1,t2,a1);
	}
	template <typename A1, typename A2>
	typename std::tr1::result_of<DF(T1,T2,A1,A2)>
	::type operator() (A1 a1, A2 a2) const {
		return f(t1,t2,a1,a2);
	}
	template <typename A1, typename A2, typename A3>
	typename std::tr1::result_of<DF(T1,T2,A1,A2,A3)>
	::type operator() (A1 a1, A2 a2, A3 a3) const {
		return f(t1,t2,a1,a2,a3);
	}
	template <typename A1, typename A2, typename A3, typename A4>
	typename std::tr1::result_of<DF(T1,T2,A1,A2,A3,A4)>
	::type operator() (A1 a1, A2 a2, A3 a3, A4 a4) const {
		return f(t1,t2,a1,a2,a3,a4);
	}

};

template <typename F, typename T1>
struct pbf_proxy<F, void(T1)> {
	typedef typename std::decay<F>::type DF;
	DF f;
	T1 t1;
	pbf_proxy (F& f_, T1 const& v1)
	: f(f_), t1(v1) {}
	
	
	template <typename A1>
	typename std::tr1::result_of<DF(T1,A1)>
	::type operator() (A1 a1) const {
		return f(t1,a1);
	}
	template <typename A1, typename A2>
	typename std::tr1::result_of<DF(T1,A1,A2)>
	::type operator() (A1 a1, A2 a2) const {
		return f(t1,a1,a2);
	}
	template <typename A1, typename A2, typename A3>
	typename std::tr1::result_of<DF(T1,A1,A2,A3)>
	::type operator() (A1 a1, A2 a2, A3 a3) const {
		return f(t1,a1,a2,a3);
	}
	template <typename A1, typename A2, typename A3, typename A4>
	typename std::tr1::result_of<DF(T1,A1,A2,A3,A4)>
	::type operator() (A1 a1, A2 a2, A3 a3, A4 a4) const {
		return f(t1,a1,a2,a3,a4);
	}
	template <typename A1, typename A2, typename A3, typename A4, typename A5>
	typename std::tr1::result_of<DF(T1,A1,A2,A3,A4,A5)>
	::type operator() (A1 a1, A2 a2, A3 a3, A4 a4, A5 a5) const {
		return f(t1,a1,a2,a3,a4,a5);
	}
	template <typename A1, typename A2, typename A3, typename A4, typename A5, typename A6>
	typename std::tr1::result_of<DF(T1,A1,A2,A3,A4,A5,A6)>
	::type operator() (A1 a1, A2 a2, A3 a3, A4 a4, A5 a5, A6 a6) const {
		return f(t1,a1,a2,a3,a4,a5,a6);
	}

};

} // impl

template <typename F, typename T1, typename T2, typename T3>
inline 
impl::pbf_proxy<F,void(T1,T2,T3)>
bind_front (F& f, T1 const& t1, T2 const& t2, T3 const& t3) {
	return impl::pbf_proxy<F,void(T1,T2,T3)>(f,t1,t2,t3);
}

template <typename F, typename T1, typename T2, typename T3>
inline 
impl::pbf_proxy<F const,void(T1,T2,T3)>
bind_front (F const& f, T1 const& t1, T2 const& t2, T3 const& t3) {
	return impl::pbf_proxy<F const,void(T1,T2,T3)>(f,t1,t2,t3);
}

/*
template <typename F, typename T1, typename T2>
inline 
impl::pbf_proxy<F,void(T1,T2)>
bind_front (F& f, T1 const& t1, T2 const& t2) {
	return impl::pbf_proxy<F,void(T1,T2)>(f,t1,t2);
}
*/

template <typename F, typename T1, typename T2>
inline 
impl::pbf_proxy<F,void(T1,T2)>
bind_front (F f, T1 const& t1, T2 const& t2) {
	return impl::pbf_proxy<F,void(T1,T2)>(f,t1,t2);
}


template <typename F, typename T1>
inline 
impl::pbf_proxy<F,void(T1)>
bind_front (F& f, T1 const& t1) {
	return impl::pbf_proxy<F,void(T1)>(f,t1);
}


}} // cxxomfort::cxxostd

    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 CXXO_IMPLSTATUS_EMULATION()

#else // pre-TR1; not implemented

    #undef  CXXOMFORT_IMPLEMENTS_p0356
    #define CXXOMFORT_IMPLEMENTS_p0356 0

#endif

#undef  CXXOMFORT_IMPLEMENTS_bind_front
#define CXXOMFORT_IMPLEMENTS_bind_front CXXOMFORT_IMPLEMENTS_p0356

#if (defined(CXXOMFORT_NO_STD_USING))
#elif (CXXOMFORT_IMPLEMENTS_bind_front > 0 && CXXOMFORT_IMPLEMENTS_bind_front < CXXO_IMPLSTATUS_NATIVE())
namespace std {
    using ::cxxomfort::cxxostd::bind_front;
}
#endif

#endif
