#ifndef CXXOMFORT_IMPL_20_ENDIAN_HPP
#define CXXOMFORT_IMPL_20_ENDIAN_HPP
/**
 * @file 
 * @brief Implementation of std::endian
 * */
#include <cxxomfort/config.hpp>

#if (defined(__cpp_lib_endian) && CXXO_COMPILER_SUPPORT_endian==0)
	#undef  CXXO_COMPILER_SUPPORT_endian
	#define CXXO_COMPILER_SUPPORT_endian 1
#endif

namespace cxxomfort { namespace cxxostd {
namespace detail {

	union endian_tester {
		uint32_t   n;
		uint8_t    p[4];
	};

	const endian_tester sample = {0x01020304}; // this initializes .n

	struct endian_finally {
		enum { 
			little = 0x00000001, 
			big    = 0x01000000,
			other  = 0x00000000 
		};
	};


} // detail
} } 

#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC)
	#include <endian.h>
#endif

// 
// values that describe endianness
// 

#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC && defined(_WIN32))
    #define CXXO_endian_members() little = 0,  big = 1,  native = little

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_GCC, < 406))
    #define CXXO_endian_members() little = __LITTLE_ENDIAN, big = __BIG_ENDIAN, native = __BYTE_ORDER

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_GCC, >= 406))
    #define CXXO_endian_members()    little = __ORDER_LITTLE_ENDIAN__, big    = __ORDER_BIG_ENDIAN__,  native = __BYTE_ORDER__

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_DMC, >= 0x9999))
	#define CXXO_BIG_ENDIAN     0x00be
	#define CXXO_LITTLE_ENDIAN  0x001e
	#if defined(_STLP_BIG_ENDIAN)
		#define CXXO_NATIVE_ENDIAN CXXO_BIG_ENDIAN
	#elif defined(_STLP_LITTLE_ENDIAN)
		#define CXXO_NATIVE_ENDIAN CXXO_LITTLE_ENDIAN
	#else
		#define CXXO_NATIVE_ENDIAN 0xffff
	#endif
    #define CXXO_endian_members()    little = CXXO_LITTLE_ENDIAN, big    = CXXO_BIG_ENDIAN,  native = CXXO_NATIVE_ENDIAN

#else
    #define CXXO_endian_members()  little = detail::endian_finally::little, big = detail::endian_finally::big , native = 0xffffffffUL & 0x1UL
#endif

// 
// endian
// 

#if defined(__cpp_lib_endian)
#elif (CXXO_COMPILER_SUPPORT_cstd_endian > 0 || CXXO_COMPILER_SUPPORT_endian > 0)
	#define __cpp_lib_endian  __cplusplus
	// nothing to do
	#define CXXOMFORT_IMPLEMENTS_p0463r1 CXXO_IMPLSTATUS_NATIVE()
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("endian is native")
	#endif

namespace cxxomfort { namespace cxxostd {
	using std::endian;
} }

#elif (CXXOMFORT_CXX_STD >= 2011)
	#define CXXOMFORT_IMPLEMENTS_p0463r1 CXXO_IMPLSTATUS_BACKPORT()
	#define CXXOMFORT_USING_endian 1
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("endian backport")
	#endif
	#define __cpp_lib_endian (CXXOMFORT_CXX_STD*100)
	
namespace cxxomfort { namespace cxxostd {
	enum class endian { CXXO_endian_members() };
} }

#elif (CXXOMFORT_CXX_STD >= 1997 && CXXO_COMPILER_SUPPORT_type_traits_intrinsic)
	#define CXXOMFORT_USING_endian
	#define CXXOMFORT_IMPLEMENTS_p0463r1 CXXO_IMPLSTATUS_EMULATION()
	#define CXXOMFORT_USING_endian 1
	#include <cxxomfort/impl/typesafe_enum-core.hpp>
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("endian emulation via typesafe-enum")
	#endif
	#define __cpp_lib_endian 1

namespace cxxomfort { namespace cxxostd {

struct myendian {
	enum type { CXXO_endian_members() };
};

typedef cxxomfort::impl::typesafe_enum<myendian> endian;

} }


#else
	#define CXXOMFORT_IMPLEMENTS_p0463r1 CXXO_IMPLSTATUS_EMULATION()
	#define CXXOMFORT_USING_endian 1
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("endian emulation via enum")
	#endif
	#define __cpp_lib_endian 1

namespace cxxomfort { namespace cxxostd {

struct myendian {
	enum type { CXXO_endian_members() };
};

typedef myendian endian;

} }

#endif

#undef  CXXOMFORT_IMPLEMENTS_endian
#define CXXOMFORT_IMPLEMENTS_endian CXXOMFORT_IMPLEMENTS_p0463r1


#if defined(CXXOMFORT_USING_endian)
	#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
	#else
namespace std { 
	//! Using in namespace std;
	using cxxomfort::cxxostd::endian;
}
	#endif // no std using

#endif


#ifdef DOXYGEN_DOC

namespace cxxomfort { namespace std { 

/**
 * @brief Endianness enumeration for C++.
 * @ingroup std0bit
 * @cxx20b Endianness
 * 
 * @c endian is a typed enumeration that describes the potential values 
 * for a system's endianness. The enumeration has three values: @c little , 
 * @c big and @c native . @c native will equal one of the other two 
 * when the system has that endianness, or will equal none of them 
 * in a system with a different (non-little, non-big) endianness.
 * 
 * The macro <tt>__cpp_lib_endian</tt> is defined when this feature is backported.
 * 
 * @see https://en.cppreference.com/w/cpp/types/endian
 * 
 * */
enum class endian {};
} }

#endif

#endif
