#ifndef CXXOMFORT_IMPL_20_STRING_HELPERS_HPP
#define CXXOMFORT_IMPL_20_STRING_HELPERS_HPP
/**
 * @file
 * @brief Implementation of string helpers from C++20.
 * 
 * This header adds global functions to access the functionality of <tt>.starts_with()</tt> and <tt>.ends_with</tt> for string types.
 * 
 */
#include <cxxomfort/base.hpp>
#include <cxxomfort/type_traits.hpp>
#include <cxxomfort/string.hpp>
#include <cxxomfort/string_view.hpp>

namespace cxxomfort { namespace string {

    #if defined(__cpp_lib_starts_ends_with)

template <typename Ch, typename CT, typename A, typename Arg>
constexpr bool 
starts_with (std::basic_string<Ch,CT,A> const& str, Arg&& a) noexcept {
    return str.starts_with(std::forward<Arg>(a));
}

template <typename Ch, typename CT, typename A, typename Arg>
constexpr bool 
ends_with (std::basic_string<Ch,CT,A> const& str, Arg&& a) noexcept {
    return str.ends_with(std::forward<Arg>(a));
}

template <typename Ch, typename CT, typename Arg>
constexpr bool 
starts_with (std::basic_string_view<Ch,CT> const str, Arg&& a) noexcept {
    return str.starts_with(std::forward<Arg>(a));
}

template <typename Ch, typename CT, typename Arg>
constexpr bool 
ends_with (std::basic_string_view<Ch,CT> const str, Arg&& a) noexcept {
    return str.ends_with(std::forward<Arg>(a));
}

    #elif (CXXOMFORT_CXX_STD >= 2014)

template <typename Ch, typename CT>
constexpr bool 
starts_with (std::basic_string_view<Ch,CT> const str, Ch const a) noexcept {
    return str.find(a) == 0;
}

template <typename Ch, typename CT>
constexpr bool 
ends_with (std::basic_string_view<Ch,CT> const str, Ch const a) noexcept {
    return !(str.empty()) && str.find(a) == str.size()-1;
}

template <typename Ch, typename CT>
constexpr bool 
starts_with (std::basic_string_view<Ch,CT> const str, std::basic_string_view<Ch,CT> const arg) noexcept {
    return str.find(arg.data(), 0, arg.size()) == 0;
}

template <typename Ch, typename CT>
constexpr bool 
ends_with (std::basic_string_view<Ch,CT> const str, std::basic_string_view<Ch,CT> const arg) noexcept {
    return (arg.size() <= str.size()) && str.find(arg.data(), 0, arg.size()) == str.size()-arg.size();
}

template <typename Ch, typename CT, typename CA, typename Arg>
constexpr bool 
starts_with (std::basic_string<Ch,CT,CA> const& str, Arg&& a) {
	return starts_with(std::basic_string_view<Ch,CT>(str) , std::forward<Arg>(a));
}

template <typename Ch, typename CT, typename CA, typename Arg>
constexpr bool 
ends_with (std::basic_string<Ch,CT,CA> const& str, Arg&& a) {
	return ends_with(std::basic_string_view<Ch,CT>(str) , std::forward<Arg>(a));
}


/*
template <typename Ch, typename T, typename A>
constexpr bool 
starts_with (std::basic_string<Ch,T,A> const& str, std::basic_string_view<Ch,T> const a) noexcept {
    // anything convertible to basic_string_view can use this one
    return str.find(a.data(), a.size()) == 0;
}
*/

    #else // 1997-2011

// char

template <typename Ch, typename CT>
bool 
starts_with (std::basic_string_view<Ch,CT> const str, Ch const a) CXXO_NOEXCEPT {
    return str.find(a) == 0;
}

template <typename Ch, typename CT>
bool 
ends_with (std::basic_string_view<Ch,CT> const str, Ch const a) CXXO_NOEXCEPT {
    return !(str.empty()) && str[str.size()-1]==a;
}

// string_view 

template <typename Ch, typename CT>
bool 
starts_with (std::basic_string_view<Ch,CT> const str, std::basic_string_view<Ch,CT> const arg) CXXO_NOEXCEPT {
    return str.find(arg.data(), 0, arg.size()) == 0;
}

template <typename Ch, typename CT>
bool 
ends_with (std::basic_string_view<Ch,CT> const str, std::basic_string_view<Ch,CT> const arg) CXXO_NOEXCEPT {
    return (arg.size() <= str.size()) && str.find(arg.data(), 0, arg.size()) == str.size()-arg.size();
}

// this grabs a right-hand argument that is convertible to basic_string_view.

template <typename Ch, typename CT, typename Arg>
bool 
starts_with (std::basic_string_view<Ch,CT> const str, Arg const& arg) CXXO_NOEXCEPT {
	return starts_with(str, std::basic_string_view<Ch,CT>(arg));
}

template <typename Ch, typename CT, typename Arg>
bool 
ends_with (std::basic_string_view<Ch,CT> const str, Arg const& arg) CXXO_NOEXCEPT {
	return ends_with(str, std::basic_string_view<Ch,CT>(arg));
}


		#if (CXXOMFORT_CXX_STD>=2011)
template <typename Ch, typename CT, typename CA, typename Arg>
constexpr bool 
starts_with (std::basic_string<Ch,CT,CA> const& str, Arg&& a) {
	using namespace std;
	return starts_with(std::basic_string_view<Ch,CT>(str) , std::forward<Arg>(a));
}

template <typename Ch, typename CT, typename CA, typename Arg>
constexpr bool 
ends_with (std::basic_string<Ch,CT,CA> const& str, Arg&& a) {
	return ends_with(std::basic_string_view<Ch,CT>(str) , std::forward<Arg>(a));
}

		#else
template <typename Ch, typename CT, typename CA, typename Arg>
bool 
starts_with (std::basic_string<Ch,CT,CA> const& str, Arg const& a) {
	return starts_with(std::basic_string_view<Ch,CT>(str) , a );
}

template <typename Ch, typename CT, typename CA, typename Arg>
bool 
ends_with (std::basic_string<Ch,CT,CA> const& str, Arg const& a) {
	return ends_with(std::basic_string_view<Ch,CT>(str) , a );
}
		#endif


/*
template <typename Ch, typename T, typename A>
bool 
starts_with (std::basic_string<Ch,T,A> const& str, std::basic_string_view<Ch,T> const a) CXXO_NOEXCEPTNOTHROW {
    // anything convertible to basic_string_view can use this one
    return str.find(a.data(), 0, a.size()) == 0;
}

template <typename Ch, typename T, typename A>
bool 
ends_with (std::basic_string<Ch,T,A> const& str, std::basic_string_view<Ch,T> const a) CXXO_NOEXCEPTNOTHROW {
    // anything convertible to basic_string_view can use this one
    using namespace std;
    if (a.size() > str.size()) return false;
    return equal(a.rbegin(), a.rend(), str.rbegin());
}

template <typename Ch, typename T, typename A>
bool 
starts_with (std::basic_string<Ch,T,A> const& str, Ch const a) CXXO_NOEXCEPTNOTHROW {
    //std::cout<< "[starts_with (char) ]"<< std::flush;
    return str.find(a) == 0;
}

template <typename Ch, typename T, typename A>
bool 
ends_with (std::basic_string<Ch,T,A> const& str, Ch const a) CXXO_NOEXCEPTNOTHROW {
    //std::cout<< "[ends_with (char) ]"<< std::flush;
    std::basic_string_view<Ch,T> const arg(&a, &a+1);
    return ends_with(str,arg);
}

template <typename Ch, typename T, typename A, typename Arg>
bool 
starts_with (std::basic_string<Ch,T,A> const& str, Arg const& a) CXXO_NOEXCEPTNOTHROW {
    // anything convertible to basic_string_view can use this one
    //std::cout<< "[starts_with (arg) ]"<< std::flush;
    //std::string_view const arg= a;
    return str.find(a) == 0;
}

template <typename Ch, typename T, typename A, typename Arg>
bool 
ends_with (std::basic_string<Ch,T,A> const& str, Arg const& a) CXXO_NOEXCEPTNOTHROW {
    // anything convertible to basic_string_view can use this one
    //std::cout<< "[ends_with (arg) ]"<< std::flush;
    std::basic_string_view<Ch,T> const arg= a;
    return ends_with(str,arg);
}


template <typename Ch, typename T>
bool 
starts_with (std::basic_string_view<Ch,T> const str, std::basic_string_view<Ch,T> const a) CXXO_NOEXCEPTNOTHROW {
    // anything convertible to basic_string_view can use this one
    return str.find(a.data(), 0, a.size()) == 0;
}

template <typename Ch, typename T>
bool 
ends_with (std::basic_string_view<Ch,T> const str, std::basic_string_view<Ch,T> const a) CXXO_NOEXCEPTNOTHROW {
    // anything convertible to basic_string_view can use this one
    using namespace std;
    if (a.size() > str.size()) return false;
    return equal(a.rbegin(), a.rend(), str.rbegin());
}

template <typename Ch, typename T>
bool 
starts_with (std::basic_string_view<Ch,T> const str, Ch const a) CXXO_NOEXCEPTNOTHROW {
    return str.find(a) == 0;
}

template <typename Ch, typename T>
bool 
ends_with (std::basic_string_view<Ch,T> const str, Ch const a) CXXO_NOEXCEPTNOTHROW {
    std::basic_string_view<Ch,T> const arg(&a, &a+1);
    return ends_with(str,arg);
}

*/    
    #endif
    
} }  //  cxxomfort::string

#endif
