#ifndef CXXOMFORT_IMPL_23_INVOKE_HPP
#define CXXOMFORT_IMPL_23_INVOKE_HPP
/**
 * @file
 * @brief Implementation of std::invoke_r
 * 
 * This header defines:
 * 
 * * invoke_r
 * 
 * The implementation for C++11-esque invoke_result from 
 * https://en.cppreference.com/mwiki/index.php?title=cpp/types/result_of&oldid=102232
 */


#include "../config.hpp"
#include <cxxomfort/base.hpp>
#include <cxxomfort/functional.hpp>
#include <functional>

#if defined(DOXYGEN_DOC)

namespace cxxomfort { namespace cxxostd {

//! @private 
struct dependent_type_t {};

/**
 * @anchor cxxostd_invoke_r
 * @brief Invokes a callable and specifies a return type. See also @c invoke() .
 * @ingroup std0functional
 * @cxx23b Invoke callables via <tt>(R)callable( args... )</tt>, specifying the return type.
 * @xrefitem std0functional "" "" Invoke callables via <tt>(R)callable( args... )</tt>, specifying the return type.
 * @see @cppref{utility/functional/invoke} 
 * @see @c invoke()
 * @see @c is_invocable
 *
 * @c invoke_r() calls a functionlike or callable @a f with the given arguments @a args... , taking care of forwarding, member function pointers, and implicitly converts the resulting value to type @a R during return.
 * 
 * <tt>invoke_r<R>(f, 1, 2, 3, ...)</tt> is in general equivalent to <tt>static_cast<R>(f(1, 2, 3, ...))</tt>. 
 * 
 * The @c invoke_r() feature is backported down to C++11 with full support, and down to C++98 with @e partial support. Unlike @c invoke() which requires type deduction for return, @c invoke_r() can be used even in compilers that lack @c typeof . The macro @c CXXOMFORT_IMPLEMENTS_invoke_r informs about implementation status.
 * 
 * @note In C++11 and C++14 mode, @c invoke_r() may not necessarily be @c constexpr .
 * 
 * @note Pre-C++11, <tt>invoke_r()</tt> only supports object-like callables. To use with eg.: member function pointers, wrap those in <tt>mem_fn()</tt>, <tt>std::function</tt>, etc.
 * 
 * */
template <typename R, typename F, typename... Args>
R invoke_r (F&& callable, Args&&... args) {
}

template <typename R, typename F, typename... Args>
struct is_invocable_r{};

} }
	/**
	 * @cppfeature{@ref @c invoke_r()}
	 * */
	#define __cpp_lib_invoke_r YYYY00L

#elif (defined(__cpp_lib_invoke_r))
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r CXXO_IMPLSTATUS_NATIVE()

#elif (CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_decltypeauto && CXXO_COMPILER_SUPPORT_variadic)
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r (CXXO_IMPLSTATUS_BACKPORT()+2)
	#define __cpp_lib_invoke_r (CXXOMFORT_CXX_STD*100)
	#include "17-invoke.hpp" // std::invoke
	#include <type_traits>

namespace cxxomfort { namespace cxxostd {

template <typename R, typename V = typename std::enable_if<std::is_void<R>::value==false>::type  , typename F, typename... Args>
R invoke_r (F&& f, Args&&... args) {
	return std::invoke( std::forward<F>(f) , std::forward<Args>(args)... );
}

template <typename R, typename V = typename std::enable_if<std::is_void<R>::value==true>::type  , typename F, typename... Args>
void invoke_r (F&& f, Args&&... args) {
	std::invoke( std::forward<F>(f) , std::forward<Args>(args)... );
}

} }

namespace std {
	using cxxomfort::cxxostd::invoke_r;
}

// 
// Base variadic implementation that doesn't rely on std::invoke 
//
#elif (CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_template_aliases && CXXO_COMPILER_SUPPORT_variadic)
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r (CXXO_IMPLSTATUS_BACKPORT()+0)
	#define __cpp_lib_invoke_r (CXXOMFORT_CXX_STD*100)
	#include <utility>
	#include <type_traits>

namespace cxxomfort { namespace cxxostd {

template <typename R, typename V = typename std::enable_if<std::is_void<R>::value==false>::type  , typename F, typename... Args>
R invoke_r (F&& f, Args&&... args) {
	return static_cast<R>(std::forward<F>(f) ( std::forward<Args>(args)... ) );
}

template <typename R, typename V = typename std::enable_if<std::is_void<R>::value==false>::type  , typename T, typename Base, typename... Args>
R invoke_r (T Base::* pm, Args&&... args) {
	return invoke_r<R>(std::mem_fn(pm), std::forward<Args>(args)... );
}

template <typename R, typename V = typename std::enable_if<std::is_void<R>::value==true>::type  , typename F, typename... Args>
void invoke_r (F&& f, Args&&... args) {
	(std::forward<F>(f) ( std::forward<Args>(args)... ) );
}

/*
template <typename R, typename V = typename std::enable_if<std::is_void<R>::value==true>::type  , typename T, typename Base, typename... Args>
R invoke_r (T Base::* pm, Args&&... args) {
	invoke_r(std::mem_fn(pm), std::forward<Args>(args)... );
}
*/


} }

namespace std {
	using cxxomfort::cxxostd::invoke_r;
}

// 
// C++0x non-variadic approach (eg.: MSVC, DMC)
// here, we also might not have default template arguments or function-level template specialization
// 
#elif (CXXO_COMPILER_SUPPORT_rvref)
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r (CXXO_IMPLSTATUS_EMULATION()+1)
	#define __cpp_lib_invoke_r (CXXOMFORT_CXX_STD*100)
	#include <utility>
	#include <type_traits>

namespace cxxomfort { namespace cxxostd {

template <typename R, typename F>
R invoke_r (F&& f) {
	return (std::forward<F>(f)( ) );
}

template <typename R, typename F, typename A1>
R invoke_r (F&& f, A1&& a1) {
	return static_cast<R>(std::forward<F>(f)( std::forward<A1>(a1) ) );
}

template <typename R, typename F, typename A1, typename A2>
R invoke_r (F&& f, A1&& a1, A2&& a2) {
	return static_cast<R>(std::forward<F>(f)( std::forward<A1>(a1),std::forward<A2>(a2) ) );
}


template <typename R, typename F, typename A1, typename A2, typename A3>
R invoke_r (F&& f, A1&& a1, A2&& a2, A3&& a3) {
	return static_cast<R>(std::forward<F>(f)( std::forward<A1>(a1),std::forward<A2>(a2),std::forward<A3>(a3) ) );
}

template <typename R, typename F, typename A1, typename A2, typename A3, typename A4>
R invoke_r (F&& f, A1&& a1, A2&& a2, A3&& a3, A4&& a4) {
	return (std::forward<F>(f)( std::forward<A1>(a1),std::forward<A2>(a2),std::forward<A3>(a3),std::forward<A4>(a4) ) );
}

template <typename R, typename F, typename A1, typename A2, typename A3, typename A4, typename A5>
R invoke_r (F&& f, A1&& a1, A2&& a2, A3&& a3, A4&& a4, A5&& a5) {
	return (std::forward<F>(f)( std::forward<A1>(a1),std::forward<A2>(a2),std::forward<A3>(a3),std::forward<A4>(a4),std::forward<A5>(a5) ) );
}

} }

namespace std {
	using cxxomfort::cxxostd::invoke_r;
}

// 
// C++03 support with cxxomfort
// 
#elif (CXXOMFORT_CXX_STD>=1997)
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r (CXXO_IMPLSTATUS_EMULATION()+1)
	#define __cpp_lib_invoke_r (CXXOMFORT_CXX_STD*100)

namespace cxxomfort { namespace cxxostd {

template <typename R, typename F>
R invoke_r (F f) { 
	return static_cast<R>((f)() );
}

template <typename R, typename F, typename A0>
R invoke_r (F f, CXXO_FWD_REF(A0) a0) { 
	return static_cast<R>((f)(std::forward<A0>(a0) ) );
}

template <typename R, typename F, typename A0, typename A1>
R invoke_r (F f, CXXO_FWD_REF(A0) a0, CXXO_FWD_REF(A1) a1) { 
	return static_cast<R>((f)(std::forward<A0>(a0), std::forward<A1>(a1) ) );
}

template <typename R, typename F, typename A0, typename A1, typename A2>
R invoke_r (F f, CXXO_FWD_REF(A0) a0, CXXO_FWD_REF(A1) a1, CXXO_FWD_REF(A2) a2) { 
	return static_cast<R>((f)(std::forward<A0>(a0), std::forward<A1>(a1),std::forward<A2>(a2) ) ); 
}

template <typename R, typename F, typename A0, typename A1, typename A2, typename A3>
R invoke_r (F f, CXXO_FWD_REF(A0) a0, CXXO_FWD_REF(A1) a1, CXXO_FWD_REF(A2) a2, CXXO_FWD_REF(A3) a3) { 
	return (f)(std::forward<A0>(a0), std::forward<A1>(a1),std::forward<A2>(a2),std::forward<A3>(a3) ); 
}

} }

namespace std {
	using cxxomfort::cxxostd::invoke_r;
}

#elif 1
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r CXXO_IMPLSTATUS_EMULATION()
	#define __cpp_lib_invoke_r 1

namespace cxxomfort { namespace cxxostd { 

template <typename R, typename F>
struct invoke_wrapper_r_0 {
	F& f;
	explicit invoke_wrapper_r_0 (F& fn)
	: f(fn) { }
	
	R operator() () const { return f(); }
	template <typename A1>
	R operator() (A1 const& a1) const {
		return static_cast<R>(f(a1));
	}
	template <typename A1, typename A2>
	R operator() (A1 const& a1, A2 const& a2) const {
		return static_cast<R>(f(a1,a2));
	}
	template <typename A1, typename A2, typename A3>
	R operator() (A1 const& a1, A2 const& a2, A3 const& a3) const {
		return static_cast<R>(f(a1,a2,a3));
	}

};

template <typename R, typename F>
R invoke_r (F f) {
	return invoke_wrapper_r_0<R,F>(f)();
}

template <typename R, typename F, typename A1>
R invoke_r (F f, A1 a1) {
	return invoke_wrapper_r_0<R,F>(f)(a1);
}

template <typename R, typename F, typename A1, typename A2>
R invoke_r (F f, A1 a1, A2 a2) {
	return invoke_wrapper_r_0<R,F>(f)(a1,a2);
}

template <typename R, typename F, typename A1, typename A2, typename A3>
R invoke_r (F f, A1 a1, A2 a2, A3 a3) {
	return invoke_wrapper_r_0<R,F>(f)(a1,a2,a3);
}


} }

namespace std {
	using cxxomfort::cxxostd::invoke_r;
}

#else
	#undef  CXXOMFORT_IMPLEMENTS_invoke_r
	#define CXXOMFORT_IMPLEMENTS_invoke_r 0

#endif // invoke_r

// 
// is_invocable_r
// 

#if (CXXO_COMPILER_SUPPORT_invoke_r)

#endif


#endif
