#ifndef CXXOMFORT_FOREACH_HPP
#define CXXOMFORT_FOREACH_HPP

/**
 * @file
 */

#include <cxxomfort/base.hpp> 
#include <cxxomfort/base/iterator.hpp>
#include <cxxomfort/impl/07-type_traits.hpp>

#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_COMPILER_SUPPORT_foreach > 0)
    // C++11 onwards, plus some compilers with c++0x support
    #define CXXOMFORT_IMPLEMENTS_n2930 CXXO_IMPLSTATUS_NATIVE()
    #undef  CXXOMFORT_IMPLEMENTS_foreach
    #define CXXOMFORT_IMPLEMENTS_foreach CXXOMFORT_IMPLEMENTS_n2930

    #define CXXO_FOREACH(VDecl,CDecl) for (VDecl : CDecl)
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>2)
		#pragma message CXXO_NOTICE("foreach is native")
	#endif

#elif (CXXOMFORT_CXX_STD >= 1997)

    #define CXXOMFORT_IMPLEMENTS_n2930 CXXO_IMPLSTATUS_EMULATION()
    #undef  CXXOMFORT_IMPLEMENTS_foreach
    #define CXXOMFORT_IMPLEMENTS_foreach CXXOMFORT_IMPLEMENTS_n2930

    #if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
        #pragma message CXXO_NOTICE("foreach emulation iteration support")
    #endif

    #include "../impl/foreach_artima.hpp"

#else

/*
 * For any other compiler, we use Artima article's basic solution
 * up to Page 2, that is, support for evaluating the container
 * only once. Any support further than that (eg.: working with
 * returned values) seems to require Boost at the moment.
 *
 * The code that follows here is a modification of the original
 * Artima article "listing_1.cpp" modified to suit the use of
 * cxxomfort's capabilities.
 *
 * - modifications from page 2 of the article to evaluate the
 * container expression once.
 * - support from iterator.hpp to use global begin(), end(),
 * which allows it to trivially support eg.: C arrays and strings.
 *
 */

    #include "../impl/foreach_artima.hpp"
    #define CXXOMFORT_USING_FOREACH_EMULATION 1
    #define CXXOMFORT_IMPLEMENTS_n2930 CXXO_IMPLSTATUS_EMULATION()
    #undef  CXXOMFORT_IMPLEMENTS_foreach
    #define CXXOMFORT_IMPLEMENTS_foreach CXXOMFORT_IMPLEMENTS_n2930
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>1)
		#pragma message CXXO_NOTICE("foreach basic emulation supportt")
	#endif

#endif // c++ mode

#if defined(DOXYGEN_DOC)
/**
 * @def CXXO_FOREACH(itemdecl,collection)
 * @brief Foreach loop construct that iterates over @p collection .
 * @ingroup cxxomfort-utils
 * @see http://www.artima.com/cppsource/foreach.html
 * 
 * See also @link std0iterator "<iterator> support" @endlink .
 *
 * A foreach macro declares iteration over @p collection , which can be
 * any STL-compatible sequence, using @p itemdecl as the iterating item.
 *
 * Usage is as follows:
 *
 * @code
list<int> L = {2, 3, 5, 7, 11};
CXXO_FOREACH (int x, L) {
  cout<< x<< ' ';
}
 * @endcode
 * 
 * When using C++11 onwards, this evaluates to the native range-for syntax, for zero overhead.
 * 
 * @note FOREACH support varies according to compiler capabilities. In particular, for compilers that do not fully support at least C++98/C++03, the macro works @b only with lvalues.
 * 
 * 
 */
#define CXXO_FOREACH(itemdecl,collection) for (implementation_specific_arguments;;)

#endif // doc

#endif

/**
 * @page cxxo-foreach 
 * @brief Foreach emulation
 * @ingroup cxxomfort-utils
 * @see http://www.artima.com/cppsource/foreach.html
 * @xrefitem cxxomfort-utils "Supplements" "Foreach emulation"
 * 
 * @code
 * #include <cxxomfort/cxxo-utils.hpp>
 * // also via
 * #include <cxxomfort/library/foreach.hpp>
 * @endcode
 * 
 * Cxxomfort implements emulation of C++11's "foreach" or 
 * <tt>for(item: collection)</tt>, for systems without C++11.
 * 
 * The interface is provided via the @c @ref CXXO_FOREACH macro and is used as follows:
 * 
 * @code
 * 
 * deque<int> dq;
 * // add elements to dq...
 * // later:
 * CXXO_FOREACH(int v , dq) { 
 *  // operate on the elements...
 * }
 * @endcode
 * 
 * This feature allows working with sequences and ranges transparently up to the limits established by the compiler. Currently foreach support in cxxomfort goes as follows:
 * 
 * 
 * * In C++11 mode onwards, this expands to the native <code>for (itemdecl : collection)</code> and is fully compatible with eg.: ref rvalues.
 * range-for iteration, incurring in no cost.
 * * In C++03 mode in compilers supporting @c __typeof__ (eg.: GCC 4.x, Clang), this makes use of typeof to implement something similar to 
 * <tt>for (unspecified_iterator b = begin(collection), e = end(collection) ; b != e ; ++b) if(itemdecl) ...</tt>. <br/> This has no extra cost but might not be fully compatible with some constructs.
 * * In C++03 mode without @c __typeof__ support (eg.: MSVC 2008-2010), a type-eraed interface is built using the same base that the original Artima article on foreach emulation provides. This has the cost of one virtual function call per iteration.
 * * In compilers without at least C++03 support (eg.: MSVC 2015, Open Watcom), a type-erased reference system is used based on original, simplified versions of the Artima article. This however works @b only with lvalues.
 *  
 * 
 * @section also See Also:
 * 
 * @see Boost.Foreach
 * @see the <a href="https://www.artima.com/cppsource/foreach.html">Eric Niebler's article on FOREACH</a>.
 * @see @ref cxxo-i12n 
 * 
 * 
 * */

