#ifndef CXXOMFORT_IMPL_FUNCTIONMETA_HPP
#define CXXOMFORT_IMPL_FUNCTIONMETA_HPP


/**
 * @file
 * @brief Meta information for function types.
 * 
 * Interfaces defined in this file:
 * 
 * * @c functiontype - traits for properties about function signatures (return type, arity, etc).
 * * @c 'function_caller' - wraps a normal C function into a zero-cost functor.
 * 
 * */

#include <cxxomfort/utility.hpp>
#include "../utils/typelist.hpp"

namespace cxxomfort { namespace impl {

template <typename T> struct functiontype {};

#if (CXXOMFORT_CXX_STD >= 2011)

#define MAKE_FUNCTIONTYPE(_ignore, ...)  \
template <typename Ret, typename... Args>  \
struct functiontype<Ret(Args...) __VA_ARGS__ > { \
    typedef Ret (* __VA_ARGS__ pointer)(Args...) ;    \
    typedef Ret (* const const_pointer)(Args...) ;  \
    typedef Ret (type)(Args...);  \
    typedef Ret return_type;  \
    typedef ::cxxomfort::typelist<Args...> argument_types; \
    enum { arity = sizeof...(Args) };  \
    template <size_t I> struct get_argument : ::cxxomfort::typelist_get<I, argument_types> { }; \
}

MAKE_FUNCTIONTYPE(0);
/*
MAKE_FUNCTIONTYPE(0, const);
	#if (0 && CXXOMFORT_CXX_STD>=2017)
MAKE_FUNCTIONTYPE(0, noexcept);
MAKE_FUNCTIONTYPE(0, const noexcept);
	#endif
*/

#undef MAKE_FUNCTIONTYPE

#else

template <typename Ret>
struct functiontype<Ret()> {
    typedef Ret(*pointer)();
    typedef Ret(* const const_pointer)();
    typedef Ret(type)();
    typedef Ret(&reference)();
    typedef Ret return_type;
    enum { arity = 0 };
    template <size_t I> struct get_argument {}; // intentionally empty
};

template <typename Ret, typename A1>
struct functiontype<Ret(A1)> {
    typedef Ret (*pointer)(A1);
    typedef Ret (type)(A1);
    typedef Ret return_type;
    typedef ::cxxomfort::typelist<A1> argument_types;
    enum { arity = 1 };
    template <size_t I> struct get_argument : ::cxxomfort::typelist_get<I, argument_types>::type { }; \
};
template <typename Ret, typename A1, typename A2>
struct functiontype<Ret(A1,A2)> {
    typedef Ret (*pointer)(A1,A2);
    typedef Ret (type)(A1,A2);
    typedef Ret return_type;
    typedef ::cxxomfort::typelist<A1,A2> argument_types;
    enum { arity = 2};
};
template <typename Ret, typename A1, typename A2, typename A3>
struct functiontype<Ret(A1,A2,A3)> {
    typedef Ret (*pointer)(A1,A2,A3);
    typedef Ret (type)(A1,A2,A3);
    typedef Ret return_type;
    typedef ::cxxomfort::typelist<A1,A2,A3> argument_types;
    enum { arity = 3};
};
template <typename Ret, typename A1, typename A2, typename A3, typename A4>
struct functiontype<Ret(A1,A2,A3,A4)> {
    typedef Ret (*pointer)(A1,A2,A3,A4);
    typedef Ret (type)(A1,A2,A3,A4);
    typedef Ret return_type;
    typedef ::cxxomfort::typelist<A1,A2,A3,A4> argument_types;
    enum { arity = 4 };
};
template <typename Ret, typename A1, typename A2, typename A3, typename A4, typename A5>
struct functiontype<Ret(A1,A2,A3,A4,A5)> {
    typedef Ret (*pointer)(A1,A2,A3,A4,A5);
    typedef Ret (type)(A1,A2,A3,A4,A5);
    typedef Ret return_type;
    typedef ::cxxomfort::typelist<A1,A2,A3,A4,A5> argument_types;
    enum { arity = 5};
};
template <typename Ret, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6>
struct functiontype<Ret(A1,A2,A3,A4,A5,A6)> {
    typedef Ret (*pointer)(A1,A2,A3,A4,A5,A6);
    typedef Ret (type)(A1,A2,A3,A4,A5,A6);
    typedef Ret return_type;
    typedef ::cxxomfort::typelist<A1,A2,A3,A4,A5,A6> argument_types;
    enum { arity = 6};
};
#endif

// 
// function_caller
// 

/*
 * function_caller wrapps a C function into a 
 * zero-cost callable unique.
 * eg.:
 * function_caller< int() , &rand > makerand;
 * 
 * originates in 
 * 
 * */
#if defined(DOXYGEN_DOC)


template <typename Signature, Signature const * fn>
struct function_caller {
	//! Invokes the function with the given @a arguments... .
	return_type operator() (arguments...);
	//! Returns a pointer to the function ([static])
	static typename functiontype<Signature>::pointer ptr() noexcept { return fn; }

};

#elif (CXXOMFORT_CXX_STD>=2011 \
|| (CXXO_COMPILER_SUPPORT_decltype && CXXO_COMPILER_SUPPORT_rvref && CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_trailing_returns ) \
)
	#define CXXOMFORT_IMPLEMENTS_function_caller 2012

template <typename Fn, Fn const * fn>
struct function_caller {
	template <typename... Args>
	auto operator() (Args&&... args) const
        -> decltype(fn(std::forward<Args...>(args...)))
    {
        return fn(std::forward<Args...>(args...));
    }
	
	typedef typename functiontype<Fn>::return_type return_type;
	static typename functiontype<Fn>::pointer ptr() noexcept { return fn; }

};

#elif (CXXOMFORT_CXX_STD >= 2011)

	#define CXXOMFORT_IMPLEMENTS_function_caller 2011

template <typename Fn, Fn const * fn>
struct function_caller {
    typedef ::cxxomfort::impl::functiontype<Fn> FT;
    typedef typename FT::return_type return_type;
    
    static typename FT::pointer ptr() noexcept { return fn; }

    template <typename... Args
    , typename = std::enable_if < FT::arity == sizeof...(Args)> 
    >
    CXXO_CONSTEXPR typename FT::return_type operator() (Args&&... args) const {
        return fn(std::forward<Args...>(args...));
    }

	#if (CXXOMFORT_CXX_STD>=2017)
    template <typename... Args>
    CXXO_CONSTEXPR typename FT::return_type operator() (Args&&... args) const noexcept {
        return fn(std::forward<Args...>(args...));
    }
	#endif

};

#elif (CXXOMFORT_CXX_STD>=1997) // c++03

	#define CXXOMFORT_IMPLEMENTS_function_caller 2007

template <typename Fn, Fn const * fn>
struct function_caller;

template <typename R, typename A1, R(* const fn)(A1)>
struct function_caller<R(A1), fn> {
    typedef ::cxxomfort::impl::functiontype<R(A1)> FT;
    typedef typename FT::return_type return_type;

    return_type operator() 
    (A1 a1) const {
        return fn(a1);
    }

	#if (CXXOMFORT_CXX_STD>=1997)
    static typename FT::pointer ptr () CXXO_NOEXCEPTNOTHROW {
		return fn;
	}
	#endif
};

template <typename R, typename A1, typename A2, R(* const fn)(A1,A2)>
struct function_caller<R(A1,A2), fn> {
    typedef ::cxxomfort::impl::functiontype<R(A1,A2)> FT;
    typedef typename FT::return_type return_type;

    return_type operator() 
    (A1 a1,A2 a2) const {
        return fn(a1,a2);
    }
    
    #if (CXXOMFORT_CXX_STD>=1997)
    static typename FT::pointer ptr () CXXO_NOEXCEPTNOTHROW { return fn; }
    #endif
};

template <typename R, typename A1, typename A2, typename A3, R(* const fn)(A1,A2,A3)>
struct function_caller<R(A1,A2,A3), fn> {
    typedef ::cxxomfort::impl::functiontype<R(A1,A2,A3)> FT;
    typedef typename FT::return_type return_type;

    return_type operator() 
    (A1 a1, A2 a2, A3 a3) const {
        return fn(a1,a2,a3);
    }

	#if (CXXOMFORT_CXX_STD>=1997)
    static typename FT::pointer ptr () CXXO_NOEXCEPTNOTHROW { return fn; }
    #endif
};


template <typename R, typename A1, typename A2, typename A3, typename A4, R(* const fn)(A1,A2,A3,A4)>
struct function_caller<R(A1,A2,A3,A4), fn> {
    typedef ::cxxomfort::impl::functiontype<R(A1,A2,A3,A4)> FT;
    typedef typename FT::return_type return_type;

    return_type operator() 
    (A1 a1, A2 a2, A3 a3, A4 a4) const {
        return fn(a1,a2,a3,a4);
    }
    
	#if (CXXOMFORT_CXX_STD>=1997)
    static typename FT::pointer ptr () CXXO_NOEXCEPTNOTHROW { return fn; }
    #endif
};

template <typename R
, typename A1, typename A2, typename A3, typename A4, typename A5, R(* const fn)(A1,A2,A3,A4,A5)>
struct function_caller<R(A1,A2,A3,A4,A5), fn> {
    typedef ::cxxomfort::impl::functiontype<R(A1,A2,A3,A4,A5)> FT;
    typedef typename FT::return_type return_type;

    return_type operator() 
    (A1 a1, A2 a2, A3 a3, A4 a4, A5 a5) const {
        return fn(a1,a2,a3,a4,a5);
    }
    
	#if (CXXOMFORT_CXX_STD>=1997)
    static typename FT::pointer ptr () CXXO_NOEXCEPTNOTHROW { return fn; }
    #endif
};


#else // CXX_STD < 1997

	#define CXXOMFORT_IMPLEMENTS_function_caller 0

#endif


// specialization for zero-argument callables

#if (CXXOMFORT_CXX_STD >= 2017)

	#if (0)
template <typename T, T(* const fn)() noexcept >
struct function_caller<T() noexcept, fn> {
		#define MODIFIER noexcept 
    typedef ::cxxomfort::impl::functiontype<T() MODIFIER> FT;  
    typedef typename FT::return_type return_type;  
	T operator() () const MODIFIER { return fn(); }  
};
		#undef MODIFIER
	#endif

#elif (CXXOMFORT_CXX_STD >= 2011)

#elif (CXXOMFORT_CXX_STD>=1997)

template <typename T, T(* const fn)() >
struct function_caller<T(), fn> {
    typedef ::cxxomfort::impl::functiontype<T()> FT;  \
    typedef typename FT::return_type return_type;  \
	T operator() () const { return fn(); }  \
};

#endif

} }


#endif
