#ifndef CXXOMFORT_IMPL_I12N_HPP
#define CXXOMFORT_IMPL_I12N_HPP

/**
 * @page cxxo-i12n "I12N (initialization)"
 * @ingroup cxxo-utils
 * 
 * Utilities for the initialization of sequence-liek objects / containers.
 * 
 * Defines the dual macros @c CXXO_I12N_BEG and @c CXXO_I12N_END 
 * that provide initialization semantics for containers similar to the 
 * initializer_list constructors in C++11.
 * 
 * It's called "i12n" from "initialization", the same way 
 * "i18n" is used for "internationalization".
 * 
 * Interfaces defined:
 * 
 * * @ref CXXO_I12N_BEG - prolog macro for initialization
 * * @ref CXXO_I12N_END - epilog macro for initialization
 * * @ref CXXO_I12N_SEQ - variadic macro for initialization
 * 
 * Sample usage:
 * 
 * @code

// Create a vector<int> with a given sequence of values.
CXXO_I12N_BEG( vector<int> , vn );
{1, 2, 3, 4, 5, 6, 7}
CXXO_I12N_END( vector<int> , vn );

// Create a deque<int> with a given sequence of values.
CXXO_I12N_SEQ( deque<int> , dn , 
 {1, 2, 3, 4, 5, 6, 7}
 );


 * @endcode
 * 
 * @b Requirements:
 * 
 * Requirements:
 * 
 * * Containers using this utility must provide an adequate @c value_type member.
 * * Containers using this utility must support iterator-range constructors.
 * * In C++03 mode, data types using this utility must be copyable.
 * 
 * Because of these requirements and other reasons intrinsic to its structure 
 * as a pseudo-container different from the C++ template sequences, 
 * i12n can not work with <code>std::array<...></code>.
 * 
 * @see
 * * @ref seq_() - generation of sequences.
 * 
**/

#if defined(DOXYGEN_DOC)

/**
 * @def CXXO_I12N_BEG(SequenceT,Var) 
 * @brief Begins an initialization sequence.
 * @param Type the type of the desired object.
 * @param Var the name of the desired object.
 * 
 * See usage case.
 * 
 */
    #define CXXO_I12N_BEG(Type, Var) _implementation_defined_ Type Var

/**
 * @def CXXO_I12N_END(SequenceT,Var) 
 * @brief Finishes an initialization sequence.
 * @param SequenceT the type of the desired object.
 * @param Var the name of the desired object.
 * 
 * See usage case.
 * 
 */
    #define CXXO_I12N_END(SequenceT, Var) _implementation_defined_ ; (void)SequenceT 


/**
 * @def CXXO_I12N_SEQ(SequenceT,Var,List...) 
 * @brief Constructs a sequence-like object.
 * @param Type the type of the desired object.
 * @param Var the name of the desired object.
 * @param List a brace-enclosed listing of elements for the object.
 * 
 * This macro wraps the usage of CXXO_I12_BEG and CXXO_I12N_END in a singular "oneliner" for compilers that have support for variadic macros.
 * 
 * 
 */
    #define CXXO_I12N_SEQ(SequenceT, Var, {List...}) _implementation_defined_ SequenceT Var = { List ... }

#elif (CXXOMFORT_CXX_STD >= 2011)

// Begins a sequence initialization construct
    #define CXXO_I12N_BEG(Type,Name) Type  Name
// Finishes a sequence initialization construct
    #define CXXO_I12N_END(Type,Name)
// Variadic sequence initialization construct
    #define CXXO_I12N_SEQ(Type,Name,...)  Type Name ({__VA_ARGS__});

#elif (CXXOMFORT_CXX_STD >= 0)

    #if defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1)
        #pragma message CXXO_NOTICE ("enabled sequence initialization helper")
    #endif

    #include <iterator>
    #include <cxxomfort/base/iterator.hpp>

namespace cxxomfort { namespace i12n_detail {

template <typename Container> struct I12N_for {
    typename Container::value_type const* starter;
    typename Container::value_type const* ender;
    template <size_t N>
    explicit I12N_for (typename Container::value_type const (&arr)[N]) 
    : starter(arr), ender(::std::end(arr)) {
    }
};

} } 

// specialization for std::valarray, C++03 onwards

    #if (CXXO_CXXSTDHEADER_valarray)
        #include <valarray>
namespace cxxomfort { namespace i12n_detail {

template <typename T> 
struct I12N_for < std::valarray<T> > {
    typedef std::valarray<T> VT;
    typename VT::value_type const* starter;
             size_t                 ender;
    template <size_t N>
    explicit I12N_for (typename std::valarray<T>::value_type const (&arr)[N]) 
    : starter(arr), ender(N) { 
    }

};

} } // cxxomfort::impl

    #endif
// end specialization

    #define CXXO_I12N_BEG(Type,Name) typedef /*typename*/ std::remove_cv<Type>::type Name##0ncv; static const Name##0ncv::value_type Name##0init []
    #define CXXO_I12N_END(Type,Name) ; Type Name ((Name##0init), cxxomfort::i12n_detail::I12N_for< Name##0ncv >(Name##0init).ender)

// If compiler supports variadic macros, this can be used:
    #if (CXXO_COMPILER_SUPPORT_va_args==1 )
        #define CXXO_I12N_SEQ(Type,Name,...) CXXO_I12N_BEG(Type,Name) =  __VA_ARGS__ CXXO_I12N_END(Type,Name)
    #endif

#else

#endif // c++11

#endif
