#ifndef CXXOMFORT_IMPL_MATH_HPP
#define CXXOMFORT_IMPL_MATH_HPP
/**
 * @file 
 * @brief Math functions particular to cxxomfort
 *  
**/

#include <cmath>
#include <utility>
#include <limits>
#include <cxxomfort/type_traits.hpp>

namespace cxxomfort { namespace impl {

// 
// sign 
// 

/**
 * @brief Returns the sign of @a n .
 * @pre @a n is of a real-like type.
 * */
template <typename N>
CXXO_CONSTEXPR 
int sign (N n) CXXO_NOEXCEPTNOTHROW {
    return n < N(0) ? -1 : n == N(0) ? 0 : +1;
}

// 
// intpow
// 


// 
// intlog
// 

/*
 * returns the integer logarithm of x in base B.
 * aborts if the base is 0 or 1
 * returns 0 if x is 0
 * */

template <typename T> CXXO_CONSTEXPR14 
typename std::enable_if< std::numeric_limits<T>::is_integer, T >::type 
intlog (T x, uint16_t B) CXXO_NOEXCEPT {
	if (B<2) throw static_cast<int>(B);
	if (x==T(0U)) return T(0U);
	return (x==T(1U)) ? T(0U) : (x == T(B)) ? T(1) : T(1) + intlog(x/B,B);
}

/*
 * returns the integer logarithm of x in base B.
 * returns 0 if x is 0
 * 
 * if clz builtin is available, it uses that for calculation
 * */
template <typename U> CXXO_CONSTEXPR 
typename std::enable_if< std::numeric_limits<U>::is_integer, U >::type 
intlogb (U u) CXXO_NOEXCEPT {
	return (u==U(0U)) ? 0 : (u==U(1U)) ? U(0U) : (u == U(2)) ? U(1) : U(1) + intlogb(u/2);
}


// 
// intsqrt
// 

// 
// gcd, lcm 
// 

// 
// extended gcd
// 

struct exgcd_t {
	uintmax_t g;
    intmax_t x, y;
};

static 
exgcd_t egcd (intmax_t a, intmax_t b) CXXO_NOEXCEPTNOTHROW {
	exgcd_t ret = { uintmax_t(b), 0, 1 };
    if (a == 0) {
        return ret; 
    } else {
		exgcd_t r2= egcd(b % a, a);
		exgcd_t r3=	{ r2.g, r2.y - intmax_t(b/a * r2.x), r2.x };
		return r3;
	}
}

// 
// factors 
// 

/**
 * @brief Returns the number of times that @a f appears as factor of @a n .
 * @pre @a n is of integer-like type; <tt>n > 0</tt>; <tt>f > 1</tt>
 * */
template <typename I>
uintmax_t factors (I n, uintmax_t f) CXXO_NOEXCEPTNOTHROW {
    if (f<=1) return 0;
    uintmax_t res= 0;
    while (n >= f) {
        n/= f; ++res;
    }
    return res;
}


// 
// round to nearest (low, up) power of k
// 

template <typename I>
I neardown_powk (I n, uintmax_t k) {
	++k;
    return n;
}

template <typename I>
I nearup_powk (I n, uintmax_t k) {
	++k;
    return n;
}


// A naive method to find modular multiplicative inverse of 
// 'a' under modulo 'm' 
static inline
uintmax_t ModInverseNaive (uintmax_t a, uintmax_t m)  { 
	a = a % m; 
	for (uintmax_t x = 1; x < m; x++) {
		if ((a * x) % m == 1) return x; 
	}
	return 0;
} 


// 
// multiplicative inverse modulo b
// (from Rosetta Code)
// 

intmax_t mul_inv(intmax_t a, intmax_t b)
{
    intmax_t t, nt, r, nr, q, tmp;
    if (b < 0) b = -b;
    if (a < 0) a = b - (-a % b);
    t = 0;  nt = 1;  r = b;  nr = a % b;
    while (nr != 0) {
      q = r/nr;
      tmp = nt;  nt = t - q*nt;  t = tmp;
      tmp = nr;  nr = r - q*nr;  r = tmp;
    }
    if (r > 1) return 0; // -1;  /* No inverse */
    if (t < 0) t += b;
    return t;
}


// 
// multiplicative inverse modulo q
// (uses: extended gcd)
// 

static inline
uintmax_t mulinvmod (uintmax_t a, uintmax_t q) CXXO_NOEXCEPTNOTHROW { 
	exgcd_t r= egcd(a, q);
    return (r.g != uintmax_t(1)) ?  uintmax_t(0) :  uintmax_t(r.x % q);
} 


} } // cxxomfort::impl

    #if (CXXOMFORT_CXX_STD >= 1997)
        #include "static_math.hpp"
    #endif

#endif

