#ifndef CXXOMFORT_IMPL_MATHBIT_HPP
#define CXXOMFORT_IMPL_MATHBIT_HPP
/**
 * @file
 *
 */

// 
// assistant functions 
// derived mostly from Hacker's Delight 2nd Ed
// where possible, these wrap bit-counting intrinsics
// 

#if (defined(_MSC_VER))
	#include <intrin.h>
#endif


namespace cxxomfort { namespace bits {

// posl1u (position of leading 1-bit) (aka ilog2)
// posr1u (position of trailing 1-bit)
// cntl1u (count of leading 1-bits)
// cntt1u (count of trailing 1-bits)
// cntl0u (count of leading 0-bits) (aka clz)
// cntt0u (count of trailing 0-bits)
// maskt0 (mask for all trailing bits 0)
// leftmost_uint64 (returns the leftmost uint64_t-size bits of n)
// leftmost_uint32 (returns the leftmost uint32_t-size bits of n)
// lshift_bb (left-shift so that leftmost 1-bit lies at a byte boundary)
// rshift_bb (right-shift so that rightmost 1-bit lies at a byte boundary)
// signext_bn (sign-extend an integer codified as a n-bit number)
// 

namespace {
template <typename U> CXXO_CONSTEXPR inline 
int32_t 	pl1_nz_impl(U u, uint32_t p) CXXO_NOEXCEPT {
	return (u & ::cxxomfort::fix::bit_masks<U>::lmb1 ? p : pl1_nz_impl(u<<1,p-1));
}
}

/**
 * @brief Obtain the bit position of the leftmost @c 1 .
 * @return The bit position (<tt>1<<p</tt>) of the leftmost @c 1 , or @c -1 if there's none.
 * @pre @a U is a native integral type.
 * @xrefitem cxxomfort-fix "" "" Bit operations: Obtain the position of the leftmost 1.
 * */
template <typename U> CXXO_CONSTEXPR inline 
int32_t 	posl1u (U u) CXXO_NOEXCEPT	{
	return u == U(0U) ? -1  : pl1_nz_impl(u,std::numeric_limits<U>::digits-1);
}

/**
 * @return The bit position (<tt>1<<p</tt>) of the leftmost @c 0 , or @c -1 if there's none.
 * @pre @a U is a native integral type.
 * @xrefitem cxxomfort-fix "" "" Bit operations: Obtain the position of the leftmost 0.
 * */
template <typename U> CXXO_CONSTEXPR inline 
int32_t 	posl0i (U u) CXXO_NOEXCEPT	{
	return posl1i(~u);
}

template <typename U> CXXO_CONSTEXPR inline 
U maskt0u (U u) CXXO_NOEXCEPT {
	return U(U(~u) & U (u - 1));
}

/**
 * @brief Obtain the count of leftmost @c 1 bits.
 * @return The number of contiguous @c 0 bits bit in @a u , starting from the left (most sifnigicant bits).
 * @pre @a U is a native integral type.
 * @xrefitem cxxomfort-fix "" "" Bit operations: Count of leading 0 bits.
 * */
template <typename U> CXXO_CONSTEXPR inline 
int32_t 	cntl0u (U u) CXXO_NOEXCEPT {
	//std::cout<< "[cntl0u invoke generic]"<< std::flush;
	return (u == U(0U)) ? std::numeric_limits<U>::digits : std::numeric_limits<U>::digits - posl1u(u) - 1;
}

template <typename U> CXXO_CONSTEXPR inline 
int32_t 	cntt0u (U u) CXXO_NOEXCEPT {
	return u==0 ? std::numeric_limits<U>::digits : u & 1 ? 0 : 1+cntt0u(u>>1);
}

#if defined(__GNUC__)
template<>  CXXO_CONSTEXPR inline 
int32_t 	cntl0u (unsigned int u) CXXO_NOEXCEPT {
	//std::cout<< "[cntl0u invoke uint = "<< __builtin_clz(u)<< " ]"<< std::flush;
	return u == 0 ? CHAR_BIT*sizeof(unsigned int) : __builtin_clz(u);
}
template<>  CXXO_CONSTEXPR inline 
int32_t 	cntl0u (unsigned long u) CXXO_NOEXCEPT {
	//std::cout<< "[cntl0u invoke ul]"<< std::flush;
	return u == 0 ? CHAR_BIT*sizeof(unsigned long) : __builtin_clzl(u);
}
template<>  CXXO_CONSTEXPR inline 
int32_t 	cntl0u (unsigned long long u) CXXO_NOEXCEPT {
	//std::cout<< "[cntl0u invoke ull]"<< std::flush;
	return u == 0 ? CHAR_BIT*sizeof(unsigned long long) : __builtin_clzll(u);
}

#elif defined(_MSC_VER)

#endif

template <typename U> CXXO_CONSTEXPR inline 
int32_t 	cntl1u (U u) CXXO_NOEXCEPT {
	return cntl0u(~u);
}


// 
// popcount 
// 



} } //bits

#endif
