#ifndef CXXOMFORT_IMPL_n4022_HPP
#define CXXOMFORT_IMPL_n4022_HPP
/**
 * @file 
 * @brief Backport implementation of @c not_fn() .
 *  
**/

#include <cxxomfort/base.hpp>
#include <cxxomfort/functional.hpp>

#if defined(DOXYGEN_DOC)

namespace std {

/**
 * @brief Creates a function wrapper that returns the negation of a given function-like / callable.
 * @param f A function-like callable with <tt>operator()</tt>.
 * @return A function object wrapper @a nf of undetermined type such that <tt>nf(args...) == !(f(args...))</tt>.
 * @xrefitem std0functional "" "" <tt>not_fn()</tt> from Fundamentals v2: a generic functor negator. 
 * @cxx17b <tt>not_fn()</tt> from Fundamentals v2: a generic functor negator. 
 * @see @cppref{utility/functional/not_fn}
 * 
 * The macro <tt>CXXOMFORT_IMPLEMENTS_not_fn</tt> informs about the implementation status. Additionally in C++>=11 mode, cxxomfort sets the macro <tt>__cpp_lib_not_fn</tt> if this facility is backported.
 * 
 * @note In C++03 mode, this wrapper can take function-likes of up to 5 arguments.
 * 
 * @note In C++03 mode, this wrapper can only take arguments by value.
 * 
 * */
template <typename Fn>
not_fn_t<Fn> not_fn (Fn&& f) {
    return not_fn_t<Fn>( std::forward<Fn>(f) );
}

} 

#elif (defined(__cpp_lib_not_fn))
	#define CXXOMFORT_IMPLEMENTS_n3699 CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD >= 2011)
	#define CXXOMFORT_IMPLEMENTS_n3699 CXXO_IMPLSTATUS_BACKPORT()
	#define __cpp_lib_not_fn 201103
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>1)
		#pragma message CXXO_NOTICE("enabled functional: not_fn<F> backport")
	#endif
	#include <type_traits>

namespace cxxomfort { namespace cxxostd {

template <typename Functor>
struct not_fn_t {
    typedef typename std::decay<Functor>::type DeFunctor; // fun name, not actually a deleter or smth
    DeFunctor _d;

	explicit not_fn_t (Functor&& f)
    : _d(std::forward<Functor>(f))
    {} 

    template <typename... Args>
    auto operator()  (Args&&... args) 
    -> decltype (!invoke(_d, std::forward<Args>(args)...)) {
        return !invoke(_d, std::forward<Args>(args)...);
    }
    template <typename... Args>
    auto operator()  (Args&&... args) const 
    -> decltype (!invoke(_d, std::forward<Args>(args)...)) {
        return !invoke(_d, std::forward<Args>(args)...);
    }

};

} }

namespace std {

template <typename Fn>
cxxomfort::cxxostd::not_fn_t<Fn> not_fn (Fn&& f) {
    return cxxomfort::cxxostd::not_fn_t<Fn>( std::forward<Fn>(f) );
}

} // std

#elif (0 && CXXO_COMPILER_SUPPORT_typeof>0)

#elif (CXXOMFORT_CXX_STD >= 1997)
	#define CXXOMFORT_IMPLEMENTS_n3699 CXXO_IMPLSTATUS_EMULATION()
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES>1)
		#pragma message CXXO_NOTICE("enabled functional: not_fn<F> emulation")
	#endif
	#define __cpp_lib_not_fn 199703


namespace cxxomfort { namespace cxxostd {

template <typename Functor>
struct not_fn_t {
    typedef typename std::decay<Functor>::type DeFunctor; // fun name, not actually a deleter or smth
    DeFunctor _d;

    explicit not_fn_t (Functor const& f)
    : _d(f)
    {}

    bool operator() () const {
        return !_d();
    }
    template <typename T1>
    bool operator() (T1 t1) const {
        return !_d(t1);
    }
    template <typename T1, typename T2>
    bool operator() (T1 t1, T2 t2) const {
        return !_d(t1,t2);
    }
    template <typename T1, typename T2, typename T3>
    bool operator() (T1 t1, T2 t2, T3 t3) const {
        return !_d(t1,t2,t3);
    }
    template <typename T1, typename T2, typename T3, typename T4>
    bool operator() (T1 t1, T2 t2, T3 t3, T4 t4) const {
        return !_d(t1,t2,t3,t4);
    }
    template <typename T1, typename T2, typename T3, typename T4, typename T5>
    bool operator() (T1 t1, T2 t2, T3 t3, T4 t4, T5 t5) const {
        return !_d(t1,t2,t3,t4,t5);
    }


};

} }

namespace std {

template <typename Fn>
cxxomfort::cxxostd::not_fn_t<Fn> not_fn (Fn const& f) {
    return cxxomfort::cxxostd::not_fn_t<Fn>(f);
}

}

#else
	#define CXXOMFORT_IMPLEMENTS_n3699 0

#endif

#define CXXOMFORT_IMPLEMENTS_n4022 CXXOMFORT_IMPLEMENTS_n3699
#define CXXOMFORT_IMPLEMENTS_not_fn CXXOMFORT_IMPLEMENTS_n4022


#endif
