#ifndef CXXOMFORT_IMPL_n4031_HPP
#define CXXOMFORT_IMPL_n4031_HPP

#include "../config.hpp"
#include "../base.hpp" // move emulation
#include <cxxomfort/utility.hpp>
#include <cxxomfort/impl/lift_type_traits.hpp> // remove_cv
#include <iostream>

//
// make_array (n4031, n4065, n4315)
// contructs an object of std::array<T,N> type with automatically calculated N.
// See proposal:  http://www.open-std.org/JTC1/SC22/WG21/docs/papers/2014/n4031.html
//
// http://stackoverflow.com/questions/6114067/how-to-emulate-c-array-initialization-int-arr-e1-e2-e3-behaviou
//


#if defined(DOXYGEN_DOC)

namespace cxxomfort {

/**
 * @brief Creates a <tt>std::array</tt> from a list of arguments.
 * @ingroup std0array
 * 
 * @xrefitem std0array "" "" Creates an array (part of <tt>experimental/lfv2</tt>)
 * 
 * The expression <tt>make_array<V>(arguments...)</tt> creates an array as follows:
 * 
 * @code
 * auto myarray = make_array<int> (3, 4, 5, 6, -7, 8);
 * // type of "myarray" is std::array<int,6>
 * @endcode
 * 
 * This is added to namespace <tt>std::experimental</tt>.
 * 
 * The accessory function @c make_cxxo_array() creates a <tt>cxxomfort::array</tt> instead.
 * 
 * @pre The @a arguments... resolve to a type that can be used to construct @a V .
 * 
 * @note In C++11 onward, tye type @a V can be ommitted, in which case it is deduced via <tt>std::common_type</tt>.
 * 
 * */
template <typename R, typename... T>
constexpr std::array<R, sizeof...(T) > make_array (T&&... t) ;

/**
 * @brief Creates a <tt>cxxomfort::array</tt> from a list of arguments.
 * @ingroup cxxomfort-fixes
 * 
 * @xrefitem std0array "" "" Creates an array (part of <tt>experimental/lfv2</tt>)
 * 
 * @xrefitem cxxomfort-fixes "" "" Creates an array (part of <tt>experimental/lfv2</tt>)
 * 
 * This function is similar to @c make_array() except it creates a <tt>cxxomfort::array</tt> instead. It is part of @ref cxxomfort-fixes .
 * 
 * */
template <typename R, typename... T>
constexpr cxxomfort::array<R, sizeof...(T) > make_cxxo_array (T&&... t) ;

}

#elif (CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_constexpr && CXXO_COMPILER_SUPPORT_rvref )

	#include <type_traits>
	
namespace cxxomfort {

    // Here type can be deduced (via common_type) or given explicitly.
    template <typename V, typename... T>
    constexpr std::array < V, sizeof...(T)> 
    make_array (T&&... t) {
        using namespace std;
        return {{ std::forward<T>(t)... }};
    }

    template <typename V, typename... T>
    constexpr ::cxxomfort::array < V, sizeof...(T)> 
    make_cxxo_array (T&&... t) {
        using namespace std;
        return {{ std::forward<T>(t)... }};
    }

} // cxxomfort

#else // C++03

namespace cxxomfort { 

    // c++03 versions, with forwarding emulation
    // for this implementation, the type should be given explicitly

    template <typename V, typename T> inline
    std::array<V,1> make_array(CXXO_FWD_REF(T) t1) {
        std::array<V,1> ret= { {std::forward<T>(t1)} };
        return ret;
    }

    template <typename V, typename T0, typename T1> inline
    std::array<V,2>  
    make_array (CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1
    ) {
        using std::forward;
        std::array<V,2> ret= { {forward<T0>(t0), forward<T1>(t1)} };
        return ret;
    }

    template <typename V, typename T0, typename T1, typename T2> inline
    std::array<V,3> 
    make_array (CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1, CXXO_FWD_REF(T2) t2
    ) {
        using std::forward;
        std::array<V,3> ret= { {forward<T0>(t0), forward<T1>(t1), forward<T2>(t2)} };
        return ret;
    }

    template <typename V, typename T0, typename T1, typename T2, typename T3> inline
    std::array<V,4> 
    make_array (CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1, CXXO_FWD_REF(T2) t2
            , CXXO_FWD_REF(T3) t3
    ) {
        using std::forward;
        std::array<V,4> ret= { {forward<T0>(t0),forward<T1>(t1),forward<T2>(t2)
            , forward<T3>(t3)} };
        return ret;
    }

    template <typename V, typename T0, typename T1, typename T2, typename T3, typename T4> inline
    std::array<V,5> 
    make_array (CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1, CXXO_FWD_REF(T2) t2
            , CXXO_FWD_REF(T3) t3, CXXO_FWD_REF(T4) t4
    ) {
        using std::forward;
        std::array<V,5> ret= { {forward<T0>(t0),forward<T1>(t1),forward<T2>(t2)
            , forward<T3>(t3), forward<T4>(t4)} };
        return ret;
    }

    template <typename V, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5> inline
    std::array<V,6> make_array(
        CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1, CXXO_FWD_REF(T2) t2
      , CXXO_FWD_REF(T3) t3, CXXO_FWD_REF(T4) t4, CXXO_FWD_REF(T5) t5
    ) {
        using std::forward;
        std::array<V,6> ret= { {forward<T0>(t0), forward<T1>(t1), forward<T2>(t2)
          , forward<T3>(t3), forward<T4>(t4), forward<T5>(t5)} };
        return ret;
    }

    template <typename V, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6> inline
    std::array<V,7> make_array(
        CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1, CXXO_FWD_REF(T2) t2
      , CXXO_FWD_REF(T3) t3, CXXO_FWD_REF(T4) t4, CXXO_FWD_REF(T5) t5
      , CXXO_FWD_REF(T6) t6
    ) {
        using std::forward;
        std::array<V,7> ret= { {forward<T0>(t0), forward<T1>(t1), forward<T2>(t2)
          , forward<T3>(t3), forward<T4>(t4), forward<T5>(t5)
          , forward<T6>(t6)} };
        return ret;
    }

    template <typename V, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7> inline
    std::array<V,8> make_array(
        CXXO_FWD_REF(T0) t0, CXXO_FWD_REF(T1) t1, CXXO_FWD_REF(T2) t2
      , CXXO_FWD_REF(T3) t3, CXXO_FWD_REF(T4) t4, CXXO_FWD_REF(T5) t5
      , CXXO_FWD_REF(T6) t6, CXXO_FWD_REF(T7) t7
    ) {
        using std::forward;
        std::array<V,8> ret= { {forward<T0>(t0), forward<T1>(t1), forward<T2>(t2)
          , forward<T3>(t3), forward<T4>(t4), forward<T5>(t5)
          , forward<T6>(t6), forward<T7>(t7) } };
        return ret;
    }


} // cxxomfort

#endif

// mirrored to std::experimental when included via <cxxomfort/array>

#endif

#ifdef DOXYGEN_DOC
namespace cxxomfort {

/**
 * @brief Makes a <tt>std::array</tt> from an argument list.
 * @param args... The list of elements that goes into the array.
 * @tparam T The element type (@c value_type ) of the returned @c array .
 * 
 * @c make_array() creates a std::array given an argument list and a type to which the arguments are converted for placing them in the array. The array returned has a size equal to the number of arguments passed.
 * 
 * The function takes one template parameter @a T , which is the element type of the returned array. Starting with C++11, if the type given here is @c void , the type of the array will be automatically deduced if possible as per <tt>std::common_type</tt>. Pre-C++11, the type @a T must be given explicitly.
 * 
 * @code
 * // here A1 is array<short,4>, 
 * // and some of the elements are narrowed
 * auto A1 = make_array<short> (1, 2, 30000000UL, 4);
 * // here A2 is array< (int)(*)(), 3>
 * // , that is, an array of function pointers
 * auto A2 = make_array<void>(rand, drand, lrand);
 * 
 * @endcode
 * 
 * */
template <typename T, typename... Args>
std::array<T, sizeof...(Args)> 
make_array (Args...&& args)
{};

}

#endif
