#ifndef CXXOMFORT_IMPL_p0040_HPP
#define CXXOMFORT_IMPL_p0040_HPP
/**
 * @file
 * @brief Implementation of std::destroy_at and allies
 * 
 */

#define CXXOMFORT_IMPLEMENTS_p0040r1 CXXO_IMPLSTATUS_BACKPORT()
#define CXXOMFORT_IMPLEMENTS_p0040r2 CXXOMFORT_IMPLEMENTS_p0040r1
#undef  CXXOMFORT_IMPLEMENTS_destroy_at
#define CXXOMFORT_IMPLEMENTS_destroy_at CXXOMFORT_IMPLEMENTS_p0040r2

#include <cxxomfort/config.hpp>
#include <memory>


#if defined(DOXYGEN_DOC)
namespace std {
#else
namespace cxxomfort {
namespace cxxostd {
#endif

/**
 * 
 * @ingroup cxx17-backports
 * @{
 * **/

/**
 * @cxx17b Invokes the destructor of an object.
 * @xrefitem std0memory "Object Management" "" Invokes the destructor of an object.
 * @see @cppref{memory/destroy_at}
 * @sa @wg21{p0040r2}
 * */
template <typename T>
void destroy_at (T* p) CXXO_NOEXCEPTNOTHROW {
    p->~T();
}

/**
 * @cxx17b Destroys a sequence of objects, invoking the destructor on each.
 * @xrefitem std0memory "Object Management" "" Destroys a sequence of objects, invoking the destructor on each.
 * @see @cppref{memory/destroy}
 * @sa http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2016/p0040r2.html
 * */
template <typename Iter, typename Iter2>
void destroy (Iter ini, Iter2 fin) {
    for (; ini != fin; ++ini) destroy_at(std::addressof(*ini));
}

/**
 * @cxx17b Destroys a sequence of objects, invoking the destructor on each.
 * @xrefitem std0memory "Object Management" "" Destroys a sequence of objects, invoking the destructor on each.
 * @see @cppref{memory/destroy}
 * @sa http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2016/p0040r2.html
 * */
template <typename Iter, typename Integer>
void destroy_n (Iter ini, Integer n) {
    for (Integer i=0; i < n; ++i, ++ini) destroy_at(std::addressof(*ini));
}


/**
 * @xrefitem std0memory "Object Management" "" Constructs objects in uninitialized memory, via the default constructor.
 * @cxx17b Constructs objects in uninitialized memory, via the default constructor.
 * @throw <tt>bad_alloc</tt>.
 * @sa @wg21{p0040r0} .
 * */
template <typename It>
It uninitialized_default_construct (It ini, It fin) {
    typedef typename std::iterator_traits<It>::value_type Value;
    It cur = ini;
    CXXO_try() {
        for (; cur != fin; ++cur) {
            ::new (static_cast<void*>(std::addressof(*cur))) Value;
        }
        return cur;
    }  CXXO_catch(...) {
        destroy(ini, cur);
        CXXO_THROW_STD();
    }
}

/**
 * @xrefitem std0memory "Object Management" "" Constructs objects in uninitialized memory, via the default constructor.
 * @cxx17b Constructs objects in uninitialized memory, via the default constructor.
 * @throw <tt>bad_alloc</tt>.
 * @sa @wg21{p0040r0} .
 * */
template <typename It, typename Integer>
It uninitialized_default_construct_n (It ini, Integer n) {
    typedef typename std::iterator_traits<It>::value_type Value;
    It cur = ini;
    CXXO_try() {
        for (; n > 0; (void) ++cur, --n) {
            ::new (static_cast<void*>(std::addressof(*cur))) Value;
        }
        //return cur;
    }  CXXO_catch(...) {
        destroy(ini, cur);
        CXXO_RETHROW_STD();
    }
    return cur;
}

template <typename It>
It uninitialized_value_construct (It ini, It fin) {
    typedef typename std::iterator_traits<It>::value_type Value;
    It cur = ini;
    CXXO_try() {
        for (; cur != fin; ++cur) {
            ::new (static_cast<void*>(std::addressof(*cur))) Value();
        }
        return cur;
    }  CXXO_catch(...) {
        destroy(ini, cur);
        CXXO_RETHROW_STD();
    }
}

template <typename It, typename Integer>
It uninitialized_value_construct_n (It ini, Integer n) {
    typedef typename std::iterator_traits<It>::value_type Value;
    It cur = ini;
    CXXO_try() {
        for (; n > 0; (void) ++cur, --n) {
            ::new (static_cast<void*>(std::addressof(*cur))) Value();
        }
    }  CXXO_catch(...) {
        destroy(ini, cur);
        CXXO_THROW_STD();
    }
    return cur;
}

/**
 * @}
 * */

#if defined(DOXYGEN_DOC)
}
#else
}}
#endif

#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (CXXOMFORT_CXX_STD < 2017)
namespace std {
    using cxxomfort::cxxostd::destroy_at;
    using cxxomfort::cxxostd::destroy;
    using cxxomfort::cxxostd::destroy_n;
    using cxxomfort::cxxostd::uninitialized_default_construct;
    using cxxomfort::cxxostd::uninitialized_default_construct_n;
    using cxxomfort::cxxostd::uninitialized_value_construct;
    using cxxomfort::cxxostd::uninitialized_value_construct_n;
}
#endif

#endif
