#ifndef CXXOMFORT_IMPL_p0792_FUNCTIONVIEW_HPP
#define CXXOMFORT_IMPL_p0792_FUNCTIONVIEW_HPP

/**
 * @file
 * @brief Implements part of p0792r0, 
 * a proposal for a non-owning reference view on functors.
 * 
 * */

#undef  CXXOMFORT_IMPLEMENTS_function_ref
#define CXXOMFORT_IMPLEMENTS_p0792 0
/* 
 * Depends:
 * C++11 reference_wrapper
 * C++11 conditionals
 * C++11 move, forward
 * C++11 basic category type traits
 * */
 
#include <functional> // std::reference_wrapper, etc
#include <utility> // forward, move, etc
//#include <cxxomfort/typeindex.hpp>
#include <cxxomfort/impl/functionmeta.hpp> // functiontype, argument_type
#include <cxxomfort/impl/07-type_traits.hpp>
#include <cxxomfort/impl/11-conditional.hpp>

#if defined(DOXYGEN_DOC)

	namespace cxxomfort { namespace impl { namespace p0792 {
/**
 * @class function_ref
 * @brief A non-owning view of an object callable with a function signature.
 * @ingroup cxxo-sup-functional
 * 
 * A @c function_ref is a non-owning view to a callable object or function pointer 
 * for a given signature.
 * 
 * @code

// note here we don't require a template for the functor
template <typename List>
vector<int> metafunction (function_ref<int(float)> F, List& L) {
    vector<int> sideeffects;
    for (typename L::reference ll : L) {
        sideeffects.push_back ( F(ll) );
    }
    return sideeffects;
}

int foo (float x) { ... };
int bar (float r) { ... };


list<float> data = ...;
vector<int> allfoos = metafunction(foo, data);
vector<int> allbars = metafunction(bar, data);


 * @endcode
 * 
 * @notr @c function_ref is a @c _ref - hence it does not work with 
 * temporaries such as bind expressions. The following doesn't work:

@code

metafunction ( bind(mem_fn(&string::size), foo ) );

@endcode

 * @note In C++03 mode, @e function_ref can only construct 
 * from function pointers explicitly.
 * 
 * @see
 * * @wg21{0792r0,function_ref proposal}
 * * @cppref{functional/function}
 * 
 * */

template <typename Signature> 
class function_ref {
	typedef _implementation_defined_t return_type;
	
	//! Set up a @c function_ref with a @c Signature -compatible @a callable object.
	template <typename Fn>
	function_ref (Fn& callable) {}
	
	//! Invoke the referenced function
	template <typename Args...>
	return_type 
	operator() (Args&&... args) {
		return caller(args...);
	}
	
	private:
	_implementation_defined_t caller;
	
};


	} /* p0792  */ } } 


#elif (CXXOMFORT_CXX_STD >= 2011 || \
(CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_rvref) \
)
	// 
	// C++11 or at least &&, variadic, trailing return support
	// 
	
	#undef  CXXOMFORT_IMPLEMENTS_p0792
	#define CXXOMFORT_IMPLEMENTS_p0792 CXXO_IMPLSTATUS_BACKPORT()

/*
 * 3/4tht taken from StackOverflow [1] 
 * 1/4th drawn from LLVM [2] 
 * We are taking extra from the LLVM because of Eric Niebler's "keep it simple" comment  https://twitter.com/ericniebler/status/880899867035394048
 * 
 *  [1] https://stackoverflow.com/questions/32918679/in-c-11-how-to-invoke-an-arbitrary-callable-object 
 * [2] http://llvm.org/doxygen/STLExtras_8h_source.html
 *
 */

namespace cxxomfort { namespace impl { namespace p0792 {

template <typename Signature> class function_ref;

// 
// function_ref 
// 
template<typename Ret, typename ...Args>
class function_ref<Ret(Args...)>  {
	typedef typename 
	cxxomfort::impl::functiontype<Ret(void *, Args...)>::type call_t;
	typedef void* voidptr;

	call_t* callback;
	voidptr callable;
	std::type_info const* tinfo/* = nullptr*/;

	template<typename Callable>
	static Ret callback_fn (voidptr callable, Args ...as) {
		Callable const* c= reinterpret_cast<Callable const*>(callable);
		return (*c)(std::forward<Args>(as)...);
	}
 
	public:
	//! deleted default ctor - a @c function_ref needs to reference something.
	function_ref() = delete;
	// function_ref(std::nullptr_t) {}

	//! Constructor - sets @c function_ref to reference a @a callable object.
	template <typename Callable>
	CXXO_CONSTEXPR14 
	function_ref(Callable&& callable,
		// callable needs not to be yet another @c function_ref
		typename std::enable_if<
			!std::is_same<typename std::remove_reference<Callable>::type,
			function_ref>::value
		>::type * = nullptr) 
	noexcept
	:  callback(callback_fn<typename std::remove_reference<Callable>::type>)
	,  callable(reinterpret_cast<voidptr>(&callable)) 
	,  tinfo(&typeid(callable))
	{}

    //! Invoke the callable.
	Ret operator() (Args&& ...as) const {
		return callback(callable, std::forward<Args>(as)...);
	}
 
	CXXO_CONSTEXPR14 
	operator bool() const noexcept { return callback; }
    
	std::type_info const& info () const noexcept {
		return *tinfo;
	}

};


} } }

#else
	// 
	// Implementation before C++03 
	// 
	#undef  CXXOMFORT_IMPLEMENTS_p0792
	#define CXXOMFORT_IMPLEMENTS_p0792 CXXO_IMPLSTATUS_EMULATION()

namespace cxxomfort { namespace impl { namespace p0792 {

template <typename Signature> class function_ref;

// function with 0 args

template <typename Ret>
class function_ref <Ret()>  {
	typedef typename functiontype<Ret()>::type function_type;
	typedef typename functiontype<Ret(void*)>::pointer call_t;
	call_t fn_;
	void* ptr_;
	std::type_info const* tinfo;
    
	template<typename Callable>
	static Ret callback_fn(void* const callable) {
		Callable const* c= reinterpret_cast<Callable* const>(callable);
		return (*c)( );
	}

	private:
	function_ref (); // = delete

	public:

	template <typename Callable>
	function_ref (Callable& callable,
		typename std::enable_if<
			!std::is_same<Callable, function_ref>::value
		>::type * = nullptr
	) 
	CXXO_NOEXCEPTNOTHROW 
	: fn_(callback_fn<typename cxxomfort::traits::remove_reference<Callable>::type>)
	, ptr_(reinterpret_cast<void*>(&callable)) 
	, tinfo(&typeid(callable))
	{ }

// MSVC <= 2010 does not generate the assignment operator correctly, needs help
	#if (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1500) && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,<1700))

    function_ref& operator= (Ret(callable)()) CXXO_NOEXCEPTNOTHROW {
        typename functiontype<Ret()>::type& f= callable;
        *this= function_ref(f);
        return *this;
    }

    template <typename U>
    function_ref& operator= (U(* callable)())     {
        typename functiontype<U()>::type& f= *callable;
        return *this= function_ref(f);
    }

	#endif

	Ret operator() () const {
		return fn_(ptr_ );
	}

	operator bool() const CXXO_NOEXCEPTNOTHROW { return fn_; }

	std::type_info const& info () const CXXO_NOEXCEPTNOTHROW {
		return *tinfo;
	}

};

// function_ref 1-argument
template <typename Ret, typename A1>
class function_ref <Ret(A1)> {
	typedef typename functiontype<Ret(void*,A1)>::pointer call_t;
	typedef void* voidptr;
	call_t fn_;
	voidptr ptr_;
	std::type_info const* tinfo;
    
	template<typename Callable>
	static Ret callback_fn(voidptr callable, A1 a1) {
		Callable* const c= reinterpret_cast<Callable* const>(callable);
		return (*c)(a1);
	}

	public:

	template <typename Callable>
	function_ref (Callable const& callable,
		typename std::enable_if<
			!std::is_same<typename std::remove_reference<Callable>::type,  function_ref>::value>
		::type * = nullptr
	) 
	CXXO_NOEXCEPTNOTHROW
	: fn_(callback_fn<typename cxxomfort::traits::remove_reference<Callable>::type>)
	, ptr_((void*)reinterpret_cast<voidptr>(&callable)) 
	{   }

	Ret operator() (CXXO_FWD_REF(A1) a1) const {
		return fn_(ptr_ , std::forward<A1>(a1));
	}
    
	operator bool() const CXXO_NOEXCEPTNOTHROW { return fn_; }

	std::type_info const& info () const CXXO_NOEXCEPTNOTHROW { return *tinfo; }
};


template <typename Ret, typename A1, typename A2>
class function_ref <Ret(A1,A2)> {
    enum { IsConst = true };
    typedef typename functiontype<Ret(void const*,A1,A2)>::pointer call_t;
    typedef typename std::conditional< IsConst, void const, void>::type voidtype;
    call_t fn_;
    voidtype* ptr_;
    std::type_info const* tinfo;
    
    template<typename Callable>
    static Ret callback_fn(voidtype* callable, A1 a1, A2 a2) {
        return (*reinterpret_cast<Callable const*>(callable))(a1,a2);
        
    }

    public:

    template <typename Callable>
    function_ref (Callable const& callable,
        typename std::enable_if<
            !std::is_same<typename std::remove_reference<Callable>::type,
              function_ref>::value>::type * = nullptr
    )
    : fn_(callback_fn<typename cxxomfort::traits::remove_reference<Callable>::type>)
    , ptr_(reinterpret_cast<voidtype*>(&callable)) 
    {   }

    Ret operator() (CXXO_FWD_REF(A1) a1,CXXO_FWD_REF(A2) a2) const {
        return fn_( ptr_ , std::forward<A1>(a1), std::forward<A2>(a2) );
    }

    operator bool() const CXXO_NOEXCEPTNOTHROW { return fn_; }

    std::type_info const& info () const CXXO_NOEXCEPTNOTHROW {
        return *tinfo;
    }
    
};


template <typename Ret, typename A1, typename A2, typename A3>
class function_ref <Ret(A1,A2,A3)> {
    enum { IsConst = true };
    typedef typename functiontype<Ret(void const*,A1,A2,A3)>::pointer call_t;
    typedef typename std::conditional< IsConst, void const, void>::type voidtype;
    call_t fn_;
    voidtype* ptr_;
    std::type_info const* tinfo;
    
    template<typename Callable>
    static Ret callback_fn(voidtype* callable, A1 a1, A2 a2, A3 a3) {
        return (*reinterpret_cast<Callable const*>(callable))(a1,a2,a3);
        
    }

    public:

    template <typename Callable>
    function_ref (Callable const& callable,
        typename std::enable_if<
            !std::is_same<typename std::remove_reference<Callable>::type,
              function_ref>::value>::type * = nullptr
    )
    : fn_(callback_fn<typename cxxomfort::traits::remove_reference<Callable>::type>)
    , ptr_(reinterpret_cast<voidtype*>(&callable)) 
    {   }

    Ret operator() (CXXO_FWD_REF(A1) a1,CXXO_FWD_REF(A2) a2, CXXO_FWD_REF(A3) a3) const {
        return fn_( ptr_ , std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3) );
    }

    operator bool() const CXXO_NOEXCEPTNOTHROW { return fn_; }

    std::type_info const& info () const CXXO_NOEXCEPTNOTHROW {
        return *tinfo;
    }
    
};

template <typename Ret, typename A1, typename A2, typename A3, typename A4>
class function_ref <Ret(A1,A2,A3,A4)> {
    enum { IsConst = true };
    typedef typename functiontype<Ret(void const*,A1,A2,A3,A4)>::pointer call_t;
    typedef typename std::conditional< IsConst, void const, void>::type voidtype;
    call_t fn_;
    voidtype* ptr_;
    std::type_info const* tinfo;
    
    template<typename Callable>
    static Ret callback_fn(voidtype* callable, A1 a1, A2 a2, A3 a3, A4 a4) {
        return (*reinterpret_cast<Callable const*>(callable))(a1,a2,a3,a4);
        
    }

    public:

    template <typename Callable>
    function_ref (Callable const& callable,
        typename std::enable_if<
            !std::is_same<typename std::remove_reference<Callable>::type,
              function_ref>::value>::type * = nullptr
    )
    : fn_(callback_fn<typename cxxomfort::traits::remove_reference<Callable>::type>)
    , ptr_(reinterpret_cast<voidtype*>(&callable)) 
    {   }

    Ret operator() (CXXO_FWD_REF(A1) a1,CXXO_FWD_REF(A2) a2, CXXO_FWD_REF(A3) a3, CXXO_FWD_REF(A4) a4) const {
        return fn_( ptr_ , std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3), std::forward<A4>(a4) );
    }

    operator bool() const CXXO_NOEXCEPTNOTHROW { return fn_; }

    std::type_info const& info () const CXXO_NOEXCEPTNOTHROW {
        return *tinfo;
    }
    
};


template <typename Ret, typename A1, typename A2, typename A3, typename A4, typename A5>
class function_ref <Ret(A1,A2,A3,A4,A5)> {
    enum { IsConst = true };
    typedef typename functiontype<Ret(void const*,A1,A2,A3,A4,A5)>::pointer call_t;
    typedef typename std::conditional< IsConst, void const, void>::type voidtype;
    call_t fn_;
    voidtype* ptr_;
    std::type_info const* tinfo;
    
    template<typename Callable>
    static Ret callback_fn(voidtype* callable, A1 a1, A2 a2, A3 a3, A4 a4, A5 a5) {
        return (*reinterpret_cast<Callable const*>(callable))(a1,a2,a3,a4,a5);
        
    }

    public:

    template <typename Callable>
    function_ref (Callable const& callable,
        typename std::enable_if<
            !std::is_same<typename std::remove_reference<Callable>::type,
              function_ref>::value>::type * = nullptr
    )
    : fn_(callback_fn<typename cxxomfort::traits::remove_reference<Callable>::type>)
    , ptr_(reinterpret_cast<voidtype*>(&callable)) 
    {   }

    Ret operator() (CXXO_FWD_REF(A1) a1,CXXO_FWD_REF(A2) a2, CXXO_FWD_REF(A3) a3, CXXO_FWD_REF(A4) a4, CXXO_FWD_REF(A5) a5) const {
        return fn_( ptr_ , std::forward<A1>(a1), std::forward<A2>(a2), std::forward<A3>(a3), std::forward<A4>(a4), std::forward<A5>(a5) );
    }

    operator bool() const CXXO_NOEXCEPTNOTHROW { return fn_; }

    std::type_info const& info () const CXXO_NOEXCEPTNOTHROW {
        return *tinfo;
    }
    
};


} } }

#endif // versions of C++ impl
#undef  CXXOMFORT_IMPLEMENTS_function_ref
#define CXXOMFORT_IMPLEMENTS_function_ref CXXOMFORT_IMPLEMENTS_p0792

#if (CXXOMFORT_IMPLEMENTS_function_ref > 0)
namespace cxxomfort {
	using cxxomfort::impl::p0792::function_ref;
}
#endif

#endif
