#ifndef CXXOMFORT_IMPL_p1402_HPP
#define CXXOMFORT_IMPL_p1402_HPP
// moved to utils/

/**
 * @file
 * @brief Implementation of <tt>cstring_view</tt> proposal
 * 
 * Implementation details:
 * 
 * * missing allocator support, raw strings only
 * 
 */

#define CXXOMFORT_IMPLEMENTS_p1402 1
#define CXXOMFORT_IMPLEMENTS_cstring_view CXXOMFORT_IMPLEMENTS_p1402

#include <cxxomfort/config.hpp>
#include <cxxomfort/string_view.hpp>
#include <ostream>
#include <cxxomfort/ostream.hpp>

namespace cxxomfort { namespace impl {

struct null_terminated_t {};
const null_terminated_t null_terminated = {};

/**
 * @brief @c string - like wrapper for a NUL-terminated C-style string.
 * @see @wg21{p1402r0} .
 * 
 * A @c basic_cstring_view is similar to a @c basic_string_view in most if not all respects, except that a @c basic_cstring_view expects NUL-termination.
 * 
 * */
template <typename Ch, typename ChT = std::char_traits<Ch> >
struct basic_cstring_view {
	private:
	typedef std::basic_string_view<Ch,ChT> sv_t;

	public:
	typedef typename sv_t::traits_type 		traits_type;
	typedef typename sv_t::value_type   	value_type;
	typedef typename sv_t::pointer      	pointer;
	typedef typename sv_t::const_pointer 	const_pointer;
	typedef typename sv_t::reference      	reference;
	typedef typename sv_t::const_reference 	const_reference;
	typedef typename sv_t::size_type  		size_type;
	typedef typename sv_t::const_iterator  	const_iterator;
	typedef typename sv_t::iterator  		iterator;
	typedef typename sv_t::const_reverse_iterator  	const_reverse_iterator;
	typedef typename sv_t::reverse_iterator  		reverse_iterator;

	public:
#if (CXXOMFORT_CXX_STD >= 2011 || DOXYGEN_DOC)
	//! default constructor
	CXXO_CONSTEXPR basic_cstring_view () noexcept = default;
	//! copy constructor
	CXXO_CONSTEXPR14 basic_cstring_view (basic_cstring_view const&) noexcept = default;
	//! assignment
	CXXO_CONSTEXPR14 basic_cstring_view& operator= (basic_cstring_view const&) noexcept = default;
#else
	basic_cstring_view () throw () : impl() {}
	basic_cstring_view (basic_cstring_view const& s) : impl(s.impl) {}
	basic_cstring_view& operator= (basic_cstring_view const& s) {
		impl= s.impl;
		return *this;
	}
#endif
	
	//! constructor with C-string
	CXXO_CONSTEXPR  basic_cstring_view (Ch const * str) 
	: impl(str) {}

	//! constructor marking a null-terminated string source
	CXXO_CONSTEXPR  basic_cstring_view (null_terminated_t, sv_t const& s) CXXO_NOEXCEPTNOTHROW 
	: impl(s) {}
	//! constructor marking a null-terminated string source
	CXXO_CONSTEXPR  basic_cstring_view (null_terminated_t, Ch const* str, size_type len)
	: impl(str,len) {}

	//! implicitly convert to the related <tt>basic_string_view</tt> type.
	CXXO_CONSTEXPR  operator sv_t () const CXXO_NOEXCEPTNOTHROW {
		return impl;
	}

	CXXO_CONSTEXPR  size_type max_size () const CXXO_NOEXCEPTNOTHROW { return impl.max_size(); }
	//! size
	CXXO_CONSTEXPR  size_type size () const CXXO_NOEXCEPTNOTHROW { return impl.size(); }
	//! length
	CXXO_CONSTEXPR  size_type length () const CXXO_NOEXCEPTNOTHROW { return impl.length(); }
	//! is empty?
	CXXO_CONSTEXPR  bool empty () const CXXO_NOEXCEPTNOTHROW { return impl.empty(); }
	//! access to data pointer
	CXXO_CONSTEXPR  const_pointer data () const CXXO_NOEXCEPT { return impl.data(); }
	//! access to data pointer
	const_pointer   c_str () const CXXO_NOEXCEPTNOTHROW { return impl.data(); }
	
	CXXO_CONSTEXPR const_reference operator[] (size_type idx) const CXXO_NOEXCEPT {
		return impl[idx];
	}
	CXXO_CONSTEXPR const_reference at (size_type idx) const CXXO_NOEXCEPT {
		return impl.at(idx);
	}

	const_reference front () const { return impl.front(); }
	const_reference back () const { return impl.back(); }

	CXXO_CONSTEXPR  const_iterator  cbegin () const CXXO_NOEXCEPTNOTHROW { return impl.cbegin(); }
	CXXO_CONSTEXPR  const_iterator  cend () const CXXO_NOEXCEPTNOTHROW { return impl.cend(); }
	CXXO_CONSTEXPR  const_iterator  begin () const CXXO_NOEXCEPTNOTHROW { return impl.begin(); }
	CXXO_CONSTEXPR  const_iterator  end () const CXXO_NOEXCEPTNOTHROW { return impl.end(); }

	CXXO_CONSTEXPR  const_reverse_iterator  crbegin () const CXXO_NOEXCEPTNOTHROW { return impl.crbegin(); }
	CXXO_CONSTEXPR  const_reverse_iterator  crend () const CXXO_NOEXCEPTNOTHROW { return impl.crend(); }
	CXXO_CONSTEXPR  const_reverse_iterator  rbegin () const CXXO_NOEXCEPTNOTHROW { return impl.rbegin(); }
	CXXO_CONSTEXPR  const_reverse_iterator  rend () const CXXO_NOEXCEPTNOTHROW { return impl.rend(); }
	
	// compare
	
	int compare (sv_t s) const CXXO_NOEXCEPTNOTHROW {
		return impl.compare(s);
	}
	int compare (pointer r) const CXXO_NOEXCEPTNOTHROW {
		return impl.compare(r);
	}
	
	// find
	
	
	// others
	
	void swap (basic_cstring_view& otro) CXXO_NOEXCEPTNOTHROW {
		std::swap(impl, otro.impl);
	}

	private:
	sv_t impl;
};


typedef basic_cstring_view<char> cstring_view;
typedef basic_cstring_view<wchar_t> wcstring_view;
#if (CXXOMFORT_CXX_STD >= 2011)
#endif



}} // cxxomfort::impl
#endif
