#ifndef CXXOMFORT_IMPL_RELATIONALS_HPP
#define CXXOMFORT_IMPL_RELATIONALS_HPP
/**
 * @file
 * @brief Macros for generating relational operators.
 */

#include <cxxomfort/config.hpp>

// This header used by cxxo-utils

/**
 * @brief Automatically generates relational operators between the classes @a L and @a R .
 * 
 * If classes @a L and @a R have working operators <tt>==</tt> and <tt><</tt>, invoking this macro 
 * <b>outside</b> their class definition will automatically generate compatible 
 * non-member operators != , <= , >, >= .
 * 
 * All generated operators have the signature <tt>bool(L const&, R const&)</tt> and are @c noexcept or @c throw() .
 * 
 * */
#define CXXO_GENERATE_RELATIONALS2(L,R) \
inline bool operator!= (L const& vl, R const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl==vr); } \
inline bool operator>  (L const& vl, R const& vr) CXXO_NOEXCEPTNOTHROW { return vr<vl; } \
inline bool operator<= (L const& vl, R const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl>vr); } \
inline bool operator>= (L const& vl, R const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl<vr); } \

/**
 * @brief Automatically generates nonmember relational operators >, >=, <= and != .
 * 
 * If class @a T has working operators <tt>==</tt> and <tt><</tt>, invoking this macro 
 * <b>outside</b> the class definition will automatically generate compatible 
 * non-member operators != , <= , >, >= .
 * 
 * All generated operators return @c bool and are @c noexcept or @c throw() .
 * 
 * */
#define CXXO_GENERATE_RELATIONALS(T) \
inline bool operator!= (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl==vr); } \
inline bool operator>  (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return vr<vl; } \
inline bool operator<= (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl>vr); } \
inline bool operator>= (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl<vr); } \

#define CXXO_GENERATE_FRIEND_RELATIONALS(T) \
friend bool operator!= (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl==vr); } \
friend bool operator>  (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return vr<vl; } \
friend bool operator<= (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl>vr); } \
friend bool operator>= (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl<vr); } \

#define CXXO_GENERATE_RELATIONALS_T(T) \
/** not_equal comparison operator for type T */ \
template <TPARAMS> \
inline bool operator!= (T<TARGS> const& vl, T<TARGS> const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl==vr); } \
/** greater comparison operator for type T */ \
template <TPARAMS> \
inline bool operator>  (T<TARGS> const& vl, T<TARGS> const& vr) CXXO_NOEXCEPTNOTHROW { return vr<vl; } \
/** less_equal comparison operator for type T */ \
template <TPARAMS> \
inline bool operator<= (T<TARGS> const& vl, T<TARGS> const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl>vr); } \
/** greater_equal comparison operator for type T */ \
template <TPARAMS> \
inline bool operator>= (T<TARGS> const& vl, T<TARGS> const& vr) CXXO_NOEXCEPTNOTHROW { return !(vl<vr); } \



#endif
