#ifndef CXXOMFORT_IMPL_STATIC_MATH_HPP
#define CXXOMFORT_IMPL_STATIC_MATH_HPP

/**
 * @file 
 * @brief Implements some static (compile-time) math functions for C++.
 * */
#include <cxxomfort/base.hpp>
#include <cxxomfort/cxxo-utils.hpp>
#include <cxxomfort/limits.hpp>
#include <cxxomfort/impl/lift_type_traits.hpp>
#include <cxxomfort/impl/11-conditional.hpp>

namespace cxxomfort {
//! Namespace for compile-time static (template based) arithmetic operations
namespace st_math {

namespace {
	using ::cxxomfort::fix::integral_limits;
}

/**
 * @ingroup cxxo-st-math
 * @{
 * */

// 
// abs, sign
// 

#if defined(DOXYGEN_DOC)
/**
 * @brief Provides the member @c value with the sign of template parameter @a T @a u .
 * @tparam u A compile-time value of type @a T used as argument.
 * @cxxoutils Compile-time sign of an integer
 * 
 * The helper types @c static_sign_s and @c statuc_sign_u omit the parameter types and operate on @c intmax_t and @c uintmax_t values, respectively.
 * 
 * @code
 * int sign_neg_999 = st_math::static_sign<short,-999>::value;
 * int sign_1 =  st_math::static_sign_s<148000000>::value;
 * int sign_2 =  st_math::static_sign_u<148000000>::value;
 * @endcode
 * 
 * See also @c static_abs .
 * 
 * */
template <typename T, T u>
struct static_sign { int value; };
#elif (CXXOMFORT_CXX_STD>=2003)
/*
 * we specialize on signedness in order to avoid spurious "unsigned comparison always returns..." warnings depending on T.
 * However, we can't do this on some older compilers that choke on the specz.
 * */
template <typename T, T u, bool If=false>
struct static_sign_impl {
	enum { value = (u==0 ? 0 : u < 0 ? -1 : +1 ) };
};

template <typename T, T u>
struct static_sign_impl<T,u,true> {
	enum { value = (u!=0) };
};

template <typename T, T u>
struct static_sign 
: public std::integral_constant<int, static_sign_impl<T,u, std::is_unsigned<T>::value>::value>
{ };

#else
template <typename T, T u>
struct static_sign 
: public std::integral_constant<int, (u > 0 ? +1 : u == 0 ? 0 : -1)>
{ };


#endif


/**
 * @brief Sign of a @c intmax_t .
 * @see static_sign
 * */
template <intmax_t i>
struct static_sign_s 
: std::integral_constant<int, (i > 0 ? +1 : i == 0 ? 0 : -1)>
{ };

/**
 * @brief Sign of a @c uintmax_t .
 * @see static_sign
 * */
template <uintmax_t i>
struct static_sign_u 
: std::integral_constant<int, (i == 0 ? 0 : +1)>
{ };

// 
// static_abs
// absolute value can be computed generically as 
// abs(x) = x * sign(x)
// which is guarantee to never cause signedness or overflow issues in unsigned types.
// 

/**
 * @brief Provides the member @c value with one of the values <tt>[-u, 0, +u]</tt> depending on the sign of @a u .
 * @cxxoutils Compile-time absolute value of an integral, pre-C++11.
 * 
 * The helper types @c static_abs_s and @c static_abs_u operate on @c intmax_t and @c uintmax_t values, respectively.
 * 
 * See also @c static_sign .
 * 
 * */
template <typename T, T u>
struct static_abs
: std::integral_constant<T, ( u * T(static_sign<T,u>::value) )>
{ };


/**
 * @brief Absolute value of a @c intmax_t .
 * @see static_abs
 * */
template <intmax_t i>
struct static_abs_s 
: std::integral_constant<intmax_t, (i < 0 ? -i : i)>
{ };

/**
 * @brief Absolute value of a @c uintmax_t .
 * @see static_abs
 * */
template <uintmax_t i>
struct static_abs_u 
: std::integral_constant<uintmax_t, i>
{ };

// 
// clamp
// 

/**
 * @brief Defines the member @a value which is @a t, constrained to the range <tt>[m,M]</tt>.
 * @cxxoutils Compile-time clamp a valur between two values, pre-C++11.
 * @tparam t A compile-time value of type @a T used as argument.
 * @tparam m,M Compile-time values of type @a T used as the minimum and maximum range in which to clamp value @a t .
 * @see std::clamp 
 * @pre <tt>m < M</tt>
 * */
template <typename T, T t, T m, T M >
struct static_clamp
: std::integral_constant<T, (t < m) ? m : (M < t ? M : t) >
{ };


// 
// minmax
// 

//! Defines the member values @c Tmin , @c Tmax with the minimum or maximum value between @a t0 and @a t1 .
template <typename T, T t0, T t1>
struct static_minmax2 {
	CXXO_CONSTEXPR static const T Tmin= t1 < t0 ? t1 : t0;
	CXXO_CONSTEXPR static const T Tmax= t1 > t0 ? t1 : t0;
};

// 
// min, max
// 

	#if (CXXOMFORT_CXX_STD < 1997)
template <typename T, T t0, T t1, T t2=t1, T t3=t1, T t4=t1>
struct static_min_impl
{
	static const T value = static_minmax2<T, 
		t0
		, static_minmax2<T, 
			t1
			, static_minmax2<T, 
				t2
				, static_minmax2<T,
					t3, t4
				>::Tmin
			>::Tmin
		>::Tmin
	>::Tmin;
};

template <typename T, T t0, T t1, T t2=t1, T t3=t1, T t4=t1> 
struct static_min 
: std::integral_constant<T, static_min_impl<T,t0,t1,t2,t3,t4>::value >
{};

	#elif (CXXOMFORT_CXX_STD < 2011)

template <typename T, T t0, T t1, T t2=t1, T t3=t2, T t4=t3>
struct static_min 
: std::integral_constant< T, static_min< T, static_minmax2<T,t0,t1>::Tmin, t2, t3, t4, t4>::value > 
{ };

template <typename T, T t0, T t1>
struct static_min<T, t0, t1, t1, t1, t1>
: std::integral_constant<T, (t0 <= t1 ? t0 : t1) >
{ };

template <typename T, T t0, T t1, T t2=t1, T t3=t2, T t4=t3>
struct static_max 
: std::integral_constant< T, static_max< T, static_minmax2<T,t0,t1>::Tmax, t2, t3, t4, t4>::value > 
{ };

template <typename T, T t0, T t1>
struct static_max<T, t0, t1, t1, t1, t1>
: std::integral_constant<T, static_minmax2<T,t0,t1>::Tmax >
{ };

	#else // c++11 onwards

template <typename T, T t0, T t1, T... txs>
struct static_min 
: std::integral_constant< T, static_min< T, (t0 <= t1 ? t0 : t1), txs...>::value >
{};

template <typename T, T t0, T t1>
struct static_min<T, t0, t1>
: std::integral_constant<T, (t0 <= t1 ? t0 : t1) >
{ };

template <typename T, T t0, T t1, T... txs>
struct static_max 
: std::integral_constant< T, static_max< T, static_minmax2<T, t0, t1>::Tmax, txs...>::value >
{ };

template <typename T, T t0, T t1>
struct static_max<T, t0, t1>
: std::integral_constant<T, (t1 > t0 ? t1 : t0) >
{ };


	#endif

// 
// safety of addition, multiplication
// 

#if (CXXOMFORT_CXX_STD>=2003)

template <intmax_t n1, intmax_t n2>
struct safety_intmax_add 
: std::integral_constant< bool
, (intmax_t(INTMAX_MAX - n1) > n2) >
{ };

template <intmax_t n1, intmax_t n2>
struct safety_intmax_mul 
: std::integral_constant< bool
, (intmax_t(INTMAX_MAX / n1) > n2) >
{ };

template <uintmax_t n1, uintmax_t n2>
struct safety_uintmax_add 
: std::integral_constant< bool
, (uintmax_t(UINTMAX_MAX - n1) > n2) >
{ };

template <uintmax_t n1, uintmax_t n2>
struct safety_uintmax_mul 
: std::integral_constant< bool
, (uintmax_t(UINTMAX_MAX / n1) > n2) >
{ };


template <typename T, T n1, T n2>
struct safety_integer_add
: std::integral_constant< bool
, (T(integral_limits<T>::const_max  - n1) > n2) >
{ };

template <typename T, T n1, T n2>
struct safety_integer_mul
: std::integral_constant< bool
, (T(integral_limits<T>::const_max  / n1) > n2) >
{ };

#endif

// 
// midpoint
// 

/**
 * @brief  Defines the member @c value , with value of type @a Int that is the numeric value at the middle between @a a and @a b. 
 * @see @ref midpoint() .
 * 
 * */
template <typename Int, Int a, Int b>
struct midpoint
: std::integral_constant<Int, 
	(
	(b < a) ? 
	a - ( ((Int)a-(Int)b)>>1) 
	: 
	a + ( ((Int)b-(Int)a)>>1) 
)>
{ };


// 
// intpow
// 
	#if defined(DOXYGEN_DOC)
/**
 * @brief  Defines the member @c value , of type @a Int and of value equal to @a x to the power of @a y .
 * @cxxoutils Compile-time integer exponentiation, pre-C++11.
 * 
 * @code
 * intmax_t mega = st_math::static_intpow<intmax_t,10,6>::value;
 * @endcode
 * 
 * @note @c static_intpow ignores overflow or underflow on the integral types used, so care must be taken to choose a type @a T that can host the result value.
 * 
 * See also: @c static_intlog , @c static_sqrt .
 * 
 * */
template <typename Int, Int x, unsigned y>
struct static_intpow
: std::integral_constant<
	Int, pow(x,y)
	>
{ };
	#elif (CXXOMFORT_CXX_STD >= 1997)
template <typename T, T x, unsigned y>
struct static_intpow_impl
{
    static const T t = static_intpow_impl<T,x,y/2>::value;
    static const T value = (y%2u==0) ? t*t : x*t*t;
};

template <typename T, T x>
struct static_intpow_impl<T,x,0>
: std::integral_constant<T, 1>
{ };

template <typename T, T x, unsigned y>
struct static_intpow
: std::integral_constant<
	T, static_intpow_impl<T,x,y>::value
	>
{ };

	#else
	#endif // std version


// 
// intlog
// 

template <size_t N, size_t Ba>
struct static_intlog_impl {
	static_assert(Ba > 1, "log base needs to be > 1");
	enum { al4 = (N/Ba/Ba/Ba/Ba > Ba) };
	static const size_t value = N==Ba ? 1 
	: ( al4 
		? 4+static_intlog_impl<N/Ba/Ba/Ba/Ba,Ba>::value 
		: 1+static_intlog_impl<N/Ba,Ba>::value 
	);
};

template <size_t Ba>
struct static_intlog_impl<1,Ba> {
	enum { value= 0 };
};

template <size_t Ba>
struct static_intlog_impl<0,Ba> {
	enum { value= 0 };
};

/**
 * @brief Provides the member @c value with the value of the logarithm of @a N in base @a Ba .
 * @cxxoutils Compile-time integer logarithm on the Naturals, pre-C++11.
 * @tparam Ba the base of the logarithm.
 * @pre <tt>Ba > 1</tt> and <tt>N > 0</tt>.
 * 
 * See also @c static_intpow , @c static_sqrt .
 * 
 * */
template <size_t N, size_t Ba>
struct static_intlog
: std::integral_constant<size_t, static_intlog_impl<N,Ba>::value>
{ };

/*
template <size_t N>
struct static_intlog<N,N>
: std::integral_constant<size_t, 1>
{ };
*/

template <size_t Ba> 
struct static_intlog<1,Ba> 
: std::integral_constant<size_t, 0 > 
{ };



// 
// intsqrt
// 

// adapted from https://baptiste-wicht.com/posts/2014/07/compile-integer-square-roots-at-compile-time-in-cpp.html

#define MID(a, b) ((a+b)/2)
#define POW(a) (a*a)

template<std::size_t res, std::size_t l = 1, std::size_t r = res>
struct ct_sqrt;

template<std::size_t res, std::size_t r>
struct ct_sqrt<res, r, r> : std::integral_constant<std::size_t, r> {};

template <std::size_t res, std::size_t l, std::size_t r>
struct ct_sqrt : std::integral_constant<std::size_t, ct_sqrt<res,
        (POW(MID(r, l)) >= res ? l : MID(r, l)+1),
        (POW(MID(r, l)) >= res ? MID(r, l) : r)>::value> {};

#undef MID
#undef POW
        
template <size_t x, size_t i, size_t f>
struct sqrtcheck
: sqrtcheck<x
	, static_intpow<size_t,midpoint<size_t,i,f>::value,2>::value >= x ? i : midpoint<size_t,i,f>::value+1
	, static_intpow<size_t,midpoint<size_t,i,f>::value,2>::value >= x ? midpoint<size_t,i,f>::value : f
>
{ 
/*
	sqrtcheck () {
		std::cerr<< "[sqrtcheck<"<< x<< ", "<< i<< ", "<< f<< ">]"<< std::endl;
	}
*/
};

template <size_t x, size_t g>
struct sqrtcheck<x, g, g>
: std::integral_constant<size_t, g>
{ };

template <size_t x>
struct static_sqrt 
//: std::integral_constant<size_t, ct_sqrt<x, 1, x>::value>
// we are operating on size_t, so we know the upper bound is either x/2, or half the max size_t
: std::integral_constant<size_t
 , sqrtcheck<x,1,static_min<size_t,x/2,0xffffffff>::value>::value >
{
/*
	static_sqrt() {
		std::cerr<< "[static_sqrt : sqrtcheck<"<< x<< ", 1, "<< static_min<size_t,x/2,0xffffffff>::value<< "> ]"<< std::endl;
		sqrtcheck<x,1,static_min<size_t,x/2,0xffffffff>::value> v;
	}
*/
};

template<>
struct static_sqrt<1>
: std::integral_constant<size_t, 1>
{ };

template<>
struct static_sqrt<0>
: std::integral_constant<size_t, 0>
{ };



// 
// static_gcd
// 


template <uintmax_t i1, uintmax_t i2>
struct static_gcd
: std::integral_constant<uintmax_t, (i2==0U) ? i1 : static_gcd<i2, i1 % i2>::value >
{ };


template <uintmax_t i1>
struct static_gcd<i1,0> 
: std::integral_constant<uintmax_t, i1> 
{ };

template <uintmax_t i1, uintmax_t i2>
struct static_lcm
: std::integral_constant<uintmax_t, i1 / static_gcd<i1,i2>::value * i2 >
{ };


template <intmax_t i1, intmax_t i2>
struct static_gcd_i
: std::integral_constant<intmax_t, (i2==intmax_t(0U)) ? i1 : static_gcd_i<i2, i1 % i2>::value >
{ };

template <intmax_t i1>
struct static_gcd_i<i1,0> 
: std::integral_constant<intmax_t, i1> 
{ };


template <typename T, T i1, T i2>
struct static_gcd_typed_impl
: std::conditional<
	::cxxomfort::traits::is_signed<T>::value , 
	std::integral_constant<T, static_gcd_i<i1,i2>::value > , 
	std::integral_constant<T, static_gcd<i1,i2>::value >
> 
{ };

template <typename T, T i1, T i2>
struct static_gcd_typed
: std::integral_constant < T, static_gcd_typed_impl<T,i1,i2>::type::value >
{ };


// 
// extended GCD 
// 

template <uintmax_t i1, uintmax_t i2, bool b= (i1!=0)>
struct ex_gcd_impl {
	typedef ex_gcd_impl<i2%i1, i1> E2;
	static const uintmax_t g= E2::g;
	static const uintmax_t x= E2::y - i2/i1 * E2::x;
	static const uintmax_t y= E2::x;
};

template <uintmax_t i1, uintmax_t i2>
struct ex_gcd_impl<i1,i2,true> {
	static uintmax_t const g= i2;
	static uintmax_t const x= 0;
	static uintmax_t const y= 1;
};

template <uintmax_t i1, uintmax_t i2>
struct static_ex_gcd
: public ex_gcd_impl<i1,i2> 
{ };


// 
// constexpr-ready functions
// 

	#if (CXXOMFORT_CXX_STD >= 2011)

//! @c constexpr math functions
namespace ce {

//! Returns the absolute value of @a u . See also @c std::abs() .
template <typename T
, typename = std::enable_if <std::numeric_limits<T>::is_specialized > 
>
constexpr 
T abs (T const u) noexcept {
	return (u < T(0) ? T(-u) : T(u) );
}

//! Returns the sign of @a u .
template <typename T
, typename = std::enable_if <std::numeric_limits<T>::is_specialized > 
>
constexpr 
int sign (T const u) noexcept {
	return T(u)==T(0) ? 0 : (T(u)<T(0) ? -1 : +1);
}


/**
 * @brief Returns the integet logarithm of @a u in base @a b .
 * @return T the floor of <tt>(log(x)/log(Ba))</tt> if @a x and @a Ba are positive and <tt>Ba > 1</tt>; fails otherwise.
 * */
template <typename T
, typename = std::enable_if <std::numeric_limits<T>::is_specialized > 
>
constexpr 
T intlog (T const x, unsigned const Ba) noexcept {
	return (Ba > 1 && x > T(0)) 
	? ( x == T(1) ? 0 : x == T(Ba)  
		? 1 : 1 + intlog(x/Ba,Ba)
	)
	: throw cxxomfort::assert_failure("intlog incorrect arguments")
	;
}



/**
 * @}
 * */

} // namespace ce

	#endif
	// c++11

} } // namespace cxxomfort::st_math



#endif
