#ifndef CXXOMFORT_IMPL_STRING_HELPERS_HPP
#define CXXOMFORT_IMPL_STRING_HELPERS_HPP
/**
 * @file
 * @brief Implementation of global string accessors for C++.
 * @ingroup cxxomfort-fixes
 * 
 * See @ref cxxo_nonmember_string_accessors for more informatin
 * 
 */
#include <cxxomfort/base.hpp>
#include <cxxomfort/string.hpp>
#include <cxxomfort/iterator.hpp>
#include <cxxomfort/string_view.hpp>
#include <utility> // pair, in_range, as_const

namespace cxxomfort {  namespace fix {
//! @ingroup cxxomfort-fix
CXXOMFORT_CXX11_CODE(inline,) namespace string_helpers { 

	#define CxxBasicString() std::basic_string<Ch,ChT,A>
	#define CxxBasicStringView() std::basic_string_view<Ch,ChT>

/**
 * @brief Calls <tt>str.clear()</tt>
 * @xrefitem cxxomfort-fixes "" "" Invokes <tt>str.clear()</tt>.
 * 
 * */
template <typename Ch, typename ChT, typename A> inline 
void clear (std::basic_string<Ch,ChT,A>& str) CXXO_NOEXCEPT {
    str.clear();
}

// compare
// compare is done in terms of string_view::compare.

/**
 * @brief Calls <tt>str1.compare(str2)</tt>
 * @xrefitem cxxomfort-fixes "" "" Compares strings as per member <tt>str1.compare(str2)</tt>.
 * 
 * */
template <typename Ch, typename ChT> inline 
int compare (std::basic_string_view<Ch,ChT> const str1
, std::basic_string_view<Ch,ChT> const str2 ) CXXO_NOEXCEPT {
    return str1.compare(str2);
}

//! @overload compare()
template <typename Ch, typename ChT> inline 
int compare (std::basic_string_view<Ch,ChT> const str1, Ch c ) CXXO_NOEXCEPT {
    std::basic_string_view<Ch,ChT> str2(&c, &c+1);
    return str1.compare(str2);
}

//! @overload compare()
template <typename Ch, typename ChT, typename R> inline 
int compare (std::basic_string_view<Ch,ChT> const str1, R const& str2) {
    typedef std::basic_string_view<Ch,ChT> SV;
    return compare(str1,SV(str2));
}

/**
 * @brief Calls <tt>str1.compare(pos,count,str2)</tt>
 * @xrefitem cxxomfort-fixes "" "" Compares strings using <tt>str1.compare(pos,count,str2)</tt>.
 * 
 * */
template <typename Ch, typename ChT> inline
int compare (std::basic_string_view<Ch,ChT> const str1, size_t pos, size_t count, std::basic_string_view<Ch,ChT> const str2 ) CXXO_NOEXCEPT {
    return str1.compare(pos,count,str2);
}

//! @overload compare()
template <typename Ch, typename ChT, typename R> inline
int compare (std::basic_string_view<Ch,ChT> const str1, size_t pos, size_t count, R const& str2 ) CXXO_NOEXCEPT {
    typedef std::basic_string_view<Ch,ChT> SV;
    return str1.compare(pos,count,SV(str2));
}


//! @overload compare()
template <typename Ch, typename ChT, typename A> inline
int compare (std::basic_string<Ch,ChT,A> const& str1, std::basic_string<Ch,ChT,A> const& str2 ) CXXO_NOEXCEPT {
    typedef std::basic_string_view<Ch,ChT> SV;
    return compare(SV(str1), SV(str2));
}

// insert, append
/* 
 * insert has overloads, then there's also push_back and push_front; we reduce them to three overloads:
 * * insert (p1, str1, str2)
 * * insert (p1, str1, c2, str2)
 * * insert (p1, str1, str2, seek2)
 * */

// insert overloads 2,4 in https://en.cppreference.com/w/cpp/string/basic_string/insert

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert character @a c at position @a p1 of string @a str1 .
 * @note Overloads 1, 3 in cppreference.
 * */
template <typename Ch, typename ChT, typename A> inline 
CxxBasicString()& insert 
(CxxBasicString()& str1, size_t p1, Ch c) {
	return str1.insert(p1,c);
}

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert character @a c at front of string @a str1 .
 * @note Overloads 1, 3 in cppreference.
 * */
template <typename Ch, typename ChT, typename A> inline 
CxxBasicString()& insert 
(CxxBasicString()& str1, ::cxxomfort::at_front_t const, Ch c) {
	return str1.insert(0,c);
}

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert character @a c at end of string @a str1 .
 * @note Overload 1 cppreference.
 * */
template <typename Ch, typename ChT, typename A> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, ::cxxomfort::at_back_t const, Ch c) {
	return str1.append(c);
}

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert string @a s2 at position @a p1 of @a str1 .
 * @note Overloads 2, 4, 10 in cppreference.
 * @note Starting with C++17, this inserter uses implicit conversion to @c basic_string_view .
 * */
#if (defined(__cpp_lib_string_view))
template <typename Ch, typename ChT, typename A, typename S2> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, size_t p1, S2 const& str2) {
	return str1.insert(p1,std::basic_string_view<Ch,ChT>(str2));
}
#else
template <typename Ch, typename ChT, typename A, typename S2> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, size_t p1, S2 const& str2) {
	return str1.insert(p1,str2.begin(),str2.end());
}
#endif

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert string @a s2 at the beginning of @a str1 .
 * @note Overloads 2, 4 in cppreference
 * @note Starting with C++17, this inserter uses implicit conversion to @c basic_string_view .
 * */
#if (defined(__cpp_lib_string_view))
template <typename Ch, typename ChT, typename A, typename S2> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, ::cxxomfort::at_front_t const, S2 const& str2) {
	return str1.insert(0,std::basic_string_view<Ch,ChT>(str2));
}
#else
template <typename Ch, typename ChT, typename A, typename S2> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, ::cxxomfort::at_front_t const, S2 const& str2) {
	return str1.insert(std::cbegin(str1),str2.begin(),str2.end());
}
#endif

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert string @a s2 at the end of @a str1 .
 * @note Overload 6 in cppreference
 * @note Starting with C++17, this inserter uses implicit conversion to @c basic_string_view .
 * */
#if (defined(__cpp_lib_string_view))
template <typename Ch, typename ChT, typename A, typename S2> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, ::cxxomfort::at_back_t const, S2 const& str2) {
	return str1.append(std::basic_string_view<Ch,ChT>(str2));
}
#else
template <typename Ch, typename ChT, typename A, typename S2> inline 
std::basic_string<Ch,ChT,A>& insert 
(std::basic_string<Ch,ChT,A>& str1, ::cxxomfort::at_back_t const, S2 const& str2) {
	return str1.append(str2.begin(), str2.end());
}
#endif

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert the elements in the range defined by the iterator pair @c r2 at position @a p1 of @a str1 .
 * @param r2 A @c pair or convertible object containing iterators to the beginning and end of the sequence.
 * @note Overload 8 in cppreference
 * */
template <typename Ch, typename ChT, typename A, typename It> inline 
typename CxxBasicString()::iterator insert 
(CxxBasicString()& str1, typename CxxBasicString()::const_iterator p1, std::pair<It,It> const& r2) {
	return str1.insert(p1, r2.first, r2.second);
}

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert the elements in the range defined by the iterator pair @c r2 at beginning of @a str1 .
 * @param r2 A @c pair or convertible object containing iterators to the beginning and end of the sequence.
 * @note Overload 8 in cppreference
 * */
template <typename Ch, typename ChT, typename A, typename It> inline 
typename CxxBasicString()::iterator insert 
(CxxBasicString()& str1, ::cxxomfort::at_front_t const, std::pair<It,It> const& r2) {
	return str1.insert(0, r2.first, r2.second);
}

/**
 * @xrefitem cxxomfort-fixes "" "String Helpers" Insert the elements in the range defined by the iterator pair @c r2 at end of @a str1 .
 * @param r2 A @c pair or convertible object containing iterators to the beginning and end of the sequence.
 * @note Overload 6 in cppreference
 * */
template <typename Ch, typename ChT, typename A, typename It> inline 
typename CxxBasicString()::iterator insert 
(CxxBasicString()& str1, ::cxxomfort::at_back_t const, std::pair<It,It> const& r2) {
	return str1.insert(std::cbegin(str1), r2.first, r2.second);
}



// replace


// substr 

#if (CXXOMFORT_CXX_STD>=1997)
template <typename Ch, typename ChT, typename A> inline 
CxxBasicString() substr 
(CxxBasicString() const& str1, size_t pos, size_t count= CxxBasicString()::npos) {
    return str1.substr(pos,count);
}

template <typename Ch, typename ChT>
std::basic_string_view<Ch,ChT> 
substr (std::basic_string_view<Ch,ChT> const str1, size_t pos, size_t count= std::basic_string<Ch,ChT>::npos) {
    return str1.substr(pos,count);
}
#else
template <typename Ch, typename ChT, typename A>
std::basic_string<Ch,ChT,A> 
substr (std::basic_string<Ch,ChT,A> const& str1, size_t pos, size_t count = -1 ) {
    return str1.substr(pos,count);
}

template <typename Ch, typename ChT>
std::basic_string_view<Ch,ChT> 
substr (std::basic_string_view<Ch,ChT> const str1, size_t pos, size_t count= -1 ) {
    return str1.substr(pos,count);
}
#endif

// copy 

template <typename Ch, typename ChT, typename A>
typename std::basic_string<Ch,ChT,A>::size_type 
copy (std::basic_string<Ch,ChT,A> const& str1, Ch* dest, size_t count, size_t pos) {
    return str1.copy(dest,count,pos);
}

template <typename Ch, typename ChT>
typename std::basic_string_view<Ch,ChT>::size_type 
copy (std::basic_string_view<Ch,ChT> str1, Ch* dest, size_t count, size_t pos) {
    return str1.copy(dest,count,pos);
}

// find 

/**
 * @brief Calls <tt>str1.find(str2, pos, str2.size() )</tt>
 * @xrefitem cxxomfort-fix "" "" Nonmember string helpers
 * @see @ref cxx11-backports
 * 
 * */
template <typename Ch, typename ChT, typename A>
typename std::basic_string<Ch,ChT,A>::size_type
find (std::basic_string<Ch,ChT,A> const& str1, std::basic_string_view<Ch,ChT> const str2, size_t pos= 0) CXXO_NOEXCEPT {
    return str1.find(str2.data(), pos, str2.size());
}

//! Finds string @a str2 (converted to @c basic_string_view ) in @a str1 starting from index @a pos .
template <typename Ch, typename ChT, typename A, typename R>
typename std::basic_string<Ch,ChT,A>::size_type
find (std::basic_string<Ch,ChT,A> const& str1, R const& str2, size_t pos= 0) CXXO_NOEXCEPT {
    typedef std::basic_string_view<Ch,ChT> SV;
    return str1.find(SV(str2), pos);
}

/**
 * @brief Calls <tt>str1.find(str2, pos)</tt>
 * @xrefitem cxxomfort-fix "" "" Nonmember string helpers
 * @see @ref cxx17-backports
 * 
 * */
template <typename Ch, typename ChT>
typename std::basic_string<Ch,ChT>::size_type
find (std::basic_string_view<Ch,ChT> const str1, std::basic_string_view<Ch,ChT> const str2, size_t pos= 0) CXXO_NOEXCEPT {
    return str1.find(str2, pos);
}

//! Finds string @a str2 (converted to @c basic_string_view ) in @a str1 starting from index @a pos .
template <typename Ch, typename ChT, typename R>
typename std::basic_string<Ch,ChT>::size_type
find (std::basic_string_view<Ch,ChT> const str1, R const& str2, size_t pos= 0) CXXO_NOEXCEPT {
    typedef std::basic_string_view<Ch,ChT> SV;
    return str1.find(SV(str2), pos);
}


// rfind  

	#undef CxxBasicString
	#undef CxxBasicStringView
} //string_helpers
} } // cxxomfort::fix

#endif
