#ifndef CXXOMFORT_UNIQUE_PTR_POOR_HPP
#define CXXOMFORT_UNIQUE_PTR_POOR_HPP


#include "../config.hpp"

#define CXXOMFORT_USING_unique_ptr
#define CXXOMFORT_IMPLEMENTS_unique_ptr CXXO_IMPLSTATUS_EMULATION()

#include "../base.hpp" // nullptr, static_assert, move, forward
#include "../impl/07-type_traits.hpp"
#include <cstddef>
#include <memory>

#if defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1)
    #pragma message CXXO_NOTICE("enabled unique_ptr POOR emulation")
#endif

/*
 * Dependencies from type_traits:
 * is_reference
 * add_reference
 * remove_reference
 * is_const
 * enable_if (provided by meta.hpp)
 */


namespace cxxomfort {
namespace cxxostd {

namespace detail_unique_ptr {

using ::cxxomfort::traits::is_reference;
using ::cxxomfort::traits::add_reference;
using ::cxxomfort::traits::remove_reference;


//
// rv wrapper class
//
template <class T>
class rv {
    T& r_;
public:
    explicit rv(T& r) : r_(r) {}
    T* operator->() {return &r_;}
    T& operator*() {return r_;}
};


template <class T1, class T2, bool E = false>
class unique_ptr_storage {
    private:
    T1 t1_;
    T2 t2_;

    typedef typename add_reference<T2>::type T2_reference;
    typedef typename add_reference<const T2>::type T2_const_reference;

    unique_ptr_storage(const unique_ptr_storage&);
    unique_ptr_storage& operator=(const unique_ptr_storage&);
public:
    operator rv<unique_ptr_storage>() {return rv<unique_ptr_storage>(*this);}

    unique_ptr_storage() : t1_(), t2_() CXXO_NOEXCEPT {}

    explicit unique_ptr_storage(T1 t1)
        : t1_(move(t1)), t2_() {}

    unique_ptr_storage(T1 t1, T2 t2)
        : t1_(move(t1)), t2_(forward<T2>(t2)) {}

          T1& first()       {return t1_;}
    const T1& first() const {return t1_;}

          T2_reference second()       {return t2_;}
    T2_const_reference second() const {return t2_;}
};


template <class T1, class T2, bool b> inline
void
swap(unique_ptr_storage<T1, T2, b>& x, unique_ptr_storage<T1, T2, b>& y) {
    using std::swap;
    swap(x.first(), y.first());
    swap(x.second(), y.second());
}


namespace pointer_type_imp {

	struct one { char x[1]; };
	struct two { char x[2]; };
	template <class U> static two test(...);
	template <class U> static one test(typename U::pointer* = 0);

}  //~namespace pointer_type_imp

template <class T>
struct has_pointer_type {
    static const bool value = sizeof(pointer_type_imp::test<T>(0)) == 1;
};


} // detail_unique_ptr


template <typename T, class D = cxxomfort::cxxostd::default_delete<T> >
class unique_ptr;


template <typename T, typename D>
detail_unique_ptr::rv< unique_ptr<T,D> > move (unique_ptr<T,D>& r) {
	return r;
}

template <typename T, class D>
class unique_ptr {
	
	public:
    typedef T element_type;
    typedef D deleter_type;
    //typedef typename detail_unique_ptr::pointer_type<element_type, deleter_type>::type pointer;
    typedef T* pointer;
	
	private:
	detail_unique_ptr::unique_ptr_storage<T,D,false> ptr_;

	private:
    //CXXO_NONCOPYABLE_MOVABLE(unique_ptr);
    CXXO_DELETED_COPY_CONSTRUCTOR(unique_ptr);
    //CXXO_DELETED_COPY_ASSIGN(unique_ptr);
	
	public:

    typedef typename cxxomfort::traits::add_reference<deleter_type>::type deleter_reference;
    typedef typename cxxomfort::traits::add_reference<const deleter_type>::type deleter_const_reference;

    //! Creates from a rvalue-wrapper (from eg.: @e move ).
    //unique_ptr(detail_unique_ptr::rv<unique_ptr> r)
    
    //unique_ptr( CXXO_RV_REF(unique_ptr) r)
    //: ptr_(r.release(), std::forward<deleter_type>(r.get_deleter())) {
    //}
    

    //! Default constructor.
    unique_ptr() CXXO_NOEXCEPTNOTHROW 
    : ptr_(nullptr) {
        static_assert (!cxxomfort::traits::is_reference<deleter_type>::value, "Trying to use the deleter of a reference type.");
        static_assert (!cxxomfort::traits::is_pointer<deleter_type>::value, "Trying to use the deleter of a pointer type.");
    }

    //! Creates a @c unique_ptr wrapping the allocated pointer @a p .
    explicit unique_ptr(pointer p) 
    : ptr_(p) {
        static_assert (!cxxomfort::traits::is_reference<deleter_type>::value, "Trying to use the deleter of a reference type.");
        static_assert (!cxxomfort::traits::is_pointer<deleter_type>::value, "Trying to use the deleter of a pointer type.");
    }

    // Destructor
    ~unique_ptr() {
        reset();
    }

	unique_ptr& operator= (unique_ptr& r) {
        reset(r.release());
        ptr_.second() = std::move(r.get_deleter());
        return *this;
	}
	
    unique_ptr& operator=(detail_unique_ptr::rv<unique_ptr> r) {
    //unique_ptr& operator=( CXXO_RV_REF(unique_ptr) r) {
        reset(r.release());
        ptr_.second() = std::move(r.get_deleter());
        return *this;
    }



    //! Provide access to member.
    typename cxxomfort::traits::add_reference<T>::type operator*() const {return *get();}
    //! Provides access to member.
    pointer operator->() const {return get();}
    //! Provides access to member.
    pointer get() const {return ptr_.first();}
    //! Provides access to deleter.
    deleter_reference       get_deleter()       {return ptr_.second();}
    deleter_const_reference get_deleter() const {return ptr_.second();}

    //operator int nat::*() const {return get() ? &nat::for_bool_ : 0;}
    CXXO_EXPLICIT_OPERATOR(bool) () const CXXO_NOEXCEPTNOTHROW {
        return this->get() ? true : false;
    }


    //! Resets / reassigns the @c unique_ptr to another allocated pointer.
    void reset (pointer p = pointer()) {
        pointer t = get();
        if (t != pointer())
            get_deleter()(t);
        ptr_.first() = p;
    }

    pointer release() {
        pointer tmp = get();
        ptr_.first() = pointer();
        return tmp;
    }

};

} //~namespace cxxostd
} //~namespace cxxomfort

namespace std {
    using ::cxxomfort::cxxostd::unique_ptr;
    using ::cxxomfort::cxxostd::move;
}


#endif  // guard
