#ifndef CXXOMFORT_IOMANIP_HPP
#define CXXOMFORT_IOMANIP_HPP
/**
 * @file
 * @brief Implementations and additions tied to Standard Header <iomanip>.
 * @xrefitem std0ios "<iomanip>" ""
 *
 */

#include <cxxomfort/base.hpp>
#include <iomanip>
#include <iostream>
#include <cxxomfort/string_view.hpp>

// 
// quoted (for output only)
// 

// for input, see 
// https://github.com/gcc-mirror/gcc/blob/master/libstdc%2B%2B-v3/include/bits/quoted_string.h

#if defined(DOXYGEN_DOC)
namespace std {

/**
 * @brief Outputs a string to a stream, quoted.
 * @cxx14b Inserts a string into a stream, quoted.
 * @xrefitem std0iomanip "" "" Inserts a string into a stream, quoted.
 * @see @cppref{io/manip/quoted}
 * 
 * 
 * */
template <typename Char>
inline implementation_defined_t 
quoted (Char const* str, Char delim='\"', Char escape='\\') {}

//! @overload quoted
template <typename Char>
inline implementation_defined_t 
quoted (std::basic_string_view<Char> str, Char delim='\"', Char escape='\\') {}

}

/**
 * @brief Macro for detection of <tt>std::quoted()</tt>.
 * @ingroup std00featuremacros 
 * @cppfeature{<tt>std::quoted()</tt>}.
 * 
 * Support for @c quoted() is basic and incomplete in cxxomfort (supporting only insertion and not extraction), thus the library sets the feature macro @ref @c __cpp_lib_quoted_string io to the value @c 1 .
 * */
 #define __cpp_lib_quoted_string_io 1


#elif defined(__cpp_lib_quoted_string_io)
#elif (CXXOMFORT_CXX_STD>=1997)
	#if (CXXOMFORT_CXX_STD >= 2011)
		#define __cpp_lib_quoted_string_io 1
	#endif

namespace cxxomfort { namespace cxxostd {
namespace detail {

template <typename Ch, typename ChT = std::char_traits<Ch> >
struct quoted_impl  {
	std::basic_string_view<Ch,ChT> msg;
	Ch const delim;
	Ch const escape;
	
	explicit quoted_impl (std::basic_string_view<Ch,ChT> s, Ch d, Ch e)
	: msg(s), delim(d), escape(e)
	{ }
	
};


// outputs a quoted string
template <typename Ch, typename ChT>
std::basic_ostream<Ch,ChT>& operator<< (std::basic_ostream<Ch,ChT>& os, quoted_impl<Ch,ChT> q) {
	os<< q.delim;
	typedef typename std::basic_string_view<Ch,ChT>::const_iterator I;
	for (I i= q.msg.begin(); i != q.msg.end(); ++i) {
		if (*i==q.escape || *i==q.delim) {
			os<< q.escape;
		}
		os<< *i;
	}
	os<< q.delim;
	return os;
}

} // detail


	#if defined(__cpp_lib_quoted_string_io)
	// add support for string_view pre-C++17
		#if defined(a)
		#else


template <typename Ch, typename ChT>
detail::quoted_impl< Ch, ChT> 
quoted (std::basic_string_view<Ch,ChT> sv, Ch delim=Ch('"'), Ch escape=Ch('\\')) {
	return detail::quoted_impl<Ch,ChT>(sv,delim,escape);
}

static inline detail::quoted_impl<char> 
quoted (char const* str, char delim='"', char escape='\\') {
	return detail::quoted_impl<char>(std::string_view(str), delim, escape);
}

		#endif

	#else

template <typename Ch, typename ChT>
detail::quoted_impl< Ch, ChT> 
quoted (Ch const* sp, Ch delim=Ch('"'), Ch escape=Ch('\\')) {
	return detail::quoted_impl<Ch,ChT>(sp,delim,escape);
}

template <typename Ch, typename ChT, typename ChA>
detail::quoted_impl< Ch, ChT> 
quoted (std::basic_string<Ch,ChT,ChA> const& st, Ch delim=Ch('"'), Ch escape=Ch('\\')) {
	return detail::quoted_impl<Ch,ChT>(st,delim,escape);
}

template <typename Ch, typename ChT>
detail::quoted_impl< Ch, ChT> 
quoted (std::basic_string_view<Ch,ChT> sv, Ch delim=Ch('"'), Ch escape=Ch('\\')) {
	return detail::quoted_impl<Ch,ChT>(sv,delim,escape);
}

	#endif

} }

namespace std {
	using ::cxxomfort::cxxostd::quoted;
}

#endif


/**
 * @page std0iomanip <iomanip>
 * @brief Backports related to Standard Header <tt><iomanip></tt> 
 * 
 * @code
 * #include <cxxomfort/iomanip.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * 
 * @section Interfaces 
 * 
 * Interfaces defined in this section:
 * 
 * * @c quoted() (C++14) (partial) - insert strings into a stream
 * * The @ref @c __cpp_lib_quoted_string_io macro if @c quoted() is implemented.
 * 
 * Non-backport interfaces:
 * 
 * * (none)
 * 
 * @section also See Also
 * 
 * * @see @cppref{header/iomanip} (cppreference)
 * 
 * */

#endif
