#ifndef CXXOMFORT_ITERATOR_HPP
#define CXXOMFORT_ITERATOR_HPP
/**
 * @file
 * @brief Implementations and backports tied to <code>\<iterator\></code>.
 * @ingroup std0iterator 
 *
 */

#include <cxxomfort/base.hpp>
#include <iterator>

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
    #pragma message CXXO_NOTICE("enabled <iterator> support.")
#endif

#define CXXO_COMPILER_SUPPORT_iterator_accessors_20 0

// 
// begin, end (C++11)
// 
#include "base/iterator.hpp"

//
// make_reverse_iterator
// (taken from cppreference.com sample code)
//

#if defined(DOXYGEN_DOC)
/**
 * @ingroup std00featuremacros
 * @cppfeature{@c make_reverse_iterator() }
 * @xrefitem std0iterator "" "" Feature Test macro.
 * */
	#define __cpp_lib_make_reverse_iterator YYYY00L.
namespace std {

/**
 * @brief Constructs the reverse_iterator for a given iterator.
 * @ingroup cxx14-backports 
 * @cxx14b Creates a @c reverse_iterator .
 * @xrefitem std0iterator "" "" Creates a @c reverse_iterator .
 * @see @cppref{iterator/make_reverse_iterator}
 * 
 * If cxxomfort adds this feature, it also sets the @ref @c __cpp_lib_make_reverse_iterator macro .
 * 
 * */
template <typename It>
CXXO_CONSTEXPR14 
std::reverse_iterator<It> make_reverse_iterator (It i) {}

}

#endif
	
namespace cxxomfort { namespace cxxostd { 

template <typename It>
CXXO_CONSTEXPR14 
std::reverse_iterator<It> make_reverse_iterator (It i) {
    return std::reverse_iterator<It>(i);
}

} }


#if (defined( CXXOFLAG_NO_STD_USING ) && CXXOFLAG_NO_STD_USING>0)
#elif (defined(__cpp_lib_make_reverse_iterator))
#else
namespace std {
	using ::cxxomfort::cxxostd::make_reverse_iterator;
}
	#define __cpp_lib_make_reverse_iterator (CXXOMFORT_CXX_STD*100)
#endif


// 
// contiguous_iterator_tag from C++20
// note this is created directly in namespace std
// 

#if defined(DOXYGEN_DOC)
namespace std {
/**
 * @brief A tag type that describes contiguous access iterators.
 * */
struct contiguous_iterator_tag : public random_access_iterator_tag { };
}
#elif (CXXOMFORT_CXX_STD >= 2020)
#elif (defined( CXXOFLAG_NO_STD_USING ) && CXXOFLAG_NO_STD_USING>0)
#else
namespace std {
	struct contiguous_iterator_tag : public random_access_iterator_tag { };
}
#endif

// 
// iterator accessors (C++14) (C++17) (C++20)
// 

#include "impl/14-iterator_access.hpp" // cbegin, cend
#define CXXOMFORT_IMPLEMENTS_iterator_accessors_17 0
#include "impl/17-iterator_access.hpp" // size, data, empty

#define CXXOMFORT_IMPLEMENTS_iterator_accessors_20 0
#if (CXXOMFORT_CXX_STD>=1997)
	#include "impl/20-iterator_access.hpp" // ssize
#endif


// 
// fixes: cdata () 
// 

namespace cxxomfort { namespace fix {

#if defined(DOXYGEN_DOC)
/**
 * @brief Returns <tt> container.data() </tt> as a const-pointer for a given container.
 * @ingroup cxxomfort-fix 
 * @param container A container or expression representing one.
 * @return An iterator of the type of <tt>container.data()</tt>; pre-C++11, a @c Container::const_pointer .
 * @pre The argument implements the <tt>.data()</tt> member with the required semantics (contiguous storage).
 * @xrefitem std0iterator "" "helpers" cdata()
 * */
template <typename T>
auto cdata (C const& container) -> decltype(container.data()) {
	return container.data();
}

#elif (CXXOMFORT_CXX_STD>=2011)
template <typename C>
auto cdata (C const& container) -> decltype(container.data()) {
	return container.data();
}

#else
template <typename C>
typename C::const_pointer cdata (C const& container) {
	return container.data();
}
#endif

} }


// 
// contiguous_access, contiguous_access_iterator
// 

#if defined(DOXYGEN_DOC)

namespace cxxomfort { namespace fix { 

/**
 * @brief Tags an iterator as a contiguous access type iterator.
 * 
 * Given an iterator type I, indicates whether the type either: 
 * 
 * * matches the @c contiguous_iterator_tag tag
 * * or represents access to a contiguous area of memory, like an array
 * 
 * Pre C++17, this can only be implemented as an heuristic and can not 
 * determine contiguousness of storage types where the iterator is 
 * being wrapped, in which cases it will always declare @c false .
 * 
 * One such example is <tt><array></tt> and <tt><vector></tt> iterators 
 * in MSVC, in versions that don't support disabling the debug iterator mode.
 * 
 */
template <typename Iterator, typename Tag= typename std::iterator_traits<Iterator>::iterator_category >
struct is_contiguous_access_iterator {};

} }

#elif (CXXOMFORT_CXX_STD >= 1997)
namespace cxxomfort { namespace fix { 

template <typename Iterator, typename Tag= typename std::iterator_traits<Iterator>::iterator_category >
struct is_contiguous_access_iterator {
	enum { value= 0, specialized= 0 };
};

template <typename Iterator>
struct is_contiguous_access_iterator<Iterator, std::contiguous_iterator_tag> 
: public std::integral_constant<bool, true> {
	enum { specialized= 0 };
};

template <typename T>
struct is_contiguous_access_iterator<T*> 
: public std::integral_constant<bool, true> {
	enum { specialized= 2 };
};

} }
#endif


// 
// default_sentinel (C++20) (for C++11 onward)
// 

#if (CXXOMFORT_CXX_STD >= 2020) 
#elif (CXXOMFORT_CXX_STD >= 2011 || CXXOMFORT_CXX_EMULATION >= 2011)
namespace std { 
	struct default_sentinel_t { };
	CXXO_CONSTEXPR14 const default_sentinel_t default_sentinel = {};
}
#endif

// 
// fixes: constexpr next, prev
// 

namespace cxxomfort { namespace fix {
// for next we require a pointer-like iterator
namespace detail {
}

#if (defined(DOXYGEN_DOC) || \
 (CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_constexpr) \
 )
	#include <type_traits>

/**
 * @brief Given a compile-time  @c const (or @c constexpr ) iterator, returns the iterator to the next position.
 * @xrefitem std0iterator "helpers" static_next
 * */
template <typename It, typename= typename std::enable_if<std::is_pointer<It>::value, bool> >
constexpr It static_next (It const p) noexcept {
	return p+1;
}

/**
 * @brief Given a compile-time  @c const (or @c constexpr ) iterator, returns the iterator to the previous position.
 * @xrefitem std0iterator "" "helpers" static_prev
 * */
template <typename It, typename= typename std::enable_if<std::is_pointer<It>::value, bool> >
constexpr It static_prev (It const p) noexcept {
	return p-1;
}

#endif

} } // cxxomfort::fix


#endif // file

/**
 * @page std0iterator <iterator>
 * @brief Backports related to Standard Header <tt><iterator></tt> 
 * 
 * @code
 * #include <cxxomfort/iterator.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Interfaces defined in this section:
 * 
 * * C++11 @ref nonmember_iterator_accessors such as @c begin() , @c end() .
 * * C++14 @ref nonmember_iterator_accessors such as @c rbegin() .
 * * @c make_reverse_iterator() (C++14).
 * * C++17 @ref nonmember_iterator_accessors such as @c size() .
 * * @c contiguous_iterator_tag (C++20).
 * * The @ref @c __cpp_lib_make_reverse_iterator macro if @c make_reverse_iterator() is implemented.
 * * The @ref @c __cpp_lib_nonmember_container_access macro if <tt>size(), data(), empty()</tt> are all implemented.
 * 
 * Non-backport interfaces (<tt>cxxomfort::fix</tt>):
 * 
 * * @link cxxomfort::fix::is_contiguous_access_iterator <tt>is_contiguous_access_iterator</tt> @endlink - A trait check for contiguous access iterators.
 * * @link cxxomfort::cxxostd::cdata() @c cdata() accessor @endlink , to supplement <tt>std::data()</tt>
 * * <tt>static_next()</tt>, <tt>static_prev()</tt> for @c constexpr -pointer iterator (C++11 onwards).
 *
 * @section nonmember_iterator_accessors Nonmember <iterator> Accessors
 * 
 * C++11 onward have introduced a number of non-member accessor functions for iterator and container elements in C++. These include global begin/end functions to access arrays, containers and sequences, and a number of accessor functions such as <tt>size()</tt> that checks for a container, array or sequence's reported size.
 * 
 * <b>Begin / End accessors</b>
 * 
 * <tt> begin() / end() </tt> : the invocation <tt>begin(sequence)</tt> returns:
 * 
 * * an automatically deduced type in C++>=11
 * * <tt>Sequence::iterator</tt> or <tt>T*</tt> in C++<11
 * * <tt>sequence.begin()</tt> when it exists
 * * <tt>&(sequence[0])</tt> when @a sequence is a native array with type <tt>T(&)[N]</tt>
 * 
 * <tt> rbegin() / rend() </tt> : the invocation <tt>rbegin(sequence)</tt> returns:
 * 
 * * an automatically deduced type in C++>=11
 * * <tt>Sequence::reverse_iterator</tt> in C++<11
 * * <tt>sequence.rbegin()</tt> when it exists
 * * <tt>&(sequence[N-1])</tt> when @a sequence is a native array with type <tt>Type(&)[N]</tt>
 * 
 * <tt> cbegin() / cend() / crbegin() / crend() </tt> : they return the <tt>const_iterator</tt> or <tt>const_pointer</tt> (or <tt>T const*</tt>) version of the above.
 * 
 * <b>Container Accessors</b>
 * 
 * C++17 added the following functions:
 * 
 * <tt>size()</tt> : the invocation <tt>size(sequence)</tt> returns:
 * 
 * * An automatically deduced type in C++>=11
 * * <tt>Sequence::size_type</tt> or <tt>size_t</tt> in C+++<11
 * * <tt>sequence.size()</tt> when it exists
 * * <tt>std:distance( sequence.begin(), sequence.end() )</tt> when a member size does not exist (see note below)
 * * <tt>N</tt> when @a sequence is a native array with type <tt>T(&)[N]</tt>
 * 
 * <tt>empty()</tt> : the invocation <tt>empty(sequence)</tt> returns:
 * 
 * * A <tt>bool</tt> value
 * * <tt>sequence.empty()</tt> when it exists
 * * <tt>0</tt> for arrays
 * 
 * <tt>data()</tt> : the invocation <tt>data(sequence)</tt> returns:
 * 
 * * A <tt>T*</tt> (or <tt>T const*</tt>) pointer value
 * * <tt>sequence.data()</tt> when it exists
 * * <tt>sequence</tt> (decayed to pointer) for arrays
 * 
 * @note cxxomfort provides size() for objects that define begin and end, such as <tt>forward_list</tt>. As per cppreference, "Custom overloads of size may be provided for classes and enumerations that do not expose a suitable size() member function, yet can be detected", so these overloads are permitted.
 * 
 * @section contiguous_access_iterator Contiguous Access Iterator Tag
 * 
 * Previous to C++20, cxxomfort defines, but can't use, the iterator tag @c contiguous_iterator_tag . This is a refinement over @c random_access_iterator_tag used in newer Standards to point out to iterator types that refer to contiguous memory storage (such as raw pointers).
 * 
 * Because it's not possible to redefine member types in std containers and iterators, at present only the following features in cxxomfort make use of the @c contiguous_iterator_tag type:
 * 
 * * @c cxxomfort::iterator_helper
 * * @ref cxxomfort::array as the category of its iterator type
 * * @ref cxxomfort::array_ref as the category of its iterator type
 * * @c basic_string_view when backported
 * * @ref cxxomfort::library::fixed_vector::fixed_vector as the category of its iterator type
 * 
 * 
 * @section also See Also
 * 
 * @see @cppref{header/iterator} (cppreference)
 * @see @cppref{iterator/begin} @cppref{iterator/end}
 * @see @ref cxxo-sup-iterator "cxxomfort <iterator> supplements"
 * 
 * */
