#ifndef CXXOMFORT_SUPP_ALGORITHM_HPP
#define CXXOMFORT_SUPP_ALGORITHM_HPP
#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <cxxomfort/algorithm.hpp>
#include <cxxomfort/functional.hpp> // bind(), not_fn()
#include <algorithm>
#include <iterator>

/**
 * @file 
 * @brief cxxomfort Supplementals for <algorithm>
 * 
 */

#include <cxxomfort/impl/valcomp.hpp>


namespace cxxomfort { namespace library { 
namespace algorithm {

/**
 * @ingroup cxxo-sup-algorithm
 * @{
 */

using ::cxxomfort::impl::valcmp;

// 
// swap_if (p2187)
// 

#undef  CXXOMFORT_LIB_swap_if
#include <cxxomfort/impl/p2187-swap_if.hpp>
#if (CXXOMFORT_IMPLEMENTS_p2187 > 0 && CXXOMFORT_IMPLEMENTS_p2187<CXXO_IMPLSTATUS_NATIVE())
	#define CXXOMFORT_LIB_swap_if (CXXOMFORT_CXX_STD*100)
namespace cxxomfort { namespace library { namespace algorithm {
	using cxxomfort::impl::swap_if;
} } }
#if (!defined(CXXO_LIB_swap_if))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c swap_if() proposal / @c p2187 ..
	//! @note @c swap_if() is also made available by @ref cxxomfort-utils under the @c cxxomfort:: namespace.
	#define CXXO_LIB_swap_if 202001
#endif

#else
#endif

// 
// is_clamped (p1440)
// 

#undef  CXXOMFORT_IMPLEMENTS_p1440
#define CXXOMFORT_IMPLEMENTS_p1440 CXXO_IMPLSTATUS_BACKPORT()

/**
 * @brief Checks if a value is clamped into a range.
 * @param less A relational <tt>operator<()</tt>-like predicate.
 * @returns @c true if @a value is within the constrained region.
 * @xrefitem cxxo-sup-algorithm "" "" Check if a value is clamped into the interval <tt>[lo,hi]</tt>.
 * @see @wg21{p1440} , @cppref{algorithm/clamp}
 * 
 * Given general invocations of <tt>clamp()</tt>, the following assertion should hold:
 * 
 * @code
 * (is_clamped(x, lo, hi)) == (x == clamp(x,lo,hi))
 * @endcode
 * 
 * @warning The behaviour is undefined if <tt>hi < lo</tt>.
 * 
 * @note Unlike the @e p1440 proposal, <tt>is_clamped()</tt> here is marked @c noexcept ; I consider there is no good scenario where relational comparison operators, or anything that mimics them, would make use of exception, and @c clamp() is to abort rather than throw upon precondition violation.
 * 
 * */
template <typename T, typename Compare> inline CXXO_CONSTEXPR 
bool 	is_clamped (T const& value, T const& lo, T const& hi, Compare less) CXXO_NOEXCEPTNOTHROW {
	return value == std::clamp(value,lo,hi,less);
}

//! @overload is_clamped()
template <typename T> inline CXXO_CONSTEXPR 
bool 	is_clamped (T const& value, T const& lo, T const& hi) CXXO_NOEXCEPTNOTHROW {
	return value == std::clamp(value,lo,hi);
}

#undef  CXXOMFORT_IMPLEMENTS_p1440
#define CXXOMFORT_IMPLEMENTS_p1440 202202U

#if (!defined(CXXO_LIB_is_clamped))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c is_clamped() from @c p1440 .
	#define CXXO_LIB_is_clamped 202209
#endif


// 
// half-open range clamp variants
// 

/**
 * @brief Clamps a value into a half-open interval.
 * @param less A relational comparison predicate.
 * @return @a val , constrained to the half-open interval <tt>[@a lo , @a hi )</tt>.
 * @xrefitem cxxo-sup-algorithm "" "" Clamps a value into a half-open interval <tt>[lo,hi)</tt>.
 * @see @cppref{algorithm/clamp}
 * 
 * <tt>clamp_halfopen(V,a,b)</tt> is similar to <tt>clamp(V,a,b)</tt> except that @a b is excluded from the interval.
 * 
 * @warning The behaviour is undefined if <tt>hi < lo</tt>.
 * 
 */
template<class T, class Compare> inline CXXO_CONSTEXPR 
T const& 	clamp_halfopen ( T const& val, T const& lo, T const& hi, Compare less) 
CXXO_NOEXCEPT {
    return assert( !less(hi, lo) ),
        less(val, lo) ? lo : less(val, hi) ? val : hi;
}

//! @overload clamp_halfopen()
template<class T> inline CXXO_CONSTEXPR 
T const& 	clamp_halfopen ( T const& val, T const& lo, T const& hi) 
CXXO_NOEXCEPT {
    return assert( !(hi< lo) ),
        (val< lo) ? lo : (val< hi) ? val : hi;
}

/**
 * @brief Checks if a value is clamped in a half-open interval
 * @xrefitem cxxo-sup-algorithm "" "" Check if a value is clamped into a half-open interval <tt>[lo,hi)</tt> as per @c clamp_halfopen() .
 * 
 * This is a supplement to p1440 "is_clamped" in that it checks if the value 
 * is clamped in a @e half-open interval.
 * 
 * Given @c clamp() , @c clamp_halfopen() , the following assertion should hold:
 * 
 * @code
 * (is_clamped_halfopen(x, lo, hi)) == (x == clamp_halfopen(x,lo,hi))
 * @endcode
 * 
 * @warning The behaviour is undefined if <tt>hi < lo</tt>.
 * 
 * @note <tt>is_clamped_halfopen()</tt> here is marked @c noexcept ; 
 * I consider there is no good scenario where relational comparison operators, or 
 * anything that mimics them would throw exceptions.
 * 
 * */
template <typename T, typename Compare> inline CXXO_CONSTEXPR 
bool 	is_clamped_halfopen (T const& value, T const& lo, T const& hi, Compare less) CXXO_NOEXCEPTNOTHROW {
	return value == clamp_halfopen(value, lo, hi, less);
}

//! @overload is_clamped_halfopen()
template <typename T> inline CXXO_CONSTEXPR 
bool 	is_clamped_halfopen (T const& value, T const& lo, T const& hi) CXXO_NOEXCEPTNOTHROW {
	return !(value < lo) && (value < hi);
}

#if (!defined(CXXO_LIB_clamp_halfopen))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of half-open interval @c clamp supplements.
	#define CXXO_LIB_clamp_halfopen 202209
#endif

}


// 
// copy_ supplementals
// 

namespace algorithm {

/**
 * @brief Copy like @c copy_if with inverted predicate (for symmetry with @c find_if_not ).
 * @xrefitem cxxo-sup-algorithm "" "" Copy values as per @c copy_if() with a negated predicate.
 * @see @cppref{algorithm/copy_if}
 * 
 * This is provided for symmetry with:  <tt>std::find_if_not()</tt>, <tt>std::copy_if()</tt>.
 * 
 * @note This is marked @c constexpr starting with C++17.
 * 
 */
template <typename OutI, typename InpI, typename Pred> inline CXXO_CONSTEXPR17 
OutI copy_if_not (InpI ini, InpI fin, OutI dest, Pred p) {
    for (; ini != fin; ++ini) { 
        if (!p(*ini)) { *dest= *ini; ++dest; }
    }
}

/**
 * @brief Like @c copy_n in that it copies @e leftmost @a ("starting") n elements from the sequence.
 * @xrefitem cxxo-sup-algorithm "" "" Copy leftmost @a n values as per @c copy_n() .
 * 
 * This function copies the first @a n elements from the sequence starting at @a ini1. While it is basically equivalent to @c copy_n() , it is provided for symmetry and completeness with @c copy_rightmost_n() .
 * 
 * @see @cppref{algorithm/copy_n} 
 * @see @c copy_rightmost_n() .
 * 
 * */
template <typename Iter1, typename Integer, typename Iter2>
Iter2 copy_leftmost_n (Iter1 ini1, Iter1 , Integer n, Iter2 ini2) {
    return copy_n(ini1,n,ini2);
}

namespace detail {

template <typename It1, typename Integer, typename It2>
It2 copy_rightmost_n (It1 ini1, It1 fin1, Integer n, It2 ini2, std::forward_iterator_tag) {
    using namespace std;
    Integer i= 0;
    It1 tra= ini1;
    for (; ini1!=fin1 && i < n; ++ini1, ++i) {}
    // at this point if ini==fin, we reached the end and we have to copy everything
    // otherwise we can keep moving to the end with a trailing iterator and copy from there.
    if (ini1==fin1) {
        return copy_n(ini1,n,ini2);
    } else {}
    
    for (; ini1!=fin1; ++ini1, ++tra) {}
    return copy_n(tra, n, ini2);
}

template <typename It1, typename Integer, typename It2>
It2 copy_rightmost_n (It1 ini1, It1 fin1, Integer n, It2 ini2, std::bidirectional_iterator_tag) {
    using namespace std;
    if ((distance(ini1,fin1)) < n) {
        return copy_n(ini1,n,ini2);
    } else {}

    return copy_n(prev(fin1,n),n,ini2);
}

} // detail

/** 
 * @brief Copy the rightmost @a n elements from a sequence.
 * @return Iterator pointing to the end of the copied sequence at the destination.
 * @xrefitem cxxo-sup-algorithm "" "" Copy rightmost @a n elements from a sequence.
 * 
 * @see copy_leftmost_n() 
 * 
 */
template <typename It1, typename Integer, typename It2>
It2 copy_rightmost_n (It1 ini1, It1 fin1, Integer n, It2 ini2) {
    const typename std::iterator_traits<It1>::iterator_category cat;
    return detail::copy_rightmost_n(ini1, fin1, n, ini2, cat);
}

// 
// find_ supplements
// 

/** 
 * @brief Finds an element @em x in a sequence such that @e x satisfies <tt>if_(x,v)</tt>, as per LWG217.
 * @anchor lwg217_find
 * @param if_ A predicate taking two arguments, where the second is bound to parameter @a v .
 * @return Iterator pointing to the found element, or else @a fin .
 * @xrefitem cxxo-sup-algorithm "" "" Finds an element @a x in a sequence such that it satisfies <tt>if_(x,v)</tt>, as per LWG217.
 * 
 * This overload to the name @c find is taken from proposal <b>lwg217</b>, which enhances the interface for <tt>find()</tt> functions.
 * 
 * @see @wg21{lwg217}
 * @see <tt>std::find_if</tt>.
 * 
 */
template <typename It, typename Pred, typename T>
It find (It ini, It fin, Pred if_, T const& v) {
	return std::find_if(ini,fin, std::bind(if_,std::placeholders::_1,v) );
}
#define CXXOMFORT_IMPLEMENTS_lwg217 202209L

#if (!defined(CXXO_LIB_find_predicate2))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c find() with bivariate predicate, from @c lwg217 .
	#define CXXO_LIB_find_predicate2 202209L
#endif

/**
 * @brief Finds the first element in the sequence not equal to @e t .
 * @return An iterator to the found element, or else @e fin .
 * @xrefitem cxxo-sup-algorithm "" "" Find an element not equal to @a t .
 * @see @c find() 
 * 
 * This is provided for homology with: 
 * <tt>std::find()</tt>, <tt>std::find_if_not()</tt>.
 * 
 * */
template <typename It, typename T>
It find_not (It ini, It fin, T const& t) {
	for (; ini != fin; ++ini) {
		if (t==*ini) {} else { return ini; }
	}
	return fin;
}

/**
 * @brief Finds an element @c r such that @a r does not match @a v in <tt>if_(r,v)</tt>.
 * @param if_ A predicate taking two arguments, such that one of them will be picked from the sequence to evaluate.
 * @xrefitem cxxo-sup-algorithm "" "" Find an item @e not matching a predicate with a value.
 * 
 * This function is the inverse of <tt>find(It,It,Pred,T)</tt> 's 4-argument from and is provided as a supplement to LWG217.
 * 
 * */
template <typename It, typename Pred, typename T>
It find_not (It ini, It fin, Pred if_, T const& v) {
	return std::find_if_not(ini, fin, std::bind(if_, std::placeholders::_1, v) );
}

namespace detail {

template <typename Iter, typename V>
Iter find_last (Iter ini, Iter fin, V const& v, std::forward_iterator_tag const) {
	Iter p= fin;
	while (ini != fin) {
		if (*ini==v) p=ini;
		++ini;
	}
	return p;
}

template <typename Iter, typename V>
Iter find_last (Iter ini, Iter fin, V const& v, std::bidirectional_iterator_tag const) {
    Iter ret(fin);
    for (; ini != fin; --fin) {
        Iter t= std::prev(fin,1);
        if (*t==v) { return t; }
    }
    return ret;
}

template<class Iter, class Pred>
Iter find_last_if (Iter ini, Iter fin, Pred f, std::forward_iterator_tag) {
    Iter ret(fin);
    for (; ini != fin; ++ini) {
        if (f(*ini)) { ret= ini; }
    }
    return ret;
}

template<class Iter, class Pred>
Iter find_last_if (Iter ini, Iter fin, Pred f, std::bidirectional_iterator_tag) {
    Iter ret(fin);
    for (; ini != fin; --fin) {
        Iter t= std::prev(fin,1);
        if (f(*t)) { return t; }
    }
    return ret;
}

} // detail

/**
 * @brief Finds the latest item in a sequence that matches a value.
 * @param ini,fin Delimiters to the sequence to be examined.
 * @return An iterator pointing to the latest item in <code>[ini,fin)</code> that fulfills @c if_ .
 * 
 * This is provided for homology with:
 * <tt>std::find()</tt>, <tt>std::find_end()</tt>.
 * 
 */
template <typename Iter, typename T>
Iter find_last (Iter ini, Iter fin, T const& t) {
	return detail::find_last(ini, fin, t, typename std::iterator_traits<Iter>::iterator_category());
}

/**
 * @brief Finds the latest item in a sequence that fits predicate @c if_ .
 * @param ini,fin Delimiters to the sequence to be examined.
 * @param if_ A predicate (a function from <var>ini...fin</var>'s alue type to @c bool ).
 * @return An iterator pointing to the latest item in <code>[ini,fin)</code> that fulfills @c if_ .
 * 
 * This is provided for homology with:
 * <tt>std::find_if()</tt>, <tt>std::find_end()</tt>.
 * 
 */
template <typename Iter, typename Pred>
Iter find_last_if (Iter ini, Iter fin, Pred fn) {
	return detail::find_last_if(ini, fin, fn, typename std::iterator_traits<Iter>::iterator_category());
}

/**
 * @brief Finds the last element @c r such that @a r does match @a v in <tt>if_(r,v)</tt>.
 * 
 * This is provided for completeness as a supplement to lwg217's <tt>find()</tt> four-argument form.
 * 
 * */
template <typename Iter, typename Pred, typename T>
Iter find_last (Iter ini, Iter fin, Pred if_, T const& v) {
	return find_last_if(ini, fin, std::bind(if_, std::placeholders::_1, v) );
}


//! @internal
namespace detail {

template <typename It, typename V>
It find_last_not (It ini, It fin, V const& v, std::forward_iterator_tag) {
    It ret(fin);
    for (; ini != fin; ++ini) {
        if (*ini!=v) { ret= ini; }
    }
    return ret;
}

template <typename It, typename V>
It find_last_not (It ini, It fin, V const& v, std::bidirectional_iterator_tag) {
    It ret(fin);
    for (; ini != fin; --fin) {
        It t= prev(fin,1);
        if (*t!=v) { return t; }
    }
    return ret;
}

template <typename It, typename Pred>
It find_last_if_not (It ini, It fin, Pred if_, std::forward_iterator_tag) {
    It ret(fin);
    for (; ini != fin; ++ini) {
        if (!if_(*ini)) { ret= ini; }
    }
    return ret;
}


} // .detail::

/**
 * @brief Finds the latest item in a sequence different from a value.
 * @param ini,fin Delimiters to the sequence to be examined.
 * @return An iterator pointing to the latest item in <code>[ini,fin)</code> that fulfills @c if_ .
 * 
 * This is provided for homology with:
 * <tt>std::find</tt>. <tt>std::find_end</tt>-
 * 
 */
template <typename Iter, typename T>
Iter find_last_not (Iter ini, Iter fin, T const& t_) {
    const typename std::iterator_traits<Iter>::iterator_category cat;
    return detail::find_last_not(ini, fin, t_, cat);
}

template <typename Iter, typename Pred>
Iter find_last_if_not (Iter ini, Iter fin, Pred if_) {
    const typename std::iterator_traits<Iter>::iterator_category cat;
    return detail::find_last_if_not(ini, fin, if_, cat);
}


// 
// count_ supplementals
// 

/**
 * @brief Count the number of elements in a sequence @e not matching a value @a v .
 * @see LWG217 @c find() , <tt>std::count()</tt>
 * @xrefitem cxxo-sup-algorithm "" "" Count the number of items @e not matching a value.
 * 
 * This function is the inverse to <tt>std::count()</tt> and is provided here for completeness of the interface compared to eg.: <tt>find()</tt> functions.
 * 
 * */
template <typename It, typename T>
typename std::iterator_traits<It>::difference_type 
count_not (It ini, It fin, T const& v) {
	typename std::iterator_traits<It>::difference_type c= 0;
	for (; ini!=fin; ++ini) {
		if ( *ini==v ) {} else {++c;}
	}
	return c;
}

template <typename It, typename Pred>
typename std::iterator_traits<It>::difference_type 
count_if_not (It ini, It fin, Pred if_) {
	typename std::iterator_traits<It>::difference_type c= 0;
	for (; ini!=fin; ++ini) {
		if ( if_(*ini) ) {} else {++c;}
	}
	return c;
}


} // algorithm

// 
// swap, inversion supplements
// 

namespace algorithm {

/*
 * @brief Finds the first element in the sequence not fitting predicate @e if_ .
 * @return An iterator, or @e fin .
 * @xrefitem cxxo-sup-algorithm "cxxo-sup-algorithm" "Searching"
 * @see @c find_if() 
 * *
template <typename It, typename Pred>
It find_if_not (It ini, It fin, Pred if_) {
	return std::find_if(ini,fin,std::not1(if_));
}
*/

/** 
 * @brief Finds a pair of adjacent elements out of order in a sequence (ie.: the second one is "less" than the first).
 * @return Iterator pointing to the first such element, or @a fin if none.
 * @xrefitem cxxo-sup-algorithm "" "Sorting" Finds contiguos elements out of sorting order in a sequence.
 * 
 * For example, for the following example @c find_inversion will return an iterator pointing to the element 6:
 * 
 * @code
vector<int> elems = { -17, 3, 6, -2, -3, -4 };
//                            ^  sorting order is broken in this pair
vector<int>::iterator br = find_inversion( begin(elems), end(elems) );
// at this point, br points to index 2 (element 6).
 * @endcode
 * 
 * See also: @c fix_inversion().
 * 
**/
template <typename Iter, typename Less> inline
Iter find_inversion (Iter ini, Iter fin, Less lt) {
    for (; ini != fin && std::next(ini) != fin; ++ini) {
		Iter nx= std::next(ini);
		if (nx==fin) return fin;
        else if (lt(*nx,*ini)) break;
    }
    return ini;
}

//! @overload find_inversion 
template <typename Iter> inline
Iter find_inversion (Iter ini, Iter fin) {
    std::less<typename Iter::value_type> const m;
    return find_inversion(ini, fin, m);
}

/** 
 * @brief Fixes a pair of misplaced elements where the second one is "less" than the first.
 * @return Iterator pointing to the first such element, or @a fin if none.
 * @xrefitem cxxo-sup-algorithm "" "Sorting" Fixes the ordering of two contiguous elements in a sequence.
 * 
 * See also: @c find_inversion().
 * 
**/
template <typename Iter, typename Less> inline
Iter fix_inversion (Iter ini, Iter fin, Less m) {
    Iter inv = find_inversion(ini,fin,m);
    if (inv!=fin) {
        std::iter_swap(inv, std::next(inv));
    }
    return inv;
}



//! count_while proposal
//! https://groups.google.com/a/isocpp.org/forum/?fromgroups#!topic/std-proposals/bXrQrkBw59c
template <class InputIterator, class UnaryPredicate>
typename std::iterator_traits<InputIterator>::difference_type
count_while (InputIterator first, InputIterator last, UnaryPredicate pred) {
    typename std::iterator_traits<InputIterator>::difference_type ret = 0;
    while (first!=last && pred(*first)) {
        ++ret;
        ++first;
    }
    return ret;
}

}

// 
// transform_ supplementals
// 

namespace algorithm {

/**
 * @brief Transform elements in a sequence in-place.
 * @param ini,fin Iterators denoting a @b writable sequence.
 * @param tf A transformation of the form <tt>TF( v ) → v</tt> where @a v is the type of the elements in the sequence.
 * @return The @e end iterator of the sequence.
 * @xrefitem cxxo-sup-algorithm "" "Transforms" Transforms elements in a sequence in-place.
 */
template <typename FIter, typename TF>
FIter transform_inplace (FIter ini, FIter fin, TF tf) {
    for (; ini != fin; ++ini) {
        *ini= tf(*ini);
    }
    return ini;
}

/**
 * @brief Transform elements in the range conditionally.
 * @param ini,fin Iterators marking a @e writable sequence.
 * @param tf A transformation of the form <code>TF(*FIter) -> *FIter</code>.
 * @param p a predicate to decide if the elements need transformation.
 * @return The @e end iterator of the sequence.
 * @xrefitem cxxo-sup-algorithm "" "Transforms"
 * 
 * This is provided for homology with: 
 * <tt>std::copy_if</tt>
 * 
 */
template <typename FIter, typename TF, typename P>
FIter transform_inplace_if (FIter ini, FIter fin, TF tf, P p) {
    for (; ini != fin; ++ini) {
        if (p(*ini)) *ini= tf(*ini);
    }
    return ini;
}

#if (!defined(CXXO_LIB_transform_inplace))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c cxxomfort::library::algorithm::transform() in-place (3-arg) version.
	#define CXXO_LIB_transform_inplace 202112
#endif

/**
 * @brief Transform elements up to n times
 * @param tf A transformation of the form <code>TF(*FIter) -> *FIter</code>.
 * @xrefitem cxxo-sup-algorithm "" "Transforms"
 * 
 * This is provided for homology with: 
 * <tt>std::copy_n</tt>.
 * 
 */
template <typename IIterator, typename OIterator, typename TF, typename Integer>
inline OIterator transform_n (IIterator ini, Integer n, OIterator dest, TF tf) {
    for (Integer i(0); i < n; ++i) { *dest= tf(*ini); }
    return dest;
}


// 
// histogramy
// 

/**
 * @brief Constructs a map of frequencies each element appears in the sequence.
 * 
 * Given a sequence S=[ini,fin), and a map @em vi of elements of type @em X 
 * with an associated mapping function @em map_int such that: 
 * @code
f "map_int" : (X) → [0, 1, ..., k]
 * @endcode
 * 
 * Then it iterates over S and increments the value at
 * <tt>vi+f(x)</tt> .
 * */
template <typename Iterator, typename ValueMapIterator, typename Mapping>
Iterator count_frequencies_map (Iterator ini, Iterator fin, ValueMapIterator vi, Mapping const& map_int) {
    for (; ini != fin; ++ini) {
        *(vi + map_int(*ini))++;
    }
    return ini;
}


/**
 * @brief Constructs a map of frequencies each element appears in the sequence.
 * @overload count_frequencies_map
 * 
 * This overload takes an array @em v of frequency holders (a bucket), 
 * accessible with <tt>operator[]</tt> for unchecked access, 
 * instead of an iterable map and a mapping function.
 * 
 * Then it iterates over S and increments the value at
 * <tt>v[f(x)]</tt> .
 * */
template <typename Iterator, typename ValueMap>
Iterator count_frequencies (Iterator ini, Iterator fin, ValueMap& v) {
    for (; ini != fin; ++ini) {
        v[*ini]++;
    }
    return ini;
}

#if (CXXOMFORT_CXX_STD>= 2011)
/**
 * @brief Checks if any of the predicates checks for @a value .
 * @ingroup cxxo-sup-algorithm
 * @return @c true if any of [ini,...,fin) evaluates @c true for @a value .
 * */
template <typename T, typename FncIter>
bool any_on (FncIter ini, FncIter fin, T const& value) {
	for (; ini != fin; ++ini) {
		if ((*ini)(value)) return true;
	}
	return false;
}

/**
 * @brief Checks if all of the predicates checks for @a value .
 * @ingroup cxxo-sup-algorithm
 * @return @c false if any of [ini,...,fin) evaluates @c false for @a value .
 * */
template <typename T, typename FncIter>
bool all_on (FncIter ini, FncIter fin, T const& value) {
	for (; ini != fin; ++ini) {
		if ((*ini)(value)!=true) return false;
	}
	return true;
}

/**
 * @brief Checks if none of the predicates checks for @a value .
 * @ingroup cxxo-sup-algorithm
 * @return @c true if all of [ini,...,fin) evaluates @c true for @a value .
 * */
template <typename T, typename FncIter>
bool none_on (FncIter ini, FncIter fin, T const& value) {
	for (; ini != fin; ++ini) {
		if ((*ini)(value)) return false;
	}
	return true;
}


#endif

/*
 * fwsort
 * sort but for non-random access iterators. Requires fw-iterators.
 */

/*
template <typename FwIter, typename Less>
void fwsort (FwIter ini, FwIter fin, Less l) {
    using namespace std;
    if (ini==fin) return;
    FwIter p= ini;
    for (; p != fin; ++p) {
        FwIter m= min_element(p, fin, l);
        if (!(p==m)) iter_swap(p,m);
    }
}
*/


/**
 * @}
 */


template <typename Cont>
struct inserter_status {
};

namespace detail {

// invokes C.insert(p, v) if the member exists

} // detail

namespace lfv2 { // library fundamentals v2

/**
 * @brief Invokes erase-remove idiom on a container to remove element @a u as per @c remove() .
 * @ingroup cxxo-sup-algorithm
 * @param C a container object (lvalue).
 * @param u the value to pass to @c C.remove() .
 * @throws exception Whatever the expression <tt>C.erase(...)</tt> throws.
 * @see @cppref{experimental/lib_extensions_2}
 * 
 * */
template <typename Container>
void erase (Container& C, typename Container::value_type const& u) {
    using namespace std;
    C.erase( std::remove (begin(C), end(C), u) , end(C) );
}

/**
 * @brief Invokes erase-remove idiom on a container to remove elements conditionally given predicate @a if_ as per @c remove_if() .
 * @ingroup cxxo-sup-algorithm
 * @param C a container object (lvalue).
 * @param if_ A predicate to pass to @c C.remove_if() .
 * @throws exception Whatever the expression <tt>C.erase(...)</tt> throws.
 * @see @cppref{experimental/lib_extensions_2}
 * 
 * */
template <typename Container, typename Pred>
void erase_if (Container& C, Pred if_) {
    using namespace std;
    C.erase( std::remove_if (begin(C), end(C), if_) , end(C) );
}

} // lfv2::

// 
// unstable_remove 
// from p0041
//  

	#if (CXXOMFORT_CXX_STD >= 1997)

namespace detail {
template <typename FwI, typename V>
FwI unstable_remove (FwI ini, FwI fin, V const& v, std::bidirectional_iterator_tag const) {
	while (true) {
		if (ini==fin) return ini;
		FwI rm = std::find(ini, fin, v);
		if (rm==fin) return fin;
		FwI dp = find_last_not(rm, fin, v);
		CXXOMFORT_CXX11_CODE( *rm= std::move(*dp) , *rm= *dp; );
		ini=rm;
		--fin;
	}
}

template <typename FwI, typename Fn>
FwI unstable_remove_if (FwI ini, FwI fin, Fn fn, std::bidirectional_iterator_tag const) {
	while (true) {
		if (ini==fin) return ini;
		FwI rm = std::find_if(ini, fin, fn);
		if (rm==fin) return fin;
		FwI dp = find_last_if_not(rm, fin, fn);
		CXXOMFORT_CXX11_CODE( *rm= std::move(*dp) , *rm= *dp; );
		ini=rm;
		--fin;
	}
}

/*
template <typename FwI, typename V>
FwI unstable_remove (FwI ini, FwI fin, V const& v, std::forward_iterator_tag const) {
}
*/

/*
template <typename FwI, typename F>
FwI unstable_remove_if (FwI ini, FwI fin, F&& fn, std::bidirectional_iterator_tag const) {
	while (true) {
		if (ini==fin) return ini;
		FwI rm = std::find_if(ini, fin, std::forward<F>(fn));
		if (rm==fin) return fin;
		FwI dp = find_last_not(rm, fin, v);
		*rm= std::move(*dp);
		ini=rm;
		--fin;
	}
}
*/

} // detail

/**
 * @brief Removes elements from a sequence without care for preserving ordering.
 * @xrefitem cxxo-sup-algorithm "" "" Removes elements from a sequence without care for preserving ordering. From @a p0041 .
 * 
 * @pre @a FwI models @c BidirectionalIterator 
 * 
 * @see @c std::remove() .
 * 
 * */
template <typename FwI, typename T> 
FwI unstable_remove (FwI ini, FwI fin, T const& value) {
	return detail::unstable_remove (ini, fin, value, std::iterator_traits<FwI>::iterator_category() );
};

/**
 * @brief Removes elements from a sequence without care for preserving ordering.
 * @xrefitem cxxo-sup-algorithm "" "" Removes elements from a sequence without care for preserving ordering. From @a p0041 .
 * @param fn A predicate on the value_type of the sequence <tt>[ini,fin)</tt>.
 * 
 * @pre @a FwI models @c BidirectionalIterator 
 * 
 * @see @c std::remove() .
 * 
 * */
template <typename FwI, typename Fn> 
FwI unstable_remove_if (FwI ini, FwI fin, Fn pr) {
	return detail::unstable_remove_if (ini, fin, pr, std::iterator_traits<FwI>::iterator_category() );
};

/*
#if (CXXOMFORT_CXX_STD >= 2011)
template <typename FwI, typename F>
FwI unstable_remove_if (FwI ini, FwI fin, F&& fn) {
	return detail::unstable_remove_if (ini, fin, std::forward<F>(fn), std::iterator_traits<FwI>::iterator_category() );
};
#else
template <typename FwI, typename F>
FwI unstable_remove_if (FwI ini, FwI fin, F& fn) {
	return detail::unstable_remove_if (ini, fin, fn, std::iterator_traits<FwI>::iterator_category() );
};
#endif
*/

		#if (!defined(CXXO_LIB_unstable_remove))
			//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c min_function_element() , @c max_function_element() .
			#define CXXO_LIB_unstable_remove 202209UL
		#endif
	#endif

// 
// min_function_element, max_function_element
// 


/**
 * @ingroup cxxo-sup-algorithm
 * @xrefitem cxxo-sup-algorithm "" "Sorting" Finds the element @a v in a sequence, that provides the minimum <tt>fn(v)</tt>.
 * @param less_ A @a comparator than can compare the result type of @a fn .
 * 
 * Feature macro: @c @ref CXXO_LIB_minmax_function_element .
 * 
 * */
template <typename Iter, typename Fn, typename Less>
Iter min_function_element (Iter ini, Iter fin, Fn fn, Less less_) {
	if (ini==fin) return fin;
	Iter fcur=ini;
	++ini;
	for (; ini != fin; ++ini) {
		if (less_(fn(*ini), fn(*fcur))) {
			fcur= ini;
		}
	}
	return fcur;
}

template <typename Iter, typename Fn>
Iter min_function_element (Iter ini, Iter fin, Fn fn) {
	typedef typename std::iterator_traits<Iter>::value_type V;
	return min_function_element(ini,fin,fn, std::less<V>());
}

/**
 * @ingroup cxxo-sup-algorithm
 * @xrefitem cxxo-sup-algorithm "" "Sorting" Finds the element @a v in a sequence, that provides the maximum <tt>fn(v)</tt>.
 * @param less_ A @a comparator than can compare the result type of @a fn .
 * 
 * Feature macro: @c @ref CXXO_LIB_minmax_function_element .
 * */
template <typename Iter, typename Fn, typename Less>
Iter max_function_element (Iter ini, Iter fin, Fn fn, Less less_) {
	if (ini==fin) return fin;
	Iter fcur=ini;
	for (; ini != fin; ++ini) {
		if (less_(fn(*ini), fn(*fcur))) {
			fcur= ini;
		}
	}
	return fcur;
}

template <typename Iter, typename Fn>
Iter max_function_element (Iter ini, Iter fin, Fn fn) {
	typedef typename std::iterator_traits<Iter>::value_type V;
	return max_function_element(ini,fin,fn, std::less<V>());
}

	#if (!defined(CXXO_LIB_minmax_function_element))
		//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c min_function_element() , @c max_function_element() .
		#define CXXO_LIB_minmax_function_element 202209UL
	#endif

} // algorithm::
} } // cxxomfort::library::


// 
// collect (find_all)
// 

namespace cxxomfort { namespace library { 
namespace algorithm {

template <typename Iter, typename V, typename Collector>
Collector&  collect (Iter ini, Iter fin, V const& v, Collector& cv) {
	while (ini != fin) {
		Iter fnd = std::find(ini, fin, v);
		if (fnd==fin) break;
		cv.push_back(fnd);
		ini= ++fnd;
	}
	return cv;
}

template <typename Iter, typename Pred, typename Collector>
Collector&  collect_if (Iter ini, Iter fin, Pred fn, Collector& cv) {
	while (ini != fin) {
		Iter fnd = std::find_if(ini, fin, fn);
		if (fnd==fin) break;
		cv.push_back(fnd);
		ini= ++fnd;
	}
	return cv;
}

template <typename Iter, typename Pred, typename V, typename Collector>
Collector&  collect (Iter ini, Iter fin, Pred fn, V const& v, Collector& cv) {
	while (ini != fin) {
		Iter fnd = find(ini, fin, fn, v);
		if (fnd==fin) break;
		cv.push_back(fnd);
		ini= ++fnd;
	}
	return cv;
}


} 
} } 

#endif

/**
 * @page cxxo-sup-algorithm Cxxomfort <algorithm> Supplements
 * @brief cxxomfort Supplementals for <algorithm>
 * 
 * @code
 * #include <cxxomfort/library/algorithm.hpp>
 * 
 * namespace algorithm2 = cxxomfort::library::algorithm;
 * 
 * @endcode
 * 
 * This component provides supplementary features for the set of 
 * algorithms present in <code><algorithm></code> that are 
 * specific to cxxomfort, as well as functions that complete the 
 * family of <code>_n</code> functions such as <code>transform_n</code>.
 * 
 * Interfaces defined here:
 * 
 * * @c valcmp() - compares two values the way eg.: @c memcmp does.
 * * @c any_on(), @c all_on(), @c none_on() - check a number of predicates on a given value.
 * * @c is_clamped() - complement to <tt>clamp()</tt> from p1440.
 * * @c clamp_halfopen() , @c is_clamped_halfopen() - half-open intervals complements to @c clamp() .
 * * @c copy_leftmost_n() , @c copy_rightmost_n () - copies from either end of a sequence.
 * * @c copy_if_not() - complementary to @c find_if_not .
 * * @c find_not() , @c find_if_not() - complements to @c find() , from p1223.
 * * @c count_not() , @c count_if_not() - complements to @c count() follwoing the same semantics as p1223.
 * * @c find() binary predicate form, from lwg127.
 * * @c find_last(), @c find_last_if() - find last elements in a sequence.
 * * @c find_last_not() - finds last elements not matching a value.
 * * @c find_inversion() , @c fix_inversion() - finds / fixes elements out of place in a sequence.
 * * @c count_while() - counts elements in a sequence.
 * * @c transform_inplace() , @c transform_inplace_if() - transforms sequences in-place.
 * * @c transform_n() - transforms sequences.
 * * @c count_frequencies_map() , @c count_frequencies() - gathers data frequency in a sequence.
 * * @c relative_search() - finds element ranges matching a differential.
 * * @c erase() , @c erase_if() from Fundamentals V2 - simplified erase-remove idiom for sequence containers.
 * * @c unstable_remove() , @c unstable_remove_if() - remove elements from a sequence without preserving ordering, from sg14 P0041.
 * * @c min_function_element() , @c max_function_element() — find the elements in a sequence for which a given function result is the minimum / maximum.
 * 
 * All function/class interfaces are defined in the namespace 
 * <tt>cxxomfort::library::algorithm::</tt>.
 * 
 * 
 * @section cxxo-sup-algorithm-ifpredicates Algorithms on Predicates
 * 
 * Cxxomfort includes the functions @c all_on() , @c any_on() , @c none_on() which are analogues to the std all/any/none_of family of functions. Where the original functions check a predicate @a f over all elements of a sequence, the new @c _on functions evalue all predicates on a sequence on an element @a el.
 * 
 * @code
 * vector< function<bool(Type)> > functions;
 * Type elem;
 * ...
 * if (all_on(begin(functions), end(functions), elem) ) {
 * 	... // 'elem' satisfies all checks
 * }
 * @endcode
 * 
 * @section cxxo-algorithm-experimental Experimental
 * 
 * Cxxomfort includes, in namespace @c cxxomfort::library::algorithm::lfv2:: the following additions from the Library Fundamentals v2 not yet accepted into Standard as of C++17:
 * 
 * * @c erase() - invokes member @c .erase() on a container.
 * * @c erase_if() - invokes member @c .erase_if() on a container.
 * 
 * @note The two functions are added generically, instead of providing per-containertype specific overloads as is the case in the Experimental version. This could be fixed in the future, although in the meantime it should not affect things as the functions are in a namespace of their own.
 * 
 * 
 * @section also See Also
 * 
 * @see @ref std0algorithm  .
 * 
 * */
