#ifndef CXXOMFORT_IMPL_FIXED_VECTOR_HPP
#define CXXOMFORT_IMPL_FIXED_VECTOR_HPP
/**
 * @file
 * @brief Implementation of a fixed-size variant of vector<>.
 * @author Luis Machuca Bezzaza
 *
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/memory.hpp>
#include <cxxomfort/iterator.hpp>
#include <cxxomfort/system_error.hpp>
#include <cxxomfort/library/iterator.hpp> // is_iterator
#include <cxxomfort/library/array_ref.hpp> // array_ref<T> support
#include <iterator>
#include <algorithm>
#include <stdexcept>
#include <memory>

#undef  CXXOMFORT_IMPLEMENTS_fixed_vector
#define CXXOMFORT_IMPLEMENTS_fixed_vector CXXO_IMPLSTATUS_LIBRARY()


namespace cxxomfort{ namespace library {

#if (CXXOMFORT_CXX_STD >= 1997)
	#define CXXO_LIB_fixed_vector 202209

template <typename T, typename Alloc> class fixed_vector;

/**
 * @brief A @anchor libfixedvector @c fixed_vector is a container for continuous and fixed-size storage of elements of type @a T .
 * 
 * @ingroup cxxomfort-library
 * @xrefitem cxxo-sup "Container Supplements" "Container Supplements" A @c vector - like container with a fixed size.
 * 
 * It behaves much like a <tt>vector<T></tt> except that capacity and size are fixed. There is no resize or reserve support;  consequently, the semantics of insert and erase also change.
 *
 * Part of the spirit of this kind of "missing" container is what originated n2648's "dynarray". However that proposal goes far beyond the mere concept of a "fixed vector".
 *
 * An instance of fixed_vector<T> stores elements of type T. The type @a T (without qualifications) is reflected as the member type @c value_type . The elements of a fixed_vector are stored contiguously, meaning that if @a a is a @c fixed_vector<T> then it obeys the identity
 * <tt>&a[n] == &a[0] + n</tt> for all <tt>0 <= n < N</tt>.
 *
 * A @c fixed_vector  presents about the same interface as @c vector, with the following key differences:
 *
 * * assignment is copy, removing the old contents.
 * * @c move  capability in C++03.
 * * no members throw unless specified.
 * * member @c insert has only one varianr which behaves similarly to @c at .
 * * member @c erase works by shifting elements to the left (via copy in C++03, move in C++>=11).
 * * member @c at_or is introduced to obtain access to an element or an alternative without throwing.
 * 
 * The following features are missing at present:
 * 
 * * Allocator support.
 * * emplace “inserts”.
 * 
 * 
 * @see 
 * * @cppref{container/vector}
 * * @ref libarrayref array_ref
 * 
 *
 */
template <typename T, typename Alloc=std::allocator<T> > class fixed_vector {
    typedef fixed_vector<T>  this_type;
    private:
    typedef Alloc        alloc_t;
    typedef ::std::allocator_traits<Alloc>     alloc_traits;
    typedef typename alloc_t::pointer        storage_pt;


    public:
    typedef T            value_type;         //!< Type of contained value
    typedef size_t       size_type;          //!< Type of an element's index.
    typedef ptrdiff_t    difference_type;    //!< Type of pointer or iterator distance.
    typedef value_type&       reference;     //!< Type of a reference (eg.: indexed access).
    typedef value_type const& const_reference; //!< Type of a reference (eg.: indexed access).
    typedef value_type*       pointer;
    typedef const value_type* const_pointer;
    typedef T*           iterator;           //!< @c iterator type
    typedef T const*     const_iterator;     //!< @c const_iterator type
    typedef std::reverse_iterator<iterator> reverse_iterator;  //!< reverse @c iterator type.
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator; //!< reverse @c const_iterator type.

    public:

	//! Default constructor. Sets up an empty (unusable) fixed-vector.
    fixed_vector ()
	: m_sz(0), m_p(nullptr)
	{}

    fixed_vector (fixed_vector const &V)
    : m_p(nullptr), m_sz(V.m_sz) {
		using namespace std;
		T* m_end;
		try {
			m_p= alloc(m_sz);
			m_end= uninitialized_copy(V.begin(), V.end(), this->begin());
		} catch(...) {
			destroy(m_p, m_end);
			delete[] m_p; m_p= nullptr;
			m_sz= 0;
			throw; // rethrow
		} // ~catch
	}

    explicit fixed_vector (size_type N);     //!< (1) count-constructor

    fixed_vector (size_type N, T const& v);  //!< (3) count-constructor


    //! iterator-constructor
    template <typename It>
    fixed_vector (It ini, It fin, typename std::enable_if<!std::is_arithmetic<It>::value,bool>::type* = nullptr);

#if (CXXOMFORT_CXX_STD >= 2011) || defined(DOXYGEN_DOC)
	/** 
	 * (7) initializer_list-ctor
	 * Constructs a fixed_vector with the following form:
	 * @code
	 * fixed_vector<int> ve = {1, 3, 5, 7, 9};
	 * @endcode
	 * */
    fixed_vector (std::initializer_list<T> const&) CXXO_NOEXCEPT;
#endif

    fixed_vector (CXXO_RV_REF(fixed_vector) p) CXXO_NOEXCEPT;  ///< move-constructor

    ~fixed_vector ();                                 ///< destructor

    // Assignment
    ////////
    
    //! Assignment operator.
    fixed_vector& operator= (fixed_vector rhs);

    //! Assign (copy) of another @c fixed_vector .
    fixed_vector& assign (fixed_vector rhs) CXXO_NOEXCEPTNOTHROW {
        this->swap(rhs);
        return *this;
    }

    // Information
    ////////
    
	//! Maximum allowed size (as per allocator).
    CXXO_CONSTEXPR 
    size_type       max_size () const CXXO_NOEXCEPTNOTHROW { return m_sz; } 
    //! Current ( = maxium) size.
    CXXO_CONSTEXPR 
    size_type 	size () const CXXO_NOEXCEPTNOTHROW { return m_sz; }
    //! Check if container is empty ( = has @c size == 0).
    CXXO_CONSTEXPR 
    bool 	empty () const CXXO_NOEXCEPTNOTHROW { return m_p == nullptr && m_sz == 0; } 
	//! Access to the data pointer.
    T const* 	data () const CXXO_NOEXCEPTNOTHROW { return m_p; }
    //! Access to the data pointer.
    T* 	data () CXXO_NOEXCEPTNOTHROW { return m_p; }

    //! Unchecked (!) direct access to elements
    CXXO_CONSTEXPR 
    const_reference 	operator[] (size_type idx) const CXXO_NOEXCEPTNOTHROW {
        return m_p[idx];
    }
    //! @overload operator[]
    reference 	operator[] (size_type idx) CXXO_NOEXCEPTNOTHROW {
        return m_p[idx];
    }

    //! Checked (throwing) access to elements
    const_reference 	at (size_type idx) const {
        if (m_sz <= idx) throw std::out_of_range("fixed_vector at(index) const");
        return m_p[idx];
    }
    //! @overload at
    reference 	at (size_type idx) {
        if (m_sz <= idx) throw std::out_of_range("fixed_vector at(index)");
        return m_p[idx];
    }
    //! Checked (non-throwing) access to elements
    const_reference 	at (size_type idx, std::error_code& ec) const CXXO_NOEXCEPTNOTHROW {
        if (m_sz <= idx) ec.assign(EINVAL, std::generic_category());
        return m_p[idx];
    }
    //! Checked (non-throwing) access to elements
    reference 	at (size_type idx, std::error_code& ec) CXXO_NOEXCEPTNOTHROW {
        if (m_sz <= idx) ec.assign(EINVAL, std::generic_category());
        return m_p[idx];
    }

	//! Checked (non-throwing) access to elements
	reference 	at_or (size_type idx, reference rx) { return idx < m_sz ? m_p[idx] : rx; }
	//! Checked (non-throwing) access to elements
	const_reference 	at_or (size_type idx, const_reference rx) const { return idx < m_sz ? m_p[idx] : rx; }
	

    //! Unchecked(!) access to front
    const_reference front () const CXXO_NOEXCEPTNOTHROW { return m_p[0]; }
    //! @overload front
    reference       front () CXXO_NOEXCEPTNOTHROW { return m_p[0]; }
    //! Unchecked (!) access to back
    const_reference back  () const CXXO_NOEXCEPTNOTHROW { return m_p[m_sz-1]; }
    //! @overload back
    reference       back  () CXXO_NOEXCEPTNOTHROW { return m_p[m_sz-1]; }

    //! Iterator to the beginning
    const_iterator  cbegin () const CXXO_NOEXCEPTNOTHROW { return m_p; }
    //! Iterator to the beginning
    const_iterator  begin  () const CXXO_NOEXCEPTNOTHROW { return cbegin(); }
    //! @overload begin
    iterator        begin  () CXXO_NOEXCEPTNOTHROW { return m_p; }
    //! Iterator to the end
    const_iterator  cend   () const CXXO_NOEXCEPTNOTHROW { return m_p+m_sz; }
    //! Iterador to the end
    const_iterator  end    () const CXXO_NOEXCEPTNOTHROW { return cend(); }
    //! @overload end
    iterator        end    () CXXO_NOEXCEPTNOTHROW { return m_p+m_sz; }

    //! Reverse iterator access to the end
    const_reverse_iterator    rbegin () const CXXO_NOEXCEPTNOTHROW {
        return const_reverse_iterator(this->end());
    }
    //! @overload rbegin
    reverse_iterator     rbegin () CXXO_NOEXCEPTNOTHROW {
        return reverse_iterator(this->end());
    }
    //! Reverse iterator access to the beginning
    const_reverse_iterator     rend () const CXXO_NOEXCEPTNOTHROW {
        return const_reverse_iterator(this->begin());
    }
    //! @overload rend
    reverse_iterator     rend () CXXO_NOEXCEPTNOTHROW {
        return reverse_iterator(this->begin());
    }

    //! fill with copies of value @em v
    void fill (T const& v) {
        for (iterator i= this->begin(); i != this->end(); ++i) {*i= v;}
    }

    //! swap with another @c fixed_vector 
    void swap (fixed_vector<T,Alloc> & rhs) {
        std::swap(this->m_p, rhs.m_p);
        std::swap(this->m_sz, rhs.m_sz);
        std::swap(this->m_a, rhs.m_a);
    }


    private:
    T* m_p;
    size_type m_sz;
    Alloc m_a;

    T* alloc (size_type n) {
        return reinterpret_cast<T*>( new char [n * sizeof(T)] );
    }
    
    void release (size_type n) {
        (void)n;
    }

};


// Constructor cantidad (N)
template <typename T, typename Alloc>
fixed_vector<T,Alloc>::fixed_vector (size_type N) // strong_throw
: m_p ( alloc(N) ), m_sz(N) {
    try {
        using namespace std;
        uninitialized_default_construct(m_p, m_p+m_sz);
    } catch(...) {
        delete[] m_p; m_p= nullptr;
        throw; // rethrow
    }//~catch
}

template <typename T, typename Alloc>
fixed_vector<T,Alloc>::fixed_vector (size_type N, T const& v) // strong_throw
: m_p ( alloc(N) ), m_sz(N) {
    using namespace std;
    try {
        uninitialized_fill(m_p, m_p+m_sz, v);
    } catch(...) {
        delete[] m_p; m_p= nullptr;
        throw; // rethrow
    }//~catch
}


template <typename T, typename Alloc>
template <typename It>
fixed_vector<T,Alloc>::fixed_vector (It ini, It fin, typename std::enable_if<!std::is_arithmetic<It>::value,bool>::type*)
:  m_p (alloc( std::distance(ini, fin) )) , m_sz ( std::distance(ini, fin) ) {
    using namespace std;
    try {
        //m_p= alloc(m_sz);
        uninitialized_copy(ini, fin, m_p);
    } catch (...) {
        delete[] m_p; m_p= nullptr;
        throw; // rethrow
    }
}

#if (CXXOMFORT_CXX_STD >= 2011)
// This assumes none of the involved constructors will throw!
template <typename T, typename Alloc>
fixed_vector<T,Alloc>::fixed_vector (std::initializer_list<T> const& L) CXXO_NOEXCEPT // shouldnt_throw
: m_p ( alloc(std::distance(L.begin(),L.end())) ), m_sz( std::distance(L.begin(),L.end()) ) {
    std::uninitialized_copy(L.begin(), L.end(), m_p);
}
#endif

// Move constructor
template <typename T, typename Alloc>
fixed_vector<T,Alloc>::fixed_vector (CXXO_RV_REF( fixed_vector ) p)  CXXO_NOEXCEPT // shouldnt_throw
: m_p(p.m_p), m_sz(p.m_sz) {
    p.m_p= nullptr;
    p.m_sz= 0;
}


// Destructor
template <typename T, typename Alloc>
fixed_vector<T,Alloc>::~fixed_vector () {
    if (this->empty())  return; // empty, nothing to do
    using namespace std;
    destroy_n(m_p, m_sz);
    release(m_sz);
}

template <typename T, typename Alloc>
fixed_vector<T,Alloc>& fixed_vector<T,Alloc>::operator= (fixed_vector<T,Alloc> rhs) {
    return this->assign(rhs);
}

//! Comparison operator @c == between two @c fixed_vector s.
template <typename T, typename Alloc>
inline bool operator== (fixed_vector<T,Alloc> const& a, fixed_vector<T,Alloc> const& b) {
    return ( a.size() == b.size() ) && std::equal(a.begin(), a.end(), b.begin()) ;
}

//! Comparison operator @c != between two @c fixed_vector s.
template <typename T, typename Alloc>
inline bool operator!= (fixed_vector<T,Alloc> const& a, fixed_vector<T,Alloc> const& b) {
    return !(a==b);
}

#endif // cxx_std

} } // cxxomfort::library

#endif
