#ifndef CXXOMFORT_CXXO_FUNCTIONAL_SUPP_HPP
#define CXXOMFORT_CXXO_FUNCTIONAL_SUPP_HPP

#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <cxxomfort/functional.hpp>
#include <iterator>
#include <utility>
/**
 * @file 
 * @ingroup cxxo-sup-functional
 * 
 * */

// 
// noop functor and function
// 
#include <cxxomfort/cxxo-utils.hpp> // noop_fn

namespace cxxomfort { namespace library {
namespace functional {

using ::cxxomfort::noop;

} } } 

#if (CXXOMFORT_CXX_STD>=1997 || defined(DOXYGEN_DOC))

// information about function types:
// their arguments, arity, return types, etc.

	#include "../impl/functionmeta.hpp"


namespace cxxomfort { namespace library {
namespace functional {

using ::cxxomfort::impl::functiontype;
using ::cxxomfort::impl::function_caller;

	#if defined (DOXYGEN_DOC)
/**
 * @brief Wraps a native C function in a stateless callable; built upon krzaq's \"<em>You don't need a stateful deleter[...]</em>" article.
 * @tparam Signature The C signature (prototype) of the chosen function.
 * @ingroup cxxo-sup-functional 
 * @xrefitem cxxo-sup-functional "" "" Wraps a C native function into its own stateless callable.
 * @see @ref @c std::reference_wrapper
 * @see @ref std::invoke()
 * @see https://dev.krzaq.cc/post/you-dont-need-a-stateful-deleter-in-your-unique_ptr-usually/
 * 
 * @c function_caller wraps a given function, which is passed as the second template argument, into a stateless (empty) callable that can be used as eg.: a @c Deleter type or a @c Hash type in template instantiations that require it.
 * 
 * For an example use, consider:
 * 
 * @code
 * function_caller< void(MyType*) , &mylibrary_free > free_me;
 * // free_me can now be used in contexts where objects (not natives) are needed
 * unique_ptr<MyType, decltype(free_me)> ptr == mylibrary_malloc(...);
 * @endcode
 * 
 * @note @c function_caller respects the function signature as given. For wrapping native / extern C functions starting with C++>=17 @e without the added exception overhead, add @c noexcept to the signature, eg.: <tt>function_caller< void*(size_t) noexcept, malloc ></tt> .
 * 
 * @note This is also available for internal cxxomfort usage as <tt>cxxomfort::impl::function_caller</tt>.
 * */
	template <typename Signature, Signature const* fn>
	struct function_caller { };
	
	#endif

// 
// pointer-to-nullary-function
// 

template <typename R>
struct nullary_function_ptrt {
    typedef ::cxxomfort::impl::functiontype<R()> FT;
    typename FT::pointer ptr;
    nullary_function_ptrt (typename FT::pointer const p) : ptr(p) {}
    R operator() () const { return ptr(); }
    
};

/**
 * @brief Overload for @c std::ptr_fun() that wrapps nullary functions.
 * @ingroup cxxo-sup-functional
 * @xrefitem cxxo-sup-functional "" "" Wraps a nullary function.
 * 
 * This is a nullary overload for <tt>std::ptr_fun()</tt> and is provided only for completeness; C++03/TR1 code would prefer <tt>std::bind()</tt>, and C++14-onward code would prefer lambda statements where possible.
 * 
 * */
template <typename R>
nullary_function_ptrt<R> ptr_fun (R(* const function)()) {
    return nullary_function_ptrt<R>(function);
}

	#define CXXO_LIB_ptr_fun_nullary 202209UL

// 
// 
// 

//! generic no-op functor with return type T.
template <typename T> 
struct noopT_t {
    CXXO_CONSTEXPR14 T operator() (...) const CXXO_NOEXCEPTNOTHROW { return T(); }
};


// 
// equivalent_to 
// 

/**
 * @brief Function object that performs equivalence ("not less, not greater") comparison.
 * @ingroup cxxo-sup-functional
 * @xrefitem cxxo-sup-functional "" "Operators" Test two objects for equivalence (partial ordering).
 * 
 * The comparison performed by a @c equivalent_to object on its arguments is basically a sort of "false equality" or non-"strict" equality, calculated only in terms of a less-than comparator function. Given two values @a va and @a vb of type @c V , they satisfy <tt>equivalent_to<V>()(va,vb) == true</tt> if @e both @a va is not less than @a vb and @a vb is not less than @a va . In other words:
 * 
 * @code

equivalent_to<T> eq;
T foo, bar;
eq(foo,bar) == !less(foo,bar) && !greater(foo,bar) == !less(foo,bar) && !less(bar,foo)
// note the above relationship does not necessarily imply!:
eq(foo,bar) == equal_to(foo,bar)

 * @endcode
 * 
 * See also: @c @cppref{utility/functional/equal_to}
 * 
 * */
template <typename T, typename Ord = std::less<T> >
struct equivalent_to {
	typedef bool result_type;
	typedef Ord ordering_type;

	//! Returns @c true if neither @a vl nor @a vr compare "less" than each other according to this object's comparison order.
	CXXO_CONSTEXPR14 bool 
	operator() (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW {
		/*static*/ const Ord ord;
		return !(ord(vl,vr)) && !(ord(vr,vl));
	}
	
};

/**
 * @brief Transparent Operator specialization for @c equivalent_to .
 * @xrefitem cxxo-sup-functional "" "Operators" Transparent operator for @c equivalent_to .
 * @see @ref transparent_operators 
 * 
 * */
template<typename Ord> 
struct equivalent_to<void,Ord> {
	typedef bool result_type;
	struct is_transparent {};
	template <typename T> 
	CXXO_CONSTEXPR14	
	bool operator() (T const& vl, T const& vr) CXXO_NOEXCEPTNOTHROW {
		/*static*/ const Ord ord;
		return !ord(vl,vr) && !ord(vr,vl);
	}
	
};


//! Generates an equivalence comparison functor on the fly.
CXXO_CONSTEXPR14 static inline 
equivalent_to<void> make_equivalence () CXXO_NOEXCEPTNOTHROW {
	return equivalent_to<void>();
}

#if (!defined(CXXO_LIB_equivalent_to))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c equivalent_to .
	#define CXXO_LIB_equivalent_to 202209UL
#endif

// 
// function_facet
// 

template <typename R, typename S, S*> struct function_facet;

template <typename R, typename U, U(* const fn)()>
struct function_facet<R(), U(), fn> {
    R operator() () const {
        return static_cast<R>(fn());
    }
};

template <typename R, typename R1, typename U, typename U1, U(* const fn)(U1)>
struct function_facet<R(R1), U(U1), fn> {
    R operator() (R1 r1) const {
        return static_cast<R>(fn(r1));
    }
};



/**
 * @brief Trait that identifies <tt>std::function<(...)></tt>.
 * @ingroup cxxo-sup-functional 
 * @xrefitem cxxo-sup-functional "" "Traits" Type trait that identifies <tt>std:function<...></tt> types.
 * 
 * @see @ref std::is_function .
 * */
template <typename T> struct is_std_function : std::false_type {};
template <typename Sig> struct is_std_function< std::function<Sig> > : std::true_type {};


/**
 * Wrapper that converts a constant value of type @a T 
 * into a function of signature <code>T(void)</code>.
 * 
 * @ingroup cxxo-sup-functional 
 * 
 * */
template <typename T>
struct constant_function {
    T const value;
    CXXO_CONSTEXPR14 constant_function (T const& t) : value(t) {}
    
    T const& operator() (...) const CXXO_NOEXCEPT { return value; }
    
};

template <typename T>
inline
constant_function<T> constant_fn (T const& t) {
	return constant_function<T>(t);
}

CXXO_CONSTEXPR14 const constant_function<bool> true_fn (true);
CXXO_CONSTEXPR14 const constant_function<bool> false_fn (false);



} // .functional:
}}

// 
// extended operators
// 

#include "operatorit.hpp" // supplement operator functors

// 
// function_ref
// 

//! @include{doc} "../impl/p0792r0-function_ref.hpp"
#include "../impl/p0792-function_ref.hpp"

namespace cxxomfort { namespace library { namespace functional {
    using ::cxxomfort::impl::p0792::function_ref;

	#if defined(DOXYGEN_DOC)
	/**
	 * @brief Wraps a callable with a given signature @c Sig .
	 * @xrefitem cxxo-sup-functional "" "" Wraps a callable with a given signature in a manner similar to that of @c std::reference_wrapper .
	 * 
	 * <tt>function_ref< Ret(A1,A2,...)></tt> wraps any callable that takes arguments <tt>(A1, A2, ...)</tt> and returns @c R , into a @b reference / view type.
	 * 
	 * @note As per C++>=17 spec, @c function_ref signatures distinguish between @c noexcept and non- @c noexcept signatures.
	 * 
	 * @note Pre-C++11, @c function_ref takes signatures of up to 5 arguments.
	 * 
	 * @see @ref function_caller , for an even stricter way to wrap C native functions.
	 * 
	 * */
	template <typename Sig> struct function_ref { 
		//! Constructor takes a callable's address
		template <typename Callable>
		explicit function_ref (Callable* p) noexcept {}
		//! Function call operator
		template <typename R, typename... Args>
		R operator() (Args&&... args) const {}
	};
	#endif
	
} } }

	#if (!defined(CXXO_LIB_function_ref))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c function_ref .
		#define CXXO_LIB_function_ref 202209UL
	#endif
#else // < 1997
	// unsupported


#endif // CXX_STD

#endif // end of file


/**
 * @page cxxo-sup-functional Cxxomfort <functional> Supplements
 * @brief Supplements to <functional>.
 * @ingroup cxxo-sup
 * 
 * @code
 * #include <cxxomfort/library/functional.hpp>
 * 
 * namespace functional2 = cxxomfort::library::functional;
 * 
 * @endcode
 *
 * This component offers a number of alternatives and supplements to features found in 
 * <code><functional></code> such as function objects to complete various tasks, 
 * a function-like reference wrapper, and 
 * function objects that complement the <code>plus</code> family.
 * 
 * * @c #noop - an object that does nothing with its arguments.
 * * @ref @c ptr_fun() — extended for nullary functions.
 * * @ref sup-functional-operators "Supplements to named operators". 
 * * Function object @c 'equivalent_to' ("not less or greater").
 * * @c is_std_function — type traits to identify <tt>std::function</tt> objects.
 * * @c constant_function .
 *   * The assist functions @c true_fn() , @c false_fn() and @c constant_fn() that create specific constants.
 * * @c #functiontype - meta information on function signatures.
 * * @ref @c cxxomfort::library::functional::function_caller  - zero-cost function object calling a named C function.
 * * @ref @c cxxomfort::library::functional::function_ref  - non-owning reference to a function-like callable.
 * * @ref @c overload() metafunction.
 * 
 * 
 * @section sup-functional-operators Supplemental Operators
 * 
 * Given a Standard C++ operator wrapper like <tt>std::plus</tt>, which wraps <tt>operator+()</tt>, cxxomfort also provides the following additional wrappers:
 * 
 * * @c predecrement — wraps <tt>--t</tt>.
 * * @c preincrement — wraps <tt>++t</tt>.
 * * @c lshift — wraps <tt>t << u</tt>.
 * * @c rshift — wraps <tt>t >> u</tt>.
 * 
 * Each operator is accompanied by their corresponding transparent operator in the vein of @ref std0functional "C++14 Transparent Operators" .
 * 
 * And their implementations are analogous to the following:
 * 
 * @code
 * template <typename T>
 * struct predecrement {
 *   T& operator() (T& t) const { return --t; }
 * };
 * template <> struct predecrement<void> {
 *   template <typename A>
 *   A& operator() (A& a) const { return --a; }
 * };
 * @endcode
 * 
 * Given a Standard C++ operator wrapper like <tt>std::plus</tt>, which wraps <tt>operator+(t,u)</tt>, cxxomfort also provides supplemental wrappers that covers the corresponding left-assign operation.
 * 
 * That is, they operate in the following way:
 * 
 * @code
// effectively invokes 'elem += 2' for every elem in List
transform (begin(List), end(List), begin(List), bind(plus_assign<void>(), placeholders::_1, 2) );

 * @endcode
 * 
 * Each operator is accompanied by their corresponding transparent operator in the vein of @ref std0functional "C++14 Transparent Operators" .
 *
 * The following names are provided:
 * 
 * * @c plus_assign (implements <tt>left += right</tt>)
 * * @c minus_assign (implements <tt>left -= right</tt>)
 * * @c multiplies_assign  (implements <tt>left *= right</tt>)
 * * @c divides_assign (implements <tt>left /= right</tt>)
 * * @c modulus_assign (implements <tt>left %= right</tt>)
 * * @c bit_and_assign (implements <tt>left &= right</tt>)
 * * @c bit_or_assign (implements <tt>left |= right</tt>)
 * * @c bit_xor_assign (implements <tt>left ^= right</tt>)
 * 
 * And their implementations are analogous to the following:
 * 
 * @code
 * template <typename T, typename U=T>
 * struct plus_assign {
 *   T& operator() (T& t, U const& u) const { return t+= u; }
 * };
 * template <> struct plus_assign<void> {
 *   template <typename A, typename B>
 *   A& operator() (A& a, B const& b) const { return a+= b; }
 * };
 * @endcode
 * 
 * Cxxomfort also provides the @ref @c 'equivalent_to' tests for potential "equivalence" between two objects - that is, none is less than the other.
 * 
 */
