#ifndef CXXOMFORT_SUPP_ITERATOR_HPP
#define CXXOMFORT_SUPP_ITERATOR_HPP
#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <iterator>
#include <functional>
#include <cxxomfort/iterator.hpp>
#include <cxxomfort/functional.hpp>
#include <cxxomfort/impl/p1227r0-ssize.hpp> // ssize

/**
 * @file
 * @ingroup cxxo-sup-iterator
 */

#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD >= 1997)


namespace cxxomfort { namespace library { 
namespace iterator {

template <typename Integer, typename I>
size_t absdistance (I i1, I i2, Integer i, std::forward_iterator_tag) {
    for (i= 0; i1 != i2; ++i, ++i1) {}
    return i; 
}

template <typename Integer, typename I>
size_t absdistance (I i1, I i2, Integer i) {
    typename std::iterator_traits<I>::iterator_category cat;
    return absdistance(i1, i2, i, cat);
}


/**
 * @brief iterates over nothing, it just stays "pointing" to the same place
 */
template <typename T>
struct fake_iterator {
    typedef std::reference_wrapper<T const> V;
    typedef std::bidirectional_iterator_tag iterator_category;
    typedef T value_type;
    typedef size_t difference_type;
    typedef T* pointer;
    typedef T& reference;
    
    // "end" state
    fake_iterator () CXXO_NOEXCEPT : v(), b(false) {}
    explicit fake_iterator (T const& t) CXXO_NOEXCEPT : v(t), b(true) {}
    T const& operator* () const {
        return v.get();
    }
    
    fake_iterator& operator++() CXXO_NOEXCEPT { return *this; }
    fake_iterator operator++(int) CXXO_NOEXCEPT { return *this; }

    fake_iterator& operator--() CXXO_NOEXCEPT { return *this; }
    fake_iterator operator--(int) CXXO_NOEXCEPT { return *this; }

    friend bool operator== (fake_iterator const& c1, fake_iterator const& c2) CXXO_NOEXCEPT {
        return (c1.b==false and c2.b==false) or (c1.b==true and c2.b==true and c1.v==c2.v);
    }

    friend bool operator!= (fake_iterator const& c1, fake_iterator const& c2) CXXO_NOEXCEPT {
        return !(c1==c2);
    }

    private:
    V const v;
    bool b;
};

template <typename T>
fake_iterator<T> make_fake_iterator(T const& t) CXXO_NOEXCEPTNOTHROW {
	return fake_iterator<T>(t);
}


/**
 * recursive_iterator<T,N> 
 * iterator that dereferences N times before dereferencing to its value
 * (N == 0 -> normal pointer dereference, N == 1 -> one dereference before pointer dereference, etc)
 */

template <typename T, unsigned N>
struct recursive_iterator {
    
};

template <typename T> struct recursive_iterator<T,0> {
    T iterator;
};


template <typename I, typename F>
struct function_iterator
: public I {
    private:
    F fn;
    typedef I base_t;
    
    public:
    function_iterator(I & i, F& f) 
    : I(i), fn(f) {}
    
    using I::operator=;
    using I::operator++;
    
    typename I::value_type operator* () const {
        I const* base= this;
        return fn(*base);
    }
    
};

template <typename I, typename F>
function_iterator<I,F> make_function_iterator(I const& i, F & f) {
    return function_iterator<I,F>(i,f);
}


#if 1
/**
 * @class container_extensions_view
 * @brief A view into a container that adds extra member functions.
 * 
 * A @c container_extensions_view initialized to a given container 
 * @a c , adds a number of member functions that are cxxomfort extensions 
 * or are otherwise not readily available.
 * 
 * * <tt>.at_or (index, object&)</tt> -- returns a reference to the element at position @a index , or @a object if none is found.
 * * <tt>.contiguous ()</tt> -- determines if the container uses continuous storage. Currently unimplemented, returns @c false .
 * 
 * */

template <typename Cont, typename If = void>
struct container_extensions_view 
{
	typedef Cont container_type;
	typedef typename container_type::value_type value_type;
	typedef typename container_type::size_type size_type;
	typedef typename container_type::iterator iterator;
	typedef typename container_type::const_iterator const_iterator;
	
	CXXO_CONSTEXPR 
	explicit container_extensions_view (Cont& ct) CXXO_NOEXCEPTNOTHROW 
	: c (ct) 
	{}

	CXXO_CONSTEXPR 
	size_type size () const CXXO_NOEXCEPTNOTHROW {
		return c.size();
	}
	
	iterator begin () {
		return c.begin();
	}

	iterator end () {
		return c.end();
	}
	
	value_type const& 
	at_or (size_type i, value_type const& o) const CXXO_NOEXCEPTNOTHROW {
		return i < c.size() ? c[i] : o;
	}

	value_type & 
	at_or (size_type i, value_type & o) CXXO_NOEXCEPTNOTHROW {
		return i < c.size() ? c[i] : o;
	}
	
	private:
	// std::reference_wrapper<Cont> c;
	Cont& c;
	
};

template <typename Cont>
CXXO_CONSTEXPR container_extensions_view<Cont const> ccexview (Cont const& c) CXXO_NOEXCEPTNOTHROW {
	return container_extensions_view<Cont const> (c);
}

template <typename Cont>
container_extensions_view<Cont> cexview (Cont & c) CXXO_NOEXCEPT {
	return container_extensions_view<Cont> (c);
}

#endif // container

/**
 * @}
 * */
// iterator
}
// cxxomfort::library
} }

// 
// extensions to cxxomfort::fix::contiguous_access
// 

#include <vector>
#include <string>

namespace cxxomfort { namespace fix {

// C++ can not into partial template specializations on subtypes?
// need to figure out how to implement this concept some other way

#if (0)
template <typename T>
struct is_contiguous_access_iterator <
	typename std::vector<T>::iterator

> {
	enum { value = (__cplusplus>=201103L), specialized = 1 };
};
#endif

// in C++11 onwards std::string and std::wstring (but no std::basic_string<...>?) are guaranteed to have contiguous storage

template<> 
struct is_contiguous_access_iterator <std::string::iterator> {
	enum { value = (CXXOMFORT_CXX_STD>=2011), specialized = 1 };
};

template<> 
struct is_contiguous_access_iterator <std::wstring::iterator> {
	enum { value = (CXXOMFORT_CXX_STD>=2011), specialized = 1 };
};


} } // namespace cxxomfort::fix


#endif


/**
 * @page cxxo-sup-iterator Cxxomfort <iterator> Supplements
 * @ingroup cxxo-sup
 * 
 * @code
 * #include <cxxomfort/library/iterator.hpp>
 * 
 * namespace iterator2 = cxxomfort::library::iterator;
 * 
 * @endcode
 * 
 * This component provides supplementary features for 
 * the utilities found in <tt><iterator></tt>.
 * 
 * @section Interfaces
 * 
 * Interfaces defined here:
 * 
 * * @c absdistance() .
 * * @c fake_iterator and @c make_fake_iterator() - an iterator that iterates over one single value.
 * * @c function_iterator and @c make_function_iterator() - an iterator that presents an object mapped via a function.
 * * @c at_or() - access element at a given index without throwing.
 * * @c index_or() - access element at a given index without throwing.
 * * @c container_extensions_view .
 * 
 * All interfaces are defined in the namespace 
 * <tt>cxxomfort::library::iterator::</tt>.
 * 
 * Interfaces previously defined here:
 * 
 * * @c ssize() from p1227r0, moved to @ref std0iterator .
 * 
 * @section also See Also
 * 
 * * @see @ref std0iterator
 * * @see @ref cxxo-sup-algorithm 
 * 
 */
#endif
