#ifndef CXXOMFORT_CXXO_NUMERIC_SUPP_HPP
#define CXXOMFORT_CXXO_NUMERIC_SUPP_HPP

#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <cxxomfort/numeric.hpp>
#include <iterator>

/**
 * @file cxxomfort/library/numeric.hpp
 * 
 */

#if (CXXOMFORT_CXX_STD>=1997 || defined(DOXYGEN_DOC))

#include <cxxomfort/impl/to_signed.hpp>
#include <cxxomfort/cxxo-utils.hpp> // static_math

namespace cxxomfort { namespace library {
//! Supplements to the standard utilities in <tt><numeric></tt>.
namespace numeric {

	/**
	 * @ingroup cxxo-sup-numeric
	 * @{
	 * */

using ::cxxomfort::traits::make_signed;
using ::cxxomfort::traits::make_unsigned;

	#if defined(DOXYGEN_DOC)
//! Converts @a t to the equivalent integral @c signed type.
template <typename T> 
typename std::make_signed<T>::type to_signed (T t);
//! Converts @a t to the equivalent integral @c unsigned type.
template <typename T> 
typename std::make_unsigned<T>::type to_unsigned (T t);
	
	#else
using ::cxxomfort::impl::numeric::to_signed;
using ::cxxomfort::impl::numeric::to_unsigned;
// using cxxomfort::impl::numeric::to_enclosing_integer;
	#endif

#if (!defined(CXXO_LIB_to_signed))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c to_signed() , @c to_unsigned() .
	#define CXXO_LIB_to_signed 202209
#endif

// 
// left fold and right fold
// 

/**
 * @brief Folds a sequence with @a f from the @b right .
 * @param f An accumulator object.
 * @param v0 An initial value.
 * @return The final accumulated value.
 * 
 * This function is exactly like <tt>std::accumulate</tt> except that the accumulator processes its state value on the right instead of on the left. In other words, if @c v carries the state, where <tt>std::accumulate</tt> calculates <tt>f(v, x_i)</tt>, @c accumulate_right calculates <tt>f(x_i, v)</tt>.
 * 
 * */
template <typename It, typename T, typename F>
T accumulate_right (It ini, It fin, T v0, F f) {
    for (; ini != fin; ++ini) {
        v0 = f(*ini, v0);
    }    
    return v0;
}

/**
 * @brief Folds a sequence with @a f from the @b right .
 * @overload accumulate_right
 * 
 * */
template <typename It, typename T>
T accumulate_right (It ini, It fin, T v0) {
    for (; ini != fin; ++ini) {
        v0 = *ini + v0;
    }    
    return v0;
}

template <typename I, I i1, I i2>
struct static_gcd_typed {
    static const I value= (i2==0) ? i1 : static_gcd_typed<I, i2, i1 % i2>::value;
};


template <unsigned i1, unsigned i2>
struct static_gcd_impl {
    static const unsigned value= (i2==0U) ? i1 : static_gcd_impl<i2, i1 % i2>::value ;
};


// compile-time math

//! Compile-time arithmethic functions, that can be used in template parameters.
namespace ctmath {

using ::cxxomfort::st_math::static_abs;
using ::cxxomfort::st_math::static_sign;

using ::cxxomfort::st_math::static_clamp;
using ::cxxomfort::st_math::static_minmax2;
using ::cxxomfort::st_math::midpoint;
using ::cxxomfort::st_math::static_min;
using ::cxxomfort::st_math::static_max;
using ::cxxomfort::st_math::static_intpow;
using ::cxxomfort::st_math::static_intlog;

using ::cxxomfort::st_math::static_gcd;
using ::cxxomfort::st_math::static_lcm;

// left fold
/*
template <class F<typename>, typename T, T t0, T t1, T t2, T t3, T t4, T t5>
struct static_leftf {
    static const T value = F<T,t0, static_leftf<F,T,t1,t2,t3,t4>::value >::value;
};
*/

/*
template <typename F, typename T, T t0, T t1>
struct static_leftf <F,T,t0,t1> {
    static const T value = F<T, t0, t1>::value;
};
*/

} // ctmath::


// 
// arithmetic series generator 
// 

template <typename T, T ax, T b>
struct arith_series_gen {
	CXXO_CONSTEXPR
	arith_series_gen () CXXO_NOEXCEPT
	: v(0), vi(0) {}
	
	CXXO_CONSTEXPR 
	explicit arith_series_gen (T v0) CXXO_NOEXCEPTNOTHROW
	: v(v0) , vi(v0) {}
	
	CXXO_CONSTEXPR 
	operator T () const CXXO_NOEXCEPTNOTHROW { return v; }
	
	T operator() () CXXO_NOEXCEPTNOTHROW {
		T ret= v;
		v= ax*v+b;
		return ret;
	}
	
	void reset () CXXO_NOEXCEPT_COND(std::is_integral<T>::value) {
		v= vi;
	}

	T v;
	T const vi;
};

// 
// 
// 

	/**
	 * @}
	 * */
} // numeric::
} } // cxxomfort::library



#else
	// unsupported
#endif // CXX_STD

#endif

/**
 * @page cxxo-sup-numeric Cxxomfort <numeric> Supplements
 * @brief cxxomfort Supplementals for <numeric>
 * 
 * @code
 * #include <cxxomfort/library/numeric.hpp>
 * 
 * namespace numeric2 = cxxomfort::library::numeric;
 * 
 * @endcode
 * 
 * 
 * This component provides supplementary features for the set of 
 * algorithms present in <code><numeric></code> that are 
 * specific to cxxomfort, in particular pre-C++11 compile-time versions 
 * of the utilities like gcd() and lcm() that were made constexpr.
 * 
 * Interfaces defined here:
 * 
 * * @c accumulate_left() , @c accumulate_right() , @c accumulate_associative() - variants of @c std::accumulate that change the invocation and use of the accumulator object.
 * * @c to_signed() , @c to_unsigned()  - convert integers to their signed or unsigned variants as per <tt>std::make_[un]signed</tt>.
 * * @c static_abs - statically obtain absolute value
 * * @c static_min , @c static_max , @link cxxomfort::library::numeric::static_minmax static_minmax @endlink
 * * @c #static_gcd , @c #static_lcm - statically obtain GCD and LCM of two numbers.
 * * @ref @c static_sum
 * 
 * All interfaces are defined in the namespace 
 * <tt>cxxomfort::library::numeric::</tt>.
 * 
 * See also: @ref std0numeric .
 * 
 * @section Static Numerics
 * 
 * The cxxomfort supplemental templates @c static_min , @c static_max and @c static_minmax provide constant-time evaluation of the 
 * minimum and maximum of a set of integers as Non-Type Template Parameters, in particular useful for C++<11 where 
 * @c constexpr and its application on <tt>std::[min|max]</tt> are not available.
 * 
 * @c static_min and @c static_max return their results via the <tt>::value</tt> member. 
 * @c static_minmax returns the members <tt>::min_value</tt> and <tt>max_value</tt>.
 * 
 * Pre-C++11, these templates take a @a type and up to 5 (five) numeric parameters.
 * 
 * The supplemental <tt>static_sum</tt> takes a @a type and up to 10 (ten) template parameters and returns 
 * as the member <tt>::value</tt> the sum (addition) of those values.
 * 
 * The supplementals @c static_gcd and @c static_lcm provide the same kind of compatible functionality to the 
 * C++17 @c constexpr implementations of <tt>std::[gcd,lcm]</tt>.
 * 
 * @section Number Conversions
 * 
 * The cxxomfort supplementals @c to_signed() and @c to_unsigned() convert a numeric expression to the 
 * corresponding signed or unsigned variant of the type. They are intended as 
 * functional supplements to type_traits' <tt>make_[signed,unsigned]</tt>.
 * 
 * Neither function performs range checking.
 * 
 */
