#ifndef CXXOMFORT_CXXO_OPERATORIT_HPP
#define CXXOMFORT_CXXO_OPERATORIT_HPP
/**
 * @file 
 * @ingroup cxxo-sup-functional
 */

#include <cxxomfort/cxxomfort.hpp>
#include <cxxomfort/functional.hpp>
#include <cxxomfort/type_traits.hpp> // add_reference

namespace cxxomfort { namespace library {
namespace functional {

/**
 * @ingroup cxxo-sup-functional 
 * @{
 */

/**
 * @brief Functoid that wraps <tt>++t</tt> for a given @a T .
 */
template <typename T>
struct preincrement
:std::unary_function< T&,T> {
    typedef T argument_type;
    typedef T result_type;
    inline T& operator() (T& t) const { return ++t; }
};

//! Heterogeneous (deduced) wrapper for @ref preincrement .
template<> struct preincrement<void> {
    template <typename T>
    inline T& operator() (T& t) const { return ++t; }
};

/**
 * @brief Functoid that wraps <tt>--t</tt> for a given @a T .
 */
template <typename T>
struct predecrement
:std::unary_function< T&,T> {
    typedef T argument_type;
    typedef T result_type;
    inline T& operator() (T& t) const { return --t; }
};

//! Heterogeneous (deduced) wrapper version of @ref predecrement.
template<> struct predecrement<void> {
    template <typename T>
    inline T& operator() (T& t) { return --t; }

};


/** 
 * @xrefitem cxxo-sup-functional "" "" Named operator that wraps <tt>a<<b</tt>.
 * @ingroup cxxo-sup-functional
 * */
template <typename T>
struct lshift {
	typedef T result_type;
	inline T operator() (T const& t1, T const& t2) const { return t1<<t2; }
};
//! Transparent operator specialization for @c lshift
//! @refitem cxxo-sup-functional 
template<> struct lshift<void> {
	struct is_transparent {};
	template <typename A, typename B>
	inline A operator() (A const& a, B const& b) const { return a<<b; }
};

/** 
 * @xrefitem cxxo-sup-functional "" "" Named operator that wraps <tt>a>>b</tt>.
 * @ingroup cxxo-sup-functional
 * */
template <typename T>
struct rshift {
	typedef T result_type;
	inline T operator() (T const& t1, T const& t2) const { return t1>>t2; }
};
//! Transparent operator specialization for @c rshift
//! @refitem cxxo-sup-functional 
template<> struct rshift<void> {
	struct is_transparent {};
	template <typename A, typename B>
	inline A operator() (A const& a, B const& b) const { return a>>b; }
};


#define CXXO_DEF_OPERATORIT(Name,Sym) \
template <typename It, typename Ot=It> \
struct CXXO_JOIN(Name,_assign) \
: std::binary_function< It& ,Ot const&,It&> {   \
    inline It& operator() (It& l, Ot const& r) const {   \
        return l Sym r;       \
    }                         \
};                            \
template<> struct CXXO_JOIN(Name,_assign)<void> \
{ \
	struct is_transparent {}; \
	template <typename L, typename R>  \
	inline L& operator() (L& l, R const& r) const {  \
		return l Sym r;  \
	}  \
};  \


/**
 * @struct plus_assign
 * @brief Implements the <tt>a+=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a+=b</tt>.
 * @see @ref std::plus 
 * */
CXXO_DEF_OPERATORIT(plus, += );
/**
 * @struct minus_assign
 * @brief Implements the <tt>a-=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a-=b</tt>.
 * @see @ref std::minus 
 * */
CXXO_DEF_OPERATORIT(minus, -= );
/**
 * @struct multiplies_assign
 * @brief Implements the <tt>a*=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a*=b</tt>.
 * @see @ref std::multiplies 
 * */
CXXO_DEF_OPERATORIT(multiplies, *= );
/**
 * @struct divides_assign
 * @brief Implements the <tt>a/=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a/=b</tt>.
 * @see @ref std::divides 
 * */
CXXO_DEF_OPERATORIT(divides, /= );
CXXO_DEF_OPERATORIT(modulus, %= );

/**
 * @struct lshift_assign
 * @brief Implements the <tt>a<<=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a<<=b</tt>.
 * @see @ref cxxomfort::library::functional::lshift 
 * */
CXXO_DEF_OPERATORIT(lshift, <<= );
/**
 * @struct rshift_assign
 * @brief Implements the <tt>a>>=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a>>=b</tt>.
 * @see @ref cxxomfort::library::functional::rshift 
 * */
CXXO_DEF_OPERATORIT(rshift, >>= );

/**
 * @struct bit_and_assign 
 * @brief Implements the <tt>a&=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a&=b</tt>.
 * @see @ref std::bit_and 
 * */
CXXO_DEF_OPERATORIT(bit_and, &= );
/**
 * @struct bit_or_assign
 * @brief Implements the <tt>a|=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a|=b</tt>.
 * @see @ref std::bit_or 
 * */
CXXO_DEF_OPERATORIT(bit_or, |= );
/**
 * @struct bit_xor_assign
 * @brief Implements the <tt>a^=b</tt> extended operator.
 * @xrefitem cxxo-sup-functional "" "Operators" @ref sup-functional-operators  implements <tt>a^=b</tt>.
 * @see @ref std::bit_xor 
 * */
CXXO_DEF_OPERATORIT(bit_xor, ^= );

#undef CXXO_DEF_OPERATORIT



// leftshift, rightshift are defined in terms of cxxomfort-operators
//CXXO_DEF_OPERATORIT(leftshift_it,<<=);
//CXXO_DEF_OPERATORIT(rightshift_it,>>=);

template <typename It, typename Ot=It> 
struct leftshift_assign : ::std::binary_function<It&,Ot const&,It&> {   
    inline It& operator() (It& l, Ot const& r) const { 
        return l<<=r;
    }
};

template <typename It, typename Ot=It> 
struct rightshift_assign : ::std::binary_function<It&,Ot const&,It&> {   
    inline It& operator() (It& l, Ot const& r) const { 
        return l>>=r;
    }
};



/**
 * @}
 */

} // functional
} } // cxxomfort::library

#define CXXO_LIB_leftassign_operators 202203

#endif

