#ifndef CXXOMFORT_CXXO_RANDOM_SUPP_HPP
#define CXXOMFORT_CXXO_RANDOM_SUPP_HPP
/**
 * @file
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "<random>" "Header File"
 */

#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <cxxomfort/random.hpp>
#include <cxxomfort/functional.hpp>
#include <cxxomfort/cstdint.hpp>
#include <cxxomfort/limits.hpp>
#include <cxxomfort/impl/math.hpp>
#include <cxxomfort/library/numeric.hpp> // make_signed
#include <ctime>

namespace cxxomfort { namespace library {
namespace random {

#if (CXXOMFORT_CXX_STD >= 2011)
inline namespace lfv2 {

/**
 * @brief Return the "random global engine" declared by the program.
 * @ingroup cxxo-sup-random
 * @see Library Fundamentals v2
 * 
 * */
std::default_random_engine& random_global_engine () {
    static std::default_random_engine Engine ( time(0) );
    return Engine;
};

//
// reseed, from Fundamentals V2
//

void reseed (std::default_random_engine::result_type value) {
    random_global_engine().seed(value);
}


//
// randint, from experimentals / Fundamentals V2
//
template <typename IntType>
IntType randint (IntType a, IntType b) CXXO_NOEXCEPTNOTHROW {
    std::uniform_int_distribution<IntType> Dis(a,b);
    if (b < a) return 0;
    return Dis(random_global_engine());
}

} // library fundamentals

#endif


// 
// linear_lrshift_feedback_engine
// 

template <typename UInt, UInt U_l, UInt U_r, UInt U_m>
struct linear_lrshift_feedback_engine {
    enum { typecheck = std::numeric_limits<UInt>::is_integer == true && std::numeric_limits<UInt>::is_signed==false };
    static_assert( typecheck, "UInt type fails unsigned checks");

    enum { ulcheck = (U_l < (std::numeric_limits<UInt>::digits-2) && U_l > 0 ) };
    static_assert( ulcheck, "Parameter U_l fails checks");

    enum { urcheck = (U_r < (std::numeric_limits<UInt>::digits-2) && U_r > 0) };
    static_assert( urcheck, "Parameter U_r fails checks");

    enum { umcheck = ((U_m & 1U) && ((U_m >> U_r) & 1U) ) };
    static_assert( umcheck, "Parameter U_m fails checks");
    //enum { uplcheck = (u_pl < (1U<<U_l)) };
    //static_assert( uplcheck, "Paramener upl fails checks");
    
    typedef UInt result_type;

    UInt uprev;
    size_t const upl; // payload to add when generating
    static size_t const upl_default = 0xdecaf0UL & ((1U<< (U_l+U_r))-1);
    
    CXXO_CONSTEXPR 
    linear_lrshift_feedback_engine () CXXO_NOEXCEPT
    : uprev(1), upl(upl_default) {}
    
    CXXO_CONSTEXPR 
    linear_lrshift_feedback_engine (UInt s0, size_t pl= 1) CXXO_NOEXCEPT 
    : uprev(s0 & U_m), upl(pl < (1U<<U_r) ? pl : upl_default) {}
    
    CXXO_CONSTEXPR14 
    result_type operator() () CXXO_NOEXCEPTNOTHROW {
	result_type n = uprev;
	n^= (uprev>> U_r);
	n^= (uprev<< U_l);
	n^= upl;
	n&= U_m;
	uprev=n;
	return n;
    }
    
    void discard (size_t n) {
	while(n-->0) this->operator()();
    }

    CXXO_CONSTEXPR14 
    void seed (result_type s0) {
	uprev= s0 & U_m;
    }
    
    CXXO_CONSTEXPR14 
    static result_type 	min () { return UInt(0U); }
    CXXO_CONSTEXPR14 
    static result_type 	max () { return U_m - 1U; }
    
    static CXXO_CONSTEXPR14 const result_type const_min = 0;
    static CXXO_CONSTEXPR14 result_type const const_max;
    
};

/*
template <typename UInt, UInt U_r, UInt U_l, UInt U_m>  
typename linear_lrshift_feedback_engine<UInt,U_r,U_l,U_m>::result_type const const_min = UInt(0U);

template <typename UInt, UInt U_r, UInt U_l, UInt U_m>  
typename linear_lrshift_feedback_engine<UInt,U_r,U_l,U_m>::result_type const const_max = UInt(U_m - 1U);
*/

typedef linear_lrshift_feedback_engine<uint32_t, 7, 3, 0x00ffffffUL> lefabo24;

// 
// shift_stretch_engine 
// 

/*
 * lomask = (1<<p)-1
 * hi = x0 & ~lomask
 * lo = x0 & lomask
 * next = (hi<< plw) | ((pl&plw)<<p) | lo
 * next = next ^ mask
 * 
 * */

template <typename UInt, UInt _um, size_t _ib, size_t _pv, size_t _pw>
struct shift_stretch_engine  {
    enum { typecheck = std::numeric_limits<UInt>::is_integer == true && std::numeric_limits<UInt>::is_signed==false };
    static_assert( typecheck, "UInt type fails unsigned checks");

    enum { ibcheck = (_ib < std::numeric_limits<UInt>::digits-1) };
    static_assert( ibcheck, "Parameter _ib (insertion point) fails checks");

    enum { pvcheck = (_pv > 0 && _pv < ::cxxomfort::fix::integral_limits<size_t>::const_max ) };
    static_assert( pvcheck, "Parameter _pv (payload) fails checks");

    enum { pwcheck = (_pw + _ib < std::numeric_limits<UInt>::digits) };
    static_assert( pwcheck, "Parameter _pw (payload width) fails checks");

    static size_t const masked_payload = (_pv & ((1U<<_pw)-1) );
    static size_t const lomask = (1U<<_ib)-1U;
    
    typedef UInt result_type;

    UInt uprev;
    static size_t const default_seed = 0xdeadecafUL;
    
    CXXO_CONSTEXPR 
    shift_stretch_engine () CXXO_NOEXCEPT
    : uprev(default_seed) {}
    
    CXXO_CONSTEXPR 
    shift_stretch_engine (UInt s0) CXXO_NOEXCEPT 
    : uprev(s0) {}
    
    CXXO_CONSTEXPR14 
    result_type operator() () CXXO_NOEXCEPTNOTHROW {
	UInt hi = uprev & ~lomask;
	hi^= uprev>>_pw;
	UInt lo = uprev & lomask;
	lo^= uprev<<_pw;
	UInt next = 0;
	next|= (hi<< _pw);
	next|= (_pv<<_ib);
	next|= (lo);
	next^= _um;
	uprev= next;
	return uprev;
    }
    
    void discard (size_t n) {
	while(n-->0) this->operator()();
    }

    CXXO_CONSTEXPR14 
    void seed (result_type s0) {
	uprev= s0;
    }
    
    CXXO_CONSTEXPR14 
    static result_type 	min () { return UInt(0U); }
    CXXO_CONSTEXPR14 
    static result_type 	max () { return std::numeric_limits<UInt>::max(); }
    
    static CXXO_CONSTEXPR14 result_type const const_min;
    static CXXO_CONSTEXPR14 result_type const const_max;
    
};

/*
template <typename UInt, UInt _um, size_t _ib, size_t _pv, size_t _pw> 
typename shift_stretch_engine<UInt,_um,_ib,_pv,_pw>::result_type const const_min = UInt(0U);

template <typename UInt, UInt _um, size_t _ib, size_t _pv, size_t _pw> 
typename shift_stretch_engine<UInt,_um,_ib,_pv,_pw>::result_type const const_max = UInt(0xffffffffUL);
*/


typedef shift_stretch_engine<uint32_t, 0xbd0430a3, 21, 0x0b3, 6> shifteree;

// 
// inverse congruential 
// 

template <typename Int, Int _q, size_t _a, size_t _c>
struct inversive_congruential_engine {
    static const Int modulus = _q;
    Int uprev;
    Int pseed;

    typedef typename ::cxxomfort::library::numeric::make_unsigned<Int>::type UInt;
    typedef UInt result_type;

    inversive_congruential_engine (result_type s0) CXXO_NOEXCEPTNOTHROW
    : uprev(s0)
    { }

#if (CXXOMFORT_CXX_STD >= 2011)
    template <typename R, typename std::enable_if<std::is_class<R>::value || std::is_function<R>::value , bool>::type = true>
    explicit inversive_congruential_engine (R& r) CXXO_NOEXCEPT 
    { this->seed(r); }
#endif
	
void seed (result_type s0) {
	uprev= s0;
}

#if (CXXOMFORT_CXX_STD >= 1997)
	#if (CXXOMFORT_CXX_STD >= 2011)
	template <typename R, typename std::enable_if< std::is_class<R>::value || std::is_function<R>::value, bool >::type = true>
	void seed (R& r) 
	#else
	template <typename R>
	void seed (R& r, typename std::enable_if< std::is_class<R>::value || std::is_function<R>::value , bool* >::type = nullptr)
	#endif
	/* */ {
		uprev= r();
	}
#endif
	
	result_type operator() () {
		result_type uinv= ::cxxomfort::impl::mul_inv(uprev, modulus);
		//result_type uinv= ::cxxomfort::impl::ModInverseNaive(uprev, _q);
		result_type unew= (uprev==0 
			? (_c) : (_a * uinv + _c) % modulus);
		uprev= unew;
		return unew;
	}
	
	void discard (uintmax_t d) {
		while (d-->0) { this->operator()(); }
	}
	
	static result_type     min () { return 0; }
	static result_type     max () { return modulus-1; }

	#if (CXXOMFORT_CXX_STD >= 1997)
	static const result_type const_min;
	static const result_type const_max;
	#endif
};

	#if (CXXOMFORT_CXX_STD >= 1997)
template <typename Int, Int _q, size_t _a, size_t _c>
typename inversive_congruential_engine<Int,_q,_a,_c>::result_type const 
inversive_congruential_engine<Int,_q,_a,_c>::const_min = 0;

template <typename Int, Int _q, size_t _a, size_t _c>
typename inversive_congruential_engine<Int,_q,_a,_c>::result_type  const 
inversive_congruential_engine<Int,_q,_a,_c>::const_max = _q - 1;
	#endif


typedef inversive_congruential_engine<unsigned int, 7, 2, 3> 
icg0;

#define SHL(p,k) (p<<k)
    
typedef inversive_congruential_engine<uint64_t, SHL(1ULL,63)-25ULL, 5520335699031059059ULL, 2752743153957480735ULL>
jdcook_sampl2020;

typedef inversive_congruential_engine<int32_t, 2147483647, 9102, (int32_t)(2147483647UL-36884165UL)> 
hellekalek1995;

#undef SHL

/*
 * Usage:
 * @code

 * @endcode
 * */
template <typename TD, typename TG>
struct random_wrapper {
    random_wrapper (TD& d, TG& g)
    : tg(g), td(d) {}

    typename TD::result_type operator() () {
        return td((TG&)tg);
    }

    std::reference_wrapper<TG> tg;
    TD td;
};

template <typename TD, typename TG>
random_wrapper<TD,TG> make_random_wrap (TD& d, TG& g) {
    return random_wrapper<TD,TG>(d,g);
}

#if defined(DOXYGEN_DOC)
/**
 * @typedef lcg_numerical_recipes_rand
 * @brief LCG as it appears in <em>Numerical Recipes</em>.
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "Generators" "" The LCG presented in <em>Numerical Recipes</em>.
 * @see https://en.wikipedia.org/wiki/Numerical_Recipes
 * 
 * This is an alternative to eg.: @c std::minstd and other linear congruential engines, using the parameters to the generator specified in <em>Numerical Recipes</em> :
 * 
 * @code
 * std::linear_congruential_engine<uint32_t, 1664525, 1013904223, 0>
 * @endcode
 * 
 * When invoked with <tt>operator()</tt> returns a @c uint32_t value.
 * */
typedef std::linear_congruential_engine<uint32_t, Parameters...> 
lcg_numerical_recipes_rand;
#else
typedef std::linear_congruential_engine<uint32_t, 1664525, 1013904223, 0> 
lcg_numerical_recipes_rand;
#endif

#if defined(DOXYGEN_DOC)
/**
 * @typedef MMIX_rand
 * @brief MMIX: LCG specification as recommended by Donald Knuth in "<em>The Art of Computer Programming – Seminumerical Algorithms</em>".
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "Generators" "" The LCG presented by Donald Knuth in <em>The Art of Computer Programming</em>.
 * @see https://arxiv.org/pdf/1811.04035.pdf
 * 
 * This is an alternative to eg.: @c std::minstd and other linear congruential engines using the parameters described in <em>The Art of Computer Programming</em>:
 * 
 * @code
 * std::linear_congruential_engine<uint64_t, 6364136223846793005ULL, 1442695040888963407ULL, 0>
 * @endcode
 * 
 * When invoked with <tt>operator()</tt> returns a @c uint64_t value.
 * */
typedef std::linear_congruential_engine<uint64_t, Parameters...>
MMIX_rand;
#else
typedef std::linear_congruential_engine<uint64_t, 6364136223846793005ULL, 1442695040888963407ULL, 0> 
MMIX_rand;
#endif

/**
 * @typedef musl_rand
 * @brief LCG specification as it appears in MUSL Libc.
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "Generators" "" The base LCG generator presented as part of musl.
 * 
 * This is an alternative to eg.: @c std::minstd and other linear congruential engines.
 * 
 * When invoked with <tt>operator()</tt> returns a @c uint64_t value.
 * */
typedef std::linear_congruential_engine<uint64_t, 6364136223846793005, 1, 0>
musl_rand;

/**
 * @typedef rand48
 * @brief LCG specification as it appears in FreeBSD.
 * @ingroup cxxo-sup-random
 * 
 * This is an alternative to eg.: @c std::minstd and other linear congruential engines.
 * 
 * When invoked with <tt>operator()</tt> returns a @c uint64_t value within a range that uses the first (lowest) 48 bits.
 * */
typedef std::linear_congruential_engine<uint_least64_t, 0x00000005deece66dULL , 0xb, 0x0001000000000000ULL >
rand48;


/**
 * @typedef xkcd221_rand
 * @brief LCG specification following Randall Munroe's <em>xkcd #221</em>.
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "Generators" "" LCG with parameters defined by Randall Munroe.
 * @see https://xkcd.com/221/
 * 
 * This is an alternative to eg.: @c std::minstd and other linear congruential engines, using as parameters to the generator the ones described in Randall Munroe's xkcd #221:
 * 
 * @code
 * std::linear_congruential_engine<unsigned int, 0, 4, 0>
 * @endcode
 * 
 * When invoked with <tt>operator()</tt> returns a @c unsigned value. The generator is equivalent to the following hand-written code:
 * 
 * @code
 * struct xkcd221_r {
 *  constexpr unsigned int () const noexcept { 
 *      return 4; // chosen by fair dice roll, guaranteed to be random
 *  } 
 * };
 * @endcode
 * 
 * */
typedef std::linear_congruential_engine<unsigned int, 0, 4, 0>
xkcd221_rand;

// 
// quickie random quality-of-life improvements
// 

/**
 * @brief Given a generator @a g , generate and choose an integer in the range <tt>[a, b]</tt>.
 * @param g A generator object (lvalue).
 * @return An @e Integer value @a r such that <tt>a <= r <= b</tt>.
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "QoL improvements" "" Choose a random number within an interval.
 * 
 * @pre @a Integer is a C++ integer type.
 * 
 * */
template <typename Gen, typename Integer>
Integer uniform_choose (Gen& g, Integer a, Integer b) {
    std::uniform_int_distribution<Integer> di(a,b);
    return di(g);
}

/**
 * @brief Choose one in a sequence of values <tt>[ini,fin)</tt>. 
 * @param ini,fin A sequence's begin and end iterators.
 * @return An iterator in the range <tt>[ini,fin)</tt>.
 * @ingroup cxxo-sup-random
 * @xrefitem cxxo-sup-random "QoL improvements" "" Pick a random element from within a sequence.
 * 
 * This function uses @c uniform_choose() on the passed generator @a g to uniformly pick one iterator within the desired range.
 * 
 * */
template <typename Gen, typename Iterator>
Iterator uniform_pick (Gen& g, Iterator ini, Iterator fin) {
    typedef typename std::iterator_traits<Iterator>::difference_type DT;
    DT dist = std::distance(ini,fin)-1;
    DT r = choose(g, 0, dist);
    return std::next(ini,r);
}

#if (!defined(CXXO_LIB_uniform_pick_choose))
	//! @xrefitem cxxomfort-libraryfeaturemacros "" "" Announces implementation of @c uniform_pick() , @c uniform_choose() .
	#define CXXO_LIB_uniform_pick_choose 201812
#endif

// 
// splitmix64 from http://xoshiro.di.unimi.it/splitmix64.c 
// recommended by  https://arvid.io/2018/07/02/better-cxx-prng/#fnref2
// 

/**
 * @brief Pseudo-RNG generator 
 * @xrefitem cxxo-sup-random "" "Vendored Utils" Xoroshiro/'splitmix64' PRNG, by Sebastiano Vigna
 * @xrefitem cxxo-vendored "" "Vendored utils" Xoroshiro/'splitmix64' PRNG, by Sebastiano Vigna
 * @see http://xoshiro.di.unimi.it/splitmix64.c
 * @see https://arvid.io/2018/07/02/better-cxx-prng
 * 
 * @c splitmix is a @e Generator compatible PRNG generator that produces 64-bit numbers.
 * 
 * */
struct splitmix64 {
    // we'll mirror the interface of eg.: minstd_rand somewhat
    private:
    uint64_t _state;
    public:
    //! the type of the integer generated
    typedef uint64_t result_type;

    //! Sets up a PRNG with a given @a seed .
    CXXO_CONSTEXPR splitmix64 (uint64_t useed) CXXO_NOEXCEPTNOTHROW 
    : _state(useed)  {}
    
 
    static CXXO_CONSTEXPR14 result_type min () CXXO_NOEXCEPTNOTHROW { 
        return 0;
    }
    static CXXO_CONSTEXPR14 result_type max () CXXO_NOEXCEPTNOTHROW {
        return ::cxxomfort::fix::integral_limits<uint64_t>::const_max;
    }

    //static const result_type const_min = min();
    //static const result_type const_max = max();
    
    //! Changes the seed of the current object
    void seed (uint64_t s) CXXO_NOEXCEPTNOTHROW {
        _state= s;
    }
    
    //! Obtains a pseudo-random number from the generator
    result_type operator() () CXXO_NOEXCEPTNOTHROW {
        uint64_t z = (_state += 0x9e3779b97f4a7c15ULL);
        z = (z ^ (z >> 30)) * 0xbf58476d1ce4e5b9ULL;
        z = (z ^ (z >> 27)) * 0x94d049bb133111ebULL;
        return z ^ (z >> 31);
    }

    //! Discards @a z "steps" from the generator
    void discard (unsigned long long z) CXXO_NOEXCEPTNOTHROW {
        while (z --> 0) { this->operator()(); }
    }
    
};

    #define CXXO_LIB_prng_splitmix 201812

#if 0 // original source code

/*  Written in 2015 by Sebastiano Vigna (vigna@acm.org)

To the extent possible under law, the author has dedicated all copyright
and related and neighboring rights to this software to the public domain
worldwide. This software is distributed without any warranty.

See <http://creativecommons.org/publicdomain/zero/1.0/>. */

#include <stdint.h>

/* This is a fixed-increment version of Java 8's SplittableRandom generator
   See http://dx.doi.org/10.1145/2714064.2660195 and 
   http://docs.oracle.com/javase/8/docs/api/java/util/SplittableRandom.html

   It is a very fast generator passing BigCrush, and it can be useful if
   for some reason you absolutely want 64 bits of state; otherwise, we
   rather suggest to use a xoroshiro128+ (for moderately parallel
   computations) or xorshift1024* (for massively parallel computations)
   generator. */

static uint64_t x; /* The state can be seeded with any value. */

uint64_t next() {
	uint64_t z = (x += 0x9e3779b97f4a7c15);
	z = (z ^ (z >> 30)) * 0xbf58476d1ce4e5b9;
	z = (z ^ (z >> 27)) * 0x94d049bb133111eb;
	return z ^ (z >> 31);
}

#endif

} // cxxomfort::library::random
}}


// 
// ran64, from numpy
// 
namespace cxxomfort { namespace library { 
namespace random {
/**
 * @c ran64 is a PRNG that is found in Python @c numpy . It produces 64-bit numbers.
 * */
struct ran64 {
    public:
    //! the type of the integer generated
    typedef uint64_t result_type;
    
    CXXO_CONSTEXPR14 ran64 (uint64_t useed= 0) CXXO_NOEXCEPT 
    : u_(0), v_(0), w_(0) 
    {
        this->seed(useed);
    }
    
    CXXO_CONSTEXPR14 
    void 	seed (uint64_t useed) CXXO_NOEXCEPTNOTHROW {
        v_= 0x38ecac5fb3251641ULL;
        w_= 1ULL;
        u_ = useed ^ v_;
        this->shuffle();
        v_= u_;
        this->shuffle();
        w_ = v_;
        this->shuffle();
    }

    
    result_type 	operator() () {
	return this->shuffle();
    }
    
    void discard (size_t n) {
	while (n-->0) this->operator()();
    }

    static CXXO_CONSTEXPR result_type min () CXXO_NOEXCEPTNOTHROW { 
        return ::cxxomfort::fix::integral_limits<uint64_t>::const_min;
    }
    static CXXO_CONSTEXPR result_type max () CXXO_NOEXCEPTNOTHROW {
        return ::cxxomfort::fix::integral_limits<uint64_t>::const_max;
    }
    
    static CXXO_CONSTEXPR result_type const const_min = ::cxxomfort::fix::integral_limits<uint64_t>::const_min;
    
    
    private:
    uint64_t u_, v_, w_;

    CXXO_CONSTEXPR14 
    result_type 	shuffle () CXXO_NOEXCEPTNOTHROW {
	
	uint64_t xu = u_;
	xu = (xu * 0x27bb2ee687b0b0fdULL + 0x9c740a0a6788d2cULL);
	u_= xu;
	
	uint64_t xv = v_;
	xv ^= (xv >> 17);
	xv &= static_cast<uint64_t>(xv << 31);
	xv ^= (xv >> 8);
	v_ = xv;
	
	uint64_t xw = w_;
	xw = (xw * 0xffffda61ULL + (xw >> 32));
	w_ = xw;
	
	uint64_t x = xu ^ (xu << 21);
	x ^= (x >> 35);
	x ^= (x << 4);
	return (x+xv)^xw;
	
    }
    
};

    #define CXXO_LIB_prng_ran64 201812

}
}}


#if defined(DOXYGEN_DOC)

namespace cxxomfort { namespace library {
namespace random {

} // random
} } 


#endif

/**
 * @page cxxo-sup-random Cxxomfort <random> Supplements
 *
 * @code
 * #include <cxxomfort/library/random.hpp>
 * // to avoid typing explosion
 * namespace random2 = cxxomfort::library::random;
 * @endcode
 *
 * @section Interfaces
 *  
 * This component provides supplementary features to the ones
 * present in <code><random></code> that are
 * specific to cxxomfort.
 *
 * * @ref cxxo-sup-random-stdgen2 - A selection of specializations supplementing the @c linear_conguential_engine in the Standard Library.
 * * @ref cxxo-sup-random-vendored — A selection of PRNGs copied or vendored from a number of sources.
 * * @c default_random_engine() — from Fundamentals v2.
 * * @c reseed() — from Fundamentals v2.
 * * @c randint() — from Fundamentals v2.
 * * @c uniform_choose() , @c uniform_pick() — simplified interfaces for random choosing.
 *
 *
 * All interfaces are defined in the namespace 
 * <tt>cxxomfort::library::random::</tt>.
 * 
 * @section cxxo-sup-random-stdgen2 Extra LCG Generators
 * 
 * The following <em>linear congruential</em> alternatives to @c std::minstd are provided in namespace @c cxxomfort::library::random :
 * 
 * * @ref @c lcg_numerical_recipes_rand - LCG as per "Numerical Recipes"
 * * @ref @c mmix_rand - LCG as per Donald Knuth in "The Art of Computer Programming"
 * * @ref @c musl_rand - LCG as used by @c musl library's @c rand .
 * * @ref @c xkcd221_rand - LCG as per Randal Munroe in "xkcd #221".
 * 
 * The generators correspond to the following specializations:
 * 
 * @code
typedef std::linear_congruential_engine<uint32_t, 1664525, 1013904223, 0> 
lcg_numerical_recipes_rand;

typedef std::linear_congruential_engine<uint64_t, 6364136223846793005ULL, 1442695040888963407ULL, 0> 
MMIX_rand;

typedef std::linear_congruential_engine<uint64_t, 6364136223846793005, 1, 0>
musl_rand;

typedef std::linear_congruential_engine<unsigned int, 0, 4, 0>
xkcd221_rand;
 * @endcode
 * 
 * @section cxxo-sup-random-vendored Vendor Tools
 * 
 * The library adds the following external PRNGs to namespace <tt>cxxomfort::library::random</tt>, and accompanies where possible with original source code in <tt>cxxomfort/vendor</tt>:
 * 
 * * @ref @c splitmix64 — Xoroshiro splitmix64 implementation, by Sebastian Virga.
 * * @ref @c ran64 — Simplified ran64 implementation from the Python <a href="https://github.com/fsssosei/Pure_PRNG">Pure_PRNG</a> project.
 * 
 * 
 * @section also See Also
 * 
 * @see @ref std0random 
 * @see https://github.com/fsssosei/Pure_PRNG 
 *
 */


#endif
