#ifndef CXXOMFORT_LIBRARY_SEQUENCES_HPP
#define CXXOMFORT_LIBRARY_SEQUENCES_HPP

/**
 * @file
 *
 */

#include "../config.hpp"
#include <cxxomfort/array.hpp>
#include <cxxomfort/iterator.hpp>
#include <cxxomfort/type_traits.hpp>
#include "../library/i12n.hpp"
#include <vector>

/**
 * @page cxxo-sequences Sequence Helpers
 * @ingroup cxxo-sup
 *
 * This header provides and centralizes a number of facilities that assist in working with sequences and containers.
 * 
 * @code
 * #include <cxxomfort/library/sequences.hpp>
 * @endcode
 * 
 * Interfaces defined in this file:
 *
 * @section sequences1 Sequence / Container Traits
 * 
 * This header provides in namespace @c cxxomfort::library a number of type traits supplementing std type traits, to identify and work with certain types of sequences:
 * 
 * 
 * * @c is_std_array - identified <tt>std::array</tt> types, in contrast to @c std::is_array .
 * * @c is_sequence_container
 * * @c is_indexable_like_array
 * * @c is_iterator
 * * @c make_array (to be moved to <array>
 * * @c seq_ , @c generic_sequence_t
 *
 *
 */


//
// is_std_array
//

namespace cxxomfort { namespace library {
} } // cxxomfort::library

//
// is_indexable_like_array operator
//
// A type is considered indexable like an array if it exposes a value_type
// and a []-operator (returning value_type [const] &).
namespace cxxomfort { namespace library {
namespace detail {

template <typename T> struct bracket_indexable_helper {

    typedef struct { int a; } yes;
    typedef char no;

    private:
    template <typename U, U> struct really_has;

    template <typename C> static yes& Test_op(really_has <typename C::value_type const& (C::*)() const,
                            &C::operator[] >*);

    template <typename> static no& Test_op(...);

    public:
    static const bool has_op    = (sizeof(Test_op<T>(0)) == sizeof(yes));

};


} // detail

/**
 * @brief Type identification trait for types that
 * are indexable via an array-like <code>operator[]</code>.
 */
template <typename T> struct is_indexable_like_array
: std::conditional< detail::bracket_indexable_helper<T>::has_op, std::true_type, std::false_type>::type
{};

template <typename T, size_t N> struct is_indexable_like_array< T[N] >
: std::true_type
{};


} } //cxxomfort::library

// 
// sequence_t and seq_ generator
// 

#include "../impl/seq_.hpp"

// 
// numeric_sequence, iterator and enumerator
// 

//#include "../impl/numeric-sequence.hpp"


//
// is_sequence -- for specializations based on ability to provide
// access to a sequence of objects
//

namespace cxxomfort { namespace library {

//
// is_iterator
//

namespace detail {

// detect iterator_category member
template <typename T> struct has_member_iterator_category {
private:
    typedef char                      yes;
    typedef yes no[2];

    template <typename C> static yes& test_ic(typename C::iterator_category*);
    template <typename C> static no&  test_ic(...);

    public:
    // iterator_category
    enum { value = sizeof(test_ic<T>(0)) == sizeof(yes) };
};

} // detail::

// Given a type @p T , inherit from @c true_type  if @p T  is a type
// considered an iterator by iterator_traits.
/*
template<typename T, typename = void>
 struct is_iterator : std::false_type {};

template<typename T>
struct is_iterator<
    T, typename std::enable_if<
        detail::has_member_iterator_category< std::iterator_traits<T> >::value
        and !std::is_same<typename std::iterator_traits<T>::value_type, void>::value
    >::type> : std::true_type {};
*/


/**
 * @struct is_iterator
 * @brief Type identification trait for iterator types as defined by the <code><iterator></code> interface.
 */
template<class T>
struct is_iterator {
    static T makeT();
    typedef void * twoptrs[2];  // sizeof(twoptrs) > sizeof(void *)
    static twoptrs & test(...); // Common case
    template<class R> static typename R::iterator_category * test(R); // Iterator
    template<class R> static void * test(R *); // Pointer

    static const bool value = sizeof(test(makeT())) == sizeof(void *);
};



/* A type S is considered a sequence if:
 * * the calls std::begin(s) and std::end(s) make sense
 * * the member type "iterator" and "value_type" are exposed
 * * the return type of those calls is specialized by iterator_traits
 * * alternatively, S is a native C array
 * * alternatively, S is a std::valarray
 */

namespace detail {

    template <typename T> struct has_sequence_members {
    private:
        typedef char                      yes;
        typedef yes no[2];

        template <typename C> static yes& test_ci(typename C::iterator*);
        template <typename C> static no&  test_ci(...);
        template <typename C> static yes& test_vt(typename C::value_type*);
        template <typename C> static no&  test_vt(...);

        public:
        // const_iterator
        enum { value_ci = sizeof(test_ci<T>(0)) == sizeof(yes) };
        // value_type
        enum { value_vt = sizeof(test_vt<T>(0)) == sizeof(yes) };
        typedef T type;

        private:
        template <typename U, U> struct really_has;

        template <typename C> static yes& Test_begin(really_has <typename C::const_iterator (C::*)() const,
                                            &C::begin>*);
        template <typename C> static yes& Test_begin(really_has <typename C::iterator (C::*)() const,
                                            &C::begin>*);

        template <typename C> static yes& Test_end(really_has <typename C::const_iterator (C::*)() const,
                                            &C::end>*);
        template <typename C> static yes& Test_end(really_has <typename C::iterator (C::*)() const,
                                            &C::end>*);

        template <typename C> static yes& Test_data(really_has <typename C::const_pointer (C::*)() const,
                                            &C::data>*);

        template <typename> static no& Test_begin(...);
        template <typename> static no& Test_end(...);
        template <typename> static no& Test_data(...);

        public:
        enum { 
		value_begin_end = (
            sizeof(Test_begin<T>(0)) == sizeof(yes) )
            && ( sizeof(Test_end<T>(0)) == sizeof(yes)
            )
        , 
        value_has_data = (
            sizeof(Test_data<T>(0)) == sizeof(yes)
            )
        };
    };

template <typename T, bool=false>
struct is_sequence_impl: traits::false_type {};

template <typename T> struct is_sequence_impl<T,true>
: traits::integral_constant<bool, is_iterator<typename T::const_iterator>::value > {};

} // detail::

// at present, a type S is a sequence if:
// 1.- has a const_iterator and value_type member types
// 1b.- has begin, end member functions returning type iterator or const_iterator (other overloads are ignored)
// 2.- or is a C array.

/**
 * @brief Type identification trait for sequence-like types:
 * * have a @c const_iterator member type.
 * * have a @c begin  and @c end  member function.
 */
template <typename S> struct is_sequence
: detail::is_sequence_impl<S,
    detail::has_sequence_members<S>::value_ci
    && detail::has_sequence_members<S>::value_vt
    && detail::has_sequence_members<S>::value_begin_end
> {};

// arrays are sequences
template <typename T, size_t N>
struct is_sequence< T[N] > : traits::true_type {};

// starting C++11, valarrays are sequences

#if (CXXOMFORT_CXX_STD >= 2011)
template <typename T>
struct is_sequence< std::valarray<T> >: std::true_type {};
#endif


//
// maybe_contiguous
//


/*
 * criteria for a sequence type to maybe be contiguous:
 *
 * 1.- is a sequence according to is_sequence
 * 2.- supports member operator[]
 * 3a.- iterator == add_pointer<value_type>
 * 3b.- iterator has random_access_iterator_tag AND sequence has a .data() member that returns pointer
 * 3c.- iterator has contiguous_iterator_tag concept (C++20) AND sequence has a .data() member that returns iterator
 * 4.- the type is registered with register_contiguous_sequence
 *
 * */

namespace {

#if (CXXOMFORT_CXX_STD >= 1997)
template <typename S, typename IfS= void>
struct contiguous_helper {
    typedef typename S::iterator I;
    typedef typename S::value_type V;
    typedef typename std::add_pointer<V>::type P;
    enum { clause1 = is_sequence<S>::value };
    enum { clause2 = is_indexable_like_array<S>::value };
    enum { clause3a = std::is_same< I, P >::value };
    enum { clause3b = std::is_same< std::random_access_iterator_tag, typename std::iterator_traits<I>::iterator_category>::value
        && detail::has_sequence_members<S>::has_data
        };

    enum { value = clause1 && clause2 && (clause3a || clause3b) };
};

template <typename T>
struct contiguous_helper<T, typename std::enable_if< std::is_fundamental<T>::value >::type >
: std::false_type {};

template <typename T, size_t N>
struct contiguous_helper <T[N], typename std::enable_if<true> ::type > {
    enum { value = true };
};

template <typename T>
struct contiguous_helper <std::valarray<T> , typename std::enable_if<true>::type > {
    enum { value = true };
};

#else
template <typename S, typename IfS= void>
struct contiguous_helper {
    enum { value = false };
};
#endif
}

template <typename T>
struct registered_contiguous
: std::bool_constant<contiguous_helper<T>::value> {
};


template <typename S>
struct is_maybe_contiguous_sequence
: public std::bool_constant<
 (false==std::is_fundamental<S>::value) && registered_contiguous<S>::value
> {
};

} } // cxxomfort::library


#endif

