#ifndef CXXOMFORT_CXXO_STRING_SUPP_HPP
#define CXXOMFORT_CXXO_STRING_SUPP_HPP
/**
 * @file
 * @brief Header providing extensions to <string>, <string_view> headers.
 * @xrefitem cxxo-sup-string "<string>" "Header"
 *
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/functional.hpp> // functors, binders, placeholders
#include <cxxomfort/string.hpp>
#include <cxxomfort/string_view.hpp>
#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD>=1997)

#include <cxxomfort/algorithm.hpp>
#include <cxxomfort/ostream.hpp> // operator<< for std::string_view 
#include <cstring>
#include <sstream>
#include "../impl/character.hpp"
#include "../impl/to_basic_string_variadic.hpp"

namespace cxxomfort { namespace library {
namespace string {

//! @xrefitem cxxo-sup-string Type trait that identifies @c basic_string specializations.
template <typename T> struct is_std_basic_string : std::false_type {};

template <typename CharType, typename ChT, typename A>
struct is_std_basic_string< std::basic_string< CharType, ChT, A > > 
: ::cxxomfort::impl::type_traits::is_character_type<CharType> {};


using ::cxxomfort::impl::string::to_basic_string; // variadic to_string
using ::cxxomfort::impl::string::to_string; // variadic to_string
// from c++
using ::cxxomfort::fix::string_helpers::compare;
using ::cxxomfort::fix::string_helpers::substr;
// from c++20
using ::cxxomfort::string::starts_with;
using ::cxxomfort::string::ends_with;
// from c++23
//using cxxomfort::string::contains;


/**
 * @brief Replaces contents in a string.
 * @param from,to Strings to change from and to.
 * @return A new constructed @c basic_string object.
 * 
 * */
template <typename Ch>
std::basic_string<Ch> 
string_replace (std::basic_string<Ch> const& from, std::basic_string<Ch> const& to, std::basic_string<Ch> s) {
    typedef std::basic_string<Ch> S;
    for (size_t pos = 0; ; pos += from.length() ) {
        pos= s.find(from, pos);
        if (pos == S::npos) break;
        s.erase (pos, from.length());
        s.insert (pos, to);
    }
    return s;

} // string_replace


// "overload" here

template <typename T, typename F>
struct string_cast_impl {
    // naive conversion. This will more often than not not suffice.
    T operator() (F const& f) const {
        using namespace std;
        return T(begin(f), end(f));
    }
};

template <typename T>
struct string_cast_impl<T,T> {
    T operator() (T const& t) const CXXO_NOEXCEPTNOTHROW {
        return t;
    }
};

// customization point for string to wstring conversion
template<>
struct string_cast_impl<std::wstring,std::string> {
    std::wstring operator() (std::string const& s) const {
        std::wstring ws(s.size(), 0); // Overestimate number of code points.
        ws.resize(std::mbstowcs(&ws[0], s.c_str(), s.size())); // Shrink to fit.        
        return ws;
    }
};

template<>
struct string_cast_impl<std::string,std::wstring> {
    std::string operator() (std::wstring const& ws) const {
        std::string s(ws.size(), 0); // Overestimate number of code points.
        s.resize(wcstombs(&s[0], ws.c_str(), ws.size())); // Shrink to fit.        
        return s;
    }
};

/**
 * @brief Converts between string representations. Common / Customization Point.
 * @tparam To A string type to convert to.
 * @param str A string of a given string type.
 * @return A new string of type @a To which is the template argument to invoke with.
 */
template <typename To, typename FromString>
To string_cast (FromString const& str) {
    static string_cast_impl<To,FromString> const c;
    return c(str);
}


} // string

/**
 * @}
 */


}} // cxxomfort::library


#else // not at least c++97
	// unsupported
	#pragma message CXXO_NOTICE("unsupported supplement - string")

#endif // CXX_STD

#endif

/**
 * @page cxxo-sup-string Cxxomfort <string> Supplements
 * @brief cxxomfort Supplementals for <string> 
 * 
 * @code
 * #include <cxxomfort/library/string.hpp>
 * 
 * namespace string2 = cxxomfort::library::string;
 * 
 * @endcode
 *
 * This section provides supplementary features to the ones
 * present in <code><string></code> that are
 * specific to cxxomfort.
 *
 * * @c chr() , @c utf8chr()  - Conversion from integer/codepoint to character.
 * * @c codepoint() - Conversion from character to integer/codepoint.
 * * @c string_cast() - convert between string representations; overload point.
 * * @c join() - Joins elements into a string.
 * * @c split_to() , @c split_to_n() - Split elements from a string.
 * * @c explode() - Splits elements from a string into a std::vector.
 * * @c l_trim() , @c r_trim() , @c trim()  - Removes elements from a string's ends.
 * * @c string_replace - Replace substrings in a string.
 * * @c string_pad - Pad a string to length.
 * * @c to_string() - Variadic version of to_string.
 *
 *
 * All interfaces are defined in the namespace <code>cxxomfort::algorithm::</code>.
 *
 * */
