#ifndef CXXOMFORT_CXXO_TEXT_SUPP_HPP
#define CXXOMFORT_CXXO_TEXT_SUPP_HPP
/**
 * @file 
 * @ingroup cxxo-sup-text
 * @xrefitem cxxo-sup-text Text Manipulation Supplements
 *
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/functional.hpp> // functors, binders, placeholders
#include <cxxomfort/string.hpp>
#include <cxxomfort/string_view.hpp>

#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD>=1997)

#include <cxxomfort/algorithm.hpp>
#include <cstring>
#include <sstream>
#include "../impl/character.hpp"
#include "../impl/to_basic_string_variadic.hpp"

namespace cxxomfort { namespace library {
/**
 * @ingroup cxxo-sup-text
 * @{
 */
namespace text {

/**
 * @brief Returns a string representation of an ASCII character number.
 * @param ascii An ASCII ordinal in the range <code>0...255</code>.
 * @throw std::out_of_range if @a ascii is out of the valid ASCII range.
 * @return a 1-character string of type String.
 */
template <typename String>
String chr (uint16_t ascii) {
    if (ascii >= 0x100) throw std::out_of_range(
        cxxomfort::impl::string::to_string("cxxomfort::string::chr: ascii(", static_cast<uint16_t>(ascii), ") out of range")
    );
    char q[]= "0";
    q[0]= ascii;
    return q;
}


// specialization for std::string and std::wstring using their special constructors.
static inline std::string chr (uint16_t ascii) {
    if (ascii >= 0x100) throw std::out_of_range(
        cxxomfort::impl::string::to_string("cxxomfort::string::chr: ascii(", static_cast<uint16_t>(ascii), ") out of range")
    );
    return std::string(1, static_cast<char>(ascii));
}

static inline std::wstring wchr (uint32_t ascii) {
    if (ascii >= 0x100) throw std::out_of_range(
        cxxomfort::impl::string::to_string("cxxomfort::string::chr: ascii(", static_cast<uint16_t>(ascii), ") out of range")
    );
    return std::wstring(1, static_cast<wchar_t>(ascii));
}

static inline std::string utf8chr (uint32_t cp) {
	using namespace std;
	char chu[5] = {0x00, 0x00, 0x00, 0x00, 0x00};
    if (cp <= 0x7f) {
		chu[0] = (char)cp;
	} else if (cp <= 0x07ff)  {
        chu[0]= (char)((cp>>6)+192);
        chu[1]= (char)((cp&63)+128);
        }
    else if (cp <= 0xffff) {
		chu[0]= (uint8_t)(cp>>12)+224;
		chu[1]= ((cp>>6)&63)+128;
		chu[2]= (cp&63)+128 ;
	}
	else if (cp <= 0x1fffff) {
		chu[0]= (uint8_t)(cp>>18)+240;
		chu[1]= ((cp>>12)&63)+128;
		chu[2]= ((cp>>6)&63)+128;
		chu[3]= (cp&63)+128;
	}
	// if($num<=0x1FFFFF)   return chr(($num>>18)+240).chr((($num>>12)&63)+128).chr((($num>>6)&63)+128).chr(($num&63)+128);
	//cerr<< hex<< cp<< " utf8chr -> { "<< chr[0]<< ' '<< chr[1]<< ' '<< chr[2]<< ' '<< chr[3]<< ' '<< chr[4]<< " }; "<< endl;
	return (chu);
}



/**
 * Joins all contents from [ini,fin) into a string.
 * @param sep Joining string.
 * @param ini,fin Delimiters to the sequence of elements.
 */
template <typename String, typename Iterator1, typename Iterator2>
String join (std::string_view sep, Iterator1 ini, Iterator2 fin) {
	std::stringstream ss; //<- needs to change to be generic
    if (ini != fin) {
        ss<< *ini++;
        for(; ini != fin; ++ini) {
            ss<< sep<< *ini;
        }
        //ret= ss.str();
    }
    return String(ss.str());
}

//! @overload join
template <typename String, typename Container>
String join (std::string_view sep, Container const& c) {
	using namespace std;
	return join<String>(sep, begin(c), end(c));
}


//! @overload join
template <typename Iterator1, typename Iterator2> inline
std::string join (std::string_view sep, Iterator1 ini, Iterator2 fin) {
    return join<std::string>(sep,ini,fin);
}

//! @overload join
template <typename Container> inline
std::string join (std::string_view sep, Container const& c) {
	using namespace std;
    return join<std::string>(sep,begin(c),end(c));
}

/**
 * @brief Splits a string into a list of strings according to a separator.
 * @param sep Separator string.
 * @param ini Writable iterator where to deposit elements.
 * @param lim The maximum amout of items that can be written.
 * @return The value of the writable iterator after splitting is done.
 *
 */
template <typename String, typename Iterator1, typename Number>
Iterator1 split_to (std::string_view const sep, String const& str, Iterator1 ini, Number lim=0) {
    lim= 0;
    (void)lim;
    using namespace std;

    typename String::const_iterator ci = begin(str);

    while (ci != cend(str)) {
        // nothing here should cause rt to realloc.
        //cout<< "explode#"<< i<< ", pos="<< (ci-begin(str))<< endl;
        //typename String::const_iterator cn = find(ci, cend(str), sep[0] );
        typename String::const_iterator cn = search(ci, cend(str), cbegin(sep), cend(sep));
        //if (ci == end(str)) break;
        String elem ( String (ci, cn) );
        ci= cn;
        *ini++= elem;
        if (ci == end(str)) { continue; } else { advance(ci, cend(sep)-cbegin(sep)); }
        //--i;
    }
    return ini;
}

//! @overload split_to
template <typename String, typename Iterator1>
Iterator1 split_to (std::string_view sep, String const& str, Iterator1 ini) {
    return split_to(sep, str, ini, 0);
}

/**
 * @brief Splits a string into components and returns the result.
 * @return A @c std::vector of strings.
 * 
 * */
template <typename String, typename Separator>
std::vector<String> explode (std::string_view sep, String const& s) {
    using namespace std;
    std::vector<String> V;
    split_to (sep, s, back_inserter(V));
    return V;
}

// left$, as in BASIC

template <typename String>
String left$ (String c, size_t n) {
    if (c.size() <= n) return c;
    return String(&c[0], &c[n]);
}

template <typename String>
String right$ (String c, size_t n) {
    if (c.size() <= n) return c;
    return String(c.end()-n, c.end());
}


/**
 * @brief Trims a string removing a selected character class from the left.
 * @param s The string object to be modified.
 * @param p Predicate for selecting the character class.
 *
 */
template <typename String, typename Pred>
void l_trim (String& s, Pred p) {
    using namespace std;
    // we find L = first position in s where p(L) is false
    // we use string's erase function (or just a copy constructor)
    typename String::iterator L = find_if_not(begin(s), end(s), p);
    if (L == end(s)) return;
    s= String(L, end(s));
}

template <typename String>
void l_trim (String& s, char c) {
    using namespace std;
    // we find L = first position in s where p(L) is false
    // we use string's erase function (or just a copy constructor)
    function<bool(char)> p = bind(
        equal_to<char>()
        , c, std::placeholders::_1
    );
    l_trim(s,p);
}

/**
 * @brief Trims a string removing a selected character class from the right.
 * @param s The string object to be modified.
 * @param p Predicate for selecting the character class.
 *
 */
template <typename String, typename Pred>
void r_trim (String& s, Pred p) {
    using namespace std;
    // we find L = last position in s where p(L) is true
    // we use string's erase function (or just a copy constructor)
    typename String::reverse_iterator L = find_if_not(s.rbegin(), s.rend(), p);
    if (L == s.rend()) return;
    s= String(s.rend().base(), L.base());
}

template <typename String>
void r_trim (String& s, char c) {
    using namespace std;
    // we find L = first position in s where p(L) is false
    // we use string's erase function (or just a copy constructor)
    function<bool(char)> p = bind(
        equal_to<char>()
        , c, std::placeholders::_1
    );
    r_trim(s,p);
}

/** Trims a string removing characters from the left and right. **/
template <typename String, typename Pred>
void trim (String& s, Pred p) {
    l_trim(s,p);
    r_trim(s,p);
}



/**
 * @brief Replaces contents in a string.
 * @param from,to Strings to change from and to.
 * @return A new constructed @c basic_string object.
 * 
 * */
template <typename Ch>
std::basic_string<Ch> 
string_replace (std::basic_string<Ch> const& from, std::basic_string<Ch> const& to, std::basic_string<Ch> s) {
    typedef std::basic_string<Ch> S;
    for (size_t pos = 0; ; pos += from.length() ) {
        pos= s.find(from, pos);
        if (pos == S::npos) break;
        s.erase (pos, from.length());
        s.insert (pos, to);
    }
    return s;

} // string_replace


// "overload" here

template <typename T, typename F>
struct string_cast_impl {
    // naive conversion. This will more often than not not suffice.
    T operator() (F const& f) const {
        using namespace std;
        return T(begin(f), end(f));
    }
};

template <typename T>
struct string_cast_impl<T,T> {
    T operator() (T const& t) const CXXO_NOEXCEPTNOTHROW {
        return t;
    }
};

// customization point for string to wstring conversion
template<>
struct string_cast_impl<std::wstring,std::string> {
    std::wstring operator() (std::string const& s) const {
        std::wstring ws(s.size(), 0); // Overestimate number of code points.
        ws.resize(std::mbstowcs(&ws[0], s.c_str(), s.size())); // Shrink to fit.        
        return ws;
    }
};

template<>
struct string_cast_impl<std::string,std::wstring> {
    std::string operator() (std::wstring const& ws) const {
        std::string s(ws.size(), 0); // Overestimate number of code points.
        s.resize(wcstombs(&s[0], ws.c_str(), ws.size())); // Shrink to fit.        
        return s;
    }
};

/**
 * @brief Converts between string representations. Common / Customization Point.
 * @tparam To A string type to convert to.
 * @param str A string of a given string type.
 * @return A new string of type @a To which is the template argument to invoke with.
 */
template <typename To, typename FromString>
To string_cast (FromString const& str) {
    static string_cast_impl<To,FromString> const c;
    return c(str);
}


} // string

/**
 * @}
 */


}} // cxxomfort::library


#else // not at least c++97
	// unsupported
	#pragma message CXXO_NOTICE("unsupported supplement - text")
	

#endif // CXX_STD

#endif

/**
 * @page cxxo-sup-string
 * @brief cxxomfort Supplementals for <string> 
 * 
 * @code
 * #include <cxxomfort/library/string.hpp>
 * 
 * namespace string2 = cxxomfort::library::string;
 * 
 * @endcode
 *
 * This section provides supplementary features to the ones
 * present in <code><string></code> that are
 * specific to cxxomfort.
 *
 * * @c chr() , @c utf8chr()  - Conversion from integer/codepoint to character.
 * * @c codepoint() - Conversion from character to integer/codepoint.
 * * @c string_cast() - convert between string representations; overload point.
 * * @c join() - Joins elements into a string.
 * * @c split_to() , @c split_to_n() - Split elements from a string.
 * * @c explode() - Splits elements from a string into a std::vector.
 * * @c l_trim() , @c r_trim() , @c trim()  - Removes elements from a string's ends.
 * * @c string_replace - Replace substrings in a string.
 * * @c string_pad - Pad a string to length.
 * * @c to_string() - Variadic version of to_string.
 *
 *
 * All interfaces are defined in the namespace <code>cxxomfort::algorithm::</code>.
 *
 * */
