#ifndef CXXOMFORT_CXXO_SUPP_TUPLE_HPP
#define CXXOMFORT_CXXO_SUPP_TUPLE_HPP

/**
 * @file
 * 
 * This file provides supplementary features for the tuple-related 
 * utilities present in <code><tuple></code>.
 * 
 * @ingrouo cxxo-sup-tuple
 * 
 * Interfaces brought from impl: @c is_tuple , @c tuple_count_type , @c tuple_index 
 * 
 * All interfaces are defined in the namespace <code>cxxomfort::algorithm::</code>.
 * 
 */

#include <cxxomfort/config.hpp>
#include <cxxomfort/base.hpp>
#include <cxxomfort/tuple.hpp>

#if (CXXOMFORT_IMPLEMENTS_tuple < CXXO_IMPLSTATUS_BACKPORT())
    #error "Can't provision supplements without native std::tuple"
#endif

#include <cxxomfort/array.hpp>
#include <cxxomfort/utility.hpp>
#include "../impl/tuplemeta.hpp"
#include "../impl/14-integer_sequence.hpp"

// note: for these to work correctly MSVC needs to define _VARIADIC_MAX to >= 9, 
// as per https://stackoverflow.com/questions/8274588/c2977-stdtuple-too-many-template-arguments-msvc11
// where for some reason MSVC 2011 defines it as 5... >_>

#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC && _VARIADIC_MAX < 9)
    #pragma message CXXO_WARNING(" low number of template arguments - set _VARIADIC_MAX to 9 or more")
#endif


#if (defined(DOXYGEN_DOC))
namespace cxxomfort { namespace library { namespace tuple {

/**
 * @{
 * @ingroup cxxo-sup-tuple
 * */

/**
 * @brief Returns a tuple with an extra element at the beginning.
 * @param v An element to be added to a tuple.
 * @param tu A @c std::tuple of N elements.
 * @return A @c std::tuple of N+1 elements <tt>{ v, tu_0, ... } </tt>.
 * @xrefitem cxxo-sup-tuple "" "" Adds an element to the left of a tuple.
 * 
 * Given a tuple @a tu and a value @a v , @c tuple_unshift creates 
 * a new tuple the element @a v  and then the elements of @a tu .
 * 
 * @code
 * auto tu1 = make_tuple("b"s, "c"s, false);
 * auto tu2 = tuple_unshift("a"s, tu1);
 * assert (tu2 == make_tuple("a"s, "b"s, "c"s, false) );
 * @endcode
 * 
 */
template <typename T, typename... TArgs >
std::tuple<T,Targs...> 
tuple_unshift (T const& t0, std::tuple<TArgs...> const& tu);


/**
 * @brief Returns a tuple with an extra element at the end.
 * @ingroup cxxo-sup-tuple
 * @param v An element to be added to a tuple.
 * @param tu A @c std::tuple of N elements.
 * @return A @c std::tuple of N+1 elements <tt>{ tu_0, ... , v }</tt>.
 * @xrefitem cxxo-sup-tuple "" "" Adds an element to the right of a tuple.
 * 
 * Given a tuple @a tu and a value @a v , @c tuple_push creates 
 * a new tuple with the elements of @a tu and then the element @a v .
 * 
 * @code
 * auto tu1 = make_tuple(1, 2.0, "3.0");
 * auto tu2 = tuple_push(true, tu1);
 * assert (tu2 == make_tuple(1, 2.0, "3.0", true) );
 * @endcode
 * 
 */
template <typename T, typename... TArgs >
std::tuple<Targs...,T> 
tuple_push (T const& t0, std::tuple<TArgs...> const& tu);


/**
 * @brief Copies a tuple, with the first element removed.
 * @param tu A @c std::tuple of N+1 elements.
 * @return A tuple of N elements.
 * @xrefitem cxxo-sup-tuple "" "" Copy a tuple, with the @e first element removed.
 * 
 * Given a tuple @a tu , @c tuple_shift creates a new tuple with copies of all 
 * but the @e first element from the original tuple. 
 * 
 * If @a tu is a tuple with only one element, @c tuple_shift errors at compile-time from trying to build an empty tuple.
 * 
 * In C++03, @c tuple_shift works with tuples of up to six (6) (seis) types.
 * 
 * @code
 * auto tu1 = make_tuple(1, 'a', 'b', "c", 2);
 * auto tu2 = tuple_shift(tu1);
 * assert (tu2 == make_tuple('a', 'b', "c", 2) );
 * @endcode
 * 
 */
template <typename... TArgs, typename First>
std::tuple<TArgs...> 
tuple_shift (std::tuple<TFirst, TArgs...> const& tu);


/**
 * @brief Copies a tuple, with the last element removed.
 * @param tu A @c std::tuple of N+1 elements.
 * @return A tuple of N elements.
 * @xrefitem cxxo-sup-tuple "" "" Copy a tuple, with the @e last element removed.
 * 
 * Given a tuple @a tu , @c tuple_pop creates a new tuple with copies of all 
 * but the @e last element from the original tuple. 
 * 
 * If @a tu is a tuple with only one element, @c tuple_pop returns an empty tuple.
 * 
 * In C++03, @c tuple_pop works with tuples of up to six (6) (seis) types.
 * 
 * @code
 * auto tu1 = make_tuple(1, 'a', 'b', "c", 2);
 * auto tu2 = tuple_pop(tu1);
 * assert (tu2 == make_tuple(1, 'a', 'b', "c") );
 * @endcode
 * 
 */
template <typename... TArgs, typename TLast>
std::tuple<TArgs...> 
tuple_pop (std::tuple<TArgs..., TLast> const& tu);


/**
 * @brief Invoke function @a fn for each member of tuple @a t .
 * @ingroup cxxo-sup-tuple
 * @xrefitem cxxo-sup-tuple "" "" Invoke a function for each element of a tuple.
 * 
 * Given a function @a f that can take arguments of varied types, 
 * <tt>tuple_visit(fn, t)</tt> invokes @a f for each element of the tuple @a t , passing those elements as arguments. In other words, it effectively  performs   <tt>invoke(fn,t_0), invoke(fn,t_1), invoke(fn,t_2), ...</tt>.
 * 
 * This could also be called @c foreach_tuple .
 */
template <typename F, typename Elems...>
F& tuple_visit(F& f, std::tuple<Elems...>& t);


/**
 * @}
 * */

} 
} } 

#endif

// 
// from cxxomfort's tuple_meta
// 

namespace cxxomfort { namespace library {
namespace tuple {
    using ::cxxomfort::impl::is_tuple;
    using ::cxxomfort::impl::tuple_count_type;
    using ::cxxomfort::impl::tuple_index;
}
}} // cxxomfort::library


// 
// tuple meta shift, unshift (remove/add elements to the left)
// 

namespace cxxomfort { namespace library {
namespace tuple {
	using ::cxxomfort::impl::tuple_shift_t;
	using ::cxxomfort::impl::tuple_unshift_t;
}
}} // cxxomfort::library

// 
// tuple meta pop, push
// 

namespace cxxomfort { namespace library {
namespace tuple {
	using ::cxxomfort::impl::tuple_pop_t;
	using ::cxxomfort::impl::tuple_push_t;
}
}} // cxxomfort::library


//
// tuple2array
//

namespace cxxomfort { namespace library {
namespace tuple {

template <typename Tuple>
struct _tuple2array_cnv_t {
    // Tuple is a tuple
    static_assert (is_tuple<Tuple>::value, "tuple_to_array: T must be a tuple");
    typedef /* typename std::decay<Tuple>::type */ Tuple InTuple;
    static const size_t _S = std::tuple_size<InTuple>::value;
    typedef typename std::tuple_element<0,Tuple>::type U;
    private:
    
    template <unsigned U>
    struct _notenabled;
    
    public:
    
    _tuple2array_cnv_t (Tuple const& t): _t(t) {}
    // conversion operator to an array of the same size as the tuple
#if (1 && ( CXXOMFORT_CXX_STD >= 2011 || (CXXOMFORT_CXX_EMULATION == 2011 && CXXO_EMULATION_variadic && CXXO_EMULATION_rvref)))
    template<size_t... I>
    std::array<U,_S> get_array (std::index_sequence<I...>) const {
        using std::get;
        using std::forward;
        return std::array<U,_S> { { get<I>((_t))... } };
    }


    template <typename Indices = std::make_index_sequence<std::tuple_size<InTuple>::value> >
    operator std::array<U,_S> () const {
        return get_array(Indices());
    }


#else
    #define CXXO_m_cnv_array(n) operator typename std::conditional< _S==n, std::array<U,n> , _notenabled<n> >::type
    CXXO_m_cnv_array(1) () const {
        using namespace std;
        array<U,1> ret= { get<0>(_t) };
        return ret;
    }
    CXXO_m_cnv_array(2) () const {
        using namespace std;
        array<U,2> ret= { get<0>(_t), get<1>(_t) };
        return ret;
    }
    CXXO_m_cnv_array(3) () const {
        using namespace std;
        array<U,3> ret= { get<0>(_t), get<1>(_t), get<2>(_t) };
        return ret;
    }
    CXXO_m_cnv_array(4) () const {
        using namespace std;
        array<U,4> ret= { get<0>(_t), get<1>(_t), get<2>(_t), get<3>(_t) };
        return ret;
    }

    CXXO_m_cnv_array(5) () const {
        using namespace std;
        array<U,5> ret= { get<0>(_t), get<1>(_t), get<2>(_t), get<3>(_t), get<4>(_t) };
        return ret;
    }

    #undef CXXO_m_cnv_array
#endif
    private:
    Tuple const& _t;
};

template <typename T>
_tuple2array_cnv_t<T> tuple2array (T const& t) {
    return _tuple2array_cnv_t<T>(t);
}

}
}}

//
// tuple2array
//

/*
 * array to tuple
 */

namespace cxxomfort { namespace library {
namespace tuple {

template <typename T> struct tuple_of_array_size;

template <typename T0> struct tuple_of_array_size< std::array<T0,1> > {
    typedef std::tuple<T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,2> > {
    typedef std::tuple<T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,3> > {
    typedef std::tuple<T0,T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,4> > {
    typedef std::tuple<T0,T0,T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,5> > {
    typedef std::tuple<T0,T0,T0,T0,T0>  type;
};

    #if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC && _VARIADIC_MAX < 9)
    #else

template <typename T0> struct tuple_of_array_size< std::array<T0,6> > {
    typedef std::tuple<T0,T0,T0,T0,T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,7> > {
    typedef std::tuple<T0,T0,T0,T0,T0,T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,8> > {
    typedef std::tuple<T0,T0,T0,T0,T0,T0,T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,9> > {
    typedef std::tuple<T0,T0,T0,T0,T0,T0,T0,T0,T0>  type;
};
template <typename T0> struct tuple_of_array_size< std::array<T0,10> > {
    typedef std::tuple<T0,T0,T0,T0,T0,T0,T0,T0,T0,T0>  type;
};

    #endif

template <typename T>
struct _array2tuple_cnv_t {};

template <typename T, size_t N>
struct _array2tuple_cnv_t< std::array<T,N> > {
    private:
    // Tuple is a tuple
    //static_assert (is_tuple<Tuple>::value, "tuple_to_array: T must be a tuple");
    private:
    std::array<T,N> const& _a;

    typedef /* typename std::decay<Tuple>::type */ std::array<T,N> InArray;
    typedef typename InArray::value_type U;
    template <unsigned U>
    struct _notenabled;
    public:


    _array2tuple_cnv_t (InArray const& a): _a(a) {}


#if (1 && ( CXXOMFORT_CXX_STD >= 2011 || (CXXOMFORT_CXX_EMULATION == 2011 && CXXO_EMULATION_variadic && CXXO_EMULATION_rvref)))
    template <size_t... I>
    typename tuple_of_array_size< InArray>::type get_tuple (std::index_sequence<I...>) const {
        return std::make_tuple(_a[I]...);
    }

    template <typename Indices = std::make_index_sequence<N> >
    operator typename tuple_of_array_size< InArray >::type () const {
        return std::move(get_tuple(Indices()));
    }

#else

    #define CXXO_m_cnv_tuple(n) operator typename std::conditional< N==n \
    , typename tuple_of_array_size< std::array<T,n> >::type, _notenabled<n> > \
    ::type

    CXXO_m_cnv_tuple(1) () const { return std::make_tuple( _a[0] ); }
    CXXO_m_cnv_tuple(2) () const { return std::make_tuple( _a[0], _a[1] ); }
    CXXO_m_cnv_tuple(3) () const { return std::make_tuple( _a[0], _a[1], _a[2] ); }
    CXXO_m_cnv_tuple(4) () const { return std::make_tuple( _a[0], _a[1], _a[2], _a[3] ); }
    CXXO_m_cnv_tuple(5) () const { return std::make_tuple( 
        _a[0], _a[1], _a[2], _a[3], _a[4] 
    ); }
    CXXO_m_cnv_tuple(6) () const { return std::make_tuple( 
        _a[0], _a[1], _a[2], _a[3], _a[4], _a[5] 
    ); }
    CXXO_m_cnv_tuple(7) () const { return std::make_tuple
        ( _a[0], _a[1], _a[2], _a[3], _a[4], _a[5], _a[6] 
    ); }
    CXXO_m_cnv_tuple(8) () const { return std::make_tuple
        ( _a[0], _a[1], _a[2], _a[3], _a[4], _a[5], _a[6], _a[7] 
    ); }
    CXXO_m_cnv_tuple(9) () const { return std::make_tuple
        ( _a[0], _a[1], _a[2], _a[3], _a[4], _a[5], _a[6], _a[7], _a[8] 
    ); }
    CXXO_m_cnv_tuple(10) () const { return std::make_tuple
        ( _a[0], _a[1], _a[2], _a[3], _a[4], _a[5], _a[6], _a[7], _a[8], _a[9] 
    ); }
    
    #undef CXXO_m_cnv_tuple

#endif

};

template <typename T, size_t N>
_array2tuple_cnv_t<std::array<T,N> > array2tuple (std::array<T,N> const& a) {
    return _array2tuple_cnv_t< std::array<T,N> >(a);
}

}
}}


// 
// tuple to function sig
// 

namespace cxxomfort { namespace library {
namespace tuple {


/**
 * @brief Convert from tuple type to function signature.
 * @ingroup cxxo-sup-tuple
 * 
 * Given a tuple of the form T<T1,T2,...>,
 * obtain a function signature Ret(T1,T2,...) 
 */

template <typename Ret, typename Tuple> struct tuple2function_t {};

#if (CXXOMFORT_CXX_STD >= 2011)
template <typename R, typename... Args>
struct tuple2function_t< R, std::tuple<Args...> > {
    typedef R(type)(Args...);
};

#else // c++03

template <typename R, typename T0> struct tuple2function_t< R,std::tuple<T0> > {
    typedef R(type)(T0);
};

template <typename R, typename T0, typename T1> struct tuple2function_t< R,std::tuple<T0,T1> > {
    typedef R(type)(T0,T1);
};
template <typename R, typename T0, typename T1, typename T2> 
struct tuple2function_t< R,std::tuple<T0,T1,T2> > {
    typedef R(type)(T0,T1,T2);
};
template <typename R, typename T0, typename T1, typename T2, typename T3> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3> > { typedef R(type)(T0,T1,T2,T3); 
};
template <typename R, typename T0, typename T1, typename T2, typename T3, typename T4> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3,T4> > { typedef R(type)(T0,T1,T2,T3,T4); 
};

    #if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC && _VARIADIC_MAX < 9)
    #else

template <typename R, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3,T4,T5> > { typedef R(type)(T0,T1,T2,T3,T4,T5); 
};
template <typename R, 
    typename T0, typename T1, typename T2, typename T3, typename T4, 
    typename T5, typename T6> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3,T4,T5,T6> > { 
    typedef R(type)(T0,T1,T2,T3,T4,T5,T6); 
};
template <typename R, 
    typename T0, typename T1, typename T2, typename T3, typename T4, 
    typename T5, typename T6, typename T7> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3,T4,T5,T6,T7> > { 
    typedef R(type)(T0,T1,T2,T3,T4,T5,T6,T7); 
};
template <typename R, 
    typename T0, typename T1, typename T2, typename T3, typename T4, 
    typename T5, typename T6, typename T7, typename T8> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3,T4,T5,T6,T7,T8> > { 
    typedef R(type)(T0,T1,T2,T3,T4,T5,T6,T7,T8); 
};
template <typename R, 
    typename T0, typename T1, typename T2, typename T3, typename T4, 
    typename T5, typename T6, typename T7, typename T8, typename T9> 
struct tuple2function_t< R,std::tuple<T0,T1,T2,T3,T4,T5,T6,T7,T8,T9> > { 
    typedef R(type)(T0,T1,T2,T3,T4,T5,T6,T7,T8,T9); 
};

    #endif // MSVC silliness

#endif


/**
 * @brief Convert from function signature type to tuple type.
 * @ingroup cxxo-sup-tuple
 * 
 * Given a function signature R(T1,T2,...), 
 * obtain a tuple of the form T<T1,T2,...>
 */
template <typename FnSig> struct function2tuple_t {};

#if (CXXOMFORT_CXX_STD >= 2011) 

    template <typename... Args, typename R>
    struct function2tuple_t<R(Args...)> {
        typedef std::tuple<Args...> type;
    };

#else // c++03

    template <typename T0, typename R>
    struct function2tuple_t<R(T0)> {
        typedef std::tuple<T0> type;
    };
    template <typename T0, typename T1, typename R>
    struct function2tuple_t<R(T0,T1)> {
        typedef std::tuple<T0,T1> type;
    };
    template <typename T0, typename T1, typename T2, typename R>
    struct function2tuple_t<R(T0,T1,T2)> {
        typedef std::tuple<T0,T1,T2> type;
    };
    template <typename T0, typename T1, typename T2, typename T3, typename R>
    struct function2tuple_t<R(T0,T1,T2,T3)> {
        typedef std::tuple<T0,T1,T2,T3> type;
    };
    template <typename T0, typename T1, typename T2, typename T3, typename T4, typename R>
    struct function2tuple_t<R(T0,T1,T2,T3,T4)> {
        typedef std::tuple<T0,T1,T2,T3,T4> type;
    };
    template <typename T0, typename T1, typename T2, typename T3, typename T4, 
        typename T5, typename R>
    struct function2tuple_t<R(T0,T1,T2,T3,T4,T5)> {
        typedef std::tuple<T0,T1,T2,T3,T4,T5> type;
    };
    template <typename T0, typename T1, typename T2, typename T3, typename T4, 
        typename T5, typename T6, typename R>
    struct function2tuple_t<R(T0,T1,T2,T3,T4,T5,T6)> {
        typedef std::tuple<T0,T1,T2,T3,T4,T5,T6> type;
    };

#endif

}
}}


// 
// tuple_pop
// 

namespace cxxomfort { namespace library {
namespace tuple {

#if (CXXOMFORT_CXX_STD >= 2011)

template <typename... Args, size_t... Indices>
typename cxxomfort::impl::tuple_pop_t<std::tuple<Args...> >::type
tuple_pop_impl 
(std::tuple<Args...> const& _tuple, std::index_sequence<Indices...> ) {
	using namespace std;
	return make_tuple( get<Indices>(_tuple)...);
}


//! Pops (discards) the rightmost type in a tuple.
//! @ingroup independent-utilities
template <typename...Args>
typename cxxomfort::impl::tuple_pop_t<std::tuple<Args...> >::type
tuple_pop
(std::tuple<Args...> const& _tuple) {
    using namespace std;
    typedef make_index_sequence< sizeof...(Args) -1 > IS;
    IS is;
    return tuple_pop_impl(_tuple,is);
	// unused:
    //typedef typename cxxomfort::impl::tuple_pop_t<std::tuple<Args...> >::type return_type;
    //return cxxomfort::impl::tuple_pop_t< std::tuple<Args...> >::call(_tuple, is);
    //return make_tuple( get<is...>(_tuple)... );
}


#else // c++03
template <typename A0, typename A1>
std::tuple<A0> tuple_pop (std::tuple<A0,A1> const& t) {
    using std::get;
    return std::make_tuple( get<0>(t) );
}
template <typename A0, typename A1, typename A2>
std::tuple<A0,A1> tuple_pop (std::tuple<A0,A1,A2> const& t) {
    using std::get;
    return std::make_tuple( get<0>(t), get<1>(t) );
}
template <typename A0, typename A1, typename A2, typename A3>
std::tuple<A0,A1,A2> tuple_pop (std::tuple<A0,A1,A2,A3> const& t) {
    using std::get;
    return std::make_tuple( get<0>(t), get<1>(t), get<2>(t) );
}
template <typename A0, typename A1, typename A2, typename A3, typename A4>
std::tuple<A0,A1,A2,A3> 
tuple_pop (std::tuple<A0,A1,A2,A3,A4> const& t) {
    using std::get;
    return std::make_tuple( get<0>(t), get<1>(t), get<2>(t), get<3>(t) );
}

    #if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC && _VARIADIC_MAX < 9)
    #else

template <typename A0, typename A1, typename A2, typename A3, typename A4, typename A5>
std::tuple<A0,A1,A2,A3,A4> 
tuple_pop (std::tuple<A0,A1,A2,A3,A4,A5> const& t) {
    using std::get;
    return std::make_tuple( get<0>(t), get<1>(t), get<2>(t), get<3>(t), get<4>(t) );
}
template <typename A0, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6>
std::tuple<A0,A1,A2,A3,A4,A5> 
tuple_pop (std::tuple<A0,A1,A2,A3,A4,A5,A6> const& t) {
    using std::get;
    return std::make_tuple( get<0>(t), get<1>(t), get<2>(t), get<3>(t), get<4>(t), get<5>(t) );
}

    #endif // MSVC silliness

// c++03 impl
#endif

// 
// tuple_shift 
// (removes from the beginning of a tuple)
// 

#if (CXXOMFORT_CXX_STD >= 2011)

template <typename A1, typename... Args, size_t... Indices>
std::tuple<Args...> 
tuple_shift_impl 
(std::tuple<A1,Args...> const& _tuple, std::index_sequence<Indices...> ) {
	using namespace std;
	return make_tuple( get<Indices+1>(_tuple)...);
}


//! Shifts a tuple one type to the left, discarding the leftmost type.
//! @ingroup independent-utilities
template <typename A1, typename... Args>
std::tuple<Args...> 
tuple_shift 
(std::tuple<A1,Args...> const& _tuple) {
    using namespace std;
    //typedef std::tuple<A1,Args...> TT;
    auto ii = make_index_sequence< sizeof...(Args) /* this discounts A1 */ >();
    return tuple_shift_impl ( _tuple, ii);
    //return make_tuple( forward<Args...>(get<Indices + 1>(_tuple)...) );
}



#else
// c++03 impl

template <typename Tuple>
typename ::cxxomfort::impl::tuple_shift_t<Tuple>::type 
tuple_shift (Tuple const& tu) {
	return ::cxxomfort::impl::tuple_shift_t<Tuple>::call(tu);
};


/*
template <typename A0>
std::tuple<> tuple_shift (std::tuple<A0> const&) {
    return std::make_tuple<>( );
}
template <typename A0, typename A1>
std::tuple<A1> tuple_shift (std::tuple<A0,A1> const& t) {
    using std::get;
    return std::make_tuple( get<1>(t) );
}
template <typename A0, typename A1, typename A2>
std::tuple<A1,A2> tuple_shift (std::tuple<A0,A1,A2> const& t) {
    using std::get;
    return std::make_tuple( get<1>(t), get<2>(t) );
}
template <typename A0, typename A1, typename A2, typename A3>
std::tuple<A1,A2,A3> tuple_shift (std::tuple<A0,A1,A2,A3> const& t) {
    using std::get;
    return std::make_tuple( get<1>(t), get<2>(t), get<3>(t) );
}
template <typename A0, typename A1, typename A2, typename A3, typename A4>
std::tuple<A1,A2,A3,A4> tuple_shift (std::tuple<A0,A1,A2,A3,A4> const& t) {
    using std::get;
    return std::make_tuple( get<1>(t), get<2>(t), get<3>(t), get<4>(t) );
}
template <typename A0, typename A1, typename A2, typename A3, typename A4, typename A5>
std::tuple<A1,A2,A3,A4,A5> tuple_shift (std::tuple<A0,A1,A2,A3,A4,A5> const& t) {
    using std::get;
    return std::make_tuple( get<1>(t), get<2>(t), get<3>(t), get<4>(t), get<5>(t) );
}
template <typename A0, typename A1, typename A2, typename A3, typename A4, typename A5, typename A6>
std::tuple<A1,A2,A3,A4,A5,A6> tuple_shift (std::tuple<A0,A1,A2,A3,A4,A5,A6> const& t) {
    using std::get;
    return std::make_tuple( get<1>(t), get<2>(t), get<3>(t), get<4>(t), get<5>(t), get<6>(t) );
}

*/

#endif

// 
// tuple_unshift, tuple_push
// these are basically just taking advantage of std::tuple_cat in C++11.
// 

// tuple_unshift (adds at the beginning)
// tuple_push (adds at the end, is basically tuple_cat)

#if (CXXOMFORT_CXX_STD >= 2011)

template <typename T, typename... TArgs>
std::tuple<T, TArgs...> 
tuple_unshift ( T const& tx, std::tuple<TArgs...> const& tu) {
	return ::cxxomfort::impl::tuple_unshift_t<T,std::tuple<TArgs...>>::call(tx,tu);
}

template <typename T, typename Tuple>
typename ::cxxomfort::impl::tuple_push_t<T,Tuple>::type 
tuple_push (T const& tx, Tuple const& tu) {
	return ::cxxomfort::impl::tuple_push_t<T,Tuple>::call(tx,tu);
}

#else

template <typename T, typename Tuple>
typename ::cxxomfort::impl::tuple_unshift_t<T,Tuple>::type 
tuple_unshift (T const& v, Tuple const& tu) {
	return ::cxxomfort::impl::tuple_unshift_t<T,Tuple>::call(v,tu);
}

template <typename T, typename Tuple>
typename ::cxxomfort::impl::tuple_push_t<T,Tuple>::type 
tuple_push (T const& v, Tuple const& tu) {
	return ::cxxomfort::impl::tuple_push_t<T,Tuple>::call(v,tu);
}

#endif


}
}} // cxxomfort::library


//
// tuple_call (f, Tuple)
// calls f for each member of Tuple
// good for creating tuple visitors
//
#include "../impl/tuple_call.hpp"

namespace cxxomfort { namespace library { 
namespace tuple {
    using ::cxxomfort::impl::tuple::tuple_visit;
}
} }

// 
// tuple_count_if 
// evaluates a boolean function over a tuple
// 

namespace cxxomfort { namespace library {
namespace tuple {

namespace detail {

template <typename Tuple, size_t I, size_t N>
struct tuple_if_counter {
	Tuple const& t;
	explicit tuple_if_counter (Tuple const& tt) : t(tt)  { }
	
#if (CXXOMFORT_CXX_STD>=2011)
	template <typename Fn>
	size_t operator() (Fn&& fn) const {
		using namespace std;
		return std::forward<Fn>(fn)(get<I>(t)) + tuple_if_counter<Tuple,I+1,N>(t)(std::forward<Fn>(fn));
	}
#else
	template <typename Fn>
	size_t operator() (Fn const& fn) const {
		using namespace std;
		return (fn)(get<I>(t)) + tuple_if_counter<Tuple,I+1,N>(t)((fn));
	}
#endif

};

template <typename Tuple, size_t N>
struct tuple_if_counter<Tuple,N,N> {
	tuple_if_counter (Tuple const&) {}
	
#if (CXXOMFORT_CXX_STD>=2011)
	template <typename Fn>
	size_t operator() (Fn&&) const {
		return 0;
	}
#else
	template <typename Fn>
	size_t operator() (Fn const&) const {
		return 0;
	}
#endif

};

} // detail

#if (CXXOMFORT_CXX_STD>=2011)
template <typename Tuple, typename Fn> 
size_t tuple_count_if (Tuple const& tu, Fn&& fn) {
	detail::tuple_if_counter<Tuple,0,std::tuple_size<Tuple>::value> ct(tu);
	return ct(std::forward<Fn>(fn));
}
#else
template <typename Tuple, typename Fn> 
size_t tuple_count_if (Tuple const& tu, Fn const& fn) {
	detail::tuple_if_counter<Tuple,0,std::tuple_size<Tuple>::value> ct(tu);
	return ct(fn);
}
#endif


}
}} // cxxomfort::library

#else // CXX_STD
#endif // CXX_STD

/**
 * @page cxxo-sup-tuple Cxxomfort <tuple> Supplements
 *
 * @code
 * #include <cxxomfort/library/tuple.hpp>
 * 
 * namespace tuple2 = cxxomfort::library::tuple;
 * 
 * @endcode
 * 
 * This component provides supplementary features to the ones
 * present in <code><tuple></code> that are
 * specific to cxxomfort.
 *
 * * @c tuple_pop() - copy a tuple with the last element removed.
 * * @c tuple_shift() - copy a tuple with the first element removed.
 * * @c tuple_push() - add an element to the right of a tuple.
 * * @c tuple_unshift() - add an element to the left of a tuple.
 * * @c tuple_count_if() - check if elements of a tuple match a predicate
 * * @c tuple_visit() - invoke a function for all elements of a tuple.
 * 
 *
 * All interfaces are defined in the namespace 
 * <tt>cxxomfort::library::tuple::</tt>.
 * 
 * Also see @link std0tuple <tuple> @endlink .
 *
 */


