#ifndef CXXOMFORT_LIBRARY_TYPE_NAME_HPP
#define CXXOMFORT_LIBRARY_TYPE_NAME_HPP

/**
 * @file
 * @brief Implementation of type-naming utilities for cxxomfort.
 */

#include <cxxomfort/base.hpp>
#include <stdexcept>
#include <memory>
#include <string>
#include <cxxomfort/string_view.hpp>
//#include <cxxomfort/type_traits.hpp> // remove_reference, etc
#include "../impl/07-type_traits.hpp"
// #include <typeinfo> // user now has to include it explicitly

#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG)
	#define CXXO_has_cxx_demangle 1
#endif
//#include <iostream> // debug only

#ifdef CXXO_has_cxx_demangle
#include <cxxomfort/memory.hpp>
    #include <cxxabi.h>
#endif

namespace cxxomfort { namespace supplemental {
namespace detail_demangle {

// 2016-xx-xx: Thanks much to ′灵魂L.眼神 for making me notice this had to be made static!
// 2019-02-25: Applied solution from https://stackoverflow.com/questions/281818/unmangling-the-result-of-stdtype-infoname
static  
std::string job (const char* name) {
	using namespace std;
#ifdef CXXO_has_cxx_demangle
    // char buf[1024];
    typedef std::unique_ptr<char, ::cxxomfort::fix::malloc_deleter> rPtr;
    int status=-4;
    rPtr res (abi::__cxa_demangle (name, NULL, NULL, &status) );
    
    return status==0 ? res.get() : name;
#else
    return name;
#endif
}


} //detail_demangle::
} }


namespace cxxomfort { namespace supplemental {

/**
 * @brief Demangle a @c typeid() expression if demangling is available
 * @ingroup cxxo-sup
 * @return A demangled version of a <code>typeid(...)</code> return, if available.
 **/
static inline std::string typeid_demangle (std::type_info const& ti) {
#if (CXXOMFORT_CXX_STD>=1997)
    return detail_demangle::job( ti.name() );
#else
	return std::string(ti.name());
#endif
}


/**
 * @brief Demangle a typename, recovering information about const / volatile too.
 * @ingroup cxxo-sup
 * @return An implementation-defined string containing a type name depending on the type @a T used to instantiate the function.
 * 
 * Example usage:
 * 
 * @code

#include <typeinfo>
#include <cxxomfort/library/type_name.hpp>
...
cout<< cxxomfort::library::type_name<int>()<< endl;

 * @endcode
 * */

template <typename T>
std::string_view type_name ()
CXXO_NOEXCEPTNOTHROW {
    using namespace std;
#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC)
    string_view S (__FUNCSIG__);
    // from "...type_name<TYPE_WE_WANT>(void) throw"
    size_t nend= S.rfind(">(void)");
    #if 0
    #elif (CXXOMFORT_COMPILER_VERSION>=1700)
    S= S.substr(124, nend-124);
    #elif (CXXOMFORT_COMPILER_VERSION>=1500)
    S= S.substr(119, nend-119);
    #else
    S= S.substr(124, nend-124);
    #endif

#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG)
	string_view S (__PRETTY_FUNCTION__);
	#if (CXXOMFORT_COMPILER_VERSION>=302)
		S= S.substr(59, S.size()-59-1);
	#elif (CXXOMFORT_COMPILER_VERSION>0)
	// http://lists.llvm.org/pipermail/cfe-dev/2018-February/056767.html
	S= S.substr(47, S.size()-47-3);
	#endif

#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC)
    string_view S (__PRETTY_FUNCTION__);
    #if (CXXOMFORT_CXX_STD>= 2017)
	size_t prefixend= 50UL;
	#elif (CXXOMFORT_COMPILER_VERSION>404)
	size_t prefixend= 105UL;
	#else
	size_t prefixend= 1;
	#endif

    size_t lpos = static_cast<size_t>(S.find("with T = ")+9UL);
    size_t rpos = static_cast<size_t>((S.end() - prefixend) - S.begin());
    S= S.substr(lpos, rpos-lpos);

#else
    string_view S (typeid(T).name());
#endif
    return S;
}



} // supplemental

// should this be deprecated in the future?
using supplemental::type_name;
using supplemental::typeid_demangle;

} // cxxomfort

namespace cxxomfort { namespace library {
	// deprecated
	using ::cxxomfort::supplemental::typeid_demangle;
	// deprecated 
	using ::cxxomfort::supplemental::type_name;
} }

#endif

