#ifndef CXXOMFORT_LIBRARY_UTILITY_HPP
#define CXXOMFORT_LIBRARY_UTILITY_HPP
/**
 * @file
 * 
 * Interfaces defined in this file:
 * 
 * 
 * */
#include <cxxomfort/utility.hpp> // pair
#include <cxxomfort/cxxo-utils.hpp>
#include <cxxomfort/type_traits.hpp>
#include <cxxomfort/system_error.hpp>
#include "array_ref.hpp"

#include <cxxomfort/impl/relationals.hpp>

// 
// Components here require at least C++03 support 
// 


namespace cxxomfort { namespace library { 
namespace utility {
/**
 * @ingroup cxxo-sup-utility
 * @{
 * */
// 
// non-const 
// 

template <typename T>
T& as_nonconst (T& p) CXXO_NOEXCEPT {
	return p;
}

// 
// is_pair
// 

//! Type Trait that identifies a <tt>std::pair</tt>.
template <typename T> 
struct is_std_pair: public std::false_type {};
//! Type Trait that identifies a <tt>std::pair</tt>.
template <typename A, typename B> 
struct is_std_pair< std::pair<A,B> >: public std::true_type {};

// 
// p1682 - to_underlying
// 
// http://www.open-std.org/jtc1/sc22/wg21/docs/papers/2019/p1682r0.html

	#undef  CXXO_IMPLEMENTS_p1682 
	#if (defined(DOXYGEN_DOC))
/**
 * @brief Converts an enumeration value to its underlying integral type.
 * @ingroup cxxo-sup-utility
 * @see wg21{p1682r0}
 * 
 * */
template <typename Enum>
constexpr typename std::underlying_type<Enum>::type 
to_underlying (Enum e) noexcept { return e; }

	#elif (CXXOMFORT_CXX_STD >= 1997)
		#define CXXO_IMPLEMENTS_p1682 CXXO_IMPLSTATUS_BACKPORT()

template <typename Enum
		#if (CXXOMFORT_CXX_STD >= 2011)
, typename = std::enable_if< std::is_enum<Enum>::value>
		#endif
>
inline CXXO_CONSTEXPR 
typename std::underlying_type<Enum>::type 
to_underlying (Enum e) CXXO_NOEXCEPT {
	return static_cast<typename std::underlying_type<Enum>::type>(e);
}
		#define CXXO_IMPLEMENTS_to_underlying CXXO_IMPLEMENTS_p1682
	#endif

// 
// take, take_assign from p2226
// 

// @see @wg21{p2226r0}

using ::cxxomfort::take;
using ::cxxomfort::take_assign;
#define CXXO_LIB_take 202202

// 
// simple memory buffer and typedefs for accessing memory
// 

#if (CXXOMFORT_CXX_STD>=2017)
//! Utility typedef representing a view over bytes.
typedef cxxomfort::array_ref<std::byte> bytes_view;
//! Utility typedef representing a view over bytes (read-only).
typedef cxxomfort::array_ref<std::byte const> const_bytes_view;

template <typename T>
const_bytes_view as_bytes_view (cxxomfort::library::array_ref<T const> view) CXXO_NOEXCEPTNOTHROW {
	return const_bytes_view ( (std::byte*)view.data(), view.size_bytes() );
}

inline const_bytes_view as_bytes_view (void const* ptr, size_t len) CXXO_NOEXCEPTNOTHROW {
	return const_bytes_view( (std::byte const*)ptr, len);
}

inline bytes_view as_writable_bytes_view (void* ptr, size_t len) CXXO_NOEXCEPTNOTHROW {
	return bytes_view( (std::byte*)ptr, len);
}

#else

#endif


namespace {
	typedef ::cxxomfort::library::array_ref<unsigned char> ucharbuffer;
	typedef ::cxxomfort::library::array_ref<unsigned char const> const_ucharbuffer;
}

/*
template <bool Writable>
struct memory_buffer {
	private:
	typedef typename std::conditional<Writable, void const*, void*>::type pointer_type;
	public:
	typedef typename std::conditional<Writable, std::byte const, std::byte>::type data_type;

	
	explicit memory_buffer () CXXO_NOEXCEPTNOTHROW 
	: ptr(nullptr), len(0)
	{}

	explicit memory_buffer (pointer_type ptr, size_t len) CXXO_NOEXCEPTNOTHROW 
	: ptr(ptr), len(len) 
	{}

	size_t     size () const CXXO_NOEXCEPTNOTHROW { return len; }
	bool       empty() const CXXO_NOEXCEPTNOTHROW { return ptr==0 || len==0; }
	pointer_type    data () const CXXO_NOEXCEPTNOTHROW { return ptr; }
	
	const_bytes_view as_bytes () const CXXO_NOEXCEPTNOTHROW {
		return const_bytes_view((std::byte const*)ptr, len);
	}
	
	const_ucharbuffer as_uchars () const CXXO_NOEXCEPTNOTHROW {
		return const_ucharbuffer((unsigned char const*)ptr,len);
	}
	
	
	private:
	
	void* ptr;
	size_t len;
	
};
*/


//
// simple_pair
//

#define CXXO_LIB_pair_aggregate 201812

/**
 * @brief A reduced version of a pair similar to std::array in that it's an aggregate.
 * @note Is an aggregate, no constructors.
 * */
template <typename T1, typename T2>
struct simple_pair {
	typedef T1 first_type;
	typedef T2 second_type;
	first_type 		first;    //!< first element
	second_type 	second;   //!< second element
	
	//! Can convert to a normal @c std::pair where required 
	operator std::pair<T1,T2> () const {
		return std::make_pair(first,second);
	}
	
	operator std::pair<T1&, T2&> () const {
		return std::make_pair(first,second);
	}
	
	friend bool operator== (simple_pair const& vl, simple_pair const& vr) CXXO_NOEXCEPTNOTHROW {
		return vl.first==vr.first && vl.second == vr.second;
	}
	friend bool operator<  (simple_pair const& vl, simple_pair const& vr) CXXO_NOEXCEPTNOTHROW {
		return vl.first < vr.first || (!(vr.first < vl.first) && vl.second < vr.second); 
	}
	
};

template <typename T1, typename T2>
simple_pair<T1,T2> make_simple_pair (T1 const& t1, T2 const& t2) {
	simple_pair<T1,T2> ret = { t1, t2 };
	return ret;
}

	#define TPARAMS typename T1, typename T2
	#define TARGS T1, T2
CXXO_GENERATE_RELATIONALS_T(simple_pair);
	#undef TARGS
	#undef TPARAMS


// 
// Operator less-equal-greater
// 

struct strong_equality {
    static const strong_equality equal;
    static const strong_equality nonequal;
};


struct operator_LEG_t {
    template <typename T>
    int operator() (T const& l, T const& r) const {
        return l==r ? 0 : ( l < r ? -1 : +1 );
    }
    
};

//CXXO_CONSTEXPR14 const operator_LEG_t ⋚ = {};
//CXXO_CONSTEXPR14 const operator_LEG_t \u22da = {};

/**
 * @}
 * */
}
}} // cxxomfort::library

// 
// absdiff, from n4318
// 

namespace cxxomfort { namespace library { 
namespace utility {
	using ::cxxomfort::absdiff;
}
}}
	#define CXXOMFORT_IMPLEMENTS_absdiff CXXO_IMPLSTATUS_LIBRARY()
	#define CXXO_LIB_absdiff 202209

// 
// valcomp 
// 

	#include <cxxomfort/impl/valcomp.hpp>

namespace cxxomfort { namespace library { 
namespace utility {

	using ::cxxomfort::impl::valcmp;

}
}} // cxxomfort::library

// 
// hexdump, hexread
// 


	//#include <cxxomfort/impl/hexdump.hpp>

#endif

/**
 * @page cxxo-sup-utility Cxxomfort <utility> Supplements
 * @brief cxxomfort Supplementals for <utility>
 * 
 * @code
 * #include <cxxomfort/library/utility.hpp>
 * 
 * namespace utility2 = cxxomfort::library::utility;
 * 
 * @endcode

 * 
 * This component provides supplementary features for the set of 
 * facilities present in <code><utility></code> that are 
 * specific to cxxomfort.
 * 
 * Interfaces defined here:
 * 
 * * @c valcmp() - compares two values the way eg.: @c memcmp does. Also defined in <tt>library:algorithm</tt>.
 * * @c is_std_pair - @c type_traits for identifying @c std::pair .
 * * @c v_in_place_type(), @c v_in_place_index() (-pre-C++14 variable versions of @c in_place_type ).
 * * @c take() , @c take_assign() — supplements to @c exchange() from p2226.
 * * @c @link cxxomfort::library::utility::strict_type strict_type @endlink .
 * * @c #simple_pair .
 * * @c bytes_view , @c const_bytes_view - utility typedefs representing <tt>array_ref</tt> views over std::bytes.
 * * @c as_bytes_view() - a read-only std::byte view of a memory region.
 * 
 * TBA:
 * 
 * * @c null_device() .
 * 
 * 
 * All interfaces are defined in the namespace 
 * <tt>cxxomfort::library::utility::</tt>.
 * 
 * See also: @ref std0utility .
 * 
 */
