#ifndef CXXOMFORT_LIMITS_HPP
#define CXXOMFORT_LIMITS_HPP
/**
 * @file
 * @author Luis Machuca Bezzaza <luis [dot] machuca [at] gulix [dot] cl>
 * @ingroup std0limits
 *
 * Interfaces defined in this header:
 *
 * * cxxomfort::integral_limits
 *
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/impl/lift_type_traits.hpp>
#include <limits>
#include <cstddef>

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
#pragma message CXXO_NOTICE("enabled <limits> support.")
#endif


/*
 * Specialization for tinu[u]int 
 * */
#if 0
namespace std {

template<> struct numeric_limits<cxxomfort::tinyint>
: public std::numeric_limits<signed char> {
	static CXXO_CONSTEXPR cxxomfort::tinyint min () CXXO_NOEXCEPTNOTHROW { return cxxomfort::tinyint(SCHAR_MIN); }
	static CXXO_CONSTEXPR cxxomfort::tinyint max () CXXO_NOEXCEPTNOTHROW { return cxxomfort::tinyint(SCHAR_MAX); }
#if (CXXOMFORT_CXX_STD>=2011)
	static CXXO_CONSTEXPR cxxomfort::tinyint lowest () CXXO_NOEXCEPTNOTHROW { return min(); }
#endif
};

template<> struct numeric_limits<cxxomfort::tinyuint>
: public std::numeric_limits<unsigned char> {
	static CXXO_CONSTEXPR cxxomfort::tinyuint min () CXXO_NOEXCEPTNOTHROW { return cxxomfort::tinyuint(0); }
	static CXXO_CONSTEXPR cxxomfort::tinyuint max () CXXO_NOEXCEPTNOTHROW { return cxxomfort::tinyuint(UCHAR_MAX); }
#if (CXXOMFORT_CXX_STD>=2011)
	static CXXO_CONSTEXPR cxxomfort::tinyuint lowest () CXXO_NOEXCEPTNOTHROW { return min(); }
#endif
};

}
// numeric_limits specializations
#endif
// tinyuint

namespace cxxomfort {
namespace impl {

template <uintmax_t N, size_t base=2> struct log_ {
    static const size_t value = 1 + log_<N/base, base>::value;
};

template <size_t base>
struct log_<1, base> {
    enum { value = 0 };
};

template <size_t base>
struct log_<0, base> {
    enum { value = 0 };
};


}

namespace fix { 

/**
 * @brief An extension to <tt>numeric_limits</tt> that deals specifically with integer-like types.
 * @xrefitem cxxomfort-fixes "" "" Compile-time constants for integral limits
 * 
 * For a given integral or integer-like type @a Ty , <tt>integral_limits<Ty></tt> 
 * provides an interface to <tt>std::numeric_limits<Ty></tt> and in fact 
 * inherits from the latter. The interface is expanded to add some useful information 
 * about the integral type that is only available at compile time as macros 
 * pre-C++11.
 * 
 * 
 * <tt>integral_limits</tt> provides the following new members: 
 * 
 * * @c const_min - minimum value as a value (<tt>::min</tt> is constexpr function in C++>=11).
 * * @c const_max - maximum value as a value (<tt>::max</tt> is constexpr function in C++>=11).
 * * @c digits_base<N> - the total number of digits required to represent any value in the given base.
 * 
 * @e Pending members : 
 * 
 * * @c narrower_type - an alias to the next smaller integral type of the same signedness.
 * * @c wider_type - an alias to the next larger integral type of the same signedness.
 * * @c lo_mask - an integral value that can be AND-ed with this type to keep only the lower bits.
 * * @c hi_mask - an integral value that can be AND-ed with this type to keep only the higher bits.
 * 
 * @subsection Specializations 
 * 
 * cxxomfort provides specializations of the new <tt>integral_limits</tt> for 
 * all the types that are defined as integral. The list explicitly includes: 
 * 
 * <tt>short,  int,  long,  long long</tt> in their signed and unsigned variants.
 * 
 * */
template <typename T>
struct integral_limits {
    static_assert (std::numeric_limits<T>::is_integer, "integral_limits is only defined for compiler integral types");
};

#define SPECIA_IMPL(Ty_,Min_,Max_,Low_) \
template <> struct integral_limits<Ty_>           \
: public std::numeric_limits<Ty_> {               \
    typedef std::numeric_limits<Ty_> Base_;       \
    static const Ty_ const_min= Min_;             \
    static const Ty_ const_lowest = Low_;         \
    static const Ty_ const_max= Max_;             \
    template <unsigned short B> struct digits_base {   \
        static const unsigned short value=             \
        /* (digits - is_signed) / impl::log_<B>::value; */  \
        impl::log_<const_max,B>::value;   \
    };                                            \
    typedef void narrower_type; \
    typedef void wider_type; \
}

#define SPECIA_IMPL_NONINT(Ty_,Min_,Max_,Low_) \
template <> struct integral_limits<Ty_>           \
: public std::numeric_limits<Ty_> {               \
    typedef std::numeric_limits<Ty_> Base_;       \
    static const Ty_ const_min;             \
    static const Ty_ const_lowest;         \
    static const Ty_ const_max;             \
    typedef void narrower_type; \
    typedef void wider_type; \
}; \
Ty_ const integral_limits<Ty_>::const_min = Ty_(Min_); \
Ty_ const integral_limits<Ty_>::const_max = Ty_(Max_); \


SPECIA_IMPL(bool,false,true,false);
SPECIA_IMPL(signed int,INT_MIN,INT_MAX,INT_MIN);
SPECIA_IMPL(unsigned int,0,UINT_MAX,0);
SPECIA_IMPL(char,CHAR_MIN,CHAR_MAX,CHAR_MIN);
SPECIA_IMPL(signed char,SCHAR_MIN,SCHAR_MAX,SCHAR_MIN);
SPECIA_IMPL(unsigned char,0,UCHAR_MAX,0);
SPECIA_IMPL(signed short,SHRT_MIN,SHRT_MAX,SHRT_MIN);
SPECIA_IMPL(unsigned short,0,USHRT_MAX,0);
SPECIA_IMPL(signed long,LONG_MIN,LONG_MAX,LONG_MIN);
SPECIA_IMPL(unsigned long,0,ULONG_MAX,0);
SPECIA_IMPL(signed long long,LLONG_MIN,LLONG_MAX,LLONG_MIN);
SPECIA_IMPL(unsigned long long,0,ULLONG_MAX,0);

// tinyint specializations


//SPECIA_IMPL_NONINT(cxxomfort::tinyint,SCHAR_MIN,SCHAR_MAX,SCHAR_MIN);
//SPECIA_IMPL_NONINT(cxxomfort::tinyuint,0,UCHAR_MAX,0);

#undef SPECIA_IMPL
#undef SPECIA_IMPL_NONINT


} // fix

} // cxxomfort



#endif

/**
 * @page std0limits <limits>
 * @brief Backports related to Standard header <tt><limits></tt> 
 * 
 * The @c limits.hpp header by itself does not add any capabilities from later versoins of the C++ Standard; instead, it provides an alternative view of the implementation of @c numeric_limits for integral types that is C++03-capable.
 * 
 * @code
 * #include <cxxomfort/limits.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Interfaces defined here:
 *
 * (none from Standard C++)
 * 
 * Non-backport interfaces (<tt>cxxomfort::fix</tt>):
 * 
 * * @link cxxomfort::fix::integral_limits <tt>integral_limits</tt> @endlink - Allows access to the constant values of integral_limits in C++ < C++11.
 *
 * @section also See Also
 * 
 * @see @cppref{header/limits}
 * @see @cppref{types/numeric_limits} (for @c integral_limits )
 *
 * */
