#ifndef CXXOMFORT_MEMORY_HPP
#define CXXOMFORT_MEMORY_HPP
/**
 * @file
 * @brief Implementations related to Standard header <tt>memory</tt>.
 * @xrefitem std0memory "<memory>" "Header"
 * @ingroup std0memory
 *
 */

#include <cxxomfort/base.hpp>
#include <memory>
#include <cstdlib>

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
	#pragma message CXXO_NOTICE("enabled <memory> support.")
#endif

// 
// header (TR1, etc)
// 

#if (CXXOMFORT_CXX_STD >= 2011)
#elif (CXXOMFORT_CXX_STD >=1997 && CXXOFLAG_namespace_tr1>0)
	#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_CLANG)
		#include <tr1/memory>
		#include <tr1/functional> // for hash
namespace std {
	using tr1::shared_ptr;
	using tr1::weak_ptr;

	#define CXXOMFORT_USING_make_shared
	
}

	#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC, < 1600) )
		#define CXXOMFORT_USING_make_shared

namespace std {
	using tr1::shared_ptr;
	using tr1::weak_ptr;
}

	#endif
#else
	// not enough C++98 support

#endif

// 
// internal: pragma packed int
// for use by eg.: aligned_storage.
// only defined if CHAR_BIT==8
// 
#if (CHAR_BIT==8)

	#include <cxxomfort/impl/11-conditional.hpp>

namespace cxxomfort { namespace impl {

template <size_t S> struct pack_t;


	#pragma pack(push)
template<> struct pack_t<1> { char u; };
	#pragma pack(2)
template<> struct pack_t<2> { uint_least16_t u; };
	#pragma pack(4)
template<> struct pack_t<4> { 
	union u {uint_least16_t du1[2]; uint_least32_t du2; }; 
};
	#pragma pack(8)
template<> struct pack_t<8> { 
	union u {uint_least32_t du1[2]; uint_least64_t du2; }; 
};
	#pragma pack(16)
template<> struct pack_t<16> { 
	union u {uint_least64_t du1[2]; long double du2; }; 
};
	#pragma pack(pop)

#endif

} } // cxxomfort::impl

// 
// make_shared 
// 

#if defined(DOXYGEN_DOC)
namespace std {
	/**
	 * @cxx11b Creates a @c shared_ptr .
	 * @xrefitem std0memory "" "" Creates a @c shared_ptr .
	 * @see @cppref{memory/shared_ptr/make_shared}
	 * @note Pre-C++11 this is provided with support for up to 5 arguments.
	 * 
	 * */
	template <typename T>
	shared_ptr<T> make_shared () {
		return shared_ptr<T>( new T() );
	}

}
#endif
#ifdef CXXOMFORT_USING_make_shared

namespace std {

	template <typename T>
	shared_ptr<T> make_shared () {
		return shared_ptr<T>( new T() );
	}

	template <typename T, typename A1>
	shared_ptr<T> make_shared (A1 const& a1) {
		return shared_ptr<T>( new T(a1) );
	}

	template <typename T, typename A1, typename A2>
	shared_ptr<T> make_shared (A1 const& a1, A2 const& a2) {
		return shared_ptr<T>( new T(a1,a2) );
	}

	template <typename T, typename A1, typename A2, typename A3>
	shared_ptr<T> make_shared (A1 const& a1, A2 const& a2, A3 const& a3) {
		return shared_ptr<T>( new T(a1,a2,a3) );
	}

	template <typename T, typename A1, typename A2, typename A3, typename A4>
	shared_ptr<T> make_shared (A1 const& a1, A2 const& a2, A3 const& a3, A4 const& a4) {
		return shared_ptr<T>( new T(a1,a2,a3,a4) );
	}

	template <typename T, typename A1, typename A2, typename A3, typename A4, typename A5>
	shared_ptr<T> make_shared (A1 const& a1, A2 const& a2, A3 const& a3, A4 const& a4, A5 const& a5) {
		return shared_ptr<T>( new T(a1,a2,a3,a4,a5) );
	}
}

#endif

//
// addressof
//


namespace cxxomfort { namespace cxxostd {
/**
 * @brief Returns the address of an object.
 * @ingroup std0memory
 * @cxx11b Take the address of an object, even in presence of overloaded @c & unary.
 * @cppref{memory/addressof}
 */
template <typename T> inline
T* addressof (T& arg) {
    return reinterpret_cast<T*>(
    &const_cast<char&>( reinterpret_cast<const volatile char&>( arg) ) 
	);
}

} }

#if defined(DOXYGEN_DOC)
namespace std {
/**
 * @brief Returns the address of an object.
 * @xrefitem std0memory "" "" Take the address of an object, even in presence of overloaded @c & unary.
 * @cxx11b Take the address of an object, even in presence of overloaded @c & unary.
 * @cppref{memory/addressof}
 */
template <typename T> inline
T* addressof (T& arg) {}
} // std

#elif (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#elif (CXXOMFORT_CXX_STD >= 2011 || CXXOMFORT_CXX_EMULATION >= 2011)
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_GCC,>=406) && CXXOMFORT_CXX_EMULATION==2011)
#else
namespace std {
    using cxxomfort::cxxostd::addressof;
}
#endif

//
// aligned_storage
//


#undef  CXXOMFORT_IMPLEMENTS_aligned_storage
#include "impl/11-conditional.hpp"

#if (CXXOMFORT_CXX_STD>=2011)
	#define CXXOMFORT_IMPLEMENTS_aligned_storage CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD>=1997)
	#include "impl/11-aligned_storage.hpp"
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_DMC,>=852))
	#include "impl/11-aligned_storage.hpp"
#else
	#define CXXOMFORT_IMPLEMENTS_aligned_storage 0
#endif


//
// aligned_union
//

//
// pointer_traits , allocator_traits
//

#include "utils/memtraits.hpp"

#if (CXXOMFORT_CXX_STD < 2011)
    #define CXXOMFORT_USING_MEMTOOLS
    #if defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1)
		#pragma message CXXO_NOTICE("memtools support.")
    #endif
#endif // c++11


//
// default_delete
//

#ifdef DOXYGEN_DOC
namespace std {
	//! @c default_delete (stateless deleter for default @c new allocations )
	//! @cxx1b Default deleter for utilities like @c unique_ptr .
	template <typename T> struct default_delete {};
}
#elif (CXXOMFORT_CXX_STD>=2011)
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600)) // MSVC 2010 already has it
#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC && CXXOMFORT_CXX_EMULATION>=2011)
#else

namespace cxxomfort { namespace cxxostd {
	template <typename T> struct default_delete {
		typedef T* pointer;

		default_delete () throw() {}

	#if (CXXO_COMPILER_SUPPORT_std_is_constructible_traits)
		template <class U>
		default_delete ( default_delete<U> const&
		, typename std::enable_if<std::is_convertible<U*, T*>::value, void>::type* = 0 )
		{ }
	#endif

		void operator() (T* ptr) const  CXXO_NOEXCEPTNOTHROW   {
			static_assert (sizeof(T) > 0, "T must be a complete type");
			delete ptr;
		}

	};

	template <typename T> struct default_delete<T[]> {

		void operator() (T* ptr) const CXXO_NOEXCEPTNOTHROW {
			static_assert (sizeof(T) > 0, "T must be a complete type");
			delete [] ptr;
		}

		private:

		template <class U> void operator()(U*) const;
	};
} } 

namespace std {
	using cxxomfort::cxxostd::default_delete;
}

#endif


//
// unique_ptr
//

#if (CXXOMFORT_CXX_STD>=2011 || CXXO_COMPILER_SUPPORT_unique_ptr)
	#undef  CXXOMFORT_IMPLEMENTS_unique_ptr
	#define CXXOMFORT_IMPLEMENTS_unique_ptr CXXO_IMPLSTATUS_NATIVE()
#elif (CXXOMFORT_CXX_STD>=1997)
	#include "impl/unique_ptr.hpp"
	#undef  CXXOMFORT_IMPLEMENTS_unique_ptr
	#define CXXOMFORT_IMPLEMENTS_unique_ptr CXXO_IMPLSTATUS_EMULATION()
#else
	//#include "impl/unique_ptr-poor.hpp"
	#undef  CXXOMFORT_IMPLEMENTS_unique_ptr
	#define CXXOMFORT_IMPLEMENTS_unique_ptr 0

#endif

// 
// std::hash<std::shared_ptr>
// 

#define CXXO_HASH_MEMBER() \
result_type operator() (argument_type const& ptr) const CXXO_NOEXCEPTNOTHROW { \
    return base_hash()(ptr.get()); \
} \

#if (CXXOMFORT_CXX_STD >= 2011)
#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC)
    #if (CXXOMFORT_COMPILER_VERSION >= 1700)

    #elif (0 && CXXOMFORT_COMPILER_VERSION >= 1500)
namespace std { 
    template <typename T> struct hash< tr1::shared_ptr<T> >    
    : private hash<T*> {
        private:
        typedef hash<T*> base_hash;
        public:
        typedef tr1::shared_ptr<T> argument_type;
        typedef size_t result_type;
        CXXO_HASH_MEMBER();
    };
} // std
    
    #endif
#elif defined(__GNUC__) || defined(__clang__)
    #if (CXXOMFORT_CXX_EMULATION==2011)
        // nothing to do, they already bring the hash spec
    #else
        // live in TR1, but we specialized in std
namespace std { namespace tr1 {
    template <typename T> 
    struct hash< std::tr1::shared_ptr<T> >    
    : private hash<T*> {
        private:
        typedef hash<T*> base_hash;
        public:
        typedef tr1::shared_ptr<T> argument_type;
        typedef size_t result_type;
        CXXO_HASH_MEMBER();
    };
} }
	#endif
    
#elif (CXXOMFORT_CXX_STD >= 1997) // other compilers

#endif


// 
// uninitialized_copy_n (C++11)
// 

#if defined(DOXYGEN_DOC)
namespace std {
/**
 * @xrefitem std0memory "" "" Copies elements to an uninitialized memory area.
 * @cxx11b Copies elements to an uninitialized memory area.
 * @see @cppref{memory/uninitialized_copy_n}
 * 
 * */
template<class InpIt, class Size, class FwdIt>
FwdIt uninitialized_copy_n (InpIt ini, Size n, FwdIt dest) {}
}
#elif (CXXOMFORT_CXX_STD >= 2011)
#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600)) // MSVC 2010
#else
namespace std {
template<class InpIt, class Size, class FwdIt>
FwdIt uninitialized_copy_n (InpIt ini, Size n, FwdIt dest) {
    typedef typename std::iterator_traits<FwdIt>::value_type Value;
    for (; n > 0; ++ini, ++dest, --n) {
        ::new (static_cast<void*>(&*dest)) Value(*ini);
    }
    return dest;
}

} // std
#endif


//
// make_unique (C++14)
//

#include "impl/14-make_unique.hpp"

#if 0
//
// alignas
//

/*
    #ifdef alignas
    #error " cxxomfort -- conflicting #define directive for 'alignas()'!"
    #endif
    #if 0
    #elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_GCC)
        #define alignas(N) __attribute__ ((aligned(N)))
    #elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC)
        #define alignas(N) __declspec(align(N))
    #endif
*/
#endif


//
// destroy, uninitialized_default_construct, ... (C++17)
//
//! @include{doc} "impl/p0040-memory_management.hpp"
#include "impl/p0040-memory_management.hpp"

// 
// to_address (C++20)
// 

#if (!defined(__cpp_lib_to_address))
namespace std {

template <class T> CXXO_CONSTEXPR14 
T* 	to_address (T* ptr) CXXO_NOEXCEPTNOTHROW {
    static_assert(!std::is_function<T>::value);
    return ptr;
}
 
template <class T> CXXO_CONSTEXPR14 
typename T::value_type const* 	to_address (T const& ptr) CXXO_NOEXCEPT {
	return to_address(ptr.operator->());
}

}
	#define __cpp_lib_to_address (CXXOMFORT_CXX_STD*100)
#endif

// 
// construct_at (C++20)
// 

#if defined(DOXYGEN_DOC)
#elif (CXXOMFORT_CXX_STD >= 2020)
#elif (CXXO_COMPILER_SUPPORT_variadic && CXXO_COMPILER_SUPPORT_rvref)
namespace std {

template <typename T, typename... Args>
T* 	construct_at (T* ptr, Args&&... args) {
	return ::new (const_cast<void*>(static_cast<const volatile void*>(ptr)))
    T(std::forward<Args>(args)...);
}
}

#else
namespace std {
template <typename T, typename A1>
T* 	construct_at (T* ptr, A1 a1) {
	return ::new (const_cast<void*>(static_cast<const volatile void*>(ptr))) T(a1);
}

} // std

#endif

//
// cxxomfort::fix assist: to_bool
//

#if defined(DOXYGEN_DOC)
namespace cxxomfort { namespace fix {
	//!Returns @c true if the @c shared_ptr currently hosts a value.
	//! @ingroup cxxomfort-fix
	template <typename T> inline 
	bool to_bool (std::shared_ptr<T> const& p) noexcept;
} }
#elif (CXXOMFORT_CXX_STD>=2011 || CXXOMFORT_CXX_EMULATION>=2011) 
	#if (CXXOFLAG_tr1)

// assist
namespace cxxomfort { namespace fix {

template <typename T>
bool to_bool (std::shared_ptr<T> const& p) CXXO_NOEXCEPTNOTHROW {
	return p.get() != nullptr;
}

} } // cxxomfort::fix

	#endif
#endif

#if defined(DOXYGEN_DOC)
namespace cxxomfort { namespace fix {
	//!Returns @c true if the @c unique_ptr currently hosts a value.
	//! @ingroup cxxomfort-fix
	template <typename T> inline 
	bool to_bool (std::unique_ptr<T> const& p) noexcept;
} }
#elif defined(CXXOMFORT_USING_unique_ptr)

// assist
namespace cxxomfort { namespace fix {

template <typename T>
bool to_bool (std::unique_ptr<T> const& p) CXXO_NOEXCEPTNOTHROW {
    return p.get() != nullptr;
}

} } // cxxomfort::fix

#endif




#if defined(CXXOMFORT_USING_MEMTOOLS)

// We define alignas, alignof depending on compiler specifics
#if 0
#elif (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_GCC)
#elif (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_CLANG)
#elif (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_MSC)
#else // unrecognized compiler, we can't go any further
	#if (defined(CXXOMFORT_NOTICES) && CXXOMFORT_NOTICES > 1)
    #pragma message CXXO_NOTICE("alignof directives not supported; compilation might fail if using <memory> utilities!")
	#endif
#endif

#else // using c++11 or emulation mode, we might still need to fix some things
// NOTE: in GCC < 4.8, alignas is not supported - not even in in c++11 mode
// (but alignof is)
	#if ((CXXOMFORT_CXX_STD >= 2011) && CXXOMFORT_COMPILER_ID==CXXOMFORT_VALUE_GCC && CXXOMFORT_COMPILER_VERSION < 408)
		#define alignas(N) __attribute__ ((aligned(N))
		#define CXXOMFORT_ALIGNMENT_TOOLS_SUPPORTED
	#endif // GCC c++11 fix

#endif // check for USING_memtools (1)


//
// assist: malloc_deleter extension
//

//#if (CXXOMFORT_CXX_STD<1997)
//	#include <malloc>
//#endif

namespace cxxomfort { namespace fix {

/**
 * @brief Deleter operator that uses <tt>free()</tt> for <tt>malloc</tt>-reserved blocks.
 * @ingroup cxxomfort-fix
 * @xrefitem std0memory "<memory>" "Memory Management"
 * */
struct malloc_deleter {
	//! Invokes <tt>free()</tt> on the pointer.
	void operator() (void* ptr) { free(ptr); }
};

} } // cxxomfort::fix::


// -- set up needed things in std::
#if (CXXOMFORT_CXX_STD < 2011 && defined(CXXOMFORT_USING_unique_ptr))
    #if (CXXOFLAG_tr1)
//
// specialize std::tr1::hash for unique_ptr<>
// (this is allowed)
// define either NO_TR1 or CXXOMFORT_NO_TR1 to disable
// using this TR1 extension in C++03
//
		#include <functional> // std::hash, from Tr1
//#include <cxxomfort/functional.hpp>

CXXO_STDTR1_OPEN() 
    template <typename T, typename D>
    struct hash< ::std::unique_ptr<T,D> >
    : private hash< typename unique_ptr<T,D>::pointer > {
        private:
        typedef hash< typename unique_ptr<T,D>::pointer > base_type;
        public:
        typedef unique_ptr<T,D> argument_type;
        typedef size_t result_type;
        size_t operator() (argument_type const& t) const {
            //return static_cast<base_type>(*this).operator()(t.get());
            return static_cast<base_type const*>(this)->operator()(t.get());
        }
    };

CXXO_STDTR1_CLOSE() 
    #endif
#endif




#endif // file



/**
 * @page std0memory <memory>
 * @brief Backports related to Standard Header <tt><memory></tt>
 *
 * @code
 * #include <cxxomfort/memory.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * @section Interfaces
 * 
 * Interfaces defined here:
 *
 * * @c alignof() (C++11) (See @e base_features ).
 * * @c addressof() (C++11) - get the memory address of an object.
 * * Pointer traits and Allocator traits (C++11) - tools for assisting in the design of containers.
 * * @link cxxomfort::cxxostd::aligned_storage @c aligned_storage @endlink </code> (C++11) - create an unitialized buffer for storing an object of a given type.
 * * @c std::default_delete (C++11).
 * * @link cxxomfort::cxxostd::unique_ptr @c unique_ptr emulation @endlink (C++11) - smart pointer that owns and holds a resource.
 * * @c @ref make_unique() (C++14) - assist for creating @c unique_ptr .
 * * @c destroy() , @c destroy_at() (C++17) - memory management by calling destructors.
 * * @c make_shared() - construct shared smart pointers safely (for some C++03 compilers).
 *
 * Interfaces repointed here when found in TR1:
 * 
 * * @c std::shared_ptr  and related utilities.
 * * @c std::weak_ptr  and related utilities.
 * * @c std::default_delete
 * * @c std::make_shared .
 * 
 * Non-backport interfaces (<tt>cxxomfort::fix</tt>):
 *
 * * <tt>to_bool()</tt> overloads to quickly check validity of a smart pointer: see @ref smartptr_accessors .
 * * <tt>#cxxomfort::fix::malloc_deleter</tt> - function object deleter that wrapps @c free() .
 * 
 * @section smartptr_accessors Smart Pointer accessors
 * 
 * cxxomfort provides, for the available smart pointer types, the special function @c cxxomfort::fix::to_bool() which takes as argument a smart pointer type and returns if it currently manages a value, as if checking their (C++>=11) <tt>operator bool</tt> property.
 * 
 * @code
 * shared_ptr<int> P1 = new int(10);
 * unique_ptr<int> P2 = new int(20);
 * 
 * using cxxomfort::fix::to_bool; 
 * 
 * if (to_bool(P1)) {
 * 	...
 * }
 * if (to_bool(P2)) {
 * 	...
 * }
 * 
 * @endcode
 * 
 * Functionally, invoking these accessors is a simpler means to checking the managed status than using eg.: <tt>explicit_cast<bool>(...)</tt>, and it has the advantage of working in C++03 and C++11/14 pre the new "@c if condition declaration" rule as well:
 * 
 * @code
 * // does not work pre-C++17:
 * 
 * shared_ptr<Foo> managedfoo = make_shared<Foo>(args...);
 * // this fails: no implicit conversion of managed pointer to bool
 * if (bool f = (managedfoo)) {
 * 		...code
 * }
 * 
 * // works in all versions of C++, no extra cost:
 * 
 * shared_ptr<Foo> managedfoo = make_shared<Foo>(args...);
 * if (bool f = to_bool(managedfoo)) {
 * 		...code
 * }
 * 
 * // note that this is basically equivalent to the following 
 * // in all versions (maps to static_cast in > C++11):
 * 
 * shared_ptr<Foo> managedfoo = make_shared<Foo>(args...);
 * if (bool f = explicit_cast<bool>(managedfoo)) {
 * 		...code
 * }
 * @endcode
 * 
 * 
 * @section Notes
 * 
 * @note <b>Cxxomfort</b> expects that the compiler provides an alignment information macro such as <tt>__alignof(T)</tt>.
 * 
 * @note <b>Cxxomfort</b> expects an implementation of <tt>std::shared_ptr</tt> 
 * to exist and work in order, as part of the general TR1 requirement.
 * 
 * @warning Before C++11, <tt>aligned_storage<S,A></tt> may not deliver a type with correct alignment when the template argument @a A is ommitted (in which case, a likely result is a type aligned to <tt>sizeof(int)</tt>). <br/>For production purposes, @a A @b always should be given explicitly, eg.: with the usual pattern of<br/><tt>aligned_storage<sizeof(T),alignof(T)></tt>.
 * 
 * @section also See Also
 *
 * @see @cppref{header/memory}
 * @see @cppref{memory/unique_ptr}
 * 
 *
 * */
