#ifndef CXXOMFORT_NUMERIC_HPP
#define CXXOMFORT_NUMERIC_HPP
/**
 * @file
 * @brief Implementations and backports tied to <tt>\<numeric\></tt>.
 * @ingroup std0numeric
 */

#include <cxxomfort/base.hpp>
#include <numeric>

#include <cxxomfort/impl/07-type_traits.hpp> // is_signed
#include <cxxomfort/impl/11-conditional.hpp> // enable_if
#if (CXXOMFORT_CXX_STD>=2011)
    #include <type_traits> // enable_if
#endif

// 
// iota, iota_n (for completeness)
// 

#if (defined(DOXYGEN_DOC))
namespace std {
#else
namespace cxxomfort { namespace cxxostd {
#endif

/**
 * @brief Fills a sequence with a sequentially growing set of values.
 * @cxx1b "Numeric" iota
 * 
 * Given the sequence <tt>[ini, fin)</tt>, @e iota assigns the values <tt>i0, i0+1, i0+2, ...</tt> to the elements of the sequence until it has been filled.
 * 
 * @pre The range <tt>[ini, fin)</tt> is valid and writable.
 * 
 * @sa cppref{algoritm/iota}
 *  
 */
template <typename FwIt, typename T>
FwIt iota (FwIt ini, FwIt fin, T i0) {
    for(; ini!=fin; ++ini, ++i0) {*ini= i0; }
    return ini;
}

/**
 * @brief Fills a sequence with a sequentially growing set of values.
 * 
 * Given the sequence starting at @a ini , @c iota assigns the values <tt>i0, i0+1, i0+2, ...</tt> to the @a n elements of the sequence until the @a n -th element has been reached.
 * 
 * @pre The range <tt>[ini, ini+n)</tt> is valid and writable.
 * 
 * @sa cppref{algoritm/iota}
 *  
 */
template <typename FwIt, typename Integer, typename T>
FwIt iota_n (FwIt ini, Integer n, T i0) {
    while (n-->0) { *ini= i0; ++ini; ++i0; }
    return ini;
}

#if defined(DOXYGEN_DOC)
}
#else
} }
#endif

// 
// gcd, lcm (C++17)
// 

#if (defined(DOXYGEN_DOC))
namespace std {

/**
 * @brief Calculates the greatest common divisor.
 * @cxx17b gcd (greatest common denominator)
 * @xrefitem std0numeric "" "" gcd (greatest common denominator)
 * @returns A numeric value of type @a Num1 .
 * @sa @c lcm()
 * @sa @cppref{numeric/gcd}
 * @note As a backport, the return type is the same as the first argument.
 */
template <typename Num1, typename Num2>
constexpr Num1 gcd (Num1 n1, Num2 n2) noexcept;

/**
 * @brief Calculates the lowest common multiple.
 * @cxx17b lcm (lowest common nultiple)
 * @xrefitem std0numeric "" "" lcm (lowest common nultiple)
 * @returns A numeric value of type @a Num1 .
 * @sa @c gcd()
 * @sa @cppref{numeric/lcm}
 * @note As a backport, the return type is the same as the first argument.
 */
template <typename Num1, typename Num2>
constexpr Num1 lcm (Num1 n1, Num2 n2) noexcept;

} }
#elif (defined(__cpp_lib_gcd_lcm))
#else
    #define __cpp_lib_gcd_lcm (CXXOMFORT_CXX_STD*100)

namespace cxxomfort { namespace cxxostd {

template <typename Num1, typename Num2>
CXXO_CONSTEXPR Num1 gcd (Num1 n1, Num2 n2) CXXO_NOEXCEPT {
    return (n2== 0) ? n1 : gcd(n2, n1 % n2);
}

template <typename Num1, typename Num2>
CXXO_CONSTEXPR Num1 lcm (Num1 n1, Num2 n2) CXXO_NOEXCEPT {
    return n1 / gcd(n1,n2) * n2;
}

} }

#endif

// 
// midpoint, from C++20
// 

// we can't set __cpp_lib_interpolate because up to this release, cxxomfort doesn't implement std::lerp()
// instead cxxomfort sets CXXOMFORT_IMPLEMENTS_midpoint 

#if defined(DOXYGEN_DOC)

namespace std {

/**
 * @ingroup cxx20-backports 
 * @cxx17b Returns the midpoint or average, avoiding overflows.
 * @xrefitem std0numeric "" "" Returns the midpoint or average, avoiding overflows.
 * @return <tt>(n1 + n2)/2</tt>, avoiding overflows.
 * @see @wg21{P0811R3}
 * @see @cppref{numeric/midpoint}
 * 
 * */
template <typename Numeric>
inline CXXO_CONSTEXPR 
Numeric midpoint (Numeric n1, Numeric n2) CXXO_NOEXCEPTNOTHROW { }

/**
 * @ingroup cxx20-backports 
 * @tparam T A pointer type (not a @e member pointer type).
 * @cxx17b Returns the midpoint position between the pointers.
 * @return A pointer @a p half the distance away from @a p1 in the direction towards @a p2 .
 * @see @wg21{P0811R3}
 * 
 * @pre @c Pointer is a pointer type.
 * 
 * */
template <typename Pointer>
Pointer midpoint (Pointer p1, Pointer p2) CXXO_NOEXCEPTNOTHROW { }

}

#elif defined(__cpp_lib_interpolate)
    // nothing to do
    #define CXXOMFORT_IMPLEMENTS_midpoint CXXO_IMPLSTATUS_NATIVE()

#elif (CXXOMFORT_CXX_STD >= 1997)
    #define CXXOMFORT_IMPLEMENTS_midpoint CXXO_IMPLSTATUS_BACKPORT()
    #include <cxxomfort/utils/traits_basic.hpp>
    #include <cxxomfort/impl/11-type_traits.hpp>
    #include <cxxomfort/impl/to_signed.hpp>

namespace cxxomfort { namespace cxxostd {

namespace detail {

// from cppcon2019 "std::midpoint - how hard it can be?"
template <typename Numeric>
inline CXXO_CONSTEXPR /* can't be constexpr */ 
Numeric midpoint (Numeric n1, Numeric n2, std::true_type const /*_signed*/ ) CXXO_NOEXCEPTNOTHROW {
    typedef typename ::cxxomfort::traits::make_unsigned<Numeric>::type U;
    return n1 > n2 ? 
    n1 - Numeric(U(n1-n2) >> 1)  :  n1 + Numeric(U(n2-n1) >> 1)
    ;

}

// for unsigneds this is easier - using half arith we can't overflow
template <typename Numeric>
inline CXXO_CONSTEXPR 
Numeric midpoint (Numeric n1, Numeric n2, std::false_type const /*_signed*/) CXXO_NOEXCEPTNOTHROW {
    return n1 > n2 ? 
    n1 - (Numeric(n1-n2) >> 1)  :  n1 + (Numeric(n2-n1) >> 1)
    ;
}

} // detail

template <typename Numeric>
inline CXXO_CONSTEXPR 
typename std::enable_if< std::is_arithmetic<Numeric>::value, Numeric>::type 
midpoint (Numeric n1, Numeric n2) CXXO_NOEXCEPTNOTHROW {
    return detail::midpoint(n1,n2, std::is_signed<Numeric>() );
    //return n1/Numeric(2) + n2/Numeric(2);
}

template <typename T>
inline typename std::enable_if<
    std::is_pointer<T>::value==true && std::is_member_pointer<T>::value == false
    , T>::type
midpoint (T p1, T p2) CXXO_NOEXCEPTNOTHROW {
    return p1+midpoint(ptrdiff_t(0) , (p2-p1));
}

} }

namespace std {
    using ::cxxomfort::cxxostd::midpoint;
}

#else
    #define CXXOMFORT_IMPLEMENTS_midpoint CXXO_IMPLSTATUS_EMULATION()
    #include <cxxomfort/utils/traits_basic.hpp>
    
namespace cxxomfort { namespace cxxostd {

template <typename Numeric>
inline Numeric
midpoint (Numeric n1, Numeric n2) CXXO_NOEXCEPTNOTHROW {
    fprintf(stdout, "invoked midpoint(%ld,%ld)", static_cast<long>(n1), static_cast<long>(n2));
    return n1 > n2 ? 
    n1 - Numeric(Numeric(n1-n2) >> 1)
    :
    n1 + Numeric(Numeric(n2-n1) >> 1)
    ;
}

template <typename T>
inline T*
midpoint (T* p1, T* p2) CXXO_NOEXCEPTNOTHROW {
    ptrdiff_t pd = p2 - p1;
    return (p1+midpoint(0 , pd));
}

} }

namespace std {
    using cxxomfort::cxxostd::midpoint;
}

#endif


// 
// p0616 - move-ing versions of numeric algorithms
// 

namespace cxxomfort {  namespace cxxostd {
namespace numeric20 {

struct numeric_move_tag {};

template <typename II, typename Ty>
inline CXXO_CONSTEXPR14 
Ty accumulate (II ini, II fin, Ty i0){
    for (; ini != fin; ++ini) {
        i0 = std::move(i0) + *ini;
    }
    return i0;
}

template <typename II, typename Ty, typename Binary>
inline CXXO_CONSTEXPR14 
Ty accumulate (II ini, II fin, Ty i0, Binary op) {
    for (; ini != fin; ++ini) {
        i0 = op(std::move(i0), *ini);
    }
    return i0;
}

template <typename II1, typename II2, typename T>
inline CXXO_CONSTEXPR14 
T inner_product (II1 ini1, II1 fin1, II2 ini2, T i0) {
    for (; ini1 != fin1; ++ini1, ++ini2) {
        i0= std::move(i0) + *ini1 * *ini2;
    }
    return i0;
}

template <typename II1, typename II2, typename T, typename Binary1, typename Binary2>
inline CXXO_CONSTEXPR14 
T inner_product (II1 ini1, II1 fin1, II2 ini2, T i0, Binary1 sum, Binary2 product) {
    for (; ini1 != fin1; ++ini1, ++ini2) {
        i0= sum(std::move(i0) , product(*ini1 , *ini2) );
    }
    return i0;
}

} // namespace numeric20
} } // cxxomfort::cxxostd


// 
// accessible numeric move synonyms
// 

#if 0

namespace cxxomfort { namespace fix {

using cxxomfort::cxxostd::numeric20::numeric_move_tag;

template <typename II, typename Ty>
inline CXXO_CONSTEXPR14 
Ty accumulate (II ini, II fin, Ty i0, numeric_move_tag) {
#if (CXXOMFORT_CXX_STD >= 2020)
    return std::accumulate(ini,fin,i0);
#else
    return cxxomfort::cxxostd::numeric20::accumulate(ini,fin,i0);
#endif
}

template <typename II, typename Ty, typename Binary>
inline CXXO_CONSTEXPR14 
Ty accumulate (II ini, II fin, Ty i0, Binary f, numeric_move_tag) {
#if (CXXOMFORT_CXX_STD >= 2020)
    return std::accumulate(ini,fin,i0,f);
#else
    return cxxomfort::cxxostd::numeric20::accumulate(ini,fin,i0,f);
#endif
}

template <typename IA, typename IB, typename Ty>
inline CXXO_CONSTEXPR14
Ty inner_product (
IA ini1, IA fin1, IB ini2, Ty t0, numeric_move_tag) {
#if (CXXOMFORT_CXX_STD >= 2020)
    return std::inner_product(ini1,fin1,ini2,t0);
#else
    return cxxomfort::cxxostd::numeric20::inner_product(ini1,fin1,ini2,t0);
#endif
}

template <typename IA, typename IB, typename Ty, typename Sum, typename Product>
inline CXXO_CONSTEXPR14
Ty inner_product (
IA ini1, IA fin1, IB ini2, Ty t0, Sum fs, Product fp, numeric_move_tag	) {
#if (CXXOMFORT_CXX_STD >= 2020)
    return std::inner_product(ini1,fin1,ini2,t0,fs,fp);
#else
    return cxxomfort::cxxostd::numeric20::inner_product(ini1,fin1,ini2,t0,fs,fp);
#endif
}

} }

#endif

// 
// 
// 

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
    #pragma message CXXO_NOTICE("enabled <numeric> support.")
#endif

#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
#else
    #if (CXXOMFORT_CXX_STD < 2011) && (CXXO_COMPILER_SUPPORT_std_cxx11_algorithms == 0)
    namespace std {
        using ::cxxomfort::cxxostd::iota;
    }
    #endif // c++

    #if (defined(__cpp_lib_gcd_lcm))
    #else
    namespace std {
        using ::cxxomfort::cxxostd::gcd;
        using ::cxxomfort::cxxostd::lcm;
    }
    #endif

    #if (CXXOMFORT_CXX_STD < 2020 || defined(CXXOMFORT_USING_midpoint))
    namespace std {
        using ::cxxomfort::cxxostd::midpoint;
    }
    #endif


#endif // std using

#endif

/**
 * @page std0numeric <numeric>
 * @brief Backports related to Standard Header <tt><numeric></tt> 
 * 
 * @code
 * #include <cxxomfort/numeric.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * #include <cxxomfort/random.hpp>
 * @endcode
 * 
 * (or automatically by including <tt><cxxomfort/backports.hpp></tt> / <tt><cxxomfort/cxxomfort.hpp></tt>)
 * 
 * @section Interfaces
 * 
 * Interfaces defined in this section:
 * 
 * * @c iota() (C++11) -  generate sequences of increasing values.
 * * @c gcd() , @c lcm() (C++17) - math functions for greatest common divisor.
 * * @c midpoint() (C++20) - midpoint of integers, floating points or pointers.
 * * Numeric algorithms @c accumulate() and @c inner_product() with <tt>std::move()</tt> versions as per C++20, under namespace <tt>cxxomfort::cxxostd::cxx20</tt>.
 * 
 * Implementation fixes:
 * 
 * * (none)
 * 
 * @section move-numerics Numeric Algorithms with std::move (c++20)
 * 
 * Proposal @e p0616r0 introduces modifications to the algorithms under <tt><numeric></tt> so that they invoke <tt>std::move()</tt> on the accumulated cursor. The proposal indicates that this was first intended for C++11 but never proposed because of a lack of "bravery" as it was thought back then that this would introduce a breaking change.
 * 
 * Essentially, the new versions of these algorithms change the underlying operation loop as follows:
 * 
 * @code
 * 
 * // before p0616:
 * accumulate (ini, fin, v0) ;
 * { // loop body
 * v0 = v0 + *ini;
 * }
 * // after p0616:
 * accumulate (ini, fin, v0) ;
 * { // loop body
 * v0 = std::move(v0) + *ini;
 * }
 * 
 * @endcode
 * 
 * Because cxxomfort can not overwrite already existing functions, an opt-in shadowing mechanism is offered instead with the new functions being offered:
 * 
 * @code
 * using namespace std; // std::accumulate
 * accumulate (begin(S), end(S), accum, S0); // normal version
 * // alternative 1: via a different name
 * using cxxomfort::move_accumulate;
 * move_accumulate (begin(S), end(S), accum, S0);
 * // alternative 2: via an extra argument
 * using namespace cxxomfort::cxxostd::numeric20;
 * accumulate (begin(S), end(S), accum, S0, numeric_move_tag() ); // move version
 * @endcode
 * 
 * @section numeric_examples Examples
 * 
 * * @c <a href="https://notabug.org/lmachucabezzaza/cxxomfort-examples/src/master/examples/gcd.cpp">gcd</a> - gcd/lcm functions (and templates for C++<11).
 * * @c <a href="https://notabug.org/lmachucabezzaza/cxxomfort-examples/src/master/examples/accumulate20.cpp">accumulate20</a> - move-capable versions of algorithms from C++20.
 * 
 * @section also See Also
 * 
 * * @see @cppref{header/numeric} (cppreference)
 * * @see @ref std0algorithm (std generic algorithms)
 * * @ref cxxo-sup-numeric (supplemental).
 */
