#ifndef CXXOMFORT_RANDOM_HPP
#define CXXOMFORT_RANDOM_HPP
/**
 * @file
 * 
 */

#include <cxxomfort/base.hpp>
#include <limits>
#include <ctime>
#include <cstdlib> // random
#include <cxxomfort/array.hpp>

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
	#pragma message CXXO_NOTICE("enabled <random> support.")
#endif

//
// randmo 
// 

namespace cxxomfort { namespace fix {

/**
 * @xrefitem cxxomfort-fix "" "" Wraps @c rand() to avoid some biases.
 * @return Returns an @c int in the range <tt>[0, RAND_MAX)</tt>.
 * 
 * Invoking <tt>randmo(n)</tt> is similar to <tt>rand() % n</tt> except some biases in the distribution is eliminated. See <a  href="https://stackoverflow.com/questions/10984974/why-do-people-say-there-is-modulo-bias-when-using-a-random-number-generator">this SO question</a> for more details.
 * 
 * @see @c rand() .
 * 
 * */
static inline int 
randmo (int mo) {
    int x= 0;
    do {
        x = rand();
    } while (x >= (RAND_MAX - RAND_MAX % mo));
    x %= mo;
    return x;
}

} } // cxxomfort::fix

// 
// include header
// 

#if (CXXOMFORT_CXX_STD >= 2011 )
	#include <random>

#elif (CXXOMFORT_CXX_EMULATION >= 2011)
	#if (CXXOFLAG_directory_tr1>0)
		#include <tr1/random>
	#else 
		//#include <random>
	#endif
#elif (CXXOFLAG_USE_INTERNAL_TR1)
#endif

// 
// bring the basic names from the included header 
// 

#if ( \
(CXXO_COMPILERREQ(CXXO_COMPILERID_GCC, >=400) && CXXOFLAG_tr1) \
|| \
(CXXO_COMPILERREQ(CXXO_COMPILERID_CLANG, <800) && CXXOFLAG_tr1) \
)
	// when in non-C++0x mode in GCC, the TR1 implementation is broken
	// in particular ranlux, which we substitute with our own
	#include <tr1/random>
	#include <cxxomfort/impl/07-random_renames.hpp>

	#define CXXO_FIX_broken_subtract_with_carry_engine 1

namespace std {
	// GCC TR1 Random only includes the following:

    using ::cxxomfort::cxxostd::random_renames::linear_congruential_engine;
    using ::cxxomfort::cxxostd::random_renames::minstd_rand;
    using ::cxxomfort::cxxostd::random_renames::minstd_rand0;
    
    using tr1::mersenne_twister;
    using tr1::mt19937;
    
    // This is included but broken, so we purposefully ignore it
    //using tr1::subtract_with_carry;
    //using tr1::ranlux24_base;
    //using tr1::ranlux48_base;

	// Not included in 4.6, 3.8 or 5.0
    //using tr1::shuffle_order_engine;
    //using tr1::knuth_b;
    
    //using tr1::generate_canonical;
    
    using ::cxxomfort::cxxostd::random_renames::uniform_int_distribution;
    // GCC's uniform_real is broken
    //using cxxomfort::cxxostd::random::uniform_real_distribution; 

    using tr1::random_device;
}

	#include <cxxomfort/impl/07-ranlux.hpp>

namespace std {
    using ::cxxomfort::cxxostd::random::subtract_with_carry_engine;
    using ::cxxomfort::cxxostd::random::ranlux24_base;
    using ::cxxomfort::cxxostd::random::ranlux48_base;
}

	#include <cxxomfort/impl/07-discard_block.hpp>

namespace std {
	using ::cxxomfort::cxxostd::discard_block_engine;
	typedef discard_block_engine<ranlux24_base, 223, 23> ranlux24;
	typedef discard_block_engine<ranlux48_base, 389, 11> ranlux48;
}

	#include <cxxomfort/impl/07-shuffle_order.hpp>

namespace std {
	using ::cxxomfort::cxxostd::shuffle_order_engine;
	using ::cxxomfort::cxxostd::knuth_b;
}

	#include <cxxomfort/impl/07-uniform_arithmetic.hpp>

namespace std {
	using ::cxxomfort::cxxostd::random::uniform_real_distribution;
}

// fix  (due to MSVC compat)

namespace cxxomfort { namespace fix {
	using ::cxxomfort::cxxostd::shuffle_order_engine;
	using ::cxxomfort::cxxostd::knuth_b;
} }


#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1700)) 
	#include <random>

	//#include <cxxomfort/impl/07-shuffle_order.hpp>

namespace cxxomfort { namespace fix {
	using std::shuffle_order_engine;
	using std::knuth_b;
} }

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600)) 
	// 
	// MSVC 2010 ships with broken shuffle_order_engine and consequently knuth_b 
	// we can include all of <random> except  cxxomfort::fix::shuffle_order_engine 
	#include <random>

	#define CXXO_FIX_broken_shuffle_order_engine 1

namespace std {
	using tr1::linear_congruential_engine;
	using tr1::minstd_rand;
	using tr1::minstd_rand0;
	
    using tr1::mersenne_twister;
    using tr1::mt19937;
    
    using tr1::subtract_with_carry_engine;
    using tr1::ranlux24_base;
    using tr1::ranlux48_base;
    
    using tr1::discard_block_engine;
    using tr1::ranlux24;
    using tr1::ranlux48;
}

	#include <cxxomfort/impl/07-shuffle_order.hpp>

#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1500) && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,<1600))
	// 
	// MSVC 2008
	//
	#include <random>
	#include "impl/07-lcg.hpp"
	#include "impl/07-random_renames.hpp"

	#define CXXO_FIX_broken_subtract_with_carry_engine 1

namespace std {
    using ::cxxomfort::cxxostd::random::linear_congruential_engine;
    using ::cxxomfort::cxxostd::random::minstd_rand;
    using ::cxxomfort::cxxostd::random::minstd_rand0;
    
    using tr1::mersenne_twister;
    using tr1::mt19937;
}

	#include "impl/07-ranlux.hpp"

namespace std {
    // this is broken in MSVC 2008, just like in GCC
    using ::cxxomfort::cxxostd::random::subtract_with_carry_engine;
    using ::cxxomfort::cxxostd::random::ranlux24_base;
    using ::cxxomfort::cxxostd::random::ranlux48_base;
}

	#include "impl/07-discard_block.hpp"

namespace std {
	using ::cxxomfort::cxxostd::discard_block_engine;
	typedef discard_block_engine<ranlux24_base, 223, 23> ranlux24;
	typedef discard_block_engine<ranlux48_base, 389, 11> ranlux48;
}
	#include <cxxomfort/impl/07-uniform_arithmetic.hpp>

namespace std {

	// in MSVC 2008 specifically, uniform_int and uniform_real cause constness errors.
	#if (CXXOMFORT_COMPILER_VERSION<1600)
    using ::cxxomfort::cxxostd::random::uniform_int_distribution;
    using ::cxxomfort::cxxostd::random::uniform_real_distribution;
    #else
    using ::cxxomfort::cxxostd::random_renames::uniform_int_distribution;
    using ::cxxomfort::cxxostd::random_renames::uniform_real_distribution;
    #endif
    
    using tr1::random_device;
}

	#include <cxxomfort/impl/07-shuffle_order.hpp>

namespace std {
	using ::cxxomfort::cxxostd::shuffle_order_engine;
	using ::cxxomfort::cxxostd::knuth_b;
}

#elif (defined(CXXOFLAG_USE_INTERNAL_TR1))

	#include <cxxomfort/impl/07-lcg.hpp>
	#include <cxxomfort/impl/07-ranlux.hpp>
	
namespace std {
	using ::cxxomfort::cxxostd::random::linear_congruential_engine;
	using ::cxxomfort::cxxostd::random::minstd_rand;
	using ::cxxomfort::cxxostd::random::minstd_rand0;
	using ::cxxomfort::cxxostd::random::subtract_with_carry_engine;
	using ::cxxomfort::cxxostd::random::ranlux24_base;
	using ::cxxomfort::cxxostd::random::ranlux48_base;
}

	#include <cxxomfort/impl/07-shuffle_order.hpp>

namespace std {
	using ::cxxomfort::cxxostd::random::shuffle_order_engine;
	using ::cxxomfort::cxxostd::random::knuth_b;
}

	#include <cxxomfort/impl/07-uniform_arithmetic.hpp>

namespace std {
	using ::cxxomfort::cxxostd::random::uniform_int_distribution;
	using ::cxxomfort::cxxostd::random::uniform_real_distribution;
}

	#include <cxxomfort/impl/07-shuffle_order.hpp>

#endif



#if (CXXOMFORT_CXX_STD < 1997)
	// nothing to do, can't support
	#undef  CXXOMFORT_IMPLEMENTS_header_random
	#define CXXOMFORT_IMPLEMENTS_header_random 0

#elif (CXXOMFORT_CXX_STD < 2011 && CXXO_CXXSTDHEADER_random==0 && CXXO_CXXSTDHEADER_tr1_random==0)
	// nothing to do, until we backport all

#elif (CXXOMFORT_CXX_STD < 2011)
	#undef  CXXOMFORT_IMPLEMENTS_header_random
	#define CXXOMFORT_IMPLEMENTS_header_random CXXO_IMPLSTATUS_BACKPORT()

	#if (CXXOFLAG_directory_tr1>0)
		#include <tr1/random>
	#else 
		#include <random>
	#endif

#elif (CXXOMFORT_CXX_STD < 9999)
    #undef  CXXOMFORT_IMPLEMENTS_header_random
    #define CXXOMFORT_IMPLEMENTS_header_random CXXO_IMPLSTATUS_NATIVE()
    #include <random>
#endif

// 
// cxxomfort fix for shuffle_order_engine
// 

#if (CXXOMFORT_CXX_STD >= 2011)
namespace cxxomfort { namespace fix {
	using std::shuffle_order_engine;
	using std::knuth_b;
} }
#elif (defined(CXXO_FIX_broken_shuffle_order_engine) )
namespace cxxomfort { namespace fix {
	using ::cxxomfort::cxxostd::random::shuffle_order_engine;
	using ::cxxomfort::cxxostd::random::knuth_b;
} }
#elif (defined(CXXOFLAG_USE_INTERNAL_TR1))
namespace cxxomfort { namespace fix {
	using ::cxxomfort::cxxostd::random::shuffle_order_engine;
	using ::cxxomfort::cxxostd::random::knuth_b;
} }
#else 
namespace cxxomfort { namespace fix {
	using std::shuffle_order_engine;
	using std::knuth_b;
} }

#endif



// TBA: Backport discrete_distribution to c++03

// 
// select default_random_engine in C++07
// 

#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_CXXSTDHEADER_random>0)
#elif (CXXOMFORT_CXX_STD >= 1997 && CXXOMFORT_IMPLEMENTS_header_random >= CXXO_IMPLSTATUS_EMULATION())

namespace std {

	typedef std::ranlux24_base default_random_engine;

}

#endif

#endif // header


#ifdef DOXYGEN_DOC
namespace std {

/**
 * @brief PRNG Linear congruential generator <tt>x_{i+1} ← (a×x_{i}+c) mod m</tt> with parameters @a a , @a c , @a m .
 * @cxx03b Linear congruential generator.
 * @xrefitem std0random "" "" Linear congruential generator.
 * 
 * @see @cppref{numeric/random/linear_congruential_engine}
 * 
 * */
template <typename UInt, UInt a, UInt c, UInt m>
struct linear_congruential_engine {};

//! @xrefitem std0random "" "" LCG with parameters given by Lewis, Goodman and Miller 1988.
typedef linear_congruential_engine<uint_fast32_t, 16807, 0, 2147483647> minstd_rand0;
//! @xrefitem std0random "" "" LCG with parameters given by Park, Miller, and Stockmeyer 1993.
typedef linear_congruential_engine<uint_fast32_t, 48271, 0, 2147483647> minstd_rand;

/**
 * @brief PRNG Subtraction engine <tt>x_{i+1} ← Y(x_{i-s}, x_{i-r})</tt> with parameters @a w , @a s , @a r .
 * @ingroup std0random
 * @cxx03b Subtract with carry engine.
 * @xrefitem std0random "" "" Subtract with carry engine.
 * 
 * @see @cppref{numeric/random/subtract_with_carry_engine}
 * */
template <typename I, size_t w, size_t s, size_r >
struct subtract_with_carry_engine {};

//! SWC engine with common parameter sets
//! @cxx03b Subtract with carry engine.
typedef subtract_with_carry_engine<uint_fast32_t, 24, 10, 24> ranlux24_base;
//! SWC engine with common parameter sets
//! @cxx03b Subtract with carry engine.
typedef subtract_with_carry_engine<uint_fast64_t, 48, 5, 12> ranlux48_base;

/**
 * @typedef default_random_engine
 * @brief cxxomfort uses @c ranlux24_base as its "default random engine".
 * @ingroup std0random
 * @xrefitem std0random "" "" Provides @c ranlux24_base as the default PRNG in cxxomfort.
 * */
typedef std::ranlux24_base default_random_engine;


} // std
#endif

#if defined(DOXYGEN_DOC)

//! @xrefitem cxxomfort-fix "" "" This macro is defined if cxxomfort has to substitute broken @c shuffle_order_engine in eg.: MSVC.
#define CXXO_FIX_broken_shuffle_order_engine 

//! xrefitem cxxomfort-fix "" "" This macro is defined if cxxomfort has to substitute broken @c ranlux in eg.: GCC < 6.0.
#define CXXO_FIX_broken_subtract_with_carry_engine 




/**
 * @brief Forwarder for @c std::shuffle_order_engine
 * @ingroup cxxomfort-fixes
 * @xrefitem std0random "" "" Forwarder.
 * 
 * Some compilers, such as MSVC 2008, 2010, don't implement @c shuffle_order_engine , and some other compilers don't implement it at all. This name will redirect to cxxomfort's implementation in those cases, or to <tt>std::shuffle_order_engine</tt> otherwise.
 * 
 * */
template <typename Engine, unsigned K> struct shuffle_order_engine {};

/**
 * @brief Forwarder for the @c std::knuth_b specialization of @c shuffle_order_engine .
 * @ingroup cxxomfort-fixes
 * 
 * See @c shuffle_order_engine above.
 * 
 * */
typedef shuffle_order_engine<minstd_rand,256> knuth_b;


} } // cxxomfort::fix


#endif

/**
 * @page std0random <random>
 * @brief Backports related to Standard Header <tt><random></tt>
 * 
 * @code
 * #include <cxxomfort/random.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * 
 * Including the header brings up <random>'s features, accounting for eg.: 
 * location being @c std::tr1::random... where set up in C++03, 
 * as well as changes to their names in certain compiler versions before C++11 was finalized.
 * 
 * @section Interfaces
 * 
 * Interfaces defined in this section:
 * 
 * * @c default_random_engine — in cxxomfort defined as a typedef for @c ranlux24_base for C++03.
 * * @c subtract_with_carry and its typedefs @c ranlux24_base and @c ranlux48_base , for pre-C++11 compilers or when it's found to be broken.
 * 
 * If @c CXXOFLAG_USE_INTERNAL_TR1 is enabled for pre-C++03 compilers, the following features are also enabled:
 * 
 * * @c linear_congruential_generator for pre-C++03 compilers.
 * * @c subtract_with_carry_engine for pre-C++03 compilers.
 * * @c shuffle_order_engine and @c knuth_b for pre-C++03 compilers.
 * 
 * Interfaces repointed here for GCC and Clang:
 * 
 * * @c linear_congruent_engine and its named typedefs @c minstd , @c minstd0 , when they appear in TR1 with a different name.
 * * @c uniform_int_distribution , @c uniform_real_distribution , when they appear in TR1 with a different name.
 * 
 * Fixes to implementation issues, in <tt>cxxomfort::fix::</tt>.
 * 
 * * @ref randmo() <tt>randmo()</tt>  - modulo-respecting call to <tt>rand()</tt>.
 * * <tt>shuffle_order_engine</tt> - for broken std in MSVC 2010.
 * * @c knuth_b - for broken std in MSVC 2010.
 *
 * Pending:
 * 
 * * @c generate_canonical complementary (C++11).
 * * @c uniform_int_distribution complementary (C++11).
 * * @c discrete_distribution complementary (C++11).
 *
 * @section unbiased_randmo Unbiased rand() ( randmo )
 * 
 * The cxxomfort library adds a simple wrapper @ref randmo() around the <tt>std::rand()</tt> function.
 * 
 * @section also See Also
 * 
 * @see @cppref{header/random} (cppreference)
 * @see @ref cxxo-sup-random (supplemental)
 * 
 * */
