#ifndef CXXOMFORT_RATIO_HPP
#define CXXOMFORT_RATIO_HPP
/**
 * @file
 * @brief Implementations, Backports and helpers for Standard header <code>\<ratio\></code>.
 * @ingroup std0ratio
 * @cxx11b Header file implementing <ratio>
 * 
 */

#include <cxxomfort/base.hpp>

#if (CXXOMFORT_CXX_STD >= 2011 || CXXO_CXXSTDHEADER_ratio)
	#include <ratio>
#else
	#define CXXO_USING_ratio
#endif

#include <cxxomfort/type_traits.hpp>

#ifdef CXXO_USING_ratio
	#include <cxxomfort/cxxo-utils.hpp>

namespace std {

    /**
     * @ingroup std0ratio
     * @{
     * */
     
    /**
     * @brief Represents finite rational numbers with numerator and denominator that fit into C++'s type system.
     * @tparam N Numerator, a value of type @c intmax_t 
     * @tparam D Denominator, a value of type @c intmax_t 
     * @cxx11b Rational arithmetic type.
     * @xrefitem std0ratio "" "" Rational arithmetic type.
     * */
    template <intmax_t N, intmax_t D = 1>
    struct ratio {
	    static_assert(D != 0, "Denominator may not be zero");
	    private:
	    static const intmax_t xG = cxxomfort::st_math::static_gcd_typed<intmax_t,N,D>::value;
	    static const intmax_t uN = cxxomfort::st_math::static_abs_s<N>::value;
	    public:
	    static const intmax_t num = cxxomfort::st_math::static_sign_s<N>::value * cxxomfort::st_math::static_sign_s<D>::value * uN / xG;
	    static const intmax_t den = D / xG;
	    //! Exposes the proper rational number type, in its reduced form.
	    typedef ratio<num,den> type;
    };

// 
// ratio prefixes
// 

// straight outta cppreference

    //! Convenience SI typedef
    typedef ratio<1,       1000000000000000000> atto;   
    //! Convenience SI typedef
    typedef ratio<1,          1000000000000000> femto;  
    //! Convenience SI typedef
    typedef ratio<1,             1000000000000> pico;   
    //! Convenience SI typedef
    typedef ratio<1,                1000000000> nano;   
    //! Convenience SI typedef
    typedef ratio<1,                   1000000> micro;  
    //! Convenience SI typedef
    typedef ratio<1,                      1000> milli;  
    //! Convenience SI typedef
    typedef ratio<1,                       100> centi;  
    //! Convenience SI typedef
    typedef ratio<1,                        10> deci;   
    //! Convenience SI typedef
    typedef ratio<                       10, 1> deca;   
    //! Convenience SI typedef
    typedef ratio<                      100, 1> hecto;  
    //! Convenience SI typedef
    typedef ratio<                     1000, 1> kilo;   
    //! Convenience SI typedef
    typedef ratio<                  1000000, 1> mega;   
    //! Convenience SI typedef
    typedef ratio<               1000000000, 1> giga;   
    //! Convenience SI typedef
    typedef ratio<            1000000000000, 1> tera;   
    //! Convenience SI typedef
    typedef ratio<         1000000000000000, 1> peta;   
    //! Convenience SI typedef
    typedef ratio<      1000000000000000000, 1> exa;    

// 
// comparators 
// 

//! @xrefitem std0ratio "" "" Equality comparison.
template< class R1, class R2 >
struct ratio_equal 
: public std::integral_constant < bool
    , R1::num == R2::num && R1::den == R2::den
> 
{ 
};

//! @xrefitem std0ratio "" "" Inequality comparison.
template< class R1, class R2 >
struct ratio_not_equal 
: public std::integral_constant < bool
	, !(ratio_equal<R1,R2>::value )
> 
{ 
};

//! @internal
namespace detail {
template <typename R1, typename R2>
struct ratio_less_impl {
	static const intmax_t N1 = R1::num;
	static const intmax_t N2 = R2::num;
	static const intmax_t GN = cxxomfort::st_math::static_gcd_i<N1,N2>::value;

	static const intmax_t D1 = R1::den;
	static const intmax_t D2 = R2::den;
	static const intmax_t GD = cxxomfort::st_math::static_gcd_i<D1,D2>::value;
	
	static const bool value = ((N1/GN) * (D2/GD)) < ((N2/GN) * (D1/GD)) ;

};
} // detail


//! Helper type that compares two @c ratio R1 and R2 for if <tt>R1 < R2</tt> (after reduction)
//! @xrefitem std0ratio "" "" Relational comparator for @c std::ratio .
template <typename R1, typename R2>
struct ratio_less 
: std::integral_constant<bool, detail::ratio_less_impl<R1,R2>::value>
{ };

//! Helper type that compares two @c ratio R1 and R2 for if <tt>R1 > R2</tt> (after reduction)
//! @xrefitem std0ratio "" "" Relational comparator for @c std::ratio .
template <typename R1, typename R2>
struct ratio_greater 
: std::integral_constant<bool, ratio_less<R2,R1>::value >
{ };

//! Helper type that compares two @c ratio R1 and R2 for if <tt>R1 <= R2</tt> (after reduction)
//! @xrefitem std0ratio "" "" Relational comparator for @c std::ratio .
template <typename R1, typename R2>
struct ratio_less_equal
: std::integral_constant<bool, !ratio_greater<R2,R1>::value >
{ };

//! Helper type that compares two @c ratio R1 and R2 for if <tt>R1 >= R2</tt> (after reduction)
//! @xrefitem std0ratio "" "" Relational comparator for @c std::ratio .
template <typename R1, typename R2>
struct ratio_greater_equal
: std::integral_constant<bool, !ratio_less<R1,R2>::value >
{ };

// 
// ratio arithmetic meta 
// 

    /**
     * @}
     * */
     
} // std
#endif // _ratio
#endif

/**
 * @page std0ratio <ratio>
 * @brief Backports related to Standard header <tt><ratio></tt>
 * 
 * @code
 * #include <cxxomfort/ratio.hpp>
 * @endcode
 * 
 * (or automatically by including <tt><cxxomfort/backports.hpp></tt> / <tt><cxxomfort/cxxomfort.hpp></tt>)
 *
 * @section Interfaces
 * 
 *The following interfaces are provided by this header:
 * 
 * * @ref @c std::ratio 
 * * @c ratio_equal , @c ratio_not_equal 
 * * @c ratio_less , @c ratio_greater
 * * The @ref ratio_si_prefixes "SI prefix names" such as @c milli .
 * 
 * @section ratio_si_prefixes SI prefixes
 * 
 * C++11 @c <ratio> also provides a number of convenience typedefs representing SI ratio/factor prefixes such as @e mega or @e centi . Some of those names are also exposed by cxxomfort:
 * 
 * * @c atto , @c femto , @c pico , @c nano , @c micro , @c milli , @c enti , @c deci
 * * @c deca , @c hecto , @c kilo , @c mega , @c giga , @c tera , @c peta , @c exa
 * 
 * As of this release the prefixes higher than <tt>10^18</tt> or smaller than <tt>10^-18</tt>, such as @c yotta , are not provided.
 * 
 * 
 * @section also See Also
 * 
 * @see @cppref{header/ratio} (cppreference)
 * @see @ref cxxo-sup-ratio (supplemental)
 * 
 * */
