#ifndef CXXOMFORT_SPAN_HPP
#define CXXOMFORT_SPAN_HPP
/*
 * @file
 * @brief Implementations and additions tied to Standard header <tt><span></tt>.
 * @ingroup std0span
 * 
 */

#include <cxxomfort/base.hpp>
#include <cxxomfort/config.hpp>

#if (CXXO_CXXSTDHEADER_span)
	#include <span>
#elif (defined(__has_include))
	#if (__has_include(<span>))
		#include <span>
	#endif
#endif

#if !defined(__cpp_lib_span)
	#include <cxxomfort/cstdint.hpp>
	#include <cxxomfort/iterator.hpp>
	#include <cxxomfort/type_traits.hpp>
	#define __cpp_lib_span (CXXOMFORT_CXX_STD*100)

namespace std {

namespace {
static CXXO_CONSTEXPR size_t const dynamic_extent = -1;
}

template <typename T, size_t Ex = dynamic_extent >
struct span {
	typedef T 	element_type;
	typedef typename ::cxxomfort::traits::remove_cv<T>::type value_type;
	typedef size_t 	size_type;
	typedef ptrdiff_t 	difference_type;
	
	typedef T* 	pointer;
	typedef T const* 	const_pointer;
	
	typedef T& 	reference;
	typedef T const& 	const_reference;
	
	typedef T* 	iterator;
	typedef T const* 	const_iterator;
	
	public:
	CXXO_CONSTEXPR 
	explicit span (pointer p) CXXO_NOEXCEPTNOTHROW : ptr(p) {}
	
	template <typename It> CXXO_CONSTEXPR14 
	explicit span (It i, size_type n
	, typename std::enable_if< std::is_same
		< value_type, typename std::remove_cv< typename std::iterator_traits<It>::value_type >::type 
		>::value , bool>::type* = 0) CXXO_NOEXCEPT 
	: ptr(i) {
		if (n != Ex) std::terminate();
	}
	
	CXXO_CONSTEXPR 
	span (element_type(&arr)[Ex]) CXXO_NOEXCEPTNOTHROW 
	: ptr(arr) { }
	
	template <typename U>
	span (std::array<U,Ex> const& arr
	, typename std::enable_if<std::is_same <value_type, typename std::remove_cv<U>::type >::value , bool*>::type = 0
	) CXXO_NOEXCEPT 
	: ptr(arr.data()) { }
	
	
	
	CXXO_CONSTEXPR 
	size_type 	size () const CXXO_NOEXCEPTNOTHROW { return Ex; }
	CXXO_CONSTEXPR 
	const_pointer 	data () const CXXO_NOEXCEPTNOTHROW { return ptr; }
	pointer 	data () CXXO_NOEXCEPT { return ptr; }
	
	CXXO_CONSTEXPR 
	const_reference 	operator[] (size_type idx) const CXXO_NOEXCEPTNOTHROW {
		return ptr[idx];
	}
	reference 	operator[] (size_type idx) CXXO_NOEXCEPT { return ptr[idx]; }
	
	const_iterator 	begin () const CXXO_NOEXCEPT { return const_iterator(ptr); }
	iterator 	begin () CXXO_NOEXCEPT { return iterator(ptr); }
	const_iterator 	end () const CXXO_NOEXCEPT { return const_iterator(ptr+Ex); }
	iterator 	end () CXXO_NOEXCEPT { return iterator(ptr+Ex); }

	private:
	span ();
	T* ptr;
}; // span

template <typename T>
struct span<T, dynamic_extent> {
	private:
	typedef ::cxxomfort::array_ref<T, size_t> ref_t;
	
	public:
	typedef T 	element_type;
	typedef typename ::cxxomfort::traits::remove_cv<T>::type value_type;
	typedef size_t 	size_type;
	typedef ptrdiff_t 	difference_type;
	
	typedef typename ref_t::pointer 	pointer;
	typedef typename ref_t::const_pointer 	const_pointer;
	
	typedef typename ref_t::reference 	reference;
	typedef typename ref_t::const_reference 	const_reference;
	
	typedef typename ref_t::iterator 	iterator;
	typedef typename ref_t::const_iterator 	const_iterator;
	
	
	public:
	CXXO_CONSTEXPR 
	explicit span (T* ptr, size_t n) CXXO_NOEXCEPT : ref(ptr,n) { }
	explicit span (const_iterator ini, const_iterator fin)
	: ref(ini,fin) { }
	
	private:
	ref_t ref;
};

}
#endif


#endif

/**
 * @page std0span <span>
 * @brief Features related to Standard header <tt><span></tt>.
 * 
 * @code
 * #include <cxxomfort/span.hpp>
 * // or as part of:
 * #include <cxxomfort/backports.hpp>
 * @endcode
 * @section Interfaces 
 * 
 * Features included in this section and mapped to global nsmespace or <tt>std::</tt>:
 * 
 * * std::span 
 * 
 * 
 * */
