#ifndef CXXOMFORT_SYSTEM_ERROR_HPP
#define CXXOMFORT_SYSTEM_ERROR_HPP
/**
 * @file
 * @xrefitem std0system_error "<system_error>" "Header"
 * 
 */

#include <cxxomfort/base.hpp>

#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
	#pragma message CXXO_NOTICE("enabled <system_error> support.")
#endif

/*
 * Notes
 * 
 * C++>C++11  has <system_error>
 * MSVC 2010, 2012 with C++<C++11 has a borked <system_error> with bad errc
 * 
 * */

#if (CXXOMFORT_CXX_STD>=2014)

	#define CXXOMFORT_IMPLEMENTS_header_system_error CXXO_IMPLSTATUS_NATIVE()
	#define CXXOMFORT_IMPLEMENTS_posix2008_errc CXXO_IMPLSTATUS_NATIVE()
	#include <system_error>

namespace cxxomfort { namespace fix {
	using std::errc;
} }


#elif (CXXO_CXXSTDHEADER_system_error>0)
	#define CXXOMFORT_IMPLEMENTS_header_system_error CXXO_IMPLSTATUS_NATIVE()
	#include <system_error>

	#if (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC, >= 1600) && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC, < 1800))
		#pragma message ("header system_error in MSVC")
		#define CXXO_FIX_broken_msvc_errc

namespace cxxomfort { namespace fix {
	using std::errc::errc;
} }
	#else

namespace cxxomfort { namespace fix {
	using std::errc;
} }

	#endif

#elif (CXXOMFORT_CXX_STD>=1997)

	#define CXXOMFORT_IMPLEMENTS_header_system_error CXXO_IMPLSTATUS_EMULATION()

/*
 * MSVC 2010 has a broken <system_error> that defines errc inside a 
 * namespace also called errc, making it impossible to 
 * declare variables of type std::errc.
 * 
 * Because of this, in the case of MSVC 2010 specifically, 
 * we probably should be overriding the native <system_error> in config/msc.hpp
 * 
 * */

	#if (defined(CXXOMFORT_NOTICES) && (CXXOMFORT_NOTICES > 1))
		#pragma message CXXO_NOTICE("backporting <system_error> support.")
	#endif

	#include <cxxomfort/cerrno.hpp>
	#include "impl/relationals.hpp"
	#include <cxxomfort/memory.hpp>
	#include "impl/11-to_string.hpp"
	#include <stdexcept>
	#include <string>
	#include <cstring> // strerror
	#include <utility> // make_pair, operator<

	#include <cxxomfort/utils/typesafe_enum.hpp>
	#include <cxxomfort/impl/errc.hpp>

	#if (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_MSC)
	#pragma warning (disable:4996) // MSVC: function may be unsafe
	#endif

	#if defined(__linux__) || defined(__unix__)
		#include <unistd.h> // POSIX
	#elif (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_MSC || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_OW)
		#include <errno.h>
		//#include "impl/errno-msvc.hpp"
	#endif

// this goes in namespace std because users need to specialize
namespace std {
// errc defined in errc.hpp
// is_error_code_enum defined in errc.hpp

template <typename T> struct is_error_condition_enum
: public std::false_type {};

template<> struct is_error_condition_enum<errc> : public std::true_type {};
//template<> struct is_error_condition_enum<errc::type> : public std::true_type {};

class error_category;
class error_condition;
class error_code;

//! Create an @c error_code .
static error_code make_error_code ( errc e ) CXXO_NOEXCEPTNOTHROW;


/**
 * @brief Describes an error category (a classification of sources for errors).
 * @see https://en.cppreference.com/w/cpp/error/error_category
 * 
 * The objects of @c error_category classes are treated as singletons, passed by reference. 
 */
class error_category {

    public:

    CXXO_DEFAULT_DEFAULT_CONSTRUCTOR(error_category);
    CXXO_DELETED_COPY_CONSTRUCTOR(error_category);
    virtual ~error_category () {};
    CXXO_DELETED_COPY_ASSIGN(error_category);
    
    //! Deriveds implement a naming schema for the category here.
    virtual const char* name () const CXXO_NOEXCEPTNOTHROW = 0;
    //! Deriveds provide a descriptive message about the error here.
    virtual std::string message (int condition ) const = 0;
    
    virtual error_condition 
    default_error_condition (int code ) const CXXO_NOEXCEPTNOTHROW;

    virtual bool 
    equivalent (int code, error_condition const& condition ) const CXXO_NOEXCEPTNOTHROW;
    
    virtual bool 
    equivalent (std::error_code const& ec, int condition) const CXXO_NOEXCEPTNOTHROW;

    friend bool operator== (error_category const& vl, error_category const& vr) CXXO_NOEXCEPT {
        return &vl == &vr;
    }
    friend bool operator< (error_category const& vl, error_category const& vr) CXXO_NOEXCEPT {
        return &vl < & vr;
    }

    private:

};

CXXO_GENERATE_RELATIONALS(error_category);

//
// error_category helpers
//

/**
 * @brief Type that describes the category for generic errors.
 * */
struct generic_category_t : public error_category {
    ~generic_category_t () {}

    const char* name () const CXXO_NOEXCEPTNOTHROW {
        return "generic";
    }
    std::string message (int condition) const {
        return std::string(name()) + ": " + std::strerror(condition);
    }
};

//! Obtains a reference to the @c error_category object for generic errors.
static inline error_category const& generic_category() CXXO_NOEXCEPTNOTHROW {
    static generic_category_t g;
    return g;
}

/**
 * @brief Type that describes the category for system errors.
 * */
struct system_category_t  : public error_category {
    ~system_category_t () {}

    const char* name () const CXXO_NOEXCEPTNOTHROW {
        return "system";
    }
    std::string message (int condition) const {
        return std::string(name()) + ": " + std::strerror(condition);
    }
};

//! Obtains a reference to the @c error_category object for errors reported by the operating system. 
static inline error_category const& system_category() CXXO_NOEXCEPTNOTHROW {
    static system_category_t g;
    return g;
}

/**
 * @brief C++ provided platform-independent error code.
 * 
 * An @c error_condition encapsulates an error code as 
 * provided by eg.: @c errc , and an error category 
 * as provided by @c error_category. 
 * 
 * In C++11 onwards @c error_condition and others are used by the 
 * Standard Library; cxxomfort provides them only for header 
 * compatibility but in C++03 mode they are not used, 
 * with the exception of @c errc .
 * 
 * */
class error_condition {
    public:
    error_condition () CXXO_NOEXCEPTNOTHROW 
    : v_error(0), p_category( &generic_category() ) 
    {}

    CXXO_DEFAULT_COPY_CONSTRUCTOR(error_condition, CXXO_NOTHROW, 
    v_error(From.v_error), p_category(From.p_category) );

    error_condition ( int val, error_category const& cat ) CXXO_NOEXCEPTNOTHROW
    : v_error(val), p_category(std::addressof(cat))
    {};

    // DMC does not support SFINAE-esques in member template it seems
	#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_OW)
    template <typename EnumE>
    error_condition (EnumE e) CXXO_NOEXCEPTNOTHROW {
        *this= make_error_condition(e);
    }
	#else
		#define CXXO_SIGCONDITION (::std::is_error_condition_enum<EnumE>::value)
    template <typename EnumE CXXO_ENABLEIF_TEMPLATE()>
    error_condition (EnumE e CXXO_ENABLEIF_FUNCTION() ) CXXO_NOEXCEPTNOTHROW 
    {
        // this is fine to do so long as we don't add bases to error_condition
        new (this) error_condition (make_error_condition(e));
    }
		#undef CXXO_SIGCONDITION
	#endif

    void       assign (int val, error_category const& cat ) CXXO_NOEXCEPTNOTHROW {
        v_error= val;
        p_category= &cat;
    }
    void       clear () CXXO_NOEXCEPTNOTHROW {
        v_error= 0;
        p_category= &generic_category();
    }
    
    int        value () const CXXO_NOEXCEPTNOTHROW { return v_error; }
    error_category const& category () const CXXO_NOEXCEPTNOTHROW { return *p_category; }
    std::string message () const;
    
    CXXO_EXPLICIT_OPERATOR(bool) () const CXXO_NOEXCEPT { return !!v_error; }

    friend bool operator== (error_condition const& l, error_condition const& r) CXXO_NOEXCEPTNOTHROW {
        return (l.p_category == r.p_category && l.v_error == r.v_error);
    }
    friend bool operator< (error_condition const& l, error_condition const& r) CXXO_NOEXCEPTNOTHROW {
        return std::make_pair(l.p_category, l.v_error) < std::make_pair(r.p_category, r.v_error);
    }

    private:
    
    int v_error;
    error_category const* p_category;
    
};

CXXO_GENERATE_RELATIONALS(error_condition);

//! Creates an @c error_condition of generic category.
#if (CXXOMFORT_CXX_STD >= 1997)
static inline 
error_condition make_error_condition ( errc e ) CXXO_NOEXCEPTNOTHROW {
    return error_condition( explicit_cast<int>(e), generic_category() );
}
#else
static inline 
error_condition make_error_condition ( errc e ) CXXO_NOEXCEPTNOTHROW {
    return error_condition( cxxomfort::impl::to_value(e), generic_category() );
}
#endif


} // namespace std

// 
// error_code
// 

namespace std {
/**
 * Represents an error code.
 */
class error_code {
    public:
    
    error_code () CXXO_NOEXCEPTNOTHROW 
    : ec(0), ct(std::addressof(system_category())) {}

    error_code (int err, error_category const& ecat) CXXO_NOEXCEPTNOTHROW
    : ec(err), ct(std::addressof(ecat)) {}

	#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC || CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_OW)
	#else
		
    template <typename EnumE>
    error_code (EnumE e , typename std::enable_if<is_error_code_enum<EnumE>::value,bool*>::type = 0 ) CXXO_NOEXCEPTNOTHROW 
    : ec(0), ct(std::addressof(system_category())) 
    {
		//std::cerr<< "[error_code::error_code(e) with ADL call]"<< std::flush;
		using std::make_error_code;
		*this= make_error_code(EnumE(e));
    }
    template <typename EnumE>
    error_code (EnumE e , typename std::enable_if<!is_error_code_enum<EnumE>::value,bool*>::type = 0 ) ;

	#endif

    void       assign (int err, error_category const& ecat) CXXO_NOEXCEPTNOTHROW {
        ec= err; ct= std::addressof(ecat);
    }
    
    void       clear () CXXO_NOEXCEPTNOTHROW {
        assign(0, system_category());
    }
    
    //! Returns the value of the error code.
    int        value () const CXXO_NOEXCEPTNOTHROW { return ec; }
    //! Returns the category of the error code.
    error_category const& category () const CXXO_NOEXCEPTNOTHROW { return *ct; }
    //! Returns the composed message for the error code.
    std::string message () const { return category().message(value()); }
    // Checks if the error value is set.
    // operator bool () const { return value() != 0; }
    
    error_condition default_error_condition () const CXXO_NOEXCEPTNOTHROW {
        return category().default_error_condition(value());
    }

    CXXO_EXPLICIT_OPERATOR(bool) () const CXXO_NOEXCEPTNOTHROW { return !!value(); }
    
    // comparison operators
    
    friend bool operator== (error_code const& vl, error_code const& vr) CXXO_NOEXCEPTNOTHROW {
        return vl.category()==vr.category() && vl.value()==vr.value();
    }
    friend bool operator<  (error_code const& vl, error_code const& vr) CXXO_NOEXCEPTNOTHROW {
        return std::make_pair(&(vl.category()),vl.value()) < std::make_pair(&(vr.category()),vr.value());
    }

    private:
    int ec;
    error_category const* ct;
};

} // std


//
// virtuals
//

namespace std {

error_condition error_category::default_error_condition (int code ) const CXXO_NOEXCEPTNOTHROW {
	return error_condition(code, *this);
}

bool error_category::equivalent (int code, error_condition const& condition ) const CXXO_NOEXCEPTNOTHROW {
	return default_error_condition(code) == condition;
};

bool error_category::equivalent (error_code const& ec, int condition) const CXXO_NOEXCEPTNOTHROW {
	return *this == ec.category() && ec.value() == condition;
}


//
// make helpers
//

static error_condition make_error_condition ( ::std::errc ) CXXO_NOEXCEPTNOTHROW;

//
// exceptions
//

class system_error
: public std::runtime_error {
    public:

    system_error (error_code ec )
    : std::runtime_error("system_error"), E(ec) , W() {}
    system_error (error_code ec, std::string const& w)
    : std::runtime_error(w), E(ec), W(w) {}
    system_error (error_code ec, const char* const w)
    : std::runtime_error(w), E(ec), W(w) {}
    system_error (int v, error_category const& ecat)
    : std::runtime_error(error_code(v,ecat).message()), E(v,ecat), W(error_code(v,ecat).message())
    {}
    system_error (int v, error_category const& ecat, const char* const w)
    : std::runtime_error(w), E(v,ecat), W(w) {}

	#if (CXXOMFORT_CXX_STD>=1997)
    // we enable construction from a std::ios::failure for conversion
    system_error (std::ios::failure const& Ex)
    : std::runtime_error(Ex.what()), E(errno,system_category()), W(Ex.what())
    {}

	#endif

    ~system_error () CXXO_NOEXCEPTNOTHROW {}

    error_code const& code() const CXXO_NOEXCEPTNOTHROW {
        return E;
    }
    
    private:
    error_code const E;
    std::string W;
};


//
// out-of-class members
//

std::string error_condition::message () const {
    return category().message(value());
}

// nonmember helpers

	#if (CXXOMFORT_CXX_STD >= 1997)
template <typename C, typename CT>
std::basic_ostream<C,CT> & operator<< (std::basic_ostream<C,CT>& os, error_code const& ec) CXXO_NOEXCEPT {
    return os << ec.category().name() << ':' << ec.value();
}


	#endif

// end namespaces

}

// these are defined in std because users can specialize

namespace std {
static inline error_code make_error_code( errc e ) CXXO_NOEXCEPTNOTHROW {
	int xe = explicit_cast<int>(e);
    return error_code(xe, generic_category());
}
}

	#if (CXXOMFORT_COMPILER_ID == CXXO_COMPILERID_MSC)
		#pragma warning (default:4996) // MSVC: function may be unsafe
	#endif


	#if (defined(CXXOFLAG_NO_STD_USING) && CXXOFLAG_NO_STD_USING>0 )
	#else

	#endif // system_error

#else // CXXOMFORT_CXX_STD < 1997

	#include <cerrno>
	#include <stdexcept>
	#include <string>
	#include <cstring> // strerror
	#include <cstdlib>
	#include <utility> // make_pair, operator<
	#include "impl/errc.hpp"
	
namespace cxxomfort { namespace cxxostd {

class error_code;
class error_condition;
class error_category;

class error_category {

    public:

    error_category () {};
    CXXO_DELETED_COPY_CONSTRUCTOR(error_category);
    virtual ~error_category () {};
    CXXO_DELETED_COPY_ASSIGN(error_category);
    
    //! Deriveds implement a naming schema for the category here.
    //virtual const char* name () const CXXO_NOEXCEPTNOTHROW = 0;
    //! Deriveds provide a descriptive message about the error here.
    //virtual std::string message (int condition ) const = 0;
    
    //virtual error_condition 
    //default_error_condition (int code ) const CXXO_NOEXCEPTNOTHROW;

    //virtual bool 
    //equivalent (int code, error_condition const& condition ) const CXXO_NOEXCEPTNOTHROW;
    
    //virtual bool 
    //equivalent (error_code const& ec, int condition) const CXXO_NOEXCEPTNOTHROW;

    friend bool operator== (error_category const& vl, error_category const& vr) CXXO_NOEXCEPT {
        return &vl == &vr;
    }
    friend bool operator< (error_category const& vl, error_category const& vr) CXXO_NOEXCEPT {
        return &vl < & vr;
    }

    private:

};

struct generic_category_t : public error_category {
    ~generic_category_t () {}

    const char* name () const CXXO_NOEXCEPTNOTHROW {
        return "generic";
    }
    std::string message (int condition) const {
        return std::string(name()) + ": " + std::strerror(condition);
    }
};

static inline error_category const& generic_category() CXXO_NOEXCEPTNOTHROW {
    static generic_category_t g;
    return g;
}


} } // cxxomfort::cxxostd

namespace std {
	using ::cxxomfort::cxxostd::errc;
	using ::cxxomfort::cxxostd::error_code;
	using ::cxxomfort::cxxostd::error_condition;
	using ::cxxomfort::cxxostd::error_category;
	
	using ::cxxomfort::cxxostd::generic_category;
	
}


#endif // CXXOMFORT_CXX_STD

//
// helper function
//

namespace cxxomfort { namespace fix {

	#ifdef DOXYGEN_DOC
	//!  @xrefitem cxxomfort-fix "errc" "" Correct mapping for <tt>std::errc</tt> in MSVC 2010~2013.
	enum class errc {};
	#elif (CXXOMFORT_CXX_STD >= 2011)
using std::errc;
	#elif (CXXO_COMPILERREQ(CXXO_COMPILERID_MSC,>=1600) && CXXO_COMPILERREQ(CXXO_COMPILERID_MSC, <1800))
using ::std::errc::errc;
	#else
using std::errc;
	#endif


/**
 * @brief Creates a value of type @c errc given a numeric error code.
 * @ingroup cxxomfort-fix
 * 
 * */
#if (CXXOMFORT_IMPLEMENTS_header_system_error>=CXXO_IMPLSTATUS_NATIVE() )
static inline errc to_errc (int E) CXXO_NOEXCEPT { return static_cast<errc>(E);
}
#elif (CXXOMFORT_IMPLEMENTS_header_system_error>=CXXO_IMPLSTATUS_BACKPORT() || CXXOMFORT_CXX_STD>=1997)
CXXO_CONSTEXPR 
static inline errc to_errc (int E) {
    return static_cast<errc::type>(E);
}
#else
static errc to_errc (int E) { return static_cast<errc>(E); }
#endif

} } // cxxomfort::fix


// 
// std::hash<std::errc>
// 


#if defined(DOXYGEN_DOC)
namespace std {
	//! cxx11b Enumeration for error codes.
	//! @xrefitem std0system_error "" "" Enumeration for error codees.
	//! @see @cppref{error/errc}
	enum class errc {};

	/**
	 * @cxx11b Base tag for categories of error codes.
	 * @xrefitem std0system_error "" "" Base tag for categories of error codes.
	 * @see @cppref{error/error_category}
	 * */
	class error_category {};
	class generic_category_t : error_category {};
	class system_category_t : error_category {};
}

#endif

#endif

/**
 * @page std0system_error <system_error>
 * @brief Backports related to Standard Header <tt><system_error></tt>
 * 
 * @code
 * #include <cxxomfort/system_error.hpp>
 * @endcode
 * 
 * (or automatically by including <tt><cxxomfort/backports.hpp></tt> / <tt><cxxomfort/cxxomfort.hpp></tt>)
 * 
 * In new C++, <tt><system_error></tt> provides general facilities for dealing with error codes and categorization of error types and error returns (eg.: @c errno ) using a schema alternative and complimentary to eg.: exceptions. 
 * 
 * @section Interfaces 
 * 
 * Interfaces defined in this file:
 * 
 * * @link cxxomfort::cxxostd::errc @c errc @endlink - enumerator type for error codes, basically a typesafe version of @c errno .
 * * @c error_category() , @c generic_category() , @c system_category() - marker functions that reference the category object that an error is being implemented in.
 * * @link cxxomfort::cxxostd::error_condition @c error_condition @endlink - a sort of platform-agnostic error code.
 * * @c @ref cxxomfort::cxxostd::error_code platform-dependent "error_code" .
 * * @c make_error_code()
 * * POSIX error codes that might be missing in some compilers (eg.: older versions of MSVC).
 * 
 * Non-backport interfaces (<tt>cxxomfort::fix</tt>):
 * 
 * * @link cxxomfort::fix::errc @c errc @endlink - Alias to @c errc::errc in MSVC 2010, normal @c std::errc otherwise.
 * * @c to_errc() - Create an @c errc value from a native integer, a sort of safe type cast.
 * 
 * @section errc errc codes
 * 
 * The listing of error codes that cxxomfort supports is compiler-specific and mostly depends on the operating system and C++ Standard conformance level. In general, when in >=C++11 mode the listing present at https://en.cppreference.com/w/cpp/error/errc is supporte.
 * 
 * In pre-C++11 mode (or in eg.: MSVC on Windows), a subset of those names are detected automatically and the extension @c cxxomfort::fix::errc provides access to some other names not normally provided by the compiler.
 * 
 * 
 * @section system_error_client_code Client Code
 * 
 * For sections of cxxomfort making use of these facilities 
 * (even in C++03 mode), see: @link cxxo-sup-iterator @endlink , <tt>at_or</tt>, 
 * @ref fixed_vector , @ref cxxo_array_ref .
 * 
 * @section See Also
 * 
 * * @cppref{header/system_error}
 * * @cppref{error/errc} (definition and values of the enumeration type)
 * 
 * */
