#ifndef CXXOMFORT_UTIL_ARRAY_HPP
#define CXXOMFORT_UTIL_ARRAY_HPP
/**
 * @file
 * @brief Includes cxxomfort's @c array implementation.
 * @ingroup cxxomfort-utils
 * */
#include <cxxomfort/base.hpp>
#include <cxxomfort/config.hpp>
#include <cxxomfort/utils/traits_basic.hpp>

namespace cxxomfort { 

/**
 * @class array
 * @brief cxxomfort's implementation of @c std::array .
 * @ingroup cxxomfort-utils
 * @cxxoutils Implementation of @c std::array
 * 
 * Cxxomfort exposes its own implementation of @c std::array so it can be used in some contexts where @c std::array can't, for example when array is not available (eg.: compilers without TR1), or when @c std::array does not expose @c constexpr access to its members (such as in some C++14 mode, C++17 mode implementations).
 * 
 * For general notes see @ref @c cxxomfort::cxxostd::array .
 * 
 * */
template <typename T, size_t N>
struct array {
    static_assert (N , "array<T,N> with N>0");
    typedef T value_type;
    typedef T const           const_value_type;
    enum{ const_size = N};

    typedef size_t            size_type;
    typedef ptrdiff_t         difference_type;

    typedef typename ::cxxomfort::traits::add_pointer<T>::type    pointer;
    typedef typename ::cxxomfort::traits::add_pointer<T const>::type const_pointer;
    
    typedef T&           reference;
    typedef T const&     const_reference;
    
    typedef pointer           iterator;
    typedef const_pointer     const_iterator;
    typedef std::reverse_iterator<iterator> reverse_iterator;
    typedef std::reverse_iterator<const_iterator> const_reverse_iterator;
    
    
    //! Access to the underlying array
    CXXO_CONSTEXPR14
    pointer         data () CXXO_NOEXCEPTNOTHROW { return elems; }
    //! Access to the underlying array
    CXXO_CONSTEXPR 
    const_pointer   data () const CXXO_NOEXCEPTNOTHROW { return elems; }
    //! Size of the array (the same as the constant @a N )
    CXXO_CONSTEXPR 
    size_t     size () const CXXO_NOEXCEPTNOTHROW { return N; }
    //! Maximum size of the array (the same as @e size )
    CXXO_CONSTEXPR 
    size_t     max_size () const CXXO_NOEXCEPTNOTHROW { return N; }
    //! Returns @c true if the array is empty 
    CXXO_CONSTEXPR 
    bool       empty () const CXXO_NOEXCEPTNOTHROW { return N==0; }

	//! Direct access to the first element
    reference       front () CXXO_NOEXCEPTNOTHROW { return elems[0]; }
    CXXO_CONSTEXPR14 
    const_reference front () const CXXO_NOEXCEPTNOTHROW { return elems[0]; }
    //! Direct access to the last element
    reference       back () CXXO_NOEXCEPTNOTHROW { return elems[N-1]; }
    CXXO_CONSTEXPR14 
    const_reference back () const CXXO_NOEXCEPTNOTHROW { return elems[N-1]; }
    
	//! Indexed access (non-throwing)
    CXXO_CONSTEXPR14 
    reference       operator[] (size_t i) CXXO_NOEXCEPTNOTHROW { return elems[i]; }
    CXXO_CONSTEXPR 
    const_reference operator[] (size_t i) const CXXO_NOEXCEPTNOTHROW { return elems[i]; }
    
    //! Indexed access (throwing)
    reference       at (size_t i) { 
		return (i < size()) ? elems[i]  : throw std::out_of_range("array");
	}

    CXXO_CONSTEXPR14 
    const_reference at (size_t i) const { 
		if (i < size()) return elems[i];
		else throw std::out_of_range("array const");
	}
    
    iterator        begin () CXXO_NOEXCEPT { return elems; }
    CXXO_CONSTEXPR 
    const_iterator  begin () const CXXO_NOEXCEPT { return elems; }
    CXXO_CONSTEXPR 
    const_iterator  cbegin () const CXXO_NOEXCEPT { return elems; }
    iterator        end () CXXO_NOEXCEPT { return elems+N; }
    CXXO_CONSTEXPR 
    const_iterator  end () const CXXO_NOEXCEPT { return elems+N; }
    CXXO_CONSTEXPR 
    const_iterator  cend () const CXXO_NOEXCEPT { return elems+N; }
    
    reverse_iterator     rbegin() CXXO_NOEXCEPT { return reverse_iterator(elems+N); }
    CXXO_CONSTEXPR14 
	const_reverse_iterator     rbegin() const CXXO_NOEXCEPT { 
        return const_reverse_iterator(elems+N); 
    }
    reverse_iterator     rend() CXXO_NOEXCEPT { return reverse_iterator(elems); }
    CXXO_CONSTEXPR14 
    const_reverse_iterator     rend() const CXXO_NOEXCEPT { 
        return const_reverse_iterator(elems); 
    }
    
    CXXO_CONSTEXPR14 
    void fill (T const& t) {
        for (size_t i=0; i < N; ++i, elems[i]=t) {}
    }
    
    CXXO_CONSTEXPR14 
    friend bool operator== (array const& a1, array const& a2) {
        //return std::equal(a1.begin(), a1.end(), a2.begin());
        for (size_t i= 0; i < N; ++i) { if (a1[i]!=a2[i]) return false; }
        return true;
    }

    friend bool operator< (array const& a1, array const& a2) {
        return std::lexicographical_compare(a1.begin(), a1.end(), a2.begin(), a2.end());
    }

    public:
    T elems[N];
};

}
#endif
