#ifndef CXXOMFORT_UTILS_ARRAYREF_HPP
#define CXXOMFORT_UTILS_ARRAYREF_HPP

/**
 * @file
 * @brief Includes cxxomfort's @c array_ref implementation and related utilities.
 * @ingroup cxxomfort-utils
 * */

#include <cxxomfort/base.hpp>
#include <cxxomfort/cxxo-utils.hpp>
#include <iterator>
#include <cxxomfort/cerrno.hpp>
#include <cxxomfort/array.hpp> // maybe do away with this
#include <cxxomfort/utils/traits_basic.hpp>
#include <cxxomfort/impl/metaprogramming.hpp>
#include <cxxomfort/impl/lift_type_traits.hpp>

#define CXXOMFORT_IMPLEMENTS_n3334 CXXO_IMPLSTATUS_LIBRARY()
#if !defined(CXXOMFORT_IMPLEMENTS_array_ref)
	#define CXXOMFORT_IMPLEMENTS_array_ref CXXOMFORT_IMPLEMENTS_n3334
#endif

//! feature-like macro for support info
#define CXXO_impl_array_ref 202210UL

/*
 * 2022xx - substr, slice, first and last alts with std::expected return value
 * 2022xx - synonymial method names for future std::span compatibility
 * 2022xx - implements slice with negative values
 * 202207 - macro introduction
 * 202210 - constexpr operations (1)
 * */

namespace cxxomfort {

/**
 * @brief A non-owning view to sequences of contiguous data of type @a T .
 * @tparam T Type of the values refered to by the view.
 * @tparam IdxType A numeric type to be used as the index, defaults to @c size_t .
 * @ingroup cxxomfort-utils 
 * @cxxoutils A non-owning view to sequences of contiguous data of type @a T .
 * 
 * A @c array_ref is a non-owning view of a sequence of contiguous elements via a simple {pointer+length} package. It is the combination of <tt>std::array</tt>'s interface and  <tt>std::reference_wrapper</tt>'s workings.
 * 
 * The utility functions @c arrayref() and @c carrayref() create array_refs from contexts where they can be needed, such as from C arrays or @c std::array .
 * 
 * See also: <tt>std::array</tt>, <tt>std::vector</tt>, @c seq_() .
 * 
 * @section n3334_differences Differences from the original proposal:
 * 
 * This implementation differs from the original proposal by implementing some changes suggested in the paper, as well as other changes from C++'s evolution.
 * 
 * * Allows for a @c array_ref of non-const objects, as suggested by the paprt.
 * * Allows for an index type shorter than @c size_t (for eg.: short alloc buffers or near/short memory models).
 * * Adds ( @c pointer , @c pointer ) constructor.
 * * Discards @c vector constructor for lower dependency overhead.
 * * Adds member functions @c front , @c back .
 * * Adds member functions @c leftmost , @c rightmost .
 * * Helper functions @c arrayref() and @c carrayref() provided as support.
 * * Relaxed @c constexpr support.
 * 
 * Other that @c .at() and the template conversion members, no member functions in @c array_ref throw exceptions. Failure indications eg.: slice an inexistent slice of an @c array_ref result in an empty @c array_ref instead.  @c .at() throws <tt>siesta::status_error</tt> with value @c invalid_index . The member @c .at_or() is provided as a checked but non-throwing alternative.
 * 
 * @section n3334_also See Also:
 * 
 * @c array_ref is based in the original publication "n3334" by Jeffrey Yasskin
 * 
 * @see @wg21{n3334,array_ref} @cppref{container/span}
 * 
 */
#pragma pack(push)
#pragma pack(2)
template <class T, class IdxType = size_t>
class array_ref {
    public:
    //! Type of the data contained in the view.
    typedef T       element_type;
    //! Type of variables of the type of the pointed-to data
    typedef typename std::remove_cv<element_type>::type value_type;
    //! @c const version of the value_type .
    typedef typename std::add_const<value_type>::type const_value_type;
    //! Type for the distance between pointers or iterators.
    typedef ptrdiff_t    difference_type;
    //! Type for subscript / indexed access (eg.: <code>operator[]</code>).
    typedef IdxType      size_type;
    
    // preconditions
    static_assert0(::cxxomfort::traits::is_unsigned<size_type>::value);

    //! Type of a pointer to a contained value; potentially writable.
    typedef typename std::add_pointer<element_type>::type 	pointer;
    //! Type of a pointer to a contained value; view-only.
    typedef typename std::add_pointer<element_type const>::type 	const_pointer;
    typedef typename ::cxxomfort::traits::add_lvalue_reference<element_type>::type 	reference;
    typedef typename ::cxxomfort::traits::add_lvalue_reference<element_type const>::type 	const_reference;
    
    //! Value that indicates an invalid index for @c size_type .
    static size_type const npos = static_cast<size_type>(-1);

    //! Iterator type; potentially writable.
    typedef pointer     iterator;
    //! Iterator type; view-only.
    typedef const_pointer const_iterator;
    //! Reverse iterator type; potentially writable.
    typedef std::reverse_iterator<iterator> reverse_iterator;
    //! Reverse iterator type; view-only.
    typedef std::reverse_iterator<const_iterator>  const_reverse_iterator;
    
    public:

    // 
    // basic constructors
    // 

#ifdef DOXYGEN_DOC
    //! Empty view (trivial constructor).
    constexpr array_ref () noexcept = default;
    //! Empty view.
    constexpr array_ref (cxxomfort::None_t) noexcept = default;
    //! Copy constructor (trivial)
    constexpr array_ref (array_ref const&) noexcept = default;
    //! Assignment operator (trivial)
    constexpr array_ref& operator= (array_ref const&) noexcept = default;
#endif
#if (CXXOMFORT_CXX_STD>=2011 || (defined(__cpp_rvalue_references) && CXXO_COMPILER_SUPPORT_default_delete>0 ))
    CXXO_CONSTEXPR array_ref () noexcept = default;
    CXXO_CONSTEXPR array_ref (array_ref const&) noexcept = default;
#else // pre-c++11 constructors
    CXXO_CONSTEXPR 	array_ref () CXXO_NOEXCEPTNOTHROW 
    : ptr_(nullptr), len_(0) 
    {}
    CXXO_CONSTEXPR 	array_ref (array_ref const& a) CXXO_NOEXCEPTNOTHROW 
    : ptr_(a.ptr_), len_(a.len_) {}
#endif
    
    // 
    // special constructors 
    // 
    //! Empty view.
    CXXO_CONSTEXPR 
    array_ref (None_t const) CXXO_NOEXCEPTNOTHROW 
    : ptr_(nullptr), len_(0) {}
    
    //! Construct a view to a buffer starting at @a p and of @a len units long.
    CXXO_CONSTEXPR 
    array_ref (pointer p, size_type len) CXXO_NOEXCEPTNOTHROW 
    : ptr_(p), len_(len) {}

    //! Construct a view to a buffer starting at @a p and ending at @a q (not in the original n3334 proposal).
    CXXO_CONSTEXPR 
    array_ref (pointer p, pointer q) CXXO_NOEXCEPTNOTHROW 
    : ptr_(p), len_(static_cast<size_type>(q-p)) {}

    //! Constructs a view over a native C array.
    template <size_t N>
    CXXO_CONSTEXPR 
    array_ref (T(&arr)[N]) CXXO_NOEXCEPTNOTHROW 
    : ptr_(arr), len_(N) {}
    
    // view from related non-const view, only for const view, for C++11-onwards
#if (CXXOMFORT_CXX_STD >= 2011)
    //! Constructs a const view over a non-const one.
    template <bool B = std::is_const<T>::value, typename std::enable_if<B,int>::type=0 > CXXO_CONSTEXPR14
    array_ref (array_ref< typename std::remove_const<T>::type > const& av) CXXO_NOEXCEPT 
    : ptr_(av.ptr_), len_(av.len_) {}
#endif

#if (CXXOMFORT_CXX_STD >= 2011)
	//! View over a <tt>std::array</tt>.
    template <size_type N>
    CXXO_CONSTEXPR 
    array_ref (std::array<T,N>& arr) CXXO_NOEXCEPTNOTHROW
    : ptr_(&arr[0]), len_(N) {}

    //! View over a <tt>std::array</tt>.
    template <size_type N> CXXO_CONSTEXPR 
    array_ref (std::array<T,N> const& arr) CXXO_NOEXCEPTNOTHROW
    // we can't invoke .data() because it's non-constexpr
    : ptr_(&arr[0]), len_(N) { }

	//! Generalized range constructor for C++11 onwards (lvalue-only for C++<11)
	#if (CXXOMFORT_CXX_STD >= 2011)
	template <typename R> CXXO_CONSTEXPR14 
	array_ref (R&& r) noexcept 
	#else
	template <typename R>
	array_ref (R& r) CXXO_NOEXCEPTNOTHROW 
	#endif
	: ptr_(r.data()), len_(r.size()) { }

    // 
    // assignment operator
    // 

#if (CXXOMFORT_CXX_STD>=2011 || (CXXO_COMPILER_SUPPORT_noexcept>0 && CXXO_COMPILER_SUPPORT_default_delete>0 ) )
    CXXO_CONSTEXPR14 
    array_ref& operator= (array_ref const&) noexcept = default;
#else // pre-c++11 constructors
	// assignment operators automatically generated
#endif

	// 
	// accessors 
	// 

    //! Unchecked access to the @a i -th element.
    reference       operator[] (size_type i) CXXO_NOEXCEPTNOTHROW { return ptr_[i]; }
    //! Unchecked access to the @a i -th element.
    CXXO_CONSTEXPR 
    const_reference operator[] (size_type i) const CXXO_NOEXCEPTNOTHROW { return ptr_[i]; }

    //! Checked access to the @a i -th element; throws @c siesta_code(invalid_index) .
    reference       at (size_type i) { check_range(i); return ptr_[i]; }
    //! Checked access to the @a i -th element; throws @c siesta_code(invalid_index) .
    CXXO_CONSTEXPR14 
    const_reference at (size_type i) const { check_range(i); return ptr_[i]; }
    
    //! Checked nonthrowing access to the @a i -th element; returns an @a alternative if it doesn't exist.
    reference       at_or (size_type i, reference alt) CXXO_NOEXCEPTNOTHROW { return i < size() ? ptr_[i] : alt; }
    //! Checked nonthrowing access to the @a i -th element; returns an @a alternative if it doesn't exist.
    CXXO_CONSTEXPR 
    const_reference at_or (size_type i, const_reference alt) const CXXO_NOEXCEPTNOTHROW { return i < size() ? ptr_[i] : alt; }


    // view info
    
    CXXO_CONSTEXPR  size_type     max_size () const CXXO_NOEXCEPTNOTHROW { return len_; }
    //! Returns size of the view as a @c size_t .
    CXXO_CONSTEXPR  size_type     size () const CXXO_NOEXCEPTNOTHROW { return len_; }
    //! Returns @c true if the view is empty (has zero elements) .
    CXXO_CONSTEXPR  bool       empty () const CXXO_NOEXCEPTNOTHROW { return size() == 0; }
    //! Returns a pointer to the view's data.
    pointer         data () CXXO_NOEXCEPTNOTHROW { return ptr_; }
    //! @return pointer to the view's data.
    CXXO_CONSTEXPR 
    const_pointer   data () const CXXO_NOEXCEPTNOTHROW { return ptr_; }
    //! @return size of the view in bytes.
    CXXO_CONSTEXPR  size_t  	size_bytes () const CXXO_NOEXCEPTNOTHROW { return static_cast<size_t>(len_)*sizeof(T); }
    
    // iterators

    //! begin
    CXXO_CONSTEXPR
    const_iterator  cbegin () const CXXO_NOEXCEPTNOTHROW { return ptr_; }
    //! end
    CXXO_CONSTEXPR
    const_iterator  cend () const CXXO_NOEXCEPTNOTHROW { return ptr_+len_; }
    //! begin
    CXXO_CONSTEXPR 
    iterator        begin () CXXO_NOEXCEPTNOTHROW { return ptr_; }
    //! end
    CXXO_CONSTEXPR 
    iterator        end () CXXO_NOEXCEPTNOTHROW { return ptr_+len_; }
    //! begin
    CXXO_CONSTEXPR 
    const_iterator  begin () const CXXO_NOEXCEPTNOTHROW { return cbegin(); }
    //! end
    CXXO_CONSTEXPR 
    const_iterator  end () const CXXO_NOEXCEPTNOTHROW { return cend(); }
    
    //! reverse-begin
    CXXO_CONSTEXPR 
    const_reverse_iterator    crbegin () const CXXO_NOEXCEPTNOTHROW { return const_reverse_iterator(ptr_+len_); }
    //! reverse-end
    CXXO_CONSTEXPR 
    const_reverse_iterator    crend () const CXXO_NOEXCEPTNOTHROW { return const_reverse_iterator(ptr_); }
    //! reverse-begin
    reverse_iterator          rbegin () CXXO_NOEXCEPTNOTHROW { return ptr_+len_; }
    //! reverse-end
    reverse_iterator          rend () CXXO_NOEXCEPTNOTHROW { return ptr_; }
    //! reverse-begin
    CXXO_CONSTEXPR 
    const_reverse_iterator    rbegin () const CXXO_NOEXCEPTNOTHROW { return crbegin(); }
    //! reverse-end
    CXXO_CONSTEXPR 
    const_reverse_iterator    rend () const CXXO_NOEXCEPTNOTHROW { return crend(); }

	//! Unchecked access to the 0th element.
    CXXO_CONSTEXPR14 
    const_iterator  front () const CXXO_NOEXCEPTNOTHROW { return ptr_[0]; }
	//! Unchecked access to the 0th element.
    iterator        front () CXXO_NOEXCEPTNOTHROW { return ptr_[0]; }
    //! Unchecked access to the last element.
    CXXO_CONSTEXPR14 
    const_iterator  back () const CXXO_NOEXCEPTNOTHROW { return ptr_[len_-1]; }
    //! Unchecked access to the last element.
    iterator  		back () CXXO_NOEXCEPTNOTHROW { return ptr_[len_-1]; }

    /**
     * @return A view to the @a count -elements subsequence starting from @a pos .
     * @pre Position @a pos falls within the view; @a count is equal or lower than the size of the view.
     * @note If indexes <tt>pos</tt> or <tt>pos+count</tt> go beyond the end of the view, returns an empty view.
     * @see A "signed" version with PHP-like semantics is available as @c slice() .
     * */
    CXXO_CONSTEXPR 
    array_ref       substr (size_type pos, size_type count) const CXXO_NOEXCEPTNOTHROW {
		return (!this->empty() && (pos<len_) && (pos+count<len_))
		? array_ref(ptr_+pos, (count < len_-pos ? count : len_-pos))
		: array_ref()
		;
	}
    /**
     * @overload substr
     * */
    CXXO_CONSTEXPR 
    array_ref       substr (size_type pos) const CXXO_NOEXCEPTNOTHROW {
		return (pos < len_) ? array_ref(ptr_+pos, len_-pos) : array_ref();
	}
    //! Returns a view to the leftmost (first) @a n elements.
    CXXO_CONSTEXPR 
    array_ref       leftmost (size_type n) const CXXO_NOEXCEPTNOTHROW {
		return substr(0, n);
	}
    //! Returns a view to the rightmost (last) @a n elements.
    CXXO_CONSTEXPR 
    array_ref       rightmost (size_type n) const CXXO_NOEXCEPTNOTHROW {
		return substr(size()-n, n);
	}
    /**
     * @return A view of @a count elements starting at index @a pos of the current view. 
     * @note If <tt>pos<0</tt>, the starting position is taken from the @e end of the view. 
     * @see @c substr() .
     * */
    CXXO_CONSTEXPR 
    array_ref       slice (difference_type pos, size_type count) const CXXO_NOEXCEPTNOTHROW {
		return (pos < 0) ? this->slice(size()+pos, count) 
		: this->substr(pos,count);
	}
    //! @overload slice 
    CXXO_CONSTEXPR 
    array_ref       slice (difference_type pos) const CXXO_NOEXCEPTNOTHROW {
		return pos >= static_cast<difference_type>(len_) ? array_ref() : array_ref(ptr_+pos, len_-pos);
	}
    
    //! Removes the last @a n elements.
    CXXO_CONSTEXPR14 
    array_ref  without_back_n (size_type n) const CXXO_NOEXCEPTNOTHROW {
        size_type newlen= (len_>n) ? len_-n : len_;
        return array_ref(ptr_, newlen);
    }

    //! Removes the first @a n elements.
    CXXO_CONSTEXPR14 
    array_ref  without_front_n (size_type n) const CXXO_NOEXCEPTNOTHROW  {
		pointer const newbegin = (len_>n) ? ptr_+n : ptr_;
		size_type const newlen= (len_>n) ? len_-n : len_;
		return array_ref(newbegin,newlen);
	}

    //! Returns a view with the last element removed.
    CXXO_CONSTEXPR14 
    array_ref  without_back () const CXXO_NOEXCEPTNOTHROW {
        return len_ > 0 ? array_ref(ptr_, len_ -1) : array_ref();
    }

    //! Returns a view with the first element removed.
    CXXO_CONSTEXPR14 
    array_ref  without_front () const CXXO_NOEXCEPTNOTHROW {
        return len_ > 0 ? array_ref(ptr_+1, len_ -1) : array_ref();
    }

#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD>=1997)
	//! Obtains a copy of the elements of the view made as a container of type @a U .
    template <typename U>
    U          copy_as () const {
        return U(data(), size());
    }
    
    //! Copies the contents of the view to the writable iterator @a to .
    template <typename DestIter>
    DestIter  copy_to (DestIter to) const {
        using namespace std;
        return copy_n(begin(), size(), to);
    }
#endif    


    private:
    CXXO_CONSTEXPR14 
    void check_range (size_type i) const {
        if (i>=len_) CXXO_THROW_STD( EINVAL ); //siesta_code(status_messages::invalid_index).throw_exception(); // ::cxxomfort::general_error::out_of_bounds;
    }
    pointer    ptr_ ;
    size_type  len_ CXXOMFORT_CXX11_CODE(= 0, ;);

}; // array_ref
#pragma pack(pop)

//! Compare <tt>array_ref</tt>s for equality.
template <typename T>
CXXO_CONSTEXPR 
bool operator== (array_ref<T> const& a1, array_ref<T> const& a2) CXXO_NOEXCEPTNOTHROW {
	using namespace cxxomfort::fix;
	return equal(cxxomfort::constexpr_tag, a1.begin(), a1.end(), a2.begin(), a2.end());
}

//! Compare <tt>array_ref</tt>s for equality.
template <typename T>
CXXO_CONSTEXPR 
bool operator!= (array_ref<T> const& a1, array_ref<T> const& a2) CXXO_NOEXCEPTNOTHROW {
	return !(a1==a2);
}

// 
// factories 
// 

/**
 * @brief Create a (potentially writeable) @c arrayref view of a buffer with size.
 * @ingroup cxxomfort-utils
 * @return A view of data of type @a Ty .
 * */
template <typename Ty, typename Sz>
CXXO_CONSTEXPR14 
array_ref<Ty,Sz> arrayref (Ty* p, Sz l) {
    return array_ref<Ty,Sz>(p,l);
}

//! @overload arrayref
template <typename Ty>
CXXO_CONSTEXPR14 
array_ref<Ty> arrayref (Ty* p, Ty* q) {
	return array_ref<Ty>(p,q);
}

//! @overload arrayref
template <typename Ty, size_t N>
CXXO_CONSTEXPR14 
array_ref<Ty> arrayref (Ty(&p)[N]) {
    return array_ref<Ty>(p,N);
}

template <typename Ty, size_t N>
CXXO_CONSTEXPR14 
array_ref<Ty> arrayref (std::array<Ty,N>& a) {
    return array_ref<Ty>(a.data(),N);
}

/**
 * @brief Create a const @c arrayref view of a buffer with size.
 * @ingroup cxxomfort-utils
 * @return A view of @c const data of type @a Ty .
 * */
template <typename Ty, typename Sz>
CXXO_CONSTEXPR 
array_ref<Ty const, Sz> carrayref (Ty* p, Sz l) {
    return array_ref<Ty const, Sz>(p,l);
}

//! @overload carrayref
template <typename Ty>
CXXO_CONSTEXPR 
array_ref<Ty const> carrayref (Ty* p, Ty* q) {
    return array_ref<Ty const>(p,q);
}

//! @overload carrayref
template <typename Ty, size_t N>
CXXO_CONSTEXPR 
array_ref<Ty const> carrayref (Ty(&p)[N]) {
    return array_ref<Ty const>(p,N);
}

template <typename Ty, size_t N>
CXXO_CONSTEXPR 
array_ref<Ty const> 
carrayref (std::array<Ty const,N> const& a) {
    return array_ref<Ty const>(a.begin(),N);
}

template <typename Ty, size_t N>
CXXO_CONSTEXPR 
array_ref<Ty const> 
carrayref (std::array<Ty,N> const& a) {
    return array_ref<Ty const>(a.data(),N);
}

//! @overload carrayref
template <typename Ty>
CXXO_CONSTEXPR 
array_ref<Ty const> 
carrayref (array_ref<Ty> const& a) {
    return array_ref<Ty const>(a.data(), a.size());
}

// 
// traits 
// 

/**
 * @brief Type trait that identifies types of @c array_ref .
 * @ingroup cxxo-utils
 * */
template <typename> struct is_array_ref: std::false_type {};
template <typename T> struct is_array_ref< array_ref<T> >
: std::true_type {};

} // cxxomfort

#endif // file
