#ifndef CXXOMFORT_UTIL_ATOR_HPP
#define CXXOMFORT_UTIL_ATOR_HPP

#include <cxxomfort/config.hpp>
#include <iterator>

/**
 * @file
 * @brief Provides the global accessor @c at_or() and @c atindex_or() .
 * @ingroup cxxomfort-utils
 * */


namespace cxxomfort {

// 
// at_or, index_or 
// 

/**
 * @ingroup cxxomfort-utils 
 * @{
 * */


/**
 * @brief Accesses the @a ind -th element of @a cont if it exists, else @a alt .
 * @param cont A container or container-like object supporting member @c .at() .
 * @param ind An index.
 * @param alt An alternative return value if a suitable element in @a cont is not found.
 * @return ref A reference to the requested element in @a cont , or the reference @a alt if such element doesn't exist.
 * @throws noexcept Never throws.
 * 
 * This function encapsulates the idiom of accessing the i-th element of C, or a suitable default A if no i-th element is found. Such an idiom already exists for other components in the form of eg.: variant/optional <tt>get_if<I>()</tt>.
 * 
 * @c at_or() is marked @c noexcept . It is intended to never throw, as it checks the precondition to access the element in @c C 's <tt>.at()</tt> before accessing. No practical implementation of <tt>.size()</tt> or relational operators could be thought that throws, so this is safe to do.
 * 
 * 
 * */
template <typename C>
typename C::const_reference 
at_or (C const& cont, typename C::size_type ind, typename C::const_reference alt) CXXO_NOEXCEPTNOTHROW {
	using namespace std;
	return (ind >= 0 && size(cont) > ind) ? cont.at(ind) : alt;
}

//! @overloads at_or
template <typename C>
typename C::reference  
at_or (C& cont, typename C::size_type ind, typename C::reference alt) CXXO_NOEXCEPTNOTHROW {
	using namespace std;
	return (ind >= 0 && size(cont) > ind) ? cont.at(ind) : alt;
}

/**
 * @brief Accesses the @a ind -th element of @a cont if it exists, else @a alt .
 * @param cont A container or container-like object supporting member @c .at() .
 * @param ind An index.
 * @param alt An alternative return value if a suitable element in @a cont is not found.
 * @return ref A reference to the requested element in @a cont , or the reference @a alt if such element doesn't exist.
 * @throws noexcept Never throws.
 * 
 * This function encapsulates the idiom of accessing the i-th element of C, or a suitable default A if no i-th element is found. Such an idiom already exists for other components in the form of eg.: variant/optional <tt>get_if<I>()</tt>.
 * 
 * @c atindex_or() is similar to @c at_or() in all respects but one, including being marked @c noexcept ; the one difference being that access is performed via member <tt>.operator[]()</tt>; as such, if that access does not exist for @a cont , compilation fails.
 * 
 * 
 * */
template <typename C>
typename C::value_type const& 
atindex_or (C const& cont, typename C::size_type ind, typename C::const_reference alt) CXXO_NOEXCEPTNOTHROW {
	using namespace std;
	return (ind >= 0 && size(cont) > ind) ? cont[ind] : alt;
}

/**
 * @}
 * */

} // cxxomfort

#endif
