#ifndef CXXOMFORT_UTILS_SIESTA_STATUSCODE_HPP
#define CXXOMFORT_UTILS_SIESTA_STATUSCODE_HPP

namespace cxxomfort { namespace siesta {

struct domain_base; 

class _generic_code_domain;

template <typename Domain>
class status_code;

template <typename Domain>
class status_error;

//! Generic status code, those mapped by eg.: @c errc .
typedef status_code<_generic_code_domain> generic_code;

// this is p1028's  "status_code_domain"
struct domain_base {
	typedef uint64_t unique_id_type;
	template <class DomainType> friend class status_code;
	protected:
	CXXO_CONSTEXPR 
	explicit domain_base (unique_id_type id) CXXO_NOEXCEPTNOTHROW 
	: _id(id) {}
	
	#if (CXXOMFORT_CXX_STD>=2011)
	domain_base& operator= (domain_base const&) = default;
	#endif
	~domain_base () CXXOMFORT_CXX11_CODE(=default, {});
	
	CXXO_CONSTEXPR 
	unique_id_type 	id () const CXXO_NOEXCEPT { return _id; }
	//! Returns a pointer to a string literal representing the name. Deriveds must implement.
	virtual char const* 	name () const CXXO_NOEXCEPT = 0;
	static domain_base const* 	get_p (domain_base const* p) { return p; }

	//! @c true if the code represents a failure code. Derived classes must implement.
	virtual 
	bool 	_do_failure (status_code<void> const& c) const CXXO_NOEXCEPT = 0;
	//! @c true if code is (non-transitively?) equivalent to code in another domain
	//virtual 
	//bool 	_do_equivalent (status_code<void> const& code1, status_code<void> const& code2) const CXXO_NOEXCEPT = 0;

	//! @c closest generic code if any, or @c unknown otherwise.
	virtual 
	generic_code 	_generic_code (status_code<void> const&) const CXXO_NOEXCEPT = 0;
	//! write an explanatory message to buffer @a dest, up to @a len characters
	virtual 
	char* 	_do_message_to (char* dest, size_t len, status_code<void> const&) const CXXO_NOEXCEPT = 0;
	
	//! throw a code as a C++ exception.
	virtual 
	void 	_do_throw_exception (status_code<void> const& code) const 
	#if (defined(CXXO_SIESTA_WITH_EXCEPTIONS))
	= 0; // deriveds must implement
	#else
	{ 
		fprintf(stdout, "siesta: exception in no-exception mode\n");
		fflush(stdout);
		abort(); 
	} // from ned14: keep a vtable slot for binary compat
	#endif
	
	public:
	CXXO_CONSTEXPR 
	bool operator== (domain_base const& r) const CXXO_NOEXCEPT { return this->id() == r.id(); }

	public:
	const uint64_t _id;

};

template <uint64_t X> 
struct domain_id_reserve_x 
: ::cxxomfort::traits::integral_constant<uint64_t, (0x3a63786fULL << 32) | X> {};

//template <class T> struct is_status_code : std::false_type {};
//template <> struct is_status_code< status_code > : std::true_type {};


/**
 * @brief Base to quickly define a status domain for a given  @a ECode type.
 * @tparam ECode The type used for the domain's status. That's what the corresponding status' <tt>.value()</tt> member will return.
 * @tparam Id A unique  @c uint64_t that identifies this domain.
 * 
 * @note For the purposes of this library, domain unique ids in the range <tt>0x3a63786fnnnnnnnn (":cxo____") are reserved by the implementation.
 * 
 * */
template <typename ECode, uint64_t Id, typename Tag = void> 
class quick_status_code_domain
: public domain_base {
	public:
	//! The type of the error code
	typedef ECode 	value_type;
	typedef Tag 	tag_type;
	static const uint64_t domain_id = Id;
	static const bool is_reserved = (Id >> 32 == 0x3a63786f);
	using domain_base::unique_id_type;
	//! Returns the unique identifier for this domain.
	//CXXO_CONSTEXPR 
	//uint64_t 	id () const CXXO_NOEXCEPT { return domain_id; }
	//! Converts an @c int to the type of the error code.
	value_type 	code (int x) const CXXO_NOEXCEPT { return static_cast<value_type>(x); }
	//! Takes a status code with this domain and prints an informative message about it to buffer @a p . Derived classes must implement this.
	//virtual char* 	message_to (char* p, size_t l, value_type) const CXXO_NOEXCEPT = 0;
	
	CXXO_CONSTEXPR 
	quick_status_code_domain () CXXO_NOEXCEPT : domain_base(Id) {}
	
	CXXO_CONSTEXPR 
	bool operator== (quick_status_code_domain const&) const CXXO_NOEXCEPT { return true; }
	
};



class _int_code_domain;


/**
 * @brief Status code
 * 
 * */
template <>
class status_code<void> {
	public:
	typedef void 	domain_type;
	typedef void 	value_type;
	public:
	template <class> friend class status_code;
	
	protected:
	status_code () : _domain(nullptr) {}
	#if (CXXOMFORT_CXX_STD >= 2011)
	status_code (status_code const&) = default;
	#else
	status_code (status_code const& s) : _domain(s._domain) {}
	#endif
	#if (CXXOMFORT_CXX_STD >= 2011)
	status_code & operator= (status_code const&) = default;
	#else
	status_code & operator= (status_code const& s) { this->_domain= s._domain; return *this; }
	#endif
	
	~status_code () CXXOMFORT_CXX11_CODE(=default, {});
	
	protected:
	CXXO_CONSTEXPR 
	explicit status_code (domain_base const* ptr) CXXO_NOEXCEPT 
	: _domain(ptr) {}
	
	public:
	domain_base const& 	domain() const CXXO_NOEXCEPT { return *_domain; }
	CXXO_CONSTEXPR14  
	bool 	empty () const CXXO_NOEXCEPT { return _domain == nullptr; }

	public:
	bool 	success () const CXXO_NOEXCEPT {
		return this->empty() ? false : !_domain->_do_failure(*this); 
	}
	bool 	failure () const CXXO_NOEXCEPT {
		return this->empty() ? false : _domain->_do_failure(*this); 
	}
	
	char* 	message_to (char* p, size_t l) const CXXO_NOEXCEPT {
		return this->empty() ? p : _domain->_do_message_to(p,l,*this);
	}
	
	#if (defined(CXXO_SIESTA_WITH_EXCEPTIONS))
	void 	throw_exception () const {
		assert(_domain != nullptr);
		_domain->_do_throw_exception(*this);
		abort();
	}
	#endif

	protected:
	domain_base const* _domain;
};


/*! A tag for an erased value type for `status_code<D>`.
Available only if `ErasedType` satisfies `traits::is_move_bitcopying<ErasedType>::value`.
*/
template <typename ErasedType>
struct erased {
	static_assert(::cxxomfort::traits::is_fundamental<ErasedType>::value || std::is_enum<ErasedType>::value, "");
	typedef ErasedType value_type;
};

namespace {

	template <class DomainType> struct get_domain_value_type
	{
	typedef  DomainType domain_type;
	typedef typename domain_type::value_type value_type;
	};
	template <class ErasedType> struct get_domain_value_type<erased<ErasedType> >
	{
	typedef domain_base domain_type;
	typedef ErasedType value_type;
	};

} // impl 

template <class DomainType> 
class status_code_storage 
: public status_code<void> {
	//static_assert(!std::is_void<DomainType>::value, "status_code_storage<void> should never occur!");
	typedef status_code<void> _Base;
	public:
	//! The type of the domain.
	typedef typename get_domain_value_type<DomainType>::domain_type domain_type;
	//! The type of the status code.
	typedef typename get_domain_value_type<DomainType>::value_type value_type;
#ifndef NDEBUG

    #if (CXXOMFORT_CXX_STD>=2011)
    static_assert(std::is_move_constructible<value_type>::value || std::is_copy_constructible<value_type>::value, "DomainType::value_type is neither move nor copy constructible!");
    static_assert(!std::is_default_constructible<value_type>::value || std::is_nothrow_default_constructible<value_type>::value, "DomainType::value_type is not nothrow default constructible!");
    static_assert(!std::is_move_constructible<value_type>::value || std::is_nothrow_move_constructible<value_type>::value, "DomainType::value_type is not nothrow move constructible!");
    static_assert(std::is_nothrow_destructible<value_type>::value, "DomainType::value_type is not nothrow destructible!");
    #endif
#endif
	// Replace the type erased implementations with type aware implementations for better codegen
	//! Return the status code domain.
	CXXO_CONSTEXPR domain_type const& domain() const CXXO_NOEXCEPT {
		return *static_cast<domain_type const *>(this->_domain);
	}
	//! Reset the code to empty.
	CXXO_CONSTEXPR14 void clear() CXXO_NOEXCEPT {
		this->_value.~value_type();
		this->_domain = nullptr;
		new(&this->_value) value_type();
	}
	//! Return a reference to the `value_type`.
	CXXO_CONSTEXPR value_type const &value() const CXXO_NOEXCEPT {
		return this->_value;
	}
	protected:
	#if (CXXOMFORT_CXX_STD>= 2011)
	status_code_storage () noexcept = default;
	#else
	status_code_storage () : _Base(), _value(0) {}
	#endif
	
    #if (CXXOMFORT_CXX_STD>=2011)
    status_code_storage(const status_code_storage &) = default;
    #else
    #endif
    #if (CXXOMFORT_CXX_STD>=2011)
    CXXO_CONSTEXPR14 status_code_storage(status_code_storage &&o) CXXO_NOEXCEPT
        : _Base(static_cast<status_code_storage &&>(o))
        , _value(static_cast<status_code_storage &&>(o)._value)
    {
      o._domain = nullptr;
    }
    #endif
    #if (CXXOMFORT_CXX_STD>=2011)
    status_code_storage &operator=(const status_code_storage &) = default;
    #else
    #endif
    
    #if (CXXOMFORT_CXX_STD>=2011)
    CXXO_CONSTEXPR14 status_code_storage &operator=(status_code_storage &&o) CXXO_NOEXCEPT
    {
      this->~status_code_storage();
      new(this) status_code_storage(static_cast<status_code_storage &&>(o));
      return *this;
    }
    #else
    #endif
    
    ~status_code_storage () CXXOMFORT_CXX11_CODE(= default, {});
    value_type _value CXXOMFORT_CXX11_CODE({} ,);
    struct _value_type_constructor
    {
    };
    #if (CXXOMFORT_CXX_STD>=2011)
    template <class... Args>
    CXXO_CONSTEXPR status_code_storage(_value_type_constructor /*unused*/, const domain_base *v, Args &&...args)
        : _Base(v), _value(static_cast<Args &&>(args)...)
    {
    }
    #else
    template <typename A1>
    status_code_storage (_value_type_constructor const, domain_base const* p, A1 a1)
    : _Base(p), _value(a1) {}
    template <typename A1, typename A2>
    status_code_storage (_value_type_constructor const, domain_base const* p, A1 a1, A2 a2)
    : _Base(p), _value(a1,a2) {}

    #endif
};

//! Type-erased, move-only status_code.
template <class ErasedT> 
class status_code< erased<ErasedT> >
: public status_code_storage< erased<ErasedT> > {
	template <class> friend class status_code;
	typedef status_code_storage< erased<ErasedT> > _Base;
	public:
	typedef void 	domain_type;
	typedef ErasedT 	value_type;
	public:
	#if (CXXO_COMPILER_SUPPORT_default_delete)
	status_code () = default;
	status_code (status_code const&) = delete; // IMPORTANT!
	status_code & operator= (status_code const&) = delete;
	#else
	status_code () {}
	private: 
	status_code (status_code const&);
	status_code& operator= (status_code const&);
	#endif
};


//! Status code for a given @a Domain .
template <typename Domain>
class status_code
: public status_code_storage<Domain> {
	template <class U> friend class status_code;
	typedef status_code_storage<Domain> _Base;
	public:
	typedef Domain 	domain_type;
	typedef typename Domain::value_type 	value_type;
	
	public:
	CXXO_CONSTEXPR 
	explicit status_code () CXXO_NOEXCEPT 
	: _Base() {}
	
	~status_code () CXXOMFORT_CXX11_CODE(= default, {});
	
	// Explicit value-type construction
	CXXO_CONSTEXPR 
	explicit status_code (value_type x) CXXO_NOEXCEPT 
	: _Base(typename _Base::_value_type_constructor(), &domain_type::get(), x)
	{
		//std::cout<< "[status_code<...>(x="<< x<< ")]"<< std::flush;
	}
	
	// Explicit in-place construction
	#if (CXXOMFORT_CXX_STD>=2011)
	template <typename... Args>
	explicit status_code (std::in_place_t /*unused*/, Args&&... args)
	: _Base(typename _Base::_value_type_constructor() , &domain_type::get(), static_cast<Args&&>(args)...) 
	{ }
	#else
	template <typename A1>
	explicit status_code (std::in_place_t, A1 a1)
	: _Base(typename _Base::_value_type_constructor(), &domain_type::get(), a1)
	{ }
	#endif

	CXXO_CONSTEXPR14 
	status_code 	clone () const { return *this; }
	
	//! Return the current value.
	value_type 	value () const CXXO_NOEXCEPT { 
		_Base const* b= this;
		return b->value();
	}
	//! Reset the current object.
	void clear() CXXO_NOEXCEPT {
		this->_value.~value_type();
		this->_domain = nullptr;
		new(&this->_value) value_type();
	}

	//! Returns the current domain ID.
	uint64_t 	domain_id () const CXXO_NOEXCEPT { return Domain::unique_id; }
	//! Write a status message text to buffer @a p of up to @a l characters.
	char* 	message_to (char*p, size_t l) const CXXO_NOEXCEPT {
		return this->domain()._do_message_to(p,l, *this);
	}
	
};

} }

#endif
