#ifndef CXXOMFORT_UTILS_TRAITS_BASIC_HPP
#define CXXOMFORT_UTILS_TRAITS_BASIC_HPP
/*
 * This is an implementation of the basic type traits, starting with the featureset in TR1 (:2007) and some cxxomfort additions. It is made available so that some simple trait metaprogramming can be used without requiring the whole of <type_traits>.
 * 
 * */

namespace cxxomfort { namespace traits {

template <bool Cond, typename T=void> struct enable_if;
template <typename T> struct enable_if<true,T> { typedef T type; };

template <bool Cond, typename T=void> struct disable_if { typedef T type; };
template <typename T> struct disable_if<false,T> ;

template <bool Cond, typename T1, typename T2> struct conditional  { typedef T2 type; };
template <typename T1, typename T2> struct conditional<true,T1,T2> { typedef T1 type; };

template <typename T> struct drop_cv { typedef T type; };
template <typename T> struct drop_cv<T const> { typedef T type; };
template <typename T> struct drop_cv<T volatile> { typedef T type; };
template <typename T> struct drop_cv<T const volatile> { typedef T type; };


template <size_t N, bool Signed=true> struct integral_of_size;

// for this we can't use the native names (int, long, long long) because their 
// sizeof matches could cause conflicts of template redefinition
template <> struct integral_of_size<sizeof(char), false> { typedef unsigned char type; enum { value = sizeof(char) }; };
template <> struct integral_of_size<sizeof(uint16_t), false> { typedef uint16_t type; enum { value = sizeof(short) }; };
template <> struct integral_of_size<sizeof(uint32_t), false> { typedef uint32_t type; enum { value = sizeof(int) }; };
template <> struct integral_of_size<sizeof(uint64_t), false> { typedef uint64_t type; enum { value = sizeof(long) }; };
template <> struct integral_of_size<sizeof(char), true> { typedef signed char type; enum { value = sizeof(char) }; };
template <> struct integral_of_size<sizeof(int16_t), true> { typedef int16_t type; enum { value = sizeof(short) }; };
template <> struct integral_of_size<sizeof(int32_t), true> { typedef int32_t type; enum { value = sizeof(int) }; };
template <> struct integral_of_size<sizeof(int64_t), true> { typedef int64_t type; enum { value = sizeof(long) }; };


// 
// basic type definitions
// 

template <typename T> struct type_identity { typedef T type; };

template <typename Integral, Integral I> struct integral_constant {
    typedef Integral value_type;
    typedef integral_constant<Integral,I> type;
    static const Integral value= I;
    CXXO_CONSTEXPR operator value_type () const CXXO_NOEXCEPTNOTHROW { return value; }
    CXXO_CONSTEXPR value_type operator() () const CXXO_NOEXCEPTNOTHROW { return value; } // from C++14
};

template <bool B>
struct bool_constant : public integral_constant<bool, B> 
{ };

struct false_type : integral_constant<bool,false> {};
struct true_type : integral_constant<bool,true> {};

#define TR1LITE_CVSPECIALIZE_TRUE(name,type) \
template <> struct name : name##_impl< typename drop_cv<type>::type >  {}
#define TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(name) \
template <typename T> struct name : name##_impl < typename drop_cv<T>::type> {}

template <typename TA, typename TB> struct is_same : public false_type {};
template <typename T> struct is_same<T,T> : public true_type {};

// 
// cv-qualitying type traits
// 

template <typename T> struct is_const : false_type {};
template <typename T> struct is_const<T const>
: true_type {};

template <typename T> struct is_volatile : false_type {};
template <typename T> struct is_volatile<T volatile>
: true_type {};

template <typename T> struct remove_const { typedef T type; };
template <typename T> struct remove_const<T const> {
    typedef T type;
};

template <typename T> struct remove_volatile { typedef T type; };
template <typename T> struct remove_volatile<T volatile> {
    typedef T type;
};

//! remove_cv
template <typename T> struct remove_cv : drop_cv<T> {};

template <typename T> struct add_const { typedef T const type; };
template <typename T> struct add_const<T const> {
    typedef T const type;
};
template <typename T> struct add_volatile { typedef T volatile type; };
template <typename T> struct add_volatile<T volatile> {
    typedef T volatile type;
};
template <typename T> struct add_cv { typedef T const volatile type; };

template <typename L, typename R> struct copy_const { typedef L type; };
template <typename L, typename R> struct copy_const<L, R const> { typedef typename add_const<L>::type type; };
template <typename L, typename R> struct copy_const<L, R volatile> { typedef typename add_volatile<L>::type type; };
template <typename L, typename R> struct copy_const<L, R const volatile> { typedef typename add_cv<L>::type type; };

// 
// pointer traits, reference traits lvalue, rvalue
// 

template <typename T> struct is_reference : false_type {};
template <typename T> struct is_reference<T&> : true_type {};
// is_lvalue_reference
template <typename T> struct is_lvalue_reference : false_type {};
template <typename T> struct is_lvalue_reference<T&> : true_type {};

template <typename T> struct is_pointer : false_type {};
template <typename T> struct is_pointer<T*> : true_type {};

template <typename T> struct remove_reference { typedef T type; };
template <typename T> struct remove_reference<T&> {
    typedef T type;
};

template <typename T> struct remove_pointer { typedef T type; };
template <typename T> struct remove_pointer<T*> {
    typedef T type;
};

template <typename T> struct add_reference { typedef T& type; };
template <typename T> struct add_reference<T&> {
    typedef T& type;
};
template <> struct add_reference<void> { typedef void type; };
template <> struct add_reference<void const> { typedef void const type; };
template <> struct add_reference<void volatile> { typedef void volatile type; };
template <> struct add_reference<void const volatile> { typedef void const volatile type; };

template <typename T> struct add_lvalue_reference : add_reference<T> {};
#if (CXXOMFORT_CXX_STD>=2011)
template <typename T> struct add_lvalue_reference<T&&> { typedef T& type; };
#endif

//template <typename T> struct add_rvalue_reference { /* intentionally empty*/ };

template <typename T> struct add_pointer { typedef T* type; };

// 
// remove_cvref 
// 
template <typename T> struct remove_cvref {
	typedef typename remove_reference<T>::type t1;
	typedef typename remove_cv<t1>::type type;
};

//
// is_{category} primary traits 
// 

template <typename T> struct is_void_impl : false_type {};
template <> struct is_void_impl<void> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_void);

template <typename T> struct is_integral_impl : public false_type {};
template <> struct is_integral_impl<signed short> : true_type {};
template <> struct is_integral_impl<unsigned short> : true_type {};
template <> struct is_integral_impl<signed int> : true_type {};
template <> struct is_integral_impl<unsigned int> : true_type {};
template <> struct is_integral_impl<signed long> : true_type {};
template <> struct is_integral_impl<unsigned long> : true_type {};
template <> struct is_integral_impl<signed long long> : true_type {};
template <> struct is_integral_impl<unsigned long long> : true_type {};
template <> struct is_integral_impl<signed char> : true_type {};
template <> struct is_integral_impl<unsigned char> : true_type {};
template <> struct is_integral_impl<wchar_t> : true_type {};
template <> struct is_integral_impl<bool> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_integral);

template <typename T> struct is_floating_point_impl : false_type {};
template <> struct is_floating_point_impl<float> : true_type {};
template <> struct is_floating_point_impl<double> : true_type {};
template <> struct is_floating_point_impl<long double> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_floating_point);

template <typename T> struct is_char_type_impl : false_type {};
template <> struct is_char_type_impl<char> : true_type {};
template <> struct is_char_type_impl<unsigned char> : true_type {};
template <> struct is_char_type_impl<signed char> : true_type {};
template <> struct is_char_type_impl<wchar_t> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_char_type);

template <typename T> struct is_arithmetic
: public integral_constant<bool, is_integral<T>::value || is_floating_point<T>::value >
{};

template <typename T> struct is_null_pointer_impl : false_type {};
template <> struct is_null_pointer_impl<std::nullptr_t> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_null_pointer);

template <typename T> struct is_fundamental
: public integral_constant<bool,
    is_arithmetic<T>::value || is_void<T>::value || is_null_pointer<T>::value
>
{};

template <typename T> struct is_compound
: public integral_constant<bool, !is_fundamental<T>::value> {};

//
// signedness
//
template <typename T> struct is_signed_impl : false_type {};
template <> struct is_signed_impl<signed short> : true_type {};
template <> struct is_signed_impl<signed int> : true_type {};
template <> struct is_signed_impl<signed long> : true_type {};
template <> struct is_signed_impl<signed long long> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_signed);

template <typename T> struct is_unsigned_impl : false_type {};
template <> struct is_unsigned_impl<unsigned short> : true_type {};
template <> struct is_unsigned_impl<unsigned int> : true_type {};
template <> struct is_unsigned_impl<unsigned long> : true_type {};
template <> struct is_unsigned_impl<unsigned long long> : true_type {};
TR1LITE_CVSPECIALIZE_TRUE_TEMPLATE(is_unsigned);

namespace {
template <typename T> struct make_signed_impl;
template <typename T> struct make_unsigned_impl;

template <> struct make_signed_impl<char> { typedef signed char type; };
template <> struct make_signed_impl<signed char> { typedef signed char type; };
template <> struct make_signed_impl<unsigned char> { typedef signed char type; };
template <> struct make_unsigned_impl<char> { typedef unsigned char type; };
template <> struct make_unsigned_impl<signed char> { typedef unsigned char type; };
template <> struct make_unsigned_impl<unsigned char> { typedef unsigned char type; };

template <> struct make_signed_impl<signed short> { typedef signed short type; };
template <> struct make_signed_impl<unsigned short> { typedef signed short type; };
template <> struct make_unsigned_impl<signed short> { typedef unsigned short type; };
template <> struct make_unsigned_impl<unsigned short> { typedef unsigned short type; };

template <> struct make_signed_impl<signed int> { typedef signed int type; };
template <> struct make_signed_impl<unsigned int> { typedef signed int type; };
template <> struct make_unsigned_impl<signed int> { typedef unsigned int type; };
template <> struct make_unsigned_impl<unsigned int> { typedef unsigned int type; };

template <> struct make_signed_impl<signed long> { typedef signed long type; };
template <> struct make_signed_impl<unsigned long> { typedef signed long type; };
template <> struct make_unsigned_impl<signed long> { typedef unsigned long type; };
template <> struct make_unsigned_impl<unsigned long> { typedef unsigned long type; };

template <> struct make_signed_impl<signed long long> { typedef signed long long type; };
template <> struct make_signed_impl<unsigned long long> { typedef signed long long type; };
template <> struct make_unsigned_impl<signed long long> { typedef unsigned long long type; };
template <> struct make_unsigned_impl<unsigned long long> { typedef unsigned long long type; };

} // local

template <typename T> struct make_signed 
: copy_const< typename make_signed_impl< typename drop_cv<T>::type>::type , T> { };

template <typename T> struct make_unsigned 
: copy_const< typename make_unsigned_impl< typename drop_cv<T>::type>::type , T> { };

// 
// is_array
// 
template <typename T> struct is_array_impl : false_type {};
template <typename T, size_t N> struct is_array_impl <T[N]> : true_type {};
#if (CXXOMFORT_CXX_STD<1997)
#else
template <typename T> struct is_array_impl <T[]> : true_type {};
#endif
template <typename T> struct is_array 
: is_array_impl< typename drop_cv<T>::type > {};

// 
// is_function
// 

template <typename T> struct is_function : false_type {};
#if (defined(__cpp_variadic_templates) )
template <typename R, typename... Args> 
struct is_function<R(Args...)> : true_type {};
#else
template <typename R> struct is_function <R()>
: true_type {};
template <typename R, typename T1>  struct is_function<R(T1)>
: true_type {};
template <typename R, typename T1, typename T2> struct is_function<R(T1,T2)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3> struct is_function<R(T1,T2,T3)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4> struct is_function<R(T1,T2,T3,T4)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5> struct is_function<R(T1,T2,T3,T4,T5)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5> struct is_function<R(T1,T2,T3,T4,T5,...)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6> 
struct is_function<R(T1,T2,T3,T4,T5,T6)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7> 
struct is_function<R(T1,T2,T3,T4,T5,T6,T7)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8> 
struct is_function<R(T1,T2,T3,T4,T5,T6,T7,T8)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9> 
struct is_function<R(T1,T2,T3,T4,T5,T6,T7,T8,T9)>
: true_type {};
template <typename R, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6, typename T7, typename T8, typename T9> 
struct is_function<R(T1,T2,T3,T4,T5,T6,T7,T8,T9, ...)>
: true_type {};
#endif

} }


#endif
