#ifndef CXXOMFORT_UTILS_TYPELIST_HPP
#define CXXOMFORT_UTILS_TYPELIST_HPP
//! @file
//! @brief Implementation of typelists.


// 
// typelist
// 

#if (defined(DOXYGEN_DOC) || CXXOMFORT_CXX_STD>=2011 || (CXXO_COMPILER_SUPPORT_variadic>0))

	#include <type_traits> // integral_constant
	
namespace cxxomfort {

/**
 @brief An implementation of a typelist.
 @cxxoutils An implementation of a typelist.
 
 A typelist is a template class that holds a number of types (in C++>=11, also a type parameter pack). Operations on the typelist are applied to the various types. A typelist is similar to a @c tuple except that the typelist only holds the types, not data.
 * 
 * @code
 * typedef cxxomfort::typelist<float, long, bool> Typelist3;
 * @endcode
 * 
 * Given a @c typelist with some types @a Types... , the following operations are implemented:
 * 
 * * <tt>@ref typelist_size < typelist ></tt>: exposes member value @c value with the size of the typelist.
 * * <tt>@ref typelist_get < I, typelist ></tt>: exposes member type @c type with the @a I -th type in the typelist, if such type exists.
 * * <tt>typelist_find< T, typelist ></tt>: exposes member type @c type which is the first type @a T in @a typelist such that <tt>typelist_get<I, typelist>::type</tt> is @a T .
 * * <tt>@ref typelist_concat <typelist1,typelist2 ></tt>: Create a new @a typelist as the concatenation of two typelists.
 * * <tt>@ref typelist_unshift < Tx, typelist ></tt>: Defines a new @a typelist which is made of { @a Tx , then the types in @a typelist }.
 * * <tt>@ref typelist_shift< typelist ></tt>: Defines a new @a typelist from all types in @a typelist except the first one.
 * * <tt>@ref typelist_push< Tx, typelist ></tt>: Defines a new @a typelist which is made of { the types in @a typelist , then @a Tx }.
 * * <tt>@ref typelist_pop< typelist ></tt>: Defines a new @a typelist from all types in @a typelist except the last one.
 * * <tt>@c typelist_all_of <typelist, P></tt>: Checks if all types @a T in the @a typelist satisfy <tt>P\<T\>::value==true</tt></tt>.
 * * <tt>@ref typelist_wrap <Wrap<>, typelist></tt>: Defines a new @typelist from the types @c Ti in @a typelist , where each new type is @c Wrap<Ti> .
 * * <tt>@ref typelist_derive <typelist></tt>: Defines a new @c type which inherits from all the types in @a typelist , in order.
 * 
 * @note [C++<11] A @c typelist can only have up to 10 types, usually numbered @c T0 to @c T9.
 * 
 */
template <typename... Args>
struct typelist { };

//! @cxxoutils Tag type that declares the end of a non-variadic @c typelist (C++<11).
struct typelist_end {};

//! @cxxoutils Obtain the number of elements in a @c typelist .
template <typename>
struct typelist_size;

template <typename... Types>
struct typelist_size< typelist<Types...> > 
: std::integral_constant<size_t, sizeof...(Types) >
{ };

template <size_t... Elems>
struct size_t_sequence {
    typedef size_t type;
}; 

namespace detail {
template <size_t N, size_t... Next>
struct ISHelper : public ISHelper<N-1U, N-1U, Next...>
{ 
};

template <size_t... Next>
struct ISHelper<0U, Next...> { 
	using type= size_t_sequence<Next...>;
};
} // detail

template <size_t N>
using make_indexes = typename detail::ISHelper<N>::type;

//template <typename> struct typelist_indexes {}; // intentionally empty

// generates the index list index<0, 1, ...> for a typelist
template <typename TypeList>
using typelist_indexes = make_indexes< typelist_size<TypeList>::value > ;


namespace detail {

//! @cxxoutils Get the @a I -th type in a @c typelist<T0,T1,...> .
template <size_t I, typename ...>
struct typelist_get_helper;

template <size_t I>
struct typelist_get_helper<I, typelist< > >;

template <typename T0, typename... Types>
struct typelist_get_helper<0, typelist<T0, Types...> > {
	typedef T0 type;
};

template <size_t I, typename T0, typename... Types>
struct typelist_get_helper<I, typelist<T0, Types...> > 
: typelist_get_helper<I-1, typelist<Types...> > {
};

} // detail

// returns the i-th type in typelist
template <size_t I, typename X>
struct typelist_get;

/** 
 * @cxxoutils If @a I is a valid index into the @a typelist , defines member type @c type as the @a I -th type in @a typelist .
 * 
 * @c typelist_get solves the old problem in C++ Standard of how to get the n-th type in a parameter pack.
 * 
 * @see https://nitter.ir/incomputable/status/1590733659694583808 .
 * @see https://cor3ntin.github.io/posts/kona22/#static_vector-sbo-vectors
 * */
template <size_t I, typename... Types>
struct typelist_get<I, typelist<Types...> > {
	//typedef typename detail::typelist_get_helper<I, 0, typelist_size< typelist<Args...> >::value, typelist<Args...> >::type type;
	typedef typename detail::typelist_get_helper<I, typelist<Types...> >::type type;
};

// ---
// shifts a typelist
//! @cxxoutils Remove the first type of a typelist: given <tt>typelist<A0,A1,A2,...></tt> defines @c type as <tt>typelist<A1,A2,...></tt>.
template <typename> struct typelist_shift;

template <>
struct typelist_shift< typelist< > > {
	// intentionally empty
};

//! @cxxoutils Removes first element: Defines a new typelist @c type as @c typelist<Ti...> for the @a typelist <tt><T0, Ti...></tt>.
template <typename T0, typename... Ti>
struct typelist_shift< typelist<T0,Ti...> > {
	typedef typelist<Ti...> type;
};

// unshifts into a typelist
template <typename Ax, typename Typelist> struct typelist_unshift {};

//! @cxxoutils Adds first element: Defines a new typelist @c type as @c typelist<Ax,Ai...> for the @a Ai elements in @a typelist.
template <typename Tn, typename... TArgs>
struct typelist_unshift< Tn, typelist<TArgs...> > {
	typedef typelist<Tn,TArgs...> type;
};

namespace detail {
template <typename S, typename... Ts> struct pop_impl_;
template <size_t... I, typename... Ts>
struct pop_impl_ < size_t_sequence<I...>, Ts...> {
	//typedef std::index_sequence<I...> IS;
	typedef typelist< typename typelist_get<I, typelist<Ts...> > ::type... > type;
};

}

// ---
// pops from a typelist
template <typename> struct typelist_pop;
//! @cxxoutils Removes last element: Defines new typelist @c type as <tt>typelist<T0, ..., Ti, ..., T_{z-1}</tt> for a @a typelist <tt><Ti..., T_{z-1}, T_{z}></tt>, where @c z is the last index in @i typelist .
template <typename... Ti> 
struct typelist_pop < typelist<Ti...> > {
	typedef make_indexes< sizeof...(Ti)-1 > IS;
	typedef typename detail::pop_impl_<IS , Ti...>::type type;
};


// pushes into a typelist
template <typename, typename...> struct typelist_push;
//! @cxxoutils Adds last element: Defines the new typelist @c type as <tt>typelist<Ti..., Tnew></tt> for a @a typelist and a new type @a Tnew .
template <typename Tnew, typename... Ti> 
struct typelist_push <Tnew, typelist<Ti...> > {
	typedef typelist<Ti...,Tnew> type;
};

// skips the first K elements in a typelist
template <size_t K, typename> struct typelist_skip;
// base case: do nothing
template <typename... Types> struct typelist_skip<0, typelist<Types...> > {
	typedef typelist<Types...> types;
};
template <size_t K, typename T0, typename... Types> struct typelist_skip<K, typelist<T0, Types...> > 
: typelist_skip<K-1, typelist<Types...> > {};


// cuts the typelist at the Kth element
template <size_t K, typename> struct typelist_cut;
// base case: do nothing
template <typename... Types> struct typelist_cut<0, typelist<Types...> > {
	typedef typelist<Types...> types;
};
template <size_t K, typename... Types> struct typelist_cut<K, typelist<Types...> > {
	static_assert( K < sizeof...(Types) , "!");
	typedef make_indexes< K > IS;
	typedef typename detail::pop_impl_<IS, Types...>::type type;
};


// ---

namespace detail {

template <typename T, size_t, size_t, typename...>
struct typelist_find_helper ;

template <typename C, size_t N, size_t S, typename... Types>
struct typelist_find_helper<C,N,S, typelist<Types...> > {
    enum { rec = (0 <= N && N < S) };
    static const size_t X = typelist_find_helper<C,N+1,S,typelist<Types...> >::value ;
    typedef typename std::enable_if<rec, bool>::type still_has_arguments_t;
    static const size_t value = 
    (std::is_same<C, typename typelist_get<N,typelist<Types...> >::type>::value
    ? N : X
    ) ;
};

template <typename C, size_t N, typename... Types>
struct typelist_find_helper<C,N,N, typelist<Types...> > {
    static CXXO_CONSTEXPR const size_t value = -1;
};

} // detail

// check if a type exists in typelist
// returns the 0-based index of the type; -1 if the type is not in Types...
template <typename T, typename>
struct typelist_find;

/**
 * @brief Finds the index of a given type @a T in a @a typelist .
 * 
 * Defines member value @c value as the 0-based index for type @a T in @a typelist if @a T is a valid type in the typelist; or @c typelist_npos (<tt>== -1</tt>) otherwise.
 * */
template <typename T, typename... Ti>
struct typelist_find <T, typelist<Ti...> > 
: std::integral_constant< size_t, detail::typelist_find_helper<T,0, sizeof...(Ti), typelist<Ti...> >::value >
{ };

// concatenate typelists

template <typename...> struct typelist_concat;

template <typename... TA> struct typelist_concat< typelist<TA...> > {
	typedef typelist<TA...> type;
};

/**
 * @brief Concatenate typelists @c typelist<Ai...> , @c typelist<Bi...> , etc.
 * @cxxoutils Concatenate typelists @c typelist<Ai...> , @c typelist<Bi...> , etc.
 * 
 * Given typelists @c typelist<Ai...> as @a Ta , @c typelist<Bi...> as @a Tb , etc., exposes the member type @c type defined as:
 * 
 * @code
 * typedef typelist<Ai..., Bi..., ...> type;
 * @endcode
 * 
 * @note [C++<11] Not available.
 * 
 * */
template <typename... TA, typename... TB> struct typelist_concat< typelist<TA...> , typelist<TB...> > {
	typedef typelist<TA..., TB...> type;
};

template <typename... TA, typename... TB, typename... Typelists >
struct typelist_concat <typelist<TA...> , typelist<TB...> , Typelists...> 
: typelist_concat < typename typelist_concat< typelist<TA...> , typelist<TB...> >::type, Typelists...> { };

// --
// return the number of items Tx in the list that satisfy F<Tx> is true

namespace detail {

template <template<typename...> class F, typename ...>
struct typelist_count_if_helper;

template <template<typename...> class F, typename T0>
struct typelist_count_if_helper<F, typelist< T0> >
: std::integral_constant< size_t, F<T0>::value ? 1 : 0>
{ };

template <template<typename...> class F, typename T0, typename... Types>
struct typelist_count_if_helper<F, typelist< T0, Types... > >
: std::integral_constant< size_t, (F<T0>::value ? 1 : 0) + (typelist_count_if_helper<F,typelist<Types...> >::value) >
{ };

} // detail

template <template<typename...> class F, typename...>
struct typelist_count_if;

template <template<typename...> class F, typename... Types>
struct typelist_count_if< F, typelist<Types...> >
: std::integral_constant<size_t, detail::typelist_count_if_helper<F, typelist<Types...> >::value > 
{ };

// check if all elements Ti in a typelist satisfy If<Ty>

template <typename TL, template<class...> class If> struct typelist_all_if;

template <template<class...> class If, typename... Types> 
struct typelist_all_if< typelist<Types...> , If> 
: public std::integral_constant<bool, (
	typelist_count_if<If, typelist<Types...> >::value == typelist_size<typelist<Types...> >::value
)> {};

// inherits from all types in a typelist

template <typename...> struct typelist_derive;

//! @cxxoutils Defines a new @b type @c type , such that it inherits from all the types in the typelist.
template <typename... Types> struct typelist_derive< typelist<Types...> > {
	struct type: Types... {};
};

// wraps each element in a typelist

template <template<class...> class, typename> struct typelist_wrap;

/**
 * @brief Defines a new @c typelist wrapping each member of a @a typelist with meta-template @a M .
 * @cxxoutils Defines a new @c typelist wrapping each member of a @a typelist with meta-template @a M .
 * 
 * Given a typelist <tt>typelist<A0,A1,...></tt> , and a meta-template @a M that can take a type as <tt>M<T></tt>, this exposes the member type @c type defined as a typelist where @a M is instantiated for every @a Ti in the typelist, as follows:
 * 
 * @code
 * typedef typelist< M<A0>, M<A1>, ...> type;
 * @endcode
 * 
 * @note [C++<11] The meta-template @a M can take only one template argument.
 * 
 * */
template <template<class...> class M, typename... Types> 
struct typelist_wrap <M, typelist<Types...> > {
	typedef typelist< M<Types>... > type;
};

// composes a typelist and a type trait

template <template<class...> class M, typename> struct typelist_compose;

//! @cxxoutils Composes a new typelist @c type such that its members are <tt>M<Ti>::type</tt> for each @a Ti in @a typelist .
//! * @note [C++<11] The meta-template @a M can take only one template argument.
template <template<class...> class M, typename... Ti> 
struct typelist_compose <M, typelist<Ti...> > {
	typedef typelist < typename M<Ti>::type... > type;
};

// reduces the first two types in a typelist using a reductor trait

template <template<class...> class X, typename> struct typelist_reduce;

template <template<class...> class X, typename T0, typename T1, typename... Tail> 
struct typelist_reduce <X, typelist<T0,T1,Tail...> > {
	typedef typelist <typename X<T0,T1>::type, Tail...> type;
};

} // cxxomfort


#else
// 
// pre-C++11 impl
// 
	#define TYPELIST_TYPES typename T0, typename T1, typename T2, typename T3, typename T4, typename T5
	#define TYPELIST_TS T0,T1,T2,T3,T4,T5
	
	#include <cxxomfort/utils/traits_basic.hpp>
	#include <cxxomfort/impl/11-type_traits.hpp> // conditional
	#include <cxxomfort/impl/lift_type_traits.hpp> // std::integral_constant

namespace cxxomfort {

struct typelist_end {};

template <typename T1, typename T2=typelist_end, typename T3=typelist_end, typename T4=typelist_end, typename T5=typelist_end, typename T6=typelist_end>
struct typelist { };

template <typename T>
struct typelist_size;

template <typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_size< typelist<T1,T2,T3,T4,T5,T6> > {
	enum { value = (!traits::is_same<T6,typelist_end>::value ? 6 : !traits::is_same<T5,typelist_end>::value ? 5 : !traits::is_same<T4,typelist_end>::value ? 4 : !traits::is_same<T3,typelist_end>::value ? 3 : !traits::is_same<T2,typelist_end>::value ? 2 : !traits::is_same<T1,typelist_end>::value ? 1 : 0) };
};

namespace detail {

template <size_t I, bool check, typename T>
struct typelist_get;


template <size_t I, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_get<I, false, typelist<T1,T2,T3,T4,T5,T6> > {
	struct _fail_I_less_than_typelist_size {};
};

template <size_t I, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_get<I, true, typelist<T1,T2,T3,T4,T5,T6> > { 
	typedef typelist<T1,T2,T3,T4,T5,T6>  Ty;
	enum { S = typelist_size<Ty>::value };
	typedef typename std::conditional<(I==0), T1
	, typename std::conditional<(I==1), T2
		, typename std::conditional<(I==2), T3
			, typename std::conditional<(I==3), T4
				, typename std::conditional<(I==4), T5
					, typename std::conditional<(I==5), T6, void // void unreachable
						>::type
					>::type
				>::type
			>::type 
		>::type 
	>::type 
	type;
};

} // detail

// returns the i-th type in typelist
template <size_t I, typename>
struct typelist_get;

template <size_t I, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_get<I, typelist<T1,T2,T3,T4,T5,T6> > {
	typedef typelist<T1,T2,T3,T4,T5,T6>  Ty;
	typedef typename detail::typelist_get<I, (I < typelist_size<Ty>::value), Ty >::type type;
};

// check if a type exists in typelist
// returns the 0-based index of the type; -1 if the type is not in Types...

namespace detail {

template <typename, size_t, size_t, typename>
struct typelist_find_helper;

template <typename T, size_t S, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_find_helper<T,S,S, typelist<T1,T2,T3,T4,T5,T6> > {
	static const size_t value = -1;
};

template <typename T, size_t N, size_t S, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_find_helper<T,N,S, typelist<T1,T2,T3,T4,T5,T6> > {
	enum { is = std::is_same<T,T1>::value };
	static const size_t value = is ? N : typelist_find_helper<T,N+1,S, typelist<T2,T3,T4,T5,T5,T6> >::value;
};


} // detail

template <typename T, typename>
struct typelist_find;

template <typename T, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_find <T, typelist<T1,T2,T3,T4,T5,T6> > 
: std::integral_constant< size_t, detail::typelist_find_helper<T,0, typelist_size<typelist< T1,T2,T3,T4,T5 > >::value, typelist<T1,T2,T3,T4,T5,T6> >::value >
{ };

// return the number of items Tx in the list for which F<Tx> is true

namespace detail {

template <template<typename> class F, size_t I, size_t S, typename>
struct typelist_count_if_helper;

template <template<typename> class F, size_t S, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_count_if_helper<F, S, S, typelist<T1,T2,T3,T4,T5,T6> >
: std::integral_constant< size_t, 0 >
{ };

template <template<typename> class F, size_t I, size_t S, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_count_if_helper<F, I, S, typelist<T1,T2,T3,T4,T5,T6> >
: std::integral_constant< size_t, (F<T1>::value ? 1 : 0) 
	+ (typelist_count_if_helper<F,I+1,S,typelist<T2,T3,T4,T5,T6> >::value) >
{ };

} // detail


template <template<typename> class F, typename>
struct typelist_count_if;

template <template<typename> class F, typename T1, typename T2, typename T3, typename T4, typename T5, typename T6>
struct typelist_count_if< F, typelist<T1,T2,T3,T4,T5,T6> >
: std::integral_constant<size_t, detail::typelist_count_if_helper<F, 0, typelist_size< typelist<T1,T2,T3,T4,T5> >::value, typelist<T1,T2,T3,T4,T5,T6> >::value > 
{ };

template <typename TL, template<class> class If, typename=void> struct typelist_all_if;

template <template<class> class If, typename TypeList> 
struct typelist_all_if< TypeList , If> 
: public std::integral_constant<bool, (
	typelist_count_if<If, TypeList >::value == typelist_size<TypeList>::value
)> {};

// wrap

template <template<class> class M, typename> struct typelist_wrap;

template <template<class> class M, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5>
struct typelist_wrap< M, typelist<T0,T1,T2,T3,T4,T5> > {
	typedef typelist< M<T0>, M<T1>, M<T2>, M<T3>, M<T4>, M<T5> > type;
};

template <template<class> class M, typename T0, typename T1, typename T2, typename T3, typename T4>
struct typelist_wrap< M, typelist<T0,T1,T2,T3,T4> > {
	typedef typelist< M<T0>, M<T1>, M<T2>, M<T3>, M<T4>  > type;
};

template <template<class> class M, typename T0, typename T1, typename T2, typename T3>
struct typelist_wrap< M, typelist<T0,T1,T2,T3> > {
	typedef typelist< M<T0>, M<T1>, M<T2>, M<T3> > type;
};

template <template<class> class M, typename T0, typename T1, typename T2>
struct typelist_wrap< M, typelist<T0,T1,T2> > {
	typedef typelist< M<T0>, M<T1>, M<T2> > type;
};

template <template<class> class M, typename T0, typename T1>
struct typelist_wrap< M, typelist<T0,T1> > {
	typedef typelist< M<T0>, M<T1> > type;
};

/*
template <template<class...> class M, typename... Types> 
struct typelist_wrap <M, typelist<Types...> > {
	typedef typelist< M<Types>... > type;
};
*/


// compose

template <template <class> class M, typename> struct typelist_compose ;

template <template <class> class M, typename T0, typename T1, typename T2, typename T3, typename T4, typename T5>
struct typelist_compose<M, typelist<T0,T1,T2,T3,T4,T5> > {
	template <typename T> struct TT { typedef typename std::conditional< std::is_same<T, typelist_end>::value, typelist_end, T>::type type; };
	typedef typename TT<T0>::type K0;
	typedef typename TT<T1>::type K1;
	typedef typename TT<T2>::type K2;
	typedef typename TT<T3>::type K3;
	typedef typename TT<T4>::type K4;
	typedef typename TT<T5>::type K5;
	typedef typelist< typename M<K0>::type, typename M<K1>::type, typename M<K2>::type, typename M<K3>::type, typename M<K4>::type, typename M<K5>::type > type;
};


// derive

template <typename> struct typelist_derive;

template <typename T0, typename T1, typename T2, typename T3, typename T4, typename T5> 
struct typelist_derive< typelist<T0,T1,T2,T3,T4,T5> > {
	struct type: T0, T1, T2, T3, T4, T5 {};
};

template <typename T0, typename T1, typename T2, typename T3, typename T4> 
struct typelist_derive< typelist<T0,T1,T2,T3,T4> > {
	struct type: T0, T1, T2, T3, T4 {};
};

template <typename T0, typename T1, typename T2, typename T3> 
struct typelist_derive< typelist<T0,T1,T2,T3> > {
	struct type: T0, T1, T2, T3 {};
};

template <typename T0, typename T1, typename T2> 
struct typelist_derive< typelist<T0,T1,T2> > {
	struct type: T0, T1, T2 {};
};

template <typename T0, typename T1> 
struct typelist_derive< typelist<T0,T1> > {
	struct type: T0, T1 {};
};

template <typename T0> 
struct typelist_derive< typelist<T0> > {
	struct type: T0 {};
};

// typelist shift, unshift

template <typename T>
struct typelist_shift {};

template < TYPELIST_TYPES >
struct typelist_shift< typelist< TYPELIST_TS > > {
	typedef typelist<T1,T2,T3,T4,T5> type;
};


template <typename T, typename> struct typelist_unshift;

template < TYPELIST_TYPES >
struct typelist_unshift <T0, typelist< T1,T2,T3,T4,T5 > > {
	typedef typelist< TYPELIST_TS> type;
};

template <typename T, typename> struct typelist_push;

template <typename T, typename T0, typename T1, typename T2, typename T3, typename T4> 
struct typelist_push< T, typelist<T0,T1,T2,T3,T4> > {
	typedef typelist<T0,T1,T2,T3,T4, T> type;
};

template <typename T, typename T0, typename T1, typename T2, typename T3> 
struct typelist_push< T, typelist<T0,T1,T2,T3> > {
	typedef typelist<T0,T1,T2,T3, T> type;
};

template <typename T, typename T0, typename T1, typename T2> 
struct typelist_push< T, typelist<T0,T1,T2> > {
	typedef typelist<T0,T1,T2, T> type;
};

template <typename> struct typelist_pop;

template <typename T0, typename T1, typename T2, typename T3, typename T4, typename T5> 
struct typelist_pop< typelist<T0,T1,T2,T3,T4,T5> > {
	typedef typelist<T0,T1,T2,T3,T4> type;
};

template <typename T0, typename T1, typename T2, typename T3, typename T4> 
struct typelist_pop< typelist<T0,T1,T2,T3,T4> > {
	typedef typelist<T0,T1,T2,T3> type;
};

template <typename T0, typename T1, typename T2, typename T3> 
struct typelist_pop< typelist<T0,T1,T2,T3> > {
	typedef typelist<T0,T1,T2> type;
};

template <typename T0, typename T1> 
struct typelist_pop< typelist<T0,T1> > {
	typedef typelist<T0> type;
};


} // cxxomfort 
	#undef TYPELIST_TS
	#undef TYPELIST_TYPES
	

#endif // typelist

#endif
