#ifndef CXXOMFORT_UTILS_TYPESAFEENUM_HPP
#define CXXOMFORT_UTILS_TYPESAFEENUM_HPP
/**
 @file
**/
#include <cxxomfort/base.hpp>
#include <cxxomfort/impl/07-type_traits.hpp>
#include <cxxomfort/impl/using_tr1_type_traits.hpp>
#include <cxxomfort/impl/11-conditional.hpp>
#include <cxxomfort/impl/11-underlying_type.hpp>

/*
 * Usage is as follows:
 */
#if 0 

// example
struct myenum_def { enum type { member1, member2, member3 }; };
// quick and easy, but not always usable in overloads
typedef typesafe_enum<myenum_def, unsigned> myenum;
// more boilerplate, but works with most overloads
struct myenum : public typesafe_enum<myenum_def, uint16_t> myenum {
	typedef typesafe_enum<myenum_def,uint16_t> Base;
	using Base::Base;
};
// quick and easy encapsulation
myenum v = myenum::member3;

#endif

/* Section: declaration of safe_enum */

/*
 * The implementation relies on some type identification measures such as is_enum. If those are not available, 'inner_type' is assumed to be a native type convertible to any enum type.
 * 
 * */

namespace cxxomfort {
namespace detail_enum {
struct safe_enum_tag {};
} // detail_enum

/**
 * @class typesafe_enum
 * @brief Wraps a native C enum into a "enum class"-like type.
 * @cxxoutils Wraps a native C enum into a "enum class"-like type.
 * */
template<typename def, typename inner = typename def::type>
class typesafe_enum : public def /*, public detail_enum::safe_enum_tag */ {
	inner val;
	public: 
	typedef typename def::type enum_type;
	typedef inner inner_type;
#if (CXXOMFORT_CXX_STD>=1997 && CXXO_COMPILER_SUPPORT_type_traits_intrinsic)
    typedef typename std::conditional<
        std::is_enum<inner>::value
        , typename std::underlying_type< /*inner*/ typename def::type >::type , inner
    >::type underlying_type;
#else
	typedef inner underlying_type;
#endif

	public:
	// as per standard sections 3.9.9, 8.5.5, 8.5.7 and 8.5.10 together, see https://stackoverflow.com/questions/17630967/user-defined-c11-enum-class-default-constructor
	// an enum class is def-ctible as if it was value-initialized to 0.

	CXXO_CONSTEXPR typesafe_enum (inner_type e = static_cast<inner_type>(0) ) CXXO_NOEXCEPTNOTHROW 
	: val(e) {}

	//! Return the enumerated value.
	CXXO_CONSTEXPR 
	enum_type value() const CXXO_NOEXCEPTNOTHROW { return static_cast<enum_type>(val); }
	CXXO_CONSTEXPR 
	inner underlying() const CXXO_NOEXCEPTNOTHROW { return val; }
	
	#if (CXXO_COMPILER_SUPPORT_explicit_operator)
	#else
	template <typename Int>
	operator ::cxxomfort::detail::explicit_cast_t<Int> () const CXXO_NOEXCEPTNOTHROW {
		return val;
	}
	#endif
    CXXO_EXPLICIT_OPERATOR(underlying_type) () const CXXO_NOEXCEPTNOTHROW {
        return static_cast<underlying_type>(val);
    }

    friend CXXO_CONSTEXPR bool
    operator== (typesafe_enum const vl, typesafe_enum const vr) CXXO_NOEXCEPTNOTHROW {
        return vl.val == vr.val;
    }
    
    // note that this HAS to be friend because a typesafe_enum can be declared anywhere:
	//friend enum_type 	to_value (typesafe_enum e) CXXO_NOEXCEPTNOTHROW {
	//	return e.value();
	//}


};

//! @cxxoutils Obtain the enumerated value of a @c typesafe_enum .
template< typename def, typename inner> CXXO_CONSTEXPR inline 
typename typesafe_enum<def,inner>::inner_type
to_value (typesafe_enum<def,inner> e) CXXO_NOEXCEPTNOTHROW {
	return e.value();
}

template< typename def, typename inner> CXXO_CONSTEXPR inline 
typename typesafe_enum<def,inner>::underlying_type
to_integer (typesafe_enum<def,inner> e) CXXO_NOEXCEPTNOTHROW {
	return e.underlying();
}

#if (CXXO_COMPILER_SUPPORT_enum_class)
template <typename E>
typename std::underlying_type<E>::type to_value (E e
, typename std::enable_if< std::is_enum<E>::value, bool* >::type = 0
) CXXO_NOEXCEPT { 
	//static_assert(std::is_enum<E>::value, "enum");
	return static_cast <typename std::underlying_type<E>::type >(e); 
}
#endif

template< typename def, typename inner>
inline CXXO_CONSTEXPR 
bool operator!= (typesafe_enum<def,inner> const e1, typesafe_enum<def,inner> const e2) CXXO_NOEXCEPTNOTHROW {
    return !(e1==e2);
}



#if (CXXOMFORT_COMPILER_ID==CXXO_COMPILERID_DMC)
#else
// construct typesafe_enum from the underlying type (useful for "to-byte"-style helpers)
// usage: TSE e = from_underlying<TSE>(u)
template <typename TSE>
TSE from_underlying ( typename TSE::underlying_type c ) {
    return static_cast<typename TSE::enum_type>(c);
}
#endif

// 
// is_typesafe_enum
// 

template <typename> struct is_typesafe_enum : std::false_type {};
template <typename T, typename Inner> struct is_typesafe_enum < typesafe_enum<T,Inner> >: 
std::true_type {};

// 
// specialize enumeration_underlying_type
// 

template <typename TSE>
struct enumeration_underlying_type<TSE , typename std::enable_if< ::cxxomfort::is_typesafe_enum<TSE>::value>::type > { 
	typedef typename TSE::inner_type type; 
};


#if 0
/* An enumerator is scoped if 
 * it is a native enum but it's not convertible 
 * to its underlying type.
 */

template <bool B, typename T>
struct is_scoped_enum_helper;

template <typename T>
struct is_scoped_enum_helper<false, T> {
    // T is not an integral type
    static const bool value = 
    !std::is_convertible<T, typename std::underlying_type<T>::type>::value
    ;
};

template <typename T>
struct is_scoped_enum_helper<true, T> {
    // T is an integral type
    static const bool value = false;
};

template <typename T>
struct is_scoped_enum 
: is_scoped_enum_helper< std::is_integral<T>::value, T > {};

#endif

} // cxxomfort::

namespace cxxomfort {
namespace {
}

template <typename E> 
struct is_enum_typesafe : std::integral_constant <bool, 
	std::is_enum<E>::value || ::cxxomfort::is_typesafe_enum<E>::value 
> {};

template <typename E, typename=void> 
struct enum_or_underlying_type;

template <typename E >
struct enum_or_underlying_type< E, typename std::enable_if< std::is_enum<E>::value , void>::type > { 
	typedef E type; 
};

template <typename D, typename I>
struct enum_or_underlying_type< typesafe_enum<D,I> > { typedef typename D::type type; };


}

#endif // SAFEENUM.HPP
